Type: Package
Package: lubridate
Title: Make Dealing with Dates a Little Easier
Version: 1.9.5
Authors@R: c(
    person("Vitalie", "Spinu", , "spinuvit@gmail.com", role = c("aut", "cre")),
    person("Garrett", "Grolemund", role = "aut"),
    person("Hadley", "Wickham", role = "aut"),
    person("Davis", "Vaughan", role = "ctb"),
    person("Ian", "Lyttle", role = "ctb"),
    person("Imanuel", "Costigan", role = "ctb"),
    person("Jason", "Law", role = "ctb"),
    person("Doug", "Mitarotonda", role = "ctb"),
    person("Joseph", "Larmarange", role = "ctb"),
    person("Jonathan", "Boiser", role = "ctb"),
    person("Chel Hee", "Lee", role = "ctb")
  )
Maintainer: Vitalie Spinu <spinuvit@gmail.com>
Description: Functions to work with date-times and time-spans: fast and
    user friendly parsing of date-time data, extraction and updating of
    components of a date-time (years, months, days, hours, minutes, and
    seconds), algebraic manipulation on date-time and time-span objects.
    The 'lubridate' package has a consistent and memorable syntax that
    makes working with dates easy and fun.
License: MIT + file LICENSE
URL: https://lubridate.tidyverse.org,
        https://github.com/tidyverse/lubridate
BugReports: https://github.com/tidyverse/lubridate/issues
Depends: methods, R (>= 3.2)
Imports: generics, timechange (>= 0.4.0)
Suggests: covr, knitr, rmarkdown, testthat (>= 2.1.0), vctrs (>= 0.6.5)
Enhances: chron, data.table, timeDate, tis, zoo
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
SystemRequirements: A system with zoneinfo data (e.g.
        /usr/share/zoneinfo). On Windows the zoneinfo included with R
        is used.
Collate: 'Dates.r' 'POSIXt.r' 'util.r' 'parse.r' 'timespans.r'
        'intervals.r' 'difftimes.r' 'durations.r' 'periods.r'
        'accessors-date.R' 'accessors-day.r' 'accessors-dst.r'
        'accessors-hour.r' 'accessors-minute.r' 'accessors-month.r'
        'accessors-quarter.r' 'accessors-second.r' 'accessors-tz.r'
        'accessors-week.r' 'accessors-year.r' 'am-pm.r' 'time-zones.r'
        'numeric.r' 'coercion.r' 'constants.r' 'cyclic_encoding.r'
        'data.r' 'decimal-dates.r' 'deprecated.r' 'format_ISO8601.r'
        'guess.r' 'hidden.r' 'instants.r' 'leap-years.r'
        'ops-addition.r' 'ops-compare.r' 'ops-division.r'
        'ops-integer-division.r' 'ops-m+.r' 'ops-modulo.r'
        'ops-multiplication.r' 'ops-subtraction.r' 'package.r'
        'pretty.r' 'round.r' 'stamp.r' 'tzdir.R' 'update.r' 'vctrs.R'
        'zzz.R'
NeedsCompilation: yes
Packaged: 2026-02-03 09:30:08 UTC; vitalie
Author: Vitalie Spinu [aut, cre],
  Garrett Grolemund [aut],
  Hadley Wickham [aut],
  Davis Vaughan [ctb],
  Ian Lyttle [ctb],
  Imanuel Costigan [ctb],
  Jason Law [ctb],
  Doug Mitarotonda [ctb],
  Joseph Larmarange [ctb],
  Jonathan Boiser [ctb],
  Chel Hee Lee [ctb]
Repository: CRAN
Date/Publication: 2026-02-04 09:00:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-27 05:40:21 UTC; unix
Archs: lubridate.so.dSYM

Type: Package
Package: lubridate
Title: Make Dealing with Dates a Little Easier
Version: 1.9.5
Authors@R: c(
    person("Vitalie", "Spinu", , "spinuvit@gmail.com", role = c("aut", "cre")),
    person("Garrett", "Grolemund", role = "aut"),
    person("Hadley", "Wickham", role = "aut"),
    person("Davis", "Vaughan", role = "ctb"),
    person("Ian", "Lyttle", role = "ctb"),
    person("Imanuel", "Costigan", role = "ctb"),
    person("Jason", "Law", role = "ctb"),
    person("Doug", "Mitarotonda", role = "ctb"),
    person("Joseph", "Larmarange", role = "ctb"),
    person("Jonathan", "Boiser", role = "ctb"),
    person("Chel Hee", "Lee", role = "ctb")
  )
Maintainer: Vitalie Spinu <spinuvit@gmail.com>
Description: Functions to work with date-times and time-spans: fast and
    user friendly parsing of date-time data, extraction and updating of
    components of a date-time (years, months, days, hours, minutes, and
    seconds), algebraic manipulation on date-time and time-span objects.
    The 'lubridate' package has a consistent and memorable syntax that
    makes working with dates easy and fun.
License: MIT + file LICENSE
URL: https://lubridate.tidyverse.org,
        https://github.com/tidyverse/lubridate
BugReports: https://github.com/tidyverse/lubridate/issues
Depends: methods, R (>= 3.2)
Imports: generics, timechange (>= 0.4.0)
Suggests: covr, knitr, rmarkdown, testthat (>= 2.1.0), vctrs (>= 0.6.5)
Enhances: chron, data.table, timeDate, tis, zoo
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
SystemRequirements: A system with zoneinfo data (e.g.
        /usr/share/zoneinfo). On Windows the zoneinfo included with R
        is used.
Collate: 'Dates.r' 'POSIXt.r' 'util.r' 'parse.r' 'timespans.r'
        'intervals.r' 'difftimes.r' 'durations.r' 'periods.r'
        'accessors-date.R' 'accessors-day.r' 'accessors-dst.r'
        'accessors-hour.r' 'accessors-minute.r' 'accessors-month.r'
        'accessors-quarter.r' 'accessors-second.r' 'accessors-tz.r'
        'accessors-week.r' 'accessors-year.r' 'am-pm.r' 'time-zones.r'
        'numeric.r' 'coercion.r' 'constants.r' 'cyclic_encoding.r'
        'data.r' 'decimal-dates.r' 'deprecated.r' 'format_ISO8601.r'
        'guess.r' 'hidden.r' 'instants.r' 'leap-years.r'
        'ops-addition.r' 'ops-compare.r' 'ops-division.r'
        'ops-integer-division.r' 'ops-m+.r' 'ops-modulo.r'
        'ops-multiplication.r' 'ops-subtraction.r' 'package.r'
        'pretty.r' 'round.r' 'stamp.r' 'tzdir.R' 'update.r' 'vctrs.R'
        'zzz.R'
NeedsCompilation: yes
Packaged: 2026-02-03 09:30:08 UTC; vitalie
Author: Vitalie Spinu [aut, cre],
  Garrett Grolemund [aut],
  Hadley Wickham [aut],
  Davis Vaughan [ctb],
  Ian Lyttle [ctb],
  Imanuel Costigan [ctb],
  Jason Law [ctb],
  Doug Mitarotonda [ctb],
  Joseph Larmarange [ctb],
  Jonathan Boiser [ctb],
  Chel Hee Lee [ctb]
Repository: CRAN
Date/Publication: 2026-02-04 09:00:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 23:16:04 UTC; unix
Archs: lubridate.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(as.POSIXct,fts)
S3method(as.POSIXct,irts)
S3method(as.POSIXct,timeSeries)
S3method(as.POSIXct,xts)
S3method(as.POSIXct,zoo)
S3method(as.POSIXlt,fts)
S3method(as.POSIXlt,irts)
S3method(as.POSIXlt,timeSeries)
S3method(as.POSIXlt,tis)
S3method(as.POSIXlt,xts)
S3method(as.POSIXlt,zoo)
S3method(as.list,Interval)
S3method(c,Duration)
S3method(c,Interval)
S3method(c,Period)
S3method(date,Period)
S3method(date,default)
S3method(decimal_date,default)
S3method(decimal_date,zoo)
S3method(dst,default)
S3method(force_tz,default)
S3method(format,Duration)
S3method(format,Interval)
S3method(format,Period)
S3method(hour,Period)
S3method(hour,default)
S3method(intersect,Interval)
S3method(mday,Period)
S3method(mday,default)
S3method(minute,Period)
S3method(minute,default)
S3method(month,Period)
S3method(month,default)
S3method(month,numeric)
S3method(months,numeric)
S3method(qday,default)
S3method(reclass_date,Date)
S3method(reclass_date,IDate)
S3method(reclass_date,ITime)
S3method(reclass_date,POSIXct)
S3method(reclass_date,POSIXlt)
S3method(reclass_date,chron)
S3method(reclass_date,ti)
S3method(reclass_date,timeDate)
S3method(second,Period)
S3method(second,default)
S3method(setdiff,Interval)
S3method(summary,Duration)
S3method(summary,Interval)
S3method(summary,Period)
S3method(tz,Date)
S3method(tz,POSIXt)
S3method(tz,character)
S3method(tz,default)
S3method(tz,logical)
S3method(tz,timeSeries)
S3method(tz,yearmon)
S3method(tz,yearqtr)
S3method(tz,zoo)
S3method(tz,zooreg)
S3method(union,Interval)
S3method(unique,Interval)
S3method(update,Date)
S3method(update,POSIXt)
S3method(wday,default)
S3method(wday,numeric)
S3method(with_tz,default)
S3method(xtfrm,Period)
S3method(yday,default)
S3method(year,Period)
S3method(year,default)
export("%--%")
export("%m+%")
export("%m-%")
export("%within%")
export("date<-")
export("day<-")
export("hour<-")
export("int_end<-")
export("int_start<-")
export("mday<-")
export("minute<-")
export("month<-")
export("qday<-")
export("second<-")
export("tz<-")
export("wday<-")
export("week<-")
export("yday<-")
export("year<-")
export(Date)
export(NA_Date_)
export(NA_POSIXct_)
export(POSIXct)
export(add_with_rollback)
export(am)
export(as.difftime)
export(as.duration)
export(as.interval)
export(as.period)
export(as_date)
export(as_datetime)
export(ceiling_date)
export(cyclic_encoding)
export(date)
export(date_decimal)
export(day)
export(days)
export(days_in_month)
export(ddays)
export(decimal_date)
export(dhours)
export(dmicroseconds)
export(dmilliseconds)
export(dminutes)
export(dmonths)
export(dmy)
export(dmy_h)
export(dmy_hm)
export(dmy_hms)
export(dnanoseconds)
export(dpicoseconds)
export(dseconds)
export(dst)
export(duration)
export(dweeks)
export(dyears)
export(dym)
export(epiweek)
export(epiyear)
export(fast_strptime)
export(fit_to_timeline)
export(floor_date)
export(force_tz)
export(force_tzs)
export(format_ISO8601)
export(guess_formats)
export(hm)
export(hms)
export(hour)
export(hours)
export(int_aligns)
export(int_diff)
export(int_end)
export(int_flip)
export(int_length)
export(int_overlaps)
export(int_shift)
export(int_standardize)
export(int_start)
export(intersect)
export(interval)
export(is.Date)
export(is.POSIXct)
export(is.POSIXlt)
export(is.POSIXt)
export(is.difftime)
export(is.duration)
export(is.instant)
export(is.interval)
export(is.period)
export(is.timepoint)
export(is.timespan)
export(isoweek)
export(isoyear)
export(leap_year)
export(local_time)
export(make_date)
export(make_datetime)
export(make_difftime)
export(mday)
export(mdy)
export(mdy_h)
export(mdy_hm)
export(mdy_hms)
export(microseconds)
export(milliseconds)
export(minute)
export(minutes)
export(month)
export(ms)
export(my)
export(myd)
export(nanoseconds)
export(now)
export(origin)
export(parse_date_time)
export(parse_date_time2)
export(period)
export(period_to_seconds)
export(picoseconds)
export(pm)
export(pretty_dates)
export(qday)
export(quarter)
export(reclass_date)
export(reclass_timespan)
export(rollback)
export(rollbackward)
export(rollforward)
export(round_date)
export(second)
export(seconds)
export(seconds_to_period)
export(semester)
export(setdiff)
export(stamp)
export(stamp_date)
export(stamp_time)
export(time_length)
export(today)
export(tz)
export(union)
export(wday)
export(week)
export(weeks)
export(with_tz)
export(yday)
export(ydm)
export(ydm_h)
export(ydm_hm)
export(ydm_hms)
export(year)
export(years)
export(ym)
export(ymd)
export(ymd_h)
export(ymd_hm)
export(ymd_hms)
export(yq)
exportClasses(Duration)
exportClasses(Interval)
exportClasses(Period)
exportClasses(Timespan)
exportMethods("$")
exportMethods("$<-")
exportMethods("%%")
exportMethods("%/%")
exportMethods("%m+%")
exportMethods("%m-%")
exportMethods("*")
exportMethods("+")
exportMethods("-")
exportMethods("/")
exportMethods("[")
exportMethods("[<-")
exportMethods("[[")
exportMethods("[[<-")
exportMethods("date<-")
exportMethods("day<-")
exportMethods("hour<-")
exportMethods("minute<-")
exportMethods("month<-")
exportMethods("second<-")
exportMethods("year<-")
exportMethods(Arith)
exportMethods(Compare)
exportMethods(as.character)
exportMethods(as.difftime)
exportMethods(as.integer)
exportMethods(as.interval)
exportMethods(as.numeric)
exportMethods(as_date)
exportMethods(as_datetime)
exportMethods(format_ISO8601)
exportMethods(reclass_timespan)
exportMethods(rep)
exportMethods(show)
importFrom(generics,as.difftime)
importFrom(generics,intersect)
importFrom(generics,setdiff)
importFrom(generics,union)
importFrom(methods,"coerce<-")
importFrom(methods,"slot<-")
importFrom(methods,Arith)
importFrom(methods,Compare)
importFrom(methods,allNames)
importFrom(methods,callGeneric)
importFrom(methods,initialize)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,setClass)
importFrom(methods,setGeneric)
importFrom(methods,show)
importFrom(methods,slot)
importFrom(methods,slotNames)
importFrom(methods,validObject)
importFrom(stats,na.omit)
importFrom(stats,setNames)
importFrom(stats,update)
importFrom(timechange,time_add)
importFrom(timechange,time_force_tz)
importFrom(timechange,time_get)
importFrom(timechange,time_update)
importFrom(utils,packageVersion)
importFrom(utils,read.delim)
useDynLib(lubridate, .registration=TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# lubridate <img src="man/figures/logo.png" align="right" />

<!-- badges: start -->

[![CRAN
version](http://www.r-pkg.org/badges/version/lubridate)](https://cran.r-project.org/package=lubridate)
[![R build
status](https://github.com/tidyverse/lubridate/workflows/R-CMD-check/badge.svg)](https://github.com/tidyverse/lubridate/actions)
[![CRAN RStudio mirror
downloads](http://cranlogs.r-pkg.org/badges/lubridate)](https://cran.r-project.org/package=lubridate)
[![R-CMD-check](https://github.com/tidyverse/lubridate/actions/workflows/check-standard.yaml/badge.svg)](https://github.com/tidyverse/lubridate/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

## Overview

Date-time data can be frustrating to work with in R. R commands for
date-times are generally unintuitive and change depending on the type of
date-time object being used. Moreover, the methods we use with
date-times must be robust to time zones, leap days, daylight savings
times, and other time related quirks, and R lacks these capabilities in
some situations. Lubridate makes it easier to do the things R does with
date-times and possible to do the things R does not.

If you are new to lubridate, the best place to start is the [date and
times chapter](https://r4ds.hadley.nz/datetimes.html) in R for data
science.

## Installation

``` r
# The easiest way to get lubridate is to install the whole tidyverse:
install.packages("tidyverse")

# Alternatively, install just lubridate:
install.packages("lubridate")

# Or the development version from GitHub:
# install.packages("devtools")
devtools::install_github("tidyverse/lubridate")
```

## Cheatsheet

<a href="https://github.com/rstudio/cheatsheets/blob/main/lubridate.pdf"><img src="https://raw.githubusercontent.com/rstudio/cheatsheets/main/pngs/thumbnails/lubridate-cheatsheet-thumbs.png" width="630" height="252"/></a>

## Features

``` r
library(lubridate, warn.conflicts = FALSE)
```

- Easy and fast parsing of date-times: `ymd()`, `ymd_hms`, `dmy()`,
  `dmy_hms`, `mdy()`, …

  ``` r
  ymd(20101215)
  #> [1] "2010-12-15"
  mdy("4/1/17")
  #> [1] "2017-04-01"
  ```

- Simple functions to get and set components of a date-time, such as
  `year()`, `month()`, `mday()`, `hour()`, `minute()` and `second()`:

  ``` r
  bday <- dmy("14/10/1979")
  month(bday)
  #> [1] 10
  wday(bday, label = TRUE)
  #> [1] Sun
  #> Levels: Sun < Mon < Tue < Wed < Thu < Fri < Sat

  year(bday) <- 2016
  wday(bday, label = TRUE)
  #> [1] Fri
  #> Levels: Sun < Mon < Tue < Wed < Thu < Fri < Sat
  ```

- Helper functions for handling time zones: `with_tz()`, `force_tz()`

  ``` r
  time <- ymd_hms("2010-12-13 15:30:30")
  time
  #> [1] "2010-12-13 15:30:30 UTC"

  # Changes printing
  with_tz(time, "America/Chicago")
  #> [1] "2010-12-13 09:30:30 CST"

  # Changes time
  force_tz(time, "America/Chicago")
  #> [1] "2010-12-13 15:30:30 CST"
  ```

Lubridate also expands the type of mathematical operations that can be
performed with date-time objects. It introduces three new time span
classes borrowed from <https://www.joda.org>.

- `durations`, which measure the exact amount of time between two points

- `periods`, which accurately track clock times despite leap years, leap
  seconds, and day light savings time

- `intervals`, a protean summary of the time information between two
  points

## Code of Conduct

Please note that the lubridate project is released with a [Contributor
Code of Conduct](https://lubridate.tidyverse.org/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
#' @include Dates.r
#' @include difftimes.r
#' @include numeric.r
#' @include POSIXt.r
NULL


add_duration_to_duration <- function(dur2, dur1)
  new("Duration", dur1@.Data + dur2@.Data)

add_duration_to_date <- function(dur, date) {
  if (is.Date(date)) {
    date <- as.POSIXct(date)
    ans <- with_tz(date + dur@.Data, "UTC")

    if (all(is.na(ans))) return(as.Date(ans))  # ALL NAs

    if (all(hour(na.omit(ans)) == 0  &
            minute(na.omit(ans)) == 0 &
            second(na.omit(ans)) == 0)) {
      return(as.Date(ans))
    }

    return(ans)
  }
  new <- date + dur@.Data
  attr(new, "tzone") <- tz(date)
  reclass_date(new, date)
}

add_period_to_period <- function(per2, per1) {
  new("Period", per1@.Data + per2@.Data,
      year = per1@year + per2@year,
      month = per1@month + per2@month,
      day = per1@day + per2@day,
      hour = per1@hour + per2@hour,
      minute = per1@minute + per2@minute)
}

add_period_to_date <- function(per, date) {

  lt <- as.POSIXlt(date)

  ## add months and years with no backwards rollover
  ms <- month(per) + year(per) * 12
  lt <- add_months(lt, ms)

  if (is.Date(date)) {
    new <- update(as.Date(lt),
                  days = mday(lt) + per@day,
                  hours = per@hour,
                  minutes = per@minute,
                  seconds = per@.Data)
    return(new)
  }

  new <- update(lt,
                days = mday(lt) + per@day,
                hours = hour(lt) + per@hour,
                minutes = minute(lt) + per@minute,
                seconds = second(lt) + per@.Data)

  reclass_date(new, date)
}

add_months <- function(mt, mos) {
  nnas <- !is.na(mos)
  if (all(mos[nnas] == 0L)) {
    return(mt)
  }
  mt$mon <- mt$mon + mos
  ndays <- as.numeric(format.POSIXlt(mt, "%d", usetz = FALSE))
  mt$mon[mt$mday != ndays] <- NA
  mt
}

add_number_to_duration <- function(num, dur) {
  new("Duration", dur@.Data + num)
}

add_number_to_period <- function(num, per) {
  slot(per, ".Data") <- per@.Data + num
  per
}

#' @export
setMethod("+", signature(e1 = "Duration", e2 = "Duration"),
          function(e1, e2) add_duration_to_duration(e2, e1))

#' @export
setMethod("+", signature(e1 = "Duration", e2 = "Date"),
          function(e1, e2) add_duration_to_date(e1, e2))

#' @export
setMethod("+", signature(e1 = "Duration", e2 = "difftime"),
          function(e1, e2) add_duration_to_duration(as.duration(e2), e1))

#' @export
setMethod("+", signature(e1 = "Duration", e2 = "numeric"),
          function(e1, e2) add_number_to_duration(e2, e1))

#' @export
setMethod("+", signature(e1 = "Duration", e2 = "POSIXct"),
          function(e1, e2) add_duration_to_date(e1, e2))

#' @export
setMethod("+", signature(e1 = "Duration", e2 = "POSIXlt"),
          function(e1, e2) add_duration_to_date(e1, e2))

#' @export
setMethod("+", signature(e1 = "Period", e2 = "Period"),
          function(e1, e2) add_period_to_period(e2, e1))

#' @export
setMethod("+", signature(e1 = "Period", e2 = "Date"),
          function(e1, e2) add_period_to_date(e1, e2))

#' @export
setMethod("+", signature(e1 = "Period", e2 = "numeric"),
          function(e1, e2) add_number_to_period(e2, e1))

#' @export
setMethod("+", signature(e1 = "Period", e2 = "POSIXct"),
          function(e1, e2) add_period_to_date(e1, e2))

#' @export
setMethod("+", signature(e1 = "Period", e2 = "POSIXlt"),
          function(e1, e2) add_period_to_date(e1, e2))

#' @export
setMethod("+", signature(e1 = "Date", e2 = "Duration"),
          function(e1, e2) add_duration_to_date(e2, e1))

#' @export
setMethod("+", signature(e1 = "Date", e2 = "Period"),
          function(e1, e2) add_period_to_date(e2, e1))

#' @export
setMethod("+", signature(e1 = "difftime", e2 = "Duration"),
          function(e1, e2) as.difftime(e2, units = "secs") + e1)

#' @export
setMethod("+", signature(e1 = "numeric", e2 = "Duration"),
          function(e1, e2) add_number_to_duration(e1, e2))

#' @export
setMethod("+", signature(e1 = "numeric", e2 = "Period"),
          function(e1, e2) add_number_to_period(e1, e2))

#' @export
setMethod("+", signature(e1 = "POSIXct", e2 = "Duration"),
          function(e1, e2) add_duration_to_date(e2, e1))

#' @export
setMethod("+", signature(e1 = "POSIXct", e2 = "Period"),
          function(e1, e2) add_period_to_date(e2, e1))

#' @export
setMethod("+", signature(e1 = "POSIXlt", e2 = "Duration"),
          function(e1, e2) add_duration_to_date(e2, e1))

#' @export
setMethod("+", signature(e1 = "POSIXlt", e2 = "Period"),
          function(e1, e2) add_period_to_date(e2, e1))

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
#' @include Dates.r
#' @include difftimes.r
#' @include numeric.r
#' @include POSIXt.r
NULL


add_duration_to_duration <- function(dur2, dur1) {
  new("Duration", dur1@.Data + dur2@.Data)
}

add_duration_to_date <- function(dur, date) {
  if (is.Date(date)) {
    date <- as.POSIXct(date)
    ans <- with_tz(date + dur@.Data, "UTC")

    if (all(is.na(ans))) {
      return(as.Date(ans))
    } # ALL NAs

    if (all(hour(na.omit(ans)) == 0 &
      minute(na.omit(ans)) == 0 &
      second(na.omit(ans)) == 0)) {
      return(as.Date(ans))
    }

    return(ans)
  }
  new <- date + dur@.Data
  attr(new, "tzone") <- tz(date)
  reclass_date(new, date)
}

add_period_to_period <- function(per2, per1) {
  new("Period", per1@.Data + per2@.Data,
    year = per1@year + per2@year,
    month = per1@month + per2@month,
    day = per1@day + per2@day,
    hour = per1@hour + per2@hour,
    minute = per1@minute + per2@minute
  )
}

add_period_to_date <- function(per, date) {
  out <-
    timechange::time_add(date,
      years = per@year,
      months = per@month,
      days = per@day,
      hours = per@hour,
      minutes = per@minute,
      seconds = per@.Data,
      roll_month = "NA",
      roll_dst = c("NA", "pre"))
  if (is.Date(date) && is_zero_hms(per@hour, per@minute, per@.Data))
    out <- as_date(out)
  out
}

add_number_to_duration <- function(num, dur) {
  new("Duration", dur@.Data + num)
}

add_number_to_period <- function(num, per) {
  slot(per, ".Data") <- per@.Data + num
  per
}

#' @export
setMethod("+", signature(e1 = "Duration", e2 = "Duration"),
  function(e1, e2) add_duration_to_duration(e2, e1)
)

#' @export
setMethod("+", signature(e1 = "Duration", e2 = "Date"),
  function(e1, e2) add_duration_to_date(e1, e2)
)

#' @export
setMethod("+", signature(e1 = "Duration", e2 = "difftime"),
  function(e1, e2) add_duration_to_duration(as.duration(e2), e1)
)

#' @export
setMethod("+", signature(e1 = "Duration", e2 = "numeric"),
  function(e1, e2) add_number_to_duration(e2, e1)
)

#' @export
setMethod("+", signature(e1 = "Duration", e2 = "Interval"),
  function(e1, e2) add_number_to_duration(e2, e1)
)

#' @export
setMethod("+", signature(e1 = "Duration", e2 = "POSIXct"),
  function(e1, e2) add_duration_to_date(e1, e2)
)

#' @export
setMethod("+", signature(e1 = "Duration", e2 = "POSIXlt"),
  function(e1, e2) add_duration_to_date(e1, e2)
)

#' @export
setMethod("+", signature(e1 = "Period", e2 = "Period"),
  function(e1, e2) add_period_to_period(e2, e1)
)

#' @export
setMethod("+", signature(e1 = "Period", e2 = "Date"),
  function(e1, e2) add_period_to_date(e1, e2)
)

#' @export
setMethod("+", signature(e1 = "Period", e2 = "numeric"),
  function(e1, e2) add_number_to_period(e2, e1)
)

#' @export
setMethod("+", signature(e1 = "Period", e2 = "Interval"),
  function(e1, e2) add_number_to_period(e2, e1)
)

#' @export
setMethod("+", signature(e1 = "Period", e2 = "POSIXct"),
  function(e1, e2) add_period_to_date(e1, e2)
)

#' @export
setMethod("+", signature(e1 = "Period", e2 = "POSIXlt"),
  function(e1, e2) add_period_to_date(e1, e2)
)

#' @export
setMethod("+", signature(e1 = "Date", e2 = "Duration"),
  function(e1, e2) add_duration_to_date(e2, e1)
)

#' @export
setMethod("+", signature(e1 = "Date", e2 = "Period"),
  function(e1, e2) add_period_to_date(e2, e1)
)

#' @export
setMethod("+", signature(e1 = "difftime", e2 = "Duration"),
  function(e1, e2) as.difftime(e2, units = "secs") + e1
)

#' @export
setMethod("+", signature(e1 = "numeric", e2 = "Duration"),
  function(e1, e2) add_number_to_duration(e1, e2)
)

#' @export
setMethod("+", signature(e1 = "numeric", e2 = "Period"),
  function(e1, e2) add_number_to_period(e1, e2)
)

#' @export
setMethod("+", signature(e1 = "Interval", e2 = "Duration"),
  function(e1, e2) add_number_to_duration(e1, e2)
)

#' @export
setMethod("+", signature(e1 = "Interval", e2 = "Period"),
  function(e1, e2) add_number_to_period(e1, e2)
)

#' @export
setMethod("+", signature(e1 = "Interval", e2 = "Interval"),
  function(e1, e2) add_number_to_period(e1, e2)
)

#' @export
setMethod("+", signature(e1 = "POSIXct", e2 = "Duration"),
  function(e1, e2) add_duration_to_date(e2, e1)
)

#' @export
setMethod("+", signature(e1 = "POSIXct", e2 = "Period"),
  function(e1, e2) add_period_to_date(e2, e1)
)

#' @export
setMethod("+", signature(e1 = "POSIXlt", e2 = "Duration"),
  function(e1, e2) add_duration_to_date(e2, e1)
)

#' @export
setMethod("+", signature(e1 = "POSIXlt", e2 = "Period"),
  function(e1, e2) add_period_to_date(e2, e1)
)

# ------------------------------------------------------------------------------
# Constructors

# S4 constructors are slow, so empty Period/Duration objects are saved for
# repeated use in `vec_ptype2()`. Interval types depend on the `tzone`, so
# the constructor must be used.

delayedAssign("lubridate_shared_empty_period", period())
delayedAssign("lubridate_shared_empty_duration", duration())

new_empty_interval <- function(tzone) {
  interval(tzone = tzone)
}

# ------------------------------------------------------------------------------
# Period - proxy / restore

# Method registered in `.onLoad()`
vec_proxy.Period <- function(x, ...) {
  out <- vec_proxy_period(x)

  # Track vector names as an extra column since data frame
  # row names must be unique
  names <- names(x)
  if (!is.null(names)) {
    out[["rcrd_names"]] <- names
  }

  out
}

# Method registered in `.onLoad()`
vec_proxy_compare.Period <- function(x, ...) {
  vec_proxy_period(x)
}

# Method registered in `.onLoad()`
vec_proxy_equal.Period <- function(x, ...) {
  vec_proxy_period(x)
}

# Method registered in `.onLoad()`
vec_restore.Period <- function(x, to, ...) {
  out <- period(
    year = x$year,
    month = x$month,
    day = x$day,
    hour = x$hour,
    minute = x$minute,
    second = x$second
  )

  names <- x$rcrd_names
  if (!is.null(names)) {
    names(out) <- names
  }

  out
}

vec_proxy_period <- function(x) {
  cols <- list(
    year = x@year,
    month = x@month,
    day = x@day,
    hour = x@hour,
    minute = x@minute,
    second = x@.Data
  )

  n <- length(x@year)

  vctrs::new_data_frame(cols, n = n)
}

# ------------------------------------------------------------------------------
# Period - ptype2

# Method registered in `.onLoad()`
vec_ptype2.Period.Period <- function(x, y, ...) {
  lubridate_shared_empty_period
}

# ------------------------------------------------------------------------------
# Period - cast

# Method registered in `.onLoad()`
vec_cast.Period.Period <- function(x, to, ...) {
  x
}

# ------------------------------------------------------------------------------
# Duration - proxy / restore

# Method registered in `.onLoad()`
vec_proxy.Duration <- function(x, ...) {
  out <- vec_proxy_duration(x)
  names(out) <- names(x)
  out
}

# Method registered in `.onLoad()`
vec_proxy_compare.Duration <- function(x, ...) {
  vec_proxy_duration(x)
}

# Method registered in `.onLoad()`
vec_proxy_equal.Duration <- function(x, ...) {
  vec_proxy_duration(x)
}

# Method registered in `.onLoad()`
vec_restore.Duration <- function(x, to, ...) {
  out <- duration(x, units = "seconds")
  names(out) <- names(x)
  out
}

vec_proxy_duration <- function(x) {
  x@.Data
}

# ------------------------------------------------------------------------------
# Duration - ptype2

# Method registered in `.onLoad()`
vec_ptype2.Duration.Duration <- function(x, y, ...) {
  lubridate_shared_empty_duration
}

# Method registered in `.onLoad()`
vec_ptype2.Duration.difftime <- function(x, y, ...) {
  lubridate_shared_empty_duration
}

# Method registered in `.onLoad()`
vec_ptype2.difftime.Duration <- function(x, y, ...) {
  lubridate_shared_empty_duration
}

# ------------------------------------------------------------------------------
# Duration - cast

# Method registered in `.onLoad()`
vec_cast.Duration.Duration <- function(x, to, ...) {
  x
}

# Method registered in `.onLoad()`
vec_cast.Duration.difftime <- function(x, to, ...) {
  as_duration_from_difftime(x)
}

# Method registered in `.onLoad()`
vec_cast.difftime.Duration <- function(x, to, ...) {
  as_difftime_from_duration(x)
}

# ------------------------------------------------------------------------------
# Interval - proxy / restore

# Method registered in `.onLoad()`
vec_proxy.Interval <- function(x, ...) {
  out <- vec_proxy_interval(x, vctrs::vec_proxy)

  # Track vector names as an extra column since data frame
  # row names must be unique
  names <- names(x)
  if (!is.null(names)) {
    out[["rcrd_names"]] <- names
  }

  out
}

# Method registered in `.onLoad()`
vec_proxy_compare.Interval <- function(x, ...) {
  vec_proxy_interval(x, vctrs::vec_proxy_compare)
}

# Method registered in `.onLoad()`
vec_proxy_equal.Interval <- function(x, ...) {
  vec_proxy_interval(x, vctrs::vec_proxy_equal)
}

# Method registered in `.onLoad()`
vec_restore.Interval <- function(x, to, ...) {
  tzone <- to@tzone

  span <- x$span

  start <- x$start
  start <- vctrs::new_datetime(start, tzone = tzone)

  out <- new("Interval", span, start = start, tzone = tzone)

  names <- x$rcrd_names
  if (!is.null(names)) {
    names(out) <- names
  }

  out
}

vec_proxy_interval <- function(x, proxy_fn) {
  start <- x@start
  span <- x@.Data

  # Proxy the underlying POSIXct as well
  start <- proxy_fn(start)

  # Ordered in such a way that the start date controls interval ordering.
  # Ties in the start date are resolved by placing shorter intervals first.
  cols <- list(
    start = start,
    span = span
  )

  n <- length(start)

  vctrs::new_data_frame(cols, n = n)
}

# ------------------------------------------------------------------------------
# Interval - ptype2

# Method registered in `.onLoad()`
vec_ptype2.Interval.Interval <- function(x, y, ...) {
  new_empty_interval(tzone = tz_union(x, y))
}

# ------------------------------------------------------------------------------
# Interval - cast

# Method registered in `.onLoad()`
vec_cast.Interval.Interval <- function(x, to, ...) {
  x_tzone <- x@tzone
  to_tzone <- to@tzone

  if (identical(x_tzone, to_tzone)) {
    return(x)
  }

  x_span <- x@.Data
  x_start <- x@start

  start <- with_tz(x_start, to_tzone)

  new("Interval", x_span, start = start, tzone = to_tzone)
}

# ------------------------------------------------------------------------------

tz_is_local <- function(x) {
  identical(x, "")
}

tz_union <- function(x, y) {
  x_tzone <- x@tzone
  y_tzone <- y@tzone

  if (tz_is_local(x_tzone)) {
    y_tzone
  } else {
    x_tzone
  }
}

# ------------------------------------------------------------------------------
# Constructors

# S4 constructors are slow, so empty Period/Duration objects are saved for
# repeated use in `vec_ptype2()`. Interval types depend on the `tzone`, so
# the constructor must be used.

delayedAssign("lubridate_shared_empty_period", period())
delayedAssign("lubridate_shared_empty_duration", duration())

new_empty_interval <- function(tzone) {
  interval(tzone = tzone)
}

# ------------------------------------------------------------------------------
# Period - proxy / restore

# Method registered in `.onLoad()`
vec_proxy.Period <- function(x, ...) {
  out <- list(
    year = x@year,
    month = x@month,
    day = x@day,
    hour = x@hour,
    minute = x@minute,
    second = x@.Data
  )

  # Push names onto the `second` column (i.e. `.Data`) so they get sliced and restored
  names <- names(x)
  if (!is.null(names)) {
    names(out$second) <- names
  }

  n <- length(out$year)

  vctrs::new_data_frame(out, n = n)
}

# Method registered in `.onLoad()`
vec_restore.Period <- function(x, to, ...) {
  out <- period(
    year = x$year,
    month = x$month,
    day = x$day,
    hour = x$hour,
    minute = x$minute,
    second = x$second
  )

  names <- names(x$second)
  if (!is.null(names)) {
    names(out) <- names
  }

  out
}

# ------------------------------------------------------------------------------
# Period - ptype2

# Method registered in `.onLoad()`
vec_ptype2.Period.Period <- function(x, y, ...) {
  lubridate_shared_empty_period
}

# ------------------------------------------------------------------------------
# Period - cast

# Method registered in `.onLoad()`
vec_cast.Period.Period <- function(x, to, ...) {
  x
}

# ------------------------------------------------------------------------------
# Duration - proxy / restore

# Method registered in `.onLoad()`
vec_proxy.Duration <- function(x, ...) {
  out <- x@.Data

  names <- names(x)
  if (!is.null(names)) {
    names(out) <- names
  }

  out
}

# Method registered in `.onLoad()`
vec_restore.Duration <- function(x, to, ...) {
  out <- duration(x, units = "seconds")

  names <- names(x)
  if (!is.null(names)) {
    names(out) <- names
  }

  out
}

# ------------------------------------------------------------------------------
# Duration - ptype2

# Method registered in `.onLoad()`
vec_ptype2.Duration.Duration <- function(x, y, ...) {
  lubridate_shared_empty_duration
}

# Method registered in `.onLoad()`
vec_ptype2.Duration.difftime <- function(x, y, ...) {
  lubridate_shared_empty_duration
}

# Method registered in `.onLoad()`
vec_ptype2.difftime.Duration <- function(x, y, ...) {
  lubridate_shared_empty_duration
}

# ------------------------------------------------------------------------------
# Duration - cast

# Method registered in `.onLoad()`
vec_cast.Duration.Duration <- function(x, to, ...) {
  x
}

# Method registered in `.onLoad()`
vec_cast.Duration.difftime <- function(x, to, ...) {
  as_duration_from_difftime(x)
}

# Method registered in `.onLoad()`
vec_cast.difftime.Duration <- function(x, to, ...) {
  as_difftime_from_duration(x)
}

# ------------------------------------------------------------------------------
# Interval - proxy / restore

# Method registered in `.onLoad()`
vec_proxy.Interval <- function(x, ...) {
  # Ordered in such a way that the start date controls interval ordering.
  # Ties in the start date are resolved by placing shorter intervals first.
  out <- list(
    start = x@start,
    span = x@.Data
  )

  # Push names onto the `span` column (i.e. `.Data`) so they get sliced and restored
  names <- names(x)
  if (!is.null(names)) {
    names(out$span) <- names
  }

  n <- length(out$start)

  vctrs::new_data_frame(out, n = n)
}

# Method registered in `.onLoad()`
vec_restore.Interval <- function(x, to, ...) {
  out <- new(
    "Interval",
    x$span,
    start = x$start,
    tzone = tz(x$start)
  )

  names <- names(x$span)
  if (!is.null(names)) {
    names(out) <- names
  }

  out
}

# ------------------------------------------------------------------------------
# Interval - ptype2

# Method registered in `.onLoad()`
vec_ptype2.Interval.Interval <- function(x, y, ...) {
  new_empty_interval(tzone = tz_union(x, y))
}

# ------------------------------------------------------------------------------
# Interval - cast

# Method registered in `.onLoad()`
vec_cast.Interval.Interval <- function(x, to, ...) {
  x_tzone <- x@tzone
  to_tzone <- to@tzone

  if (identical(x_tzone, to_tzone)) {
    return(x)
  }

  x_span <- x@.Data
  x_start <- x@start

  start <- with_tz(x_start, to_tzone)

  new("Interval", x_span, start = start, tzone = to_tzone)
}

# ------------------------------------------------------------------------------

tz_is_local <- function(x) {
  identical(x, "")
}

tz_union <- function(x, y) {
  x_tzone <- x@tzone
  y_tzone <- y@tzone

  if (tz_is_local(x_tzone)) {
    y_tzone
  } else {
    x_tzone
  }
}

#' Get date-time in a different time zone
#'
#' with_tz returns a date-time as it would appear in a different time zone.
#' The actual moment of time measured does not change, just the time zone it is
#' measured in. with_tz defaults to the Universal Coordinated time zone (UTC)
#' when an unrecognized time zone is inputted. See [Sys.timezone()]
#' for more information on how R recognizes time zones.
#'
#' @param time a POSIXct, POSIXlt, Date, chron date-time object or a data.frame
#'   object. When a data.frame all POSIXt elements of a data.frame are processed
#'   with `with_tz()` and new data.frame is returned.
#' @param tzone a character string containing the time zone to convert to. R
#'   must recognize the name contained in the string as a time zone on your
#'   system.
#' @param ... Parameters passed to other methods.
#' @return a POSIXct object in the updated time zone
#' @keywords chron manip
#' @seealso [force_tz()]
#' @examples
#' x <- ymd_hms("2009-08-07 00:00:01", tz = "America/New_York")
#' with_tz(x, "GMT")
#' @export
with_tz <- function(time, tzone = "", ...) {
    UseMethod("with_tz")
}

#' @rdname with_tz
#' @export
with_tz.default <- function(time, tzone = "", ...) {
  if (!timechange:::C_valid_tz(tzone)) {
    warning(sprintf("Unrecognized time zone '%s'", tzone))
  }
  if (is.data.frame(time)) {
    for (nm in names(time)) {
      if (is.POSIXt(time[[nm]])) {
        time[[nm]] <- .with_tz(time[[nm]], tzone = tzone)
      }
    }
    time
  } else {
    .with_tz(time, tzone)
  }
}

.with_tz <- function(time, tzone = "") {
  out <- if (!is.POSIXct(time)) as.POSIXct(time) else time
  attr(out, "tzone") <- tzone
  if (is.POSIXlt(time)) {
    out <- as.POSIXlt(out)
  }
  out
}

#' Replace time zone to create new date-time
#'
#' `force_tz` returns the date-time that has the same clock time as input time,
#'  but in the new time zone. `force_tzs` is the parallel version of `force_tz`,
#'  meaning that every element from `time` argument is matched with the
#'  corresponding time zone in `tzones` argument.
#'
#'  Although the new date-time has the same clock time (e.g. the same values in
#'  the year, month, days, etc. elements) it is a different moment of time than
#'  the input date-time.
#'
#'  As R date-time vectors cannot hold elements with non-uniform time zones,
#'  `force_tzs` returns a vector with time zone `tzone_out`, UTC by default.
#'
#' @param time a POSIXct, POSIXlt, Date, chron date-time object, or a data.frame
#'   object. When a data.frame all POSIXt elements of a data.frame are processed
#'   with `force_tz()` and new data.frame is returned.
#' @param tzone a character string containing the time zone to convert to. R
#'   must recognize the name contained in the string as a time zone on your
#'   system.
#' @param roll deprecated, same as `roll_dst` parameter.
#' @param ... Parameters passed to other methods.
#' @return a POSIXct object in the updated time zone
#' @keywords chron manip
#' @seealso [with_tz()], [local_time()]
#' @inheritParams timechange::time_add
#' @examples
#' x <- ymd_hms("2009-08-07 00:00:01", tz = "America/New_York")
#' force_tz(x, "UTC")
#' force_tz(x, "Europe/Amsterdam")
#'
#' ## DST skip:
#' y <- ymd_hms("2010-03-14 02:05:05 UTC")
#' force_tz(y, "America/New_York", roll_dst = "NA")
#' force_tz(y, "America/New_York", roll_dst = "pre")
#' force_tz(y, "America/New_York", roll_dst = "boundary")
#' force_tz(y, "America/New_York", roll_dst = "post")
#'
#' ## DST repeat
#' y <- ymd_hms("2014-11-02 01:35:00", tz = "UTC")
#' force_tz(y, "America/New_York", roll_dst = "NA")
#' force_tz(y, "America/New_York", roll_dst = "pre")
#' force_tz(y, "America/New_York", roll_dst = "boundary")
#' force_tz(y, "America/New_York", roll_dst = "post")
#'
#' ## DST skipped and repeated
#' y <- ymd_hms("2010-03-14 02:05:05 UTC", "2014-11-02 01:35:00", tz = "UTC")
#' force_tz(y, "America/New_York", roll_dst = c("NA", "pre"))
#' force_tz(y, "America/New_York", roll_dst = c("boundary", "post"))
#' @export
force_tz <- function(time, tzone = "", ...) {
    UseMethod("force_tz")
}

#' @rdname force_tz
#' @export
force_tz.default <- function(time, tzone = "", roll_dst = c("NA", "post"), roll = NULL, ...) {
  tzone <- as.character(tzone)
  if (is.data.frame(time)) {
    for (nm in names(time)) {
      if (is.POSIXt(time[[nm]])) {
        time[[nm]] <- force_tz(time[[nm]], tzone = tzone)
      }
    }
    time
  } else {
    roll_dst <- normalize_roll_dst(roll_dst, roll)
    if (is.POSIXct(time)) {
      timechange::time_force_tz(time, tz = tzone, roll_dst = roll_dst)
    } else if (is.Date(time)) {
      timechange::time_force_tz(date_to_posix(time), tz = tzone, roll_dst = roll_dst)
    } else {
      out <- timechange::time_force_tz(as.POSIXct(time, tz = tz(time)), tz = tzone, roll_dst = roll_dst)
      reclass_date(out, time)
    }
  }
}

#' @param tzones character vector of timezones to be "enforced" on `time` time
#'   stamps. If `time` and `tzones` lengths differ, the smaller one is recycled
#'   in accordance with usual R conventions.
#' @param tzone_out timezone of the returned date-time vector (for `force_tzs`).
#' @rdname force_tz
#' @examples
#'
#' ## Heterogeneous time-zones:
#'
#' x <- ymd_hms(c("2009-08-07 00:00:01", "2009-08-07 01:02:03"))
#' force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"))
#' force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"), tzone_out = "America/New_York")
#'
#' x <- ymd_hms("2009-08-07 00:00:01")
#' force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"))
#' @export
force_tzs <- function(time, tzones, tzone_out = "UTC", roll_dst = c("NA", "post"), roll = NULL) {
  if (length(tzones) < length(time)) {
    tzones <- rep_len(tzones, length(time))
  } else if (length(tzones) > length(time)) {
    attr <- attributes(time)
    time <- rep_len(time, length(tzones))
    attributes(time) <- attr
  }
  roll_dst <- normalize_roll_dst(roll_dst, roll)
  out <- timechange::time_force_tz(as.POSIXct(time), tzones, tzone_out, roll_dst)
  reclass_date(out, time)
}

#' Get local time from a date-time vector.
#'
#' `local_time` retrieves day clock time in specified time zones. Computation is
#' vectorized over both `dt` and `tz` arguments, the shortest is recycled in
#' accordance with standard R rules.
#'
#' @param dt a date-time object.
#' @param tz a character vector of timezones for which to compute the local time.
#' @param units passed directly to [as.difftime()].
#' @examples
#'
#' x <- ymd_hms(c("2009-08-07 01:02:03", "2009-08-07 10:20:30"))
#' local_time(x, units = "secs")
#' local_time(x, units = "hours")
#' local_time(x, "Europe/Amsterdam")
#' local_time(x, "Europe/Amsterdam") == local_time(with_tz(x, "Europe/Amsterdam"))
#'
#' x <- ymd_hms("2009-08-07 01:02:03")
#' local_time(x, c("America/New_York", "Europe/Amsterdam", "Asia/Shanghai"), unit = "hours")
#' @export
local_time <- function(dt, tz = NULL, units = "secs") {
  timechange::time_clock_at_tz(dt, tz, units)
}

test_that("with_tz works as expected", {
  x_ct <- as.POSIXct("2008-08-03 10:01:59", tz = "America/New_York")
  y_ct <- as.POSIXct("2008-08-03 14:01:59", tz = "UTC")

  expect_equal(with_tz(x_ct, "UTC"), y_ct)
  expect_s3_class(with_tz(x_ct, "UTC"), "POSIXct")

  x_lt <- as.POSIXlt("2008-08-03 10:01:59", tz = "America/New_York")
  y_lt <- as.POSIXlt("2008-08-03 14:01:59", tz = "UTC")

  expect_equal(with_tz(x_lt, "UTC"), y_lt)
  expect_equal(as.POSIXct(with_tz(y_lt, "America/New_York")), x_ct)
  expect_equal(x_ct, as.POSIXct(with_tz(y_lt, "America/New_York")))
  expect_s3_class(with_tz(y_lt, "UTC"), "POSIXlt")
})

test_that("with_tz handles data.frames", {
  x <- as.POSIXct("2008-08-03 10:01:59", tz = "America/New_York")
  df <- data.frame(x = x, y = as.POSIXlt(x), z = "blabla")
  df <- with_tz(df, "UTC")
  x_out <- as.POSIXct(format(as.POSIXct(x), tz = "UTC"), tz = "UTC")
  expect_equal(df$x, x_out)
  expect_equal(df$y, x_out)
})

test_that("with_tz throws warning on unrecognized time zones", {
  t <- now()
  expect_warning(with_tz(t, "blablabla"))
  expect_silent(with_tz(t, "UTC"))
  expect_silent(with_tz(t, ""))
  expect_silent(with_tz(t, "America/New_York"))
})

# force_tz ----------------------------------------------------------------

test_that("force_tz works as expected", {
  x_ct <- as.POSIXct("2008-08-03 10:01:59", tz = "America/New_York")
  y_ct <- as.POSIXct("2008-08-03 10:01:59", tz = "UTC")

  expect_equal(force_tz(x_ct, "UTC"), y_ct)
  expect_s3_class(force_tz(x_ct, "UTC"), "POSIXct")

  x_lt <- as.POSIXlt(x_ct)
  y_lt <- as.POSIXlt(y_ct)

  expect_equal(force_tz(y_lt, "UTC"), y_lt)
  expect_s3_class(force_tz(y_lt, "UTC"), "POSIXlt")
})

test_that("force_tz handles data.frames", {
  x <- as.POSIXct("2008-08-03 10:01:59", tz = "America/New_York")
  x_out <- as.POSIXct(format(as.POSIXct(x)), tz = "UTC")
  df <- data.frame(x = x, y = as.POSIXlt(x), z = "blabla")
  df <- force_tz(df, "UTC")
  expect_equal(df$x, x_out)
  expect_equal(df$y, x_out)
})


test_that("force_tzs works as expected", {
  x <- ymd_hms(c("2009-08-07 00:00:01", "2009-08-07 00:00:01"))
  expect_equal(
    force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam")),
    ymd_hms("2009-08-07 04:00:01 UTC", "2009-08-06 22:00:01 UTC")
  )
  expect_equal(
    force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"), tzone_out = "America/New_York"),
    ymd_hms("2009-08-07 00:00:01 EDT", "2009-08-06 18:00:01 EDT", tz = "America/New_York")
  )
  ## recycling
  expect_equal(
    force_tzs(x, tzones = "America/New_York", tzone_out = "UTC"),
    ymd_hms("2009-08-07 04:00:01 UTC", "2009-08-07 04:00:01 UTC")
  )
  x <- ymd_hms("2009-08-07 00:00:01")
  expect_equal(
    force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam")),
    ymd_hms("2009-08-07 04:00:01 UTC", "2009-08-06 22:00:01 UTC")
  )
  expect_equal(
    force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"), tzone_out = "America/New_York"),
    ymd_hms("2009-08-07 00:00:01 EDT", "2009-08-06 18:00:01 EDT", tz = "America/New_York")
  )
})

test_that("force_tzs is robusts against overflow", {
  x <- ymd_hms(c(
    "2038-01-19 03:14:06", "2038-01-19 03:14:07", "2038-01-19 03:14:08",
    "2038-01-19 03:14:09", "2038-01-19 03:14:10"
  ))
  expect_equal(
    force_tzs(x, tzones = "America/New_York"),
    with_tz(
      ymd_hms(c(
        "2038-01-19 03:14:06", "2038-01-19 03:14:07", "2038-01-19 03:14:08",
        "2038-01-19 03:14:09", "2038-01-19 03:14:10"
      ),
      tz = "America/New_York"
      ),
      "UTC"
    )
  )
})

test_that("force_tz doesn't return NA just because new time zone uses DST", {
  poslt <- as.POSIXlt("2009-03-14 02:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  poslt2 <- force_tz(poslt, tzone = "America/New_York")

  expect_true(!is.na(poslt2))
})

test_that("force_tz works within repeated DST", {
  ## "2014-11-02 00:00:00 CST"
  ch <- ymd_hms("2014-11-02 00:00:00", tz = "America/Chicago")
  eu <- ymd_hms("2014-11-02 00:00:00", tz = "Europe/Berlin")

  expect_equal(
    force_tz(eu + dminutes(95), "America/Chicago", roll_dst = "pre"),
    ch + dminutes(95))
  expect_equal(
    force_tz(eu + dminutes(95), "America/Chicago", roll_dst = "post"),
    ch + dminutes(155))
  expect_equal(
    force_tz(eu + dminutes(95), "America/Chicago", roll_dst = "boundary"),
    ch + dminutes(120))

  expect_equal(
    force_tz(eu + dminutes(95), "America/Chicago", roll_dst = "post"),
    ch + dminutes(155))
  expect_equal(
    force_tz(eu + dminutes(95), "America/Chicago", roll_dst = "pre"),
    ch + dminutes(95))
  expect_equal(
    force_tz(eu + dminutes(95), "America/Chicago", roll_dst = "boundary"),
    ch + dminutes(120))

  expect_equal(
    force_tz(ch + dminutes(95), "Europe/Berlin", roll_dst = "post"),
    eu + dminutes(95))
  expect_equal(
    force_tz(ch + dminutes(95), "Europe/Berlin", roll_dst = "pre"),
    eu + dminutes(95))
  expect_equal(
    force_tz(ch + dminutes(95), "Europe/Berlin", roll_dst = "boundary"),
    eu + dminutes(95))

  ## default roll_dst is "post"
  expect_equal(
    force_tz(eu + dminutes(95), "America/Chicago"),
    ch + dminutes(155))
  expect_equal(
    force_tz(ch + dminutes(95), "Europe/Berlin"),
    eu + dminutes(95))
  expect_equal(
    force_tz(ch + dminutes(155), "Europe/Berlin"),
    eu + dminutes(95))

})

# local_time --------------------------------------------------------------

test_that("local_time works as expected", {
  x <- ymd_hms(c("2009-08-07 01:02:03", "2009-08-07 10:20:30"))
  expect_equal(
    local_time(x, units = "secs"),
    as.difftime(c(3723, 37230), units = "secs")
  )
  expect_equal(
    local_time(x, units = "hours"),
    as.difftime(c(3723, 37230) / 3600, units = "hours")
  )
  expect_equal(
    local_time(x, "Europe/Amsterdam"),
    local_time(with_tz(x, "Europe/Amsterdam"))
  )

  x1 <- ymd_hms("2009-08-07 01:02:03")
  x2 <- local_time(x1, c("America/New_York", "Europe/Amsterdam", "Asia/Shanghai"))

  expect_equal(x2[1], local_time(with_tz(x1, "America/New_York")))
  expect_equal(x2[2], local_time(with_tz(x1, "Europe/Amsterdam")))
  expect_equal(x2[3], local_time(with_tz(x1, "Asia/Shanghai")))
})

# tz --------------------------------------------------------------------

test_that("tz() returns UTC on special objects", {
  expect_equal(tz(NA), "UTC")
  expect_equal(tz("character"), "UTC")
  expect_equal(tz(today()), "UTC")
  expect_equal(tz(POSIXct()), "UTC")
  expect_equal(tz(Date()), "UTC")
})


test_that("If original post/pre is known, ignore roll_DST", {

  tt <- ymd_hms("2014-11-02 01:35:00")
  chpst <- force_tz(tt, "America/Chicago", roll_dst = "post")
  chpdt <- force_tz(tt, "America/Chicago", roll_dst = "pre")
  lapst <- force_tz(tt, "America/Los_Angeles", roll_dst = "post")
  lapdt <- force_tz(tt, "America/Los_Angeles", roll_dst = "pre")

  expect_equal(lapst, force_tz(chpst, "America/Los_Angeles", roll_dst = "pre"))
  expect_equal(lapst, force_tz(chpst, "America/Los_Angeles", roll_dst = "post"))
  expect_equal(lapst, force_tz(chpst, "America/Los_Angeles", roll_dst = "boundary"))

  expect_equal(lapdt, force_tz(chpdt, "America/Los_Angeles", roll_dst = "pre"))
  expect_equal(lapdt, force_tz(chpdt, "America/Los_Angeles", roll_dst = "post"))
  expect_equal(lapdt, force_tz(chpdt, "America/Los_Angeles", roll_dst = "boundary"))

  expect_equal(lapst, force_tzs(chpst, "America/Los_Angeles", "America/Los_Angeles", roll_dst = "pre"))
  expect_equal(lapst, force_tzs(chpst, "America/Los_Angeles", "America/Los_Angeles", roll_dst = "post"))
  expect_equal(lapst, force_tzs(chpst, "America/Los_Angeles", "America/Los_Angeles", roll_dst = "boundary"))
  expect_equal(lapst, force_tzs(chpst, "America/Los_Angeles", "America/Los_Angeles", roll_dst = "pre"))
  expect_equal(lapst, force_tzs(chpst, "America/Los_Angeles", "America/Los_Angeles", roll_dst = "post"))
  expect_equal(lapst, force_tzs(chpst, "America/Los_Angeles", "America/Los_Angeles", roll_dst = "boundary"))

  expect_equal(with_tz(c(tt, chpdt, lapdt), "America/Chicago"),
    force_tzs(tt,  #c(tt, chpdt, lapdt),
      c("UTC", "America/Chicago", "America/Los_Angeles"),
      tzone_out = "America/Chicago",
      roll_dst = "pre"))

  expect_equal(with_tz(c(tt, chpst, lapst), "America/Chicago"),
    force_tzs(tt,  #c(tt, chpdt, lapdt),
      c("UTC", "America/Chicago", "America/Los_Angeles"),
      tzone_out = "America/Chicago",
      roll_dst = "post"))

  expect_equal(with_tz(c(tt, chpdt, lapdt), "America/Chicago"),
    force_tzs(chpdt,
      c("UTC", "America/Chicago", "America/Los_Angeles"),
      tzone_out = "America/Chicago",
      roll_dst = "pre"))

  expect_equal(with_tz(c(tt, chpdt, lapdt), "America/Chicago"),
    force_tzs(chpdt,
      c("UTC", "America/Chicago", "America/Los_Angeles"),
      tzone_out = "America/Chicago",
      roll_dst = "post"))

  expect_equal(with_tz(c(tt, chpst, lapst), "America/Chicago"),
    force_tzs(chpst,
      c("UTC", "America/Chicago", "America/Los_Angeles"),
      tzone_out = "America/Chicago",
      roll_dst = "post"))

  expect_equal(with_tz(c(tt, chpst + 100, chpst + 3600), "America/Chicago"),
    force_tzs(with_tz(c(chpst, chpst + 100, chpdt), "America/Chicago"),
      c("UTC", "America/Chicago", "America/Los_Angeles"),
      tzone_out = "America/Chicago",
      roll_dst = "post"))
})


test_that("as_datetime does not loose time zone", {
  ## #1085
  x <- ymd_hms("2022-01-01 23:40:36", tz = "US/Pacific")
  expect_identical(x, as_datetime(x))
  expect_identical(with_tz(x, "Europe/Amsterdam"), as_datetime(x, tz = "Europe/Amsterdam"))
  x <- as.POSIXlt(x)
  expect_identical(x, as_datetime(x))
  expect_identical(with_tz(x, "Europe/Amsterdam"), as_datetime(x, tz = "Europe/Amsterdam"))
})

test_that("as_datetime parsers with exact formats", {
  ## #1097
  expect_equal(
    as_datetime("blabla2022-xyz01-03", format = "blabla%Y-xyz%m-%d"),
    ymd("2022-01-03", tz = "UTC"))
  expect_equal(
    as_datetime(c("2022-01-01", "blabla2022-xyz01-03"), format = c("%Y-%m-%d", "blabla%Y-xyz%m-%d")),
    ymd(c("2022-01-01", "2022-01-03"), tz = "UTC"))
})

context("Time zones")

test_that("with_tz works as expected", {
  x_ct <- as.POSIXct("2008-08-03 10:01:59", tz = "America/New_York")
  y_ct <- as.POSIXct("2008-08-03 14:01:59", tz = "UTC")

  expect_equal(with_tz(x_ct, "UTC"), y_ct)
  expect_s3_class(with_tz(x_ct, "UTC"), "POSIXct")

  x_lt <- as.POSIXlt("2008-08-03 10:01:59", tz = "America/New_York")
  y_lt <- as.POSIXlt("2008-08-03 14:01:59", tz = "UTC")

  expect_equal(with_tz(x_lt, "UTC"), y_lt)
  expect_equal(with_tz(y_lt, "America/New_York"), x_lt)
  expect_equal(x_ct, as.POSIXct(with_tz(y_lt, "America/New_York")))
  expect_s3_class(with_tz(y_lt, "UTC"), "POSIXlt")
})

test_that("with_tz handles data.frames", {
  x <- as.POSIXct("2008-08-03 10:01:59", tz = "America/New_York")
  df <- data.frame(x = x, y = as.POSIXlt(x), z = "blabla")
  df <- with_tz(df, "UTC")
  x_out <- as.POSIXct(format(as.POSIXct(x), tz = "UTC"), tz = "UTC")
  expect_equal(df$x, x_out)
  expect_equal(df$y, x_out)
})

test_that("with_tz throws warning on unrecognized time zones", {
  t <- now()
  expect_warning(with_tz(t, "blablabla"))
  expect_silent(with_tz(t, "UTC"))
  expect_silent(with_tz(t, ""))
  expect_silent(with_tz(t, "America/New_York"))
})

# force_tz ----------------------------------------------------------------

test_that("force_tz works as expected", {
  x_ct <- as.POSIXct("2008-08-03 10:01:59", tz = "America/New_York")
  y_ct <- as.POSIXct("2008-08-03 10:01:59", tz = "UTC")

  expect_equal(force_tz(x_ct, "UTC"), y_ct)
  expect_s3_class(force_tz(x_ct, "UTC"), "POSIXct")

  x_lt <- as.POSIXlt(x_ct)
  y_lt <- as.POSIXlt(y_ct)

  expect_equal(force_tz(y_lt, "UTC"), y_lt)
  expect_s3_class(force_tz(y_lt, "UTC"), "POSIXlt")
})

test_that("force_tz handles data.frames", {
  x <- as.POSIXct("2008-08-03 10:01:59", tz = "America/New_York")
  x_out <- as.POSIXct(format(as.POSIXct(x)), tz = "UTC")
  df <- data.frame(x = x, y = as.POSIXlt(x), z = "blabla")
  df <- force_tz(df, "UTC")
  expect_that(df$x, equals(x_out))
  expect_that(df$y, equals(x_out))
})


test_that("force_tzs works as expected", {
  x <- ymd_hms(c("2009-08-07 00:00:01", "2009-08-07 00:00:01"))
  expect_equal(force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam")),
               ymd_hms("2009-08-07 04:00:01 UTC", "2009-08-06 22:00:01 UTC"))
  expect_equal(force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"), tzone_out = "America/New_York"),
               ymd_hms("2009-08-07 00:00:01 EDT", "2009-08-06 18:00:01 EDT", tz = "America/New_York"))
  ## recycling
  expect_equal(force_tzs(x, tzones = "America/New_York", tzone_out = "UTC"),
               ymd_hms("2009-08-07 04:00:01 UTC", "2009-08-07 04:00:01 UTC"))
  x <- ymd_hms("2009-08-07 00:00:01")
  expect_equal(force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam")),
               ymd_hms("2009-08-07 04:00:01 UTC", "2009-08-06 22:00:01 UTC"))
  expect_equal(force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"), tzone_out = "America/New_York"),
               ymd_hms("2009-08-07 00:00:01 EDT", "2009-08-06 18:00:01 EDT", tz = "America/New_York"))
})

test_that("force_tzs is robusts against overflow", {
  x <- ymd_hms(c("2038-01-19 03:14:06", "2038-01-19 03:14:07", "2038-01-19 03:14:08",
                 "2038-01-19 03:14:09", "2038-01-19 03:14:10"))
  expect_equal(force_tzs(x, tzones = "America/New_York"),
               with_tz(ymd_hms(c("2038-01-19 03:14:06", "2038-01-19 03:14:07", "2038-01-19 03:14:08",
                                 "2038-01-19 03:14:09", "2038-01-19 03:14:10"),
                               tz = "America/New_York"),
                       "UTC"))
})

test_that("force_tz doesn't return NA just because new time zone uses DST", {
  poslt <- as.POSIXlt("2009-03-14 02:59:59", tz = "UTC", format
                      = "%Y-%m-%d %H:%M:%S")
  poslt2 <- force_tz(poslt, tzone = "America/New_York")

  expect_true(!is.na(poslt2))
})

# local_time --------------------------------------------------------------

test_that("local_time works as expected", {
  x <- ymd_hms(c("2009-08-07 01:02:03", "2009-08-07 10:20:30"))
  expect_equal(local_time(x, units = "secs"),
               as.difftime(c(3723, 37230), units = "secs"))
  expect_equal(local_time(x, units = "hours"),
               as.difftime(c(3723, 37230)/3600, units = "hours"))
  expect_equal(local_time(x, "Europe/Amsterdam"),
               local_time(with_tz(x, "Europe/Amsterdam")))

  x1 <- ymd_hms("2009-08-07 01:02:03")
  x2 <- local_time(x1, c("America/New_York", "Europe/Amsterdam", "Asia/Shanghai"))

  expect_equal(x2[1], local_time(with_tz(x1, "America/New_York")))
  expect_equal(x2[2], local_time(with_tz(x1, "Europe/Amsterdam")))
  expect_equal(x2[3], local_time(with_tz(x1, "Asia/Shanghai")))
})

# tz --------------------------------------------------------------------

test_that("tz() returns UTC on special objects", {
  expect_equal(tz(NA), "UTC")
  expect_equal(tz("character"), "UTC")
  expect_equal(tz(today()), "UTC")
  expect_equal(tz(POSIXct()), "UTC")
  expect_equal(tz(Date()), "UTC")
})

## CCTZ needs zoneinfo. On windows we set it to R's own zoneinfo. On unix like
## it's in "/usr/share/zoneinfo" where CCTZ looks by default. On some systems
## (solaris, osx) it might be in a different location. So, help ourselves by
## setting the TZDIR env var, but only if it's not already set.
##
## As of R4.0.3 Sys.timezone() intrusively sets TZDIR to non path values.
## https://github.com/wch/r-source/commit/92712b530e5f0f3c19b1e83ed86554bbfb6d9095#diff-883a8df98e17cbfd43513e7378ae4899bf504d8dbd8e47d021d8493602ab4b8fR52-R66
tzdir_set <- function() {
  # On MacOS, this chooses between system and built-in zoneinfo based on
  # which is newer
  Sys.timezone()

  tzdir <- tzdir_get()
  if (tzdir == "internal") {
    Sys.setenv(TZDIR = tzdir_internal())
  } else if (tzdir %in% c("macOS", "")) {
    dir <- tzdir_find()
    if (is.null(dir)) {
      warning("Failed to locate timezone database", call. = FALSE)
    } else {
      Sys.setenv(TZDIR = dir)
    }
  }
}

tzdir_get <- function() {
  Sys.getenv("TZDIR")
}

## adapted from Syz.timezone and OlsonNames function
tzdir_find <- function() {
  if (.Platform$OS.type == "windows") {
    return(tzdir_internal())
  }

  tzdirs <- c(
    "/usr/share/zoneinfo",
    "/usr/share/lib/zoneinfo",
    "/usr/lib/zoneinfo",
    "/usr/local/etc/zoneinfo",
    "/etc/zoneinfo",
    "/usr/etc/zoneinfo",
    "/usr/share/zoneinfo.default",
    "/var/db/timezone/zoneinfo",
    tzdir_internal()
  )
  tzdirs <- tzdirs[file.exists(tzdirs)]
  if (length(tzdirs)) {
    tzdirs[[1]]
  } else {
    NULL
  }
}

tzdir_internal <- function() {
  file.path(R.home("share"), "zoneinfo")
}

#' Get date-time in a different time zone
#'
#' with_tz returns a date-time as it would appear in a different time zone.
#' The actual moment of time measured does not change, just the time zone it is
#' measured in. with_tz defaults to the Universal Coordinated time zone (UTC)
#' when an unrecognized time zone is inputted. See [Sys.timezone()]
#' for more information on how R recognizes time zones.
#'
#' @param time a POSIXct, POSIXlt, Date, chron date-time object or a data.frame
#'   object. When a data.frame all POSIXt elements of a data.frame are processed
#'   with `with_tz()` and new data.frame is returned.
#' @param tzone a character string containing the time zone to convert to. R
#'   must recognize the name contained in the string as a time zone on your
#'   system.
#' @return a POSIXct object in the updated time zone
#' @keywords chron manip
#' @seealso [force_tz()]
#' @examples
#' x <- ymd_hms("2009-08-07 00:00:01", tz = "America/New_York")
#' with_tz(x, "GMT")
#' @export
with_tz <- function(time, tzone = "") {
  if (!C_valid_tz(tzone))
    warning(sprintf("Unrecognized time zone '%s'", tzone))
  if (is.data.frame(time)) {
    for (nm in names(time)) {
      if (is.POSIXt(time[[nm]])) {
        time[[nm]] <- .with_tz(time[[nm]], tzone = tzone)
      }
    }
    time
  } else {
    .with_tz(time, tzone)
  }
}

.with_tz <- function(time, tzone = "") {
  out <- if (!is.POSIXct(time)) as.POSIXct(time) else time
  attr(out, "tzone") <- tzone
  if (is.POSIXlt(time)) {
    out <- as.POSIXlt(out)
  }
  out
}

#' Replace time zone to create new date-time
#'
#' `force_tz` returns the date-time that has the same clock time as input time,
#'  but in the new time zone. `force_tzs` is the parallel version of `force_tz`,
#'  meaning that every element from `time` argument is matched with the
#'  corresponding time zone in `tzones` argument.
#'
#'  Although the new date-time has the same clock time (e.g. the same values in
#'  the year, month, days, etc. elements) it is a different moment of time than
#'  the input date-time.
#'
#'  As R date-time vectors cannot hold elements with non-uniform time zones,
#'  `force_tzs` returns a vector with time zone `tzone_out`, UTC by default.
#'
#' @param time a POSIXct, POSIXlt, Date, chron date-time object, or a data.frame
#'   object. When a data.frame all POSIXt elements of a data.frame are processed
#'   with `force_tz()` and new data.frame is returned.
#' @param tzone a character string containing the time zone to convert to. R
#'   must recognize the name contained in the string as a time zone on your
#'   system.
#' @param roll logical. If TRUE, and `time` falls into the DST-break, assume
#'   the next valid civil time, otherwise return NA. See examples.
#' @return a POSIXct object in the updated time zone
#' @keywords chron manip
#' @seealso [with_tz()], [local_time()]
#' @examples
#' x <- ymd_hms("2009-08-07 00:00:01", tz = "America/New_York")
#' force_tz(x, "UTC")
#' force_tz(x, "Europe/Amsterdam")
#'
#' ## DST skip:
#'
#' y <- ymd_hms("2010-03-14 02:05:05 UTC")
#' force_tz(y, "America/New_York", roll=FALSE)
#' force_tz(y, "America/New_York", roll=TRUE)
#' @export
force_tz <- function(time, tzone = "", roll = FALSE) {
  tzone <- as.character(tzone)
  if (is.data.frame(time)) {
    for (nm in names(time)) {
      if (is.POSIXt(time[[nm]])) {
        time[[nm]] <- force_tz(time[[nm]], tzone = tzone)
      }
    }
    time
  } else {
    if (is.POSIXct(time))
      C_force_tz(time, tz = tzone, roll)
    else if (is.Date(time))
      C_force_tz(date_to_posix(time), tz = tzone, roll)
    else {
      out <- C_force_tz(as.POSIXct(time, tz = tz(time)), tz = tzone, roll)
      reclass_date(out, time)
    }
  }
}

#' @param tzones character vector of timezones to be "enforced" on `time` time
#'   stamps. If `time` and `tzones` lengths differ, the smaller one is recycled
#'   in accordance with usual R conventions.
#' @param tzone_out timezone of the returned date-time vector (for `force_tzs`).
#' @rdname force_tz
#' @examples
#'
#' ## Heterogeneous time-zones:
#'
#' x <- ymd_hms(c("2009-08-07 00:00:01", "2009-08-07 01:02:03"))
#' force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"))
#' force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"), tzone_out = "America/New_York")
#'
#' x <- ymd_hms("2009-08-07 00:00:01")
#' force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"))
#' @export
force_tzs <- function(time, tzones, tzone_out = "UTC", roll = FALSE) {
  if (length(tzones) < length(time))
    tzones <- rep_len(tzones, length(time))
  else if (length(tzones) > length(time)) {
    attr <- attributes(time)
    time <- rep_len(time, length(tzones))
    attributes(time) <- attr
  }
  out <- C_force_tzs(as.POSIXct(time), tzones, tzone_out, roll)
  reclass_date(out, time)
}

#' Get local time from a date-time vector.
#'
#' `local_time` retrieves day clock time in specified time zones. Computation is
#' vectorized over both `dt` and `tz` arguments, the shortest is recycled in
#' accordance with standard R rules.
#'
#' @param dt a date-time object.
#' @param tz a character vector of timezones for which to compute the local time.
#' @param units passed directly to [as.difftime()].
#' @examples
#'
#' x <- ymd_hms(c("2009-08-07 01:02:03", "2009-08-07 10:20:30"))
#' local_time(x, units = "secs")
#' local_time(x, units = "hours")
#' local_time(x, "Europe/Amsterdam")
#' local_time(x, "Europe/Amsterdam") == local_time(with_tz(x, "Europe/Amsterdam"))
#'
#' x <- ymd_hms("2009-08-07 01:02:03")
#' local_time(x, c("America/New_York", "Europe/Amsterdam", "Asia/Shanghai"), unit = "hours")
#' @export
local_time <- function(dt, tz = NULL, units = "secs") {
  if (is.null(tz))
    tz <- tz(dt)
  if (length(tz) < length(dt))
    tz <- rep_len(tz, length(dt))
  else if (length(tz) > length(dt)) {
    attr <- attributes(dt)
    dt <- rep_len(dt, length(tz))
    attributes(dt) <- attr
  }
  secs <- C_local_time(as.POSIXct(dt), tz)
  out <- structure(secs, units = "secs", class = "difftime")
  units(out) <- units
  out
}

match_lengths <- function(x, y) {
  n.x <- length(x)
  n.y <- length(y)
  n.max <- max(n.x, n.y)
  n.min <- min(n.x, n.y)
  if (n.max %% n.min != 0L) {
    stop("longer object length is not a multiple of shorter object length")
  } else {
    if (n.x < n.y) {
      x <- rep(x, length.out = n.y)
    } else {
      y <- rep(y, length.out = n.x)
    }
  }
  list(x, y)
}

standardise_date_names <- function(x) {
  dates <- c("second", "minute", "hour", "mday", "wday", "yday", "day", "week", "month", "year", "tz")
  y <- gsub("(.)s$", "\\1", x)
  res <- dates[pmatch(y, dates, duplicates.ok = TRUE)]
  if (any(is.na(res))) {
    stop("Invalid unit name: ", paste(x[is.na(res)], collapse = ", "),
      call. = FALSE)
  }
  res
}

standardise_difftime_names <- function(x) {
  dates <- c("secs", "mins", "hours", "days", "weeks")
  y <- gsub("(.)s$", "\\1", x)
  y <- substr(y, 1, 3)
  res <- dates[pmatch(y, dates, duplicates.ok = TRUE)]
  if (any(is.na(res))) {
    stop("Invalid difftime name: ", paste(x[is.na(res)], collapse = ", "),
      call. = FALSE)
  }
  res
}

standardise_period_names <- function(x) {
  dates <- c("second", "minute", "hour", "day", "week", "month", "year",
             ## these ones are used for rounding only
             "bimonth", "quarter", "halfyear", "season")
  y <- gsub("(.)s$", "\\1", x)
  y <- substr(y, 1, 3)
  res <- dates[pmatch(y, dates)]
  if (any(is.na(res))) {
    stop("Invalid period name: ", paste(x[is.na(res)], collapse = ", "),
      call. = FALSE)
  }
  res
}

standardise_lt_names <- function(x) {
  if (length(x) == 0L)
    stop("No unit names supplied.")
  dates <- c("sec", "min", "hour", "day", "mday", "wday", "yday", "mon", "year", "tz")
  y <- gsub("(.)s$", "\\1", x)
  y <- substr(y, 1, 3)
  res <- dates[pmatch(y, dates)]
  if (any(is.na(res))) {
    stop("Invalid unit name: ", paste(x[is.na(res)], collapse = ", "),
         call. = FALSE)
  }
  res
}

## return list(n=nr_units, unit="unit_name")
parse_period_unit <- function(unit) {

  if (length(unit) > 1) {
    warning("Unit argument longer than 1. Taking first element.")
    unit <- unit[[1]]
  }

  p <- .Call(C_parse_period, as.character(unit))

  if (!is.na(p[[1]])) {

    period_units <- c("second", "minute", "hour", "day", "week", "month", "year")

    wp <- which(p > 0)
    if (length(wp) > 1) {
      ## Fractional units are actually supported but only when it leads to one
      ## final unit.
      stop("Cannot't parse heterogenuous or fractional units larger than one minute.")
    }

    list(n = p[wp], unit = period_units[wp])

  } else {
    ## this part is for backward compatibility and allows for bimonth, halfyear
    ## and quarter

    m <- regexpr(" *(?<n>[0-9.,]+)? *(?<unit>[^ \t\n]+)", unit[[1]], perl = T)
    if (m > 0) {
      ## should always match
      nms <- attr(m, "capture.names")
      nms <- nms[nzchar(nms)]
      start <- attr(m, "capture.start")
      end <- start + attr(m, "capture.length") - 1L
      n <- if (end[[1]] >= start[[1]]) {
             as.integer(substr(unit, start[[1]], end[[1]]))
           } else {
             1
           }
      unit <- substr(unit, start[[2]], end[[2]])
      list(n = n, unit = unit)
    } else {
      stop(sprintf("Invalid unit specification '%s'", unit))
    }

  }
}

date_to_posix <- function(date, tz = "UTC") {
  utc <- .POSIXct(unclass(date) * 86400, tz = "UTC")
  if (is_utc(tz)) utc
  else force_tz(utc, tz)
}

# UTC-equivalent timezones can be treated as UTC;
#   check grep('UTC|GMT', OlsonNames(), value = TRUE)
is_utc = function(tz) {
  utc_tz = c("UTC", "GMT", "Etc/UTC", "Etc/GMT", "GMT-0", "GMT+0", "GMT0")
  if (is.null(tz)) tz = Sys.timezone()
  return(tz %in% utc_tz)
}

# minimal custom str_sub function to replicate stringr::str_sub without the full dependency.
.str_sub <- function(x, start, end, replace_with = "") {

  # get the parts of the string to the left and right of the replacement.
  start.c = substr(x, 1, start - 1)
  end.c = substr(x, end + 1, nchar(x))

  # paste with replacement in the middle.
  x <- paste(start.c, replace_with, end.c, sep = "")

  x
}

is_verbose <- function() {
  isTRUE(getOption("lubridate.verbose"))
}

stop_incompatible_classes <- function(x, y, method) {
  stop(paste0(
    "Incompatible classes: <", is(x)[[1]], "> ", method, " <", is(y)[[1]], ">\n"
  ), call. = FALSE)
}

test_that("leap_year correctly identifies leap years", {
  x <- as.POSIXct("2009-08-03 12:01:59", tz = "UTC")
  y <- as.POSIXct("2008-08-03 12:01:59", tz = "UTC")

  expect_false(leap_year(x))
  expect_true(leap_year(y))
})

test_that("leap_year handles various date-time vectors", {
  x <- as.POSIXct(c("2008-08-03 12:01:59", "2009-08-03 12:01:59"), tz = "UTC")

  expect_equal(leap_year(x), c(TRUE, FALSE))
  expect_equal(leap_year(as.Date(x)), c(TRUE, FALSE))
  expect_equal(leap_year(as.POSIXlt(x)), c(TRUE, FALSE))
})

test_that("standardise_date_names handles repeated units", {
  expect_equal(
    standardise_date_names(c("seconds", "seconds")),
    c("second", "second")
  )
})

context("Utilities")

test_that("leap_year correctly identifies leap years", {
  x <- as.POSIXct("2009-08-03 12:01:59", tz = "UTC")
  y <- as.POSIXct("2008-08-03 12:01:59", tz = "UTC")

  expect_false(leap_year(x))
  expect_true(leap_year(y))
})

test_that("leap_year handles various date-time vectors", {
  x <- as.POSIXct(c("2008-08-03 12:01:59", "2009-08-03 12:01:59"), tz = "UTC")

  expect_equal(leap_year(x), c(TRUE, FALSE))
  expect_equal(leap_year(as.Date(x)), c(TRUE, FALSE))
  expect_equal(leap_year(as.POSIXlt(x)), c(TRUE, FALSE))
})

test_that("standardise_date_names handles repeated units", {
  expect_equal(
    standardise_date_names(c("seconds", "seconds")),
    c("second", "second")
  )
})

normalize_roll_dst <- function(roll_dst, roll = NULL, n_call = 1) {
  if (!is.null(roll)) {
    .deprecated_arg("roll", "1.8.4")
    roll <- as.logical(roll)
    roll_dst <- if (roll) c("boundary", "post") else c("NA", "post")
  }
  roll_dst
}

match_lengths <- function(x, y) {
  n.x <- length(x)
  n.y <- length(y)
  n.max <- max(n.x, n.y)
  n.min <- min(n.x, n.y)
  if (n.max %% n.min != 0L) {
    stop("longer object length is not a multiple of shorter object length")
  } else {
    if (n.x < n.y) {
      x <- rep(x, length.out = n.y)
    } else {
      y <- rep(y, length.out = n.x)
    }
  }
  list(x, y)
}

standardise_date_names <- function(x) {
  dates <- c("second", "minute", "hour", "mday", "wday", "yday", "day", "week", "month", "year", "tz")
  y <- gsub("(.)s$", "\\1", x)
  res <- dates[pmatch(y, dates, duplicates.ok = TRUE)]
  if (any(is.na(res))) {
    stop("Invalid unit name: ", paste(x[is.na(res)], collapse = ", "),
      call. = FALSE
    )
  }
  res
}

standardise_difftime_names <- function(x) {
  dates <- c("secs", "mins", "hours", "days", "weeks")
  y <- gsub("(.)s$", "\\1", x)
  y <- substr(y, 1, 3)
  res <- dates[pmatch(y, dates, duplicates.ok = TRUE)]
  if (any(is.na(res))) {
    stop("Invalid difftime name: ", paste(x[is.na(res)], collapse = ", "),
      call. = FALSE
    )
  }
  res
}

standardise_period_names <- function(x) {
  dates <- c(
    "second", "minute", "hour", "day", "week", "month", "year",
    ## these ones are used for rounding only
    "bimonth", "quarter", "halfyear", "season"
  )
  y <- gsub("(.)s$", "\\1", x)
  y <- substr(y, 1, 3)
  res <- dates[pmatch(y, dates)]
  if (any(is.na(res))) {
    stop("Invalid period name: ", paste(x[is.na(res)], collapse = ", "),
      call. = FALSE
    )
  }
  res
}

standardise_lt_names <- function(x) {
  if (length(x) == 0L) {
    stop("No unit names supplied.")
  }
  dates <- c("sec", "min", "hour", "day", "mday", "wday", "yday", "mon", "year", "tz")
  y <- gsub("(.)s$", "\\1", x)
  y <- substr(y, 1, 3)
  res <- dates[pmatch(y, dates)]
  if (any(is.na(res))) {
    stop("Invalid unit name: ", paste(x[is.na(res)], collapse = ", "),
      call. = FALSE
    )
  }
  res
}

## return list(n=nr_units, unit="unit_name")
parse_period_unit <- function(unit) {
  if (length(unit) > 1) {
    warning("Unit argument longer than 1. Taking first element.")
    unit <- unit[[1]]
  }

  p <- .Call(C_parse_period, as.character(unit))

  if (!is.na(p[[1]])) {
    period_units <- c("second", "minute", "hour", "day", "week", "month", "year")

    wp <- which(p > 0)
    if (length(wp) > 1) {
      ## Fractional units are actually supported but only when it leads to one
      ## final unit.
      stop("Cannot't parse heterogenuous or fractional units larger than one minute.")
    }

    list(n = p[wp], unit = period_units[wp])
  } else {
    ## this part is for backward compatibility and allows for bimonth, halfyear
    ## and quarter

    m <- regexpr(" *(?<n>[0-9.,]+)? *(?<unit>[^ \t\n]+)", unit[[1]], perl = T)
    if (m > 0) {
      ## should always match
      nms <- attr(m, "capture.names")
      nms <- nms[nzchar(nms)]
      start <- attr(m, "capture.start")
      end <- start + attr(m, "capture.length") - 1L
      n <- if (end[[1]] >= start[[1]]) {
        as.integer(substr(unit, start[[1]], end[[1]]))
      } else {
        1
      }
      unit <- substr(unit, start[[2]], end[[2]])
      list(n = n, unit = unit)
    } else {
      stop(sprintf("Invalid unit specification '%s'", unit))
    }
  }
}

date_to_posix <- function(date, tz = "UTC") {
  utc <- .POSIXct(unclass(date) * 86400, tz = "UTC")
  if (is_utc(tz)) {
    utc
  } else {
    force_tz(utc, tz)
  }
}

# UTC-equivalent timezones can be treated as UTC;
#   check grep('UTC|GMT', OlsonNames(), value = TRUE)
is_utc <- function(tz) {
  utc_tz <- c("UTC", "GMT", "Etc/UTC", "Etc/GMT", "GMT-0", "GMT+0", "GMT0")
  if (is.null(tz)) tz <- Sys.timezone()
  return(tz %in% utc_tz)
}

# minimal custom str_sub function to replicate stringr::str_sub without the full dependency.
.str_sub <- function(x, start, end, replace_with = "") {

  # get the parts of the string to the left and right of the replacement.
  start.c <- substr(x, 1, start - 1)
  end.c <- substr(x, end + 1, nchar(x))

  # paste with replacement in the middle.
  x <- paste(start.c, replace_with, end.c, sep = "")

  x
}

is_verbose <- function() {
  isTRUE(getOption("lubridate.verbose"))
}

stop_incompatible_classes <- function(x, y, method) {
  stop(paste0(
    "Incompatible classes: <", is(x)[[1]], "> ", method, " <", is(y)[[1]], ">\n"
  ), call. = FALSE)
}

stop_incompatible_size <- function(x, y, msg = NULL) {
  lx <- length(x)
  ly <- length(y)
  if (!(lx == ly || lx == 1 || ly == 1)) {
    if (is.null(msg)) {
      msg <- sprintf("Incompatible size of '%s'(%s) and '%s'(%s)",
        deparse(substitute(x)), lx, deparse(substitute(y)), ly
      )
    }
    stop(msg, call. = FALSE)
  }
}

#' Format in ISO8601 character format
#'
#' @param x An object to convert to ISO8601 character format.
#' @param usetz Include the time zone in the formatting (of outputs including
#'   time; date outputs never include time zone information).
#' @param precision The amount of precision to represent with substrings of
#'   "ymdhms", as "y"ear, "m"onth, "d"ay, "h"our, "m"inute, and "s"econd. (e.g.
#'   "ymdhm" would show precision through minutes.  When \code{NULL}, full
#'   precision for the object is shown.
#' @param ... Additional arguments to methods.
#' @return A character vector of ISO8601-formatted text.
#' @references \url{https://en.wikipedia.org/wiki/ISO_8601}
#' @examples
#' format_ISO8601(as.Date("02-01-2018", format="%m-%d-%Y"))
#' format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz="EST"), usetz=TRUE)
#' format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz="EST"), precision="ymdhm")
#' @aliases format_ISO8601,Date-method format_ISO8601,POSIXt-method format_ISO8601,Interval-method
#' format_ISO8601,Duration-method format_ISO8601,Period-method
#' @export
setGeneric(name = "format_ISO8601",
           def = function(x, usetz=FALSE, precision=NULL, ...) standardGeneric("format_ISO8601"))

#' @export
setMethod("format_ISO8601", signature="Date",
          function(x, usetz=FALSE, precision=NULL, ...) {
            precision_format <-
              format_ISO8601_precision_check(precision=precision, max_precision="ymd", usetz=FALSE)
            as.character(x, format=precision_format, usetz=FALSE)
          })

#' @export
setMethod("format_ISO8601", signature="POSIXt",
          function(x, usetz=FALSE, precision=NULL, ...) {
            precision_format <-
              format_ISO8601_precision_check(precision=precision, max_precision="ymdhms", usetz=usetz)
            # Note that the usetz argument to as.character is always FALSE because the time zone is handled in the precision argument.
            as.character(x, format=precision_format, usetz=FALSE)
          })

#' @export
setMethod("format_ISO8601", signature="Interval",
          function(x, usetz=FALSE, precision=NULL, ...) {
            precision_format <-
              format_ISO8601_precision_check(precision=precision, max_precision="ymdhms", usetz=usetz)
            # Note that the usetz argument to as.character is always FALSE because the time zone is handled in the precision argument.
            sprintf("%s/%s",
                    as.character(x@start, format=precision_format, usetz=FALSE),
                    as.character(x@start + x@.Data, format=precision_format, usetz=FALSE))
          })

#' @export
setMethod("format_ISO8601", signature="Duration",
          function(x, usetz=FALSE, precision=NULL, ...) {
            if (!is.null(precision)) {
              warning("precision is not used for Duration objects")
            }
            sprintf("PT%sS", format(x@.Data))
          })

#' @export
setMethod("format_ISO8601", signature="Period",
          function(x, usetz=FALSE, precision=NULL, ...) {
            if (!is.null(precision)) {
              warning("precision is not used for Period objects")
            }
            date_part <-
              paste0(
                ifelse(x@year != 0,
                       paste0(x@year, "Y"),
                       ""),
                ifelse(x@month != 0,
                       paste0(x@month, "M"),
                       ""),
                ifelse(x@day != 0,
                       paste0(x@day, "D"),
                       ""))
            time_part <-
              paste0(
                ifelse(x@hour != 0,
                       paste0(x@hour, "H"),
                       ""),
                ifelse(x@minute != 0,
                       paste0(x@minute, "M"),
                       ""),
                ifelse(x@.Data != 0,
                       paste0(x@.Data, "S"),
                       ""))
            mask_neither <- (nchar(date_part) == 0) & (nchar(time_part) == 0)
            time_part[mask_neither] <- "0S"
            ifelse(nchar(time_part),
                   paste0("P", date_part, "T", time_part),
                   paste0("P", date_part))
          })

ISO8601_precision_map <-
  list(y="%Y",
       ym="%Y-%m",
       ymd="%Y-%m-%d",
       ymdh="%Y-%m-%dT%H",
       ymdhm="%Y-%m-%dT%H:%M",
       ymdhms="%Y-%m-%dT%H:%M:%S")

#' Provide a format for ISO8601 dates and times with the requested precision.
#'
#' @param precision The amount of precision to represent with substrings of
#'   "ymdhms", as "y"ear, "m"onth, "d"ay, "h"our, "m"inute, and
#'   "s"econd. (e.g. "ymdhm" would show precision through minutes. When NULL,
#'   `max_precision` is returned.  When `precision` is more precise than
#'   `max_precision`, a warning is given and `max_precision` is returned.
#' @param max_precision The maximum precision allowed to be output.
#' @param usetz Include the timezone in the output format
#' @keywords internal
format_ISO8601_precision_check <- function(precision, max_precision, usetz=FALSE) {
  if (!(max_precision %in% names(ISO8601_precision_map))) {
    stop("Invalid value for max_precision provided: ", max_precision)
  }
  if (is.null(precision)) {
    precision <- max_precision
  }
  if (length(precision) != 1) {
    stop("precision must be a scalar")
  }
  if (nchar(precision) > nchar(max_precision)) {
    warning("More precision requested (", precision, ") ",
            "than allowed (", max_precision, ") for this format.  ",
            "Using maximum allowed precision.")
    precision <- max_precision
  }
  if (is.null(ret <- ISO8601_precision_map[[precision]])) {
    stop("Invalid value for precision provided: ", precision)
  }
  if (usetz) {
    ret <- paste0(ret, "%z")
  }
  ret
}

test_that("Formatting a date works", {
  expect_equal(format_ISO8601(as.Date("02-01-2018", format = "%m-%d-%Y")),
    "2018-02-01",
    info = "Standard date formatting works"
  )
  expect_equal(format_ISO8601(as.Date("02-01-2018", format = "%m-%d-%Y"), usetz = TRUE),
    "2018-02-01",
    info = "usetz is ignored with dates"
  )
  expect_equal(format_ISO8601(as.Date("02-01-2018", format = "%m-%d-%Y"), usetz="Z"),
               "2018-02-01",
               info="usetz='Z' is ignored with dates")
  expect_equal(format_ISO8601(as.Date("02-01-2018", format = "%m-%d-%Y"), precision = "y"),
    "2018",
    info = "precision is respected (y)"
  )
  expect_equal(format_ISO8601(as.Date("02-01-2018", format = "%m-%d-%Y"), precision = "ym"),
    "2018-02",
    info = "precision is respected (ym)"
  )
  expect_equal(format_ISO8601(as.Date("02-01-2018", format = "%m-%d-%Y"), precision = "ymd"),
    "2018-02-01",
    info = "precision is respected (ymd)"
  )
  expect_warning(
    expect_equal(format_ISO8601(as.Date("02-01-2018", format = "%m-%d-%Y"), precision = "ymdh"),
      "2018-02-01",
      info = "precision is respected (ymdh)"
    ),
    info = "Requesting too much precision gives a warning and the maximum precision."
  )
})

test_that("Formatting a datetime works", {
  expect_equal(format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz = "EST")),
    "2018-02-01T03:04:05",
    info = "Standard datetime formatting works"
  )
  expect_equal(format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz = "EST"), usetz = TRUE),
    "2018-02-01T03:04:05-0500",
    info = "usetz is respected with datetimes"
  )
  expect_equal(format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz="EST"), usetz="Z"),
               "2018-02-01T08:04:05Z",
               info="usetz='Z' is respected with datetimes")
  expect_equal(format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz = "EST"), precision = "y"),
    "2018",
    info = "precision is respected (y)"
  )
  # Uncertain if this is the best result; including a test so that if it changes
  # the change is caught.
  expect_equal(format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz = "EST"), precision = "y", usetz = TRUE),
    "2018-0500",
    info = "precision is respected (y) with timezone"
  )
  expect_equal(format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz = "EST"), precision = "ymdhm"),
    "2018-02-01T03:04",
    info = "precision is respected (ymdhm)"
  )
})

test_that("Formatting a Duration works", {
  expect_equal(format_ISO8601(duration(20, units = "seconds")),
    "PT20S",
    info = "Standard duration formatting works"
  )
  expect_equal(format_ISO8601(duration(20, units = "minutes")),
    paste0("PT", 20 * 60, "S"),
    info = "Duration always formats as seconds to ensure precision."
  )
  expect_warning(format_ISO8601(duration(20, units = "minutes"), precision = "y"),
    regexp = "precision is not used for Duration objects",
    fixed = TRUE
  )
})

test_that("Formatting a Period works", {
  expect_equal(format_ISO8601(period(1, units = "seconds")),
    "PT1S",
    info = "A period of 1 second works"
  )
  expect_equal(format_ISO8601(period(1, units = "minute")),
    "PT1M",
    info = "A period of 1 minute works"
  )
  expect_equal(format_ISO8601(period(1, units = "hour")),
    "PT1H",
    info = "A period of 1 hour works"
  )
  expect_equal(format_ISO8601(period(1, units = "day")),
    "P1D",
    info = "A period of 1 day works"
  )
  expect_equal(format_ISO8601(period(1, units = "month")),
    "P1M",
    info = "A period of 1 month works"
  )
  expect_equal(format_ISO8601(period(1, units = "year")),
    "P1Y",
    info = "A period of 1 year works"
  )

  expect_equal(format_ISO8601(period(1, units = "seconds") + period(1, units = "year")),
    "P1YT1S",
    info = "More than one unit works"
  )

  expect_equal(format_ISO8601(period(0, units = "seconds")),
    "PT0S",
    info = "zero-length period (seconds specified)"
  )

  expect_equal(format_ISO8601(period(0, units = "year")),
    "PT0S",
    info = "zero-length period (years specified)"
  )
})

test_that("Formatting an Interval works", {
  expect_equal(format_ISO8601(interval(start = "2018-02-01", end = "2018-03-04")),
    "2018-02-01T00:00:00/2018-03-04T00:00:00",
    info = "Standard interval formatting works"
  )
  expect_equal(format_ISO8601(interval(start = "2018-02-01", end = "2018-03-04"), precision = "ymd"),
    "2018-02-01/2018-03-04",
    info = "interval formatting respects precision (ymd)"
  )
  expect_equal(format_ISO8601(interval(start = "2018-02-01", end = "2018-03-04", tzone = "EST"), usetz = TRUE),
    "2018-02-01T00:00:00-0500/2018-03-04T00:00:00-0500",
    info = "interval formatting respects timezone"
  )
  expect_equal(format_ISO8601(interval(start="2018-02-01", end="2018-03-04", tzone="EST"), usetz="Z"),
               "2018-02-01T05:00:00Z/2018-03-04T05:00:00Z",
               info="interval formatting respects usetz='Z'")
})

test_that("formatting precision provides accurate format strings", {
  expect_equal(format_ISO8601_precision_check(precision = "y", max_precision = "ymd", usetz = FALSE),
    "%Y",
    info = "simple formatting works"
  )
  expect_equal(format_ISO8601_precision_check(precision = "y", max_precision = "ymd", usetz = TRUE),
    "%Y%z",
    info = "timezone formatting works"
  )
  expect_equal(format_ISO8601_precision_check(precision = NULL, max_precision = "ymd", usetz = FALSE),
    "%Y-%m-%d",
    info = "NULL yields max_precision"
  )
  expect_error(format_ISO8601_precision_check(precision = "foo", max_precision = "ymd", usetz = FALSE),
    regexp = "Invalid value for precision provided: foo",
    fixed = TRUE,
    info = "invalid precision is an error"
  )
  expect_error(format_ISO8601_precision_check(precision = NULL, max_precision = "foo", usetz = FALSE),
    regexp = "Invalid value for max_precision provided: foo",
    fixed = TRUE,
    info = "invalid max_precision is an error"
  )
  expect_error(format_ISO8601_precision_check(precision = "ymd", max_precision = "foo", usetz = FALSE),
    regexp = "Invalid value for max_precision provided: foo",
    fixed = TRUE,
    info = "invalid max_precision is an error (even if it's not used)"
  )
  expect_error(format_ISO8601_precision_check(precision = c("ymd", "ymdh"), max_precision = "ymdh", usetz = FALSE),
    regexp = "precision must be a scalar"
  )
  expect_warning(
    expect_equal(format_ISO8601_precision_check(precision = "ymdh", max_precision = "ymd", usetz = FALSE),
      "%Y-%m-%d",
      info = "max_precision is respected"
    ),
    regexp = "More precision requested (ymdh) than allowed (ymd) for this format.  Using maximum allowed precision.",
    fixed = TRUE,
    info = "max_precision is warned"
  )
})

context("format_ISO8601")

test_that("Formatting a date works", {
  expect_equal(format_ISO8601(as.Date("02-01-2018", format="%m-%d-%Y")),
               "2018-02-01",
               info="Standard date formatting works")
  expect_equal(format_ISO8601(as.Date("02-01-2018", format="%m-%d-%Y"), usetz=TRUE),
               "2018-02-01",
               info="usetz is ignored with dates")
  expect_equal(format_ISO8601(as.Date("02-01-2018", format="%m-%d-%Y"), precision="y"),
               "2018",
               info="precision is respected (y)")
  expect_equal(format_ISO8601(as.Date("02-01-2018", format="%m-%d-%Y"), precision="ym"),
               "2018-02",
               info="precision is respected (ym)")
  expect_equal(format_ISO8601(as.Date("02-01-2018", format="%m-%d-%Y"), precision="ymd"),
               "2018-02-01",
               info="precision is respected (ymd)")
  expect_warning(
    expect_equal(format_ISO8601(as.Date("02-01-2018", format="%m-%d-%Y"), precision="ymdh"),
                 "2018-02-01",
                 info="precision is respected (ymdh)"),
    info="Requesting too much precision gives a warning and the maximum precision.")
})

test_that("Formatting a datetime works", {
  expect_equal(format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz="EST")),
               "2018-02-01T03:04:05",
               info="Standard datetime formatting works")
  expect_equal(format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz="EST"), usetz=TRUE),
               "2018-02-01T03:04:05-0500",
               info="usetz is respected with datetimes")
  expect_equal(format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz="EST"), precision="y"),
               "2018",
               info="precision is respected (y)")
  # Uncertain if this is the best result; including a test so that if it changes
  # the change is caught.
  expect_equal(format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz="EST"), precision="y", usetz=TRUE),
               "2018-0500",
               info="precision is respected (y) with timezone")
  expect_equal(format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz="EST"), precision="ymdhm"),
               "2018-02-01T03:04",
               info="precision is respected (ymdhm)")
})

test_that("Formatting a Duration works", {
  expect_equal(format_ISO8601(duration(20, units="seconds")),
               "PT20S",
               info="Standard duration formatting works")
  expect_equal(format_ISO8601(duration(20, units="minutes")),
               paste0("PT", 20*60,"S"),
               info="Duration always formats as seconds to ensure precision.")
  expect_warning(format_ISO8601(duration(20, units="minutes"), precision="y"),
                 regexp="precision is not used for Duration objects",
                 fixed=TRUE)
})

test_that("Formatting a Period works", {
  expect_equal(format_ISO8601(period(1, units="seconds")),
               "PT1S",
               info="A period of 1 second works")
  expect_equal(format_ISO8601(period(1, units="minute")),
               "PT1M",
               info="A period of 1 minute works")
  expect_equal(format_ISO8601(period(1, units="hour")),
               "PT1H",
               info="A period of 1 hour works")
  expect_equal(format_ISO8601(period(1, units="day")),
               "P1D",
               info="A period of 1 day works")
  expect_equal(format_ISO8601(period(1, units="month")),
               "P1M",
               info="A period of 1 month works")
  expect_equal(format_ISO8601(period(1, units="year")),
               "P1Y",
               info="A period of 1 year works")

  expect_equal(format_ISO8601(period(1, units="seconds") + period(1, units="year")),
               "P1YT1S",
               info="More than one unit works")

  expect_equal(format_ISO8601(period(0, units="seconds")),
               "PT0S",
               info="zero-length period (seconds specified)")

  expect_equal(format_ISO8601(period(0, units="year")),
               "PT0S",
               info="zero-length period (years specified)")
})

test_that("Formatting an Interval works", {
  expect_equal(format_ISO8601(interval(start="2018-02-01", end="2018-03-04")),
               "2018-02-01T00:00:00/2018-03-04T00:00:00",
               info="Standard interval formatting works")
  expect_equal(format_ISO8601(interval(start="2018-02-01", end="2018-03-04"), precision="ymd"),
               "2018-02-01/2018-03-04",
               info="interval formatting respects precision (ymd)")
  expect_equal(format_ISO8601(interval(start="2018-02-01", end="2018-03-04", tzone="EST"), usetz=TRUE),
               "2018-02-01T00:00:00-0500/2018-03-04T00:00:00-0500",
               info="interval formatting respects timezone")
})

test_that("formatting precision provides accurate format strings", {
  expect_equal(format_ISO8601_precision_check(precision="y", max_precision="ymd", usetz=FALSE),
               "%Y",
               info="simple formatting works")
  expect_equal(format_ISO8601_precision_check(precision="y", max_precision="ymd", usetz=TRUE),
               "%Y%z",
               info="timezone formatting works")
  expect_equal(format_ISO8601_precision_check(precision=NULL, max_precision="ymd", usetz=FALSE),
               "%Y-%m-%d",
               info="NULL yields max_precision")
  expect_error(format_ISO8601_precision_check(precision="foo", max_precision="ymd", usetz=FALSE),
               regexp="Invalid value for precision provided: foo",
               fixed=TRUE,
               info="invalid precision is an error")
  expect_error(format_ISO8601_precision_check(precision=NULL, max_precision="foo", usetz=FALSE),
               regexp="Invalid value for max_precision provided: foo",
               fixed=TRUE,
               info="invalid max_precision is an error")
  expect_error(format_ISO8601_precision_check(precision="ymd", max_precision="foo", usetz=FALSE),
               regexp="Invalid value for max_precision provided: foo",
               fixed=TRUE,
               info="invalid max_precision is an error (even if it's not used)")
  expect_error(format_ISO8601_precision_check(precision=c("ymd", "ymdh"), max_precision="ymdh", usetz=FALSE),
                 regexp="precision must be a scalar")
  expect_warning(
    expect_equal(format_ISO8601_precision_check(precision="ymdh", max_precision="ymd", usetz=FALSE),
                 "%Y-%m-%d",
                 info="max_precision is respected"),
    regexp="More precision requested (ymdh) than allowed (ymd) for this format.  Using maximum allowed precision.",
    fixed=TRUE,
    info="max_precision is warned")
})

#' Format in ISO8601 character format
#'
#' @param x An object to convert to ISO8601 character format.
#' @inheritParams format_ISO8601_precision_check
#' @param ... Additional arguments to methods.
#' @return A character vector of ISO8601-formatted text.
#' @references \url{https://en.wikipedia.org/wiki/ISO_8601}
#' @examples
#' format_ISO8601(as.Date("02-01-2018", format = "%m-%d-%Y"))
#' format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz = "EST"), usetz = TRUE)
#' format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz = "EST"), precision = "ymdhm")
#' @aliases format_ISO8601,Date-method format_ISO8601,POSIXt-method format_ISO8601,Interval-method
#' format_ISO8601,Duration-method format_ISO8601,Period-method
#' @export
setGeneric(
  name = "format_ISO8601",
  def = function(x, usetz = FALSE, precision = NULL, ...) standardGeneric("format_ISO8601")
)

#' @export
setMethod("format_ISO8601",
  signature = "Date",
  function(x, usetz = FALSE, precision = NULL, ...) {
    precision_format <-
      format_ISO8601_precision_check(precision = precision, max_precision = "ymd", usetz = FALSE)
    format(x, format = precision_format, usetz = FALSE)
  }
)

#' @export
setMethod("format_ISO8601",
  signature = "POSIXt",
  function(x, usetz = FALSE, precision = NULL, ...) {
    precision_format <-
      format_ISO8601_precision_check(
        precision = precision, max_precision = "ymdhms", usetz = usetz)
    if (identical(usetz, "Z")) {
      x <- with_tz(x, tzone="UTC")
    }
    # usetz argument is FALSE because the time zone is handled in the precision argument.
    format(x, format = precision_format, usetz = FALSE)
  }
)

#' @export
setMethod("format_ISO8601",
  signature = "Interval",
  function(x, usetz = FALSE, precision = NULL, ...) {
    precision_format <-
      format_ISO8601_precision_check(
        precision = precision, max_precision = "ymdhms", usetz = usetz)
    x_start <- x@start
    x_end <- x@start + x@.Data
    if (identical(usetz, "Z")) {
      x_start <- with_tz(x_start, tzone="UTC")
      x_end <- with_tz(x_end, tzone="UTC")
    }
    # usetz argument is FALSE because the time zone is handled in the precision argument.
    sprintf(
      "%s/%s",
      format(x_start, format = precision_format, usetz = FALSE),
      format(x_end, format = precision_format, usetz = FALSE)
    )
  }
)

#' @export
setMethod("format_ISO8601",
  signature = "Duration",
  function(x, usetz = FALSE, precision = NULL, ...) {

    if (!is.null(precision)) {
      warning("precision is not used for Duration objects")
    }
    sprintf("PT%sS", format(x@.Data))
  }
)

#' @export
setMethod("format_ISO8601",
  signature = "Period",
  function(x, usetz = FALSE, precision = NULL, ...) {
    if (!is.null(precision)) {
      warning("precision is not used for Period objects")
    }
    date_part <-
      paste0(
        ifelse(x@year != 0,
          paste0(x@year, "Y"),
          ""
        ),
        ifelse(x@month != 0,
          paste0(x@month, "M"),
          ""
        ),
        ifelse(x@day != 0,
          paste0(x@day, "D"),
          ""
        )
      )
    time_part <-
      paste0(
        ifelse(x@hour != 0,
          paste0(x@hour, "H"),
          ""
        ),
        ifelse(x@minute != 0,
          paste0(x@minute, "M"),
          ""
        ),
        ifelse(x@.Data != 0,
          paste0(x@.Data, "S"),
          ""
        )
      )
    mask_neither <- (nchar(date_part) == 0) & (nchar(time_part) == 0)
    time_part[mask_neither] <- "0S"
    ifelse(nchar(time_part),
      paste0("P", date_part, "T", time_part),
      paste0("P", date_part)
    )
  }
)

ISO8601_precision_map <-
  list(
    y = "%Y",
    ym = "%Y-%m",
    ymd = "%Y-%m-%d",
    ymdh = "%Y-%m-%dT%H",
    ymdhm = "%Y-%m-%dT%H:%M",
    ymdhms = "%Y-%m-%dT%H:%M:%S"
  )

#' Provide a format for ISO8601 dates and times with the requested precision.
#'
#' @param usetz Include the time zone in the formatting.  If
#'   \code{usetz} is \code{TRUE}, the time zone is included. If
#'   \code{usetz} is \code{"Z"}, the time is converted to "UTC" and
#'   the time zone is indicated with "Z" ISO8601 notation.
#' @param precision The amount of precision to represent with
#'   substrings of "ymdhms", as "y"ear, "m"onth, "d"ay, "h"our,
#'   "m"inute, and "s"econd. (e.g. "ymdhm" would show precision
#'   through minutes.  When \code{NULL}, full precision for the object
#'   is shown.
#' @param max_precision The maximum precision allowed to be output.
#' @keywords internal
format_ISO8601_precision_check <- function(precision, max_precision, usetz = FALSE) {
  if (!(max_precision %in% names(ISO8601_precision_map))) {
    stop("Invalid value for max_precision provided: ", max_precision)
  }
  if (is.null(precision)) {
    precision <- max_precision
  }
  if (length(precision) != 1) {
    stop("precision must be a scalar")
  }
  if (nchar(precision) > nchar(max_precision)) {
    warning(
      "More precision requested (", precision, ") ",
      "than allowed (", max_precision, ") for this format.  ",
      "Using maximum allowed precision."
    )
    precision <- max_precision
  }
  if (is.null(ret <- ISO8601_precision_map[[precision]])) {
    stop("Invalid value for precision provided: ", precision)
  }
  if (identical(usetz, "Z")) {
    ret <- paste0(ret, "Z")
  } else if (usetz) {
    ret <- paste0(ret, "%z")
  }
  ret
}

# nocov start

.onLoad <- function(libname, pkgname) {
  tzdir_set()

  on_package_load("vctrs", {
    register_s3_method("vctrs", "vec_proxy", "Period")
    register_s3_method("vctrs", "vec_restore", "Period")
    register_s3_method("vctrs", "vec_ptype2", "Period.Period")
    register_s3_method("vctrs", "vec_cast", "Period.Period")

    register_s3_method("vctrs", "vec_proxy", "Duration")
    register_s3_method("vctrs", "vec_restore", "Duration")
    register_s3_method("vctrs", "vec_ptype2", "Duration.Duration")
    register_s3_method("vctrs", "vec_ptype2", "Duration.difftime")
    register_s3_method("vctrs", "vec_ptype2", "difftime.Duration")
    register_s3_method("vctrs", "vec_cast", "Duration.Duration")
    register_s3_method("vctrs", "vec_cast", "Duration.difftime")
    register_s3_method("vctrs", "vec_cast", "difftime.Duration")

    register_s3_method("vctrs", "vec_proxy", "Interval")
    register_s3_method("vctrs", "vec_restore", "Interval")
    register_s3_method("vctrs", "vec_ptype2", "Interval.Interval")
    register_s3_method("vctrs", "vec_cast", "Interval.Interval")
  })
}

register_s3_method <- function(pkg, generic, class, fun = NULL) {
  stopifnot(is.character(pkg), length(pkg) == 1)
  stopifnot(is.character(generic), length(generic) == 1)
  stopifnot(is.character(class), length(class) == 1)

  if (is.null(fun)) {
    fun <- get(paste0(generic, ".", class), envir = parent.frame())
  } else {
    stopifnot(is.function(fun))
  }

  if (pkg %in% loadedNamespaces()) {
    registerS3method(generic, class, fun, envir = asNamespace(pkg))
  }

  # Always register hook in case package is later unloaded & reloaded
  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      registerS3method(generic, class, fun, envir = asNamespace(pkg))
    }
  )
}

on_package_load <- function(pkg, expr) {
  if (isNamespaceLoaded(pkg)) {
    expr
  } else {
    thunk <- function(...) expr
    setHook(packageEvent(pkg, "onLoad"), thunk)
  }
}

# nocov end

# nocov start

.onLoad <- function(libname, pkgname) {
  ## CCTZ needs zoneinfo. On windows we set it to R's own zoneinfo. On unix like
  ## it's in "/usr/share/zoneinfo" where CCTZ looks by default. On some systems
  ## (solaris, osx) it might be in a different location. So, help ourselves by
  ## setting the TZDIR env var, but only if it's not already set.

  ## adapted from Syz.timezone and OlsonNames function
  .find_tzdir <- function() {
    ## Initialize Sys.timezone() cache to avoid resetting TZDIR at a later stage.
    ## As of R4.0.3 Sys.timezone() intrusively sets TZDIR to non path values.
    Sys.timezone()
    if (.Platform$OS.type == "windows")
      return(file.path(R.home("share"), "zoneinfo"))
    tzdirs <- c("/usr/share/zoneinfo",
                "/usr/share/lib/zoneinfo",
                "/usr/lib/zoneinfo",
                "/usr/local/etc/zoneinfo",
                "/etc/zoneinfo",
                "/usr/etc/zoneinfo",
                "/usr/share/zoneinfo.default",
                "/var/db/timezone/zoneinfo",
                file.path(R.home("share"), "zoneinfo"))
    tzdirs <- tzdirs[file.exists(tzdirs)]
    if (length(tzdirs)) tzdirs[[1]]
    else NULL
  }

  tzdir <- Sys.getenv("TZDIR")
  if (tzdir == "internal") {
    Sys.setenv(TZDIR = file.path(R.home("share"), "zoneinfo"))
  } else if (tzdir == "macOS") {
    if (!is.null(dir <- .find_tzdir()))
      Sys.setenv(TZDIR = dir)
  } else if (tzdir == "") {
    # not checking for /usr/share/zoneinfo which is the default in CCTZ
    if (!file.exists("/usr/share/zoneinfo")) {
      if (!is.null(dir <- .find_tzdir()))
        Sys.setenv(TZDIR = dir)
    }
  }

  on_package_load("vctrs", {
    register_s3_method("vctrs", "vec_proxy", "Period")
    register_s3_method("vctrs", "vec_proxy_compare", "Period")
    register_s3_method("vctrs", "vec_proxy_equal", "Period")
    register_s3_method("vctrs", "vec_restore", "Period")
    register_s3_method("vctrs", "vec_ptype2", "Period.Period")
    register_s3_method("vctrs", "vec_cast", "Period.Period")

    register_s3_method("vctrs", "vec_proxy", "Duration")
    register_s3_method("vctrs", "vec_proxy_compare", "Duration")
    register_s3_method("vctrs", "vec_proxy_equal", "Duration")
    register_s3_method("vctrs", "vec_restore", "Duration")
    register_s3_method("vctrs", "vec_ptype2", "Duration.Duration")
    register_s3_method("vctrs", "vec_ptype2", "Duration.difftime")
    register_s3_method("vctrs", "vec_ptype2", "difftime.Duration")
    register_s3_method("vctrs", "vec_cast", "Duration.Duration")
    register_s3_method("vctrs", "vec_cast", "Duration.difftime")
    register_s3_method("vctrs", "vec_cast", "difftime.Duration")

    register_s3_method("vctrs", "vec_proxy", "Interval")
    register_s3_method("vctrs", "vec_proxy_compare", "Interval")
    register_s3_method("vctrs", "vec_proxy_equal", "Interval")
    register_s3_method("vctrs", "vec_restore", "Interval")
    register_s3_method("vctrs", "vec_ptype2", "Interval.Interval")
    register_s3_method("vctrs", "vec_cast", "Interval.Interval")
  })
}

register_s3_method <- function(pkg, generic, class, fun = NULL) {
  stopifnot(is.character(pkg), length(pkg) == 1)
  stopifnot(is.character(generic), length(generic) == 1)
  stopifnot(is.character(class), length(class) == 1)

  if (is.null(fun)) {
    fun <- get(paste0(generic, ".", class), envir = parent.frame())
  } else {
    stopifnot(is.function(fun))
  }

  if (pkg %in% loadedNamespaces()) {
    registerS3method(generic, class, fun, envir = asNamespace(pkg))
  }

  # Always register hook in case package is later unloaded & reloaded
  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      registerS3method(generic, class, fun, envir = asNamespace(pkg))
    }
  )
}

on_package_load <- function(pkg, expr) {
  if (isNamespaceLoaded(pkg)) {
    expr
  } else {
    thunk <- function(...) expr
    setHook(packageEvent(pkg, "onLoad"), thunk)
  }
}

#nocov end

setOldClass("Date")

#' Various date utilities
#'
#' [Date()] mirrors primitive contructors in base R ([double()], [character()]
#' etc.)
#'
#' @rdname date_utils
#' @param x an R object
#' @param length A non-negative number specifying the desired length. Supplying
#'   an argument of length other than one is an error.
#' @seealso [is.instant()], [is.timespan()], [is.POSIXt()], [POSIXct()]
#' @keywords logic chron
#' @examples
#' is.Date(as.Date("2009-08-03")) # TRUE
#' is.Date(difftime(now() + 5, now())) # FALSE
#' @export
is.Date <- function(x) is(x, "Date")

#' @rdname date_utils
#' @export
Date <- function(length = 0L) {
  structure(rep.int(NA_real_, length), class = "Date")
}

#' @rdname date_utils
#' @export
NA_Date_ <- structure(NA_real_, class = "Date")

.recursive_date_unclass <- function(x) {
  if (length(x) == 0) {
    Date()
  } else if (is.recursive(x)) {
    lapply(x, .recursive_date_unclass)
  } else {
    as_date(x)
  }
}

#' @method c Date
c.Date <- function(..., recursive = FALSE) {
  structure(c(unlist(lapply(list(...), .recursive_date_unclass))),
    class = "Date"
  )
}

evalqOnLoad({
  registerS3method("c", "Date", c.Date)
})

#' @name hidden_aliases
#' @aliases
#'   day<-,Date-method hour<-,Date-method minute<-,Date-method
#'   month<-,Date-method second<-,Date-method year<-,Date-method
NULL

context("Dates")

test_that("is.Date works as expected", {
  expect_false(is.Date(234))
  expect_false(is.Date(as.POSIXct("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.Date(as.POSIXlt("2008-08-03 13:01:59", tz = "UTC")))
  expect_true(is.Date(Sys.Date()))
  expect_false(is.Date(minutes(1)))
  expect_false(is.Date(dminutes(1)))
  expect_false(is.Date(interval(
    as.POSIXct("2008-08-03 13:01:59", tz = "UTC"),
    as.POSIXct("2009-08-03 13:01:59", tz = "UTC"))))
})

test_that("as_date works", {
  dt1 <- as.POSIXct("2010-08-03 00:59:59.23")
  dt2 <- as.POSIXct("2010-08-03 00:59:59.23", tz = "Europe/London")
  dt3 <- as.POSIXct("2010-11-03 00:59:59.23")
  dt4 <- as.POSIXct("2010-11-03 00:59:59.23", tz = "Europe/London")
  expect_equal(as_date(dt1), as_date("2010-08-03"))
  expect_equal(as_date(dt2), as_date("2010-08-03"))
  expect_equal(as_date(dt3), as_date("2010-11-03"))
  expect_equal(as_date(dt4), as_date("2010-11-03"))
  expect_equal(as_date(dt1), as.Date("2010-08-03"))
  expect_equal(as_date(dt2), as.Date("2010-08-03"))
  expect_equal(as_date(dt3), as.Date("2010-11-03"))
  expect_equal(as_date(dt4), as.Date("2010-11-03"))
  expect_equal(as_date(10), ymd("1970-01-11"))

  expect_equal(as_date("2010-08-03 00:59:59.23"), as_date("2010-08-03"))
  expect_equal(as_date("2010-11-03 00:59:59.23"), as_date("2010-11-03"))

  ## tz is ignored
  expect_equal(as_date("2010-08-03 00:59:59.23"), as_date("2010-08-03"))

  ## can supply custom format
  expect_equal(as_date("03/04/2015", format = "%d/%m/%Y"), as_date("2015-04-03"))

  ## Zulu time is part of the instant spec, so we compute on the instant object
  ## and not on the textual representation.
  expect_equal(as_date("2010-08-03 00:59:59.23Z-02"), as_date("2010-08-03"))
  expect_equal(as_date("2010-08-03 00:59:59.23Z+02"), as_date("2010-08-02"))
  expect_equal(as_date("2010-08-03 00:59:59.23Z+08"), as_date("2010-08-02"))

})

setOldClass("Date")

#' Various date utilities
#'
#' [Date()] mirrors primitive contructors in base R ([double()], [character()]
#' etc.)
#'
#' @rdname date_utils
#' @param x an R object
#' @param length A non-negative number specifying the desired length. Supplying
#'   an argument of length other than one is an error.
#' @seealso [is.instant()], [is.timespan()], [is.POSIXt()], [POSIXct()]
#' @keywords logic chron
#' @examples
#' is.Date(as.Date("2009-08-03")) # TRUE
#' is.Date(difftime(now() + 5, now())) # FALSE
#' @export
is.Date <- function(x) is(x, "Date")

#' @rdname date_utils
#' @export
Date <- function(length = 0L) {
  structure(rep.int(NA_real_, length), class = "Date")
}

#' @rdname date_utils
#' @export
NA_Date_ <- structure(NA_real_, class = "Date")

#' Get/set time zone component of a date-time
#'
#' @description
#' Conveniently get and set the time zone of a date-time.
#'
#' `tz<-` is an alias for [force_tz()], which preserves the local time,
#' creating a different instant in time. Use [with_tz()] if you want keep
#' the instant the same, but change the printed representation.
#'
#' @section Valid time zones:
#' Time zones are stored in system specific database, so are not guaranteed
#' to be the same on every system (however, they are usually pretty similar
#' unless your system is very out of date). You can see a complete list with
#' [OlsonNames()].
#'
#' @export
#' @param x A date-time vector, usually of class `POSIXct` or `POSIXlt`.
#' @return A character vector of length 1. An empty string (`""`) represents
#'   your current time zone.
#'
#'   For backward compatibility, the time zone of a date, `NA`, or
#'   character vector is `"UTC"`.
#' @seealso See [DateTimeClasses] for a description of the underlying
#'   `tzone` attribute..
#' @keywords utilities manip chron methods
#' @examples
#' x <- y <- ymd_hms("2012-03-26 10:10:00", tz = "UTC")
#' tz(x)
#'
#' # Note that setting tz() preserved the clock time, which implies
#' # that the actual instant in time is changing
#' tz(y) <- "Pacific/Auckland"
#' y
#' x - y
#'
#' # This is the same as force_tz()
#' force_tz(x, "Pacific/Auckland")
#'
#' # Use with_tz() if you want to change the time zone, leave
#' # the instant in time the same
#' with_tz(x, "Pacific/Auckland")
tz <- function(x) {
  UseMethod("tz")
}

#' @export
tz.POSIXt <- function(x) {
  tzone <- attr(x, "tzone")
  if (is.null(tzone)) {
    ""
  } else {
    tzone[[1]]
  }
}

#' @export
tz.Date <- function(x) {
  "UTC"
}

#' @export
tz.character <- function(x) {
  "UTC"
}

#' @export
tz.logical <- function(x) {
  if (all(is.na(x))) {
    "UTC"
  } else {
    NextMethod()
  }
}

#' @export
tz.default <- function(x) {
  warning(
    "tz(): Don't know how to compute timezone for object of class ",
    paste0(class(x), collapse = "/"),
    "; returning \"UTC\".",
    call. = FALSE
  )
  "UTC"
}

#' @export
tz.zoo <- function(x) {
  tz(zoo::index(x))
}

#' @export
tz.zooreg <- function(x) {
  tz(zoo::index(x))
}

#' @export
tz.yearmon <- function(x) {
  "UTC"
}

#' @export
tz.yearqtr <- function(x) {
  "UTC"
}

#' @export
tz.timeSeries <- function(x) {
  tz(x@FinCenter)
}

#' @rdname tz
#' @param value New value of time zone.
#' @export
"tz<-" <- function(x, value) {
  force_tz(x, value)
}

#' Get/set time zone component of a date-time
#'
#' @description
#' Conveniently get and set the time zone of a date-time.
#'
#' `tz<-` is an alias for [force_tz()], which preserves the local time,
#' creating a different instant in time. Use [with_tz()] if you want keep
#' the instant the same, but change the printed representation.
#'
#' @section Valid time zones:
#' Time zones are stored in system specific database, so are not guaranteed
#' to be the same on every system (however, they are usually pretty similar
#' unless your system is very out of date). You can see a complete list with
#' [OlsonNames()].
#'
#' @export
#' @param x A date-time vector, usually of class `POSIXct` or `POSIXlt`.
#' @return A character vector of length 1. An empty string (`""`) represents
#'   your current time zone.
#'
#'   For backward compatibility, the time zone of a date, `NA`, or
#'   character vector is `"UTC"`.
#' @seealso See [DateTimeClasses] for a description of the underlying
#'   `tzone` attribute..
#' @keywords utilities manip chron methods
#' @examples
#' x <- y <- ymd_hms("2012-03-26 10:10:00", tz = "UTC")
#' tz(x)
#'
#' # Note that setting tz() preserved the clock time, which implies
#' # that the actual instant in time is changing
#' tz(y) <- "Pacific/Auckland"
#' y
#' x - y
#'
#' # This is the same as force_tz()
#' force_tz(x, "Pacific/Auckland")
#'
#' # Use with_tz() if you want to change the time zone, leave
#' # the instant in time the same
#' with_tz(x, "Pacific/Auckland")
tz <- function(x) {
  UseMethod("tz")
}

#' @export
tz.POSIXt <- function(x) {
  tzone <- attr(x, "tzone")
  if (is.null(tzone)) {
    ""
  } else {
    tzone[[1]]
  }
}

#' @export
tz.Date <- function(x) {
  "UTC"
}

#' @export
tz.character <- function(x) {
  "UTC"
}

#' @export
tz.logical <- function(x) {
  if (all(is.na(x))) {
    "UTC"
  } else {
    NextMethod()
  }
}

#' @export
tz.default <- function(x) {
  ### FIXME: make this `stop` in next major release
  ## VS[21-03-2020]: v.1.7.8 already sent notification to broken packages
  warning(
    "tz(): Don't know how to compute timezone for object of class ",
    paste0(class(x), collapse = "/"),
    "; returning \"UTC\". This warning will become an error in the next major version of lubridate.",
    call. = FALSE
  )
  "UTC"
}

#' @export
tz.zoo <- function(x) {
  tz(zoo::index(x))
}

#' @export
tz.zooreg <- function(x) {
  tz(zoo::index(x))
}

#' @export
tz.yearmon <- function(x) {
  "UTC"
}

#' @export
tz.yearqtr <- function(x) {
  "UTC"
}

#' @export
tz.timeSeries <- function(x) {
  tz(x@FinCenter)
}

#' @rdname tz
#' @param value New value of time zone.
#' @export
"tz<-" <- function(x, value) {
  force_tz(x, value)
}

#' Does date time occur in the am or pm?
#'
#' @export
#' @param x a date-time object
#' @return TRUE or FALSE depending on whether x occurs in the am or pm
#' @keywords chron
#' @examples
#' x <- ymd("2012-03-26")
#' am(x)
#' pm(x)
am <- function(x) hour(x) < 12

#' @export
#' @rdname am
pm <- function(x) !am(x)

#' Timespan class
#'
#' Timespan is an S4 class with no slots. It is extended by the
#' [Interval-class], [Period-class], and [Duration-class]
#' classes.
#'
#' @export
#' @keywords internal
#' @aliases *,Timespan,Timespan-method
#' @aliases %/%,Timespan,Timespan-method
#' @aliases %/%,difftime,Timespan-method
#' @aliases %/%,Interval,Interval-method
#' @aliases %/%,Period,Period-method
#' @aliases %/%,Interval,Duration-method
#' @aliases %/%,Duration,Interval-method
#' @aliases %/%,Interval,Period-method
#' @aliases %/%,Period,Interval-method
setClass("Timespan")

#' Is x a length of time?
#'
#' @export is.timespan
#' @aliases is.timespan
#' @param x an R object
#' @return TRUE if x is a period, interval, duration, or difftime object, FALSE otherwise.
#' @seealso [is.instant()], [is.duration()], [is.difftime()], [is.period()], [is.interval()]
#' @keywords logic chron
#' @examples
#' is.timespan(as.Date("2009-08-03")) # FALSE
#' is.timespan(duration(second = 1)) # TRUE
is.timespan <- function(x) is(x, "Timespan")

#' Description of time span classes in lubridate
#'
#' @description
#' A time span can be measured in three ways: as a duration, an interval, or a
#' period.
#'
#' * [duration]s record the exact number of seconds in a time span.
#'   They measure the exact passage of time but do not always align with
#'   human measurements like hours, months and years.
#'
#' * [period]s record the change in the clock time between two date-times.
#'   They are measured in human units: years, months, days, hours, minutes,
#'   and seconds.
#'
#' * [intervals] are time spans bound by two real date-times. Intervals can be
#'   accurately converted to periods and durations.
#'
#' @aliases timespans
#' @keywords classes chron
#' @examples
#' duration(3690, "seconds")
#' period(3690, "seconds")
#' period(second = 30, minute = 1, hour = 1)
#' interval(ymd_hms("2009-08-09 13:01:30"), ymd_hms("2009-08-09 12:00:00"))
#'
#' date <- ymd_hms("2009-03-08 01:59:59") # DST boundary
#' date + days(1)
#' date + ddays(1)
#'
#' date2 <- ymd_hms("2000-02-29 12:00:00")
#' date2 + years(1)
#' # self corrects to next real day
#'
#' date3 <- ymd_hms("2009-01-31 01:00:00")
#' date3 + c(0:11) * months(1)
#'
#' span <- date2 %--% date # creates interval
#'
#' date <- ymd_hms("2009-01-01 00:00:00")
#' date + years(1)
#' date - days(3) + hours(6)
#' date + 3 * seconds(10)
#'
#' months(6) + days(1)
#' @name timespan
NULL

#' Compute the exact length of a time span
#'
#'
#' @param x a duration, period, difftime or interval
#' @param unit a character string that specifies with time units to use
#' @return the length of the interval in the specified unit. A negative number
#' connotes a negative interval or duration
#'
#' @details When `x` is an [Interval-class] object and
#' `unit` are years or months, `time_length()` takes into account
#' the fact that all months and years don't have the same number of days.
#'
#' When `x` is a [Duration-class], [Period-class]
#' or [difftime()] object, length in months or years is based on their
#' most common lengths in seconds (see [timespan()]).
#' @seealso [timespan()]
#' @keywords chron math period methods
#' @examples
#' int <- interval(ymd("1980-01-01"), ymd("2014-09-18"))
#' time_length(int, "week")
#'
#' # Exact age
#' time_length(int, "year")
#'
#' # Age at last anniversary
#' trunc(time_length(int, "year"))
#'
#' # Example of difference between intervals and durations
#' int <- interval(ymd("1900-01-01"), ymd("1999-12-31"))
#' time_length(int, "year")
#' time_length(as.duration(int), "year")
#' @export
setGeneric("time_length",
  useAsDefault =
    function(x, unit = "second") {
      as.duration(x) / duration(num = 1, units = unit)
    }
)

#' Converts a date to a decimal of its year
#'
#' @export
#' @param date a POSIXt or Date object
#' @return a numeric object where the date is expressed as a fraction of its year
#' @keywords manip chron methods
#' @examples
#' date <- ymd("2009-02-10")
#' decimal_date(date) # 2009.11
decimal_date <- function(date) {
  UseMethod("decimal_date")
}

#' @export
decimal_date.default <- function(date) {
  if (any(!inherits(date, c("POSIXt", "POSIXct", "POSIXlt", "Date")))) {
    stop("date(s) not in POSIXt or Date format")
  }

  Y <- year(date)
  start <- make_datetime(Y, 1L, 1L, tz = tz(date))
  end <- make_datetime(Y + 1L, 1L, 1L, tz = tz(date))
  sofar <- as.numeric(difftime(date, start, units = "secs"))
  total <- as.numeric(difftime(end, start, units = "secs"))
  Y + sofar / total
}

#' @export
decimal_date.zoo <- function(date) {
  decimal_date(zoo::index(date))
}

## #' @export
## decimal_date.its <- function(date)
##   decimal_date.default(attr(date, "dates"))


#' Converts a decimal to a date
#'
#' @export
#' @param decimal a numeric object
#' @param tz the time zone required
#' @return a POSIXct object, whose year corresponds to the integer part of
#' decimal. The months, days, hours, minutes and seconds elements are picked so
#' the date-time will accurately represent the fraction of the year expressed by
#' decimal.
#' @keywords manip chron methods
#' @examples
#' date <- ymd("2009-02-10")
#' decimal <- decimal_date(date) # 2009.11
#' date_decimal(decimal) # "2009-02-10 UTC"
date_decimal <- function(decimal, tz = "UTC") {
  Y <- trunc(decimal)
  ## parsing is much faster than updating
  start <- make_datetime(Y, 1L, 1L, tz = tz)
  end <- make_datetime(Y + 1L, 1L, 1L, tz = tz)
  seconds <- as.numeric(difftime(end, start, units = "secs"))
  frac <- decimal - Y
  end <- start + seconds * frac
  return(end)
}

test_that("decimal_date handles vectors", {
  x <- as.POSIXct(c("2008-08-03 13:01:59", "2009-08-03 10:01:59"), tz = "UTC")
  expect_equal(round(decimal_date(x), 3), c(2008.589, 2009.587))
})

test_that("decimal_date handles first second of year without returning NaN", {
  expect_equal(decimal_date(ymd("2012-01-01")), 2012)
})

test_that("date_decimal works as expected", {
  y <- 2008.58846
  expect_equal(as.numeric(date_decimal(y)), 1217754118)
})

test_that("date_decimal handles vectors", {
  y <- c(2008.589, 2009.587)
  expect_equal(as.numeric(date_decimal(y)), c(1217771194, 1249279632))
})

test_that("date_decimal reverses decimal_date", {
  date <- ymd("2012-03-01", tz = "UTC")
  decimal <- decimal_date(date)
  expect_equal(date_decimal(decimal), date)

  date2 <- ymd("2012-03-01")
  decimal2 <- decimal_date(date2)
  expect_equal(as.Date(date_decimal(decimal2)), date2)

  ## before day light shift
  date3 <- ymd_hms("2008-02-03 10:01:59", tz = "America/New_York")
  decimal3 <- decimal_date(date3)
  date_decimal(decimal3, tz = "America/New_York")
  expect_equal(date_decimal(decimal3, tz = "America/New_York"), date3)

  ## after
  date <- ymd_hms("2008-05-03 10:01:59", tz = "America/New_York")
  decimal <- decimal_date(date)
  date_decimal(decimal, tz = "America/New_York")
  expect_equal(date_decimal(decimal, tz = "America/New_York"), date)
})

test_that("decimal_date returns NA on NA date input", {
  expect_equal(decimal_date(as.Date(NA)), as.numeric(NA))
})

test_that("decimal_date correctly handles daylight savings time", {
  date4 <- ymd_hms("2016-03-13 01:00:00", tz = "America/New_York")
  date4 <- date4 + dhours(-1:1)
  decimal4 <- decimal_date(date4)
  expect_equal(c(2016.196721, 2016.1968351, 2016.196948), decimal4)
  diff.decimal <- diff(decimal4)
  expect_equal(diff.decimal[1], diff.decimal[2])
  expect_equal(c(2016.197290, 2016.197404, 2016.1975182), decimal_date(with_tz(date4, "UTC")))
})

context("Decimal_date")

test_that("decimal_date handles vectors", {
  x <- as.POSIXct(c("2008-08-03 13:01:59", "2009-08-03 10:01:59"), tz = "UTC")
  expect_equal(round(decimal_date(x), 3), c(2008.589, 2009.587))
})

test_that("decimal_date handles first second of year without returning NaN", {
  expect_equal(decimal_date(ymd("2012-01-01")), 2012)
})

test_that("date_decimal works as expected", {
  y <- 2008.58846
  expect_equal(as.numeric(date_decimal(y)), 1217754118)
})

test_that("date_decimal handles vectors", {
  y <- c(2008.589, 2009.587)
  expect_equal(as.numeric(date_decimal(y)), c(1217771194, 1249279632))
})

test_that("date_decimal reverses decimal_date", {
  date <- ymd("2012-03-01", tz = "UTC")
  decimal <- decimal_date(date)
  expect_equal(date_decimal(decimal), date)

  date2 <- ymd("2012-03-01")
  decimal2 <- decimal_date(date2)
  expect_equal(as.Date(date_decimal(decimal2)), date2)

  ## before day light shift
  date3 <- ymd_hms("2008-02-03 10:01:59", tz = "America/New_York")
  decimal3 <- decimal_date(date3)
  date_decimal(decimal3, tz = "America/New_York")
  expect_equal(date_decimal(decimal3, tz = "America/New_York"), date3)

  ## after
  date <- ymd_hms("2008-05-03 10:01:59", tz = "America/New_York")
  decimal <- decimal_date(date)
  date_decimal(decimal, tz = "America/New_York")
  expect_equal(date_decimal(decimal, tz = "America/New_York"), date)
})

test_that("decimal_date returns NA on NA date input", {
  expect_equal(decimal_date(as.Date(NA)), as.numeric(NA))
})

test_that("decimal_date correctly handles daylight savings time", {
  date4 <- ymd_hms("2016-03-13 01:00:00", tz = "America/New_York")
  date4        <- date4 + dhours(-1:1)
  decimal4     <- decimal_date(date4)
  expect_equal(c(2016.196721, 2016.1968351, 2016.196948), decimal4)
  diff.decimal <- diff(decimal4)
  expect_equal(diff.decimal[1], diff.decimal[2])
  expect_equal(c(2016.197290, 2016.197404, 2016.1975182), decimal_date(with_tz(date4, "UTC")))
})

#' Timespan class
#'
#' Timespan is an S4 class with no slots. It is extended by the
#' [Interval-class], [Period-class], and [Duration-class]
#' classes.
#'
#' @export
#' @keywords internal
#' @aliases *,Timespan,Timespan-method
#' @aliases %/%,Timespan,Timespan-method
#' @aliases %/%,difftime,Timespan-method
#' @aliases %/%,Interval,Interval-method
#' @aliases %/%,Period,Period-method
#' @aliases %/%,Interval,Duration-method
#' @aliases %/%,Duration,Interval-method
#' @aliases %/%,Interval,Period-method
#' @aliases %/%,Period,Interval-method
setClass("Timespan")

#' Is x a length of time?
#'
#' @export is.timespan
#' @aliases is.timespan
#' @param x an R object
#' @return TRUE if x is a period, interval, duration, or difftime object, FALSE otherwise.
#' @seealso [is.instant()], [is.duration()], [is.difftime()], [is.period()], [is.interval()]
#' @keywords logic chron
#' @examples
#' is.timespan(as.Date("2009-08-03")) # FALSE
#' is.timespan(duration(second = 1)) # TRUE
is.timespan <- function(x) is(x, "Timespan")

#' Description of time span classes in lubridate
#'
#' @description
#' A time span can be measured in three ways: as a duration, an interval, or a
#' period.
#'
#' * [duration]s record the exact number of seconds in a time span.
#'   They measure the exact passage of time but do not always align with
#'   human measurements like hours, months and years.
#'
#' * [period]s record the change in the clock time between two date-times.
#'   They are measured in human units: years, months, days, hours, minutes,
#'   and seconds.
#'
#' * [intervals] are time spans bound by two real date-times. Intervals can be
#'   accurately converted to periods and durations.
#'
#' @aliases timespans
#' @keywords classes chron
#' @examples
#' duration(3690, "seconds")
#' period(3690, "seconds")
#' period(second = 30, minute = 1, hour = 1)
#' interval(ymd_hms("2009-08-09 13:01:30"), ymd_hms("2009-08-09 12:00:00"))
#'
#' date <- ymd_hms("2009-03-08 01:59:59") # DST boundary
#' date + days(1)
#' date + ddays(1)
#'
#' date2 <- ymd_hms("2000-02-29 12:00:00")
#' date2 + years(1)
#' # self corrects to next real day
#'
#' date3 <- ymd_hms("2009-01-31 01:00:00")
#' date3 + c(0:11) * months(1)
#'
#' span <- date2 %--% date  #creates interval
#'
#' date <- ymd_hms("2009-01-01 00:00:00")
#' date + years(1)
#' date - days(3) + hours(6)
#' date + 3 * seconds(10)
#'
#' months(6) + days(1)
#' @name timespan
NULL

#' Compute the exact length of a time span
#'
#'
#' @param x a duration, period, difftime or interval
#' @param unit a character string that specifies with time units to use
#' @return the length of the interval in the specified unit. A negative number
#' connotes a negative interval or duration
#'
#' @details When `x` is an [Interval-class] object and
#' `unit` are years or months, `time_length()` takes into account
#' the fact that all months and years don't have the same number of days.
#'
#' When `x` is a [Duration-class], [Period-class]
#' or [difftime()] object, length in months or years is based on their
#' most common lengths in seconds (see [timespan()]).
#' @seealso [timespan()]
#' @keywords chron math period methods
#' @examples
#' int <- interval(ymd("1980-01-01"), ymd("2014-09-18"))
#' time_length(int, "week")
#'
#' # Exact age
#' time_length(int, "year")
#'
#' # Age at last anniversary
#' trunc(time_length(int, "year"))
#'
#' # Example of difference between intervals and durations
#' int <- interval(ymd("1900-01-01"), ymd("1999-12-31"))
#' time_length(int, "year")
#' time_length(as.duration(int), "year")
#' @export
setGeneric("time_length",
           useAsDefault =
             function(x, unit = "second") {
               as.duration(x) / duration(num = 1, units = unit)
             })

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
NULL

modulo_duration_by_duration <- function(dur, dur2) {
  dur - dur %/% dur2 * dur2
}

modulo_interval_by_duration <- function(int, dur) {
  interval(int@start + (int@.Data %/% dur@.Data * dur), int_end(int))
}

modulo_interval_by_period <- function(int, per) {
  interval(add_with_rollback(int@start, (int %/% per) * per), int_end(int))
}

modulo_period_by_period <- function(per, per2) {
  per - per %/% per2 * per2
}

#' @export
setMethod(
  "%%", signature(e1 = "Duration", e2 = "Duration"),
  function(e1, e2) modulo_duration_by_duration(e1, e2)
)

#' @export
setMethod(
  "%%", signature(e1 = "Interval", e2 = "Duration"),
  function(e1, e2) modulo_interval_by_duration(e1, e2)
)

#' @export
setMethod(
  "%%", signature(e1 = "Interval", e2 = "Interval"),
  function(e1, e2) stop("%% not defined for Interval class\nConsider setdiff()")
)

#' @export
setMethod(
  "%%", signature(e1 = "Interval", e2 = "Period"),
  function(e1, e2) modulo_interval_by_period(e1, e2)
)

#' @export
setMethod(
  "%%", signature(e1 = "Period", e2 = "Period"),
  function(e1, e2) modulo_period_by_period(e1, e2)
)

#' @export
setMethod(
  "%%", signature(e1 = "Period", e2 = "Interval"),
  function(e1, e2) modulo_period_by_period(e1, e2)
)

#' Converts a date to a decimal of its year
#'
#' @export
#' @param date a POSIXt or Date object
#' @return a numeric object where the date is expressed as a fraction of its year
#' @keywords manip chron methods
#' @examples
#' date <- ymd("2009-02-10")
#' decimal_date(date)  # 2009.11
decimal_date <- function(date)
  UseMethod("decimal_date")

#' @export
decimal_date.default <- function(date) {

  if (any(!inherits(date, c("POSIXt", "POSIXct", "POSIXlt", "Date"))))
    stop("date(s) not in POSIXt or Date format")

  Y <- year(date)
  start <- make_datetime(Y, 1L, 1L, tz = tz(date))
  end <- make_datetime(Y + 1L, 1L, 1L, tz = tz(date))
  sofar <- as.numeric(difftime(date, start, units = "secs"))
  total <- as.numeric(difftime(end, start, units = "secs"))
  Y + sofar/total
}

#' @export
decimal_date.zoo <- function(date)
  decimal_date(zoo::index(date))

## #' @export
## decimal_date.its <- function(date)
##   decimal_date.default(attr(date, "dates"))


#' Converts a decimal to a date
#'
#' @export
#' @param decimal a numeric object
#' @param tz the time zone required
#' @return a POSIXct object, whose year corresponds to the integer part of
#' decimal. The months, days, hours, minutes and seconds elements are picked so
#' the date-time will accurately represent the fraction of the year expressed by
#' decimal.
#' @keywords manip chron methods
#' @examples
#' date <- ymd("2009-02-10")
#' decimal <- decimal_date(date)  # 2009.11
#' date_decimal(decimal) # "2009-02-10 UTC"
date_decimal <- function(decimal, tz = "UTC") {
  Y <- trunc(decimal)
  ## parsing is much faster than updating
  start <- make_datetime(Y, 1L, 1L, tz = tz)
  end <- make_datetime(Y + 1L, 1L, 1L, tz = tz)
  seconds <- as.numeric(difftime(end, start, units = "secs"))
  frac <- decimal - Y
  end <- start + seconds*frac
  return(end)
}

#' @include timespans.r
#' @include util.r
#' @include durations.r
NULL

check_period <- function(object) {
  errors <- character()

  length(object@.Data) -> n
  lengths <- c(
    length(object@year), length(object@month),
    length(object@day), length(object@hour), length(object@minute)
  )

  if (any(lengths != n)) {
    msg <- paste("Inconsistent lengths: year = ", lengths[1],
      ", month = ", lengths[2],
      ", day = ", lengths[3],
      ", hour = ", lengths[4],
      ", minute = ", lengths[5],
      ", second = ", n,
      sep = ""
    )
    errors <- c(errors, msg)
  }

  values <- c(object@year, object@month, object@day, object@hour, object@minute)
  values <- na.omit(values)
  if (sum(values - trunc(values))) {
    msg <- "periods must have integer values"
    errors <- c(errors, msg)
  }

  if (length(errors) == 0) {
    TRUE
  } else {
    errors
  }
}

#' Period class
#'
#' Period is an S4 class that extends the [Timespan-class] class.
#' Periods track the change in the "clock time" between two date-times. They
#' are measured in common time related units: years, months, days, hours,
#' minutes, and seconds. Each unit except for seconds must be expressed in
#' integer values.
#'
#' The exact length of a period is not defined until the period is placed at a
#' specific moment of time. This is because the precise length of one year,
#' month, day, etc. can change depending on when it occurs due to daylight savings,
#' leap years, and other conventions. A period can be
#' associated with a specific moment in time by coercing it to an
#' [Interval-class] object with [as.interval()] or by adding
#' it to a date-time with "+".
#'
#' Periods provide a method for measuring generalized timespans when we wish to
#' model clock times. Periods will attain intuitive results at this task even
#' when leap years, leap seconds, gregorian days, daylight savings changes, and
#' other events happen during the period.
#'
#' Because Period represents imprecise amount of time it cannot be compared to
#' precise timestamps as Durations and Intervals are. You need to explicitely
#' convert to durations. See [Duration-class].
#'
#' The logic that guides arithmetic with periods can be unintuitive. Starting
#' with version 1.3.0, \pkg{lubridate} enforces the reversible property of arithmetic
#' (e.g. a date + period - period = date) by returning an NA if you create an
#' implausible date by adding periods with months or years units to a date. For
#' example,  adding one month to January 31st, 2013 results in February 31st,
#' 2013, which is not a real date. \pkg{lubridate} users have argued in the past that
#' February 31st, 2013 should be rolled over to March 3rd, 2013 or rolled back
#' to February 28, 2013. However, each of these corrections would destroy the
#' reversibility of addition (Mar 3 - one month == Feb 3 != Jan 31, Feb 28 - one
#' month == Jan 28 != Jan 31). If you would like to add and subtract months in a
#' way that rolls the results back to the last day of a month (when appropriate)
#' use the special operators, \code{\link{\%m+\%}},  \code{\link{\%m-\%}} or a
#' bit more flexible [add_with_rollback()].
#'
#' Period class objects have six slots. 1) .Data, a numeric object. The
#' apparent amount of seconds to add to the period. 2) minute, a numeric object.
#' The apparent amount of minutes to add to the period. 3) hour, a numeric object.
#' The apparent amount of hours to add to the period.4) day, a numeric object.
#' The apparent amount of days to add to the period.5) month, a numeric object.
#' The apparent amount of months to add to the period. 6) year, a numeric object.
#' The apparent amount of years to add to the period.
#'
#' @export
#' @keywords internal
setClass("Period",
  contains = c("Timespan", "numeric"),
  slots = c(
    year = "numeric", month = "numeric", day = "numeric",
    hour = "numeric", minute = "numeric"
  ),
  prototype = prototype(year = 0, month = 0, day = 0, hour = 0, minute = 0),
  validity = check_period
)

#' @name hidden_aliases
#' @aliases Arith,ANY,Period-method Arith,Duration,Period-method
#'   Arith,Period,Duration-method Compare,Period,Duration-method
#'   Compare,numeric,Period-method Compare,difftime,Period-method
#'   Compare,Period,difftime-method Compare,Period,Period-method
#'   Compare,Period,character-method Compare,Period,numeric-method
#'   Compare,character,Period-method second,Period-method second<-,Period-method
#'   minute,Period-method minute<-,Period-method hour,Period-method hour<-,Period-method
#'   Arith,Period,ANY-method day,Period-method day<-,Period-method month,Period-method
#'   month<-,Period-method year,Period-method year<-,Period-method date,Period-method
#'   date<-,Period-method as.numeric,Period-method as.integer,Period-method
#'   show,Period-method c,Period-method rep,Period-method [,Period-method
#'   [<-,Period,ANY,ANY,Period-method [[,Period-method [[<-,Period,ANY,ANY,Period-method
#'   $,Period-method $<-,Period-method as.difftime,Period-method
#'   as.character,Period-method +,Period,Duration-method +,Period,Interval-method
#'   +,Period,Period-method +,Period,Date-method +,Date,Period-method
#'   +,Period,difftime-method +,difftime,Period-method +,Period,numeric-method
#'   +,numeric,Period-method +,Period,POSIXct-method +,POSIXct,Period-method
#'   +,Period,POSIXlt-method +,POSIXlt,Period-method /,Period,Duration-method
#'   /,Period,Interval-method /,Period,Period-method /,Period,difftime-method
#'   /,difftime,Period-method /,Period,numeric-method /,numeric,Period-method
#'   *,Period,ANY-method *,ANY,Period-method -,Period,ANY-method -,Period,missing-method
#'   -,ANY,Period-method -,Period,Period-method %%,Period,Duration-method
#'   %%,Period,Interval-method %%,Period,Period-method >,Period,Period-method
#'   >=,Period,Period-method ==,Period,Period-method !=,Period,Period-method
#'   <=,Period,Period-method <,Period,Period-method >,Period,Duration-method
#'   >=,Period,Duration-method ==,Period,Duration-method !=,Period,Duration-method
#'   <=,Period,Duration-method <,Period,Duration-method >,Duration,Period-method
#'   >=,Duration,Period-method ==,Duration,Period-method !=,Duration,Period-method
#'   <=,Duration,Period-method <,Duration,Period-method >,Period,numeric-method
#'   >=,Period,numeric-method ==,Period,numeric-method !=,Period,numeric-method
#'   <=,Period,numeric-method <,Period,numeric-method >,numeric,Period-method
#'   >=,numeric,Period-method ==,numeric,Period-method !=,numeric,Period-method
#'   <=,numeric,Period-method <,numeric,Period-method !=,Duration,Period
#'   !=,Period,Duration !=,Period,Period !=,Period,numeric !=,numeric,Period
#'   %%,Period,Duration %%,Period,Interval %%,Period,Period *,ANY,Period *,Period,ANY
#'   -,ANY,Period -,Period,Interval -,Period,missing /,numeric,Period <,Duration,Period
#'   <,Period,Duration <,Period,Period <,Period,numeric <,numeric,Period
#'   <=,Duration,Period <=,Period,Duration <=,Period,Period <=,Period,numeric
#'   <=,numeric,Period ==,Duration,Period ==,Period,Duration ==,Period,Period
#'   ==,Period,numeric ==,numeric,Period >,Duration,Period >,Period,Duration
#'   >,Period,Period >,Period,numeric >,numeric,Period >=,Duration,Period
#'   >=,Period,Duration >=,Period,Period >=,Period,numeric >=,numeric,Period
#'   -,Period,Date-method -,Period,Duration-method -,Period,POSIXct-method
#'   -,Period,POSIXlt-method -,Period,numeric-method -,Date,Period-method
#'   -,numeric,Period-method -,POSIXlt,Period-method -,POSIXct,Period-method
NULL

setMethod("initialize", "Period", function(.Object, ...) {
  dots <- list(...)
  names(dots)[!nzchar(allNames(dots))] <- ".Data"
  lens <- unlist(lapply(dots, length), F, F)

  ## if any 0-length components, the entire object is 0-length
  if (any(lens == 0)) {
    for (nm in slotNames(.Object)) {
      slot(.Object, nm) <- numeric()
    }
    validObject(.Object)
    return(.Object)
  }

  len <- max(lens)
  nas <- FALSE
  for (nm in slotNames(.Object)) {
    el <-
      if (is.null(obj <- dots[[nm]])) {
        rep.int(0L, len)
      } else {
        if (length(obj) < len) {
          rep_len(obj, len)
        } else {
          obj
        }
      }
    nas <- nas | is.na(el)
    slot(.Object, nm) <- el
  }

  ## If any component has NAs the entire object should have those NAs
  if (any(nas)) {
    for (nm in slotNames(.Object)) {
      slot(.Object, nm)[nas] <- NA_real_
    }
  }

  validObject(.Object)
  .Object
})

#' @export
setMethod("show", signature(object = "Period"), function(object) {
  if (length(object@.Data) == 0) {
    cat("<Period[0]>\n")
  } else {
    print(format(object))
  }
})

#' @export
format.Period <- function(x, ...) {
  if (length(x) == 0) {
    return(character())
  }

  show <- paste(
    x@year, "y ", x@month, "m ", x@day, "d ",
    x@hour, "H ", x@minute, "M ", x@.Data, "S",
    sep = ""
  )
  start <- regexpr("[-1-9]|(0\\.)", show)
  show <- ifelse(start > 0, substr(show, start, nchar(show)), "0S")

  show[is.na(x)] <- NA
  show
}

#' @export
xtfrm.Period <- function(x) {
  xtfrm(period_to_seconds(x))
}

#' @export
c.Period <- function(...) {
  dots <- list(...)
  nempty <- sapply(dots, length) != 0
  elements <- lapply(dots[nempty], as.period)
  seconds <- unlist(lapply(elements, slot, ".Data"))
  # don't keep names for slots
  elements <- unname(elements)
  years <- unlist(lapply(elements, slot, "year"))
  months <- unlist(lapply(elements, slot, "month"))
  days <- unlist(lapply(elements, slot, "day"))
  hours <- unlist(lapply(elements, slot,  "hour"))
  minutes <- unlist(lapply(elements, slot, "minute"))
  new("Period",
    seconds, year = years, month = months,
    day = days, hour = hours, minute = minutes
  )
}

#' @export
setMethod("rep", signature(x = "Period"), function(x, ...) {
  new("Period", rep(x@.Data, ...),
    year = rep(x@year, ...),
    month = rep(x@month, ...), day = rep(x@day, ...),
    hour = rep(x@hour, ...), minute = rep(x@minute, ...)
  )
})

#' @export
setMethod(
  "[", signature(x = "Period"),
  function(x, i, j, ..., drop = TRUE) {
    new("Period", x@.Data[i],
      year = x@year[i], month = x@month[i],
      day = x@day[i], hour = x@hour[i], minute = x@minute[i]
    )
  }
)

#' @export
setMethod(
  "[[", signature(x = "Period"),
  function(x, i, j, ..., exact = TRUE) {
    new("Period", x@.Data[i],
      year = x@year[i], month = x@month[i],
      day = x@day[i], hour = x@hour[i], minute = x@minute[i]
    )
  }
)

#' @export
setMethod(
  "[<-", signature(x = "Period", value = "Period"),
  function(x, i, j, ..., value) {
    x@.Data[i] <- value@.Data
    x@year[i] <- value@year
    x@month[i] <- value@month
    x@day[i] <- value@day
    x@hour[i] <- value@hour
    x@minute[i] <- value@minute
    x
  }
)

#' @export
setMethod(
  "[[<-", signature(x = "Period", value = "Period"),
  function(x, i, j, ..., value) {
    x@.Data[i] <- value@.Data
    x@year[i] <- value@year
    x@month[i] <- value@month
    x@day[i] <- value@day
    x@hour[i] <- value@hour
    x@minute[i] <- value@minute
    x
  }
)

#' @export
setMethod("$", signature(x = "Period"), function(x, name) {
  if (name == "second") name <- ".Data"
  slot(x, name)
})

#' @export
setMethod("$<-", signature(x = "Period"), function(x, name, value) {
  if (name == "second") name <- ".Data"
  slot(x, name) <- rep_len(value, length(x))
  x
})

#' Create or parse period objects
#'
#' `period()` creates or parses a period object with the specified values.
#'
#' Within a Period object, time units do not have a fixed length (except for
#' seconds) until they are added to a date-time. The length of each time unit
#' will depend on the date-time to which it is added. For example, a year that
#' begins on 2009-01-01 will be 365 days long.  A year that begins on 2012-01-01
#' will be 366 days long. When math is performed with a period object, each unit
#' is applied separately. How the length of a period is distributed among its
#' units is non-trivial. For example, when leap seconds occur 1 minute is longer
#' than 60 seconds.
#'
#' Periods track the change in the "clock time" between two date-times. They
#' are measured in common time related units: years, months, days, hours,
#' minutes, and seconds. Each unit except for seconds must be expressed in
#' integer values.
#'
#' Besides the main constructor and parser [period()], period objects can also
#' be created with the specialized functions [years()], [months()], [weeks()],
#' [days()], [hours()], [minutes()], and [seconds()]. These objects can be added
#' to and subtracted to date-times to create a user interface similar to object
#' oriented programming.
#'
#' Note: Arithmetic with periods can result in undefined behavior when
#' non-existent dates are involved (such as February 29th in non-leap years).
#' Please see [Period-class] for more details and \code{\link{\%m+\%}} and
#' [add_with_rollback()] for alternative operations.
#'
#' @name period
#' @aliases periods
#' @param num a numeric or character vector. A character vector can specify
#'   periods in a convenient shorthand format or ISO 8601 specification. All
#'   unambiguous name units and abbreviations are supported, "m" stands for
#'   months, "M" for minutes unless ISO 8601 "P" modifier is present (see
#'   examples). Fractional units are supported but the fractional part is always
#'   converted to seconds.
#' @param units a character vector that lists the type of units to be used. The
#'   units in units are matched to the values in num according to their
#'   order. When `num` is character, this argument is ignored.
#' @param ... a list of time units to be included in the period and their
#'   amounts. Seconds, minutes,  hours, days, weeks, months, and years are
#'   supported. Normally only one of `num` or `...` are present. If both are
#'   present, the periods are concatenated.
#' @param x Any R object for `is.periods` and a numeric value of the number of
#'   units for elementary constructors. With the exception of seconds(), x must
#'   be an integer.
#' @param abbreviate Ignored. For consistency with S3 generic in base namespace.
#' @seealso [Period-class], [period()], \code{\link{\%m+\%}},
#'   [add_with_rollback()]
#' @return a period object
#' @keywords chron classes
#' @examples
#'
#' ### Separate period and units vectors
#'
#' period(c(90, 5), c("second", "minute"))
#' #  "5M 90S"
#' period(-1, "days")
#' period(c(3, 1, 2, 13, 1), c("second", "minute", "hour", "day", "week"))
#' period(c(1, -60), c("hour", "minute"))
#' period(0, "second")
#'
#' ### Units as arguments
#'
#' period(second = 90, minute = 5)
#' period(day = -1)
#' period(second = 3, minute = 1, hour = 2, day = 13, week = 1)
#' period(hour = 1, minute = -60)
#' period(second = 0)
#' period(c(1, -60), c("hour", "minute"), hour = c(1, 2), minute = c(3, 4))
#'
#' ### Lubridate style parsing
#'
#' period("2M 1sec")
#' period("2hours 2minutes 1second")
#' period("2d 2H 2M 2S")
#' period("2days 2hours 2mins 2secs")
#' period("2 days, 2 hours, 2 mins, 2 secs")
#' # Missing numerals default to 1. Repeated units are added up.
#' duration("day day")
#'
#' ### ISO 8601 parsing
#'
#' period("P10M23DT23H") # M stands for months
#' period("10DT10M") # M stands for minutes
#' period("P3Y6M4DT12H30M5S") # M for both minutes and months
#' period("P23DT60H 20min 100 sec") # mixing ISO and lubridate style parsing
#'
#' ### Comparison with characters (from v1.6.0)
#'
#' duration("day 2 sec") > "day 1sec"
#'
#' ### Elementary Constructors
#'
#' x <- ymd("2009-08-03")
#' x + days(1) + hours(6) + minutes(30)
#' x + days(100) - hours(8)
#'
#' class(as.Date("2009-08-09") + days(1)) # retains Date class
#' as.Date("2009-08-09") + hours(12)
#' class(as.Date("2009-08-09") + hours(12))
#' # converts to POSIXt class to accomodate time units
#'
#' years(1) - months(7)
#' c(1:3) * hours(1)
#' hours(1:3)
#'
#' # sequencing
#' y <- ymd(090101) # "2009-01-01 CST"
#' y + months(0:11)
#'
#' # compare DST handling to durations
#' boundary <- ymd_hms("2009-03-08 01:59:59", tz = "America/Chicago")
#' boundary + days(1) # period
#' boundary + ddays(1) # duration
#' @export
period <- function(num = NULL, units = "second", ...) {
  if (is.character(num)) {
    parse_period(num)
  } else {
    out1 <- .period_from_num(num, units)
    out2 <- .period_from_units(list(...))
    if (is.null(out1) && is.null(out2)) {
      new("Period", numeric())
    } else if (is.null(out1)) {
      out2
    } else if (is.null(out2)) {
      out1
    } else {
      c(out1, out2)
    }
  }
}

parse_period <- function(x) {
  out <- .Call(C_parse_period, as.character(x))
  new("Period",
    out[1, ],
    minute = out[2, ],
    hour   = out[3, ],
    day    = out[4, ] + 7L * out[5, ],
    month  = out[6, ],
    year   = out[7, ]
  )
}

.period_from_num <- function(num, units) {
  if (length(num) == 0) {
    return(NULL)
  }

  if (!is.numeric(num)) {
    stop(sprintf("First argument to `period()` constructor must be character or numeric. Supplied object of class '%s'", class(num)))
  }

  ## qucik check for common wrongdoings: #462
  if (inherits(num, c("Interval", "Duration"))) {
    stop("Interval or Durations objects cannot be used as input to 'period()' constructor. Plese use 'as.period()'.")
  }

  stop_incompatible_size(num, units)
  if (length(units) > length(num)) {
    num <- rep_len(num, length(units))
  }

  .period_from_units(structure(num, names = rep_len(units, length(num))))
}

.period_from_units <- function(pieces) {
  if (length(pieces) == 0) {
    return(NULL)
  }

  if (!is.numeric(pieces)) {
    pieces <- lapply(pieces, as.numeric)
  }

  out <- list(
    second = 0, minute = 0, hour = 0, day = 0,
    week = 0, month = 0, year = 0
  )

  unit <- standardise_date_names(names(pieces))
  for (i in seq_along(unit)) {
    nm <- unit[[i]]
    out[[nm]] <- out[[nm]] + pieces[[i]]
  }
  out$day <- out$day + 7 * out$week

  new("Period", out$second,
    year = out$year, month = out$month,
    day = out$day, hour = out$hour, minute = out$minute
  )
}

#' @rdname period
#' @examples
#' is.period(as.Date("2009-08-03")) # FALSE
#' is.period(period(months = 1, days = 15)) # TRUE
#' @export
is.period <- function(x) is(x, "Period")

#' @export seconds minutes hours days weeks years milliseconds microseconds microseconds nanoseconds picoseconds
#' @rdname period
seconds <- function(x = 1) period(second = x)
#' @rdname period
minutes <- function(x = 1) period(minute = x)
#' @rdname period
hours <- function(x = 1) period(hour = x)
#' @rdname period
days <- function(x = 1) period(day = x)
#' @rdname period
weeks <- function(x = 1) period(week = x)
#' @rdname period
years <- function(x = 1) period(year = x)
#' @rdname period
milliseconds <- function(x = 1) seconds(x / 1000)
#' @rdname period
microseconds <- function(x = 1) seconds(x / 1000000)
#' @rdname period
nanoseconds <- function(x = 1) seconds(x / 1e9)
#' @rdname period
picoseconds <- function(x = 1) seconds(x / 1e12)

#' @rdname period
#' @export
months.numeric <- function(x, abbreviate) {
  period(month = x)
}

#' Contrive a period to/from a given number of seconds
#'
#' `period_to_seconds()` approximately converts a period to seconds assuming
#' there are 365.25 days in a calendar year and 365.25/12 days in a month.
#'
#' @param x A numeric object. The number of seconds to coerce into a period.
#' @return A number (period) that roughly equates to the period (seconds) given.
#' @export
period_to_seconds <- function(x) {
  x@.Data +
    60 * x@minute +
    60 * 60 * x@hour +
    60 * 60 * 24 * x@day +
    60 * 60 * 24 * 365.25 / 12 * x@month +
    60 * 60 * 24 * 365.25 * x@year
}

#' @description
#' `seconds_to_period()` create a period that has the maximum number of
#' non-zero elements (days, hours, minutes, seconds). This computation is exact
#' because it doesn't involve years or months.
#' @rdname period_to_seconds
#' @export
seconds_to_period <- function(x) {
  span <- as.double(x)
  remainder <- abs(span)
  newper <- period(second = rep(0, length(x)))

  slot(newper, "day") <- remainder %/% (3600 * 24)
  remainder <- remainder %% (3600 * 24)

  slot(newper, "hour") <- remainder %/% (3600)
  remainder <- remainder %% (3600)

  slot(newper, "minute") <- remainder %/% (60)

  slot(newper, ".Data") <- remainder %% (60)

  newper * sign(span)
}

#' @export
summary.Period <- function(object, ...) {
  nas <- is.na(object)
  object <- object[!nas]
  persecs <- period_to_seconds(object)
  qq <- stats::quantile(persecs)
  qq <- c(qq[1L:3L], mean(persecs), qq[4L:5L])
  qq <- seconds_to_period(qq)
  qq <- as.character(qq)
  names(qq) <- c(
    "Min.", "1st Qu.", "Median", "Mean", "3rd Qu.",
    "Max."
  )
  if (any(nas)) {
    c(qq, `NA's` = sum(nas))
  } else {
    qq
  }
}

#' @export
setMethod("Arith", signature(e1 = "Period", e2 = "ANY"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, .Generic)
})

#' @export
setMethod("Arith", signature(e1 = "ANY", e2 = "Period"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, .Generic)
})

## duration is.numeric. So we need these explicits here:
#' @export
setMethod("Arith", signature(e1 = "Duration", e2 = "Period"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, .Generic)
})

#' @export
setMethod("Arith", signature(e1 = "Period", e2 = "Duration"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, .Generic)
})

#' @export
setMethod(
  "Compare", signature(e1 = "Period", e2 = "Period"),
  function(e1, e2) {
    callGeneric(period_to_seconds(e1), period_to_seconds(e2))
  }
)

#' @export
setMethod(
  "Compare", signature(e1 = "Period", e2 = "character"),
  function(e1, e2) {
    callGeneric(e1, as.period(e2))
  }
)

#' @export
setMethod(
  "Compare", signature(e1 = "character", e2 = "Period"),
  function(e1, e2) {
    callGeneric(as.period(e1), e2)
  }
)

#' @export
setMethod(
  "Compare", signature(e1 = "Period", e2 = "Duration"),
  function(e1, e2) {
    callGeneric(as.duration(e1), e2)
  }
)

#' @export
setMethod(
  "Compare", signature(e1 = "Duration", e2 = "Period"),
  function(e1, e2) {
    callGeneric(e1, as.duration(e2))
  }
)

#' @export
setMethod(
  "Compare", signature(e1 = "Period", e2 = "numeric"),
  function(e1, e2) {
    callGeneric(as.numeric(e1, "secs"), e2)
  }
)

#' @export
setMethod(
  "Compare", signature(e1 = "numeric", e2 = "Period"),
  function(e1, e2) {
    callGeneric(e1, as.numeric(e2, "secs"))
  }
)

#' @export
setMethod(
  "Compare", c(e1 = "Period", e2 = "difftime"),
  function(e1, e2) {
    callGeneric(as.numeric(e1, units = "secs"), as.numeric(e2, units = "secs"))
  }
)

#' @export
setMethod(
  "Compare", c(e1 = "difftime", e2 = "Period"),
  function(e1, e2) {
    callGeneric(as.numeric(e1, units = "secs"), as.numeric(e2, units = "secs"))
  }
)

context("Namespace")

if (file.exists("../../00check.log")) {
  # test was invoked by R CMD check -> package is already built
  R_test_lib <- normalizePath("../..")
  env <- paste0("R_LIBS=", R_test_lib, ":", Sys.getenv("R_LIBS"))
} else {
  # We're testing in the source directory -> need to build and install
  R_test_lib <- file.path(tempdir(), "Rlib")
  env <- ""
  ## dir.create(R_test_lib)
  ## on.exit(unlink(R_test_lib, recursive = TRUE))
  ## system2(
  ##   "R",
  ##   c("CMD install",
  ##     paste0("--library=", R_test_lib),
  ##     "--no-docs --no-help --no-demo --no-data --no-test-load",
  ##     normalizePath("../..")),
  ##   stdout = TRUE, stderr = TRUE)
}

do_Rscript <- function(expr) {
  rscript <- sprintf("%s/bin/Rscript", Sys.getenv("R_HOME"))
  paste(
    system2(rscript,
            args = c("--vanilla", "--default-packages=NULL", "-e", shQuote(expr)),
            env = c("R_TESTS=", env),
            stdout = TRUE, stderr = TRUE),
    collapse = "\n")
}

test_that("methods is not attached", {
  skip_on_cran()
  skip_on_os("windows")
  # Checking test assumptions.
  # If this fails, namespace tests may not be needed anymore!
  expect_match(
    do_Rscript("'package:methods' %in% search()"),
    "FALSE")
})

test_that("lubridate:: calls work when methods is not attached", {
  skip_on_cran()
  skip_on_os("windows")
  expect_match( # https://github.com/tidyverse/lubridate/issues/314
    do_Rscript(
      "lubridate::round_date(as.POSIXct('2017-10-03 03:01:13Z'), 'minute')"),
    as.character(round_date(as.POSIXct('2017-10-03 03:01:13Z'), 'minute')),
    fixed = TRUE)
  expect_match( # https://github.com/tidyverse/lubridate/issues/407
    do_Rscript("lubridate::days(1)"),
    as.character(days(1)),
    fixed = TRUE)
  expect_match( # https://github.com/tidyverse/lubridate/issues/499
    do_Rscript("lubridate::dhours(22)"),
    as.character(dhours(22)),
    fixed = TRUE)
})

if (file.exists("../../00check.log")) {
  # test was invoked by R CMD check -> package is already built
  R_test_lib <- normalizePath("../..")
  env <- paste0("R_LIBS=", R_test_lib, ":", Sys.getenv("R_LIBS"))
} else {
  # We're testing in the source directory -> need to build and install
  R_test_lib <- file.path(tempdir(), "Rlib")
  env <- ""
  ## dir.create(R_test_lib)
  ## on.exit(unlink(R_test_lib, recursive = TRUE))
  ## system2(
  ##   "R",
  ##   c("CMD install",
  ##     paste0("--library=", R_test_lib),
  ##     "--no-docs --no-help --no-demo --no-data --no-test-load",
  ##     normalizePath("../..")),
  ##   stdout = TRUE, stderr = TRUE)
}

do_Rscript <- function(expr) {
  rscript <- sprintf("%s/bin/Rscript", Sys.getenv("R_HOME"))
  paste(
    system2(rscript,
      args = c("--vanilla", "--default-packages=NULL", "-e", shQuote(expr)),
      env = c("R_TESTS=", env),
      stdout = TRUE, stderr = TRUE
    ),
    collapse = "\n"
  )
}

test_that("methods is not attached", {
  skip_on_cran()
  skip_on_os("windows")
  # Checking test assumptions.
  # If this fails, namespace tests may not be needed anymore!
  expect_match(
    do_Rscript("'package:methods' %in% search()"),
    "FALSE"
  )
})

test_that("lubridate:: calls work when methods is not attached", {
  skip_on_cran()
  skip_on_os("windows")
  expect_match( # https://github.com/tidyverse/lubridate/issues/314
    do_Rscript(
      "lubridate::round_date(as.POSIXct('2017-10-03 03:01:13Z'), 'minute')"
    ),
    as.character(round_date(as.POSIXct("2017-10-03 03:01:13Z"), "minute")),
    fixed = TRUE
  )
  expect_match( # https://github.com/tidyverse/lubridate/issues/407
    do_Rscript("lubridate::days(1)"),
    as.character(days(1)),
    fixed = TRUE
  )
  expect_match( # https://github.com/tidyverse/lubridate/issues/499
    do_Rscript("lubridate::dhours(22)"),
    as.character(dhours(22)),
    fixed = TRUE
  )
})

test_that("period() returns zero-length vector", {
  x <- period()
  expect_s4_class(x, "Period")
  expect_length(x, 0)
  expect_equal(format(x), character())
  expect_output(print(x), "<Period[0]>", fixed = TRUE)
})

test_that("period(0-len-vec, ...) returns zero-length period", {
  x <- period(character())
  expect_s4_class(x, "Period")
  expect_length(x, 0)

  x <- period(hour = numeric())
  expect_s4_class(x, "Period")
  expect_length(x, 0)

  x <- period(numeric(), hour = numeric())
  expect_s4_class(x, "Period")
  expect_length(x, 0)

  expect_equal(
    period(c(30, 20), units = c("secs", "days")),
    period(c(30, 20), units = c("secs", "days"), days = numeric())
  )

  expect_equal(
    period(numeric(), days = 10),
    period(days = 10)
  )

  expect_equal(
    period(days = 10, secs = 1, hours = numeric()),
    period()
  )

  expect_equal(
    period(days = 10, secs = numeric(), hours = 3),
    period()
  )
})

test_that("period constructor doesn't accept non-numeric or non-character inputs", {
  expect_error(period(interval(ymd("2014-01-01"), ymd("2015-01-01"))))
})

test_that("period constructor works with duplicated units", {
  expect_equal(
    period(c(1, 1, 2, 10), c("days", "sec", "seconds", "day")),
    period("11d 3s")
  )
  expect_equal(
    period(sec = 1, day = 1, seconds = 2, days = 10),
    period("11d 3s")
  )
  expect_equal(
    period(sec = c(1, 2), day = c(1, 2), seconds = 2, days = c(10, 20, 30, 40)),
    period(c("11d 3s", "22d 4s", "31d 3s", "42d 4s"))
  )
})

test_that("period parsing works", {
  expect_equal(
    period("1min 2sec 2secs 1H 2M 1d"),
    period(seconds = 4, minutes = 3, hours = 1, days = 1)
  )
  expect_equal(
    period("day day"),
    period(days = 2)
  )
  expect_equal(
    period("S M H d m y"),
    period(seconds = 1, minutes = 1, hours = 1, days = 1, months = 1, years = 1)
  )
  expect_equal(
    period("2S 3M 4H 5d 6w 7m 8y"),
    period(seconds = 2, minutes = 3, hours = 4, days = 5, weeks = 6, months = 7, years = 8)
  )
  expect_equal(period("K")@.Data, NA_real_)
  expect_equal(period("K")@hour, NA_real_)
  expect_equal(period("ksfdsfds")@.Data, NA_real_)
  expect_equal(period("ksfdsfds")@day, NA_real_)
})

test_that("ISO ISO 8601 period parsing works", {
  expect_equal(period("")@.Data, NA_real_)
  expect_equal(period("P")@.Data, NA_real_)
  expect_equal(
    period("P3Y6M4DT12H30M5S"),
    period(years = 3, months = 6, days = 4, hours = 12, minutes = 30, seconds = 5)
  )
  expect_equal(
    period("P23DT23H"),
    period(days = 23, hours = 23)
  )
  expect_equal(
    period("P23DT60H"),
    period(days = 23, hours = 60)
  )
  expect_equal(
    period("P23DT60H20minutes 100 sec"),
    period(days = 23, hours = 60, minutes = 20, seconds = 100)
  )
  expect_equal(
    period(c("P23DT23H", "PT0S", "P0D")),
    period(days = c(23, 0, 0), hours = c(23, 0, 0))
  )
})

test_that("fractional parsing works as expected", {
  expect_equal(
    period("1.1min 2.3sec 2.3secs 1.0H 2.2M 1.5d"),
    period(seconds = 43222.6, minutes = 3, hours = 1, days = 1)
  )
  expect_equal(
    period("day 1.2days"),
    period(days = 2, seconds = 17280)
  )
})

test_that("is.period works", {
  expect_false(is.period(234))
  expect_false(is.period(as.POSIXct("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.period(as.POSIXlt("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.period(Sys.Date()))
  expect_true(is.period(minutes(1)))
  expect_false(is.period(dminutes(1)))
  expect_false(is.period(interval(
    as.POSIXct("2008-08-03 13:01:59", tz = "UTC"),
    as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  )))
})

test_that("period works as expected", {
  per <- period(second = 90, minute = 5)

  expect_s4_class(per, "Period")
  expect_equal(period(hours = 25), hours(25))
  expect_equal(per@year, 0)
  expect_equal(per@month, 0)
  expect_equal(per@day, 0)
  expect_equal(per@hour, 0)
  expect_equal(per@minute, 5)
  expect_equal(per@.Data, 90)
})

test_that("period works as expected", {
  per <- period(c(90, 5), c("second", "minute"))

  expect_s4_class(per, "Period")
  expect_equal(period(hours = 25), hours(25))
  expect_equal(per@year, 0)
  expect_equal(per@month, 0)
  expect_equal(per@day, 0)
  expect_equal(per@hour, 0)
  expect_equal(per@minute, 5)
  expect_equal(per@.Data, 90)
})

test_that("period handles vector input", {
  per <- period(second = c(90, 60), minute = 5)

  expect_s4_class(per, "Period")
  expect_equal(length(per), 2)
  expect_equal(per@year, c(0, 0))
  expect_equal(per@month, c(0, 0))
  expect_equal(per@day, c(0, 0))
  expect_equal(per@hour, c(0, 0))
  expect_equal(per@minute, c(5, 5))
  expect_equal(per@.Data, c(90, 60))
})


test_that("period objects handle vector input", {
  x <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  expect_equal(
    as.numeric(x + minutes(c(1, 3, 4))),
    as.numeric(x + c(60, 180, 240))
  )
})

test_that("format.Period works as expected", {
  per <- period(second = 90, minute = 5)
  per2 <- period(days = 0)
  per3 <- period(years = 1, days = NA)
  expect_match(format(per), "5M 90S")
  expect_match(format(per2), "0S")
  expect_equal(format(per3), as.character(NA))
})

test_that("as.numeric and as.duration correctly handle periods", {
  tt <- period(24, "hours") + period(60, "minutes")
  expect_equal(as.numeric(tt, "hour"), 25)
  tt <- period(hours = 24, minutes = 6)
  expect_equal(as.numeric(tt, "hour"), 24.1)
  expect_equal(
    as.numeric(tt, "hour"),
    as.numeric(as.duration(tt), "hour")
  )
})

test_that("as.period handles interval objects", {
  start <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  end <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  int <- interval(start, end)
  int_neg <- interval(end, start)

  expect_equal(as.period(int), years(1))
  expect_equal(as.period(int_neg), years(-1))
})

test_that("as.period handles intervals across years correctly", {
  ## https://github.com/tidyverse/lubridate/issues/371

  t1 <- ymd("1985-11-30")
  t2 <- ymd("1986-02-01")
  expect_equal(t2, t1 + as.period(interval(t1, t2)))

  ## negative
  t1 <- ymd("1986-02-01")
  t2 <- ymd("1985-11-30")
  expect_equal(t2, t1 + as.period(interval(t1, t2)))

  t1 <- "1985-11-01"
  t2 <- "2007-08-01"
  as.period(interval(t1, t2))
  ymd(t1) + as.period(interval(t1, t2, "UTC"))

  ## negative
  t1 <- "2007-08-01"
  t2 <- "1985-11-01"
  as.period(interval(t1, t2))
  ymd(t1) + as.period(interval(t1, t2, "UTC"))

  t1 <- ymd("1986-02-01")
  t2 <- ymd("1986-01-05")
  expect_equal(t2, t1 + as.period(interval(t1, t2)))
})

test_that("as.period handles vector interval objects", {
  ## https://github.com/tidyverse/lubridate/issues/349
  ints <- c(
    interval(ymd("2001-01-01"), ymd("2002-01-01")),
    interval(ymd("2001-01-01"), ymd("2004-01-01"))
  )
  expect_equal(as.period(ints), period(years = c(1, 3)))

  ints <- c(
    interval(ymd("2001-01-01"), ymd("2002-03-05")),
    interval(ymd("2001-01-01"), ymd_hms("2004-12-31 3:2:1"))
  )
  expect_equal(
    as.period(ints),
    period(
      years = c(1, 3), months = c(2, 11), days = c(4, 30),
      hours = c(0, 3), minutes = c(0, 2), seconds = c(0, 1)
    )
  )
})

test_that("as.period handles don't produce negative periods", {
  p2 <- as.period(interval(ymd("1985-12-31"), ymd("1986-02-01")))
  expect_equal(month(p2), 1)

  p3 <- as.period(interval(ymd("1985-12-31"), ymd("1986-03-01")))
  expect_equal(month(p3), 2)
  expect_equal(day(p3), 1)
})

test_that("as.period handles interval objects with special start dates", {
  start <- ymd("1992-02-29")
  end <- ymd("2010-12-05")
  int <- interval(start, end)

  expect_equal(as.period(int), period(c(18, 9, 6), c("year", "month", "day")))
  expect_equal(as.period(int) + start, end)
})


test_that("as.period with different units handles interval objects", {
  start <- ymd("1992-02-29")
  end <- ymd_hms("2010-12-05 01:02:03")
  int <- interval(start, end)

  expect_equal(
    as.period(int),
    period(c(18, 9, 6, 1, 2, 3), c("year", "month", "day", "hour", "minute", "second"))
  )
  expect_equal(as.period(int) + start, end)

  expect_equal(
    as.period(int, "months"),
    period(c(225, 6, 1, 2, 3), c("month", "day", "hour", "minute", "second"))
  )
  expect_equal(as.period(int, "months") + start, end)

  expect_equal(as.period(int, "hours"), period(c(164497, 2, 3), c("hour", "minute", "second")))
  expect_equal(as.period(int, "hours") + start, end)

  expect_equal(as.period(int, "minute"), period(c(9869822, 3), c("minute", "second")))
  expect_equal(as.period(int, "minute") + start, end)

  expect_equal(as.period(int, "second"), period(c(592189323), c("second")))
  expect_equal(as.period(int, "second") + start, end)
})

test_that("period constructor works num-unit args", {
  expect_equal(period(1, c("sec", "hour")), period("1H 1S"))
  expect_equal(period(1:3, c("sec", "hour", "min")), period("2H 1S 3M"))
  expect_error(period(1:3, c("sec", "hour")), "Incompatible")
})

test_that("as.period with different units handles negative interval objects", {
  end <- ymd("1992-02-29", tz = "UTC")
  start <- ymd_hms("2010-12-05 01:02:03")
  int <- interval(start, end)

  expect_equal(
    as.period(int),
    period(-c(18, 9, 5, 1, 2, 3), c("year", "month", "day", "hour", "minute", "second"))
  )
  expect_equal(as.period(int) + start, end)

  expect_equal(
    as.period(int, "months"),
    period(-c(225, 5, 1, 2, 3), c("month", "day", "hour", "minute", "second"))
  )
  expect_equal(as.period(int, "months") + start, end)

  expect_equal(as.period(int, "hours"), period(-c(164497, 2, 3), c("hour", "minute", "second")))
  expect_equal(as.period(int, "hours") + start, end)

  expect_equal(as.period(int, "minute"), period(-c(9869822, 3), c("minute", "second")))
  expect_equal(as.period(int, "minute") + start, end)

  expect_equal(as.period(int, "second"), period(-c(592189323), c("second")))
  expect_equal(as.period(int, "second") + start, end)
})

test_that("as.period handles tricky intervals", {
  expect_equal(
    as.period(interval(ymd("1986-01-31"), ymd("1986-02-01"))),
    period(days = 1)
  )

  expect_equal(
    as.period(interval(ymd("1984-01-30"), ymd("1986-02-01"))),
    period(years = 2, days = 2)
  )

  expect_equal(
    as.period(interval(ymd("1984-01-30"), ymd("1986-03-01"))),
    period(years = 2, months = 1, days = 1)
  )

  expect_equal(
    as.period(interval(ymd("1985-01-30"), ymd("1986-03-30"))),
    period(years = 1, months = 2)
  )

  expect_equal(
    as.period(interval(ymd("1985-01-28"), ymd("1986-03-28"))),
    period(years = 1, months = 2)
  )

  expect_equal(
    as.period(interval(ymd("1985-01-31"), ymd("1986-03-28"))),
    period(years = 1, months = 1, days = 28)
  )

  expect_equal(
    as.period(interval(ymd("1984-01-28"), ymd("1984-02-28"))),
    period(months = 1)
  )

  expect_equal(
    as.period(interval(ymd("1984-01-28"), ymd("1984-02-29"))),
    period(months = 1, days = 1)
  )

  expect_equal(
    as.period(interval(ymd_hms("1984-01-28 5:0:0"), ymd_hms("1984-02-29 3:0:0"))),
    period(months = 1, hours = 22)
  )

  expect_equal(
    as.period(interval(ymd("1984-01-28"), ymd("1984-03-01"))),
    period(months = 1, days = 2)
  )

  expect_equal(
    as.period(interval(ymd_hms("1984-01-28 5:0:0"), ymd_hms("1984-03-01 3:0:0"))),
    period(months = 1, days = 1, hours = 22)
  )
})



test_that("as.period handles tricky negative intervals", {
  expect_equal(
    as.period(interval(ymd("1986-02-01"), ymd("1986-01-31"))),
    period(days = -1)
  )

  expect_equal(
    as.period(interval(ymd("1986-02-01"), ymd("1984-01-30"))),
    period(years = -2, days = -2)
  )

  expect_equal(
    as.period(interval(ymd("1986-03-01"), ymd("1984-01-30"))),
    period(years = -2, months = -1, days = -2)
  )

  expect_equal(
    as.period(interval(ymd("1986-03-30"), ymd("1985-01-30"))),
    period(years = -1, months = -2)
  )

  expect_equal(
    as.period(interval(ymd("1986-03-28"), ymd("1985-01-28"))),
    period(years = -1, months = -2)
  )

  expect_equal(
    as.period(interval(ymd("1986-03-28"), ymd("1985-01-31"))),
    period(years = -1, months = -1, days = -28)
  )

  expect_equal(
    as.period(interval(ymd("1986-02-01"), ymd("1985-12-31"))),
    period(months = -1, days = -1)
  )

  expect_equal(
    as.period(interval(ymd("1984-03-01"), ymd("1984-01-31"))),
    period(months = -1, days = -1)
  )

  expect_equal(
    as.period(interval(ymd("1984-03-01"), ymd("1984-01-30"))),
    period(months = -1, days = -2)
  )

  expect_equal(
    as.period(interval(ymd("1984-03-01"), ymd("1984-01-29"))),
    period(months = -1, days = -3)
  )

  expect_equal(
    as.period(interval(ymd_hms("1984-03-01 3:0:0"), ymd_hms("1984-01-28 5:0:0"))),
    period(months = -1, days = -3, hours = -22)
  )
})

test_that("as.period handles NA in interval objects", {
  one_missing_date <- as.POSIXct(NA_real_, origin = origin)
  one_missing_interval <- interval(
    one_missing_date,
    one_missing_date
  )
  several_missing_dates <- rep(as.POSIXct(NA_real_, origin = origin), 2)
  several_missing_intervals <- interval(
    several_missing_dates,
    several_missing_dates
  )
  start_missing_intervals <- interval(several_missing_dates, origin)
  end_missing_intervals <- interval(origin, several_missing_dates)
  na.per <- period(
    sec = NA, min = NA, hour = NA, day = NA,
    month = NA, year = NA
  )

  expect_equal(as.period(one_missing_interval, "year"), na.per)
  expect_equal(as.period(several_missing_intervals, "year"), c(na.per, na.per))
  expect_equal(as.period(start_missing_intervals, "year"), c(na.per, na.per))
  expect_equal(as.period(end_missing_intervals, "year"), c(na.per, na.per))
})

test_that("as.period handles NA duration objects", {
  na.per <- period(
    sec = NA, min = NA, hour = NA, day = NA,
    month = NA, year = NA
  )

  expect_equal(suppressMessages(as.period(dyears(NA))), na.per)
  expect_equal(suppressMessages(as.period(dyears(c(NA, NA)))), c(na.per, na.per))
  expect_equal(suppressMessages(as.period(dyears(c(1, NA)))), c(years(1), na.per))
})

test_that("as.period handles NA objects", {
  na.per <- seconds(NA)
  expect_equal(as.period(NA), na.per)
})

test_that("as.period handles vectors", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  time3 <- as.POSIXct("2010-08-03 13:01:59", tz = "UTC")
  int <- interval(c(time2, time1), time3)

  dur <- duration(seconds = 5, minutes = c(30, 59))

  expect_equal(as.period(int), years(1:2))
  expect_equal(as.period(dur), seconds(5) +
    minutes(c(30, 59)))
})

test_that("as.period handles duration objects", {
  dur <- duration(seconds = 5, minutes = 30)
  expect_equal(as.period(dur), seconds(5) + minutes(30))
})

test_that("as.period handles period objects", {
  per <- minutes(1000) + seconds(50) + seconds(11)
  expect_equal(as.period(per), per)
  expect_equal(as.period(per, "minute"), period(c(1001, 1), c("minute", "second")))
  expect_equal(as.period(per, "hour"), period(c(16, 41, 1), c("hour", "minute", "second")))
})

test_that("[<- can subset periods with new periods", {
  Time <- data.frame(Time = c(hms("01:01:01"), hms("02:02:02")))
  Time[1, 1] <- Time[1, 1] + hours(1)

  times <- days(1:3)
  times[1] <- times[1] + hours(2)

  expect_equal(Time[1, 1], hms("02:01:01"))
  expect_equal(times[1], period(days = 1, hours = 2))
})

test_that("period correctly handles week units", {
  expect_equal(period(1, "week"), days(7))
  expect_equal(period(8, "days"), days(8))
  expect_equal(period(c(3, 2), c("days", "week")), days(c(17)))
})

test_that("format.period correctly displays negative units", {
  expect_match(format(days(-2)), "-2d 0H 0M 0S")
  expect_match(format(period(second = -1, hour = -2, day = 3)), "3d -2H 0M -1S")
})

test_that("c.Period correctly handles NAs", {
  per <- period(seconds = 5)
  expect_true(is.na(c(per, NA)[2]))
})

test_that("NA components propagate to all components of a period", {
  p <- period(secs = c(1, 2, 3), days = c(1, 2, NA), months = c(1, NA, 3))
  expect_equal(
    p,
    new("Period",
      .Data = c(1, NA, NA), year = c(0, NA, NA),
      month = c(1, NA, NA), day = c(1, NA, NA),
      hour = c(0, NA, NA), minute = c(0, NA, NA)
    )
  )
})

test_that("c.Period doesn't fail with empty elements", {
  expect_equal(
    c(period(), period(3), NULL, logical(), period(1, "days")),
    c(period(3), period(days = 1))
  )
})

test_that("as.integer works with periods", {
  p <- period("2d 0H 2M 1s")
  expect_equal(as.integer(p, units = "minute"), as.integer(as.numeric(p, units = "minute")))
  expect_equal(as.integer(p, units = "hour"), as.integer(as.numeric(p, units = "hour")))
  expect_equal(as.integer(p), as.numeric(p))
  expect_type(as.integer(p), "integer")
})

test_that("c.Period works with named elements", {
  nmd <- c(a = hours(1), b = minutes(30))
  umd <- c(hours(1), minutes(30))
  names(umd) <- c("a", "b")
  expect_equal(nmd, umd)

  nmd <- c(a = hours(c(1, 1)), b = minutes(c(30, 31)))
  umd <- c(hours(1), hours(1), minutes(30), minutes(31))
  names(umd) <- c("a1", "a2", "b1", "b2")
  expect_equal(nmd, umd)

  ## empty elements
  nmd <- c(a = hours(1), NULL, b = minutes(30), seconds(10))
  umd <- c(hours(1), minutes(30), seconds(10))
  names(umd) <- c("a", "b", "")
  expect_equal(nmd, umd)
})

test_that("summary.Period creates useful summary", {
  per <- period(minutes = 5)
  text <- c(rep("5M 0S", 6), 1)
  names(text) <- c("Min.", "1st Qu.", "Median", "Mean", "3rd Qu.", "Max.", "NA's")

  expect_equal(summary(c(per, NA)), text)
})



test_that("idempotentcy between period_to_seconds and seconds_to_period holds", {
  period <- period(years = 2, months = 3, days = 4, hour = 5, minutes = 6, seconds = 7)
  seconds <- 71319967
  expect_equal(period_to_seconds(seconds_to_period(seconds)), seconds)
})


test_that("direct reation of periods works as expected", {
  expect_equal(
    new("Period", 1:4),
    period(seconds = 1:4)
  )

  expect_equal(
    new("Period", 1:4, day = 1:2),
    period(seconds = 1:4, days = 1:2)
  )
})

context("Periods")

test_that("period() returns zero-length vector", {
  x <- period()
  expect_s4_class(x, "Period")
  expect_length(x, 0)
  expect_equal(format(x), character())
  expect_output(print(x), "<Period[0]>", fixed = TRUE)
})

test_that("period(0-len-vec, ...) returns zero-length period", {
  x <- period(character())
  expect_s4_class(x, "Period")
  expect_length(x, 0)

  x <- period(hour = numeric())
  expect_s4_class(x, "Period")
  expect_length(x, 0)

  x <- period(numeric(), hour = numeric())
  expect_s4_class(x, "Period")
  expect_length(x, 0)

  expect_equal(
    period(c(30, 20), units = c("secs", "days")),
    period(c(30, 20), units = c("secs", "days"), days = numeric()))

  expect_equal(
    period(numeric(), days = 10),
    period(days = 10))

  expect_equal(
    period(days = 10, secs = 1, hours = numeric()),
    period())

  expect_equal(
    period(days = 10, secs = numeric(), hours = 3),
    period())
})

test_that("period constructor doesn't accept non-numeric or non-character inputs", {
  expect_error(period(interval(ymd("2014-01-01"), ymd("2015-01-01"))))
})

test_that("period constructor works with duplicated units", {
  expect_equal(
    period(c(1, 1, 2, 10), c("days", "sec", "seconds", "day")),
    period("11d 3s"))
  expect_equal(
    period(sec = 1, day = 1, seconds = 2, days = 10),
    period("11d 3s"))
  expect_equal(
    period(sec = c(1, 2), day = c(1, 2), seconds = 2, days = c(10, 20, 30, 40)),
    period(c("11d 3s", "22d 4s", "31d 3s", "42d 4s")))
})

test_that("period parsing works", {
  expect_equal(
    period("1min 2sec 2secs 1H 2M 1d"),
    period(seconds = 4, minutes = 3, hours = 1, days = 1))
  expect_equal(
    period("day day"),
    period(days = 2))
  expect_equal(
    period("S M H d m y"),
    period(seconds = 1, minutes = 1, hours = 1, days = 1, months = 1, years = 1))
  expect_equal(
    period("2S 3M 4H 5d 6w 7m 8y"),
    period(seconds = 2, minutes = 3, hours = 4, days = 5, weeks = 6, months = 7, years = 8))
  expect_equal(period("K")@.Data, NA_real_)
  expect_equal(period("K")@hour, NA_real_)
  expect_equal(period("ksfdsfds")@.Data, NA_real_)
  expect_equal(period("ksfdsfds")@day, NA_real_)
})

test_that("ISO ISO 8601 period parsing works", {
  expect_equal(period("")@.Data, NA_real_)
  expect_equal(period("P")@.Data, NA_real_)
  expect_equal(
    period("P3Y6M4DT12H30M5S"),
    period(years = 3, months = 6, days = 4, hours = 12, minutes = 30, seconds = 5))
  expect_equal(
    period("P23DT23H"),
    period(days = 23, hours = 23))
  expect_equal(
    period("P23DT60H"),
    period(days = 23, hours = 60))
  expect_equal(
    period("P23DT60H20minutes 100 sec"),
    period(days = 23, hours = 60, minutes = 20, seconds = 100))
  expect_equal(
    period(c("P23DT23H", "PT0S", "P0D")),
    period(days = c(23, 0, 0), hours = c(23, 0, 0)))
})

test_that("fractional parsing works as expected", {
  expect_equal(
    period("1.1min 2.3sec 2.3secs 1.0H 2.2M 1.5d"),
    period(seconds = 43222.6, minutes = 3, hours = 1, days = 1))
  expect_equal(
    period("day 1.2days"),
    period(days = 2, seconds = 17280))
})

test_that("is.period works", {
  expect_false(is.period(234))
  expect_false(is.period(as.POSIXct("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.period(as.POSIXlt("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.period(Sys.Date()))
  expect_true(is.period(minutes(1)))
  expect_false(is.period(dminutes(1)))
  expect_false(is.period(interval(
    as.POSIXct("2008-08-03 13:01:59", tz = "UTC"),
    as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  )))
})

test_that("period works as expected", {
  per <- period(second = 90, minute = 5)

  expect_is(per, "Period")
  expect_equal(period(hours = 25), hours(25))
  expect_equal(per@year, 0)
  expect_equal(per@month, 0)
  expect_equal(per@day, 0)
  expect_equal(per@hour, 0)
  expect_equal(per@minute, 5)
  expect_equal(per@.Data, 90)
})

test_that("period works as expected", {
  per <- period(c(90, 5), c("second", "minute"))

  expect_is(per, "Period")
  expect_equal(period(hours = 25), hours(25))
  expect_equal(per@year, 0)
  expect_equal(per@month, 0)
  expect_equal(per@day, 0)
  expect_equal(per@hour, 0)
  expect_equal(per@minute, 5)
  expect_equal(per@.Data, 90)
})

test_that("period handles vector input", {
  per <- period(second = c(90, 60), minute = 5)

  expect_is(per, "Period")
  expect_equal(length(per), 2)
  expect_equal(per@year, c(0, 0))
  expect_equal(per@month, c(0, 0))
  expect_equal(per@day, c(0, 0))
  expect_equal(per@hour, c(0, 0))
  expect_equal(per@minute, c(5, 5))
  expect_equal(per@.Data, c(90, 60))
})


test_that("period objects handle vector input", {
  x <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  expect_equal(
    as.numeric(x + minutes(c(1, 3, 4))),
    as.numeric(x + c(60, 180, 240))
  )
})

test_that("format.Period works as expected", {
  per <- period(second = 90, minute = 5)
  per2 <- period(days = 0)
  per3 <- period(years = 1, days = NA)
  expect_match(format(per), "5M 90S")
  expect_match(format(per2), "0S")
  expect_equivalent(format(per3), as.character(NA))
})

test_that("as.numeric and as.duration correctly handle periods", {
  tt <- period(24, "hours") + period(60, "minutes")
  expect_equal(as.numeric(tt, "hour"), 25)
  tt <- period(hours = 24, minutes = 6)
  expect_equal(as.numeric(tt, "hour"), 24.1)
  expect_equal(
    as.numeric(tt, "hour"),
    as.numeric(as.duration(tt), "hour")
  )
})

test_that("as.period handles interval objects", {
  start <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  end <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  int <- interval(start, end)
  int_neg <- interval(end, start)

  expect_equal(as.period(int), years(1))
  expect_equal(as.period(int_neg), years(-1))
})

test_that("as.period handles intervals across years correctly", {
  ## https://github.com/tidyverse/lubridate/issues/371

  t1 <- ymd("1985-11-30")
  t2 <- ymd("1986-02-01")
  expect_equal(t2, t1 + as.period(interval(t1, t2)))

  ## negative
  t1 <- ymd("1986-02-01")
  t2 <- ymd("1985-11-30")
  expect_equal(t2, t1 + as.period(interval(t1, t2)))

  t1 <- "1985-11-01"
  t2 <- "2007-08-01"
  as.period(interval(t1, t2))
  ymd(t1) + as.period(interval(t1, t2, "UTC"))

  ## negative
  t1 <- "2007-08-01"
  t2 <- "1985-11-01"
  as.period(interval(t1, t2))
  ymd(t1) + as.period(interval(t1, t2, "UTC"))

  t1 <- ymd("1986-02-01")
  t2 <- ymd("1986-01-05")
  expect_equal(t2, t1 + as.period(interval(t1, t2)))
})

test_that("as.period handles vector interval objects", {
  ## https://github.com/tidyverse/lubridate/issues/349
  ints <- c(
    interval(ymd("2001-01-01"), ymd("2002-01-01")),
    interval(ymd("2001-01-01"), ymd("2004-01-01"))
  )
  expect_equal(as.period(ints), period(years = c(1, 3)))

  ints <- c(
    interval(ymd("2001-01-01"), ymd("2002-03-05")),
    interval(ymd("2001-01-01"), ymd_hms("2004-12-31 3:2:1"))
  )
  expect_equal(
    as.period(ints),
    period(
      years = c(1, 3), months = c(2, 11), days = c(4, 30),
      hours = c(0, 3), minutes = c(0, 2), seconds = c(0, 1)
    )
  )
})

test_that("as.period handles don't produce negative periods", {
  p2 <- as.period(interval(ymd("1985-12-31"), ymd("1986-02-01")))
  expect_equal(month(p2), 1)

  p3 <- as.period(interval(ymd("1985-12-31"), ymd("1986-03-01")))
  expect_equal(month(p3), 2)
  expect_equal(day(p3), 1)
})

test_that("as.period handles interval objects with special start dates", {
  start <- ymd("1992-02-29")
  end <- ymd("2010-12-05")
  int <- interval(start, end)

  expect_equal(as.period(int), period(c(18, 9, 6), c("year", "month", "day")))
  expect_equal(as.period(int) + start, end)
})


test_that("as.period with different units handles interval objects", {
  start <- ymd("1992-02-29")
  end <- ymd_hms("2010-12-05 01:02:03")
  int <- interval(start, end)

  expect_equal(
    as.period(int),
    period(c(18, 9, 6, 1, 2, 3), c("year", "month", "day", "hour", "minute", "second"))
  )
  expect_equal(as.period(int) + start, end)

  expect_equal(
    as.period(int, "months"),
    period(c(225, 6, 1, 2, 3), c("month", "day", "hour", "minute", "second"))
  )
  expect_equal(as.period(int, "months") + start, end)

  expect_equal(as.period(int, "hours"), period(c(164497, 2, 3), c("hour", "minute", "second")))
  expect_equal(as.period(int, "hours") + start, end)

  expect_equal(as.period(int, "minute"), period(c(9869822, 3), c("minute", "second")))
  expect_equal(as.period(int, "minute") + start, end)

  expect_equal(as.period(int, "second"), period(c(592189323), c("second")))
  expect_equal(as.period(int, "second") + start, end)
})


test_that("as.period with different units handles negative interval objects", {
  end <- ymd("1992-02-29", tz = "UTC")
  start <- ymd_hms("2010-12-05 01:02:03")
  int <- interval(start, end)

  expect_equal(
    as.period(int),
    period(-c(18, 9, 5, 1, 2, 3), c("year", "month", "day", "hour", "minute", "second"))
  )
  expect_equal(as.period(int) + start, end)

  expect_equal(
    as.period(int, "months"),
    period(-c(225, 5, 1, 2, 3), c("month", "day", "hour", "minute", "second"))
  )
  expect_equal(as.period(int, "months") + start, end)

  expect_equal(as.period(int, "hours"), period(-c(164497, 2, 3), c("hour", "minute", "second")))
  expect_equal(as.period(int, "hours") + start, end)

  expect_equal(as.period(int, "minute"), period(-c(9869822, 3), c("minute", "second")))
  expect_equal(as.period(int, "minute") + start, end)

  expect_equal(as.period(int, "second"), period(-c(592189323), c("second")))
  expect_equal(as.period(int, "second") + start, end)
})

test_that("as.period handles tricky intervals", {
  expect_equal(
    as.period(interval(ymd("1986-01-31"), ymd("1986-02-01"))),
    period(days = 1)
  )

  expect_equal(
    as.period(interval(ymd("1984-01-30"), ymd("1986-02-01"))),
    period(years = 2, days = 2)
  )

  expect_equal(
    as.period(interval(ymd("1984-01-30"), ymd("1986-03-01"))),
    period(years = 2, months = 1, days = 1)
  )

  expect_equal(
    as.period(interval(ymd("1985-01-30"), ymd("1986-03-30"))),
    period(years = 1, months = 2)
  )

  expect_equal(
    as.period(interval(ymd("1985-01-28"), ymd("1986-03-28"))),
    period(years = 1, months = 2)
  )

  expect_equal(
    as.period(interval(ymd("1985-01-31"), ymd("1986-03-28"))),
    period(years = 1, months = 1, days = 28)
  )

  expect_equal(
    as.period(interval(ymd("1984-01-28"), ymd("1984-02-28"))),
    period(months = 1)
  )

  expect_equal(
    as.period(interval(ymd("1984-01-28"), ymd("1984-02-29"))),
    period(months = 1, days = 1)
  )

  expect_equal(
    as.period(interval(ymd_hms("1984-01-28 5:0:0"), ymd_hms("1984-02-29 3:0:0"))),
    period(months = 1, hours = 22)
  )

  expect_equal(
    as.period(interval(ymd("1984-01-28"), ymd("1984-03-01"))),
    period(months = 1, days = 2)
  )

  expect_equal(
    as.period(interval(ymd_hms("1984-01-28 5:0:0"), ymd_hms("1984-03-01 3:0:0"))),
    period(months = 1, days = 1, hours = 22)
  )
})



test_that("as.period handles tricky negative intervals", {
  expect_equal(
    as.period(interval(ymd("1986-02-01"), ymd("1986-01-31"))),
    period(days = -1)
  )

  expect_equal(
    as.period(interval(ymd("1986-02-01"), ymd("1984-01-30"))),
    period(years = -2, days = -2)
  )

  expect_equal(
    as.period(interval(ymd("1986-03-01"), ymd("1984-01-30"))),
    period(years = -2, months = -1, days = -2)
  )

  expect_equal(
    as.period(interval(ymd("1986-03-30"), ymd("1985-01-30"))),
    period(years = -1, months = -2)
  )

  expect_equal(
    as.period(interval(ymd("1986-03-28"), ymd("1985-01-28"))),
    period(years = -1, months = -2)
  )

  expect_equal(
    as.period(interval(ymd("1986-03-28"), ymd("1985-01-31"))),
    period(years = -1, months = -1, days = -28)
  )

  expect_equal(
    as.period(interval(ymd("1986-02-01"), ymd("1985-12-31"))),
    period(months = -1, days = -1)
  )

  expect_equal(
    as.period(interval(ymd("1984-03-01"), ymd("1984-01-31"))),
    period(months = -1, days = -1)
  )

  expect_equal(
    as.period(interval(ymd("1984-03-01"), ymd("1984-01-30"))),
    period(months = -1, days = -2)
  )

  expect_equal(
    as.period(interval(ymd("1984-03-01"), ymd("1984-01-29"))),
    period(months = -1, days = -3)
  )

  expect_equal(
    as.period(interval(ymd_hms("1984-03-01 3:0:0"), ymd_hms("1984-01-28 5:0:0"))),
    period(months = -1, days = -3, hours = -22)
  )
})

test_that("as.period handles NA in interval objects", {
  one_missing_date <- as.POSIXct(NA_real_, origin = origin)
  one_missing_interval <- interval(
    one_missing_date,
    one_missing_date
  )
  several_missing_dates <- rep(as.POSIXct(NA_real_, origin = origin), 2)
  several_missing_intervals <- interval(
    several_missing_dates,
    several_missing_dates
  )
  start_missing_intervals <- interval(several_missing_dates, origin)
  end_missing_intervals <- interval(origin, several_missing_dates)
  na.per <- period(
    sec = NA, min = NA, hour = NA, day = NA,
    month = NA, year = NA
  )

  expect_equal(as.period(one_missing_interval, "year"), na.per)
  expect_equal(as.period(several_missing_intervals, "year"), c(na.per, na.per))
  expect_equal(as.period(start_missing_intervals, "year"), c(na.per, na.per))
  expect_equal(as.period(end_missing_intervals, "year"), c(na.per, na.per))
})

test_that("as.period handles NA duration objects", {
  na.per <- period(
    sec = NA, min = NA, hour = NA, day = NA,
    month = NA, year = NA
  )

  expect_equal(suppressMessages(as.period(dyears(NA))), na.per)
  expect_equal(suppressMessages(as.period(dyears(c(NA, NA)))), c(na.per, na.per))
  expect_equal(suppressMessages(as.period(dyears(c(1, NA)))), c(years(1), na.per))
})

test_that("as.period handles NA objects", {
  na.per <- seconds(NA)
  expect_equal(as.period(NA), na.per)
})

test_that("as.period handles vectors", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  time3 <- as.POSIXct("2010-08-03 13:01:59", tz = "UTC")
  int <- interval(c(time2, time1), time3)

  dur <- duration(seconds = 5, minutes = c(30, 59))

  expect_equal(as.period(int), years(1:2))
  expect_equal(as.period(dur), seconds(5) +
    minutes(c(30, 59)))
})

test_that("as.period handles duration objects", {
  dur <- duration(seconds = 5, minutes = 30)
  expect_equal(as.period(dur), seconds(5) + minutes(30))
})

test_that("as.period handles period objects", {
  per <- minutes(1000) + seconds(50) + seconds(11)
  expect_equal(as.period(per), per)
  expect_equal(as.period(per, "minute"), period(c(1001, 1), c("minute", "second")))
  expect_equal(as.period(per, "hour"), period(c(16, 41, 1), c("hour", "minute", "second")))
})

test_that("[<- can subset periods with new periods", {
  Time <- data.frame(Time = c(hms("01:01:01"), hms("02:02:02")))
  Time[1, 1] <- Time[1, 1] + hours(1)

  times <- days(1:3)
  times[1] <- times[1] + hours(2)

  expect_equal(Time[1, 1], hms("02:01:01"))
  expect_equal(times[1], period(days = 1, hours = 2))
})

test_that("period correctly handles week units", {
  expect_equal(period(1, "week"), days(7))
  expect_equal(period(8, "days"), days(8))
  expect_equal(period(c(3, 2), c("days", "week")), days(c(17)))
})

test_that("format.period correctly displays negative units", {
  expect_match(format(days(-2)), "-2d 0H 0M 0S")
  expect_match(format(period(second = -1, hour = -2, day = 3)), "3d -2H 0M -1S")
})

test_that("c.Period correctly handles NAs", {
  per <- period(seconds = 5)
  expect_true(is.na(c(per, NA)[2]))
})

test_that("NA components propagate to all components of a period", {
  p <- period(secs = c(1, 2, 3), days = c(1, 2, NA), months = c(1, NA, 3))
  expect_equal(p,
               new("Period", .Data = c(1, NA, NA), year = c(0, NA, NA),
                   month = c(1, NA, NA), day = c(1, NA, NA),
                   hour = c(0, NA, NA), minute = c(0, NA, NA)))
})

test_that("c.Period doesn't fail with empty elements", {
  expect_equal(
    c(period(), period(3), NULL, logical(), period(1, "days")),
    c(period(3), period(days = 1))
  )
})

test_that("summary.Period creates useful summary", {
  per <- period(minutes = 5)
  text <- c(rep("5M 0S", 6), 1)
  names(text) <- c("Min.", "1st Qu.", "Median", "Mean", "3rd Qu.", "Max.", "NA's")

  expect_equal(summary(c(per, NA)), text)
})

test_that("idempotentcy between period_to_seconds and seconds_to_period holds", {
  period <- period(years = 2, months = 3, days = 4, hour = 5, minutes = 6, seconds = 7)
  seconds <- 71319967
  expect_equal(period_to_seconds(seconds_to_period(seconds)), seconds)
})


test_that("direct reation of periods works as expected", {
  expect_equal(
    new("Period", 1:4),
    period(seconds = 1:4)
  )

  expect_equal(
    new("Period", 1:4, day = 1:2),
    period(seconds = 1:4, days = 1:2)
  )
})

test_that("addition handles daylight savings time", {
  x <- as.POSIXct("2010-03-14 00:00:00", tz = "America/New_York")
  y <- as.POSIXct("2010-03-15 01:00:00", tz = "America/New_York")

  expect_equal(x + days(1), as.POSIXct(
    "2010-03-15 00:00:00",
    tz = "America/New_York"
  ))
  expect_equal(x + ddays(1), y)
})

test_that("subtraction handles daylight savings time", {
  x <- as.POSIXct("2010-03-15 00:00:00", tz = "America/New_York")
  y <- as.POSIXct("2010-03-13 23:00:00", tz = "America/New_York")

  expect_equal(x - days(1), as.POSIXct(
    "2010-03-14 00:00:00",
    tz = "America/New_York"
  ))
  expect_equal(x - ddays(1), y)
})

test_that("addition works as expected for instants", {
  x <- as.POSIXct("2008-01-01 00:00:00", tz = "UTC")
  y <- as.POSIXlt("2008-01-01 00:00:00", tz = "UTC")
  z <- as.Date("2008-01-01")

  expect_equal(x + years(1), as.POSIXct("2009-01-01 00:00:00", tz = "UTC"))
  expect_equal(y + years(1), as.POSIXlt("2009-01-01 00:00:00", tz = "UTC"))
  expect_equal(z + years(1), as.Date("2009-01-01"))

  expect_equal(x + dyears(1), as.POSIXct("2008-12-31 06:00:00", tz = "UTC"))
  expect_equal(y + dyears(1), as.POSIXlt("2008-12-31 06:00:00", tz = "UTC"))
  expect_equal(z + dyears(1), as.POSIXct("2008-12-31 06:00:00", tz = "UTC"))

  time1 <- as.POSIXct("2008-08-02 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  int <- interval(time1, time2)
  num <- as.numeric(time2) - as.numeric(time1)

  expect_error(x + int)
  expect_error(y + int)
  expect_error(z + int)
})

test_that("addition with instants returns correct class", {
  x <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  y <- as.POSIXlt("2008-01-02 00:00:00", tz = "UTC")
  z <- as.Date("2008-01-01")

  expect_s3_class(x + years(1), "POSIXct")
  expect_s3_class(y + years(1), "POSIXlt")
  expect_s3_class(z + years(1), "Date")

  expect_s3_class(x + dyears(1), "POSIXct")
  expect_s3_class(y + dyears(1), "POSIXlt")
  expect_s3_class(z + dyears(1), "POSIXct")
})



test_that("addition works as expected for periods", {
  time1 <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  int <- interval(time1, time2)

  expect_equal(years(1) + 1, period(seconds = 1, years = 1))

  expect_equal(
    years(1) + as.POSIXct("2008-01-01 00:00:00", tz = "UTC"),
    as.POSIXct("2009-01-01 00:00:00", tz = "UTC")
  )

  expect_equal(
    years(1) + as.POSIXlt("2008-01-01 00:00:00", tz = "UTC"),
    as.POSIXlt("2009-01-01 00:00:00", tz = "UTC")
  )

  expect_equal(years(1) + minutes(3), period(
    minutes = 3,
    years = 1
  ))

  expect_error(years(1) + dyears(1))
  expect_error(years(1) + int)
})

test_that("addition with periods returns correct class", {
  expect_s4_class(years(1) + 1, "Period")

  expect_s3_class(years(1) + as.POSIXct(
    "2008-01-01 00:00:00",
    tz = "UTC"
  ), "POSIXt")

  expect_s3_class(years(1) + as.POSIXlt(
    "2008-01-01 00:00:00",
    tz = "UTC"
  ), "POSIXlt")

  expect_s4_class(years(1) + minutes(3), "Period")
})


test_that("addition works as expected for durations", {
  x <- as.POSIXct("2008-01-01 00:00:00", tz = "UTC")
  y <- as.POSIXct("2008-12-31 06:00:00", tz = "UTC")
  time1 <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  int <- interval(time1, time2)

  expect_equal(dyears(1) + 1, duration(31557601))
  expect_equal(dyears(1) + x, y)
  expect_equal(dyears(1) + dyears(1), dyears(2))
  expect_error(dyears(1) + minutes(3))
  expect_error(dyears(1) + int)
})

test_that("addition with durations returns correct class", {
  ct <- as.POSIXct("2008-01-01 00:00:00", tz = "UTC")
  lt <- as.POSIXlt("2008-01-01 00:00:00", tz = "UTC")

  expect_s4_class(dyears(1) + 1, "Duration")
  expect_s3_class(dyears(1) + ct, "POSIXct")
  expect_s3_class(dyears(1) + lt, "POSIXlt")
  expect_s4_class(dyears(1) + dyears(1), "Duration")
})

test_that("addition works as expected for intervals", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  time3 <- as.POSIXct("2008-01-01 00:00:00", tz = "UTC")
  diff <- difftime(time2, time1)

  int <- interval(time1, time2)

  expect_error(int + time3)
  expect_error(int + time1)
  expect_error(int + 1)
  expect_error(int + minutes(3))
  expect_error(int + dyears(1))

  time5 <- as.POSIXct("2010-08-03 00:00:00", tz = "UTC")
  int3 <- interval(time2, time5)

  expect_error(int + int3)
})


#### Vectors

test_that("adding vectors works as expected for instants", {
  x <- as.POSIXct(c("2008-01-01 00:00:00", "2009-01-01 00:00:00"), tz = "UTC")
  y <- as.POSIXlt(c("2008-01-01 00:00:00", "2009-01-01 00:00:00"), tz = "UTC")
  z <- c(as.Date("2008-01-01"), as.Date("2008-01-10"))

  expect_equal(x + years(1), ymd_hms(c("2009-01-01 00:00:00", "2010-01-01 00:00:00")))
  expect_equal(y + years(1), as.POSIXlt(c("2009-01-01 00:00:00", "2010-01-01 00:00:00"), tz = "UTC"))
  expect_equal(z + years(1), as.Date(c("2009-01-01", "2009-01-10")))

  expect_equal(x + dyears(1), ymd_hms(c("2008-12-31 06:00:00", "2010-01-01 06:00:00")))
  expect_equal(y + dyears(1), as.POSIXlt(c("2008-12-31 06:00:00", "2010-01-01 06:00:00"), tz = "UTC"))
  expect_equal(z + dyears(1), ymd_hms(c("2008-12-31 06:00:00", "2009-01-09 06:00:00")))

  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  int <- interval(time1, time2)
  num <- as.numeric(time2) - as.numeric(time1)

  expect_error(x + int)
  expect_error(y + int)
  expect_error(z + int)
})

test_that("adding vectors works as expected for periods", {
  expect_equal(years(1:2) + 1, period(seconds = 1, years = c(1, 2)))

  expect_equal(
    years(1:2) + as.POSIXct("2008-01-01 00:00:00", tz = "UTC"),
    as.POSIXct(c("2009-01-01 00:00:00", "2010-01-01 00:00:00"), tz = "UTC")
  )

  expect_equal(years(1:2) + as.POSIXlt("2008-01-01 00:00:00",
    tz = "UTC"
  ), as.POSIXlt(c(
    "2009-01-01 00:00:00",
    "2010-01-01 00:00:00"
  ), tz = "UTC"))


  expect_equal(years(1:2) + minutes(3), period(
    minutes = 3, years = c(1, 2)
  ))

  expect_error(years(1:2) + dyears(1))

  time1 <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  int <- interval(time1, time2)

  expect_error(years(1:2) + int)
})

test_that("adding vectors works as expected for durations", {
  w <- as.POSIXct("2007-01-01 00:00:00", tz = "UTC")
  x <- as.POSIXct("2008-01-01 00:00:00", tz = "UTC")
  y <- as.POSIXct(c("2008-01-01 00:00:00", "2008-12-31 00:00:00"), tz = "UTC")
  dur <- dminutes(1:2) + 1
  ddif <- ddays(c(.25, .5))
  expect_equal(dur@.Data, c(61, 121))
  expect_equal(dyears(1:2) + w - ddif, y)
  expect_equal(dyears(1:2) + as.POSIXlt(w) - ddif, as.POSIXlt(y))
  expect_error(dyears(1:2) + minutes(3))
  expect_equal(dyears(1:2) + dyears(1), dyears(2:3))

  time1 <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  int <- interval(time1, time2)
  expect_error(dyears(1:2) + int)
})

test_that("adding vectors works as expected for intervals", {
  time1 <- as.POSIXct("2008-08-03 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  time3 <- as.POSIXct("2010-08-03 00:00:00", tz = "UTC")
  int <- interval(c(time1, time2), time3)
  int2 <- interval(time1, time2)

  expect_error(int + time4)
  expect_error(int + 1)
  expect_error(int + minutes(3))
  expect_error(int + dyears(1))
  expect_error(int + int2)
})

test_that("%m+% correctly adds months without rollover", {
  jan <- ymd_hms("2010-01-31 03:04:05")
  ends <- ymd_hms(c("2010-02-28 03:04:05", "2010-03-31 03:04:05", "2010-04-30 03:04:05"))

  expect_equal(jan %m+% months(1:3), ends)
})

test_that("%m+% correctly adds years without rollover", {
  leap <- ymd("2012-02-29")
  next1 <- ymd("2013-02-28")
  next2 <- ymd("2013-03-29")
  expect_equal(leap %m+% years(1), next1)
  expect_equal(leap %m+% period(years = 1, months = 1), next2)
})

test_that("%m+% correctly adds years, months, days and HMS (#286)", {
  date <- ymd("2012-02-29")
  per_all <- period(years = 1, months = 1, days = 2, hours = 1, minutes = 20, seconds = 30)
  per_major <- period(years = 1, months = 1)
  per_minor <- period(days = 2, hours = 1, minutes = 20, seconds = 30)
  expect_equal(
    date %m+% per_all,
    date %m+% per_major + per_minor
  )

  date <- ymd("2012-03-31")
  per_all <- period(months = 3, days = 30, hours = 1, minutes = 20, seconds = 30)
  per_major <- period(months = 3)
  per_minor <- period(days = 30, hours = 1, minutes = 20, seconds = 30)
  expect_equal(
    date %m+% per_all,
    date %m+% per_major + per_minor
  )
})

test_that("add_with_rollback doesn't chock on DTS", {
  x <- ymd_hms(c("2019-03-29 02:03:00", "2019-03-29 02:11:00"), tz = "Europe/Paris")
  y <- ymd_hms(c("2020-03-29 03:00:00", "2020-03-29 03:00:00"), tz = "Europe/Paris")
  expect_equal(add_with_rollback(x, years(1)), y)
})

test_that("%m+% correctly adds negative months without rollover", {
  may <- ymd_hms("2010-05-31 03:04:05")
  ends <- ymd_hms(c("2010-04-30 03:04:05", "2010-03-31 03:04:05", "2010-02-28 03:04:05"))

  expect_equal(may %m+% -months(1:3), ends)
})

test_that("%m+% correctly adds negative years without rollover", {
  leap <- ymd("2012-02-29")
  next1 <- ymd("2011-02-28")
  next2 <- ymd("2011-01-29")
  next3 <- ymd("2011-03-29")

  expect_equal(leap %m+% years(-1), next1)
  expect_equal(leap %m+% period(years = -1, months = -1), next2)
  expect_equal(leap %m+% period(years = -1, months = 1), next3)
})


test_that("addition with period months and years returns NA when appropriate", {
  jan <- ymd("2013-01-31", tz = "America/Chicago")
  feb <- ymd("2013-02-28", tz = "America/Chicago")
  mar <- ymd("2013-03-28", tz = "America/Chicago")
  leap <- ymd("2012-02-29", tz = "America/Chicago")

  mos <- ymd(c(
    "2013-01-31", NA, "2013-03-31", NA,
    "2013-05-31", NA, "2013-07-31",
    "2013-08-31", NA, "2013-10-31", NA,
    "2013-12-31"
  ), tz = "America/Chicago")
  yrs <- ymd(c("2012-02-29", NA, NA, NA, "2016-02-29"), tz = "America/Chicago")

  mos2 <- ymd(c(
    "2013-01-31", "2012-12-31", NA,
    "2012-10-31", NA, "2012-08-31",
    "2012-07-31", NA, "2012-05-31", NA,
    "2012-03-31", NA
  ), tz = "America/Chicago")
  yrs2 <- ymd(c("2012-02-29", NA, NA, NA, "2008-02-29"), tz = "America/Chicago")

  expect_equal(jan + months(0:11), mos)
  expect_equal(feb + months(1), mar)
  expect_equal(leap + years(0:4), yrs)

  expect_equal(jan - months(0:11), mos2)
  expect_equal(mar - months(1), feb)
  expect_equal(leap - years(0:4), yrs2)

  # Again with Dates (#1069)
  jan <- as_date(jan)
  feb <- as_date(feb)
  mar <- as_date(mar)
  leap <- as_date(leap)
  mos <- as_date(mos)
  yrs <- as_date(yrs)
  mos2 <- as_date(mos2)
  yrs2 <- as_date(yrs2)

  expect_equal(jan + months(0:11), mos)
  expect_equal(feb + months(1), mar)
  expect_equal(leap + years(0:4), yrs)

  expect_equal(jan - months(0:11), mos2)
  expect_equal(mar - months(1), feb)
  expect_equal(leap - years(0:4), yrs2)
})

test_that("addition with period months recycles correctly", {
  x <- as_date(c("2019-01-01", "2019-01-02"))
  y <- as_datetime(x, tz = "UTC")

  period <- months(1)
  x_expect <- as_date(c("2019-02-01", "2019-02-02"))
  y_expect <- as_datetime(x_expect, tz = "UTC")
  expect_identical(x + period, x_expect)
  expect_identical(y + period, y_expect)

  period <- months(NA_integer_)
  x_expect <- as_date(c(NA, NA))
  y_expect <- as_datetime(x_expect, tz = "UTC")
  expect_identical(x + period, x_expect)
  expect_identical(y + period, y_expect)

  period <- months(integer())
  expect_error(x + period)
  expect_error(y + period)

  period <- months(1:3)
  expect_error(x + period)
  expect_error(y + period)

  period <- months(c(1, 2, NA))
  expect_error(x + period)
  expect_error(y + period)
})

test_that("adding multi unit periods produces NAs on intermediate computation", {
  expect_equal(ymd("2021-01-31") + period("1m1d"), NA_Date_)
  expect_equal(ymd("2020-02-29") + period("1y1d"), NA_Date_)
  expect_equal(ymd("2020-01-29") + period("1y1m1d"), NA_Date_)
  expect_equal(ymd("2021-01-31", tz = "UTC") + period("1m1d"), NA_POSIXct_)
  expect_equal(ymd("2020-02-29", tz = "UTC") + period("1y1d"), NA_POSIXct_)
  expect_equal(ymd("2020-01-29", tz = "UTC") + period("1y1m1d"), NA_POSIXct_)
})

test_that("addition with singleton NAs periods work", {
  days1 <- days(NA_real_)
  months1 <- months(NA_real_)
  years1 <- years(NA_real_)

  x <- as.Date("2019-01-01")
  y <- as.POSIXct("2019-01-01", "UTC")

  expect_equal(x + days1, as.Date(NA))
  expect_equal(y + days1, NA_POSIXct_)
  expect_equal(x + months1, as.Date(NA))
  expect_equal(y + months1, NA_POSIXct_)
  expect_equal(x + years1, as.Date(NA))
  expect_equal(y + years1, NA_POSIXct_)
})

test_that("addition with durations containing NA", {
  dt <- ymd(20161018)
  dt_1 <- ymd(20161019)

  dd_1na <- ddays(c(1, NA))
  dd_na1 <- ddays(c(NA, 1))
  dd_nana <- ddays(c(NA, NA))

  ans_1na <- add_duration_to_date(dd_1na, dt)
  ans_na1 <- add_duration_to_date(dd_na1, dt)
  ans_nana <- add_duration_to_date(dd_nana, dt)

  expect_s3_class(ans_1na, "Date")
  expect_s3_class(ans_na1, "Date")
  expect_s3_class(ans_nana, "Date")

  expect_equal(ans_1na, c(dt_1, NA))
  expect_equal(ans_na1, rev(c(dt_1, NA)))
  expect_equal(ans_nana, as.Date(c(NA, NA)))
})


test_that("addition works correctly for DST transitions", {
  ref <- ymd("2017-10-01", tz = "Australia/Melbourne")
  expect_equal(ref + hours(1:3), ref + dhours(c(1, NA, 2)))
  ref <- ymd_hms("2022-10-30 00:00:00", tz = "Europe/Amsterdam")
  expect_equal(ref + hours(1:3), ref + dhours(c(1, 2, 4)))
  ref <- ymd_hms("2022-03-27 00:00:00", tz = "Europe/Amsterdam")
  expect_equal(ref + hours(1:3), ref + dhours(c(1, NA, 2)))
})

context("Addition operations")

test_that("addition handles daylight savings time", {
  x <- as.POSIXct("2010-03-14 00:00:00", tz = "America/New_York")
  y <- as.POSIXct("2010-03-15 01:00:00", tz = "America/New_York")

  expect_that(x + days(1), equals(as.POSIXct(
    "2010-03-15 00:00:00", tz = "America/New_York")))
  expect_that(x + ddays(1), equals(y))
})

test_that("subtraction handles daylight savings time", {
  x <- as.POSIXct("2010-03-15 00:00:00", tz = "America/New_York")
  y <- as.POSIXct("2010-03-13 23:00:00", tz = "America/New_York")

  expect_that(x - days(1), equals(as.POSIXct(
    "2010-03-14 00:00:00", tz = "America/New_York")))
  expect_that(x - ddays(1), equals(y))
})

test_that("addition works as expected for instants", {

  x <- as.POSIXct("2008-01-01 00:00:00", tz = "UTC")
  y <- as.POSIXlt("2008-01-01 00:00:00", tz = "UTC")
  z <- as.Date("2008-01-01")

  expect_equal(x + years(1), as.POSIXct("2009-01-01 00:00:00", tz = "UTC"))
  expect_equal(y + years(1), as.POSIXlt("2009-01-01 00:00:00", tz = "UTC"))
  expect_equal(z + years(1), as.Date("2009-01-01"))

  expect_that(x + dyears(1), equals(as.POSIXct("2008-12-31 06:00:00", tz = "UTC")))
  expect_that(y + dyears(1), equals(as.POSIXct("2008-12-31 06:00:00", tz = "UTC")))
  expect_that(z + dyears(1), equals(as.POSIXct("2008-12-31 06:00:00", tz = "UTC")))

  time1 <- as.POSIXct("2008-08-02 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  int <- interval(time1, time2)
  num <- as.numeric(time2) - as.numeric(time1)

  expect_error(x + int)
  expect_error(y + int)
  expect_error(z + int)
})

test_that("addition with instants returns correct class", {
  x <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  y <- as.POSIXlt("2008-01-02 00:00:00", tz = "UTC")
  z <- as.Date("2008-01-01")

  expect_is(x + years(1), "POSIXct")
  expect_is(y + years(1), "POSIXlt")
  expect_is(z + years(1), "Date")

  expect_that(x + dyears(1), is_a("POSIXct"))
  expect_that(y + dyears(1), is_a("POSIXlt"))
  expect_that(z + dyears(1), is_a("POSIXct"))
})



test_that("addition works as expected for periods", {
  time1 <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  int <- interval(time1, time2)

  expect_that(years(1) + 1, equals(period(seconds = 1,
    years = 1)))

  expect_that(years(1) + as.POSIXct("2008-01-01 00:00:00", tz = "UTC"),
    equals(as.POSIXct("2009-01-01 00:00:00", tz = "UTC")))

  expect_that(years(1) + as.POSIXlt("2008-01-01 00:00:00", tz = "UTC"),
    equals(as.POSIXlt("2009-01-01 00:00:00", tz = "UTC")))

  expect_that(years(1) + minutes(3), equals(period(minutes = 3,
    years = 1)))

  expect_error(years(1) + dyears(1))
  expect_error(years(1) + int)
})

test_that("addition with periods returns correct class", {

  expect_is(years(1) + 1, "Period")

  expect_is(years(1) + as.POSIXct(
    "2008-01-01 00:00:00", tz = "UTC"), "POSIXt")

  expect_is(years(1) + as.POSIXlt(
    "2008-01-01 00:00:00", tz = "UTC"), "POSIXlt")

  expect_is(years(1) + minutes(3), "Period")
})


test_that("addition works as expected for durations", {

  x <- as.POSIXct("2008-01-01 00:00:00", tz = "UTC")
  y <- as.POSIXct("2008-12-31 06:00:00", tz = "UTC")
  time1 <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  int <- interval(time1, time2)

  expect_equal(dyears(1) + 1, duration(31557601))
  expect_equal(dyears(1) + x, y)
  expect_equal(dyears(1) + dyears(1), dyears(2))
  expect_error(dyears(1) + minutes(3))
  expect_error(dyears(1) + int)
})

test_that("addition with durations returns correct class", {
  ct <- as.POSIXct("2008-01-01 00:00:00", tz = "UTC")
  lt <- as.POSIXlt("2008-01-01 00:00:00", tz = "UTC")

  expect_that(dyears(1) + 1, is_a("Duration"))
  expect_that(dyears(1) + ct, is_a("POSIXct"))
  expect_that(dyears(1) + lt, is_a("POSIXlt"))
  expect_that(dyears(1) + dyears(1), is_a("Duration"))
})

test_that("addition works as expected for intervals", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  time3 <- as.POSIXct("2008-01-01 00:00:00", tz = "UTC")
  diff <- difftime(time2, time1)

  int <- interval(time1, time2)

  expect_error(int + time3)
  expect_error(int + time1)
  expect_error(int + 1)
  expect_error(int + minutes(3))
  expect_error(int + dyears(1))

  time5 <- as.POSIXct("2010-08-03 00:00:00", tz = "UTC")
  int3 <- interval(time2, time5)

  expect_error(int + int3)
})


#### Vectors

test_that("adding vectors works as expected for instants", {
  x <- as.POSIXct(c("2008-01-01 00:00:00", "2009-01-01 00:00:00"), tz = "UTC")
  y <- as.POSIXlt(c("2008-01-01 00:00:00", "2009-01-01 00:00:00"), tz = "UTC")
  z <- c(as.Date("2008-01-01"), as.Date("2008-01-10"))

  expect_equal(x + years(1), ymd_hms(c("2009-01-01 00:00:00", "2010-01-01 00:00:00")))
  expect_equal(y + years(1), as.POSIXlt(c( "2009-01-01 00:00:00", "2010-01-01 00:00:00"), tz = "UTC"))
  expect_equal(z + years(1), as.Date(c("2009-01-01", "2009-01-10")))

  expect_equal(x + dyears(1), ymd_hms(c("2008-12-31 06:00:00", "2010-01-01 06:00:00")))
  expect_equal(y + dyears(1), as.POSIXlt(c("2008-12-31 06:00:00", "2010-01-01 06:00:00"), tz = "UTC"))
  expect_equal(z + dyears(1), ymd_hms(c("2008-12-31 06:00:00", "2009-01-09 06:00:00")))

  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  int <- interval(time1, time2)
  num <- as.numeric(time2) - as.numeric(time1)

  expect_error(x + int)
  expect_error(y + int)
  expect_error(z + int)
})

test_that("adding vectors works as expected for periods", {

  expect_that(years(1:2) + 1, equals(period(seconds = 1,
    years = c(1, 2))))

  expect_that(years(1:2) + as.POSIXct("2008-01-01 00:00:00", tz = "UTC"),
    equals(as.POSIXct(c("2009-01-01 00:00:00",
    "2010-01-01 00:00:00"), tz = "UTC")))

  expect_that(years(1:2) + as.POSIXlt("2008-01-01 00:00:00",
    tz = "UTC"), equals(as.POSIXlt(c("2009-01-01 00:00:00",
    "2010-01-01 00:00:00"), tz = "UTC")))


  expect_that(years(1:2) + minutes(3), equals(period(
    minutes = 3, years = c(1, 2))))

  expect_error(years(1:2) + dyears(1))

  time1 <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  int <- interval(time1, time2)

  expect_error(years(1:2) + int)
})

test_that("adding vectors works as expected for durations", {
  w <- as.POSIXct("2007-01-01 00:00:00", tz = "UTC")
  x <- as.POSIXct("2008-01-01 00:00:00", tz = "UTC")
  y <- as.POSIXct(c("2008-01-01 00:00:00", "2008-12-31 00:00:00"), tz = "UTC")
  dur <- dminutes(1:2) + 1
  ddif <- ddays(c(.25, .5))
  expect_equal(dur@.Data, c(61, 121))
  expect_equal(dyears(1:2) + w - ddif, y)
  expect_that(dyears(1:2) + as.POSIXlt(w) - ddif, equals(as.POSIXlt(y)))
  expect_error(dyears(1:2) + minutes(3))
  expect_that(dyears(1:2) + dyears(1), equals(dyears(2:3)))

  time1 <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  int <- interval(time1, time2)
  expect_error(dyears(1:2) + int)
})

test_that("adding vectors works as expected for intervals", {
  time1 <- as.POSIXct("2008-08-03 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  time3 <- as.POSIXct("2010-08-03 00:00:00", tz = "UTC")
  int <- interval(c(time1, time2), time3)
  int2 <- interval(time1, time2)

  expect_error(int + time4)
  expect_error(int + 1)
  expect_error(int + minutes(3))
  expect_error(int + dyears(1))
  expect_error(int + int2)

})

test_that("%m+% correctly adds months without rollover", {
  jan <- ymd_hms("2010-01-31 03:04:05")
  ends <- ymd_hms(c("2010-02-28 03:04:05", "2010-03-31 03:04:05", "2010-04-30 03:04:05"))

  expect_equal(jan %m+% months(1:3), ends)
})

test_that("%m+% correctly adds years without rollover", {
  leap <- ymd("2012-02-29")
  next1 <- ymd("2013-02-28")
  next2 <- ymd("2013-03-29")
  expect_equal(leap %m+% years(1), next1)
  expect_equal(leap %m+% period(years = 1, months = 1), next2)
})

test_that("%m+% correctly adds years, months, days and HMS (#286)", {
  date <- ymd("2012-02-29")
  per_all <- period(years = 1, months = 1, days = 2, hours = 1, minutes = 20, seconds = 30)
  per_major <- period(years = 1, months = 1)
  per_minor <- period(days = 2, hours = 1, minutes = 20, seconds = 30)
  expect_equal(date %m+% per_all,
               date %m+% per_major + per_minor)

  date <- ymd("2012-03-31")
  per_all <- period(months = 3, days = 30, hours = 1, minutes = 20, seconds = 30)
  per_major <- period(months = 3)
  per_minor <- period(days = 30, hours = 1, minutes = 20, seconds = 30)
  expect_equal(date %m+% per_all,
               date %m+% per_major + per_minor)
})

test_that("add_with_rollback doesn't chock on DTS", {
  x <- ymd_hms(c("2019-03-29 02:03:00", "2019-03-29 02:11:00"), tz = "Europe/Paris")
  y <- ymd_hms(c("2020-03-29 03:00:00", "2020-03-29 03:00:00"), tz = "Europe/Paris")
  expect_equal(add_with_rollback(x, years(1)), y)
})

test_that("%m+% correctly adds negative months without rollover", {
  may <- ymd_hms("2010-05-31 03:04:05")
  ends <- ymd_hms(c("2010-04-30 03:04:05", "2010-03-31 03:04:05", "2010-02-28 03:04:05"))

  expect_equal(may %m+% -months(1:3), ends)
})

test_that("%m+% correctly adds negative years without rollover", {
  leap <- ymd("2012-02-29")
  next1 <- ymd("2011-02-28")
  next2 <- ymd("2011-01-29")
  next3 <- ymd("2011-03-29")

  expect_equal(leap %m+% years(-1), next1)
  expect_equal(leap %m+% period(years = -1, months = -1), next2)
  expect_equal(leap %m+% period(years = -1, months = 1), next3)
})


test_that("addition with period months and years returns NA when appropriate", {
  jan <- ymd("2013-01-31", tz = "America/Chicago")
  feb <- ymd("2013-02-28", tz = "America/Chicago")
  mar <- ymd("2013-03-28", tz = "America/Chicago")
  leap <- ymd("2012-02-29", tz = "America/Chicago")

  mos <- ymd(c("2013-01-31", NA, "2013-03-31", NA,
               "2013-05-31", NA, "2013-07-31",
               "2013-08-31", NA, "2013-10-31", NA,
               "2013-12-31"), tz = "America/Chicago")
  yrs <- ymd(c("2012-02-29", NA, NA, NA, "2016-02-29"), tz = "America/Chicago")

  mos2 <- ymd(c("2013-01-31", "2012-12-31", NA,
               "2012-10-31", NA, "2012-08-31",
               "2012-07-31", NA, "2012-05-31", NA,
               "2012-03-31", NA), tz = "America/Chicago")
  yrs2 <- ymd(c("2012-02-29", NA, NA, NA, "2008-02-29"), tz = "America/Chicago")



  expect_equal(jan + months(0:11), mos)
  expect_equal(feb + months(1), mar)
  expect_equal(leap + years(0:4), yrs)

  expect_equal(jan - months(0:11), mos2)
  expect_equal(mar - months(1), feb)
  expect_equal(leap - years(0:4), yrs2)

})

test_that("addition with singleton NAs periods work", {
  days1 <- days(NA_real_)
  months1 <- months(NA_real_)
  years1 <- years(NA_real_)

  x <- as.Date("2019-01-01")
  y <- as.POSIXct("2019-01-01", "UTC")

  expect_equal(x + days1, as.Date(NA))
  expect_equal(y + days1, NA_POSIXct_)
  expect_equal(x + months1, as.Date(NA))
  expect_equal(y + months1, NA_POSIXct_)
  expect_equal(x + years1, as.Date(NA))
  expect_equal(y + years1, NA_POSIXct_)
})

test_that("addition with durations containing NA", {
  dt <- ymd(20161018)
  dt_1 <- ymd(20161019)

  dd_1na  <- ddays(c(1, NA))
  dd_na1  <- ddays(c(NA, 1))
  dd_nana <- ddays(c(NA, NA))

  ans_1na  <- add_duration_to_date(dd_1na, dt)
  ans_na1  <- add_duration_to_date(dd_na1, dt)
  ans_nana <- add_duration_to_date(dd_nana, dt)

  expect_is(ans_1na, "Date")
  expect_is(ans_na1, "Date")
  expect_is(ans_nana, "Date")

  expect_equal(ans_1na, c(dt_1, NA))
  expect_equal(ans_na1, rev(c(dt_1, NA)))
  expect_equal(ans_nana, as.Date(c(NA, NA)))
})

test_that("modulo operations return correct class", {
  int <- ymd("2010-01-01") %--% ymd("2011-01-01")

  expect_error(int %% int)
  expect_s4_class(int %% months(1), "Interval")
  expect_s4_class(int %% ddays(10), "Interval")

  expect_error(months(3) %% int)
  expect_s4_class(days(3) %% hours(2), "Period")
  expect_error(days(5) %% dminutes(300))

  expect_error(dyears(3) %% int)
  expect_error(ddays(2) %% weeks(1))
  expect_s4_class(ddays(3) %% ddays(1), "Duration")
})


test_that("modulo operations synchronize with integer division", {
  int <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- interval(int_start(int) + int %/% years(2) * years(2), int_end(int))
  int3 <- interval(int_start(int) + int %/% dweeks(20) * dweeks(20), int_end(int))

  expect_error(int %% int)
  expect_equal(int %% years(2), int2)
  expect_equal(int %% dweeks(20), int3)

  expect_error(years(2) %% int)
  expect_equal(years(2) %% months(5) + years(2) %/% months(5) * months(5), years(2))
  expect_error(years(2) %% dweeks(20))

  expect_error(dweeks(20) %% int)
  expect_error(dweeks(20) %% years(2))
  expect_equal(dweeks(20) %% ddays(20) + dweeks(20) %/% ddays(20) * ddays(20), dweeks(20))
})


test_that("modulo operations work for vectors", {
  int <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
  int3 <- interval(int_start(int) + int %/% years(2) * years(2), int_end(int))
  int4 <- interval(int_start(int) + int %/% dweeks(20) * dweeks(20), int_end(int))

  expect_error(c(int, int) %% int2)
  expect_equal(c(int, int) %% years(2), c(int3, int3))
  expect_equal(c(int, int) %% dweeks(20), c(int4, int4))

  expect_error(int %% c(int2, int2))
  expect_equal(int %% years(c(2, 2)), c(int3, int3))
  expect_equal(int %% dweeks(c(20, 20)), c(int4, int4))

  expect_error(years(2:3) %% int)
  expect_equal(years(2:3) %% months(5) + years(2:3) %/% months(5) * months(5), years(2:3))
  expect_error(years(2:3) %% dweeks(20))
  expect_error(years(2) %% c(int, int))
  expect_equal(years(2) %% months(5:6) + years(2) %/% months(5:6) * months(5:6), years(c(2, 2)))
  expect_error(years(2) %% dweeks(20:21))

  expect_error(dweeks(20:21) %% int)
  expect_error(dweeks(20:21) %% years(2))
  expect_equal(dweeks(20:21) %% ddays(20) + dweeks(20:21) %/% ddays(20) * ddays(20), dweeks(20:21))
  expect_error(dweeks(20) %% c(int, int))
  expect_error(dweeks(20) %% years(2:3))
  expect_equal(dweeks(20) %% ddays(20:21) + dweeks(20) %/% ddays(20:21) * ddays(20:21), c(dweeks(20), dweeks(20)))
})

context("Modulo operations")

test_that("modulo operations return correct class", {
        int <- ymd("2010-01-01") %--% ymd("2011-01-01")

        expect_error(int %% int)
        expect_is(int %% months(1), "Interval")
        expect_is(int %% ddays(10), "Interval")

        expect_error(months(3) %% int)
        expect_is(days(3) %% hours(2), "Period")
        expect_error(days(5) %% dminutes(300))

        expect_error(dyears(3) %% int)
        expect_error(ddays(2) %% weeks(1))
        expect_is(ddays(3) %% ddays(1), "Duration")

})


test_that("modulo operations synchronize with integer division", {
        int <- ymd("2010-01-01") %--% ymd("2011-01-01")
        int2 <- interval(int_start(int) + int %/% years(2) * years(2), int_end(int))
        int3 <- interval(int_start(int) + int %/% dweeks(20) * dweeks(20), int_end(int))

        expect_error(int %% int)
        expect_equal(int %% years(2), int2)
        expect_equal(int %% dweeks(20), int3)

        expect_error(years(2) %% int)
        expect_equal(years(2) %% months(5) + years(2) %/% months(5) * months(5), years(2))
        expect_error(years(2) %% dweeks(20))

        expect_error(dweeks(20) %% int)
        expect_error(dweeks(20) %% years(2))
        expect_equal(dweeks(20) %% ddays(20) + dweeks(20) %/% ddays(20) * ddays(20), dweeks(20))

})


test_that("modulo operations work for vectors", {
        int <- ymd("2010-01-01") %--% ymd("2011-01-01")
        int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
        int3 <- interval(int_start(int) + int %/% years(2) * years(2), int_end(int))
        int4 <- interval(int_start(int) + int %/% dweeks(20) * dweeks(20), int_end(int))

        expect_error(c(int, int) %% int2)
        expect_equal(c(int, int) %% years(2), c(int3, int3))
        expect_equal(c(int, int) %% dweeks(20), c(int4, int4))

        expect_error(int %% c(int2, int2))
        expect_equal(int %% years(c(2, 2)), c(int3, int3))
        expect_equal(int %% dweeks(c(20, 20)), c(int4, int4))

        expect_error(years(2:3) %% int)
        expect_equal(years(2:3) %% months(5) + years(2:3) %/% months(5) * months(5), years(2:3))
        expect_error(years(2:3) %% dweeks(20))
        expect_error(years(2) %% c(int, int))
        expect_equal(years(2) %% months(5:6) + years(2) %/% months(5:6) * months(5:6), years(c(2, 2)))
        expect_error(years(2) %% dweeks(20:21))

        expect_error(dweeks(20:21) %% int)
        expect_error(dweeks(20:21) %% years(2))
        expect_equal(dweeks(20:21) %% ddays(20) + dweeks(20:21) %/% ddays(20) * ddays(20), dweeks(20:21))
        expect_error(dweeks(20) %% c(int, int))
        expect_error(dweeks(20) %% years(2:3))
        expect_equal(dweeks(20) %% ddays(20:21) + dweeks(20) %/% ddays(20:21) * ddays(20:21), c(dweeks(20), dweeks(20)))


})

test_that("integer division concatenates and recycles correctly", {
  int1 <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
  int3 <- ymd("2009-01-01") %--% ymd("2010-01-01")

  expect_equal(int1 %/% ddays(1:2), c(365, 182))
  expect_equal(int1 %/% days(1:2), c(365, 182))
  expect_equal(c(int1, int2) %/% ddays(2), c(182, 365))
  expect_equal(c(int1, int2) %/% days(2), c(182, 365))
  expect_equal(dyears(1) %/% ddays(1:2), c(365, 182))
  expect_equal(dyears(1:2) %/% ddays(2), c(182, 365))
  expect_equal(months(c(13, 26)) %/% years(2), c(0, 1))
  expect_equal(months(13) %/% years(c(1, 2)), c(1, 0))
})

test_that("integer division works for interval numerator", {
  verify_output(test_path("test-ops-integer-division.txt"), {
    int <- ymd("2010-01-01") %--% ymd("2010-06-01")
    dur <- ddays(1)
    per <- months(1)

    "# Interval numerator"
    int %/% int
    int %/% dur
    int %/% per

    "# Duration numerator"
    dur %/% int
    dur %/% dur
    dur %/% per

    "# Period numeric"
    per %/% int
    per %/% dur
    per %/% per
  })
})

test_that("is.POSIXt works as expected", {
  expect_false(is.POSIXt(234))
  expect_true(is.POSIXt(as.POSIXct("2008-08-03 13:01:59", tz = "UTC")))
  expect_true(is.POSIXt(as.POSIXlt("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.POSIXt(Sys.Date()))
  expect_false(is.POSIXt(minutes(1)))
  expect_false(is.POSIXt(dminutes(1)))
  expect_false(is.POSIXt(interval(
    as.POSIXct("2008-08-03 13:01:59", tz = "UTC"),
    as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  )))
})

test_that("is.POSIXt handles vectors", {
  expect_true(is.POSIXt(c(
    as.POSIXct("2008-08-03 13:01:59", tz = "UTC"),
    as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  )))
})

test_that("c.POSIXct deals correctly with heterogeneous date-time classes", {
  d <- make_date(2000, 1, 1)
  dt <- make_datetime(2000, 1, 1, tz = "Europe/Berlin")
  expect_equal(c(dt, d), make_datetime(c(2000, 2000), 1, 1, tz = "Europe/Berlin"))
  expect_equal(c(dt, list(d)), make_datetime(c(2000, 2000), 1, 1, tz = "Europe/Berlin"))
  expect_equal(c(dt, list(d, list(d))), make_datetime(c(2000, 2000, 2000), 1, 1, tz = "Europe/Berlin"))
  dt <- make_datetime(2000, 1, 1, tz = "UTC")
  expect_equal(c(dt, d), make_datetime(c(2000, 2000), 1, 1, tz = "UTC"))


  ct <- as.POSIXct("1999-01-01 01:02:03", tz = "America/New_York")
  lt <- as.POSIXlt("2001-01-30 01:02:03", tz = "Europe/Berlin")
  expect_equal(c(ct, lt), ymd_hms(c("1999-01-01 01:02:03", "2001-01-29 19:02:03"), tz = "America/New_York"))
  expect_equal(c(lt, ct), as.POSIXlt(ymd_hms(c("2001-01-30 01:02:03", "1999-01-01 07:02:03"), tz = "Europe/Berlin")))
})


test_that("c.POSIXct deals correctly with empty vectors", {
  expect_equal(c(POSIXct()), POSIXct())
  expect_equal(c(POSIXct(), POSIXct()), POSIXct())
  expect_equal(c(POSIXct(), Date()), POSIXct())
  expect_equal(c(POSIXct(tz = "America/New_York")), POSIXct(tz = "America/New_York"))
  expect_equal(
    c(ymd("2021-01-01", tz = "America/New_York"), NULL, c()),
    ymd("2021-01-01", tz = "America/New_York")
  )
  expect_equal(
    c(
      ymd("2021-01-01", tz = "America/New_York"), POSIXct(),
      ymd("2021-01-02", tz = "America/New_York")
    ),
    ymd(c("2021-01-01", "2021-01-02"), tz = "America/New_York")
  )
  expect_equal(
    c(ymd("2021-01-01", tz = "UTC"), POSIXct(), ymd("2021-01-02"), NULL),
    ymd(c("2021-01-01", "2021-01-02"), tz = "UTC")
  )
})

# as_datetime -------------------------------------------------------------

test_that("converts numeric", {
  dt <- as_datetime(0)
  expect_s3_class(dt, "POSIXct")
  expect_equal(tz(dt), "UTC")
  expect_equal(unclass(dt)[[1]], 0)
})

test_that("converts date", {
  dt <- as_datetime(as.Date("1970-01-01"))
  expect_s3_class(dt, "POSIXct")
  expect_equal(tz(dt), "UTC")
  expect_equal(unclass(dt)[[1]], 0)
})

test_that("as_datetime.Date respects tz and sets HMS to 00:00:00", {
  d <- ymd(c("2000-01-01", "2020-10-10"))
  expect_equal(
    as_datetime(d, tz = "Europe/Berlin"),
    ymd(c("2000-01-01", "2020-10-10"), tz = "Europe/Berlin")
  )
  expect_equal(
    as_datetime(d),
    ymd(c("2000-01-01", "2020-10-10"), tz = "UTC")
  )
})

test_that("converts character", {
  chars <- c("2017-03-22T15:48:00.000Z", "2017-03-01 0:0:0", "2017-03-01 0:0:0.23", "2017-03-01 0:0:0.23")
  dt <- as_datetime(chars)
  expect_s3_class(dt, "POSIXct")
  expect_equal(tz(dt), "UTC")
  expect_equal(dt, ymd_hms(chars))
  expect_equal(suppressMessages(as_datetime(chars, tz = "Europe/Amsterdam")), ymd_hms(chars, tz = "Europe/Amsterdam", quiet = TRUE))
})

test_that("changes timezone of POSIXct", {
  dt <- as_datetime(make_datetime(tz = "America/Chicago"), tz = "UTC")
  expect_equal(tz(dt), "UTC")
})

test_that("addition of large seconds doesn't overflow", {
  from_period <- origin + seconds(2^31 + c(-2:2))
  from_char <- ymd_hms(c(
    "2038-01-19 03:14:06", "2038-01-19 03:14:07", "2038-01-19 03:14:08",
    "2038-01-19 03:14:09", "2038-01-19 03:14:10"
  ))
  expect_equal(from_period, from_char)
})

test_that("as_datetime works correctly", {
  x <- c(
    "17-01-20", "2017-01-20 01:02:03", "2017-03-22T15:48:00.000Z",
    "2017-01-20", "2017-01-20 01:02:03", "2017-03-22T15:48:00.000Z"
  )

  y <- c(
    "2017-01-20 00:00:00 UTC", "2017-01-20 01:02:03 UTC",
    "2017-03-22 15:48:00 UTC", "2017-01-20 00:00:00 UTC",
    "2017-01-20 01:02:03 UTC", "2017-03-22 15:48:00 UTC"
  )

  zns <- c(3, 6)
  putc <- ymd_hms(y)
  pus <- ymd_hms(y[-zns], tz = "America/Chicago")
  expect_equal(as_datetime(x), putc)
  expect_equal(as_datetime(x[-zns], tz = "America/Chicago"), pus)

  for (i in seq_along(x)) {
    expect_equal(as_datetime(x[[i]]), putc[[i]])
  }

  for (i in seq_along(pus)) {
    expect_equal(as_datetime(x[-zns][[i]], tz = "America/Chicago"), pus[[i]])
  }
})

test_that("as_datetime() always returns POSIXct", {
  expect_s3_class(as_datetime("2010-01-01"), "POSIXct")
  expect_s3_class(as_datetime("2010-01-01", format = "%Y-%m-%d"), "POSIXct")
})

context("Integer division operations")

test_that("integer division concatenates and recycles correctly", {

  int1 <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
  int3 <- ymd("2009-01-01") %--% ymd("2010-01-01")

  expect_equal(int1 %/% ddays(1:2), c(365, 182))
  expect_equal(int1 %/% days(1:2), c(365, 182))
  expect_equal(c(int1, int2) %/% ddays(2), c(182, 365))
  expect_equal(c(int1, int2) %/% days(2), c(182, 365))
  expect_equal(dyears(1) %/% ddays(1:2), c(365, 182))
  expect_equal(dyears(1:2) %/% ddays(2), c(182, 365))
  expect_equal(months(c(13, 26)) %/% years(2), c(0, 1))
  expect_equal(months(13) %/% years(c(1, 2)), c(1, 0))

})

test_that("integer division works for interval numerator", {

  verify_output(test_path("test-ops-integer-division.txt"), {
    int <- ymd("2010-01-01") %--% ymd("2010-06-01")
    dur <- ddays(1)
    per <- months(1)

    "# Interval numerator"
    int %/% int
    int %/% dur
    int %/% per

    "# Duration numerator"
    dur %/% int
    dur %/% dur
    dur %/% per

    "# Period numeric"
    per %/% int
    per %/% dur
    per %/% per
  })
})

test_that("force_tz returns NA for a time that falls in the spring gap", {
  x <- structure(1268532305, class = c("POSIXct", "POSIXt"), tzone = "UTC")
  y <- structure(1268618705, class = c("POSIXct", "POSIXt"), tzone = "UTC")
  z <- structure(1268633105, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  expect_true(is.na(force_tz(x, "America/New_York")))
  expect_equal(force_tz(c(y, x), "America/New_York"), c(z, NA))
})

test_that("force_tz behaves consistently for the fall overlap", {
  y <- structure(1289111405, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))
  y2 <- structure(1289115005, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  y3 <- structure(1289104205, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")

  z <- structure(1289197805, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  z2 <- structure(1289201405, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  z3 <- structure(1289194205, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")

  cdst <- structure(1352012400, tzone = "America/Chicago", class = c("POSIXct", "POSIXt"))
  edst <- structure(1352008800, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))

  expect_equal(y + hours(1), y2)
  expect_equal(c(y, z) + hours(1), c(y2, z2))
  expect_equal(y - hours(1), y3)
  expect_equal(c(y, z) - hours(1), c(y3, z3))
  expect_equal(force_tz(cdst, "America/New_York"), edst)
})

test_that("addition handles daylight savings time for spring gap", {
  x <- structure(1268542800, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px <- structure(1268625600, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px2 <- structure(1268712000, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex <- structure(1268629200, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex2 <- structure(1268715600, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))

  expect_equal(x + days(1), px)
  expect_equal(c(x, px) + days(1), c(px, px2))
  expect_equal(x + ddays(1), ex)
  expect_equal(c(x, ex) + ddays(1), c(ex, ex2))
})


test_that("subtraction handles daylight savings time for spring gap", {
  x <- structure(1268625600, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px <- structure(1268542800, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px2 <- structure(1268456400, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex <- structure(1268539200, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex2 <- structure(1268452800, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))

  expect_equal(x - days(1), px)
  expect_equal(c(x, px) - days(1), c(px, px2))
  expect_equal(x - ddays(1), ex)
  expect_equal(c(x, ex) - ddays(1), c(ex, ex2))
})

test_that("addition handles daylight savings time for fall overlap", {
  x <- structure(1289102400, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px <- structure(1289192400, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px2 <- structure(1289278800, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex <- structure(1289188800, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex2 <- structure(1289275200, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")

  expect_equal(x + days(1), px)
  expect_equal(c(x, px) + days(1), c(px, px2))
  expect_equal(x + ddays(1), ex)
  expect_equal(c(x, ex) + ddays(1), c(ex, ex2))
})


test_that("subtraction handles daylight savings time for fall overlap", {
  x <- structure(1289192400, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px <- structure(1289102400, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px2 <- structure(1289016000, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex <- structure(1289106000, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))
  ex2 <- structure(1289019600, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))

  expect_equal(x - days(1), px)
  expect_equal(c(x, px) - days(1), c(px, px2))
  expect_equal(x - ddays(1), ex)
  expect_equal(c(x, ex) - ddays(1), c(ex, ex2))
})


test_that("update returns NA for date-times in the spring dst gap", {
  poslt <- structure(list(
    sec = 59, min = 59L, hour = 1L, mday = 14L, mon = 2L,
    year = 110L, wday = 0L, yday = 72L, isdst = 0L
  ),
  .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
  class = c("POSIXlt", "POSIXt"), tzone = c("America/New_York", "EST", "EDT")
  )
  x <- structure(list(
    sec = 59, min = 59L, hour = 1L, mday = 15L, mon = 2L,
    year = 110L, wday = 1L, yday = 73L, isdst = 1L
  ),
  .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
  class = c("POSIXlt", "POSIXt"), tzone = c("America/New_York", "EST", "EDT")
  )
  x2 <- structure(list(
    sec = 59, min = 59L, hour = 2L, mday = 15L, mon = 2L,
    year = 110L, wday = 1L, yday = 73L, isdst = 1L
  ),
  .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
  class = c("POSIXlt", "POSIXt"), tzone = c("America/New_York", "EST", "EDT")
  )
  xu <- structure(list(
    sec = 59, min = 59L, hour = 2L, mday = 15L, mon = 2L,
    year = 110L, wday = 1L, yday = 73L, isdst = 1L
  ),
  .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
  class = c("POSIXlt", "POSIXt"), tzone = "UTC"
  )

  expect_true(is.na(update(poslt, seconds = 61)))
  expect_true(is.na(update(poslt, minutes = 61)))
  expect_true(is.na(update(poslt, hours = 2)))
  expect_equal(update(c(x, poslt), hours = 2), c(x2, NA))

  poslt <- structure(list(
    sec = 59, min = 59L, hour = 2L, mday = 13L, mon = 2L,
    year = 110L, wday = 6L, yday = 71L, isdst = 0L
  ),
  .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
  class = c("POSIXlt", "POSIXt"), tzone = c("America/New_York", "EST", "EDT")
  )
  expect_true(is.na(update(poslt, mdays = 14)))
  expect_true(is.na(update(poslt, wdays = 8)))
  expect_true(is.na(update(poslt, ydays = 73)))

  poslt <- structure(list(
    sec = 59, min = 59L, hour = 2L, mday = 14L, mon = 1L,
    year = 110L, wday = 0L, yday = 44L, isdst = 0L
  ),
  .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
  class = c("POSIXlt", "POSIXt"), tzone = c("America/New_York", "EST", "EDT")
  )
  expect_true(is.na(update(poslt, months = 3)))

  ## poslt <- structure(list(sec = 59, min = 59L, hour = 2L, mday = 14L, mon = 2L,
  ##   year = 109L, wday = 6L, yday = 72L, isdst = 0L),
  ##   .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
  ##   class = c("POSIXlt", "POSIXt"), tzone = c("America/New_York", "EST", "EDT"))
  ## expect_true(is.na(update(poslt, years = 2010)))

  poslt <- structure(list(
    sec = 59, min = 59L, hour = 2L, mday = 14L, mon = 2L,
    year = 110L, wday = 0L, yday = 72L, isdst = 0L
  ),
  .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
  class = c("POSIXlt", "POSIXt"), tzone = "UTC"
  )
  expect_true(is.na(update(poslt, tzs = "America/New_York")))
  expect_equal(update(c(xu, poslt), tzs = "America/New_York"), c(x2, NA))
})

test_that("update rollovers perform correctly across the fall overlap", {
  timest <- structure(1289113199, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))
  posct <- structure(1289109659, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))
  timest2 <- structure(1289116919, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  timest3 <- structure(1289124119, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))
  timest0 <- structure(1289116859, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))
  expect_equal(update(timest, minutes = 121), timest2)
  expect_equal(update(c(timest, timest2), minutes = 121), c(timest2, timest3))
  expect_equal(update(timest, minutes = 0), posct)
  expect_equal(update(c(timest, timest2), minutes = 0), c(posct, timest0))
})

test_that("arithmetic handles fall gap in timespan appropriate way", {
  cdst <- structure(1352012400, tzone = "America/Chicago", class = c("POSIXct", "POSIXt"))

  expect_equal(cdst - seconds(1), cdst - 3601)
  expect_equal(cdst - dseconds(1), cdst - 1)
})

test_that("division operations work for interval numerator", {
  int <- ymd("2009-02-03") %--% ymd("2010-01-01")
  smaller_int <- ymd("2009-12-01") %--% ymd("2010-01-01")
  bigger_int <- ymd("2009-01-01") %--% ymd("2010-02-01")
  smaller_dur <- ddays(20) + dhours(4)
  bigger_dur <- dyears(1) + dseconds(2)
  smaller_diff <- make_difftime(days = 100)
  bigger_diff <- make_difftime(days = 400)
  expect_error(int / smaller_int)
  expect_error(int / bigger_int)
  expect_equal(int_start(int) + days(int / days(1)), int_end(int))
  expect_equal(int_start(int) + ddays(int / ddays(1)), int_end(int))
  expect_equal(int / smaller_diff, 28684800 / 8640000)
  expect_equal(int / bigger_diff, 28684800 / 34560000)

  # #926
  int1 <- interval(ymd("2017-01-29") + years(1:3), ymd("2019-01-30"))
  int2 <- interval(ymd("2018-01-29") + years(1:3), ymd("2020-01-30"))
  expect_equal(int1 / months(1), c(12.033333333, 0.0333333333, -11.9666666))
  expect_equal(int2 / months(1), c(12.032258064, 0.0322580645, -11.9677419))
})

test_that("division works for interval numerator with vectors", {
  int1 <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
  int3 <- ymd("2009-01-01") %--% ymd("2010-01-01")
  smaller_diff <- make_difftime(days = 365)
  expect_error(c(int1, int2) / int1)
  expect_error(int1 / c(int1, int2))
  expect_equal(c(int1, int2) / years(1), c(1, 2))
  expect_equal(c(int1, int2) / dyears(1), c(0.9993155, 1.9986311), tolerance = 6)
  expect_equal(int1 / years(1:2), c(1, 0.5))
  expect_equal(int1 / dyears(1:2), c(0.9993155, 0.4996578), tolerance = 6)
  expect_equal(c(int1, int2) / smaller_diff, c(1, 2))
  expect_equal(int2 / c(1, 2), c(int2, int3))
  expect_equal(c(int2, int2) / 2, c(int3, int3))
})

test_that("arguments are recycled in interval division", {
  expect_equal(
    interval(c("2015-03-01", "2000-03-02", "2015-03-02", "2015-03-03"), "2016-10-01") / months(1),
    c(19.00000000, 198.96666667, 18.96666667, 18.93333333)
  )
  expect_equal(
    interval(c("2015-03-01", "2015-03-02", "2015-03-03", "2015-03-04"), "2016-10-01") / months(1:2),
    interval(c("2015-03-01", "2015-03-02", "2015-03-03", "2015-03-04"), "2016-10-01") / months(c(1:2, 1:2))
  )
  expect_equal(
    interval("2015-03-02", "2016-10-01") / months(1:3),
    interval(c("2015-03-02", "2015-03-02", "2015-03-02"), "2016-10-01") / months(1:3)
  )
})

test_that("NA and Interval(0) propagate in interval division", {
  t <- ymd_hms("2019-03-01 12:30:50")
  int0 <- interval(character(0), t)
  expect_equal(int0 / days(3), numeric(0))
  intNA <- interval(NA, t)
  expect_equal(intNA / days(3), NA_real_)
})

test_that("division works for interval numerator with period denominator", {
  int1 <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
  int3 <- ymd("2009-01-01") %--% ymd("2010-01-01")
  expect_equal(c(int1, int2) / days(1), time_length(c(int1, int2), "days"))
  expect_equal(c(int1, int2) / minutes(1), time_length(c(int1, int2), "mins"))
  expect_equal(c(int1, int2) / months(1), c(12, 24))
  expect_equal(c(int1, int2) / months(6), c(2, 4))
  expect_equal(c(int1, int2) / years(1), c(1, 2))
  expect_equal(c(int1, int2) / years(2), c(.5, 1))
})

test_that("dividision works for interval vector with NAs", {
  int1 <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
  int3 <- ymd(NA, quiet = TRUE) %--% ymd(NA, quiet = TRUE)
  ints <- c(int1, int2, int3)
  nas <- as.numeric(c(NA, NA, NA))
  expect_equal(ints / years(1), c(1, 2, NA))
  expect_equal(ints / dyears(1), c(0.9993155, 1.9986311, NA), tolerance = 6)
  expect_equal(ints / years(NA), nas)
  expect_equal(ints / dyears(NA), nas)
})

test_that("division operations work for period numerator", {
  int1 <- ymd("2009-01-01") %--% ymd("2010-01-01")
  expect_error(years(2) / int1)
  expect_equal(days(2) / hours(12), 4)
  expect_equal(years(1) / months(1), 12)
  expect_error(days(2) / dhours(12))
  expect_error(days(1) / make_difftime(days = 1))
  expect_equal(days(2) / 2, days(1))
})

test_that("division works for period numerator with vectors", {
  int1 <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
  expect_error(years(1:2) / int2)
  expect_error(years(1) / c(int1, int2))
  expect_equal(days(1:2) / hours(12), c(2, 4))
  expect_equal(days(1) / hours(c(12, 24)), c(2, 1))
  expect_error(days(1:2) / dhours(12))
  expect_error(days(1) / dhours(c(12, 24)))
  expect_error(days(1:2) / make_difftime(days = 1))
  expect_equal(days(2) / c(1, 2), c(days(2), days(1)))
  expect_equal(days(c(2, 4)) / c(2), c(days(1), days(2)))
})

test_that("division operations work for duration numerator", {
  int <- ymd("2009-12-01") %--% ymd("2010-01-01")
  expect_error(ddays(31) / int)
  expect_error(ddays(20) / days(20))
  expect_equal(ddays(20) / ddays(1), 20)
  expect_equal(ddays(20) / make_difftime(days = 1), 20)
})

test_that("division works for duration numerator with vectors", {
  int1 <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
  expect_error(ddays(365 * c(1:2)) / int2)
  expect_error(ddays(365) / c(int1, int2))
  expect_equal(ddays(1:2) / dhours(12), c(2, 4))
  expect_equal(ddays(1) / dhours(c(12, 24)), c(2, 1))
  expect_error(ddays(1:2) / hours(12))
  expect_error(ddays(1) / hours(c(12, 24)))
  expect_equal(ddays(1:2) / make_difftime(days = 1), c(1, 2))
  expect_equal(ddays(2) / c(1, 2), c(ddays(2), ddays(1)))
  expect_equal(ddays(c(2, 4)) / c(2), c(ddays(1), ddays(2)))
})

test_that("division operations work for difftime numerator", {
  diff <- make_difftime(days = 365)
  int <- ymd("2010-01-01") %--% ymd("2011-01-01")
  expect_error(diff / int)
  expect_error(diff / days(1))
  expect_equal(diff / ddays(365), 1)
})

test_that("division works as expected for periods", {
  expect_error(3 / months(1))
  expect_error(months(1) / 3)
  expect_equal(months(9) / 3, months(3))
  expect_s4_class(months(9) / 3, "Period")
})

test_that("dividing vectors works for periods", {
  expect_error(c(2, 3) / months(1))
  expect_error(months(1) / c(2, 3))
  expect_equal(months(9) / c(3, 1), months(c(3, 9)))
  expect_s4_class(months(9) / c(3, 1), "Period")
})

test_that("division works as expected for durations", {
  expect_error(3 / dhours(1))
  expect_equal(dhours(9) / 3, dhours(3))
  expect_s4_class(dhours(9) / 3, "Duration")
})

test_that("dividing vectors works for durations", {
  expect_error(c(2, 3) / dhours(1))
  expect_equal(dhours(9) / c(3, 1), dhours(c(3, 9)))
  expect_s4_class(dhours(9) / c(3, 1), "Duration")
})

context("POSIXt")

test_that("is.POSIXt works as expected", {
  expect_false(is.POSIXt(234))
  expect_true(is.POSIXt(as.POSIXct("2008-08-03 13:01:59", tz = "UTC")))
  expect_true(is.POSIXt(as.POSIXlt("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.POSIXt(Sys.Date()))
  expect_false(is.POSIXt(minutes(1)))
  expect_false(is.POSIXt(dminutes(1)))
  expect_false(is.POSIXt(interval(
    as.POSIXct("2008-08-03 13:01:59", tz = "UTC"),
    as.POSIXct("2009-08-03 13:01:59", tz = "UTC"))))
})

test_that("is.POSIXt handles vectors", {
  expect_true(is.POSIXt(c(
    as.POSIXct("2008-08-03 13:01:59", tz = "UTC"),
    as.POSIXct("2009-08-03 13:01:59", tz = "UTC"))))
})

# as_datetime -------------------------------------------------------------

test_that("converts numeric", {
  dt <- as_datetime(0)
  expect_s3_class(dt, "POSIXct")
  expect_equal(tz(dt), "UTC")
  expect_equal(unclass(dt)[[1]], 0)
})

test_that("converts date", {
  dt <- as_datetime(as.Date("1970-01-01"))
  expect_s3_class(dt, "POSIXct")
  expect_equal(tz(dt), "UTC")
  expect_equal(unclass(dt)[[1]], 0)
})

test_that("converts character", {
  chars <- c("2017-03-22T15:48:00.000Z", "2017-03-01 0:0:0", "2017-03-01 0:0:0.23", "2017-03-01 0:0:0.23")
  dt <- as_datetime(chars)
  expect_s3_class(dt, "POSIXct")
  expect_equal(tz(dt), "UTC")
  expect_equal(dt, ymd_hms(chars))
  expect_equal(as_datetime(chars, tz = "Europe/Amsterdam"), ymd_hms(chars, tz = "Europe/Amsterdam", quiet = TRUE))
})

test_that("changes timezone of POSIXct", {
  dt <- as_datetime(make_datetime(tz = "America/Chicago"))
  expect_equal(tz(dt), "UTC")
})

test_that("addition of large seconds doesn't overflow", {
  from_period <- origin + seconds(2 ^ 31 + c(-2:2))
  from_char <- ymd_hms(c("2038-01-19 03:14:06", "2038-01-19 03:14:07", "2038-01-19 03:14:08",
                         "2038-01-19 03:14:09", "2038-01-19 03:14:10"))
  expect_equal(from_period, from_char)
})

test_that("as_datetime works correctly", {

  x <- c("17-01-20", "2017-01-20 01:02:03", "2017-03-22T15:48:00.000Z",
         "2017-01-20", "2017-01-20 01:02:03", "2017-03-22T15:48:00.000Z")

  y <- c("2017-01-20 00:00:00 UTC", "2017-01-20 01:02:03 UTC",
         "2017-03-22 15:48:00 UTC", "2017-01-20 00:00:00 UTC",
         "2017-01-20 01:02:03 UTC", "2017-03-22 15:48:00 UTC")

  zns <- c(3, 6)
  putc <- ymd_hms(y)
  pus <- ymd_hms(y[-zns], tz = "America/Chicago")
  expect_equal(as_datetime(x), putc)
  expect_equal(as_datetime(x[-zns], tz = "America/Chicago"), pus)

  for (i in seq_along(x)) {
    expect_equal(as_datetime(x[[i]]), putc[[i]])
  }

  for (i in seq_along(pus)) {
    expect_equal(as_datetime(x[-zns][[i]], tz = "America/Chicago"), pus[[i]])
  }
})

test_that("as_datetime() always returns POSIXct", {
  expect_s3_class(as_datetime("2010-01-01"), "POSIXct")
  expect_s3_class(as_datetime("2010-01-01", format = "%Y-%m-%d"), "POSIXct")
})

###### subtraction for everything

test_that("subtraction works as expected for instants", {
  x <- as.POSIXct("2008-01-01 00:00:02", tz = "UTC")
  y <- as.POSIXlt("2008-01-01 00:00:02", tz = "UTC")
  z <- as.Date("2008-01-03")

  expect_equal(x - years(1), as.POSIXct("2007-01-01 00:00:02", tz = "UTC"))
  expect_equal(y - years(1), as.POSIXlt("2007-01-01 00:00:02", tz = "UTC"))
  expect_equal(z - years(1), as.Date("2007-01-03"))

  expect_equal(x - dyears(1), as.POSIXct("2006-12-31 18:00:02", tz = "UTC"))
  expect_equal(y - dyears(1), as.POSIXlt("2006-12-31 18:00:02", tz = "UTC"))
  expect_equal(z - dyears(1), as.POSIXct("2007-01-02 18:00:00", tz = "UTC"))

  time1 <- as.POSIXct("2007-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2008-01-01 00:00:59", tz = "UTC")
  int <- interval(time1, time2)
  num <- as.numeric(time2) - as.numeric(time1)

  expect_error(x - int)
  expect_error(y - int)
  expect_error(z - int)
})

test_that("subtraction with instants returns correct class", {
  x <- as.POSIXct("2008-01-01 12:00:00", tz = "UTC")
  y <- as.POSIXlt("2008-01-01 12:00:00", tz = "UTC")
  z <- as.Date("2008-01-01")

  expect_s3_class(x - years(1), "POSIXct")
  expect_s3_class(y - years(1), "POSIXlt")
  expect_s3_class(z - years(1), "Date")
  expect_s3_class(x - dyears(1), "POSIXct")
  expect_s3_class(y - dyears(1), "POSIXlt")
})


test_that("subtraction works as expected for periods", {
  time1 <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-02-02 00:00:00", tz = "UTC")
  int <- interval(time1, time2)

  expect_equal(years(1) - 1, period(seconds = -1, years = 1))
  expect_error(years(1) - as.POSIXct("2008-01-01 00:00:00", tz = "UTC"))
  expect_error(years(1) - as.POSIXct("2008-01-01 00:00:00", tz = "UTC"))
  expect_equal(years(1) - minutes(3), period(minutes = -3, years = 1))
  expect_error(years(1) - dyears(1))
  expect_error(years(1) - int)
})

test_that("subtraction with periods returns correct class", {
  expect_s4_class(years(1) - 1, "Period")
  expect_s4_class(years(1) - minutes(3), "Period")
})


test_that("subtraction works as expected for durations", {
  time1 <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  int <- interval(time1, time2)

  expect_equal(dyears(1) - 1, duration(31557599))
  expect_error(dyears(1) - as.POSIXct("2008-01-01 00:00:00", tz = "UTC"))
  expect_error(dyears(1) - minutes(3))
  expect_equal(dyears(2) - dyears(1), dyears(1))
  expect_error(dyears(1) - int)
})

test_that("subtraction with durations returns correct class", {
  expect_s4_class(dyears(1) - 1, "Duration")
  expect_s4_class(dyears(1) - dyears(1), "Duration")
})


test_that("subtraction works as expected for intervals", {
  time1 <- as.POSIXct("2008-08-03 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  time3 <- as.POSIXct("2008-11-02 00:00:00", tz = "UTC")
  int <- interval(time1, time2)
  int2 <- interval(time3, time2)

  expect_error(time2 - int)
  expect_error(int - 1)
  expect_error(int - time1)
  expect_error(int - minutes(3))
  expect_error(int - dminutes(3))
  expect_error(int - int2)
})


test_that("%m-% correctly subtracts months without rollover", {
  may <- ymd_hms("2010-05-31 03:04:05")
  ends <- ymd_hms(c("2010-04-30 03:04:05", "2010-03-31 03:04:05", "2010-02-28 03:04:05"))

  expect_equal(may %m-% months(1:3), ends)
})

test_that("%m-% correctly subtracts years without rollover", {
  leap <- ymd("2012-02-29")
  next1 <- ymd("2011-02-28")
  next2 <- ymd("2011-01-29")
  next3 <- ymd("2011-03-29")

  expect_equal(leap %m-% years(1), next1)
  expect_equal(leap %m-% period(years = 1, months = 1), next2)
  expect_equal(leap %m-% period(years = 1, months = -1), next3)
})

test_that("%m-% correctly subtract negative months without rollover", {
  jan <- ymd_hms("2010-01-31 03:04:05")
  ends <- ymd_hms(c("2010-02-28 03:04:05", "2010-03-31 03:04:05", "2010-04-30 03:04:05"))

  expect_equal(jan %m-% -months(1:3), ends)
})

test_that("%m-% correctly subtracts negative years without rollover", {
  leap <- ymd("2012-02-29")
  next1 <- ymd("2013-02-28")
  next2 <- ymd("2013-03-29")
  next3 <- ymd("2013-01-29")

  expect_equal(leap %m-% years(-1), next1)
  expect_equal(leap %m-% period(years = -1, months = -1), next2)
  expect_equal(leap %m-% period(years = -1, months = 1), next3)
})

library(vctrs)

# ------------------------------------------------------------------------------
# Common ptype2 / cast

test_that("no common type when mixing Period/Duration/Interval", {
  verify_errors({
    expect_error(vec_ptype2(period(), duration()), class = "vctrs_error_incompatible_type")
    expect_error(vec_ptype2(duration(), period()), class = "vctrs_error_incompatible_type")

    expect_error(vec_ptype2(period(), interval()), class = "vctrs_error_incompatible_type")
    expect_error(vec_ptype2(interval(), period()), class = "vctrs_error_incompatible_type")

    expect_error(vec_ptype2(duration(), interval()), class = "vctrs_error_incompatible_type")
    expect_error(vec_ptype2(interval(), duration()), class = "vctrs_error_incompatible_type")
  })
})

test_that("can't cast between Period/Duration/Interval", {
  verify_errors({
    expect_error(vec_cast(period(), duration()), class = "vctrs_error_incompatible_type")
    expect_error(vec_cast(duration(), period()), class = "vctrs_error_incompatible_type")

    expect_error(vec_cast(period(), interval()), class = "vctrs_error_incompatible_type")
    expect_error(vec_cast(interval(), period()), class = "vctrs_error_incompatible_type")

    expect_error(vec_cast(duration(), interval()), class = "vctrs_error_incompatible_type")
    expect_error(vec_cast(interval(), duration()), class = "vctrs_error_incompatible_type")
  })
})

# ------------------------------------------------------------------------------
# Period - proxy / restore

test_that("proxy is a data frame", {
  x <- period(year = 1:2, day = 3:4)

  expect <- list(
    year = x@year, month = x@month, day = x@day,
    hour = x@hour, minute = x@minute, second = x@.Data
  )

  expect <- new_data_frame(expect)

  expect_identical(vec_proxy(x), expect)
})

test_that("proxy stores names on the `$second` column (allowing duplicates)", {
  skip_if_cant_set_s4_names()

  x <- stats::setNames(days(1:3), c("x", "y", "x"))

  proxy <- vec_proxy(x)

  expect_named(proxy$second, names(x))
})

test_that("comparison / equality proxies have the names too", {
  skip_if_cant_set_s4_names()

  x <- stats::setNames(days(1:3), c("x", "y", "x"))

  expect_named(vec_proxy_compare(x)$second, names(x))
  expect_named(vec_proxy_equal(x)$second, names(x))
})

test_that("restore method works", {
  x <- period(year = 1:2, day = 3:4)
  expect_identical(vec_restore(vec_proxy(x), x), x)
})

test_that("restore method retains names", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(days(1), "x")
  expect_named(vec_restore(vec_proxy(x), x), "x")
})

# ------------------------------------------------------------------------------
# Period - ptype2

test_that("Period default ptype2 method falls through to `vec_default_ptype2()`", {
  verify_errors({
    expect_error(vec_ptype2(period(), 1), class = "vctrs_error_incompatible_type")
    expect_error(vec_ptype2(1, period()), class = "vctrs_error_incompatible_type")
  })
})

test_that("common type of Period and Period exists", {
  expect_identical(vec_ptype2(period(), period()), period())
})

test_that("common type of Period and NULL exists", {
  expect_identical(vec_ptype2(period(), NULL), period())
  expect_identical(vec_ptype2(NULL, period()), period())
})

test_that("common type of Period and unspecified exists", {
  expect_identical(vec_ptype2(period(), NA), period())
  expect_identical(vec_ptype2(NA, period()), period())

  expect_identical(vec_ptype2(period(), vctrs::unspecified()), period())
  expect_identical(vec_ptype2(vctrs::unspecified(), period()), period())
})

# ------------------------------------------------------------------------------
# Period - cast

test_that("Period default cast method falls through to `vec_default_cast()`", {
  verify_errors({
    expect_error(vec_cast(period(), 1), class = "vctrs_error_incompatible_type")
    expect_error(vec_cast(1, period()), class = "vctrs_error_incompatible_type")
  })
})

test_that("Period can be cast to Period", {
  expect_identical(vec_cast(days(1), months(1)), days(1))
})

test_that("can cast around `NULL`", {
  expect_identical(vec_cast(NULL, period()), NULL)
  expect_identical(vec_cast(period(), NULL), period())
})

test_that("can cast unspecified to Period", {
  expect_identical(vec_cast(NA, period()), period()[NA_real_])
  expect_error(vec_cast(period(), NA), class = "vctrs_error_incompatible_type")
})

# ------------------------------------------------------------------------------
# Period - vctrs functionality

test_that("can slice Period objects", {
  expect_identical(vec_slice(days(3:4), 2:1), days(4:3))
})

test_that("slicing preserves names", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(days(1:2), c("x", "y"))
  expect_named(vec_slice(x, c(1, 1, 2)), c("x", "x", "y"))
})

test_that("can combine Period objects", {
  expect_identical(vec_c(days(1), days(2)), days(1:2))
})

test_that("can combine Period objects when some are named (#1072)", {
  skip_if_cant_set_s4_names()
  x <- days(1)
  y <- stats::setNames(days(2), "y")
  expect_named(vec_c(x, y), c("", "y"))
})

test_that("can row bind Period objects", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(days(1), "x")
  expect_identical(vec_rbind(x, x), data.frame(x = c(x, x)))
})

test_that("can row bind data frames with Period objects", {
  expect_identical(
    vec_rbind(data.frame(x = days(1)), data.frame(x = days(1))),
    data.frame(x = days(c(1, 1)))
  )
})

test_that("can column bind Period objects", {
  expect_identical(
    vec_cbind(x = days(1), y = days(1:2)),
    data.frame(x = days(c(1, 1)), y = days(1:2))
  )
})

test_that("can column bind data frames with Period objects", {
  expect_identical(
    vec_cbind(data.frame(x = days(1)), data.frame(y = days(1:2))),
    data.frame(x = days(c(1, 1)), y = days(1:2))
  )
})

test_that("Period objects can be ordered", {
  expect_identical(vec_order(vec_c(years(1), days(1))), c(2L, 1L))
  expect_identical(vec_order(vec_c(days(2), days(1))), c(2L, 1L))
})

# ------------------------------------------------------------------------------
# Duration - proxy / restore

test_that("proxy is the underlying number of seconds", {
  x <- ddays(1:2)
  expect_identical(vec_proxy(x), x@.Data)
})

test_that("proxy stores the names", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(ddays(1:3), c("x", "y", "x"))
  expect_named(vec_proxy(x), c("x", "y", "x"))
})

test_that("comparison / equality proxies store names too", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(ddays(1:3), c("x", "y", "x"))
  expect_named(vec_proxy_compare(x), names(x))
  expect_named(vec_proxy_equal(x), names(x))
})

test_that("restore method works", {
  x <- ddays(1:2)
  expect_identical(vec_restore(vec_proxy(x), x), x)
})

test_that("restore method retains names", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(ddays(1), "x")
  expect_named(vec_restore(vec_proxy(x), x), "x")
})

# ------------------------------------------------------------------------------
# Duration - ptype2

test_that("Duration default ptype2 method falls through to `vec_default_ptype2()`", {
  verify_errors({
    expect_error(vec_ptype2(duration(), 1), class = "vctrs_error_incompatible_type")
    expect_error(vec_ptype2(1, duration()), class = "vctrs_error_incompatible_type")
  })
})

test_that("common type of Duration and Duration exists", {
  expect_identical(vec_ptype2(duration(), duration()), duration())
})

test_that("common type of Duration and NULL exists", {
  expect_identical(vec_ptype2(duration(), NULL), duration())
  expect_identical(vec_ptype2(NULL, duration()), duration())
})

test_that("common type of Duration and unspecified exists", {
  expect_identical(vec_ptype2(duration(), NA), duration())
  expect_identical(vec_ptype2(NA, duration()), duration())

  expect_identical(vec_ptype2(duration(), unspecified()), duration())
  expect_identical(vec_ptype2(unspecified(), duration()), duration())
})

test_that("common type of Duration and difftime is Duration", {
  expect_identical(vec_ptype2(duration(), new_duration()), duration())
  expect_identical(vec_ptype2(new_duration(), duration()), duration())
})

# ------------------------------------------------------------------------------
# Duration - cast

test_that("Duration default cast method falls through to `vec_default_cast()`", {
  verify_errors({
    expect_error(vec_cast(duration(), 1), class = "vctrs_error_incompatible_type")
    expect_error(vec_cast(1, duration()), class = "vctrs_error_incompatible_type")
  })
})

test_that("Duration can be cast to Duration", {
  expect_identical(vec_cast(ddays(1), dmonths(1)), ddays(1))
})

test_that("can cast around `NULL`", {
  expect_identical(vec_cast(NULL, duration()), NULL)
  expect_identical(vec_cast(duration(), NULL), duration())
})

test_that("can cast unspecified to Duration", {
  expect_identical(vec_cast(NA, duration()), duration()[NA_real_])
  expect_error(vec_cast(duration(), NA), class = "vctrs_error_incompatible_type")
})

test_that("Duration can be cast to and from difftime", {
  expect_identical(vec_cast(duration(), new_duration()), new_duration())
  expect_identical(vec_cast(new_duration(), duration()), duration())
})

# ------------------------------------------------------------------------------
# Duration - vctrs functionality

test_that("can slice Duration objects", {
  expect_identical(vec_slice(ddays(3:4), 2:1), ddays(4:3))
})

test_that("slicing preserves names", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(ddays(1:2), c("x", "y"))
  expect_named(vec_slice(x, c(1, 1, 2)), c("x", "x", "y"))
})

test_that("can combine Duration objects", {
  expect_identical(vec_c(ddays(1), ddays(2)), ddays(1:2))
})

test_that("can combine Duration objects when some are named (#1072)", {
  skip_if_cant_set_s4_names()
  x <- ddays(1)
  y <- stats::setNames(ddays(2), "y")
  expect_named(vec_c(x, y), c("", "y"))
})

test_that("can row bind Duration objects", {
  skip_if_cant_set_s4_names()
  x <- ddays(1)
  x_named <- stats::setNames(x, "x")
  expect_identical(vec_rbind(x_named, x_named), data.frame(x = c(x, x)))
})

test_that("can row bind data frames with Duration objects", {
  expect_identical(
    vec_rbind(data.frame(x = ddays(1)), data.frame(x = ddays(1))),
    data.frame(x = ddays(c(1, 1)))
  )
})

test_that("can column bind Duration objects", {
  expect_identical(
    vec_cbind(x = ddays(1), y = ddays(1:2)),
    data.frame(x = ddays(c(1, 1)), y = ddays(1:2))
  )
})

test_that("can column bind data frames with Duration objects", {
  expect_identical(
    vec_cbind(data.frame(x = ddays(1)), data.frame(y = ddays(1:2))),
    data.frame(x = ddays(c(1, 1)), y = ddays(1:2))
  )
})

# ------------------------------------------------------------------------------
# Interval - proxy / restore

test_that("proxy is a data frame", {
  x <- interval(tzone = "UTC")

  expect <- list(start = POSIXct(tz = "UTC"), span = numeric())
  expect <- new_data_frame(expect)

  expect_identical(vec_proxy(x), expect)
})

test_that("proxy stores names on the `$span` column (allowing duplicates)", {
  skip_if_cant_set_s4_names()

  x <- c("2019-01-01", "2019-01-02", "2019-01-03")
  x <- stats::setNames(interval(x), c("x", "y", "x"))

  proxy <- vec_proxy(x)

  expect_named(proxy$span, names(x))
})

test_that("comparison / equality proxies have the names too", {
  skip_if_cant_set_s4_names()

  x <- c("2019-01-01", "2019-01-02", "2019-01-03")
  x <- stats::setNames(interval(x), c("x", "y", "x"))

  expect_named(vec_proxy_compare(x)$span, names(x))
  expect_named(vec_proxy_equal(x)$span, names(x))
})

test_that("restore method works", {
  x <- interval(c("2019-01-01", "2019-01-02"), c("2020-01-01", "2020-01-02"))
  expect_identical(vec_restore(vec_proxy(x), x), x)
})

test_that("restore method retains names", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(interval("2019-01-01"), "x")
  expect_named(vec_restore(vec_proxy(x), x), "x")
})

# ------------------------------------------------------------------------------
# Interval - ptype2

test_that("Interval default ptype2 method falls through to `vec_default_ptype2()`", {
  verify_errors({
    expect_error(vec_ptype2(interval(), 1), class = "vctrs_error_incompatible_type")
    expect_error(vec_ptype2(1, interval()), class = "vctrs_error_incompatible_type")
  })
})

test_that("common type of Interval and Interval exists", {
  expect_identical(vec_ptype2(interval(), interval()), interval())

  x <- interval(tzone = "America/Los_Angeles")
  expect_identical(vec_ptype2(x, x), x)
})

test_that("common tzone uses non-local tzone", {
  x <- interval(tzone = "")
  y <- interval(tzone = "America/Los_Angeles")

  expect_identical(vec_ptype2(x, y)@tzone, "America/Los_Angeles")
  expect_identical(vec_ptype2(y, x)@tzone, "America/Los_Angeles")

  expect_identical(tz(int_start(vec_ptype2(x, y))), "America/Los_Angeles")
  expect_identical(tz(int_start(vec_ptype2(y, x))), "America/Los_Angeles")
})

test_that("common tzone is order dependent", {
  x <- interval(tzone = "America/New_York")
  y <- interval(tzone = "America/Los_Angeles")

  expect_identical(vec_ptype2(x, y)@tzone, "America/New_York")
  expect_identical(vec_ptype2(y, x)@tzone, "America/Los_Angeles")

  expect_identical(tz(int_start(vec_ptype2(x, y))), "America/New_York")
  expect_identical(tz(int_start(vec_ptype2(y, x))), "America/Los_Angeles")
})

test_that("common type of Interval and NULL exists", {
  expect_identical(vec_ptype2(interval(), NULL), interval())
  expect_identical(vec_ptype2(NULL, interval()), interval())
})

test_that("common type of Interval and unspecified exists", {
  expect_identical(vec_ptype2(interval(), NA), interval())
  expect_identical(vec_ptype2(NA, interval()), interval())

  expect_identical(vec_ptype2(interval(), unspecified()), interval())
  expect_identical(vec_ptype2(unspecified(), interval()), interval())
})

# ------------------------------------------------------------------------------
# Interval - cast

test_that("Interval default cast method falls through to `vec_default_cast()`", {
  verify_errors({
    expect_error(vec_cast(interval(), 1), class = "vctrs_error_incompatible_type")
    expect_error(vec_cast(1, interval()), class = "vctrs_error_incompatible_type")
  })
})

test_that("Interval can be cast to Interval", {
  expect_identical(vec_cast(interval(), interval()), interval())
})

test_that("can cast to a different tzone", {
  x_tzone <- "America/Los_Angeles"
  x_start <- as.POSIXct("1970-01-01", tz = x_tzone)
  x_end <- as.POSIXct("1970-01-02", tz = x_tzone)
  x <- interval(x_start, x_end, x_tzone)

  to_tzone <- "America/New_York"
  to <- interval(tzone = to_tzone)

  expect_start <- with_tz(x_start, to_tzone)
  expect_end <- with_tz(x_end, to_tzone)
  expect <- interval(expect_start, expect_end, to_tzone)

  expect_identical(vec_cast(x, to), expect)
})

test_that("can cast around `NULL`", {
  expect_identical(vec_cast(NULL, interval()), NULL)
  expect_identical(vec_cast(interval(), NULL), interval())
})

test_that("can cast unspecified to Interval", {
  expect_identical(vec_cast(NA, interval()), interval()[NA_real_])
  expect_error(vec_cast(interval(), NA), class = "vctrs_error_incompatible_type")
})

# ------------------------------------------------------------------------------
# Interval - vctrs functionality

test_that("can slice Interval objects", {
  x <- interval(c("1970-01-01", "1970-01-02"))
  expect_identical(vec_slice(x, 2:1), x[2:1])
})

test_that("slicing preserves names", {
  skip_if_cant_set_s4_names()
  x <- interval(c("1970-01-01", "1970-01-02"))
  x <- stats::setNames(x, c("x", "y"))
  expect_named(vec_slice(x, c(1, 1, 2)), c("x", "x", "y"))
})

test_that("can combine Interval objects", {
  x <- interval("1970-01-01")
  y <- interval("1970-01-02")
  expect <- interval(c("1970-01-01", "1970-01-02"))
  expect_identical(vec_c(x, y), expect)
})

test_that("can combine Interval objects when some are named (#1072)", {
  skip_if_cant_set_s4_names()
  x <- interval("1970-01-01")
  y <- stats::setNames(interval("1970-01-02"), "y")
  expect_named(vec_c(x, y), c("", "y"))
})

test_that("can row bind Interval objects", {
  skip_if_cant_set_s4_names()
  x <- interval("1970-01-01")
  x_named <- stats::setNames(x, "x")
  expect_identical(vec_rbind(x_named, x_named), data.frame(x = c(x, x)))
})

test_that("can row bind data frames with Interval objects", {
  x <- interval("1970-01-01")

  expect_identical(
    vec_rbind(data.frame(x = x), data.frame(x = x)),
    data.frame(x = vec_c(x, x))
  )
})

test_that("can column bind Interval objects", {
  x <- interval("1970-01-01")
  y <- interval(c("1970-01-01", "1970-01-02"))

  expect_identical(
    vec_cbind(x = x, y = y),
    data.frame(x = vec_c(x, x), y = y)
  )
})

test_that("can column bind data frames with Interval objects", {
  x <- interval("1970-01-01")
  y <- interval(c("1970-01-01", "1970-01-02"))

  expect_identical(
    vec_cbind(data.frame(x = x), data.frame(y = y)),
    data.frame(x = vec_c(x, x), y = y)
  )
})

test_that("Interval objects can be ordered", {
  x <- interval("1970-01-01", "1970-01-02")
  y <- interval("1970-01-02", "1970-01-03")
  z <- interval("1970-01-02", "1970-01-04")

  # Different from `order()`!
  expect_identical(vec_order(vec_c(y, x)), c(2L, 1L))

  expect_identical(vec_order(vec_c(z, y)), c(2L, 1L))
})

# ------------------------------------------------------------------------------
# Output

test_that("vctrs methods have informative errors", {
  # partial match warning from [<-.data.frame messes up snapshots
  skip_if_not(getRversion() >= "3.6.0")

  expect_snapshot(error = TRUE, {
    "# no common type when mixing Period/Duration/Interval"
    vec_ptype2(period(), duration())
    vec_ptype2(duration(), period())

    vec_ptype2(period(), interval())
    vec_ptype2(interval(), period())

    vec_ptype2(duration(), interval())
    vec_ptype2(interval(), duration())

    "# can't cast between Period/Duration/Interval"
    vec_cast(period(), duration())
    vec_cast(duration(), period())

    vec_cast(period(), interval())
    vec_cast(interval(), period())

    vec_cast(duration(), interval())
    vec_cast(interval(), duration())

    "# Period default ptype2 method falls through to `vec_default_ptype2()`"
    vec_ptype2(period(), 1)
    vec_ptype2(1, period())

    "# Period default cast method falls through to `vec_default_cast()`"
    vec_cast(period(), 1)
    vec_cast(1, period())

    "# Duration default ptype2 method falls through to `vec_default_ptype2()`"
    vec_ptype2(duration(), 1)
    vec_ptype2(1, duration())

    "# Duration default cast method falls through to `vec_default_cast()`"
    vec_cast(duration(), 1)
    vec_cast(1, duration())

    "# Interval default ptype2 method falls through to `vec_default_ptype2()`"
    vec_ptype2(interval(), 1)
    vec_ptype2(1, interval())

    "# Interval default cast method falls through to `vec_default_cast()`"
    vec_cast(interval(), 1)
    vec_cast(1, interval())
  })
})

test_that("duration() returns zero-length vector", {
  x <- duration()
  expect_s4_class(x, "Duration")
  expect_length(x, 0)
  expect_equal(format(x), character())
  expect_output(print(x), "<Duration[0]>", fixed = TRUE)
})

test_that("duration(...) returns zero-length with on 0-length inputs", {
  x <- duration(character())
  expect_s4_class(x, "Duration")
  expect_length(x, 0)

  x <- duration(hour = numeric())
  expect_s4_class(x, "Duration")
  expect_length(x, 0)

  x <- duration(numeric(), hour = numeric())
  expect_s4_class(x, "Duration")
  expect_length(x, 0)

  expect_equal(
    duration(c(30, 20)),
    duration(c(30, 20), days = numeric())
  )

  expect_equal(
    duration(numeric(), days = 10),
    duration(days = 10)
  )
})

test_that("duration constructor doesn't accept non-numeric or non-character inputs", {
  expect_error(duration(interval(ymd("2014-01-01"), ymd("2015-01-01"))))
})

test_that("make_difftime works as expected", {
  x <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  y <- difftime(x + 5 + 30 * 60 + 60 * 60 + 14 * 24 * 60 * 60, x, tz = "UTC")
  attr(y, "tzone") <- NULL
  diff <- make_difftime(seconds = 5, minutes = 30, days = 0, hour = 1, weeks = 2)
  expect_equal(diff, y)
})

test_that("Duration parsing works", {
  expect_equal(
    duration("1min 2sec 2secs 1H 2M 1d"),
    duration(seconds = 4, minutes = 3, hours = 1, days = 1)
  )
  expect_equal(
    duration("day day"),
    duration(days = 2)
  )
  expect_equal(
    duration("S M H d w"),
    duration(seconds = 1, minutes = 1, hours = 1, days = 1, weeks = 1)
  )
  expect_equal(
    duration("1S 2M 3H 4d 5w"),
    duration(seconds = 1, minutes = 2, hours = 3, days = 4, weeks = 5)
  )
})

test_that("fractional parsing works as expected", {
  expect_equal(
    duration("1.1min 2.3sec 2.3secs 1.0H 2.2M 1.5d"),
    duration(seconds = 43222.6, minutes = 3, hours = 1, days = 1)
  )
  expect_equal(
    duration("day 1.2days"),
    duration(days = 2, seconds = 17280)
  )
})

test_that("sub-unit fractional parsing works as expected", {
  expect_identical(
    duration(".1min .3sec .3secs .0H .2M .5d"),
    duration("0.1min 0.3sec 0.3secs 0.0H 0.2M 0.5d")
  )

  expect_identical(
    duration(".1min .3sec .3secs .0H .2M .5d"),
    duration(seconds = 6.6, minutes = .2, hours = 12)
  )
})

test_that("parsing with 0 units works as expected", {
  expect_equal(as.numeric(duration("2d 0H 0M 1s")), 2 * 24 * 3600 + 1)
  expect_equal(as.numeric(duration("0d 0H 0M 1s")), 1)
  expect_equal(period("2d 0H 0M 1s"), days(2) + seconds(1))
  expect_equal(period("y 0m 2d 0H 0M 1s"), years(1) + days(2) + seconds(1))
})

test_that("as.integer works with durrations", {
  d1 <- duration("2d 0H 2M 1s")
  expect_equal(as.integer(d1, units = "minute"), as.integer(as.numeric(d1, units = "minute")))
  expect_equal(as.integer(d1, units = "hour"), as.integer(as.numeric(d1, units = "hour")))
  expect_equal(as.integer(d1), as.numeric(d1))
  expect_type(as.integer(d1), "integer")
})

test_that("make_difftime handles vectors", {
  x <- as.POSIXct(c("2008-08-03 13:01:59", "2008-08-03 13:01:59"), tz = "UTC")
  y <- difftime(x + c(
    5 + 30 * 60 + 60 * 60 + 14 * 24 * 60 * 60,
    1 + 3 * 24 * 60 * 60 + 60 * 60
  ), x, tz = "UTC")
  attr(y, "tzone") <- NULL
  z <- difftime(x + c(
    5 + 30 * 60 + 60 * 60 + 14 * 24 * 60 * 60,
    5 + 30 * 60 + 60 * 60 + 14 * 24 * 60 * 60 + 3 * 24 * 60 * 60
  ),
  x,
  tz = "UTC"
  )
  attr(z, "tzone") <- NULL

  expect_equal(
    make_difftime(
      seconds = c(5, 1), minutes = c(30, 0),
      days = c(0, 3), hour = c(1, 1), weeks = c(2, 0)
    ),
    y
  )

  expect_equal(
    make_difftime(
      seconds = 5, minutes = 30,
      days = c(0, 3), hour = 1, weeks = 2
    ),
    z
  )
})

test_that("duration works as expected", {
  dur <- duration(
    seconds = 5, minutes = 30, days = 0,
    hour = 1, weeks = 2
  )

  expect_equal(dur@.Data, 1215005)
  expect_s4_class(dur, "Duration")
})

test_that("duration handles vectors", {
  dur1 <- duration(
    seconds = c(5, 1), minutes = c(30, 0),
    days = c(0, 3), hour = c(1, 1), weeks = c(2, 0)
  )
  dur2 <- duration(
    seconds = 5, minutes = 30, days =
      c(0, 3), hour = 1, weeks = 2
  )

  expect_equal(dur1@.Data, c(1215005, 262801))
  expect_equal(dur2@.Data, c(1215005, 1474205))
  expect_s4_class(dur1, "Duration")
  expect_s4_class(dur2, "Duration")
})

test_that("as.duration handles vectors", {
  expect_equal(as.duration(minutes(1:3)), dminutes(1:3))
})

test_that("as.duration handles periods", {
  expect_equal(as.duration(seconds(1)), dseconds(1))
  expect_equal(as.duration(minutes(2)), dminutes(2))
  expect_equal(as.duration(hours(3)), dhours(3))
  expect_equal(as.duration(days(4)), ddays(4))
  expect_equal(as.duration(weeks(5)), dweeks(5))
  expect_equal(as.duration(months(1)), dseconds(60 * 60 * 24 * 365.25 / 12))
  expect_equal(as.duration(years(1)), dseconds(60 * 60 * 24 * 365.25))
  expect_equal(as.duration(seconds(1) + minutes(4)), dseconds(1) + dminutes(4))
})

test_that("parsing duration's allows for a full roundtrip", {
  expect_equal(as.duration(lubridate:::parse_period("31557600s (~1 years)")), dseconds(31557600))
  durs <- duration(c("1m 2s, 3days", "2.5min 1day"))
  expect_equal(durs, as.duration(as.character(durs)))
})

test_that("as.duration handles intervals", {
  time1 <- as.POSIXct("2009-01-02 12:24:03", tz = "UTC")
  time2 <- as.POSIXct("2010-02-03 14:31:42", tz = "UTC")
  dur <- as.duration(interval(time1, time2))
  y <- as.numeric(time2 - time1, units = "secs")

  expect_equal(dur@.Data, y)
  expect_s4_class(dur, "Duration")
})

test_that("as.duration handles difftimes", {
  x <- difftime(
    as.POSIXct("2010-02-03 14:31:42", tz = "UTC"),
    as.POSIXct("2009-01-02 12:24:03", tz = "UTC")
  )
  dur <- as.duration(x)
  y <- as.numeric(x, units = "secs")

  expect_equal(dur@.Data, y)
  expect_s4_class(dur, "Duration")
})

test_that("eobjects handle vectors", {
  dur <- dseconds(c(1, 3, 4))

  expect_equal(dur@.Data, c(1, 3, 4))
  expect_s4_class(dur, "Duration")
})

test_that("is.duration works as expected", {
  ct_time <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  lt_time <- as.POSIXlt("2009-08-03 13:01:59", tz = "UTC")

  expect_false(is.duration(234))
  expect_false(is.duration(ct_time))
  expect_false(is.duration(lt_time))
  expect_false(is.duration(Sys.Date()))
  expect_false(is.duration(minutes(1)))
  expect_true(is.duration(dminutes(1)))
  expect_false(is.duration(make_difftime(1000)))
  expect_false(is.duration(interval(lt_time, ct_time)))
})

test_that("format.Duration works as expected", {
  dur <- duration(seconds = c(5, NA, 10, -10, 1000, -1000))
  expect_equal(
    format(dur),
    c(
      "5s", NA, "10s", "-10s",
      "1000s (~16.67 minutes)",
      "-1000s (~-16.67 minutes)"
    )
  )
})

test_that("summary.Duration creates useful summary", {
  dur <- dminutes(5)
  text <- c(rep("300s (~5 minutes)", 6), 1)
  names(text) <- c("Min.", "1st Qu.", "Median", "Mean", "3rd Qu.", "Max.", "NA's")
  expect_equal(summary(c(dur, NA)), text)
})

test_that("format.Duration works with NA values", {
  expect_equal(
    format(duration(c(NA, 1, 100, 10000, 100000, 100000000, 75.25, 75.5001)), scientific = 7),
    c(
      NA, "1s", "100s (~1.67 minutes)", "10000s (~2.78 hours)", "1e+05s (~1.16 days)",
      "1e+08s (~3.17 years)",
      "75.25s (~1.25 minutes)", "75.5001s (~1.26 minutes)"
    )
  )
  expect_equal(
    format(duration(c(75.25, 75.5001))),
    c("75.25s (~1.25 minutes)", "75.5001s (~1.26 minutes)")
  )
})

test_that("as.duration handles NA interval objects", {
  one_missing_date <- as.POSIXct(NA_real_, origin = origin)
  one_missing_interval <- interval(
    one_missing_date,
    one_missing_date
  )
  several_missing_dates <- rep(as.POSIXct(NA_real_, origin = origin), 2)
  several_missing_intervals <- interval(
    several_missing_dates,
    several_missing_dates
  )
  start_missing_intervals <- interval(several_missing_dates, origin)
  end_missing_intervals <- interval(origin, several_missing_dates)
  na.dur <- dseconds(NA)

  expect_equal(as.duration(one_missing_interval), na.dur)
  expect_equal(as.duration(several_missing_intervals), c(na.dur, na.dur))
  expect_equal(as.duration(start_missing_intervals), c(na.dur, na.dur))
  expect_equal(as.duration(end_missing_intervals), c(na.dur, na.dur))
})

test_that("as.duration handles NA period objects", {
  na.dur <- dseconds(NA)
  expect_equal(as.duration(years(NA)), na.dur)
  expect_equal(as.duration(years(c(NA, NA))), c(na.dur, na.dur))
  expect_equal(as.duration(years(c(1, NA))), c(dyears(1), na.dur))
})

test_that("as.duration handles NA objects", {
  na.dur <- dseconds(NA)
  expect_equal(as.duration(NA), na.dur)
})


test_that("c.duration works with named elements", {
  nmd <- c(a = dhours(1), b = dminutes(30))
  umd <- c(dhours(1), dminutes(30))
  names(umd) <- c("a", "b")
  expect_equal(nmd, umd)

  nmd <- c(a = dhours(c(1, 1)), b = dminutes(c(30, 31)))
  umd <- c(dhours(1), dhours(1), dminutes(30), dminutes(31))
  names(umd) <- c("a1", "a2", "b1", "b2")
  expect_equal(nmd, umd)

  ## empty elements
  nmd <- c(a = dhours(1), NULL, b = dminutes(30), seconds(10))
  umd <- c(dhours(1), dminutes(30), seconds(10))
  names(umd) <- c("a", "b", "")
  expect_equal(nmd, umd)
})

test_that("is.timespan works as expected", {
  expect_false(is.timespan(234))
  expect_false(is.timespan(as.POSIXct("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.timespan(as.POSIXlt("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.timespan(Sys.Date()))
  expect_true(is.timespan(minutes(1)))
  expect_true(is.timespan(dminutes(1)))
  expect_true(is.timespan(interval(
    as.POSIXct("2008-08-03 13:01:59", tz = "UTC"),
    as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  )))
})

test_that("time_length works as expected", {
  expect_equal(
    time_length(period(1, "day")), # period
    86400
  )
  expect_equal(
    time_length(ymd("2014-12-15") - ymd("2014-11-30"), "day"), # difftime
    15
  )
  expect_equal(
    time_length(ymd("2014-12-15") - ymd("2014-11-30"), "week"),
    15 / 7
  )
  expect_equal(
    time_length(duration(3, "months"), "month"), # duration
    3
  )
  expect_equal(
    time_length(interval(ymd("2014-11-30"), ymd("2014-12-15")), "weeks"), # interval
    15 / 7
  )
  expect_lt(
    time_length(interval(ymd("1900-01-01"), ymd("1999-12-31")), "years"),
    100
  )
  expect_lt(
    time_length(as.duration(interval(ymd("1900-01-01"), ymd("1999-12-31"))), "years"),
    100
  )
  expect_equal(
    -time_length(interval(ymd("1900-01-01"), ymd("2000-01-01")), "days"),
    time_length(int_flip(interval(ymd("1900-01-01"), ymd("2000-01-01"))), "days")
  )
})

test_that("time_length works with intervals on DST", {
  times <- ymd_hms(c("2019-03-29 02:03:00", "2019-03-29 02:10:00"), tz = "Europe/Paris")
  expect_equal(time_length(interval(times[[1]], times[[2]]), "years"), 1.32818e-05)
})

test_that("time_length works with intervals on DST (2)", {
  skip("FIXME: time_length should be same on sides of DST?")
  expect_equal(
    time_length(interval(times[[1]], times[[2]]), "years"),
    time_length(interval(times2[[1]], times2[[2]]), "years"))
  expect_equal(
    time_length(interval(times[[1]], times[[2]]), "years"),
    time_length(interval(times3[[1]], times3[[2]]), "years"))
})

test_that("time_length works with missing intervals", {
  expect_equal(time_length(interval(NA, ymd("2016-01-01")), unit = "year"), NA_real_)
})

test_that("time_length works with birth date 29 Feb ", {
  expect_equal(
    round(time_length(interval(ymd("1992-02-29"), ymd("1999-02-28")), "years"), 4),
    6.9973
  )
  expect_equal(
    round(time_length(interval(ymd("1992-02-29"), ymd("1999-03-31")), "years"), 4),
    7.0822
  )
  expect_equal(
    time_length(interval(ymd("1992-02-29"), ymd("1999-03-01")), "years"),
    7
  )
  expect_equal(
    round(time_length(interval(ymd("1992-02-29"), ymd("2000-03-01")), "years"), 4),
    8.0027
  )
  expect_equal(
    time_length(interval(ymd("1992-02-29"), ymd("2000-02-29")), "years"),
    8
  )
  expect_gt(
    time_length(interval(ymd_hms("1992-02-29 12:00:00"), ymd_hms("1999-03-01 05:00:00")), "years"),
    7
  )
})

test_that("time_length works with negative interals", {
  expect_equal(
    -time_length(interval(ymd("1992-02-28"), ymd("2000-01-01")), "days"),
    time_length(int_flip(interval(ymd("1992-02-28"), ymd("2000-01-01"))), "days")
  )
  expect_equal(
    -time_length(interval(ymd("1992-02-28"), ymd("2000-03-01")), "days"),
    time_length(int_flip(interval(ymd("1992-02-28"), ymd("2000-03-01"))), "days")
  )

  ## If both ends include leap years Febs, the lenths are not identical
  int <- interval(ymd("1992-02-28"), ymd("2000-03-01"))
  expect_equal(-time_length(int, "years"), time_length(int_flip(int), "years"))

  ## or if both ends don't include leap years Febs, the lenths are identical
  int <- interval(ymd("1994-02-28"), ymd("2002-03-01"))
  expect_equal(-time_length(int, "years"), time_length(int_flip(int), "years"))

  ## ... otherwise not
  int <- interval(ymd("1992-02-28"), ymd("2002-01-01"))
  expect_gt(-time_length(int, "years"), time_length(int_flip(int), "years"))
})

test_that("time_length handles vectors", {
  expect_equal(
    time_length(days(1:3), unit = "days"),
    1:3
  )
  expect_equal(
    time_length(as.interval(days(1:3), start = today()), unit = "days"),
    1:3
  )
  expect_equal(
    time_length(as.interval(days(c(1:3, NA)), start = today()), unit = "days"),
    c(1:3, NA)
  )

  ints <- new("Interval",
    .Data = c(
      6363040, 1646253613, 159312603, -510371200, -381983229, 1369742727,
      -190254638, -1118394043, -2091814107
    ),
    start = structure(c(
      64156663, 326788049, 462064351, 909110414, 778538359,
      395846311, 1415640365, 1697694335, 2193040469
    ),
    class = c("POSIXct", "POSIXt"), tzone = "UTC"
    ),
    tzone = "UTC"
  )

  expect_equal(
    round(time_length(ints, "years"), 5),
    c(
      0.20122, 52.16674, 5.04903, -16.17281, -12.10439, 43.40412,
      -6.03011, -35.44092, -66.28715
    )
  )

  expect_equal(
    round(time_length(ints[ints > 0], "years"), 5),
    c(0.20122, 52.16674, 5.04903, 43.40412)
  )

  expect_equal(
    round(time_length(ints[ints < 0], "years"), 5),
    c(-16.17281, -12.10439, -6.03011, -35.44092, -66.28715)
  )

  expect_equal(
    round(time_length(ints, "month"), 5),
    c(
      2.4402, 625.99552, 60.57729, -194.06691, -145.2291, 520.81679,
      -72.35552, -425.27918, -795.43557
    )
  )
})

test_that("time_length handles 0 length intervals", {
  date <- ymd_hms("2014-12-30 20:35:13")
  int <- interval(c(date, date), c(date, date))
  expect_equal(time_length(int), c(0, 0))
  expect_equal(time_length(int, "years"), c(0, 0))
  expect_equal(time_length(int, "months"), c(0, 0))
})

context("Daylight savings times")

test_that("force_tz returns NA for a time that falls in the spring gap", {
  x <- structure(1268532305, class = c("POSIXct", "POSIXt"), tzone = "UTC")
  y <- structure(1268618705, class = c("POSIXct", "POSIXt"), tzone = "UTC")
  z <- structure(1268633105, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  expect_true(is.na(force_tz(x, "America/New_York")))
  expect_equal(force_tz(c(y, x), "America/New_York"), c(z, NA))
})

test_that("force_tz behaves consistently for the fall overlap", {

  y <- structure(1289111405, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))
  y2 <- structure(1289115005, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  y3 <- structure(1289104205, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")

  z <- structure(1289197805, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  z2 <- structure(1289201405, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  z3 <- structure(1289194205, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")

  cdst <- structure(1352012400, tzone = "America/Chicago", class = c("POSIXct", "POSIXt"))
  edst <- structure(1352008800, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))

  expect_equal(y + hours(1), y2)
  expect_equal(c(y, z) + hours(1), c(y2, z2))
  expect_equal(y - hours(1), y3)
  expect_equal(c(y, z) - hours(1), c(y3, z3))
  expect_equal(force_tz(cdst, "America/New_York"), edst)

})

test_that("addition handles daylight savings time for spring gap", {
  x <- structure(1268542800, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px <- structure(1268625600, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px2 <- structure(1268712000, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex <- structure(1268629200, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex2 <- structure(1268715600, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))

  expect_equal(x + days(1), px)
  expect_equal(c(x, px) + days(1), c(px, px2))
  expect_equal(x + ddays(1), ex)
  expect_equal(c(x, ex) + ddays(1), c(ex, ex2))

})


test_that("subtraction handles daylight savings time for spring gap", {
  x <- structure(1268625600, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px <- structure(1268542800, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px2 <- structure(1268456400, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex <- structure(1268539200, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex2 <- structure(1268452800, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))

  expect_equal(x - days(1), px)
  expect_equal(c(x, px) - days(1), c(px, px2))
  expect_equal(x - ddays(1), ex)
  expect_equal(c(x, ex) - ddays(1), c(ex, ex2))

})

test_that("addition handles daylight savings time for fall overlap", {
  x <- structure(1289102400, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px <- structure(1289192400, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px2 <- structure(1289278800, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex <- structure(1289188800, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex2 <- structure(1289275200, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")

  expect_equal(x + days(1), px)
  expect_equal(c(x, px) + days(1), c(px, px2))
  expect_equal(x + ddays(1), ex)
  expect_equal(c(x, ex) + ddays(1), c(ex, ex2))
})


test_that("subtraction handles daylight savings time for fall overlap", {
  x <- structure(1289192400, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px <- structure(1289102400, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  px2 <- structure(1289016000, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  ex <- structure(1289106000, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))
  ex2 <- structure(1289019600, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))

  expect_equal(x - days(1), px)
  expect_equal(c(x, px) - days(1), c(px, px2))
  expect_equal(x - ddays(1), ex)
  expect_equal(c(x, ex) - ddays(1), c(ex, ex2))

})


test_that("update returns NA for date-times in the spring dst gap", {

  poslt <- structure(list(sec = 59, min = 59L, hour = 1L, mday = 14L, mon = 2L,
    year = 110L, wday = 0L, yday = 72L, isdst = 0L),
    .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
    class = c("POSIXlt", "POSIXt"), tzone = c("America/New_York", "EST", "EDT"))
  x <- structure(list(sec = 59, min = 59L, hour = 1L, mday = 15L, mon = 2L,
    year = 110L, wday = 1L, yday = 73L, isdst = 1L),
    .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
    class = c("POSIXlt", "POSIXt"), tzone = c("America/New_York", "EST", "EDT"))
  x2 <- structure(list(sec = 59, min = 59L, hour = 2L, mday = 15L, mon = 2L,
    year = 110L, wday = 1L, yday = 73L, isdst = 1L),
    .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
    class = c("POSIXlt", "POSIXt"), tzone = c("America/New_York", "EST", "EDT"))
  xu <- structure(list(sec = 59, min = 59L, hour = 2L, mday = 15L, mon = 2L,
    year = 110L, wday = 1L, yday = 73L, isdst = 1L),
    .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
    class = c("POSIXlt", "POSIXt"), tzone = "UTC")
  expect_true(is.na(update(poslt, seconds = 61)))
  expect_true(is.na(update(poslt, minutes = 61)))
  expect_true(is.na(update(poslt, hours = 2)))
  expect_equal(update(c(x, poslt), hours = 2), c(x2, NA))

  poslt <- structure(list(sec = 59, min = 59L, hour = 2L, mday = 13L, mon = 2L,
    year = 110L, wday = 6L, yday = 71L, isdst = 0L),
    .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
    class = c("POSIXlt", "POSIXt"), tzone = c("America/New_York", "EST", "EDT"))
  expect_true(is.na(update(poslt, mdays = 14)))
  expect_true(is.na(update(poslt, wdays = 8)))
  expect_true(is.na(update(poslt, ydays = 73)))

  poslt <- structure(list(sec = 59, min = 59L, hour = 2L, mday = 14L, mon = 1L,
    year = 110L, wday = 0L, yday = 44L, isdst = 0L),
    .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
    class = c("POSIXlt", "POSIXt"), tzone = c("America/New_York", "EST", "EDT"))
  expect_true(is.na(update(poslt, months = 3)))

  ## poslt <- structure(list(sec = 59, min = 59L, hour = 2L, mday = 14L, mon = 2L,
  ##   year = 109L, wday = 6L, yday = 72L, isdst = 0L),
  ##   .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
  ##   class = c("POSIXlt", "POSIXt"), tzone = c("America/New_York", "EST", "EDT"))
  ## expect_true(is.na(update(poslt, years = 2010)))

  poslt <- structure(list(sec = 59, min = 59L, hour = 2L, mday = 14L, mon = 2L,
    year = 110L, wday = 0L, yday = 72L, isdst = 0L),
    .Names = c("sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst"),
    class = c("POSIXlt", "POSIXt"), tzone = "UTC")
  expect_true(is.na(update(poslt, tzs = "America/New_York")))
  expect_equal(update(c(xu, poslt), tzs = "America/New_York"), c(x2, NA))

})

test_that("update rollovers perform correctly across the fall overlap", {
  timest <- structure(1289113199, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))
  posct <- structure(1289109659, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))
  timest2 <- structure(1289116919, class = c("POSIXct", "POSIXt"), tzone = "America/New_York")
  timest3 <- structure(1289124119, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))
  timest0 <- structure(1289116859, tzone = "America/New_York", class = c("POSIXct", "POSIXt"))
  expect_equal(update(timest, minutes = 121), timest2)
  expect_equal(update(c(timest, timest2), minutes = 121), c(timest2, timest3))
  expect_equal(update(timest, minutes = 0), posct)
  expect_equal(update(c(timest, timest2), minutes = 0), c(posct, timest0))
})

test_that("arithmetic handles fall gap in timespan appropriate way", {
  cdst <- structure(1352012400, tzone = "America/Chicago", class = c("POSIXct",                                                                  "POSIXt"))

  expect_equal(cdst - seconds(1), cdst - 3601)
  expect_equal(cdst - dseconds(1), cdst - 1)
})

context("Division operators")

test_that("division operations work for interval numerator", {
  int <- ymd("2009-02-03") %--% ymd("2010-01-01")
  smaller_int <- ymd("2009-12-01") %--% ymd("2010-01-01")
  bigger_int <- ymd("2009-01-01") %--% ymd("2010-02-01")
  smaller_dur <- ddays(20) + dhours(4)
  bigger_dur <- dyears(1) + dseconds(2)
  smaller_diff <- make_difftime(days = 100)
  bigger_diff <- make_difftime(days = 400)
  expect_error(int/smaller_int)
  expect_error(int/bigger_int)
  expect_equal(int_start(int) + days(int/days(1)), int_end(int))
  expect_equal(int_start(int) + ddays(int/ddays(1)), int_end(int))
  expect_equal(int/smaller_diff, 28684800/8640000)
  expect_equal(int/bigger_diff, 28684800/34560000)
})

test_that("division works for interval numerator with vectors", {
  int1 <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
  int3 <- ymd("2009-01-01") %--% ymd("2010-01-01")
  smaller_diff <- make_difftime(days = 365)
  expect_error(c(int1, int2) / int1)
  expect_error(int1/c(int1, int2))
  expect_equal(c(int1, int2) / years(1), c(1, 2))
  expect_equal(c(int1, int2) / dyears(1), c(0.9993155, 1.9986311), tolerance = 6)
  expect_equal(int1/years(1:2), c(1, 0.5))
  expect_equal(int1/dyears(1:2), c(0.9993155, 0.4996578), tolerance = 6)
  expect_equal(c(int1, int2) / smaller_diff, c(1, 2))
  expect_equal(int2/c(1, 2), c(int2, int3))
  expect_equal(c(int2, int2) / 2, c(int3, int3))
})

test_that("arguments are recycled in interval division", {
expect_equal(interval(c('2015-03-01', '2015-03-02', '2015-03-03'), '2016-10-01') / months(1),
             c(19.00000000, 18.96666667, 18.93333333))
expect_equal(interval(c('2015-03-01', '2015-03-02', '2015-03-03', '2015-03-04'), '2016-10-01') / months(1:2),
             interval(c('2015-03-01', '2015-03-02', '2015-03-03', '2015-03-04'), '2016-10-01') / months(c(1:2, 1:2)))
expect_equal(interval('2015-03-02', '2016-10-01') / months(1:3),
             interval(c('2015-03-02', '2015-03-02', '2015-03-02'), '2016-10-01') / months(1:3))
})

test_that("NA and Interval(0) propagate in interval division", {
  t <- ymd_hms("2019-03-01 12:30:50")
  int0 <- interval(character(0), t)
  expect_equal(int0/days(3), numeric(0))
  intNA <- interval(NA, t)
  expect_equal(intNA/days(3), NA_real_)
})

test_that("division works for interval numerator with period denominator", {
  int1 <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
  int3 <- ymd("2009-01-01") %--% ymd("2010-01-01")
  expect_equal(c(int1, int2) / days(1), time_length(c(int1, int2), "days"))
  expect_equal(c(int1, int2) / minutes(1), time_length(c(int1, int2), "mins"))
  expect_equal(c(int1, int2) / months(1), c(12, 24))
  expect_equal(c(int1, int2) / months(6), c(2, 4))
  expect_equal(c(int1, int2) / years(1), c(1, 2))
  expect_equal(c(int1, int2) / years(2), c(.5, 1))
})

test_that("dividision works for interval vector with NAs", {
  int1 <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
  int3 <- ymd(NA, quiet = TRUE) %--% ymd(NA, quiet = TRUE)
  ints <- c(int1, int2, int3)
  nas <- as.numeric(c(NA, NA, NA))
  expect_equal(ints / years(1), c(1, 2, NA))
  expect_equal(ints / dyears(1), c(0.9993155, 1.9986311, NA), tolerance = 6)
  expect_equal(ints / years(NA), nas)
  expect_equal(ints / dyears(NA), nas)
})

test_that("division operations work for period numerator", {
  int1 <- ymd("2009-01-01") %--% ymd("2010-01-01")
  expect_error(years(2) / int1)
  expect_equal(days(2) / hours(12), 4)
  expect_equal(years(1) / months(1), 12)
  expect_error(days(2) / dhours(12))
  expect_error(days(1) / make_difftime(days = 1))
  expect_equal(days(2) / 2, days(1))
})

test_that("division works for period numerator with vectors", {
  int1 <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
  expect_error(years(1:2) / int2)
  expect_error(years(1) / c(int1, int2))
  expect_equal(days(1:2) / hours(12), c(2, 4))
  expect_equal(days(1) / hours(c(12, 24)), c(2, 1))
  expect_error(days(1:2) / dhours(12))
  expect_error(days(1) / dhours(c(12, 24)))
  expect_error(days(1:2) / make_difftime(days = 1))
  expect_equal(days(2) / c(1, 2), c(days(2), days(1)))
  expect_equal(days(c(2, 4)) / c(2), c(days(1), days(2)))

})

test_that("division operations work for duration numerator", {
  int <-  ymd("2009-12-01") %--% ymd("2010-01-01")
  expect_error(ddays(31)/int)
  expect_error(ddays(20)/ days(20))
  expect_equal(ddays(20)/ ddays(1), 20)
  expect_equal(ddays(20)/ make_difftime(days = 1), 20)
})

test_that("division works for duration numerator with vectors", {
  int1 <- ymd("2010-01-01") %--% ymd("2011-01-01")
  int2 <- ymd("2009-01-01") %--% ymd("2011-01-01")
  expect_error(ddays(365*c(1:2)) / int2)
  expect_error(ddays(365) / c(int1, int2))
  expect_equal(ddays(1:2) / dhours(12), c(2, 4))
  expect_equal(ddays(1) / dhours(c(12, 24)), c(2, 1))
  expect_error(ddays(1:2) / hours(12))
  expect_error(ddays(1) / hours(c(12, 24)))
  expect_equal(ddays(1:2) / make_difftime(days = 1), c(1, 2))
  expect_equal(ddays(2) / c(1, 2), c(ddays(2), ddays(1)))
  expect_equal(ddays(c(2, 4)) / c(2), c(ddays(1), ddays(2)))
})

test_that("division operations work for difftime numerator", {
  diff <- make_difftime(days = 365)
  int <- ymd("2010-01-01") %--% ymd("2011-01-01")
  expect_error(diff/int)
  expect_error(diff/days(1))
  expect_equal(diff/ddays(365), 1)
})

test_that("division works as expected for periods", {
  expect_error(3/months(1))
  expect_error(months(1)/3)
  expect_equal(months(9)/3, months(3))
  expect_is(months(9)/3, "Period")
})

test_that("dividing vectors works for periods", {
  expect_error(c(2, 3)/months(1))
  expect_error(months(1)/c(2, 3))
  expect_equal(months(9)/c(3, 1), months(c(3, 9)))
  expect_is(months(9)/c(3, 1), "Period")
})

test_that("division works as expected for durations", {
  expect_error(3/dhours(1))
  expect_equal(dhours(9)/3, dhours(3))
  expect_is(dhours(9)/3, "Duration")

})

test_that("dividing vectors works for durations", {
  expect_error(c(2, 3)/dhours(1))
  expect_equal(dhours(9)/c(3, 1), dhours(c(3, 9)))
  expect_is(dhours(9)/c(3, 1), "Duration")
})

context("Subtraction operations")

###### subtraction for everything

test_that("subtraction works as expected for instants", {
  x <- as.POSIXct("2008-01-01 00:00:02", tz = "UTC")
  y <- as.POSIXlt("2008-01-01 00:00:02", tz = "UTC")
  z <- as.Date("2008-01-03")

  expect_equal(x - years(1), as.POSIXct("2007-01-01 00:00:02", tz = "UTC"))
  expect_equal(y - years(1), as.POSIXlt("2007-01-01 00:00:02", tz = "UTC"))
  expect_equal(z - years(1), as.Date("2007-01-03"))

  expect_equal(x - dyears(1), as.POSIXct("2006-12-31 18:00:02", tz = "UTC"))
  expect_equal(y - dyears(1), as.POSIXlt("2006-12-31 18:00:02", tz = "UTC"))
  expect_equal(z - dyears(1), as.POSIXct("2007-01-02 18:00:00", tz = "UTC"))

  time1 <- as.POSIXct("2007-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2008-01-01 00:00:59", tz = "UTC")
  int <- interval(time1, time2)
  num <- as.numeric(time2) - as.numeric(time1)

  expect_error(x - int)
  expect_error(y - int)
  expect_error(z - int)

})

test_that("subtraction with instants returns correct class", {
  x <- as.POSIXct("2008-01-01 12:00:00", tz = "UTC")
  y <- as.POSIXlt("2008-01-01 12:00:00", tz = "UTC")
  z <- as.Date("2008-01-01")

  expect_is(x - years(1), "POSIXct")
  expect_is(y - years(1), "POSIXlt")
  expect_is(z - years(1), "Date")
  expect_is(x - dyears(1), "POSIXct")
  expect_is(y - dyears(1), "POSIXlt")
})


test_that("subtraction works as expected for periods", {
  time1 <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-02-02 00:00:00", tz = "UTC")
  int <- interval(time1, time2)

  expect_equal(years(1) - 1, period(seconds = -1, years = 1))
  expect_error(years(1) - as.POSIXct("2008-01-01 00:00:00", tz = "UTC"))
  expect_error(years(1) - as.POSIXct("2008-01-01 00:00:00", tz = "UTC"))
  expect_equal(years(1) - minutes(3), period(minutes = -3, years = 1))
  expect_error(years(1) - dyears(1))
  expect_error(years(1) - int)

})

test_that("subtraction with periods returns correct class", {

  expect_is(years(1) - 1, "Period")
  expect_is(years(1) - minutes(3), "Period")

})


test_that("subtraction works as expected for durations", {
  time1 <- as.POSIXct("2008-01-02 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  int <- interval(time1, time2)

  expect_equal(dyears(1) - 1, duration(31557599))
  expect_error(dyears(1) - as.POSIXct("2008-01-01 00:00:00", tz = "UTC"))
  expect_error(dyears(1) - minutes(3))
  expect_equal(dyears(2) - dyears(1), dyears(1))
  expect_error(dyears(1) - int)
})

test_that("subtraction with durations returns correct class", {

  expect_is(dyears(1) - 1, "Duration")
  expect_is(dyears(1) - dyears(1), "Duration")

})


test_that("subtraction works as expected for intervals", {
  time1 <- as.POSIXct("2008-08-03 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  time3 <- as.POSIXct("2008-11-02 00:00:00", tz = "UTC")
  int <- interval(time1, time2)
  int2 <- interval(time3, time2)

  expect_error(time2 - int)
  expect_error(int - 1)
  expect_error(int - time1)
  expect_error(int - minutes(3))
  expect_error(int - dminutes(3))
  expect_error(int - int2)

})


test_that("%m-% correctly subtracts months without rollover", {
  may <- ymd_hms("2010-05-31 03:04:05")
  ends <- ymd_hms(c("2010-04-30 03:04:05", "2010-03-31 03:04:05", "2010-02-28 03:04:05"))

  expect_equal(may %m-% months(1:3), ends)
})

test_that("%m-% correctly subtracts years without rollover", {
  leap <- ymd("2012-02-29")
  next1 <- ymd("2011-02-28")
  next2 <- ymd("2011-01-29")
  next3 <- ymd("2011-03-29")

  expect_equal(leap %m-% years(1), next1)
  expect_equal(leap %m-% period(years = 1, months = 1), next2)
  expect_equal(leap %m-% period(years = 1, months = -1), next3)
})

test_that("%m-% correctly subtract negative months without rollover", {
  jan <- ymd_hms("2010-01-31 03:04:05")
  ends <- ymd_hms(c("2010-02-28 03:04:05", "2010-03-31 03:04:05", "2010-04-30 03:04:05"))

  expect_equal(jan %m-% -months(1:3), ends)
})

test_that("%m-% correctly subtracts negative years without rollover", {
  leap <- ymd("2012-02-29")
  next1 <- ymd("2013-02-28")
  next2 <- ymd("2013-03-29")
  next3 <- ymd("2013-01-29")

  expect_equal(leap %m-% years(-1), next1)
  expect_equal(leap %m-% period(years = -1, months = -1), next2)
  expect_equal(leap %m-% period(years = -1, months = 1), next3)

})

context("vctrs")

library(vctrs)

# ------------------------------------------------------------------------------
# Common ptype2 / cast

test_that("no common type when mixing Period/Duration/Interval", {
  verify_errors({
    expect_error(vec_ptype2(period(), duration()), class = "vctrs_error_incompatible_type")
    expect_error(vec_ptype2(duration(), period()), class = "vctrs_error_incompatible_type")

    expect_error(vec_ptype2(period(), interval()), class = "vctrs_error_incompatible_type")
    expect_error(vec_ptype2(interval(), period()), class = "vctrs_error_incompatible_type")

    expect_error(vec_ptype2(duration(), interval()), class = "vctrs_error_incompatible_type")
    expect_error(vec_ptype2(interval(), duration()), class = "vctrs_error_incompatible_type")
  })
})

test_that("can't cast between Period/Duration/Interval", {
  verify_errors({
    expect_error(vec_cast(period(), duration()), class = "vctrs_error_incompatible_type")
    expect_error(vec_cast(duration(), period()), class = "vctrs_error_incompatible_type")

    expect_error(vec_cast(period(), interval()), class = "vctrs_error_incompatible_type")
    expect_error(vec_cast(interval(), period()), class = "vctrs_error_incompatible_type")

    expect_error(vec_cast(duration(), interval()), class = "vctrs_error_incompatible_type")
    expect_error(vec_cast(interval(), duration()), class = "vctrs_error_incompatible_type")
  })
})

# ------------------------------------------------------------------------------
# Period - proxy / restore

test_that("proxy is a data frame", {
  x <- period(year = 1:2, day = 3:4)

  expect <- list(
    year = x@year, month = x@month, day = x@day,
    hour = x@hour, minute = x@minute, second = x@.Data
  )

  expect <- new_data_frame(expect)

  expect_identical(vec_proxy(x), expect)
})

test_that("proxy can optionally store vector names in the last column (allowing duplicates)", {
  skip_if_cant_set_s4_names()

  x <- stats::setNames(days(1:3), c("x", "y", "x"))

  proxy <- vec_proxy(x)

  expect_identical(proxy$rcrd_names, names(x))
  expect_identical(match("rcrd_names", names(proxy)), ncol(proxy))
})

test_that("comparison / equality proxies don't have the names column", {
  skip_if_cant_set_s4_names()

  x <- stats::setNames(days(1:3), c("x", "y", "x"))

  expect_null(vec_proxy_compare(x)$rcrd_names)
  expect_null(vec_proxy_equal(x)$rcrd_names)
})

test_that("restore method works", {
  x <- period(year = 1:2, day = 3:4)
  expect_identical(vec_restore(vec_proxy(x), x), x)
})

test_that("restore method retains names", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(days(1), "x")
  expect_named(vec_restore(vec_proxy(x), x), "x")
})

# ------------------------------------------------------------------------------
# Period - ptype2

test_that("Period default ptype2 method falls through to `vec_default_ptype2()`", {
  verify_errors({
    expect_error(vec_ptype2(period(), 1), class = "vctrs_error_incompatible_type")
    expect_error(vec_ptype2(1, period()), class = "vctrs_error_incompatible_type")
  })
})

test_that("common type of Period and Period exists", {
  expect_identical(vec_ptype2(period(), period()), period())
})

test_that("common type of Period and NULL exists", {
  expect_identical(vec_ptype2(period(), NULL), period())
  expect_identical(vec_ptype2(NULL, period()), period())
})

test_that("common type of Period and unspecified exists", {
  expect_identical(vec_ptype2(period(), NA), period())
  expect_identical(vec_ptype2(NA, period()), period())

  expect_identical(vec_ptype2(period(), vctrs::unspecified()), period())
  expect_identical(vec_ptype2(vctrs::unspecified(), period()), period())
})

# ------------------------------------------------------------------------------
# Period - cast

test_that("Period default cast method falls through to `vec_default_cast()`", {
  verify_errors({
    expect_error(vec_cast(period(), 1), class = "vctrs_error_incompatible_type")
    expect_error(vec_cast(1, period()), class = "vctrs_error_incompatible_type")
  })
})

test_that("Period can be cast to Period", {
  expect_identical(vec_cast(days(1), months(1)), days(1))
})

test_that("can cast around `NULL`", {
  expect_identical(vec_cast(NULL, period()), NULL)
  expect_identical(vec_cast(period(), NULL), period())
})

test_that("can cast unspecified to Period", {
  expect_identical(vec_cast(NA, period()), period()[NA_real_])
  expect_error(vec_cast(period(), NA), class = "vctrs_error_incompatible_type")
})

# ------------------------------------------------------------------------------
# Period - vctrs functionality

test_that("can slice Period objects", {
  expect_identical(vec_slice(days(3:4), 2:1), days(4:3))
})

test_that("slicing preserves names", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(days(1:2), c("x", "y"))
  expect_named(vec_slice(x, c(1, 1, 2)), c("x", "x", "y"))
})

test_that("can combine Period objects", {
  expect_identical(vec_c(days(1), days(2)), days(1:2))
})

test_that("can row bind Period objects", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(days(1), "x")
  expect_identical(vec_rbind(x, x), data.frame(x = c(x, x)))
})

test_that("can row bind data frames with Period objects", {
  expect_identical(
    vec_rbind(data.frame(x = days(1)), data.frame(x = days(1))),
    data.frame(x = days(c(1, 1)))
  )
})

test_that("can column bind Period objects", {
  expect_identical(
    vec_cbind(x = days(1), y = days(1:2)),
    data.frame(x = days(c(1, 1)), y = days(1:2))
  )
})

test_that("can column bind data frames with Period objects", {
  expect_identical(
    vec_cbind(data.frame(x = days(1)), data.frame(y = days(1:2))),
    data.frame(x = days(c(1, 1)), y = days(1:2))
  )
})

test_that("Period objects can be ordered", {
  expect_identical(vec_order(vec_c(years(1), days(1))), c(2L, 1L))
  expect_identical(vec_order(vec_c(days(2), days(1))), c(2L, 1L))
})

# ------------------------------------------------------------------------------
# Duration - proxy / restore

test_that("proxy is the underlying number of seconds", {
  x <- ddays(1:2)
  expect_identical(vec_proxy(x), x@.Data)
})

test_that("proxy stores the names", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(ddays(1:3), c("x", "y", "x"))
  expect_named(vec_proxy(x), c("x", "y", "x"))
})

test_that("comparison / equality proxies don't store names", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(ddays(1:3), c("x", "y", "x"))
  expect_named(vec_proxy_compare(x), NULL)
  expect_named(vec_proxy_equal(x), NULL)
})

test_that("restore method works", {
  x <- ddays(1:2)
  expect_identical(vec_restore(vec_proxy(x), x), x)
})

test_that("restore method retains names", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(ddays(1), "x")
  expect_named(vec_restore(vec_proxy(x), x), "x")
})

# ------------------------------------------------------------------------------
# Duration - ptype2

test_that("Duration default ptype2 method falls through to `vec_default_ptype2()`", {
  verify_errors({
    expect_error(vec_ptype2(duration(), 1), class = "vctrs_error_incompatible_type")
    expect_error(vec_ptype2(1, duration()), class = "vctrs_error_incompatible_type")
  })
})

test_that("common type of Duration and Duration exists", {
  expect_identical(vec_ptype2(duration(), duration()), duration())
})

test_that("common type of Duration and NULL exists", {
  expect_identical(vec_ptype2(duration(), NULL), duration())
  expect_identical(vec_ptype2(NULL, duration()), duration())
})

test_that("common type of Duration and unspecified exists", {
  expect_identical(vec_ptype2(duration(), NA), duration())
  expect_identical(vec_ptype2(NA, duration()), duration())

  expect_identical(vec_ptype2(duration(), unspecified()), duration())
  expect_identical(vec_ptype2(unspecified(), duration()), duration())
})

test_that("common type of Duration and difftime is Duration", {
  expect_identical(vec_ptype2(duration(), new_duration()), duration())
  expect_identical(vec_ptype2(new_duration(), duration()), duration())
})

# ------------------------------------------------------------------------------
# Duration - cast

test_that("Duration default cast method falls through to `vec_default_cast()`", {
  verify_errors({
    expect_error(vec_cast(duration(), 1), class = "vctrs_error_incompatible_type")
    expect_error(vec_cast(1, duration()), class = "vctrs_error_incompatible_type")
  })
})

test_that("Duration can be cast to Duration", {
  expect_identical(vec_cast(ddays(1), dmonths(1)), ddays(1))
})

test_that("can cast around `NULL`", {
  expect_identical(vec_cast(NULL, duration()), NULL)
  expect_identical(vec_cast(duration(), NULL), duration())
})

test_that("can cast unspecified to Duration", {
  expect_identical(vec_cast(NA, duration()), duration()[NA_real_])
  expect_error(vec_cast(duration(), NA), class = "vctrs_error_incompatible_type")
})

test_that("Duration can be cast to and from difftime", {
  expect_identical(vec_cast(duration(), new_duration()), new_duration())
  expect_identical(vec_cast(new_duration(), duration()), duration())
})

# ------------------------------------------------------------------------------
# Duration - vctrs functionality

test_that("can slice Duration objects", {
  expect_identical(vec_slice(ddays(3:4), 2:1), ddays(4:3))
})

test_that("slicing preserves names", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(ddays(1:2), c("x", "y"))
  expect_named(vec_slice(x, c(1, 1, 2)), c("x", "x", "y"))
})

test_that("can combine Duration objects", {
  expect_identical(vec_c(ddays(1), ddays(2)), ddays(1:2))
})

test_that("can row bind Duration objects", {
  skip_if_cant_set_s4_names()
  x <- ddays(1)
  x_named <- stats::setNames(x, "x")
  expect_identical(vec_rbind(x_named, x_named), data.frame(x = c(x, x)))
})

test_that("can row bind data frames with Duration objects", {
  expect_identical(
    vec_rbind(data.frame(x = ddays(1)), data.frame(x = ddays(1))),
    data.frame(x = ddays(c(1, 1)))
  )
})

test_that("can column bind Duration objects", {
  expect_identical(
    vec_cbind(x = ddays(1), y = ddays(1:2)),
    data.frame(x = ddays(c(1, 1)), y = ddays(1:2))
  )
})

test_that("can column bind data frames with Duration objects", {
  expect_identical(
    vec_cbind(data.frame(x = ddays(1)), data.frame(y = ddays(1:2))),
    data.frame(x = ddays(c(1, 1)), y = ddays(1:2))
  )
})

# ------------------------------------------------------------------------------
# Interval - proxy / restore

test_that("proxy is a data frame", {
  x <- interval(tzone = "UTC")

  expect <- list(start = POSIXct(tz = "UTC"), span = numeric())
  expect <- new_data_frame(expect)

  expect_identical(vec_proxy(x), expect)
})

test_that("proxy can optionally store vector names in the last column (allowing duplicates)", {
  skip_if_cant_set_s4_names()

  x <- c("2019-01-01", "2019-01-02", "2019-01-03")
  x <- stats::setNames(interval(x), c("x", "y", "x"))

  proxy <- vec_proxy(x)

  expect_identical(proxy$rcrd_names, names(x))
  expect_identical(match("rcrd_names", names(proxy)), ncol(proxy))
})

test_that("comparison / equality proxies don't have the names column", {
  skip_if_cant_set_s4_names()

  x <- c("2019-01-01", "2019-01-02", "2019-01-03")
  x <- stats::setNames(interval(x), c("x", "y", "x"))

  expect_null(vec_proxy_compare(x)$rcrd_names)
  expect_null(vec_proxy_equal(x)$rcrd_names)
})

test_that("restore method works", {
  x <- interval(c("2019-01-01", "2019-01-02"), c("2020-01-01", "2020-01-02"))
  expect_identical(vec_restore(vec_proxy(x), x), x)
})

test_that("restore method retains names", {
  skip_if_cant_set_s4_names()
  x <- stats::setNames(interval("2019-01-01"), "x")
  expect_named(vec_restore(vec_proxy(x), x), "x")
})

# ------------------------------------------------------------------------------
# Interval - ptype2

test_that("Interval default ptype2 method falls through to `vec_default_ptype2()`", {
  verify_errors({
    expect_error(vec_ptype2(interval(), 1), class = "vctrs_error_incompatible_type")
    expect_error(vec_ptype2(1, interval()), class = "vctrs_error_incompatible_type")
  })
})

test_that("common type of Interval and Interval exists", {
  expect_identical(vec_ptype2(interval(), interval()), interval())

  x <- interval(tzone = "America/Los_Angeles")
  expect_identical(vec_ptype2(x, x), x)
})

test_that("common tzone uses non-local tzone", {
  x <- interval(tzone = "")
  y <- interval(tzone = "America/Los_Angeles")

  expect_identical(vec_ptype2(x, y)@tzone, "America/Los_Angeles")
  expect_identical(vec_ptype2(y, x)@tzone, "America/Los_Angeles")

  expect_identical(tz(int_start(vec_ptype2(x, y))), "America/Los_Angeles")
  expect_identical(tz(int_start(vec_ptype2(y, x))), "America/Los_Angeles")
})

test_that("common tzone is order dependent", {
  x <- interval(tzone = "America/New_York")
  y <- interval(tzone = "America/Los_Angeles")

  expect_identical(vec_ptype2(x, y)@tzone, "America/New_York")
  expect_identical(vec_ptype2(y, x)@tzone, "America/Los_Angeles")

  expect_identical(tz(int_start(vec_ptype2(x, y))), "America/New_York")
  expect_identical(tz(int_start(vec_ptype2(y, x))), "America/Los_Angeles")
})

test_that("common type of Interval and NULL exists", {
  expect_identical(vec_ptype2(interval(), NULL), interval())
  expect_identical(vec_ptype2(NULL, interval()), interval())
})

test_that("common type of Interval and unspecified exists", {
  expect_identical(vec_ptype2(interval(), NA), interval())
  expect_identical(vec_ptype2(NA, interval()), interval())

  expect_identical(vec_ptype2(interval(), unspecified()), interval())
  expect_identical(vec_ptype2(unspecified(), interval()), interval())
})

# ------------------------------------------------------------------------------
# Interval - cast

test_that("Interval default cast method falls through to `vec_default_cast()`", {
  verify_errors({
    expect_error(vec_cast(interval(), 1), class = "vctrs_error_incompatible_type")
    expect_error(vec_cast(1, interval()), class = "vctrs_error_incompatible_type")
  })
})

test_that("Interval can be cast to Interval", {
  expect_identical(vec_cast(interval(), interval()), interval())
})

test_that("can cast to a different tzone", {
  x_tzone <- "America/Los_Angeles"
  x_start <- as.POSIXct("1970-01-01", tz = x_tzone)
  x_end <- as.POSIXct("1970-01-02", tz = x_tzone)
  x <- interval(x_start, x_end, x_tzone)

  to_tzone <- "America/New_York"
  to <- interval(tzone = to_tzone)

  expect_start <- with_tz(x_start, to_tzone)
  expect_end <- with_tz(x_end, to_tzone)
  expect <- interval(expect_start, expect_end, to_tzone)

  expect_identical(vec_cast(x, to), expect)
})

test_that("can cast around `NULL`", {
  expect_identical(vec_cast(NULL, interval()), NULL)
  expect_identical(vec_cast(interval(), NULL), interval())
})

test_that("can cast unspecified to Interval", {
  expect_identical(vec_cast(NA, interval()), interval()[NA_real_])
  expect_error(vec_cast(interval(), NA), class = "vctrs_error_incompatible_type")
})

# ------------------------------------------------------------------------------
# Interval - vctrs functionality

test_that("can slice Interval objects", {
  x <- interval(c("1970-01-01", "1970-01-02"))
  expect_identical(vec_slice(x, 2:1), x[2:1])
})

test_that("slicing preserves names", {
  skip_if_cant_set_s4_names()
  x <- interval(c("1970-01-01", "1970-01-02"))
  x <- stats::setNames(x, c("x", "y"))
  expect_named(vec_slice(x, c(1, 1, 2)), c("x", "x", "y"))
})

test_that("can combine Interval objects", {
  x <- interval("1970-01-01")
  y <- interval("1970-01-02")
  expect <- interval(c("1970-01-01", "1970-01-02"))
  expect_identical(vec_c(x, y), expect)
})

test_that("can row bind Interval objects", {
  skip_if_cant_set_s4_names()
  x <- interval("1970-01-01")
  x_named <- stats::setNames(x, "x")
  expect_identical(vec_rbind(x_named, x_named), data.frame(x = c(x, x)))
})

test_that("can row bind data frames with Interval objects", {
  x <- interval("1970-01-01")

  expect_identical(
    vec_rbind(data.frame(x = x), data.frame(x = x)),
    data.frame(x = vec_c(x, x))
  )
})

test_that("can column bind Interval objects", {
  x <- interval("1970-01-01")
  y <- interval(c("1970-01-01", "1970-01-02"))

  expect_identical(
    vec_cbind(x = x, y = y),
    data.frame(x = vec_c(x, x), y = y)
  )
})

test_that("can column bind data frames with Interval objects", {
  x <- interval("1970-01-01")
  y <- interval(c("1970-01-01", "1970-01-02"))

  expect_identical(
    vec_cbind(data.frame(x = x), data.frame(y = y)),
    data.frame(x = vec_c(x, x), y = y)
  )
})

test_that("Interval objects can be ordered", {
  x <- interval("1970-01-01", "1970-01-02")
  y <- interval("1970-01-02", "1970-01-03")
  z <- interval("1970-01-02", "1970-01-04")

  # Different from `order()`!
  expect_identical(vec_order(vec_c(y, x)), c(2L, 1L))

  expect_identical(vec_order(vec_c(z, y)), c(2L, 1L))
})

# ------------------------------------------------------------------------------
# Output

test_that("vctrs methods have informative errors", {
  verify_output(test_path("output", "test-vctrs.txt"), {
    "# no common type when mixing Period/Duration/Interval"
    vec_ptype2(period(), duration())
    vec_ptype2(duration(), period())

    vec_ptype2(period(), interval())
    vec_ptype2(interval(), period())

    vec_ptype2(duration(), interval())
    vec_ptype2(interval(), duration())

    "# can't cast between Period/Duration/Interval"
    vec_cast(period(), duration())
    vec_cast(duration(), period())

    vec_cast(period(), interval())
    vec_cast(interval(), period())

    vec_cast(duration(), interval())
    vec_cast(interval(), duration())

    "# Period default ptype2 method falls through to `vec_default_ptype2()`"
    vec_ptype2(period(), 1)
    vec_ptype2(1, period())

    "# Period default cast method falls through to `vec_default_cast()`"
    vec_cast(period(), 1)
    vec_cast(1, period())

    "# Duration default ptype2 method falls through to `vec_default_ptype2()`"
    vec_ptype2(duration(), 1)
    vec_ptype2(1, duration())

    "# Duration default cast method falls through to `vec_default_cast()`"
    vec_cast(duration(), 1)
    vec_cast(1, duration())

    "# Interval default ptype2 method falls through to `vec_default_ptype2()`"
    vec_ptype2(interval(), 1)
    vec_ptype2(1, interval())

    "# Interval default cast method falls through to `vec_default_cast()`"
    vec_cast(interval(), 1)
    vec_cast(1, interval())
  })
})

context("Timespans")

test_that("is.timespan works as expected", {
  expect_false(is.timespan(234))
  expect_false(is.timespan(as.POSIXct("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.timespan(as.POSIXlt("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.timespan(Sys.Date()))
  expect_true(is.timespan(minutes(1)))
  expect_true(is.timespan(dminutes(1)))
  expect_true(is.timespan(interval(
    as.POSIXct("2008-08-03 13:01:59", tz = "UTC"),
    as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  )))
})

test_that("time_length works as expected", {
  expect_that(
    time_length(period(1, "day")), # period
    equals(86400)
  )
  expect_that(
    time_length(ymd("2014-12-15") - ymd("2014-11-30"), "day"), # difftime
    equals(15)
  )
  expect_that(
    time_length(ymd("2014-12-15") - ymd("2014-11-30"), "week"),
    equals(15 / 7)
  )
  expect_that(
    time_length(duration(3, "months"), "month"), # duration
    equals(3)
  )
  expect_that(
    time_length(interval(ymd("2014-11-30"), ymd("2014-12-15")), "weeks"), # interval
    equals(15 / 7)
  )
  expect_that(
    time_length(interval(ymd("1900-01-01"), ymd("1999-12-31")), "years"),
    is_less_than(100)
  )
  expect_that(
    time_length(as.duration(interval(ymd("1900-01-01"), ymd("1999-12-31"))), "years"),
    is_less_than(100)
  )
  expect_that(
    -time_length(interval(ymd("1900-01-01"), ymd("2000-01-01")), "days"),
    equals(time_length(int_flip(interval(ymd("1900-01-01"), ymd("2000-01-01"))), "days"))
  )
})

test_that("time_length works with intervals on DTS", {
  times <- ymd_hms(c("2019-03-29 02:03:00", "2019-03-29 02:10:00"), tz = "Europe/Paris")
  expect_equal(time_length(interval(times[[1]], times[[2]]), "years"), 1.32818e-05)
})

test_that("time_length works with missing intervals", {
  expect_equal(time_length(interval(NA, ymd("2016-01-01")), unit = "year"), NA_real_)
})

test_that("time_length works with birth date 29 Feb ", {
  expect_that(
    round(time_length(interval(ymd("1992-02-29"), ymd("1999-02-28")), "years"), 4),
    equals(6.9973)
  )
  expect_that(
    round(time_length(interval(ymd("1992-02-29"), ymd("1999-03-31")), "years"), 4),
    equals(7.0822)
  )
  expect_that(
    time_length(interval(ymd("1992-02-29"), ymd("1999-03-01")), "years"),
    equals(7)
  )
  expect_that(
    round(time_length(interval(ymd("1992-02-29"), ymd("2000-03-01")), "years"), 4),
    equals(8.0027)
  )
  expect_that(
    time_length(interval(ymd("1992-02-29"), ymd("2000-02-29")), "years"),
    equals(8)
  )
  expect_that(
    time_length(interval(ymd_hms("1992-02-29 12:00:00"), ymd_hms("1999-03-01 05:00:00")), "years"),
    is_more_than(7)
  )
})

test_that("time_length works with negative interals", {
  expect_that(
    -time_length(interval(ymd("1992-02-28"), ymd("2000-01-01")), "days"),
    equals(time_length(int_flip(interval(ymd("1992-02-28"), ymd("2000-01-01"))), "days"))
  )
  expect_that(
    -time_length(interval(ymd("1992-02-28"), ymd("2000-03-01")), "days"),
    equals(time_length(int_flip(interval(ymd("1992-02-28"), ymd("2000-03-01"))), "days"))
  )

  ## If both ends include leap years Febs, the lenths are not identical
  int <- interval(ymd("1992-02-28"), ymd("2000-03-01"))
  expect_equal(-time_length(int, "years"), time_length(int_flip(int), "years"))

  ## or if both ends don't include leap years Febs, the lenths are identical
  int <- interval(ymd("1994-02-28"), ymd("2002-03-01"))
  expect_equal(-time_length(int, "years"), time_length(int_flip(int), "years"))

  ## ... otherwise not
  int <- interval(ymd("1992-02-28"), ymd("2002-01-01"))
  expect_gt(-time_length(int, "years"), time_length(int_flip(int), "years"))
})

test_that("time_length handles vectors", {
  expect_that(
    time_length(days(1:3), unit = "days"),
    equals(1:3)
  )
  expect_that(
    time_length(as.interval(days(1:3), start = today()), unit = "days"),
    equals(1:3)
  )
  expect_that(
    time_length(as.interval(days(c(1:3, NA)), start = today()), unit = "days"),
    equals(c(1:3, NA))
  )

  ints <- new("Interval",
    .Data = c(
      6363040, 1646253613, 159312603, -510371200, -381983229, 1369742727,
      -190254638, -1118394043, -2091814107
    ),
    start = structure(c(
      64156663, 326788049, 462064351, 909110414, 778538359,
      395846311, 1415640365, 1697694335, 2193040469
    ),
    class = c("POSIXct", "POSIXt"), tzone = "UTC"
    ),
    tzone = "UTC"
  )

  expect_equal(
    round(time_length(ints, "years"), 5),
    c(
      0.20122, 52.16674, 5.04903, -16.17281, -12.10439, 43.40412,
      -6.03011, -35.44092, -66.28715
    )
  )

  expect_equal(
    round(time_length(ints[ints > 0], "years"), 5),
    c(0.20122, 52.16674, 5.04903, 43.40412)
  )

  expect_equal(
    round(time_length(ints[ints < 0], "years"), 5),
    c(-16.17281, -12.10439, -6.03011, -35.44092, -66.28715)
  )

  expect_equal(
    round(time_length(ints, "month"), 5),
    c(
      2.4402, 625.99552, 60.57729, -194.06691, -145.2291, 520.81679,
      -72.35552, -425.27918, -795.43557
    )
  )
})

test_that("time_length handles 0 length intervals", {
  date <- ymd_hms("2014-12-30 20:35:13")
  int <- interval(c(date, date), c(date, date))
  expect_equal(time_length(int), c(0, 0))
  expect_equal(time_length(int, "years"), c(0, 0))
  expect_equal(time_length(int, "months"), c(0, 0))
})

#' @include periods.r
NULL

#' Get/set months component of a date-time
#'
#' Date-time must be a POSIXct, POSIXlt, Date, Period, chron, yearmon, yearqtr, zoo,
#' zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
#'
#' @param x a date-time object
#' @param label logical. TRUE will display the month as a character string such
#'   as "January." FALSE will display the month as a number.
#' @param abbr logical. FALSE will display the month as a character string
#'   label, such as "January". TRUE will display an abbreviated version of the
#'  label, such as "Jan". abbr is disregarded if label = FALSE.
#' @param value a numeric object
#' @param locale for month, locale to use for month names. Default to current locale.
#' @return If `label = FALSE`: month as number (1-12, 1 = January, 12 = December),
#'   otherwise as an ordered factor.
#' @keywords utilities manip chron methods
#' @examples
#' x <- ymd("2012-03-26")
#' month(x)
#' month(x) <- 1
#' month(x) <- 13
#' month(x) > 3
#'
#' month(ymd(080101))
#' month(ymd(080101), label = TRUE)
#' month(ymd(080101), label = TRUE, abbr = FALSE)
#' month(ymd(080101) + months(0:11), label = TRUE)
#' @export
month <- function(x, label = FALSE, abbr = TRUE, locale = Sys.getlocale("LC_TIME")) {
  UseMethod("month")
}

#' @export
month.default <- function(x, label = FALSE, abbr = TRUE, locale = Sys.getlocale("LC_TIME")) {
  month(as.POSIXlt(x, tz = tz(x))$mon + 1, label, abbr, locale = locale)
}

#' @export
month.numeric <- function(x, label = FALSE, abbr = TRUE, locale = Sys.getlocale("LC_TIME")) {
  if (!all(x[!is.na(x)] %in% 1:12)) {
    stop("Values are not in 1:12")
  }

  if (!label) {
    return(x)
  }

  names <- .get_locale_regs(locale)$month_names
  labels <- if (abbr) names$abr else names$full

  ordered(x, levels = 1:12, labels = labels)
}

#' @export
month.Period <- function(x, label = FALSE, abbr = TRUE, locale = Sys.getlocale("LC_TIME")) {
  slot(x, "month")
}

as_month <- function(value) {
  ## FIXME: use same technique as in as_week_start to localize this
  if (is.character(value)) {
    value <- pmatch(
      tolower(value),
      c(
        "january", "february", "march",
        "june", "july", "august", "september",
        "october", "november", "december"
      )
    )
  }
  value
}

#' @rdname month
#' @export
setGeneric("month<-",
  function (x, value) standardGeneric("month<-"),
  useAsDefault = function(x, value) {
    y <- update_datetime(as.POSIXct(x), months = value, roll_month = "NAym")
    reclass_date(y, x)
  }
)

#' @export
setMethod("month<-", "Duration", function(x, value) {
  x <- x + months(as_month(value) - month(x))
})

#' @export
setMethod("month<-", signature("Period"), function(x, value) {
  slot(x, "month") <- as_month(value)
  x
})

#' @export
setMethod("month<-", "Interval", function(x, value) {
  x <- x + months(as_month(value) - month(x))
})

#' @export
setMethod("month<-", "POSIXt", function(x, value) {
  update_datetime(x, months = value, roll_month = "NAym")
})

#' @export
setMethod("month<-", "Date", function(x, value) {
  update_datetime(x, months = value, roll_month = "NAym")
})

#' Get the number of days in the month of a date-time
#'
#' Date-time must be a POSIXct, POSIXlt, Date, chron, yearmon, yearqtr,
#' zoo, zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
#'
#' @export
#' @param x a date-time object
#' @return An integer of the number of days in the month component of the date-time object.
days_in_month <- function(x) {
  month_x <- month(x, label = TRUE, locale = "C")
  n_days <- N_DAYS_IN_MONTHS[month_x]
  n_days[month_x == "Feb" & leap_year(x)] <- 29L
  n_days
}

## fixme: integrate with above, this oen is needed internally
.days_in_month <- function(m, y) {
  n_days <- N_DAYS_IN_MONTHS[m]
  n_days[m == 2L & leap_year(y)] <- 29L
  n_days
}

## tothink: export?
days_in_months_so_far <- function(month, leap) {
  ## if month is negative, compute from the end of the year
  cum_days_pos <- c(0, cumsum(N_DAYS_IN_MONTHS)[-12])
  cum_days_neg <- c(0, cumsum(rev(N_DAYS_IN_MONTHS))[-12])
  negative <- month < 0
  positive <- month > 0
  sofar <- integer(length(month))
  sofar[negative] <- cum_days_neg[-month[negative]]
  sofar[positive] <- cum_days_pos[month[positive]]
  adjust <- leap & ((negative & month == -12) | (positive & month > 2))
  sofar[adjust] <- sofar[adjust] + 1L
  sofar
}
## days_in_months_so_far(c(1, 2, 3, -10, -11, -12), rep.int(T, 6))
## [1]   0  31  60 275 306 335
## days_in_months_so_far(c(1, 2, 3, -10, -11, -12), rep.int(F, 6))
## [1]   0  31  59 275 306 334

#' Is x a date-time object?
#'
#' An instant is a specific moment in time. Most common date-time
#' objects (e.g, POSIXct, POSIXlt, and Date objects) are instants.
#'
#' @aliases instant instants
#' @export
#' @param x an R object
#' @return TRUE if x is a POSIXct, POSIXlt, or Date object, FALSE otherwise.
#' @seealso [is.timespan()], [is.POSIXt()], [is.Date()]
#' @keywords logic chron
#' @examples
#' is.instant(as.Date("2009-08-03")) # TRUE
#' is.timepoint(5) # FALSE
is.instant <- function(x) inherits(x, c("POSIXt", "POSIXct", "POSIXlt", "Date"))

#' @export
#' @rdname is.instant
is.timepoint <- is.instant

#' The current day and time
#'
#' @param tzone a character vector specifying which time zone you would like the
#'   current time in. tzone defaults to your computer's system timezone. You can
#'   retrieve the current time in the Universal Coordinated Time (UTC) with
#'   now("UTC").
#' @return `now` - the current datetime as a `POSIXct` object
#'
#' @keywords chron utilities
#' @examples
#' now()
#' now("GMT")
#' now("")
#' now() == now() # would be TRUE if computer processed both at the same instant
#' now() < now() # TRUE
#' now() > now() # FALSE
#' @export
now <- function(tzone = "") {
  with_tz(Sys.time(), tzone)
}


#' @rdname now
#' @examples
#' today()
#' today("GMT")
#' today() == today("GMT") # not always true
#' today() < as.Date("2999-01-01") # TRUE  (so far)
#' @export
today <- function(tzone = "") {
  as_date(now(tzone))
}

#' 1970-01-01 UTC
#'
#' Origin is the date-time for 1970-01-01 UTC in POSIXct format. This date-time
#' is the origin for the numbering system used by POSIXct, POSIXlt, chron, and
#' Date classes.
#'
#' @keywords data chron
#' @examples
#' origin
#' @export origin
origin <- structure(0, class = c("POSIXct", "POSIXt"), tzone = "UTC")

.rep_maybe <- function(x, N) {
  if (N > 1 && length(x) > 1 && length(x) != N) {
    out <- rep_len(x, N)
    ## repl_len doesn't preserve attributes
    if (is.POSIXct(x)) {
      attributes(out) <- attributes(x)
    }
    out
  } else {
    x
  }
}

##' Efficient creation of date-times from numeric representations
##'
##' `make_datetime()` is a very fast drop-in replacement for
##' [base::ISOdate()] and [base::ISOdatetime()]. `make_date()` produces
##' objects of class `Date`.
##'
##' Input vectors are silently recycled. All inputs except `sec` are
##' silently converted to integer vectors; `sec` can be either integer or
##' double.
##'
##' @param year numeric year
##' @param month numeric month
##' @param day numeric day
##' @param hour numeric hour
##' @param min numeric minute
##' @param sec numeric second
##' @param tz time zone. Defaults to UTC.
##' @export
##' @examples
##' make_datetime(year = 1999, month = 12, day = 22, sec = 10)
##' make_datetime(year = 1999, month = 12, day = 22, sec = c(10, 11))
make_datetime <- function(year = 1970L, month = 1L, day = 1L, hour = 0L, min = 0L, sec = 0, tz = "UTC") {
  lengths <- vapply(list(year, month, day, hour, min, sec), length, 1, USE.NAMES = FALSE)
  if (min(lengths) == 0L) {
    .POSIXct(numeric(), tz = tz)
  } else {
    N <- max(lengths)
    timechange::time_update(
      time = .rep_maybe(origin, N),
      year = .rep_maybe(year, N),
      month = .rep_maybe(month, N),
      mday = .rep_maybe(day, N),
      hour = .rep_maybe(hour, N),
      minute = .rep_maybe(min, N),
      second = .rep_maybe(sec, N),
      tz = tz,
      roll_month = "full"
    )
  }
}

##' @rdname make_datetime
##' @export
make_date <- function(year = 1970L, month = 1L, day = 1L) {
  lengths <- vapply(list(year, month, day), length, 1, USE.NAMES = FALSE)
  if (min(lengths) == 0L) {
    as.Date(integer(), origin = origin)
  } else {
    N <- max(lengths)
    secs <- .Call(
      C_make_d,
      rep_len(as.integer(year), N),
      rep_len(as.integer(month), N),
      rep_len(as.integer(day), N)
    )
    structure(secs / 86400L, class = "Date")
  }
}

test_that("is.instant/is.timepoint works as expected", {
  expect_false(is.instant(234))
  expect_true(is.instant(as.POSIXct("2008-08-03 13:01:59", tz = "UTC")))
  expect_true(is.instant(as.POSIXlt("2008-08-03 13:01:59", tz = "UTC")))
  expect_true(is.instant(Sys.Date()))
  expect_false(is.instant(minutes(1)))
  expect_true(is.timespan(interval(
    as.POSIXct("2008-08-03 13:01:59", tz = "UTC"),
    as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  )))
})

test_that("now() handles time zone input correctly", {
  nt <- now("UTC")
  st <- Sys.time()
  expect_identical(
    floor_date(nt, "minute"),
    floor_date(as.POSIXct(format(as.POSIXct(st), tz = "UTC"), tz = "UTC"), "minute")
  )
})

## test_that("today() works correctly", {
##   ## Sys.Date is not correct
##   ## expect_identical(today(), Sys.Date())
## })


test_that("make_datetime returns same values as ISOdatetime", {
  set.seed(1000)
  N <- 1e4
  y <- as.integer(runif(N, 1800, 2200))
  m <- as.integer(runif(N, 1, 12))
  d <- as.integer(runif(N, 1, 28))
  H <- as.integer(runif(N, 0, 23))
  M <- as.integer(runif(N, 0, 59))
  S <- runif(N, 0, 59)
  out1 <- ISOdatetime(y, m, d, H, M, S, tz = "UTC")
  out2 <- make_datetime(y, m, d, H, M, S)
  expect_equal(out1, out2)
  S <- as.integer(runif(N, 0, 59))
  out1 <- ISOdatetime(y, m, d, H, M, S, tz = "UTC")
  out2 <- make_datetime(y, m, d, H, M, S)
  expect_equal(out1, out2)
  out3 <- make_date(y, m, d)
  expect_equal(as.Date(out1), out3)
})

test_that("make_datetime replicates as expected", {
  expect_equal(
    make_datetime(year = 1999, month = c(11, 12), day = 22, sec = c(10, 11)),
    as.POSIXct(c("1999-11-22 00:00:10 UTC", "1999-12-22 00:00:11 UTC"), tz = "UTC")
  )
  expect_equal(
    make_datetime(year = c(1999, 2000, 3000), month = c(11, 12), day = 22, sec = c(10, 11, 13, 13)),
    ymd_hms(c("1999-11-22 00:00:10", "2000-12-22 00:00:11", "3000-11-22 00:00:13", "1999-12-22 00:00:13"), tz = "UTC")
  )
  expect_equal(
    make_datetime(
      year = c(1999, 2000, 3000), month = c(11, 12), day = 22, sec = c(10, 11, 13, 13),
      tz = "America/New_York"
    ),
    ymd_hms(c("1999-11-22 00:00:10", "2000-12-22 00:00:11", "3000-11-22 00:00:13", "1999-12-22 00:00:13"),
      tz = "America/New_York"
    )
  )
})

test_that("make_datetime propagates NAs as expected", {
  expect_equal(
    make_datetime(year = 1999, month = c(11, NA), day = 22, sec = c(10, 11, NA)),
    as.POSIXct(c("1999-11-22 00:00:10 UTC", NA, NA), tz = "UTC")
  )
})


test_that("make_datetime rolls over month boundary", {
  expect_equal(make_datetime(month = 1, day = 31, hour = 24), ymd("1970-02-01", tz = "UTC"))
  expect_equal(
    make_datetime(month = 1, day = 31, hour = 25, min = 61, sec = 61),
    ymd_hms("1970-02-01 02:02:01", tz = "UTC"))
})

context("Instants")

test_that("is.instant/is.timepoint works as expected", {
  expect_false(is.instant(234))
  expect_true(is.instant(as.POSIXct("2008-08-03 13:01:59", tz = "UTC")))
  expect_true(is.instant(as.POSIXlt("2008-08-03 13:01:59", tz = "UTC")))
  expect_true(is.instant(Sys.Date()))
  expect_false(is.instant(minutes(1)))
  expect_true(is.timespan(interval(
    as.POSIXct("2008-08-03 13:01:59", tz = "UTC"),
    as.POSIXct("2009-08-03 13:01:59", tz = "UTC"))))
})

test_that("now() handles time zone input correctly", {
  nt <- now("UTC")
  st <- Sys.time()
  expect_identical(
    floor_date(nt, "minute"),
    floor_date(as.POSIXct(format(as.POSIXct(st), tz = "UTC"), tz = "UTC"), "minute")
  )
})

## test_that("today() works correctly", {
##   ## Sys.Date is not correct
##   ## expect_identical(today(), Sys.Date())
## })


test_that("make_datetime returns same values as ISOdatetime", {
  set.seed(1000)
  N <- 1e4
  y <- as.integer(runif(N, 1800, 2200))
  m <- as.integer(runif(N, 1, 12))
  d <- as.integer(runif(N, 1, 28))
  H <- as.integer(runif(N, 0, 23))
  M <- as.integer(runif(N, 0, 59))
  S <- runif(N, 0, 59)
  out1 <- ISOdatetime(y, m, d, H, M, S, tz = "UTC")
  out2 <- make_datetime(y, m, d, H, M, S)
  expect_equal(out1, out2)
  S <- as.integer(runif(N, 0, 59))
  out1 <- ISOdatetime(y, m, d, H, M, S, tz = "UTC")
  out2 <- make_datetime(y, m, d, H, M, S)
  expect_equal(out1, out2)
  out3 <- make_date(y, m, d)
  expect_equal(as.Date(out1), out3)
})

test_that("make_datetime replicates as expected", {
  expect_equal(make_datetime(year = 1999, month = c(11, 12), day = 22, sec = c(10, 11)),
               as.POSIXct(c("1999-11-22 00:00:10 UTC", "1999-12-22 00:00:11 UTC"), tz = "UTC"))
  expect_equal(make_datetime(year = c(1999, 2000, 3000), month = c(11, 12), day = 22, sec = c(10, 11, 13, 13)),
               ymd_hms(c("1999-11-22 00:00:10", "2000-12-22 00:00:11", "3000-11-22 00:00:13", "1999-12-22 00:00:13"), tz = "UTC"))
  expect_equal(make_datetime(year = c(1999, 2000, 3000), month = c(11, 12), day = 22, sec = c(10, 11, 13, 13),
                             tz = "America/New_York"),
               ymd_hms(c("1999-11-22 00:00:10", "2000-12-22 00:00:11", "3000-11-22 00:00:13", "1999-12-22 00:00:13"),
                       tz = "America/New_York"))
})

test_that("make_datetime propagates NAs as expected", {
  expect_equal(make_datetime(year = 1999, month = c(11, NA), day = 22, sec = c(10, 11, NA)),
               as.POSIXct(c("1999-11-22 00:00:10 UTC", NA, NA), tz = "UTC"))
})

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
NULL

modulo_duration_by_duration <- function(dur, dur2)
  dur - dur %/% dur2 * dur2

modulo_interval_by_duration <- function(int, dur)
  interval(int@start + (int@.Data %/% dur@.Data * dur), int_end(int))

modulo_interval_by_period <- function(int, per)
  interval(add_with_rollback(int@start, (int %/% per) * per), int_end(int))

modulo_period_by_period <- function(per, per2)
  per - per %/% per2 * per2

#' @export
setMethod("%%", signature(e1 = "Duration", e2 = "Duration"),
          function(e1, e2) modulo_duration_by_duration(e1, e2))

#' @export
setMethod("%%", signature(e1 = "Interval", e2 = "Duration"),
          function(e1, e2) modulo_interval_by_duration(e1, e2))

#' @export
setMethod("%%", signature(e1 = "Interval", e2 = "Interval"),
          function(e1, e2) stop("%% not defined for Interval class\nConsider setdiff()"))

#' @export
setMethod("%%", signature(e1 = "Interval", e2 = "Period"),
          function(e1, e2) modulo_interval_by_period(e1, e2))

#' @export
setMethod("%%", signature(e1 = "Period", e2 = "Period"),
          function(e1, e2) modulo_period_by_period(e1, e2))

#' @include periods.r
NULL

#' Get/set days component of a date-time
#'
#' @details `mday()` and `yday()` return the day of the month and day of the year
#'   respectively. `day()` and `day<-()` are aliases for `mday()` and `mday<-()`.
#' @param x a POSIXct, POSIXlt, Date, chron, yearmon, yearqtr, zoo, zooreg, timeDate,
#'   xts, its, ti, jul, timeSeries, or fts object.
#' @param label logical. Only available for wday. TRUE will display the day of the week
#'   as an ordered factor of character strings, such as "Sunday." FALSE will display the
#'   day of the week as a number.
#' @param abbr logical. Only available for wday. FALSE will display the day of the week
#'   as an ordered factor of character strings, such as "Sunday." TRUE will display an
#'   abbreviated version of the label, such as "Sun". abbr is disregarded if label =
#'   FALSE.
#' @param value (for `wday<-`) a numeric or a string giving the name of the day in the
#'   current locale or in English. Can be abbreviated. When a string, the value of
#'   `week_start` is ignored.
#' @param week_start day on which week starts following ISO conventions: 1 means Monday
#'   and 7 means Sunday (default). When `label = FALSE` and `week_start = 7`, the number
#'   returned for Sunday is 1, for Monday is 2, etc. When `label = TRUE`, the returned
#'   value is a factor with the first level being the week start (e.g. Sunday if
#'   `week_start = 7`). You can set `lubridate.week.start` option to control this
#'   parameter globally.
#' @param locale locale to use for day names. Default to current locale.
#' @return `wday()` returns the day of the week as a decimal number or an ordered factor
#'   if label is `TRUE`.
#' @keywords utilities manip chron methods
#' @examples
#' x <- as.Date("2009-09-02")
#' wday(x) # 4
#' wday(x, label = TRUE) # Wed
#'
#' wday(x, week_start = 1) # 3
#' wday(x, week_start = 7) # 4
#'
#' wday(x, label = TRUE, week_start = 7) # Wed (Sun is the first level)
#' wday(x, label = TRUE, week_start = 1) # Wed (Mon is the first level)
#'
#' wday(ymd(080101))
#' wday(ymd(080101), label = TRUE, abbr = FALSE)
#' wday(ymd(080101), label = TRUE, abbr = TRUE)
#' wday(ymd(080101) + days(-2:4), label = TRUE, abbr = TRUE)
#'
#' x <- as.Date("2009-09-02")
#' yday(x) # 245
#' mday(x) # 2
#' yday(x) <- 1 # "2009-01-01"
#' yday(x) <- 366 # "2010-01-01"
#' mday(x) > 3
#' @export day mday
day <- function(x) {
  UseMethod("mday")
}

#' @rdname day
#' @export
mday <- day

#' @rdname day
#' @export
wday <- function(x, label = FALSE, abbr = TRUE,
                 week_start = getOption("lubridate.week.start", 7),
                 locale = Sys.getlocale("LC_TIME")) {
  UseMethod("wday")
}

#' @export
wday.default <- function(x, label = FALSE, abbr = TRUE,
                         week_start = getOption("lubridate.week.start", 7),
                         locale = Sys.getlocale("LC_TIME")) {
  wday.numeric(
    as.POSIXlt(x, tz = tz(x))$wday + 1,
    label = label,
    abbr = abbr,
    locale = locale,
    week_start = week_start)
}

.shift_wday_names <- function(names, week_start = 7) {
  if (week_start != 7) {
    c(names[(week_start + 1):7], names[1:week_start])
  } else {
    names
  }
}

## FIXME: Remove export. This is an internal function and has very
## confusing semantics. It works by assuming that 1 = Sunday always.
## https://github.com/tidyverse/lubridate/issues/1025
#' @export
wday.numeric <- function(x, label = FALSE, abbr = TRUE,
                         week_start = getOption("lubridate.week.start", 7),
                         locale = Sys.getlocale("LC_TIME")) {
  start <- as_week_start(week_start)

  if (start != 7) {
    x <- 1 + (x + (6 - start)) %% 7
  }

  if (!label) {
    return(x)
  }

  names <- .get_locale_regs(locale)$wday_names
  labels <- if (abbr) names$abr else names$full
  ordered(x, levels = 1:7, labels = .shift_wday_names(labels, week_start = start))
}

#' @export
mday.default <- function(x) {
  as.POSIXlt(x, tz = tz(x))$mday
}

#' @export
mday.Period <- function(x) {
  slot(x, "day")
}

#' @rdname day
#' @export
qday <- function(x) {
  UseMethod("qday")
}

#' @export
qday.default <- function(x) {
  x <- as_date(x)
  as.integer(x - floor_date(x, "quarter")) + 1
}

#' @rdname day
#' @export
yday <- function(x) {
  UseMethod("yday")
}

#' @export
yday.default <- function(x) {
  as.POSIXlt(x, tz = tz(x))$yday + 1
}

#' @rdname day
#' @export
setGeneric("day<-",
  function (x, value) standardGeneric("day<-"),
  useAsDefault = function(x, value) {
    y <- update_datetime(as.POSIXct(x), days = value)
    reclass_date(y, x)
  }
)

#' @export
setMethod("day<-", "Duration", function(x, value) {
  x <- x + days(value - day(x))
})

#' @export
setMethod("day<-", signature("Period"), function(x, value) {
  slot(x, "day") <- value
  x
})

#' @export
setMethod("day<-", "Interval", function(x, value) {
  x <- x + days(value - day(x))
})

#' @export
setMethod("day<-", "POSIXt", function(x, value) {
  update_datetime(x, days = value)
})

#' @export
setMethod("day<-", "Date", function(x, value) {
  update_datetime(x, days = value)
})

#' @rdname day
#' @export
`mday<-` <- function(x, value) {
  day(x) <- value
  x
}

#' @rdname day
#' @export
`qday<-` <- function(x, value) standardGeneric("qday<-")

#' @rdname day
#' @export
setGeneric("qday<-", useAsDefault = function(x, value) {
  x <- x + days(value - qday(x))
})

#' @rdname day
#' @export
"wday<-" <- function(x, week_start = getOption("lubridate.week.start", 7), value) {
  week_start <- as_week_start(week_start)
  if (is.character(value)) {
    value <- as_week_start(value)
    update(x, wdays = value, week_start = 1)
  } else {
    update(x, wdays = value, week_start = week_start)
  }
}

#' @rdname day
#' @export
"yday<-" <- function(x, value) {
  update(x, ydays = value)
}

context("Accessors")

test_that("seconds accessor extracts correct second", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(second(poslt), 59)
  expect_equal(second(posct), 59)
  expect_equal(second(date), 0)

})

test_that("minutes accessor extracts correct minute", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(minute(poslt), 45)
  expect_equal(minute(posct), 45)
  expect_equal(minute(date), 0)

})

test_that("hours accessor extracts correct hour", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(hour(poslt), 13)
  expect_equal(hour(posct), 13)
  expect_equal(hour(date), 0)

})

test_that("days accessors extract correct days", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(day(poslt), 3)
  expect_equal(day(posct), 3)
  expect_equal(day(date), 3)

  expect_equal(mday(poslt), 3)
  expect_equal(mday(posct), 3)
  expect_equal(mday(date), 3)

  expect_equal(yday(poslt), 34)
  expect_equal(yday(posct), 34)
  expect_equal(yday(date), 34)

  expect_equal(wday(poslt), 4)
  expect_equal(wday(posct), 4)
  expect_equal(wday(date), 4)

  expect_equal(qday(poslt), 34)
  expect_equal(qday(poslt), 34)
  expect_equal(qday(poslt), 34)
})

test_that("empty argument date() works", {
  expect_silent(date())
})

test_that("day accessors work on character inputs", {
  x <- "2017-05-07 GMT"
  d <- ymd(x)
  expect_equal(wday(x), wday(d))
  expect_equal(mday(x), mday(d))
  expect_equal(qday(x), qday(d))
  expect_equal(yday(x), yday(d))
})

test_that("wday works with various start values", {

  days <- days2 <-
    ymd(c("2005-01-01", "2005-01-02", "2005-12-31", "2007-01-01", "2007-12-30",
          "2007-12-31", "2008-01-01", "2008-12-28", "2008-12-29", "2008-12-30",
          "2008-12-31", "2009-01-01", "2009-12-31", "2010-01-01", "2010-01-02"))

  expect_equal(as.character(wday(days, label = T, week_start = 1)),
               as.character(wday(days, label = T, week_start = 3)))

  expect_equal(as.character(wday(days, label = T, week_start = 1)),
               as.character(wday(days, label = T, week_start = 7)))

  expect_equal(as.character(wday(days, label = T, locale = "C"))[1], "Sat")
  expect_equal(as.character(wday(days, label = T, abbr = FALSE, locale = "C"))[1], "Saturday")

  expect_equal(wday(days, label = F, week_start = 1),
               c(6, 7, 6, 1, 7, 1, 2, 7, 1, 2, 3, 4, 4, 5, 6))

  expect_equal(wday(days, label = F, week_start = 7),
               c(7, 1, 7, 2, 1, 2, 3, 1, 2, 3, 4, 5, 5, 6, 7))

  set.seed(1000)
  new_days <- sample(1:7, length(days2), replace = T)
  wday(days2, week_start = 1) <- new_days
  expect_equal(wday(days2, week_start = 1), new_days)
  wday(days2, week_start = 7) <- new_days
  expect_equal(wday(days2, week_start = 7), new_days)
  wday(days2, week_start = 3) <- new_days
  expect_equal(wday(days2, week_start = 3), new_days)
})

test_that("weeks accessor extracts correct week", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(week(poslt), 5)
  expect_equal(week(posct), 5)
  expect_equal(week(date), 5)

})

test_that("isoweek accessor extracts correct ISO8601 week", {
  poslt <- as.POSIXlt("2010-01-01 13:45:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(isoweek(poslt), 53)
  expect_equal(isoweek(posct), 53)
  expect_equal(isoweek(date), 53)

  df <- read.table(textConnection(
    "Sat 1 Jan 2005 	2005-01-01 	2004-W53-6
     Sun 2 Jan 2005 	2005-01-02 	2004-W53-7
     Sat 31 Dec 2005 	2005-12-31 	2005-W52-6
     Mon 1 Jan 2007 	2007-01-01 	2007-W01-1 	Both years 2007 start with the same day.
     Sun 30 Dec 2007 	2007-12-30 	2007-W52-7
     Mon 31 Dec 2007 	2007-12-31 	2008-W01-1
     Tue 1 Jan 2008 	2008-01-01 	2008-W01-2 	Gregorian year 2008 is a leap year. ISO year 2008 is 2 days shorter: 1 day longer at the start, 3 days shorter at the end.
     Sun 28 Dec 2008 	2008-12-28 	2008-W52-7 	ISO year 2009 begins three days before the end of Gregorian 2008.
     Mon 29 Dec 2008 	2008-12-29 	2009-W01-1
     Tue 30 Dec 2008 	2008-12-30 	2009-W01-2
     Wed 31 Dec 2008 	2008-12-31 	2009-W01-3
     Thu 1 Jan 2009 	2009-01-01 	2009-W01-4
     Thu 31 Dec 2009 	2009-12-31 	2009-W53-4 	ISO year 2009 has 53 weeks and ends three days into Gregorian year 2010.
     Fri 1 Jan 2010 	2010-01-01 	2009-W53-5
     Sat 2 Jan 2010 	2010-01-02 	2009-W53-6
     Sun 3 Jan 2010 	2010-01-03 	2009-W53-7"),
    sep = "\t", fill = T, stringsAsFactors = F, header = F)

  names(df) <- c("Gregorian", "ymd", "iso", "note")

  df <- within(df, {
    ymd <- ymd(ymd)
    isoweek <- as.numeric(gsub(".*W([0-9]+).*", "\\1", iso))
    isoyear <- as.numeric(gsub("^([0-9]+).*", "\\1", iso))
  })

  expect_equal(isoweek(df$ymd), df$isoweek)
  expect_equal(isoyear(df$ymd), df$isoyear)
})

test_that("epiweek computes dates correctly", {

  df <- read.table(textConnection(
    "ew	date	year
     1	12/30/07	8
     2	01/06/08	8
     3	01/13/08	8
     52	01/02/10	9
     1	01/03/10	10
     2	01/10/10	10
     3	01/17/10	10
     4	01/24/10	10
     5	01/31/10	10
     1	01/04/09	9
     2	01/11/09	9
     3	01/18/09	9
     50 	12/13/09	9
     51 	12/20/09	9
     52 	12/27/09	9
     50 	12/12/10	10
     51 	12/19/10	10
     52 	12/26/10	10
     50 	12/07/08	8
     51 	12/14/08	8
     52 	12/21/08	8"),
    header = T, sep = "\t", stringsAsFactors = F)

  date <- mdy(df$date)
  expect_equal(epiweek(date), df$ew)
  expect_equal(epiyear(date), df$year + 2000)
})

test_that("isoweek returns correct value for non-UTC time zone (#311)", {
  cest <- ymd_hms("2015-04-14 16:45:00", tz = "Europe/Copenhagen")
  utc <-  ymd_hms("2015-04-14 16:45:00", tz = "UTC")
  expect_equal(isoweek(cest), 16L)
  expect_equal(isoweek(utc), 16L)
})

test_that("months accessor extracts correct month", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(month(poslt), 2)
  expect_equal(month(posct), 2)
  expect_equal(month(date), 2)

  expect_equal(as.character(month(date, label = TRUE, locale = "C")), "Feb")
  expect_equal(as.character(month(date, label = TRUE, abbr = FALSE, locale = "C")),
              "February")

})

test_that("month() on numeric input doesn't accept invalid values", {
  expect_error(month(0:13))
  expect_error(month(c(0:13, NA)))
  expect_error(month(0))
  expect_error(month(0), label = TRUE)
  expect_equal(month(1:12), 1:12)
  expect_equal(month(c(1:12, NA)), c(1:12, NA))
  expect_equal(month(as.double(1:12)), 1:12)
})

test_that("quarters accessor extracts correct quarter", {
  posct <- ymd_hms("2010-11-03 13:45:59")
  poslt <- as.POSIXlt(posct)
  date <- as.Date(poslt)

  expect_equal(quarter(poslt), 4)
  expect_equal(quarter(poslt, with_year = TRUE), 2010.4)
  expect_equal(quarter(poslt, fiscal_start = 11), 1)
  expect_equal(quarter(poslt, with_year = TRUE, fiscal_start = -2), 2011.1)
  expect_equal(quarter(poslt, with_year = TRUE, fiscal_start = 11), 2011.1)

  expect_equal(quarter(posct), 4)
  expect_equal(quarter(posct, with_year = TRUE), 2010.4)
  expect_equal(quarter(posct, fiscal_start = 11), 1)
  expect_equal(quarter(posct, with_year = TRUE, fiscal_start = -2), 2011.1)
  expect_equal(quarter(posct, with_year = TRUE, fiscal_start = 11), 2011.1)

  expect_equal(quarter(date), 4)
  expect_equal(quarter(date, with_year = TRUE), 2010.4)
  expect_equal(quarter(date, fiscal_start = 11), 1)
  expect_equal(quarter(date, with_year = TRUE, fiscal_start = -2), 2011.1)
  expect_equal(quarter(date, with_year = TRUE, fiscal_start = 11), 2011.1)

  x <- ymd(c("2012-03-01", "2012-03-26", "2012-05-04", "2012-09-23", "2012-12-31"))
  expect_equal(quarter(x, with_year = TRUE, fiscal_start = 4),
               c(2012.4, 2012.4, 2013.1, 2013.2, 2013.3))
  expect_equal(quarter(x, with_year = TRUE, fiscal_start = 11),
               c(2012.2, 2012.2, 2012.3, 2012.4, 2013.1))

  x <- ymd("2010-01-01") + months(0:11)
  expect_equal(quarter(x, with_year = TRUE, fiscal_start = 6),
               c(2010.3, 2010.3, 2010.4, 2010.4, 2010.4, 2011.1, 2011.1, 2011.1,
                 2011.2, 2011.2, 2011.2, 2011.3))
  expect_equal(quarter(x, with_year = TRUE, fiscal_start = 6),
               quarter(x, with_year = TRUE, fiscal_start = -6))
  x <- ymd("2010-01-01") + months(seq(0, 23, by = 3))
  expect_equal(quarter(x, with_year = TRUE, fiscal_start = 10),
               c(2010.2, 2010.3, 2010.4, 2011.1, 2011.2, 2011.3, 2011.4, 2012.1))
  expect_equal(quarter(x, with_year = TRUE, fiscal_start = -2),
               quarter(x, with_year = TRUE, fiscal_start = 10))

  x <- ymd(c("2018-07-15", "2018-12-27", "2019-01-01",
             "2019-06-01", "2019-07-01", "2019-10-16",
             "2019-12-31", "2020-01-01", "2020-06-30",
             "2020-07-01", "2020-11-09", "2021-01-19",
             "2021-06-30", "2021-07-01"))
  expect_equal(quarter(x, with_year = T, fiscal_start = 7),
               c(2019.1, 2019.2, 2019.3, 2019.4, 2020.1, 2020.2, 2020.2, 2020.3,
                 2020.4, 2021.1, 2021.2, 2021.3, 2021.4, 2022.1))
})

test_that("years accessor extracts correct year", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(year(poslt), 2010)
  expect_equal(year(posct), 2010)
  expect_equal(year(date), 2010)

})

test_that("isoyear accessor extracts correct ISO8601 year", {
  poslt <- c(as.POSIXlt("2014-12-28 13:45:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S"),
             as.POSIXlt("2014-12-29 01:12:08", tz = "UTC", format = "%Y-%m-%d %H:%M:%S"))
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(isoyear(poslt), c(2014, 2015))
  expect_equal(isoyear(posct), c(2014, 2015))
  expect_equal(isoyear(date), c(2014, 2015))

})

test_that("date accessor extracts correct date", {
  poslt <- as.POSIXlt("2010-02-03 23:45:59", tz = "Etc/GMT+8", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)

  date <- as.Date(poslt)
  expect_equal(date(poslt), as.Date("2010-02-03"))
  expect_equal(date(posct), as.Date("2010-02-03"))
  expect_equal(date(date), as.Date("2010-02-03"))
})

test_that("timezone accessor extracts correct timezone", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_match(tz(poslt), "UTC")
  expect_match(tz(posct), "UTC")
  expect_match(tz(date), "UTC")

})


test_that("accessors handle vectors", {
  poslt <- as.POSIXlt(c("2001-01-01 01:01:01",
    "2002-02-02 02:02:02", "2003-03-03 03:03:03"),
    tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(second(poslt), c(1, 2, 3))
  expect_equal(second(posct), c(1, 2, 3))
  expect_equal(second(date), c(0, 0, 0))

  expect_equal(minute(poslt), c(1, 2, 3))
  expect_equal(minute(posct), c(1, 2, 3))
  expect_equal(minute(date), c(0, 0, 0))

  expect_equal(hour(poslt), c(1, 2, 3))
  expect_equal(hour(posct), c(1, 2, 3))
  expect_equal(hour(date), c(0, 0, 0))

  expect_equal(mday(poslt), c(1, 2, 3))
  expect_equal(mday(posct), c(1, 2, 3))
  expect_equal(mday(date), c(1, 2, 3))

  expect_equal(wday(poslt), c(2, 7, 2))
  expect_equal(wday(posct), c(2, 7, 2))
  expect_equal(wday(date), c(2, 7, 2))

  expect_equal(yday(poslt), c(1, 33, 62))
  expect_equal(yday(posct), c(1, 33, 62))
  expect_equal(yday(date), c(1, 33, 62))

  expect_equal(week(poslt), c(1, 5, 9))
  expect_equal(week(posct), c(1, 5, 9))
  expect_equal(week(date), c(1, 5, 9))

  expect_equal(isoweek(poslt), c(1, 5, 10))
  expect_equal(isoweek(posct), c(1, 5, 10))
  expect_equal(isoweek(date), c(1, 5, 10))

  expect_equal(month(poslt), c(1, 2, 3))
  expect_equal(month(posct), c(1, 2, 3))
  expect_equal(month(date), c(1, 2, 3))

  expect_equal(year(poslt), c(2001, 2002, 2003))
  expect_equal(year(posct), c(2001, 2002, 2003))
  expect_equal(year(date), c(2001, 2002, 2003))

  expect_equal(date(poslt), as.Date(c("2001-01-01", "2002-02-02", "2003-03-03")))
  expect_equal(date(posct), as.Date(c("2001-01-01", "2002-02-02", "2003-03-03")))
  expect_equal(date(date), as.Date(c("2001-01-01", "2002-02-02", "2003-03-03")))

  expect_match(tz(poslt), "UTC")
  expect_match(tz(posct), "UTC")
  expect_match(tz(date), "UTC")
})

test_that("accessors handle Period objects", {
  per <- period(seconds = 1, minutes = 2, hours = 3, days = 4, months = 5,  years = 6)
  pers <- c(per, per)

  expect_equal(second(per), 1)
  expect_equal(minute(per), 2)
  expect_equal(hour(per), 3)
  expect_equal(day(per), 4)
  expect_equal(month(per), 5)
  expect_equal(year(per), 6)
  expect_error(date(per), "date is undefined for Period objects")

  expect_equal(second(pers), c(1, 1))
  expect_equal(minute(pers), c(2, 2))
  expect_equal(hour(pers), c(3, 3))
  expect_equal(day(pers), c(4, 4))
  expect_equal(month(pers), c(5, 5))
  expect_equal(year(pers), c(6, 6))
  expect_error(date(pers), "date is undefined for Period objects")

  second(per) <- 2
  minute(per) <- 3
  hour(per) <- 4
  day(per) <- 5
  month(per) <- 6
  year(per) <- 7

  expect_equal(per@.Data, 2)
  expect_equal(per@minute, 3)
  expect_equal(per@hour, 4)
  expect_equal(per@day, 5)
  expect_equal(per@month, 6)
  expect_equal(per@year, 7)

  second(pers) <- c(2, 3)
  minute(pers) <- c(3, 4)
  hour(pers) <- c(4, 5)
  day(pers) <- c(5, 6)
  month(pers) <- c(6, 7)
  year(pers) <- c(7, 8)

  expect_equal(pers@.Data, c(2, 3))
  expect_equal(pers@minute, c(3, 4))
  expect_equal(pers@hour, c(4, 5))
  expect_equal(pers@day, c(5, 6))
  expect_equal(pers@month, c(6, 7))
  expect_equal(pers@year, c(7, 8))

})



context("Test days_in_month")

test_that(
  "days in month works for non leap years",
{
  x <- seq(ymd("2011-01-01"), ymd("2011-12-01"), "1 month")
  expected <- c(
    Jan = 31L, Feb = 28L, Mar = 31L,
    Apr = 30L, May = 31L, Jun = 30L,
    Jul = 31L, Aug = 31L, Sep = 30L,
    Oct = 31L, Nov = 30L, Dec = 31L
  )
  expect_equal(days_in_month(x), expected)
}
)

test_that(
  "days in month works for leap years",
{
  x <- seq(ymd("2012-01-01"), ymd("2012-12-01"), "1 month")
  expected <- c(
    Jan = 31L, Feb = 29L, Mar = 31L,
    Apr = 30L, May = 31L, Jun = 30L,
    Jul = 31L, Aug = 31L, Sep = 30L,
    Oct = 31L, Nov = 30L, Dec = 31L
  )
  expect_equal(days_in_month(x), expected)
}
)

test_that("seconds accessor extracts correct second", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(second(poslt), 59)
  expect_equal(second(posct), 59)
  expect_equal(second(date), 0)
})

test_that("minutes accessor extracts correct minute", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(minute(poslt), 45)
  expect_equal(minute(posct), 45)
  expect_equal(minute(date), 0)
})

test_that("hours accessor extracts correct hour", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(hour(poslt), 13)
  expect_equal(hour(posct), 13)
  expect_equal(hour(date), 0)
})

test_that("days accessors extract correct days", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(day(poslt), 3)
  expect_equal(day(posct), 3)
  expect_equal(day(date), 3)

  expect_equal(mday(poslt), 3)
  expect_equal(mday(posct), 3)
  expect_equal(mday(date), 3)

  expect_equal(yday(poslt), 34)
  expect_equal(yday(posct), 34)
  expect_equal(yday(date), 34)

  expect_equal(wday(poslt), 4)
  expect_equal(wday(posct), 4)
  expect_equal(wday(date), 4)

  expect_equal(qday(poslt), 34)
  expect_equal(qday(poslt), 34)
  expect_equal(qday(poslt), 34)
})

test_that("empty argument date() works", {
  expect_silent(date())
})

test_that("day accessors work on character inputs", {
  x <- "2017-05-07 GMT"
  d <- ymd(x)
  expect_equal(wday(x), wday(d))
  expect_equal(mday(x), mday(d))
  expect_equal(qday(x), qday(d))
  expect_equal(yday(x), yday(d))
})

test_that("wday works with various start values", {

  days <- days2 <-
    ymd(c(
      "2005-01-01", "2005-01-02", "2005-12-31", "2007-01-01", "2007-12-30",
      "2007-12-31", "2008-01-01", "2008-12-28", "2008-12-29", "2008-12-30",
      "2008-12-31", "2009-01-01", "2009-12-31", "2010-01-01", "2010-01-02"
    ))

  expect_equal(
    as.character(wday(days, label = T, week_start = 1)),
    as.character(wday(days, label = T, week_start = 3))
  )

  expect_equal(
    as.character(wday(days, label = T, week_start = 1)),
    as.character(wday(days, label = T, week_start = 7))
  )

  expect_equal(as.character(wday(days, label = T, locale = "C"))[1], "Sat")
  expect_equal(as.character(wday(days, label = T, abbr = FALSE, locale = "C"))[1], "Saturday")

  expect_equal(
    wday(days, label = F, week_start = 1),
    c(6, 7, 6, 1, 7, 1, 2, 7, 1, 2, 3, 4, 4, 5, 6)
  )

  expect_equal(
    wday(days, label = F, week_start = 7),
    c(7, 1, 7, 2, 1, 2, 3, 1, 2, 3, 4, 5, 5, 6, 7)
  )

  set.seed(1000)
  days2 <- days
  new_days <- sample(1:7, length(days2), replace = T)
  wday(days2, week_start = 1) <- new_days
  expect_equal(wday(days2, week_start = 1), new_days)
  wday(days2, week_start = 7) <- new_days
  expect_equal(wday(days2, week_start = 7), new_days)
  wday(days2, week_start = 3) <- new_days
  expect_equal(wday(days2, week_start = 3), new_days)
})

test_that("weeks accessor extracts correct week", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(week(poslt), 5)
  expect_equal(week(posct), 5)
  expect_equal(week(date), 5)
})

test_that("isoweek accessor extracts correct ISO8601 week", {
  poslt <- as.POSIXlt("2010-01-01 13:45:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(isoweek(poslt), 53)
  expect_equal(isoweek(posct), 53)
  expect_equal(isoweek(date), 53)

  df <- read.table(textConnection(
    "Sat 1 Jan 2005 	2005-01-01 	2004-W53-6
     Sun 2 Jan 2005 	2005-01-02 	2004-W53-7
     Sat 31 Dec 2005 	2005-12-31 	2005-W52-6
     Mon 1 Jan 2007 	2007-01-01 	2007-W01-1 	Both years 2007 start with the same day.
     Sun 30 Dec 2007 	2007-12-30 	2007-W52-7
     Mon 31 Dec 2007 	2007-12-31 	2008-W01-1
     Tue 1 Jan 2008 	2008-01-01 	2008-W01-2 	Gregorian year 2008 is a leap year. ISO year 2008 is 2 days shorter: 1 day longer at the start, 3 days shorter at the end.
     Sun 28 Dec 2008 	2008-12-28 	2008-W52-7 	ISO year 2009 begins three days before the end of Gregorian 2008.
     Mon 29 Dec 2008 	2008-12-29 	2009-W01-1
     Tue 30 Dec 2008 	2008-12-30 	2009-W01-2
     Wed 31 Dec 2008 	2008-12-31 	2009-W01-3
     Thu 1 Jan 2009 	2009-01-01 	2009-W01-4
     Thu 31 Dec 2009 	2009-12-31 	2009-W53-4 	ISO year 2009 has 53 weeks and ends three days into Gregorian year 2010.
     Fri 1 Jan 2010 	2010-01-01 	2009-W53-5
     Sat 2 Jan 2010 	2010-01-02 	2009-W53-6
     Sun 3 Jan 2010 	2010-01-03 	2009-W53-7"
  ),
  sep = "\t", fill = T, stringsAsFactors = F, header = F
  )

  names(df) <- c("Gregorian", "ymd", "iso", "note")

  df <- within(df, {
    ymd <- ymd(ymd)
    isoweek <- as.numeric(gsub(".*W([0-9]+).*", "\\1", iso))
    isoyear <- as.numeric(gsub("^([0-9]+).*", "\\1", iso))
  })

  expect_equal(isoweek(df$ymd), df$isoweek)
  expect_equal(isoyear(df$ymd), df$isoyear)
})

test_that("epiweek computes dates correctly", {
  df <- read.table(textConnection(
    "ew	date	year
     1	12/30/07	8
     2	01/06/08	8
     3	01/13/08	8
     52	01/02/10	9
     1	01/03/10	10
     2	01/10/10	10
     3	01/17/10	10
     4	01/24/10	10
     5	01/31/10	10
     1	01/04/09	9
     2	01/11/09	9
     3	01/18/09	9
     50 	12/13/09	9
     51 	12/20/09	9
     52 	12/27/09	9
     50 	12/12/10	10
     51 	12/19/10	10
     52 	12/26/10	10
     50 	12/07/08	8
     51 	12/14/08	8
     52 	12/21/08	8"
  ),
  header = T, sep = "\t", stringsAsFactors = F
  )

  date <- mdy(df$date)
  expect_equal(epiweek(date), df$ew)
  expect_equal(epiyear(date), df$year + 2000)
})

test_that("isoweek returns correct value for non-UTC time zone (#311)", {
  cest <- ymd_hms("2015-04-14 16:45:00", tz = "Europe/Copenhagen")
  utc <- ymd_hms("2015-04-14 16:45:00", tz = "UTC")
  expect_equal(isoweek(cest), 16L)
  expect_equal(isoweek(utc), 16L)
})

test_that("months accessor extracts correct month", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(month(poslt), 2)
  expect_equal(month(posct), 2)
  expect_equal(month(date), 2)

  expect_equal(as.character(month(date, label = TRUE, locale = "C")), "Feb")
  expect_equal(
    as.character(month(date, label = TRUE, abbr = FALSE, locale = "C")),
    "February"
  )
})

test_that("month() on numeric input doesn't accept invalid values", {
  expect_error(month(0:13))
  expect_error(month(c(0:13, NA)))
  expect_error(month(0))
  expect_error(month(0), label = TRUE)
  expect_equal(month(1:12), 1:12)
  expect_equal(month(c(1:12, NA)), c(1:12, NA))
  expect_equal(month(as.double(1:12)), 1:12)
})

test_that("quarters accessor extracts correct quarter", {
  posct <- ymd_hms("2010-11-03 13:45:59")
  poslt <- as.POSIXlt(posct)
  date <- as.Date(poslt)

  expect_equal(quarter(posct), 4)
  expect_equal(quarter(posct, type = "quarter"), 4)
  expect_equal(quarter(posct, type = "year.quarter"), 2010.4)
  expect_equal(quarter(posct, type = "date_first"), ymd("2010-10-01"))
  expect_equal(quarter(posct, type = "date_last"), ymd("2010-12-31"))
  expect_equal(quarter(posct, with_year = TRUE), 2010.4)
  expect_equal(quarter(posct, fiscal_start = 11), 1)
  expect_equal(quarter(posct, fiscal_start = -2), 1)
  expect_equal(quarter(posct, type = "year.quarter", fiscal_start = -2), 2011.1)
  expect_equal(quarter(posct, type = "year.quarter", fiscal_start = 11), 2011.1)
  expect_equal(quarter(posct, type = "date_first", fiscal_start = 11), ymd("2010-11-01"))
  expect_equal(quarter(posct, type = "date_last", fiscal_start = 11), ymd("2011-01-31"))

  expect_equal(quarter(poslt), 4)
  expect_equal(quarter(poslt, type = "quarter"), 4)
  expect_equal(quarter(poslt, type = "year.quarter"), 2010.4)
  expect_equal(quarter(poslt, type = "date_first"), ymd("2010-10-01"))
  expect_equal(quarter(poslt, type = "date_last"), ymd("2010-12-31"))
  expect_equal(quarter(poslt, with_year = TRUE), 2010.4)
  expect_equal(quarter(poslt, fiscal_start = 11), 1)
  expect_equal(quarter(poslt, fiscal_start = -2), 1)
  expect_equal(quarter(poslt, type = "year.quarter", fiscal_start = -2), 2011.1)
  expect_equal(quarter(poslt, type = "year.quarter", fiscal_start = 11), 2011.1)
  expect_equal(quarter(poslt, type = "date_first", fiscal_start = 11), ymd("2010-11-01"))
  expect_equal(quarter(poslt, type = "date_last", fiscal_start = 11), ymd("2011-01-31"))

  expect_equal(quarter(date), 4)
  expect_equal(quarter(date, type = "quarter"), 4)
  expect_equal(quarter(date, type = "year.quarter"), 2010.4)
  expect_equal(quarter(date, type = "date_first"), ymd("2010-10-01"))
  expect_equal(quarter(date, type = "date_last"), ymd("2010-12-31"))
  expect_equal(quarter(date, with_year = TRUE), 2010.4)
  expect_equal(quarter(date, fiscal_start = 11), 1)
  expect_equal(quarter(date, fiscal_start = -2), 1)
  expect_equal(quarter(date, type = "year.quarter", fiscal_start = -2), 2011.1)
  expect_equal(quarter(date, type = "year.quarter", fiscal_start = 11), 2011.1)
  expect_equal(quarter(date, type = "date_first", fiscal_start = 11), ymd("2010-11-01"))
  expect_equal(quarter(date, type = "date_last", fiscal_start = 11), ymd("2011-01-31"))

  x <- ymd(c("2012-03-01", "2012-03-26", "2012-05-04", "2012-09-23", "2012-12-31"))
  expect_equal(quarter(x, FALSE, fiscal_start = 11), c(2, 2, 3, 4, 1))
  expect_equal(quarter(x, fiscal_start = 11, FALSE), c(2, 2, 3, 4, 1))
  expect_equal(quarter(x, TRUE, fiscal_start = 9), c(2012.3, 2012.3, 2012.3, 2013.1, 2013.2))
  expect_equal(quarter(x, TRUE, fiscal_start = 9), c(2012.3, 2012.3, 2012.3, 2013.1, 2013.2))

  expect_equal(
    quarter(x, type = "quarter", fiscal_start = 4),
    c(4, 4, 1, 2, 3)
  )
  expect_equal(
    quarter(x, type = "year.quarter", fiscal_start = 4),
    c(2012.4, 2012.4, 2013.1, 2013.2, 2013.3)
  )
  expect_equal(
    quarter(x, type = "year.quarter", fiscal_start = 11),
    c(2012.2, 2012.2, 2012.3, 2012.4, 2013.1)
  )
  expect_equal(
    quarter(x, type = "date_first", fiscal_start = 4),
    ymd(c("2012-01-01", "2012-01-01", "2012-04-01", "2012-07-01", "2012-10-01"))
  )
  expect_equal(
    quarter(x, type = "date_first", fiscal_start = 11),
    ymd(c("2012-02-01", "2012-02-01", "2012-05-01", "2012-08-01", "2012-11-01"))
  )
  expect_equal(
    quarter(x, type = "date_last", fiscal_start = 4),
    ymd(c("2012-03-31", "2012-03-31", "2012-06-30", "2012-09-30", "2012-12-31"))
  )
  expect_equal(
    quarter(x, type = "date_last", fiscal_start = 11),
    ymd(c("2012-04-30", "2012-04-30", "2012-07-31", "2012-10-31", "2013-01-31"))
  )

  x <- ymd("2010-01-01") + months(0:11)
  expect_equal(
    quarter(x, type = "year.quarter", fiscal_start = 6),
    c(
      2010.3, 2010.3, 2010.4, 2010.4, 2010.4, 2011.1, 2011.1, 2011.1,
      2011.2, 2011.2, 2011.2, 2011.3
    )
  )
  expect_equal(
    quarter(x, type = "year.quarter", fiscal_start = 6),
    quarter(x, type = "year.quarter", fiscal_start = -6)
  )
  expect_equal(
    quarter(x, type = "date_first", fiscal_start = 6),
    ymd(c(
      "2009-12-01", "2009-12-01", "2010-03-01", "2010-03-01",
      "2010-03-01", "2010-06-01", "2010-06-01", "2010-06-01",
      "2010-09-01", "2010-09-01", "2010-09-01", "2010-12-01"
    ))
  )
  expect_equal(
    quarter(x, type = "date_first", fiscal_start = 6),
    quarter(x, type = "date_first", fiscal_start = -6)
  )
  expect_equal(
    quarter(x, type = "date_last", fiscal_start = 6),
    ymd(c(
      "2010-02-28", "2010-02-28", "2010-05-31", "2010-05-31",
      "2010-05-31", "2010-08-31", "2010-08-31", "2010-08-31",
      "2010-11-30", "2010-11-30", "2010-11-30", "2011-02-28"
    ))
  )
  expect_equal(
    quarter(x, type = "date_last", fiscal_start = 6),
    quarter(x, type = "date_last", fiscal_start = -6)
  )

  x <- ymd("2010-01-01") + months(seq(0, 23, by = 3))
  expect_equal(
    quarter(x, type = "year.quarter", fiscal_start = 10),
    c(2010.2, 2010.3, 2010.4, 2011.1, 2011.2, 2011.3, 2011.4, 2012.1)
  )
  expect_equal(
    quarter(x, type = "year.quarter", fiscal_start = -2),
    quarter(x, type = "year.quarter", fiscal_start = 10)
  )
  expect_equal(
    quarter(x, type = "date_first", fiscal_start = 10),
    ymd(c(
      "2010-01-01", "2010-04-01", "2010-07-01", "2010-10-01",
      "2011-01-01", "2011-04-01", "2011-07-01", "2011-10-01"
    ))
  )
  expect_equal(
    quarter(x, type = "date_first", fiscal_start = -2),
    quarter(x, type = "date_first", fiscal_start = 10)
  )
  expect_equal(
    quarter(x, type = "date_last", fiscal_start = 10),
    ymd(c(
      "2010-03-31", "2010-06-30", "2010-09-30", "2010-12-31",
      "2011-03-31", "2011-06-30", "2011-09-30", "2011-12-31"
    ))
  )
  expect_equal(
    quarter(x, type = "date_last", fiscal_start = -2),
    quarter(x, type = "date_last", fiscal_start = 10)
  )

  x <- ymd(c(
    "2018-07-15", "2018-12-27", "2019-01-01",
    "2019-06-01", "2019-07-01", "2019-10-16",
    "2019-12-31", "2020-01-01", "2020-06-30",
    "2020-07-01", "2020-11-09", "2021-01-19",
    "2021-06-30", "2021-07-01"
  ))
  expect_equal(
    quarter(x, type = "year.quarter", fiscal_start = 7),
    c(
      2019.1, 2019.2, 2019.3, 2019.4, 2020.1, 2020.2, 2020.2, 2020.3,
      2020.4, 2021.1, 2021.2, 2021.3, 2021.4, 2022.1
    )
  )
  out <- quarter(x, type = "date_first", fiscal_start = 7)
  expect_s3_class(out, "Date")
  expect_equal(
    out,
    ymd(c(
      "2018-07-01", "2018-10-01", "2019-01-01", "2019-04-01",
      "2019-07-01", "2019-10-01", "2019-10-01", "2020-01-01",
      "2020-04-01", "2020-07-01", "2020-10-01", "2021-01-01",
      "2021-04-01", "2021-07-01"
    ))
  )
  expect_equal(
    quarter(x, type = "date_last", fiscal_start = 7),
    ymd(c(
      "2018-09-30", "2018-12-31", "2019-03-31", "2019-06-30",
      "2019-09-30", "2019-12-31", "2019-12-31", "2020-03-31",
      "2020-06-30", "2020-09-30", "2020-12-31", "2021-03-31",
      "2021-06-30", "2021-09-30"
    ))
  )
})

test_that("years accessor extracts correct year", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(year(poslt), 2010)
  expect_equal(year(posct), 2010)
  expect_equal(year(date), 2010)
})

test_that("isoyear accessor extracts correct ISO8601 year", {
  poslt <- c(
    as.POSIXlt("2014-12-28 13:45:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S"),
    as.POSIXlt("2014-12-29 01:12:08", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(isoyear(poslt), c(2014, 2015))
  expect_equal(isoyear(posct), c(2014, 2015))
  expect_equal(isoyear(date), c(2014, 2015))
})

test_that("date accessor extracts correct date", {
  poslt <- as.POSIXlt("2010-02-03 23:45:59",
    tz = "Etc/GMT+8", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)

  date <- as.Date(poslt)
  expect_equal(date(poslt), as.Date("2010-02-03"))
  expect_equal(date(posct), as.Date("2010-02-03"))
  expect_equal(date(date), as.Date("2010-02-03"))
})

test_that("timezone accessor extracts correct timezone", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_match(tz(poslt), "UTC")
  expect_match(tz(posct), "UTC")
  expect_match(tz(date), "UTC")
})


test_that("accessors handle vectors", {
  poslt <- as.POSIXlt(c(
    "2001-01-01 01:01:01",
    "2002-02-02 02:02:02", "2003-03-03 03:03:03"
  ),
  tz = "UTC", format = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  expect_equal(second(poslt), c(1, 2, 3))
  expect_equal(second(posct), c(1, 2, 3))
  expect_equal(second(date), c(0, 0, 0))

  expect_equal(minute(poslt), c(1, 2, 3))
  expect_equal(minute(posct), c(1, 2, 3))
  expect_equal(minute(date), c(0, 0, 0))

  expect_equal(hour(poslt), c(1, 2, 3))
  expect_equal(hour(posct), c(1, 2, 3))
  expect_equal(hour(date), c(0, 0, 0))

  expect_equal(mday(poslt), c(1, 2, 3))
  expect_equal(mday(posct), c(1, 2, 3))
  expect_equal(mday(date), c(1, 2, 3))

  expect_equal(wday(poslt), c(2, 7, 2))
  expect_equal(wday(posct), c(2, 7, 2))
  expect_equal(wday(date), c(2, 7, 2))

  expect_equal(yday(poslt), c(1, 33, 62))
  expect_equal(yday(posct), c(1, 33, 62))
  expect_equal(yday(date), c(1, 33, 62))

  expect_equal(week(poslt), c(1, 5, 9))
  expect_equal(week(posct), c(1, 5, 9))
  expect_equal(week(date), c(1, 5, 9))

  expect_equal(isoweek(poslt), c(1, 5, 10))
  expect_equal(isoweek(posct), c(1, 5, 10))
  expect_equal(isoweek(date), c(1, 5, 10))

  expect_equal(month(poslt), c(1, 2, 3))
  expect_equal(month(posct), c(1, 2, 3))
  expect_equal(month(date), c(1, 2, 3))

  expect_equal(year(poslt), c(2001, 2002, 2003))
  expect_equal(year(posct), c(2001, 2002, 2003))
  expect_equal(year(date), c(2001, 2002, 2003))

  expect_equal(date(poslt), as.Date(c("2001-01-01", "2002-02-02", "2003-03-03")))
  expect_equal(date(posct), as.Date(c("2001-01-01", "2002-02-02", "2003-03-03")))
  expect_equal(date(date), as.Date(c("2001-01-01", "2002-02-02", "2003-03-03")))

  expect_match(tz(poslt), "UTC")
  expect_match(tz(posct), "UTC")
  expect_match(tz(date), "UTC")

})

test_that("accessors<- handle vectors", {

  poslt <- as.POSIXlt(c(
    "1999-01-01 00:00:00",
    "1999-01-01 00:00:00",
    "1999-01-01 00:00:00"
  ), tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  wday(poslt) <- c(2, 7, 2)
  wday(posct) <- c(2, 7, 2)
  wday(date) <- c(2, 7, 2)
  expect_equal(wday(poslt), c(2, 7, 2))
  expect_equal(wday(posct), c(2, 7, 2))
  expect_equal(wday(date), c(2, 7, 2))

  second(poslt) <- 1:3
  second(posct) <- 1:3
  second(date) <- 1:3
  expect_equal(second(poslt), c(1, 2, 3))
  expect_equal(second(posct), c(1, 2, 3))
  expect_equal(second(date), c(1, 2, 3))

  minute(poslt) <- 1:3
  minute(posct) <- 1:3
  minute(date) <- 1:3
  expect_equal(minute(poslt), c(1, 2, 3))
  expect_equal(minute(posct), c(1, 2, 3))
  expect_equal(minute(date), c(1, 2, 3))

  hour(poslt) <- 1:3
  hour(posct) <- 1:3
  hour(date) <- 1:3
  expect_equal(hour(poslt), c(1, 2, 3))
  expect_equal(hour(posct), c(1, 2, 3))
  expect_equal(hour(date), c(1, 2, 3))

  mday(poslt) <- 1:3
  mday(posct) <- 1:3
  mday(date) <- 1:3
  expect_equal(mday(poslt), c(1, 2, 3))
  expect_equal(mday(posct), c(1, 2, 3))
  expect_equal(mday(date), c(1, 2, 3))

  wday(poslt) <- c(2, 7, 2)
  wday(posct) <- c(2, 7, 2)
  wday(date) <- c(2, 7, 2)
  expect_equal(wday(poslt), c(2, 7, 2))
  expect_equal(wday(posct), c(2, 7, 2))
  expect_equal(wday(date), c(2, 7, 2))

  yday(poslt) <- c(1, 33, 62)
  yday(posct) <- c(1, 33, 62)
  yday(date) <- c(1, 33, 62)
  expect_equal(yday(poslt), c(1, 33, 62))
  expect_equal(yday(posct), c(1, 33, 62))
  expect_equal(yday(date), c(1, 33, 62))

  week(poslt) <- c(1, 5, 9)
  week(posct) <- c(1, 5, 9)
  week(date) <- c(1, 5, 9)
  expect_equal(week(poslt), c(1, 5, 9))
  expect_equal(week(posct), c(1, 5, 9))
  expect_equal(week(date), c(1, 5, 9))

  month(poslt) <- c(1, 2, 3)
  month(posct) <- c(1, 2, 3)
  month(date) <- c(1, 2, 3)
  expect_equal(month(poslt), c(1, 2, 3))
  expect_equal(month(posct), c(1, 2, 3))
  expect_equal(month(date), c(1, 2, 3))

  year(poslt) <- c(2001, 2002, 2003)
  year(posct) <- c(2001, 2002, 2003)
  year(date) <- c(2001, 2002, 2003)
  expect_equal(year(poslt), c(2001, 2002, 2003))
  expect_equal(year(posct), c(2001, 2002, 2003))
  expect_equal(year(date), c(2001, 2002, 2003))

})

test_that("accessors handle Period objects", {
  per <- period(seconds = 1, minutes = 2, hours = 3, days = 4, months = 5, years = 6)
  pers <- c(per, per)

  expect_equal(second(per), 1)
  expect_equal(minute(per), 2)
  expect_equal(hour(per), 3)
  expect_equal(day(per), 4)
  expect_equal(month(per), 5)
  expect_equal(year(per), 6)
  expect_error(date(per), "date is undefined for Period objects")

  expect_equal(second(pers), c(1, 1))
  expect_equal(minute(pers), c(2, 2))
  expect_equal(hour(pers), c(3, 3))
  expect_equal(day(pers), c(4, 4))
  expect_equal(month(pers), c(5, 5))
  expect_equal(year(pers), c(6, 6))
  expect_error(date(pers), "date is undefined for Period objects")

  second(per) <- 2
  minute(per) <- 3
  hour(per) <- 4
  day(per) <- 5
  month(per) <- 6
  year(per) <- 7

  expect_equal(per@.Data, 2)
  expect_equal(per@minute, 3)
  expect_equal(per@hour, 4)
  expect_equal(per@day, 5)
  expect_equal(per@month, 6)
  expect_equal(per@year, 7)

  second(pers) <- c(2, 3)
  minute(pers) <- c(3, 4)
  hour(pers) <- c(4, 5)
  day(pers) <- c(5, 6)
  month(pers) <- c(6, 7)
  year(pers) <- c(7, 8)

  expect_equal(pers@.Data, c(2, 3))
  expect_equal(pers@minute, c(3, 4))
  expect_equal(pers@hour, c(4, 5))
  expect_equal(pers@day, c(5, 6))
  expect_equal(pers@month, c(6, 7))
  expect_equal(pers@year, c(7, 8))
})

test_that(
  "days in month works for non leap years",
  {
    x <- seq(ymd("2011-01-01"), ymd("2011-12-01"), "1 month")
    expected <- c(
      Jan = 31L, Feb = 28L, Mar = 31L,
      Apr = 30L, May = 31L, Jun = 30L,
      Jul = 31L, Aug = 31L, Sep = 30L,
      Oct = 31L, Nov = 30L, Dec = 31L
    )
    expect_equal(days_in_month(x), expected)
  }
)

test_that(
  "days in month works for leap years",
  {
    x <- seq(ymd("2012-01-01"), ymd("2012-12-01"), "1 month")
    expected <- c(
      Jan = 31L, Feb = 29L, Mar = 31L,
      Apr = 30L, May = 31L, Jun = 30L,
      Jul = 31L, Aug = 31L, Sep = 30L,
      Oct = 31L, Nov = 30L, Dec = 31L
    )
    expect_equal(days_in_month(x), expected)
  }
)

test_that("seconds settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  second(poslt) <- 10
  second(posct) <- 10
  second(date) <- 10

  expect_equal(second(poslt), 10)
  expect_equal(second(posct), 10)
  expect_equal(second(date), 10)
})

test_that("seconds settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  second(poslt) <- 69
  second(posct) <- 69
  second(date) <- 69

  expect_equal(second(poslt), 9)
  expect_equal(minute(poslt), 0)
  expect_equal(hour(poslt), 0)
  expect_equal(mday(poslt), 1)
  expect_equal(wday(poslt), 7)
  expect_equal(yday(poslt), 1)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 9)
  expect_equal(minute(posct), 0)
  expect_equal(hour(posct), 0)
  expect_equal(mday(posct), 1)
  expect_equal(wday(posct), 7)
  expect_equal(yday(posct), 1)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 9)
  expect_equal(minute(date), 1)
  expect_equal(hour(date), 0)
  expect_equal(mday(date), 31)
  expect_equal(wday(date), 6)
  expect_equal(yday(date), 365)
  expect_equal(month(date), 12)
  expect_equal(year(date), 2010)
  expect_equal(tz(date), "UTC")
})

test_that("seconds settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  second(poslt) <- 1
  second(posct) <- 1
  second(date) <- 1

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

  second(poslt) <- 69
  second(posct) <- 69
  second(date) <- 69

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("seconds settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-14 01:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posxct <- as.POSIXct(poslt)

  second(poslt) <- 69
  second(posxct) <- 69

  expect_true(is.na(poslt))
  expect_true(is.na(posxct))
})

test_that("seconds settor retains object class for datetimes", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)

  second(poslt) <- 10
  second(posct) <- 10

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")

  second(poslt) <- 70
  second(posct) <- 70

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
})

test_that("seconds settor makes POSTXct from Date", {
  date <- as.Date("2010-02-14", tz = "GMT", format = "%Y-%m-%d")

  second(date) <- 10

  expect_s3_class(date, "POSIXct")

  second(date) <- 70

  expect_s3_class(date, "POSIXct")
})


test_that("minutes settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  minute(poslt) <- 10
  minute(posct) <- 10
  minute(date) <- 10

  expect_equal(minute(poslt), 10)
  expect_equal(minute(posct), 10)
  expect_equal(minute(date), 10)
})

test_that("minutes settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  minute(poslt) <- 69
  minute(posct) <- 69
  minute(date) <- 69

  expect_equal(second(poslt), 59)
  expect_equal(minute(poslt), 9)
  expect_equal(hour(poslt), 0)
  expect_equal(mday(poslt), 1)
  expect_equal(wday(poslt), 7)
  expect_equal(yday(poslt), 1)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 59)
  expect_equal(minute(posct), 9)
  expect_equal(hour(posct), 0)
  expect_equal(mday(posct), 1)
  expect_equal(wday(posct), 7)
  expect_equal(yday(posct), 1)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 0)
  expect_equal(minute(date), 9)
  expect_equal(hour(date), 1)
  expect_equal(mday(date), 31)
  expect_equal(wday(date), 6)
  expect_equal(yday(date), 365)
  expect_equal(month(date), 12)
  expect_equal(year(date), 2010)
  expect_equal(tz(date), "UTC")
})

test_that("minutes settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  minute(poslt) <- 1
  minute(posct) <- 1
  minute(date) <- 1

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

  minute(poslt) <- 70
  minute(posct) <- 70
  minute(date) <- 70

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("minutes settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-14 01:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  minute(poslt) <- 70
  minute(posct) <- 70

  expect_true(is.na(poslt))
  expect_true(is.na(posct))
})

test_that("minutes settor retains object class for datetimes", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)

  minute(poslt) <- 1
  minute(posct) <- 1

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")

  minute(poslt) <- 70
  minute(posct) <- 70

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
})

test_that("minutes settor makes POSIXct from Date", {
  date <- as.Date("2010-02-14", tz = "GMT", format = "%Y-%m-%d")

  minute(date) <- 70

  expect_s3_class(date, "POSIXct")
})


test_that("hours settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  hour(poslt) <- 10
  hour(posct) <- 10
  hour(date) <- 10

  expect_equal(hour(poslt), 10)
  expect_equal(hour(posct), 10)
  expect_equal(hour(date), 10)
})

test_that("hours settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  hour(poslt) <- 25
  hour(posct) <- 25
  hour(date) <- 25

  expect_equal(second(poslt), 59)
  expect_equal(minute(poslt), 59)
  expect_equal(hour(poslt), 1)
  expect_equal(mday(poslt), 1)
  expect_equal(wday(poslt), 7)
  expect_equal(yday(poslt), 1)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 59)
  expect_equal(minute(posct), 59)
  expect_equal(hour(posct), 1)
  expect_equal(mday(posct), 1)
  expect_equal(wday(posct), 7)
  expect_equal(yday(posct), 1)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 0)
  expect_equal(minute(date), 0)
  expect_equal(hour(date), 1)
  expect_equal(mday(date), 1)
  expect_equal(wday(date), 7)
  expect_equal(yday(date), 1)
  expect_equal(month(date), 01)
  expect_equal(year(date), 2011)
  expect_equal(tz(date), "UTC")
})

test_that("hours settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  hour(poslt) <- 1
  hour(posct) <- 1
  hour(date) <- 1

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

  hour(poslt) <- 70
  hour(posct) <- 70
  hour(date) <- 70

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("hours settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-14 01:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  hour(poslt) <- 2
  hour(posct) <- 2

  expect_true(is.na(poslt))
  expect_true(is.na(posct))
})

test_that("hours settor retains object class for datetimes", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)

  hour(poslt) <- 2
  hour(posct) <- 2

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")

  hour(poslt) <- 25
  hour(posct) <- 25

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
})

test_that("hours settor makes POSIXct from Date", {
  date <- as.Date("2010-02-14", tz = "GMT", format = "%Y-%m-%d")

  hour(date) <- 2

  expect_s3_class(date, "POSIXct")

  hour(date) <- 25

  expect_s3_class(date, "POSIXct")
})


test_that("mdays settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  mday(poslt) <- 10
  mday(posct) <- 10
  mday(date) <- 10

  expect_equal(mday(poslt), 10)
  expect_equal(mday(posct), 10)
  expect_equal(mday(date), 10)

  expect_equal(yday(poslt), 41)
  expect_equal(yday(posct), 41)
  expect_equal(yday(date), 41)

  expect_equal(wday(poslt), 4)
  expect_equal(wday(posct), 4)
  expect_equal(wday(date), 4)
})

test_that("mdays settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  mday(poslt) <- 32
  mday(posct) <- 32
  mday(date) <- 32

  expect_equal(second(poslt), 59)
  expect_equal(minute(poslt), 59)
  expect_equal(hour(poslt), 23)
  expect_equal(mday(poslt), 1)
  expect_equal(wday(poslt), 7)
  expect_equal(yday(poslt), 1)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 59)
  expect_equal(minute(posct), 59)
  expect_equal(hour(posct), 23)
  expect_equal(mday(posct), 1)
  expect_equal(wday(posct), 7)
  expect_equal(yday(posct), 1)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 0)
  expect_equal(minute(date), 0)
  expect_equal(hour(date), 0)
  expect_equal(mday(date), 1)
  expect_equal(wday(date), 7)
  expect_equal(yday(date), 1)
  expect_equal(month(date), 1)
  expect_equal(year(date), 2011)
  expect_equal(tz(date), "UTC")
})

test_that("mdays settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  mday(poslt) <- 1
  mday(posct) <- 1
  mday(date) <- 1

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

  mday(poslt) <- 32
  mday(posct) <- 32
  mday(date) <- 32

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("mdays settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-13 02:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  mday(poslt) <- 14
  mday(posct) <- 14

  expect_true(is.na(poslt))
  expect_true(is.na(posct))
})

test_that("mdays settor retains object class", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  mday(poslt) <- 3
  mday(posct) <- 3
  mday(date) <- 3

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
  expect_s3_class(date, "Date")

  mday(poslt) <- 32
  mday(posct) <- 32
  mday(date) <- 32

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
  expect_s3_class(date, "Date")
})



test_that("ydays settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  yday(poslt) <- 41
  yday(posct) <- 41
  yday(date) <- 41

  expect_equal(yday(poslt), 41)
  expect_equal(yday(posct), 41)
  expect_equal(yday(date), 41)

  expect_equal(mday(poslt), 10)
  expect_equal(mday(posct), 10)
  expect_equal(mday(date), 10)

  expect_equal(wday(poslt), 4)
  expect_equal(wday(posct), 4)
  expect_equal(wday(date), 4)
})

test_that("ydays settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  yday(poslt) <- 366
  yday(posct) <- 366
  yday(date) <- 366

  expect_equal(second(poslt), 59)
  expect_equal(minute(poslt), 59)
  expect_equal(hour(poslt), 23)
  expect_equal(mday(poslt), 1)
  expect_equal(wday(poslt), 7)
  expect_equal(yday(poslt), 1)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 59)
  expect_equal(minute(posct), 59)
  expect_equal(hour(posct), 23)
  expect_equal(mday(posct), 1)
  expect_equal(wday(posct), 7)
  expect_equal(yday(posct), 1)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 0)
  expect_equal(minute(date), 0)
  expect_equal(hour(date), 0)
  expect_equal(mday(date), 1)
  expect_equal(wday(date), 7)
  expect_equal(yday(date), 1)
  expect_equal(month(date), 1)
  expect_equal(year(date), 2011)
  expect_equal(tz(date), "UTC")
})

test_that("ydays settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  yday(poslt) <- 1
  yday(posct) <- 1
  yday(date) <- 1

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

  yday(poslt) <- 366
  yday(posct) <- 366
  yday(date) <- 366

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("ydays settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-13 02:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  yday(poslt) <- 73
  yday(posct) <- 73

  expect_true(is.na(poslt))
  expect_true(is.na(posct))
})

test_that("ydays settor retains object class", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  yday(poslt) <- 3
  yday(posct) <- 3
  yday(date) <- 3

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
  expect_s3_class(date, "Date")

  yday(poslt) <- 366
  yday(posct) <- 366
  yday(date) <- 366

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
  expect_s3_class(date, "Date")
})



test_that("wdays settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  wday(poslt) <- 5
  wday(posct) <- 5
  wday(date) <- 5

  expect_equal(wday(poslt), 5)
  expect_equal(wday(posct), 5)
  expect_equal(wday(date), 5)

  expect_equal(mday(poslt), 4)
  expect_equal(mday(posct), 4)
  expect_equal(mday(date), 4)

  expect_equal(mday(poslt), 4)
  expect_equal(mday(posct), 4)
  expect_equal(mday(date), 4)
})

test_that("wdays settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  wday(poslt) <- 8
  wday(posct) <- 8
  wday(date) <- 8

  expect_equal(second(poslt), 59)
  expect_equal(minute(poslt), 59)
  expect_equal(hour(poslt), 23)
  expect_equal(mday(poslt), 2)
  expect_equal(wday(poslt), 1)
  expect_equal(yday(poslt), 2)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 59)
  expect_equal(minute(posct), 59)
  expect_equal(hour(posct), 23)
  expect_equal(mday(posct), 2)
  expect_equal(wday(posct), 1)
  expect_equal(yday(posct), 2)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 0)
  expect_equal(minute(date), 0)
  expect_equal(hour(date), 0)
  expect_equal(mday(date), 2)
  expect_equal(wday(date), 1)
  expect_equal(yday(date), 2)
  expect_equal(month(date), 1)
  expect_equal(year(date), 2011)
  expect_equal(tz(date), "UTC")
})

test_that("wdays settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  wday(poslt) <- 2
  wday(posct) <- 2
  wday(date) <- 2

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

  wday(poslt) <- 8
  wday(posct) <- 8
  wday(date) <- 8

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("wdays settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-13 02:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  wday(poslt) <- 8
  wday(posct) <- 8

  expect_true(is.na(poslt))
  expect_true(is.na(posct))
})

test_that("wdays settor retains object class", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  wday(poslt) <- 3
  wday(posct) <- 3
  wday(date) <- 3

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
  expect_s3_class(date, "Date")

  wday(poslt) <- 8
  wday(posct) <- 8
  wday(date) <- 8

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
  expect_s3_class(date, "Date")
})


test_that("months settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  month(poslt) <- 5
  month(posct) <- 5
  month(date) <- 5

  expect_equal(month(poslt), 5)
  expect_equal(month(posct), 5)
  expect_equal(month(date), 5)
})

test_that("months settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  month(poslt) <- 13
  month(posct) <- 13
  month(date) <- 13

  expect_equal(second(poslt), 59)
  expect_equal(minute(poslt), 59)
  expect_equal(hour(poslt), 23)
  expect_equal(mday(poslt), 31)
  expect_equal(wday(poslt), 2)
  expect_equal(yday(poslt), 31)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 59)
  expect_equal(minute(posct), 59)
  expect_equal(hour(posct), 23)
  expect_equal(mday(posct), 31)
  expect_equal(wday(posct), 2)
  expect_equal(yday(posct), 31)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 0)
  expect_equal(minute(date), 0)
  expect_equal(hour(date), 0)
  expect_equal(mday(date), 31)
  expect_equal(wday(date), 2)
  expect_equal(yday(date), 31)
  expect_equal(month(date), 1)
  expect_equal(year(date), 2011)
  expect_equal(tz(date), "UTC")
})

test_that("months settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  month(poslt) <- 1
  month(posct) <- 1
  month(date) <- 1

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

  month(poslt) <- 13
  month(posct) <- 13
  month(date) <- 13

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("months settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-02-14 02:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  month(poslt) <- 3
  month(posct) <- 3

  expect_true(is.na(poslt))
  expect_true(is.na(posct))
})

test_that("months settor retains object class", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  month(poslt) <- 1
  month(posct) <- 1
  month(date) <- 1

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
  expect_s3_class(date, "Date")

  month(poslt) <- 13
  month(posct) <- 13
  month(date) <- 13

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
  expect_s3_class(date, "Date")
})


test_that("years settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  year(poslt) <- 2000
  year(posct) <- 2000
  year(date) <- 2000

  expect_equal(year(poslt), 2000)
  expect_equal(year(posct), 2000)
  expect_equal(year(date), 2000)
})


test_that("years settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  year(poslt) <- 2000
  year(posct) <- 2000
  year(date) <- 2000

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("years settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2009-03-14 02:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  year(poslt) <- 2010
  year(posct) <- 2010

  expect_true(is.na(poslt))
  expect_true(is.na(posct))
})

test_that("years settor retains object class", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  year(poslt) <- 2000
  year(posct) <- 2000
  year(date) <- 2000

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
  expect_s3_class(date, "Date")
})


test_that("dates settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  date(poslt) <- as.Date("2000-01-01")
  date(posct) <- as.Date("2000-01-01")
  date(date) <- as.Date("2000-01-01")

  expect_equal(date(poslt), as.Date("2000-01-01"))
  expect_equal(date(posct), as.Date("2000-01-01"))
  expect_equal(date(date), as.Date("2000-01-01"))
})

test_that("dates settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  date(poslt) <- as.Date("2000-01-01")
  date(posct) <- as.Date("2000-01-01")
  date(date) <- as.Date("2000-01-01")

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})

test_that("dates settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2009-03-14 02:59:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  date(poslt) <- as.Date("2010-03-14")
  date(posct) <- as.Date("2010-03-14")

  expect_true(is.na(poslt))
  expect_true(is.na(posct))
})

test_that("dates settor retains object class", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  date(poslt) <- as.Date("2000-01-01")
  date(posct) <- as.Date("2000-01-01")
  date(date) <- as.Date("2000-01-01")

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
  expect_s3_class(date, "Date")
})


test_that("time zone settor correctly performs simple updates", {
  posct <- ymd_hms("2010-02-03 13:45:59", tz = "UTC")
  poslt <- as.POSIXlt(posct)
  date <- as.Date(posct)

  tz(poslt) <- "GMT"
  tz(posct) <- "GMT"
  tz(date) <- "GMT"

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
  expect_s3_class(date, "POSIXct")

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "GMT")
})

test_that("time zone settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-14 02:30:59",
    tz = "UTC", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)

  tz(poslt) <- "America/New_York"
  tz(posct) <- "America/New_York"

  expect_true(is.na(poslt))
  expect_true(is.na(posct))
})

test_that("settors handle vectors", {
  poslt <- as.POSIXlt(c("2010-02-14 01:59:59", "2010-02-15 01:59:59", "2010-02-16 01:59:59"),
    tz = "UTC", format = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  second(poslt) <- 1
  second(posct) <- 1
  second(date) <- 1

  expect_equal(second(poslt), c(1, 1, 1))
  expect_equal(second(posct), c(1, 1, 1))
  expect_equal(second(date), c(1, 1, 1))

  minute(poslt) <- 1
  minute(posct) <- 1
  minute(date) <- 1

  expect_equal(minute(poslt), c(1, 1, 1))
  expect_equal(minute(posct), c(1, 1, 1))
  expect_equal(minute(date), c(1, 1, 1))

  hour(poslt) <- 1
  hour(posct) <- 1
  hour(date) <- 1

  expect_equal(hour(poslt), c(1, 1, 1))
  expect_equal(hour(posct), c(1, 1, 1))
  expect_equal(hour(date), c(1, 1, 1))

  mday(poslt) <- 1
  mday(posct) <- 1
  mday(date) <- 1

  expect_equal(mday(poslt), c(1, 1, 1))
  expect_equal(mday(posct), c(1, 1, 1))
  expect_equal(mday(date), c(1, 1, 1))

  wday(poslt) <- 1
  wday(posct) <- 1
  wday(date) <- 1

  expect_equal(wday(poslt), c(1, 1, 1))
  expect_equal(wday(posct), c(1, 1, 1))
  expect_equal(wday(date), c(1, 1, 1))

  yday(poslt) <- 1
  yday(posct) <- 1
  yday(date) <- 1

  expect_equal(yday(poslt), c(1, 1, 1))
  expect_equal(yday(posct), c(1, 1, 1))
  expect_equal(yday(date), c(1, 1, 1))

  week(poslt) <- 2
  week(posct) <- 2
  week(date) <- 2

  expect_equal(week(poslt), c(2, 2, 2))
  expect_equal(week(posct), c(2, 2, 2))
  expect_equal(week(date), c(2, 2, 2))

  month(poslt) <- 10
  month(posct) <- 10
  month(date) <- 10

  expect_equal(month(poslt), c(10, 10, 10))
  expect_equal(month(posct), c(10, 10, 10))
  expect_equal(month(date), c(10, 10, 10))

  year(poslt) <- 2001
  year(posct) <- 2001
  year(date) <- 2001

  expect_equal(year(poslt), c(2001, 2001, 2001))
  expect_equal(year(posct), c(2001, 2001, 2001))
  expect_equal(year(date), c(2001, 2001, 2001))

  date(poslt) <- as.Date("2001-01-01")
  date(posct) <- as.Date("2001-01-01")
  date(date) <- as.Date("2001-01-01")

  expect_equal(date(poslt), as.Date(c("2001-01-01", "2001-01-01", "2001-01-01")))
  expect_equal(date(posct), as.Date(c("2001-01-01", "2001-01-01", "2001-01-01")))
  expect_equal(date(date), as.Date(c("2001-01-01", "2001-01-01", "2001-01-01")))

  tz(poslt) <- "GMT"
  tz(posct) <- "GMT"
  tz(date) <- "GMT" # date has been made POSIXlt with sec, etc

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "GMT")
})

test_that("qdays settors correctly performs simple updates and rolls over as expected", {
  poslt <- as.POSIXlt(c("2010-02-14 01:59:59", "2010-04-15 01:59:59", "2010-10-16
  01:59:59"), tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  qday(poslt) <- 2
  qday(posct) <- 2
  qday(date) <- 2

  expect_equal(month(poslt), c(1, 4, 10))
  expect_equal(month(posct), c(1, 4, 10))
  expect_equal(month(date), c(1, 4, 10))

  expect_equal(mday(poslt), c(2, 2, 2))
  expect_equal(mday(posct), c(2, 2, 2))
  expect_equal(mday(date), c(2, 2, 2))

  qday(poslt) <- 95
  qday(posct) <- 95
  qday(date) <- 95

  expect_equal(month(poslt), c(4, 7, 1))
  expect_equal(month(posct), c(4, 7, 1))
  expect_equal(month(date), c(4, 7, 1))

  expect_equal(mday(poslt), c(5, 4, 3))
  expect_equal(mday(posct), c(5, 4, 3))
  expect_equal(mday(date), c(5, 4, 3))

  expect_equal(year(poslt), c(2010, 2010, 2011))
  expect_equal(year(posct), c(2010, 2010, 2011))
  expect_equal(year(date), c(2010, 2010, 2011))
})


test_that("year<- and month<- roll but produce NAs on invalid days",  {
  dd <- ymd(c("2020-02-29", "2020-03-31"))
  pp <- ymd(c("2020-02-29", "2020-03-31"), tz = "America/New_York")

  dd1 <- dd
  year(dd1) <- 2021
  expect_equal(dd1, ymd(c(NA, "2021-03-31")))

  dd1 <- dd
  month(dd1) <- 4
  expect_equal(dd1, ymd(c("2020-04-29", NA)))

  dd1 <- dd
  month(dd1) <- 15
  expect_equal(dd1, ymd(c("2021-03-29", "2021-03-31")))

  dd1 <- dd
  month(dd1) <- 14
  expect_equal(dd1, ymd(c(NA, NA)))

  dd1 <- dd
  yday(dd1) <- 370
  expect_equal(dd1, ymd(c("2021-01-04", "2021-01-04")))

  dd1 <- dd
  day(dd1) <- c(30, 32)
  expect_equal(dd1, ymd(c("2020-03-01", "2020-04-01")))

  # update recycles instead of producing NAs
  expect_equal(
    update(dd, months = 4),
    ymd(c("2020-04-29", "2020-05-01")))
  expect_equal(
    update(dd, days = 31),
    ymd(c("2020-03-02", "2020-03-31")))
  expect_equal(
    update(dd, months = 4, days = 32),
    ymd(c("2020-05-02", "2020-05-02")))

  expect_equal(
    update(pp, months = 4),
    ymd(c("2020-04-29", "2020-05-01"), tz = "America/New_York"))
  expect_equal(
    update(pp, days = 31),
    ymd(c("2020-03-02", "2020-03-31"), tz = "America/New_York"))
  expect_equal(
    update(pp, months = 4, days = 32),
    ymd(c("2020-05-02", "2020-05-02"), tz = "America/New_York"))

})

test_that("year<- update rolls over dates but not update", {

  x <- ymd("2020-02-29")
  expect_equal(update(x, years = 2019), ymd("2019-03-01"))
  year(x) <- 2017
  expect_equal(x, NA_Date_)

  x <- ymd("2020-02-29", tz = "America/New_York")
  expect_equal(update(x, years = 2019), ymd("2019-03-01", tz = "America/New_York"))
  year(x) <- 2017
  expect_equal(x, with_tz(NA_POSIXct_, "America/New_York"))

})

test_that("day<- rolls over", {

  x <- ymd("2017-02-20")
  expect_equal(update(x, days = 29), ymd("2017-03-01"))
  expect_equal(update(x, days = 100), ymd("2017-05-11"))
  day(x) <- 29
  expect_equal(x, ymd("2017-03-01"))
  x <- ymd("2017-02-20")
  day(x) <- 100
  expect_equal(x, ymd("2017-05-11"))

  x <- ymd("2017-02-20", tz = "America/New_York")
  expect_equal(update(x, mdays = 29), ymd("2017-03-01", tz = "America/New_York"))
  expect_equal(update(x, mdays = 100), ymd("2017-05-11", tz = "America/New_York"))
  day(x) <- 29
  expect_equal(x, ymd("2017-03-01", tz = "America/New_York"))
  x <- ymd("2017-02-20", tz = "America/New_York")
  day(x) <- 100
  expect_equal(x, ymd("2017-05-11", tz = "America/New_York"))

})

context("Settors")

test_that("seconds settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  second(poslt) <- 10
  second(posct) <- 10
  second(date) <- 10

  expect_equal(second(poslt), 10)
  expect_equal(second(posct), 10)
  expect_equal(second(date), 10)

})

test_that("seconds settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  second(poslt) <- 69
  second(posct) <- 69
  second(date) <- 69

  expect_equal(second(poslt), 9)
  expect_equal(minute(poslt), 0)
  expect_equal(hour(poslt), 0)
  expect_equal(mday(poslt), 1)
  expect_equal(wday(poslt), 7)
  expect_equal(yday(poslt), 1)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 9)
  expect_equal(minute(posct), 0)
  expect_equal(hour(posct), 0)
  expect_equal(mday(posct), 1)
  expect_equal(wday(posct), 7)
  expect_equal(yday(posct), 1)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 9)
  expect_equal(minute(date), 1)
  expect_equal(hour(date), 0)
  expect_equal(mday(date), 31)
  expect_equal(wday(date), 6)
  expect_equal(yday(date), 365)
  expect_equal(month(date), 12)
  expect_equal(year(date), 2010)
  expect_equal(tz(date), "UTC")

})

test_that("seconds settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  second(poslt) <- 1
  second(posct) <- 1
  second(date) <- 1

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

  second(poslt) <- 69
  second(posct) <- 69
  second(date) <- 69

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("seconds settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-14 01:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posxct <- as.POSIXct(poslt)

  second(poslt) <- 69
  second(posxct) <- 69

  expect_true(is.na(poslt))
  expect_true(is.na(posxct))

})

test_that("seconds settor retains object class for datetimes", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)

  second(poslt) <- 10
  second(posct) <- 10

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")

  second(poslt) <- 70
  second(posct) <- 70

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")

})

test_that("seconds settor makes POSTXct from Date", {
  date <- as.Date("2010-02-14", tz = "GMT", format = "%Y-%m-%d")

  second(date) <- 10

  expect_s3_class(date, "POSIXct")

  second(date) <- 70

  expect_s3_class(date, "POSIXct")

})


test_that("minutes settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  minute(poslt) <- 10
  minute(posct) <- 10
  minute(date) <- 10

  expect_equal(minute(poslt), 10)
  expect_equal(minute(posct), 10)
  expect_equal(minute(date), 10)

})

test_that("minutes settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  minute(poslt) <- 69
  minute(posct) <- 69
  minute(date) <- 69

  expect_equal(second(poslt), 59)
  expect_equal(minute(poslt), 9)
  expect_equal(hour(poslt), 0)
  expect_equal(mday(poslt), 1)
  expect_equal(wday(poslt), 7)
  expect_equal(yday(poslt), 1)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 59)
  expect_equal(minute(posct), 9)
  expect_equal(hour(posct), 0)
  expect_equal(mday(posct), 1)
  expect_equal(wday(posct), 7)
  expect_equal(yday(posct), 1)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 0)
  expect_equal(minute(date), 9)
  expect_equal(hour(date), 1)
  expect_equal(mday(date), 31)
  expect_equal(wday(date), 6)
  expect_equal(yday(date), 365)
  expect_equal(month(date), 12)
  expect_equal(year(date), 2010)
  expect_equal(tz(date), "UTC")

})

test_that("minutes settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

    minute(poslt) <- 1
  minute(posct) <- 1
  minute(date) <- 1

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

  minute(poslt) <- 70
  minute(posct) <- 70
  minute(date) <- 70

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("minutes settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-14 01:59:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  minute(poslt) <- 70
  minute(posct) <- 70

  expect_true(is.na(poslt))
  expect_true(is.na(posct))

})

test_that("minutes settor retains object class for datetimes", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)

  minute(poslt) <- 1
  minute(posct) <- 1

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")

  minute(poslt) <- 70
  minute(posct) <- 70

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
})

test_that("minutes settor makes POSIXct from Date", {
  date <- as.Date("2010-02-14", tz = "GMT", format = "%Y-%m-%d")

  minute(date) <- 70

  expect_is(date, "POSIXct")
})


test_that("hours settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  hour(poslt) <- 10
  hour(posct) <- 10
  hour(date) <- 10

  expect_equal(hour(poslt), 10)
  expect_equal(hour(posct), 10)
  expect_equal(hour(date), 10)

})

test_that("hours settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  hour(poslt) <- 25
  hour(posct) <- 25
  hour(date) <- 25

  expect_equal(second(poslt), 59)
  expect_equal(minute(poslt), 59)
  expect_equal(hour(poslt), 1)
  expect_equal(mday(poslt), 1)
  expect_equal(wday(poslt), 7)
  expect_equal(yday(poslt), 1)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 59)
  expect_equal(minute(posct), 59)
  expect_equal(hour(posct), 1)
  expect_equal(mday(posct), 1)
  expect_equal(wday(posct), 7)
  expect_equal(yday(posct), 1)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 0)
  expect_equal(minute(date), 0)
  expect_equal(hour(date), 1)
  expect_equal(mday(date), 1)
  expect_equal(wday(date), 7)
  expect_equal(yday(date), 1)
  expect_equal(month(date), 01)
  expect_equal(year(date), 2011)
  expect_equal(tz(date), "UTC")

})

test_that("hours settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  hour(poslt) <- 1
  hour(posct) <- 1
  hour(date) <- 1

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

  hour(poslt) <- 70
  hour(posct) <- 70
  hour(date) <- 70

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("hours settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-14 01:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  hour(poslt) <- 2
  hour(posct) <- 2

  expect_true(is.na(poslt))
  expect_true(is.na(posct))

})

test_that("hours settor retains object class for datetimes", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)

  hour(poslt) <- 2
  hour(posct) <- 2

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")

  hour(poslt) <- 25
  hour(posct) <- 25

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
})

test_that("hours settor makes POSIXct from Date", {
  date <- as.Date("2010-02-14", tz = "GMT", format = "%Y-%m-%d")

  hour(date) <- 2

  expect_s3_class(date, "POSIXct")

  hour(date) <- 25

  expect_s3_class(date, "POSIXct")
})


test_that("mdays settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  mday(poslt) <- 10
  mday(posct) <- 10
  mday(date) <- 10

  expect_equal(mday(poslt), 10)
  expect_equal(mday(posct), 10)
  expect_equal(mday(date), 10)

  expect_equal(yday(poslt), 41)
  expect_equal(yday(posct), 41)
  expect_equal(yday(date), 41)

  expect_equal(wday(poslt), 4)
  expect_equal(wday(posct), 4)
  expect_equal(wday(date), 4)

})

test_that("mdays settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59", tz = "UTC", format
                      = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  mday(poslt) <- 32
  mday(posct) <- 32
  mday(date) <- 32

  expect_equal(second(poslt), 59)
  expect_equal(minute(poslt), 59)
  expect_equal(hour(poslt), 23)
  expect_equal(mday(poslt), 1)
  expect_equal(wday(poslt), 7)
  expect_equal(yday(poslt), 1)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 59)
  expect_equal(minute(posct), 59)
  expect_equal(hour(posct), 23)
  expect_equal(mday(posct), 1)
  expect_equal(wday(posct), 7)
  expect_equal(yday(posct), 1)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 0)
  expect_equal(minute(date), 0)
  expect_equal(hour(date), 0)
  expect_equal(mday(date), 1)
  expect_equal(wday(date), 7)
  expect_equal(yday(date), 1)
  expect_equal(month(date), 1)
  expect_equal(year(date), 2011)
  expect_equal(tz(date), "UTC")

})

test_that("mdays settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  mday(poslt) <- 1
  mday(posct) <- 1
  mday(date) <- 1

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

  mday(poslt) <- 32
  mday(posct) <- 32
  mday(date) <- 32

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("mdays settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-13 02:59:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  mday(poslt) <- 14
  mday(posct) <- 14

  expect_true(is.na(poslt))
  expect_true(is.na(posct))

})

test_that("mdays settor retains object class", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  mday(poslt) <- 3
  mday(posct) <- 3
  mday(date) <- 3

  expect_is(poslt, "POSIXlt")
  expect_is(posct, "POSIXct")
  expect_is(date, "Date")

  mday(poslt) <- 32
  mday(posct) <- 32
  mday(date) <- 32

  expect_is(poslt, "POSIXlt")
  expect_is(posct, "POSIXct")
  expect_is(date, "Date")

})



test_that("ydays settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  yday(poslt) <- 41
  yday(posct) <- 41
  yday(date) <- 41

  expect_equal(yday(poslt), 41)
  expect_equal(yday(posct), 41)
  expect_equal(yday(date), 41)

  expect_equal(mday(poslt), 10)
  expect_equal(mday(posct), 10)
  expect_equal(mday(date), 10)

  expect_equal(wday(poslt), 4)
  expect_equal(wday(posct), 4)
  expect_equal(wday(date), 4)
})

test_that("ydays settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  yday(poslt) <- 366
  yday(posct) <- 366
  yday(date) <- 366

  expect_equal(second(poslt), 59)
  expect_equal(minute(poslt), 59)
  expect_equal(hour(poslt), 23)
  expect_equal(mday(poslt), 1)
  expect_equal(wday(poslt), 7)
  expect_equal(yday(poslt), 1)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 59)
  expect_equal(minute(posct), 59)
  expect_equal(hour(posct), 23)
  expect_equal(mday(posct), 1)
  expect_equal(wday(posct), 7)
  expect_equal(yday(posct), 1)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 0)
  expect_equal(minute(date), 0)
  expect_equal(hour(date), 0)
  expect_equal(mday(date), 1)
  expect_equal(wday(date), 7)
  expect_equal(yday(date), 1)
  expect_equal(month(date), 1)
  expect_equal(year(date), 2011)
  expect_equal(tz(date), "UTC")

})

test_that("ydays settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  yday(poslt) <- 1
  yday(posct) <- 1
  yday(date) <- 1

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

  yday(poslt) <- 366
  yday(posct) <- 366
  yday(date) <- 366

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("ydays settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-13 02:59:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  yday(poslt) <- 73
  yday(posct) <- 73

  expect_true(is.na(poslt))
  expect_true(is.na(posct))

})

test_that("ydays settor retains object class", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  yday(poslt) <- 3
  yday(posct) <- 3
  yday(date) <- 3

  expect_is(poslt, "POSIXlt")
  expect_is(posct, "POSIXct")
  expect_is(date, "Date")

  yday(poslt) <- 366
  yday(posct) <- 366
  yday(date) <- 366

  expect_is(poslt, "POSIXlt")
  expect_is(posct, "POSIXct")
  expect_is(date, "Date")

})



test_that("wdays settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  wday(poslt) <- 5
  wday(posct) <- 5
  wday(date) <- 5

  expect_equal(wday(poslt), 5)
  expect_equal(wday(posct), 5)
  expect_equal(wday(date), 5)

  expect_equal(mday(poslt), 4)
  expect_equal(mday(posct), 4)
  expect_equal(mday(date), 4)

  expect_equal(mday(poslt), 4)
  expect_equal(mday(posct), 4)
  expect_equal(mday(date), 4)

})

test_that("wdays settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  wday(poslt) <- 8
  wday(posct) <- 8
  wday(date) <- 8

  expect_equal(second(poslt), 59)
  expect_equal(minute(poslt), 59)
  expect_equal(hour(poslt), 23)
  expect_equal(mday(poslt), 2)
  expect_equal(wday(poslt), 1)
  expect_equal(yday(poslt), 2)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 59)
  expect_equal(minute(posct), 59)
  expect_equal(hour(posct), 23)
  expect_equal(mday(posct), 2)
  expect_equal(wday(posct), 1)
  expect_equal(yday(posct), 2)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 0)
  expect_equal(minute(date), 0)
  expect_equal(hour(date), 0)
  expect_equal(mday(date), 2)
  expect_equal(wday(date), 1)
  expect_equal(yday(date), 2)
  expect_equal(month(date), 1)
  expect_equal(year(date), 2011)
  expect_equal(tz(date), "UTC")

})

test_that("wdays settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  wday(poslt) <- 2
  wday(posct) <- 2
  wday(date) <- 2

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

  wday(poslt) <- 8
  wday(posct) <- 8
  wday(date) <- 8

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("wdays settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-13 02:59:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  wday(poslt) <- 8
  wday(posct) <- 8

  expect_true(is.na(poslt))
  expect_true(is.na(posct))

})

test_that("wdays settor retains object class", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  wday(poslt) <- 3
  wday(posct) <- 3
  wday(date) <- 3

  expect_is(poslt, "POSIXlt")
  expect_is(posct, "POSIXct")
  expect_is(date, "Date")

  wday(poslt) <- 8
  wday(posct) <- 8
  wday(date) <- 8

  expect_is(poslt, "POSIXlt")
  expect_is(posct, "POSIXct")
  expect_is(date, "Date")

})


test_that("months settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  month(poslt) <- 5
  month(posct) <- 5
  month(date) <- 5

  expect_equal(month(poslt), 5)
  expect_equal(month(posct), 5)
  expect_equal(month(date), 5)
})

test_that("months settor rolls over as expected", {
  poslt <- as.POSIXlt("2010-12-31 23:59:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

    month(poslt) <- 13
  month(posct) <- 13
  month(date) <- 13

  expect_equal(second(poslt), 59)
  expect_equal(minute(poslt), 59)
  expect_equal(hour(poslt), 23)
  expect_equal(mday(poslt), 31)
  expect_equal(wday(poslt), 2)
  expect_equal(yday(poslt), 31)
  expect_equal(month(poslt), 1)
  expect_equal(year(poslt), 2011)
  expect_equal(tz(poslt), "UTC")

  expect_equal(second(posct), 59)
  expect_equal(minute(posct), 59)
  expect_equal(hour(posct), 23)
  expect_equal(mday(posct), 31)
  expect_equal(wday(posct), 2)
  expect_equal(yday(posct), 31)
  expect_equal(month(posct), 1)
  expect_equal(year(posct), 2011)
  expect_equal(tz(posct), "UTC")

  expect_equal(second(date), 0)
  expect_equal(minute(date), 0)
  expect_equal(hour(date), 0)
  expect_equal(mday(date), 31)
  expect_equal(wday(date), 2)
  expect_equal(yday(date), 31)
  expect_equal(month(date), 1)
  expect_equal(year(date), 2011)
  expect_equal(tz(date), "UTC")

})

test_that("months settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

    month(poslt) <- 1
  month(posct) <- 1
  month(date) <- 1

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")

    month(poslt) <- 13
  month(posct) <- 13
  month(date) <- 13

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("months settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-02-14 02:59:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  month(poslt) <- 3
  month(posct) <- 3

  expect_true(is.na(poslt))
  expect_true(is.na(posct))
})

test_that("months settor retains object class", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  month(poslt) <- 1
  month(posct) <- 1
  month(date) <- 1

  expect_is(poslt, "POSIXlt")
  expect_is(posct, "POSIXct")
  expect_is(date, "Date")

  month(poslt) <- 13
  month(posct) <- 13
  month(date) <- 13

  expect_is(poslt, "POSIXlt")
  expect_is(posct, "POSIXct")
  expect_is(date, "Date")

})


test_that("years settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  year(poslt) <- 2000
  year(posct) <- 2000
  year(date) <- 2000

  expect_equal(year(poslt), 2000)
  expect_equal(year(posct), 2000)
  expect_equal(year(date), 2000)
})


test_that("years settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  year(poslt) <- 2000
  year(posct) <- 2000
  year(date) <- 2000

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})


test_that("years settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2009-03-14 02:59:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  year(poslt) <- 2010
  year(posct) <- 2010

  expect_true(is.na(poslt))
  expect_true(is.na(posct))

})

test_that("years settor retains object class", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  year(poslt) <- 2000
  year(posct) <- 2000
  year(date) <- 2000

  expect_is(poslt, "POSIXlt")
  expect_is(posct, "POSIXct")
  expect_is(date, "Date")
})


test_that("dates settor correctly performs simple updates", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  date(poslt) <- as.Date("2000-01-01")
  date(posct) <- as.Date("2000-01-01")
  date(date) <- as.Date("2000-01-01")

  expect_equal(date(poslt), as.Date("2000-01-01"))
  expect_equal(date(posct), as.Date("2000-01-01"))
  expect_equal(date(date), as.Date("2000-01-01"))
})

test_that("dates settor does not change time zone", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  date(poslt) <- as.Date("2000-01-01")
  date(posct) <- as.Date("2000-01-01")
  date(date) <- as.Date("2000-01-01")

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "UTC")
})

test_that("dates settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2009-03-14 02:59:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  posct <- as.POSIXct(poslt)

  date(poslt) <- as.Date("2010-03-14")
  date(posct) <- as.Date("2010-03-14")

  expect_true(is.na(poslt))
  expect_true(is.na(posct))
})

test_that("dates settor retains object class", {
  poslt <- as.POSIXlt("2010-02-14 01:59:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  date(poslt) <- as.Date("2000-01-01")
  date(posct) <- as.Date("2000-01-01")
  date(date) <- as.Date("2000-01-01")

  expect_is(poslt, "POSIXlt")
  expect_is(posct, "POSIXct")
  expect_is(date, "Date")
})


test_that("time zone settor correctly performs simple updates", {
  posct <- ymd_hms("2010-02-03 13:45:59", tz = "UTC")
  poslt <- as.POSIXlt(posct)
  date <- as.Date(posct)

  tz(poslt) <- "GMT"
  tz(posct) <- "GMT"
  tz(date) <- "GMT"

  expect_s3_class(poslt, "POSIXlt")
  expect_s3_class(posct, "POSIXct")
  expect_s3_class(date, "POSIXct")

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "GMT")
})

test_that("time zone settor returns NA for spring dst gap", {
  poslt <- as.POSIXlt("2010-03-14 02:30:59", tz = "UTC", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)

  tz(poslt) <- "America/New_York"
  tz(posct) <- "America/New_York"

  expect_true(is.na(poslt))
  expect_true(is.na(posct))

})

test_that("settors handle vectors", {
  poslt <- as.POSIXlt(c("2010-02-14 01:59:59", "2010-02-15 01:59:59", "2010-02-16 01:59:59"),
                      tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  second(poslt) <- 1
  second(posct) <- 1
  second(date) <- 1

  expect_equal(second(poslt), c(1, 1, 1))
  expect_equal(second(posct), c(1, 1, 1))
  expect_equal(second(date), c(1, 1, 1))

  minute(poslt) <- 1
  minute(posct) <- 1
  minute(date) <- 1

  expect_equal(minute(poslt), c(1, 1, 1))
  expect_equal(minute(posct), c(1, 1, 1))
  expect_equal(minute(date), c(1, 1, 1))

  hour(poslt) <- 1
  hour(posct) <- 1
  hour(date) <- 1

  expect_equal(hour(poslt), c(1, 1, 1))
  expect_equal(hour(posct), c(1, 1, 1))
  expect_equal(hour(date), c(1, 1, 1))

  mday(poslt) <- 1
  mday(posct) <- 1
  mday(date) <- 1

  expect_equal(mday(poslt), c(1, 1, 1))
  expect_equal(mday(posct), c(1, 1, 1))
  expect_equal(mday(date), c(1, 1, 1))

  wday(poslt) <- 1
  wday(posct) <- 1
  wday(date) <- 1

  expect_equal(wday(poslt), c(1, 1, 1))
  expect_equal(wday(posct), c(1, 1, 1))
  expect_equal(wday(date), c(1, 1, 1))

  yday(poslt) <- 1
  yday(posct) <- 1
  yday(date) <- 1

  expect_equal(yday(poslt), c(1, 1, 1))
  expect_equal(yday(posct), c(1, 1, 1))
  expect_equal(yday(date), c(1, 1, 1))

  week(poslt) <- 2
  week(posct) <- 2
  week(date) <- 2

  expect_equal(week(poslt), c(2, 2, 2))
  expect_equal(week(posct), c(2, 2, 2))
  expect_equal(week(date), c(2, 2, 2))

  month(poslt) <- 10
  month(posct) <- 10
  month(date) <- 10

  expect_equal(month(poslt), c(10, 10, 10))
  expect_equal(month(posct), c(10, 10, 10))
  expect_equal(month(date), c(10, 10, 10))

  year(poslt) <- 2001
  year(posct) <- 2001
  year(date) <- 2001

  expect_equal(year(poslt), c(2001, 2001, 2001))
  expect_equal(year(posct), c(2001, 2001, 2001))
  expect_equal(year(date), c(2001, 2001, 2001))

  date(poslt) <- as.Date("2001-01-01")
  date(posct) <- as.Date("2001-01-01")
  date(date) <- as.Date("2001-01-01")

  expect_equal(date(poslt), as.Date(c("2001-01-01", "2001-01-01", "2001-01-01")))
  expect_equal(date(posct), as.Date(c("2001-01-01", "2001-01-01", "2001-01-01")))
  expect_equal(date(date), as.Date(c("2001-01-01", "2001-01-01", "2001-01-01")))

  tz(poslt) <- "GMT"
  tz(posct) <- "GMT"
  tz(date) <- "GMT" # date has been made POSIXlt with sec, etc

  expect_match(tz(poslt), "GMT")
  expect_match(tz(posct), "GMT")
  expect_match(tz(date), "GMT")
  })

test_that("qdays settors correctly performs simple updates and rolls over as expected", {
  poslt <- as.POSIXlt(c("2010-02-14 01:59:59", "2010-04-15 01:59:59", "2010-10-16
  01:59:59"), tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)

  qday(poslt) <- 2
  qday(posct) <- 2
  qday(date) <- 2

  expect_equal(month(poslt), c(1, 4, 10))
  expect_equal(month(posct), c(1, 4, 10))
  expect_equal(month(date), c(1, 4, 10))

  expect_equal(mday(poslt), c(2, 2, 2))
  expect_equal(mday(posct), c(2, 2, 2))
  expect_equal(mday(date), c(2, 2, 2))

  qday(poslt) <- 95
  qday(posct) <- 95
  qday(date) <- 95

  expect_equal(month(poslt), c(4, 7, 1))
  expect_equal(month(posct), c(4, 7, 1))
  expect_equal(month(date), c(4, 7, 1))

  expect_equal(mday(poslt), c(5, 4, 3))
  expect_equal(mday(posct), c(5, 4, 3))
  expect_equal(mday(date), c(5, 4, 3))

  expect_equal(year(poslt), c(2010, 2010, 2011))
  expect_equal(year(posct), c(2010, 2010, 2011))
  expect_equal(year(date), c(2010, 2010, 2011))
})

test_that("update.Date returns a date object", {
  date <- as.Date("05/05/2010", "%m/%d/%Y")
  expect_s3_class(update(date, days = 1), "Date")
  expect_s3_class(update(date, ydays = 1), "Date")
  expect_s3_class(update(date, mdays = 1), "Date")
  expect_s3_class(update(date, wdays = 1), "Date")
  expect_s3_class(update(date, months = 1), "Date")
  expect_s3_class(update(date, years = 2001), "Date")
  ## expect_s3_class(update(date, tzs = "UTC"), "POSIXct")
  tz(date) <- "UTC"
  expect_s3_class(date, "POSIXct")
})

test_that("update.Date returns a posix object if time is manipulated", {
  date <- as.Date("05/05/2010", "%m/%d/%Y")
  expect_s3_class(update(date, seconds = 1), "POSIXt")
  expect_s3_class(update(date, minutes = 1), "POSIXt")
  expect_s3_class(update(date, hours = 1), "POSIXt")
})

test_that("update.POSIXlt returns a POSIXlt object", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  expect_s3_class(update(poslt, seconds = 1), "POSIXlt")
  expect_s3_class(update(poslt, minutes = 1), "POSIXlt")
  expect_s3_class(update(poslt, hours = 1), "POSIXlt")
  expect_s3_class(update(poslt, days = 1), "POSIXlt")
  expect_s3_class(update(poslt, ydays = 1), "POSIXlt")
  expect_s3_class(update(poslt, mdays = 1), "POSIXlt")
  expect_s3_class(update(poslt, wdays = 1), "POSIXlt")
  expect_s3_class(update(poslt, months = 1), "POSIXlt")
  expect_s3_class(update(poslt, years = 2001), "POSIXlt")
  expect_s3_class(update(poslt, tzs = "UTC"), "POSIXlt")
})

test_that("update.POSIXct returns a POSIXct object", {
  posct <- as.POSIXct("2010-02-03 13:45:59",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  expect_s3_class(update(posct, seconds = 1), "POSIXct")
  expect_s3_class(update(posct, minutes = 1), "POSIXct")
  expect_s3_class(update(posct, hours = 1), "POSIXct")
  expect_s3_class(update(posct, days = 1), "POSIXct")
  expect_s3_class(update(posct, ydays = 1), "POSIXct")
  expect_s3_class(update(posct, mdays = 1), "POSIXct")
  expect_s3_class(update(posct, wdays = 1), "POSIXct")
  expect_s3_class(update(posct, months = 1), "POSIXct")
  expect_s3_class(update(posct, years = 2001), "POSIXct")
  expect_s3_class(update(posct, tzs = "UTC"), "POSIXct")
})

test_that("update.Date performs simple operation as expected", {
  date <- as.Date("05/05/2010", "%m/%d/%Y")
  expect_equal(second(update(date, seconds = 1)), 1)
  expect_equal(minute(update(date, minutes = 1)), 1)
  expect_equal(hour(update(date, hours = 1)), 1)
  expect_equal(mday(update(date, mdays = 1)), 1)
  expect_equal(wday(update(date, mdays = 1)), 7)
  expect_equal(yday(update(date, mdays = 1)), 121)
  expect_equal(yday(update(date, ydays = 1)), 1)
  expect_equal(mday(update(date, ydays = 1)), 1)
  expect_equal(wday(update(date, ydays = 1)), 6)
  expect_equal(wday(update(date, wdays = 1)), 1)
  expect_equal(yday(update(date, wdays = 1)), 122)
  expect_equal(mday(update(date, wdays = 1)), 2)
  expect_equal(month(update(date, months = 1)), 1)
  expect_equal(year(update(date, years = 2000)), 2000)
  expect_match(tz(update(date, tzs = "UTC")), "UTC")
})

test_that("update.POSIXt performs simple operation as expected", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct("2010-02-03 13:45:59", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  expect_equal(second(update(poslt, seconds = 1)), 1)
  expect_equal(minute(update(poslt, minutes = 1)), 1)
  expect_equal(hour(update(poslt, hours = 1)), 1)
  expect_equal(mday(update(poslt, mdays = 1)), 1)
  expect_equal(wday(update(poslt, mdays = 1)), 2)
  expect_equal(yday(update(poslt, mdays = 1)), 32)
  expect_equal(yday(update(poslt, ydays = 1)), 1)
  expect_equal(mday(update(poslt, ydays = 1)), 1)
  expect_equal(wday(update(poslt, ydays = 1)), 6)
  expect_equal(wday(update(poslt, wdays = 1)), 1)
  expect_equal(yday(update(poslt, wdays = 1)), 31)
  expect_equal(mday(update(poslt, wdays = 1)), 31)
  expect_equal(month(update(poslt, months = 1)), 1)
  expect_equal(year(update(poslt, years = 2000)), 2000)
  expect_match(tz(update(poslt, tzs = "UTC")), "UTC")
  expect_equal(second(update(posct, seconds = 1)), 1)
  expect_equal(minute(update(posct, minutes = 1)), 1)
  expect_equal(hour(update(posct, hours = 1)), 1)
  expect_equal(mday(update(posct, mdays = 1)), 1)
  expect_equal(wday(update(posct, mdays = 1)), 2)
  expect_equal(yday(update(posct, mdays = 1)), 32)
  expect_equal(yday(update(posct, ydays = 1)), 1)
  expect_equal(mday(update(posct, ydays = 1)), 1)
  expect_equal(wday(update(posct, ydays = 1)), 6)
  expect_equal(wday(update(posct, wdays = 1)), 1)
  expect_equal(yday(update(posct, wdays = 1)), 31)
  expect_equal(mday(update(posct, wdays = 1)), 31)
  expect_equal(month(update(posct, months = 1)), 1)
  expect_equal(year(update(posct, years = 2000)), 2000)
  expect_match(tz(update(posct, tzs = "UTC")), "UTC")
})

test_that("update.POSIXt works on wdays", {
  date <- ymd("2017-05-07") ## sunday
  ct <- as.POSIXct("2010-02-03 13:45:59", tz = "America/New_York", format = "%Y-%m-%d %H:%M:%S") ## Wednesday
  expect_equal(wday(update(ct, wdays = 1)), 1)
  expect_equal(wday(update(ct, wdays = 2)), 2)
  expect_equal(wday(update(ct, wdays = 5)), 5)
  expect_equal(wday(update(ct, wdays = 7)), 7)
  expect_equal(wday(update(date, wdays = 1)), 1)
  expect_equal(wday(update(date, wdays = 2)), 2)
  expect_equal(wday(update(date, wdays = 5)), 5)
  expect_equal(wday(update(date, wdays = 7)), 7)

  ws <- 1
  expect_equal(wday(update(ct, wdays = 1, week_start = ws)), 2)
  expect_equal(wday(update(ct, wdays = 2, week_start = ws)), 3)
  expect_equal(wday(update(ct, wdays = 5, week_start = ws)), 6)
  expect_equal(wday(update(ct, wdays = 7, week_start = ws)), 1)
  expect_equal(wday(update(date, wdays = 1, week_start = ws)), 2)
  expect_equal(wday(update(date, wdays = 2, week_start = ws)), 3)
  expect_equal(wday(update(date, wdays = 5, week_start = ws)), 6)
  expect_equal(wday(update(date, wdays = 7, week_start = ws)), 1)

  ws <- 1
  expect_equal(wday(update(ct, wdays = 1, week_start = ws), week_start = ws), 1)
  expect_equal(wday(update(ct, wdays = 2, week_start = ws), week_start = ws), 2)
  expect_equal(wday(update(ct, wdays = 5, week_start = ws), week_start = ws), 5)
  expect_equal(wday(update(ct, wdays = 7, week_start = ws), week_start = ws), 7)
  expect_equal(wday(update(date, wdays = 1, week_start = ws), week_start = ws), 1)
  expect_equal(wday(update(date, wdays = 2, week_start = ws), week_start = ws), 2)
  expect_equal(wday(update(date, wdays = 5, week_start = ws), week_start = ws), 5)
  expect_equal(wday(update(date, wdays = 7, week_start = ws), week_start = ws), 7)

  ws <- 3
  expect_equal(wday(update(ct, wdays = 1, week_start = ws), week_start = ws), 1)
  expect_equal(wday(update(ct, wdays = 2, week_start = ws), week_start = ws), 2)
  expect_equal(wday(update(ct, wdays = 5, week_start = ws), week_start = ws), 5)
  expect_equal(wday(update(ct, wdays = 7, week_start = ws), week_start = ws), 7)
  expect_equal(wday(update(date, wdays = 1, week_start = ws), week_start = ws), 1)
  expect_equal(wday(update(date, wdays = 2, week_start = ws), week_start = ws), 2)
  expect_equal(wday(update(date, wdays = 5, week_start = ws), week_start = ws), 5)
  expect_equal(wday(update(date, wdays = 7, week_start = ws), week_start = ws), 7)
})

test_that("updates on ydays works correctly with leap years", {
  expect_equal(
    update(ymd("15-02-03", tz = "UTC"), years = 2000, ydays = 1),
    ymd("2000-01-01", tz = "UTC")
  )
  expect_equal(
    update(ymd("15-02-03", tz = "UTC"), years = 2015, ydays = 1),
    ymd("2015-01-01", tz = "UTC")
  )
  expect_equal(
    update(ymd("15-02-03", tz = "UTC"), years = 2016, ydays = 10),
    ymd("2016-01-10", tz = "UTC")
  )
  expect_equal(
    update(ymd("15-02-03", tz = "America/New_York"), years = 2000, ydays = 1),
    ymd("2000-01-01", tz = "America/New_York")
  )
  expect_equal(
    update(ymd("15-02-03", tz = "America/New_York"), years = 2015, ydays = 1),
    ymd("2015-01-01", tz = "America/New_York")
  )
  expect_equal(
    update(ymd("15-02-03", tz = "America/New_York"), years = 2016, ydays = 10),
    ymd("2016-01-10", tz = "America/New_York")
  )

  expect_equal(
    update(ymd(c("2016-02-29", "2016-03-01")), ydays = 1),
    ymd(c("2016-01-01", "2016-01-01"))
  )
  expect_equal(
    update(ymd(c("2016-02-29", "2016-03-01"), tz = "America/New_York"), ydays = 1),
    ymd(c("2016-01-01", "2016-01-01"), tz = "America/New_York")
  )
  expect_equal(
    update(ymd_hms(c("2016-02-29 1:2:3", "2016-03-01 10:20:30")), ydays = 1),
    ymd_hms(c("2016-01-01 1:2:3", "2016-01-01 10:20:30"))
  )
  expect_equal(
    update(ymd_hms(c("2016-02-29 1:2:3", "2016-03-01 10:20:30"), tz = "America/New_York"), ydays = 1),
    ymd_hms(c("2016-01-01 1:2:3", "2016-01-01 10:20:30"), tz = "America/New_York")
  )
})


test_that("update performs roll overs correctly for Date objects", {
  date <- as.Date("05/05/2010", "%m/%d/%Y")
  expect_equal(second(update(date, seconds = 61)), 1)
  expect_equal(minute(update(date, seconds = 61)), 1)
  expect_equal(minute(update(date, minutes = 61)), 1)
  expect_equal(hour(update(date, minutes = 61)), 1)
  expect_equal(hour(update(date, hours = 25)), 1)
  expect_equal(mday(update(date, hours = 25)), 6)
  expect_equal(yday(update(date, hours = 25)), 126)
  expect_equal(wday(update(date, hours = 25)), 5)
  expect_equal(mday(update(date, mdays = 32)), 1)
  expect_equal(month(update(date, mdays = 32)), 6)
  expect_equal(wday(update(date, wdays = 31)), 3)
  expect_equal(month(update(date, wdays = 31)), 6)
  expect_equal(yday(update(date, ydays = 366)), 1)
  expect_equal(month(update(date, ydays = 366)), 1)
  expect_equal(month(update(date, months = 13)), 1)
  expect_equal(year(update(date, months = 13)), 2011)
  expect_match(tz(update(date, months = 13)), "UTC")
})

test_that("update performs roll overs correctly for POSIXlt objects", {
  poslt <- as.POSIXlt("2010-05-05 00:00:00", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  expect_equal(second(update(poslt, seconds = 61)), 1)
  expect_equal(minute(update(poslt, seconds = 61)), 1)
  expect_equal(minute(update(poslt, minutes = 61)), 1)
  expect_equal(hour(update(poslt, minutes = 61)), 1)
  expect_equal(hour(update(poslt, hours = 25)), 1)
  expect_equal(mday(update(poslt, hours = 25)), 6)
  expect_equal(yday(update(poslt, hours = 25)), 126)
  expect_equal(wday(update(poslt, hours = 25)), 5)
  expect_equal(mday(update(poslt, mdays = 32)), 1)
  expect_equal(month(update(poslt, mdays = 32)), 6)
  expect_equal(wday(update(poslt, wdays = 31)), 3)
  expect_equal(month(update(poslt, wdays = 31)), 6)
  expect_equal(yday(update(poslt, ydays = 366)), 1)
  expect_equal(month(update(poslt, ydays = 366)), 1)
  expect_equal(month(update(poslt, months = 13)), 1)
  expect_equal(year(update(poslt, months = 13)), 2011)
  expect_match(tz(update(poslt, months = 13)), "GMT")
})

test_that("update performs roll overs correctly for POSIXct objects", {
  posct <- as.POSIXct("2010-05-05 00:00:00", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  expect_equal(second(update(posct, seconds = 61)), 1)
  expect_equal(minute(update(posct, seconds = 61)), 1)
  expect_equal(minute(update(posct, minutes = 61)), 1)
  expect_equal(hour(update(posct, minutes = 61)), 1)
  expect_equal(hour(update(posct, hours = 25)), 1)
  expect_equal(mday(update(posct, hours = 25)), 6)
  expect_equal(yday(update(posct, hours = 25)), 126)
  expect_equal(wday(update(posct, hours = 25)), 5)
  expect_equal(mday(update(posct, mdays = 32)), 1)
  expect_equal(month(update(posct, mdays = 32)), 6)
  expect_equal(wday(update(posct, wdays = 31)), 3)
  expect_equal(month(update(posct, wdays = 31)), 6)
  expect_equal(yday(update(posct, ydays = 366)), 1)
  expect_equal(month(update(posct, ydays = 366)), 1)
  expect_equal(month(update(posct, months = 13)), 1)
  expect_equal(year(update(posct, months = 13)), 2011)
  expect_match(tz(update(posct, months = 13)), "GMT")
})

test_that("update performs consecutive roll overs correctly for
  Date objects regardless of order", {
  date <- update(as.Date("11/01/2010", "%m/%d/%Y"),
    months = 13, days = 32, hours = 25, minutes = 61, seconds
    = 61
  )
  expect_equal(second(date), 1)
  expect_equal(minute(date), 2)
  expect_equal(hour(date), 2)
  expect_equal(mday(date), 2)
  expect_equal(wday(date), 4)
  expect_equal(yday(date), 33)
  expect_equal(month(date), 2)
  expect_equal(year(date), 2011)
  expect_match(tz(date), "UTC")
  date2 <- update(as.Date("11/01/2010", "%m/%d/%Y"),
    seconds = 61, minutes = 61, hours = 25, days = 32, months = 13
  )
  expect_equal(second(date2), 1)
  expect_equal(minute(date2), 2)
  expect_equal(hour(date2), 2)
  expect_equal(mday(date2), 2)
  expect_equal(wday(date2), 4)
  expect_equal(yday(date2), 33)
  expect_equal(month(date2), 2)
  expect_equal(year(date2), 2011)
  expect_match(tz(date2), "UTC")
})

test_that("update performs consecutive roll overs correctly for
  POSIXlt objects", {
  posl <- as.POSIXlt("2010-11-01 00:00:00",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  poslt <- update(posl,
    months = 13, days = 32, hours = 25,
    minutes = 61, seconds = 61
  )

  expect_equal(second(poslt), 1)
  expect_equal(minute(poslt), 2)
  expect_equal(hour(poslt), 2)
  expect_equal(mday(poslt), 2)
  expect_equal(wday(poslt), 4)
  expect_equal(yday(poslt), 33)
  expect_equal(month(poslt), 2)
  expect_equal(year(poslt), 2011)
  expect_match(tz(poslt), "GMT")

  poslt2 <- update(posl,
    seconds = 61, minutes = 61, hours = 25,
    days = 32, months = 13
  )

  expect_equal(second(poslt2), 1)
  expect_equal(minute(poslt2), 2)
  expect_equal(hour(poslt2), 2)
  expect_equal(mday(poslt2), 2)
  expect_equal(wday(poslt2), 4)
  expect_equal(yday(poslt2), 33)
  expect_equal(month(poslt2), 2)
  expect_equal(year(poslt2), 2011)
  expect_match(tz(poslt2), "GMT")
})

test_that("update performs consecutive roll overs correctly for POSIXct objects", {
  posc <- as.POSIXct("2010-11-01 00:00:00",
    tz = "GMT", format
    = "%Y-%m-%d %H:%M:%S"
  )
  posct <- update(posc,
    months = 13, days = 32, hours = 25,
    minutes = 61, seconds = 61
  )

  expect_equal(second(posct), 1)
  expect_equal(minute(posct), 2)
  expect_equal(hour(posct), 2)
  expect_equal(mday(posct), 2)
  expect_equal(wday(posct), 4)
  expect_equal(yday(posct), 33)
  expect_equal(month(posct), 2)
  expect_equal(year(posct), 2011)
  expect_match(tz(posct), "GMT")

  posct2 <- update(posc,
    seconds = 61, minutes = 61, hours = 25,
    days = 32, months = 13
  )

  expect_equal(second(posct2), 1)
  expect_equal(minute(posct2), 2)
  expect_equal(hour(posct2), 2)
  expect_equal(mday(posct2), 2)
  expect_equal(wday(posct2), 4)
  expect_equal(yday(posct2), 33)
  expect_equal(month(posct2), 2)
  expect_equal(year(posct2), 2011)
  expect_match(tz(posct2), "GMT")
})

test_that("update returns NA for date-times in the spring dst gap", {
  poslt <- as.POSIXlt("2010-03-14 01:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  expect_true(is.na(update(poslt, seconds = 65)))
  expect_true(is.na(update(poslt, minutes = 65)))
  expect_true(is.na(update(poslt, hours = 2)))
  poslt <- as.POSIXlt("2010-03-13 02:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  expect_true(is.na(update(poslt, mdays = 14)))
  expect_true(is.na(update(poslt, wdays = 8)))
  expect_true(is.na(update(poslt, ydays = 73)))
  poslt <- as.POSIXlt("2010-02-14 02:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  expect_true(is.na(update(poslt, months = 3)))
  poslt <- as.POSIXlt("2009-03-14 02:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  expect_true(is.na(update(poslt, years = 2010)))
  poslt <- as.POSIXlt("2010-03-14 02:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  expect_true(is.na(update(poslt, tzs = "America/New_York")))
})

test_that("update handles repeated DST transitions", {
  repref <- ymd_hms("2022-10-30 01:00:00", tz = "Europe/Amsterdam")
  rep <- ymd_hms("2022-10-29 00:01:02", tz = "Europe/Amsterdam")

  expect_equal(update(rep, days = 30, hours = 2, minutes = 3, roll_dst = "pre"),
    repref + 3600 + 3*60 + 2)
  expect_equal(update(rep, days = 30, hours = 2, minutes = 3, roll_dst = "boundary"),
    repref + 2*3600)
  expect_equal(update(rep, days = 30, hours = 2, minutes = 3, roll_dst = "post"),
    repref + 2*3600 + 3*60 + 2)
  expect_equal(update(rep, days = 30, hours = 2, seconds = 3.35, roll_dst = "pre"),
    repref + 3600 + 60 + 3.35)
  expect_equal(update(rep, days = 30, hours = 2, seconds = 3.35, roll_dst = "boundary"),
    repref + 2*3600)
  expect_equal(update(rep, days = 30, hours = 2, seconds = 3.35, roll_dst = "post"),
    repref + 2*3600 + 60 + 3.35)

  ## our default roll_dst is post
  repref <- ymd_hms("2022-10-30 01:00:00", tz = "Europe/Amsterdam")
  rep <- ymd_hms("2022-10-29 00:10:02.35", tz = "Europe/Amsterdam")
  expect_equal(update(rep, days = 30, hours = 2, minutes = 3),
    repref + 2*3600 + 3*60 + 2.35)
  expect_equal(update(rep, days = 30, hours = 2, seconds = 5.6),
    repref + 2*3600 + 10*60 + 5.6)
  expect_equal(update(rep, days = 30, hours = 2, seconds = 65.6),
    repref + 2*3600 + 11*60 + 5.6)

})

test_that("update handles vectors of dates", {
  poslt <- as.POSIXlt(c("2010-02-14 01:59:59", "2010-02-15 01:59:59", "2010-02-16 01:59:59"),
    tz = "UTC", format = "%Y-%m-%d %H:%M:%S"
  )
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)
  expect_equal(second(update(poslt, seconds = 1)), c(1, 1, 1))
  expect_equal(second(update(posct, seconds = 1)), c(1, 1, 1))
  expect_equal(day(update(date, days = 1)), c(1, 1, 1))
})

test_that("update handles vectors of dates and conformable vector of inputs", {
  poslt <- as.POSIXlt(c("2010-02-14 01:59:59", "2010-02-15 01:59:59", "2010-02-16
    01:59:59"), tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)
  expect_equal(second(update(poslt, seconds = c(1, 2, 3))), c(1, 2, 3))
  expect_equal(second(update(posct, seconds = c(1, 2, 3))), c(1, 2, 3))
  expect_equal(day(update(date, days = c(1, 2, 3))), c(1, 2, 3))
})

test_that("update handles single vector of inputs", {
  poslt <- as.POSIXlt("2010-03-14 01:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)
  expect_equal(second(update(poslt, seconds = c(1, 2, 3))), c(1, 2, 3))
  expect_equal(second(update(posct, seconds = c(1, 2, 3))), c(1, 2, 3))
  expect_equal(day(update(date, days = c(1, 2, 3))), c(1, 2, 3))
})

test_that("update handles conformable vectors of inputs", {
  ## FIXME: these actually should be an error in conformance with vctrs semantics
  poslt <- as.POSIXlt("2010-03-10 01:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)
  expect_equal(second(update(poslt, seconds = c(1, 2), minutes = c(1, 2, 3, 4))), c(1, 2, 1, 2))
  expect_equal(second(update(posct, seconds = c(1, 2), minutes = c(1, 2, 3, 4))), c(1, 2, 1, 2))
  expect_equal(day(update(date, days = c(1, 2), months = c(1, 2, 3, 4))), c(1, 2, 1, 2))
})

test_that("update.POSIXct returns input of length zero when given input of length zero", {
  x <- structure(vector(mode = "numeric"), class = c("POSIXct", "POSIXt"))
  expect_equal(update(x, days = 1), x)
})

test_that("update.POSIXlt returns input of length zero when given input of length zero", {
  x <- as.POSIXlt(structure(vector(mode = "numeric"), class = c("POSIXct", "POSIXt")))
  expect_equal(update(x, days = 1), x)
})

test_that("update correctly handles 60 seconds on 59 minute (bug #313)", {
  expect_equal(
    ymd_hms("2015-01-01 00:59:00") + seconds(60),
    ymd_hms("2015-01-01 01:00:00")
  )
  expect_equal(
    ymd_hms("2015-01-01 01:59:00") + seconds(60),
    ymd_hms("2015-01-01 02:00:00")
  )
  expect_equal(
    ymd_hms("2015-01-01 23:59:00") + seconds(60),
    ymd_hms("2015-01-02 00:00:00")
  )
  expect_equal(
    ymd_hms("2015-01-01 00:59:05") + seconds(55),
    ymd_hms("2015-01-01 01:00:00")
  )
  expect_equal(
    ymd_hms("2015-01-01 00:59:59") + seconds(1),
    ymd_hms("2015-01-01 01:00:00")
  )
})

test_that("Updateing on seconds doesn't affect hours", {
  ## https://github.com/tidyverse/lubridate/issues/619

  tt2 <- tt <- Sys.time()
  second(tt2) <- 5
  expect_equal(hour(tt), hour(tt2))
})

test_that("Subtracting months to March 1 produces correct results", {
  ## #1037
  time <- ymd("2022-04-01", tz = "America/New_York")
  expect_equal(time - months(1), ymd("2022-03-01", tz = "America/New_York"))
  time <- ymd("2021-04-01", tz = "America/New_York")
  expect_equal(time - months(1), ymd("2021-03-01", tz = "America/New_York"))
  time <- ymd("2022-05-01", tz = "America/New_York")
  expect_equal(time - months(2), ymd("2022-03-01", tz = "America/New_York"))
  time <- ymd_hms("2022-04-02 04:01:01", tz = "America/New_York")
  expect_equal(time - period("1m 1d 4H 1M 1S"), ymd("2022-03-01", tz = "America/New_York"))
})

## ## bug #319
## x <- structure(list(sec = 0, min = 0, hour = 0, mday = -212, mon = 7L,
##                         year = 108L, wday = NA_integer_, yday = NA_integer_, isdst = 0L,
##                         zone = "EST", gmtoff = -18000L),
##                    .Names = c("sec", "min",
##                        "hour", "mday", "mon", "year", "wday", "yday", "isdst", "zone",
##                        "gmtoff"),
##                    tzone = c("America/New_York", "EST", "EDT"),
##                    class = c("POSIXlt", "POSIXt"))
## update(x, year = year(x) + 1)

context("Updating dates")

test_that("update.Date returns a date object", {
  date <- as.Date("05/05/2010", "%m/%d/%Y")
  expect_is(update(date, days = 1), "Date")
  expect_is(update(date, ydays = 1), "Date")
  expect_is(update(date, mdays = 1), "Date")
  expect_is(update(date, wdays = 1), "Date")
  expect_is(update(date, months = 1), "Date")
  expect_is(update(date, years = 2001), "Date")
  expect_is(update(date, tzs = "UTC"), "Date")
})

test_that("update.Date returns a posix object if time is manipulated", {
  date <- as.Date("05/05/2010", "%m/%d/%Y")
  expect_is(update(date, seconds = 1), "POSIXt")
  expect_is(update(date, minutes = 1), "POSIXt")
  expect_is(update(date, hours = 1), "POSIXt")
})

test_that("update.POSIXlt returns a POSIXlt object", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  expect_is(update(poslt, seconds = 1), "POSIXlt")
  expect_is(update(poslt, minutes = 1), "POSIXlt")
  expect_is(update(poslt, hours = 1), "POSIXlt")
  expect_is(update(poslt, days = 1), "POSIXlt")
  expect_is(update(poslt, ydays = 1), "POSIXlt")
  expect_is(update(poslt, mdays = 1), "POSIXlt")
  expect_is(update(poslt, wdays = 1), "POSIXlt")
  expect_is(update(poslt, months = 1), "POSIXlt")
  expect_is(update(poslt, years = 2001), "POSIXlt")
  expect_is(update(poslt, tzs = "UTC"), "POSIXlt")
})

test_that("update.POSIXct returns a POSIXct object", {
  posct <- as.POSIXct("2010-02-03 13:45:59", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  expect_is(update(posct, seconds = 1), "POSIXct")
  expect_is(update(posct, minutes = 1), "POSIXct")
  expect_is(update(posct, hours = 1), "POSIXct")
  expect_is(update(posct, days = 1), "POSIXct")
  expect_is(update(posct, ydays = 1), "POSIXct")
  expect_is(update(posct, mdays = 1), "POSIXct")
  expect_is(update(posct, wdays = 1), "POSIXct")
  expect_is(update(posct, months = 1), "POSIXct")
  expect_is(update(posct, years = 2001), "POSIXct")
  expect_is(update(posct, tzs = "UTC"), "POSIXct")
})

test_that("update.Date performs simple operation as expected", {
  date <- as.Date("05/05/2010", "%m/%d/%Y")
  expect_equal(second(update(date, seconds = 1)), 1)
  expect_equal(minute(update(date, minutes = 1)), 1)
  expect_equal(hour(update(date, hours = 1)), 1)
  expect_equal(mday(update(date, mdays = 1)), 1)
  expect_equal(wday(update(date, mdays = 1)), 7)
  expect_equal(yday(update(date, mdays = 1)), 121)
  expect_equal(yday(update(date, ydays = 1)), 1)
  expect_equal(mday(update(date, ydays = 1)), 1)
  expect_equal(wday(update(date, ydays = 1)), 6)
  expect_equal(wday(update(date, wdays = 1)), 1)
  expect_equal(yday(update(date, wdays = 1)), 122)
  expect_equal(mday(update(date, wdays = 1)), 2)
  expect_equal(month(update(date, months = 1)), 1)
  expect_equal(year(update(date, years = 2000)), 2000)
  expect_match(tz(update(date, tzs = "UTC")), "UTC")
})

test_that("update.POSIXt performs simple operation as expected", {
  poslt <- as.POSIXlt("2010-02-03 13:45:59", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct("2010-02-03 13:45:59", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  expect_equal(second(update(poslt, seconds = 1)), 1)
  expect_equal(minute(update(poslt, minutes = 1)), 1)
  expect_equal(hour(update(poslt, hours = 1)), 1)
  expect_equal(mday(update(poslt, mdays = 1)), 1)
  expect_equal(wday(update(poslt, mdays = 1)), 2)
  expect_equal(yday(update(poslt, mdays = 1)), 32)
  expect_equal(yday(update(poslt, ydays = 1)), 1)
  expect_equal(mday(update(poslt, ydays = 1)), 1)
  expect_equal(wday(update(poslt, ydays = 1)), 6)
  expect_equal(wday(update(poslt, wdays = 1)), 1)
  expect_equal(yday(update(poslt, wdays = 1)), 31)
  expect_equal(mday(update(poslt, wdays = 1)), 31)
  expect_equal(month(update(poslt, months = 1)), 1)
  expect_equal(year(update(poslt, years = 2000)), 2000)
  expect_match(tz(update(poslt, tzs = "UTC")), "UTC")
  expect_equal(second(update(posct, seconds = 1)), 1)
  expect_equal(minute(update(posct, minutes = 1)), 1)
  expect_equal(hour(update(posct, hours = 1)), 1)
  expect_equal(mday(update(posct, mdays = 1)), 1)
  expect_equal(wday(update(posct, mdays = 1)), 2)
  expect_equal(yday(update(posct, mdays = 1)), 32)
  expect_equal(yday(update(posct, ydays = 1)), 1)
  expect_equal(mday(update(posct, ydays = 1)), 1)
  expect_equal(wday(update(posct, ydays = 1)), 6)
  expect_equal(wday(update(posct, wdays = 1)), 1)
  expect_equal(yday(update(posct, wdays = 1)), 31)
  expect_equal(mday(update(posct, wdays = 1)), 31)
  expect_equal(month(update(posct, months = 1)), 1)
  expect_equal(year(update(posct, years = 2000)), 2000)
  expect_match(tz(update(posct, tzs = "UTC")), "UTC")
})

test_that("update.POSIXt works on wdays", {

  date <- ymd("2017-05-07") ## sunday
  ct <- as.POSIXct("2010-02-03 13:45:59", tz = "America/New_York", format = "%Y-%m-%d %H:%M:%S") ## Wednesday
  expect_equal(wday(update(ct, wdays = 1)), 1)
  expect_equal(wday(update(ct, wdays = 2)), 2)
  expect_equal(wday(update(ct, wdays = 5)), 5)
  expect_equal(wday(update(ct, wdays = 7)), 7)
  expect_equal(wday(update(date, wdays = 1)), 1)
  expect_equal(wday(update(date, wdays = 2)), 2)
  expect_equal(wday(update(date, wdays = 5)), 5)
  expect_equal(wday(update(date, wdays = 7)), 7)

  ws <- 1
  expect_equal(wday(update(ct, wdays = 1, week_start = ws)), 2)
  expect_equal(wday(update(ct, wdays = 2, week_start = ws)), 3)
  expect_equal(wday(update(ct, wdays = 5, week_start = ws)), 6)
  expect_equal(wday(update(ct, wdays = 7, week_start = ws)), 1)
  expect_equal(wday(update(date, wdays = 1, week_start = ws)), 2)
  expect_equal(wday(update(date, wdays = 2, week_start = ws)), 3)
  expect_equal(wday(update(date, wdays = 5, week_start = ws)), 6)
  expect_equal(wday(update(date, wdays = 7, week_start = ws)), 1)

  ws <- 1
  expect_equal(wday(update(ct, wdays = 1, week_start = ws), week_start = ws), 1)
  expect_equal(wday(update(ct, wdays = 2, week_start = ws), week_start = ws), 2)
  expect_equal(wday(update(ct, wdays = 5, week_start = ws), week_start = ws), 5)
  expect_equal(wday(update(ct, wdays = 7, week_start = ws), week_start = ws), 7)
  expect_equal(wday(update(date, wdays = 1, week_start = ws), week_start = ws), 1)
  expect_equal(wday(update(date, wdays = 2, week_start = ws), week_start = ws), 2)
  expect_equal(wday(update(date, wdays = 5, week_start = ws), week_start = ws), 5)
  expect_equal(wday(update(date, wdays = 7, week_start = ws), week_start = ws), 7)

  ws <- 3
  expect_equal(wday(update(ct, wdays = 1, week_start = ws), week_start = ws), 1)
  expect_equal(wday(update(ct, wdays = 2, week_start = ws), week_start = ws), 2)
  expect_equal(wday(update(ct, wdays = 5, week_start = ws), week_start = ws), 5)
  expect_equal(wday(update(ct, wdays = 7, week_start = ws), week_start = ws), 7)
  expect_equal(wday(update(date, wdays = 1, week_start = ws), week_start = ws), 1)
  expect_equal(wday(update(date, wdays = 2, week_start = ws), week_start = ws), 2)
  expect_equal(wday(update(date, wdays = 5, week_start = ws), week_start = ws), 5)
  expect_equal(wday(update(date, wdays = 7, week_start = ws), week_start = ws), 7)

})

test_that("updates on ydays works correctly with leap years", {
  expect_equal(update(ymd("15-02-03", tz = "UTC"), years = 2000, ydays = 1),
               ymd("2000-01-01", tz = "UTC"))
  expect_equal(update(ymd("15-02-03", tz = "UTC"), years = 2015, ydays = 1),
               ymd("2015-01-01", tz = "UTC"))
  expect_equal(update(ymd("15-02-03", tz = "UTC"), years = 2016, ydays = 10),
               ymd("2016-01-10", tz = "UTC"))
  expect_equal(update(ymd("15-02-03", tz = "America/New_York"), years = 2000, ydays = 1),
               ymd("2000-01-01", tz = "America/New_York"))
  expect_equal(update(ymd("15-02-03", tz = "America/New_York"), years = 2015, ydays = 1),
               ymd("2015-01-01", tz = "America/New_York"))
  expect_equal(update(ymd("15-02-03", tz = "America/New_York"), years = 2016, ydays = 10),
               ymd("2016-01-10", tz = "America/New_York"))

  expect_equal(update(ymd(c("2016-02-29", "2016-03-01")), ydays = 1),
               ymd(c("2016-01-01", "2016-01-01")))
  expect_equal(update(ymd(c("2016-02-29", "2016-03-01"), tz = "America/New_York"), ydays = 1),
               ymd(c("2016-01-01", "2016-01-01"), tz = "America/New_York"))
  expect_equal(update(ymd_hms(c("2016-02-29 1:2:3", "2016-03-01 10:20:30")), ydays = 1),
               ymd_hms(c("2016-01-01 1:2:3", "2016-01-01 10:20:30")))
  expect_equal(update(ymd_hms(c("2016-02-29 1:2:3", "2016-03-01 10:20:30"), tz = "America/New_York"), ydays = 1),
               ymd_hms(c("2016-01-01 1:2:3", "2016-01-01 10:20:30"), tz = "America/New_York"))

})


test_that("update performs roll overs correctly for Date objects", {
  date <- as.Date("05/05/2010", "%m/%d/%Y")
  expect_equal(second(update(date, seconds = 61)), 1)
  expect_equal(minute(update(date, seconds = 61)), 1)
  expect_equal(minute(update(date, minutes = 61)), 1)
  expect_equal(hour(update(date, minutes = 61)), 1)
  expect_equal(hour(update(date, hours = 25)), 1)
  expect_equal(mday(update(date, hours = 25)), 6)
  expect_equal(yday(update(date, hours = 25)), 126)
  expect_equal(wday(update(date, hours = 25)), 5)
  expect_equal(mday(update(date, mdays = 32)), 1)
  expect_equal(month(update(date, mdays = 32)), 6)
  expect_equal(wday(update(date, wdays = 31)), 3)
  expect_equal(month(update(date, wdays = 31)), 6)
  expect_equal(yday(update(date, ydays = 366)), 1)
  expect_equal(month(update(date, ydays = 366)), 1)
  expect_equal(month(update(date, months = 13)), 1)
  expect_equal(year(update(date, months = 13)), 2011)
  expect_match(tz(update(date, months = 13)), "UTC")
})

test_that("update performs roll overs correctly for POSIXlt objects", {
  poslt <- as.POSIXlt("2010-05-05 00:00:00", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  expect_equal(second(update(poslt, seconds = 61)), 1)
  expect_equal(minute(update(poslt, seconds = 61)), 1)
  expect_equal(minute(update(poslt, minutes = 61)), 1)
  expect_equal(hour(update(poslt, minutes = 61)), 1)
  expect_equal(hour(update(poslt, hours = 25)), 1)
  expect_equal(mday(update(poslt, hours = 25)), 6)
  expect_equal(yday(update(poslt, hours = 25)), 126)
  expect_equal(wday(update(poslt, hours = 25)), 5)
  expect_equal(mday(update(poslt, mdays = 32)), 1)
  expect_equal(month(update(poslt, mdays = 32)), 6)
  expect_equal(wday(update(poslt, wdays = 31)), 3)
  expect_equal(month(update(poslt, wdays = 31)), 6)
  expect_equal(yday(update(poslt, ydays = 366)), 1)
  expect_equal(month(update(poslt, ydays = 366)), 1)
  expect_equal(month(update(poslt, months = 13)), 1)
  expect_equal(year(update(poslt, months = 13)), 2011)
  expect_match(tz(update(poslt, months = 13)), "GMT")
})

test_that("update performs roll overs correctly for POSIXct objects", {
  posct <- as.POSIXct("2010-05-05 00:00:00", tz = "GMT", format = "%Y-%m-%d %H:%M:%S")
  expect_equal(second(update(posct, seconds = 61)), 1)
  expect_equal(minute(update(posct, seconds = 61)), 1)
  expect_equal(minute(update(posct, minutes = 61)), 1)
  expect_equal(hour(update(posct, minutes = 61)), 1)
  expect_equal(hour(update(posct, hours = 25)), 1)
  expect_equal(mday(update(posct, hours = 25)), 6)
  expect_equal(yday(update(posct, hours = 25)), 126)
  expect_equal(wday(update(posct, hours = 25)), 5)
  expect_equal(mday(update(posct, mdays = 32)), 1)
  expect_equal(month(update(posct, mdays = 32)), 6)
  expect_equal(wday(update(posct, wdays = 31)), 3)
  expect_equal(month(update(posct, wdays = 31)), 6)
  expect_equal(yday(update(posct, ydays = 366)), 1)
  expect_equal(month(update(posct, ydays = 366)), 1)
  expect_equal(month(update(posct, months = 13)), 1)
  expect_equal(year(update(posct, months = 13)), 2011)
  expect_match(tz(update(posct, months = 13)), "GMT")
})

test_that("update performs consecutive roll overs correctly for
  Date objects regardless of order", {
  date <- update(as.Date("11/01/2010", "%m/%d/%Y"),
    months = 13, days = 32, hours = 25, minutes = 61, seconds
     = 61)
  expect_equal(second(date), 1)
  expect_equal(minute(date), 2)
  expect_equal(hour(date), 2)
  expect_equal(mday(date), 2)
  expect_equal(wday(date), 4)
  expect_equal(yday(date), 33)
  expect_equal(month(date), 2)
  expect_equal(year(date), 2011)
  expect_match(tz(date), "UTC")
  date2 <- update(as.Date("11/01/2010", "%m/%d/%Y"),
    seconds = 61, minutes = 61, hours = 25, days = 32, months = 13)
  expect_equal(second(date2), 1)
  expect_equal(minute(date2), 2)
  expect_equal(hour(date2), 2)
  expect_equal(mday(date2), 2)
  expect_equal(wday(date2), 4)
  expect_equal(yday(date2), 33)
  expect_equal(month(date2), 2)
  expect_equal(year(date2), 2011)
  expect_match(tz(date2), "UTC")
})

test_that("update performs consecutive roll overs correctly for
  POSIXlt objects", {
  posl <- as.POSIXlt("2010-11-01 00:00:00", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  poslt <- update(posl, months = 13, days = 32, hours = 25,
    minutes = 61, seconds = 61)

  expect_equal(second(poslt), 1)
  expect_equal(minute(poslt), 2)
  expect_equal(hour(poslt), 2)
  expect_equal(mday(poslt), 2)
  expect_equal(wday(poslt), 4)
  expect_equal(yday(poslt), 33)
  expect_equal(month(poslt), 2)
  expect_equal(year(poslt), 2011)
  expect_match(tz(poslt), "GMT")

  poslt2 <- update(posl, seconds = 61, minutes = 61, hours = 25,
    days = 32, months = 13)

  expect_equal(second(poslt2), 1)
  expect_equal(minute(poslt2), 2)
  expect_equal(hour(poslt2), 2)
  expect_equal(mday(poslt2), 2)
  expect_equal(wday(poslt2), 4)
  expect_equal(yday(poslt2), 33)
  expect_equal(month(poslt2), 2)
  expect_equal(year(poslt2), 2011)
  expect_match(tz(poslt2), "GMT")
})

test_that("update performs consecutive roll overs correctly for POSIXct objects", {
  posc <- as.POSIXct("2010-11-01 00:00:00", tz = "GMT", format
     = "%Y-%m-%d %H:%M:%S")
  posct <- update(posc, months = 13, days = 32, hours = 25,
    minutes = 61, seconds = 61)

  expect_equal(second(posct), 1)
  expect_equal(minute(posct), 2)
  expect_equal(hour(posct), 2)
  expect_equal(mday(posct), 2)
  expect_equal(wday(posct), 4)
  expect_equal(yday(posct), 33)
  expect_equal(month(posct), 2)
  expect_equal(year(posct), 2011)
  expect_match(tz(posct), "GMT")

  posct2 <- update(posc, seconds = 61, minutes = 61, hours = 25,
    days = 32, months = 13)

  expect_equal(second(posct2), 1)
  expect_equal(minute(posct2), 2)
  expect_equal(hour(posct2), 2)
  expect_equal(mday(posct2), 2)
  expect_equal(wday(posct2), 4)
  expect_equal(yday(posct2), 33)
  expect_equal(month(posct2), 2)
  expect_equal(year(posct2), 2011)
  expect_match(tz(posct2), "GMT")
})

test_that("update returns NA for date-times in the spring dst gap", {
  poslt <- as.POSIXlt("2010-03-14 01:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  expect_true(is.na(update(poslt, seconds = 65)))
  expect_true(is.na(update(poslt, minutes = 65)))
  expect_true(is.na(update(poslt, hours = 2)))
  poslt <- as.POSIXlt("2010-03-13 02:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  expect_true(is.na(update(poslt, mdays = 14)))
  expect_true(is.na(update(poslt, wdays = 8)))
  expect_true(is.na(update(poslt, ydays = 73)))
  poslt <- as.POSIXlt("2010-02-14 02:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  expect_true(is.na(update(poslt, months = 3)))
  poslt <- as.POSIXlt("2009-03-14 02:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  poslt <- force_tz(poslt, tzone = "America/New_York")
  expect_true(is.na(update(poslt, years = 2010)))
  poslt <- as.POSIXlt("2010-03-14 02:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  expect_true(is.na(update(poslt, tzs = "America/New_York")))
})

test_that("update handles vectors of dates", {
  poslt <- as.POSIXlt(c("2010-02-14 01:59:59", "2010-02-15 01:59:59", "2010-02-16 01:59:59"),
                      tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)
  expect_equal(second(update(poslt, seconds = 1)), c(1, 1, 1))
  expect_equal(second(update(posct, seconds = 1)), c(1, 1, 1))
  expect_equal(day(update(date, days = 1)), c(1, 1, 1))
})

test_that("update handles vectors of dates and conformable vector of inputs", {
  poslt <- as.POSIXlt(c("2010-02-14 01:59:59", "2010-02-15 01:59:59", "2010-02-16
    01:59:59"), tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)
  expect_equal(second(update(poslt, seconds = c(1, 2, 3))), c(1, 2, 3))
  expect_equal(second(update(posct, seconds = c(1, 2, 3))), c(1, 2, 3))
  expect_equal(day(update(date, days = c(1, 2, 3))), c(1, 2, 3))
})

test_that("update handles single vector of inputs", {
  poslt <- as.POSIXlt("2010-03-14 01:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)
  expect_equal(second(update(poslt, seconds = c(1, 2, 3))), c(1, 2, 3))
  expect_equal(second(update(posct, seconds = c(1, 2, 3))), c(1, 2, 3))
  expect_equal(day(update(date, days = c(1, 2, 3))), c(1, 2, 3))
})

test_that("update handles conformable vectors of inputs", {
  poslt <- as.POSIXlt("2010-03-10 01:59:59", tz = "UTC", format = "%Y-%m-%d %H:%M:%S")
  posct <- as.POSIXct(poslt)
  date <- as.Date(poslt)
  expect_equal(second(update(poslt, seconds = c(1, 2), minutes = c(1, 2, 3, 4))), c(1, 2, 1, 2))
  expect_equal(second(update(posct, seconds = c(1, 2), minutes = c(1, 2, 3, 4))), c(1, 2, 1, 2))
  expect_equal(day(update(date, days = c(1, 2), months = c(1, 2, 3, 4))), c(1, 2, 1, 2))
})

test_that("update.POSIXct returns input of length zero when given input of length zero", {
  x <- structure(vector(mode = "numeric"), class = c("POSIXct", "POSIXt"))
  expect_equal(update(x, days = 1), x)
})

test_that("update.POSIXlt returns input of length zero when given input of length zero", {
  x <- as.POSIXlt(structure(vector(mode = "numeric"), class = c("POSIXct", "POSIXt")))
  expect_equal(update(x, days = 1), x)
})

test_that("update correctly handles 60 seconds on 59 minute (bug #313)", {
  expect_equal(ymd_hms("2015-01-01 00:59:00") + seconds(60),
               ymd_hms("2015-01-01 01:00:00"))
  expect_equal(ymd_hms("2015-01-01 01:59:00") + seconds(60),
               ymd_hms("2015-01-01 02:00:00"))
  expect_equal(ymd_hms("2015-01-01 23:59:00") + seconds(60),
               ymd_hms("2015-01-02 00:00:00"))
  expect_equal(ymd_hms("2015-01-01 00:59:05") + seconds(55),
               ymd_hms("2015-01-01 01:00:00"))
  expect_equal(ymd_hms("2015-01-01 00:59:59") + seconds(1),
               ymd_hms("2015-01-01 01:00:00"))
})

test_that("Updateing on seconds doesn't affect hours", {
  ## https://github.com/tidyverse/lubridate/issues/619

  tt2 <- tt <- Sys.time()
  second(tt2) <- 5
  expect_equal(hour(tt), hour(tt2))
})


## ## bug #319
## x <- structure(list(sec = 0, min = 0, hour = 0, mday = -212, mon = 7L,
##                         year = 108L, wday = NA_integer_, yday = NA_integer_, isdst = 0L,
##                         zone = "EST", gmtoff = -18000L),
##                    .Names = c("sec", "min",
##                        "hour", "mday", "mon", "year", "wday", "yday", "isdst", "zone",
##                        "gmtoff"),
##                    tzone = c("America/New_York", "EST", "EDT"),
##                    class = c("POSIXlt", "POSIXt"))
## update(x, year = year(x) + 1)

#' @include timespans.r
#' @include util.r
#' @include durations.r
NULL

check_period <- function(object) {
  errors <- character()

  length(object@.Data) -> n
  lengths <- c(length(object@year), length(object@month),
               length(object@day), length(object@hour), length(object@minute))

  if (any(lengths != n)) {
    msg <- paste("Inconsistent lengths: year = ", lengths[1],
      ", month = ", lengths[2],
      ", day = ", lengths[3],
      ", hour = ", lengths[4],
      ", minute = ", lengths[5],
      ", second = ", n,
      sep = "")
    errors <- c(errors, msg)
  }

  values <- c(object@year, object@month, object@day, object@hour, object@minute)
  values <- na.omit(values)
  if (sum(values - trunc(values))) {
    msg <- "periods must have integer values"
    errors <- c(errors, msg)
  }

  if (length(errors) == 0)
    TRUE
  else
    errors
}

#' Period class
#'
#' Period is an S4 class that extends the [Timespan-class] class.
#' Periods track the change in the "clock time" between two date-times. They
#' are measured in common time related units: years, months, days, hours,
#' minutes, and seconds. Each unit except for seconds must be expressed in
#' integer values.
#'
#' The exact length of a period is not defined until the period is placed at a
#' specific moment of time. This is because the precise length of one year,
#' month, day, etc. can change depending on when it occurs due to daylight savings,
#' leap years, and other conventions. A period can be
#' associated with a specific moment in time by coercing it to an
#' [Interval-class] object with [as.interval()] or by adding
#' it to a date-time with "+".
#'
#' Periods provide a method for measuring generalized timespans when we wish to
#' model clock times. Periods will attain intuitive results at this task even
#' when leap years, leap seconds, gregorian days, daylight savings changes, and
#' other events happen during the period.
#'
#' Because Period represents imprecise amount of time it cannot be compared to
#' precise timestamps as Durations and Intervals are. You need to explicitely
#' convert to durations. See [Duration-class].
#'
#' The logic that guides arithmetic with periods can be unintuitive. Starting
#' with version 1.3.0, \pkg{lubridate} enforces the reversible property of arithmetic
#' (e.g. a date + period - period = date) by returning an NA if you create an
#' implausible date by adding periods with months or years units to a date. For
#' example,  adding one month to January 31st, 2013 results in February 31st,
#' 2013, which is not a real date. \pkg{lubridate} users have argued in the past that
#' February 31st, 2013 should be rolled over to March 3rd, 2013 or rolled back
#' to February 28, 2013. However, each of these corrections would destroy the
#' reversibility of addition (Mar 3 - one month == Feb 3 != Jan 31, Feb 28 - one
#' month == Jan 28 != Jan 31). If you would like to add and subtract months in a
#' way that rolls the results back to the last day of a month (when appropriate)
#' use the special operators, \code{\link{\%m+\%}},  \code{\link{\%m-\%}} or a
#' bit more flexible [add_with_rollback()].
#'
#' Period class objects have six slots. 1) .Data, a numeric object. The
#' apparent amount of seconds to add to the period. 2) minute, a numeric object.
#' The apparent amount of minutes to add to the period. 3) hour, a numeric object.
#' The apparent amount of hours to add to the period.4) day, a numeric object.
#' The apparent amount of days to add to the period.5) month, a numeric object.
#' The apparent amount of months to add to the period. 6) year, a numeric object.
#' The apparent amount of years to add to the period.
#'
#' @export
#' @keywords internal
setClass("Period", contains = c("Timespan", "numeric"),
  slots = c(year = "numeric", month = "numeric", day = "numeric",
    hour = "numeric", minute = "numeric"),
  prototype = prototype(year = 0, month = 0, day = 0, hour = 0, minute = 0),
  validity = check_period)

#' @name hidden_aliases
#' @aliases Arith,ANY,Period-method Arith,Duration,Period-method
#'   Arith,Period,Duration-method Compare,Period,Duration-method
#'   Compare,numeric,Period-method Compare,difftime,Period-method
#'   Compare,Period,difftime-method Compare,Period,Period-method
#'   Compare,Period,character-method Compare,Period,numeric-method
#'   Compare,character,Period-method second,Period-method second<-,Period-method
#'   minute,Period-method minute<-,Period-method hour,Period-method
#'   hour<-,Period-method Arith,Period,ANY-method day,Period-method
#'   day<-,Period-method month,Period-method month<-,Period-method
#'   year,Period-method year<-,Period-method date,Period-method
#'   date<-,Period-method as.numeric,Period-method show,Period-method
#'   c,Period-method rep,Period-method [,Period-method
#'   [<-,Period,ANY,ANY,Period-method [[,Period-method
#'   [[<-,Period,ANY,ANY,Period-method $,Period-method $<-,Period-method
#'   as.difftime,Period-method as.character,Period-method
#'   +,Period,Duration-method +,Period,Interval-method +,Period,Period-method
#'   +,Period,Date-method +,Date,Period-method +,Period,difftime-method
#'   +,difftime,Period-method +,Period,numeric-method +,numeric,Period-method
#'   +,Period,POSIXct-method +,POSIXct,Period-method +,Period,POSIXlt-method
#'   +,POSIXlt,Period-method /,Period,Duration-method /,Period,Interval-method
#'   /,Period,Period-method /,Period,difftime-method /,difftime,Period-method
#'   /,Period,numeric-method /,numeric,Period-method *,Period,ANY-method
#'   *,ANY,Period-method -,Period,ANY-method -,Period,missing-method
#'   -,ANY,Period-method %%,Period,Duration-method %%,Period,Interval-method
#'   %%,Period,Period-method >,Period,Period-method >=,Period,Period-method
#'   ==,Period,Period-method !=,Period,Period-method <=,Period,Period-method
#'   <,Period,Period-method >,Period,Duration-method >=,Period,Duration-method
#'   ==,Period,Duration-method !=,Period,Duration-method
#'   <=,Period,Duration-method <,Period,Duration-method >,Duration,Period-method
#'   >=,Duration,Period-method ==,Duration,Period-method
#'   !=,Duration,Period-method <=,Duration,Period-method
#'   <,Duration,Period-method >,Period,numeric-method >=,Period,numeric-method
#'   ==,Period,numeric-method !=,Period,numeric-method <=,Period,numeric-method
#'   <,Period,numeric-method >,numeric,Period-method >=,numeric,Period-method
#'   ==,numeric,Period-method !=,numeric,Period-method <=,numeric,Period-method
#'   <,numeric,Period-method !=,Duration,Period !=,Period,Duration
#'   !=,Period,Period !=,Period,numeric !=,numeric,Period %%,Period,Duration
#'   %%,Period,Interval %%,Period,Period *,ANY,Period *,Period,ANY -,ANY,Period
#'   -,Period,Interval -,Period,missing /,numeric,Period <,Duration,Period
#'   <,Period,Duration <,Period,Period <,Period,numeric <,numeric,Period
#'   <=,Duration,Period <=,Period,Duration <=,Period,Period <=,Period,numeric
#'   <=,numeric,Period ==,Duration,Period ==,Period,Duration ==,Period,Period
#'   ==,Period,numeric ==,numeric,Period >,Duration,Period >,Period,Duration
#'   >,Period,Period >,Period,numeric >,numeric,Period >=,Duration,Period
#'   >=,Period,Duration >=,Period,Period >=,Period,numeric >=,numeric,Period
NULL

setMethod("initialize", "Period", function(.Object, ...) {
  dots <- list(...)
  names(dots)[!nzchar(allNames(dots))] <- ".Data"
  lens <- unlist(lapply(dots, length), F, F)

  ## if any 0-length components, the entire object is 0-length
  if (any(lens == 0)) {
    for (nm in slotNames(.Object))
      slot(.Object, nm) <- numeric()
    validObject(.Object)
    return(.Object)
  }

  len <- max(lens)
  nas <- FALSE
  for (nm in slotNames(.Object)) {
    el <-
      if (is.null(obj <- dots[[nm]])) {
        rep.int(0L, len)
      } else {
        if (length(obj) < len) rep_len(obj, len)
        else obj
      }
    nas <- nas | is.na(el)
    slot(.Object, nm) <- el
  }

  ## If any component has NAs the entire object should have those NAs
  if (any(nas)) {
    for (nm in slotNames(.Object)) {
      slot(.Object, nm)[nas] <- NA_real_
    }
  }

  validObject(.Object)
  .Object
})

#' @export
setMethod("show", signature(object = "Period"), function(object) {
  if (length(object@.Data) == 0) {
    cat("<Period[0]>\n")
  } else {
    print(format(object))
  }
})

#' @export
format.Period <- function(x, ...) {
  if (length(x) == 0) {
    return(character())
  }

  show <- paste(
    x@year, "y ", x@month, "m ", x@day, "d ",
    x@hour, "H ", x@minute, "M ", x@.Data, "S",
    sep = ""
  )
  start <- regexpr("[-1-9]|(0\\.)", show)
  show <- ifelse(start > 0, substr(show, start, nchar(show)), "0S")

  show[is.na(x)] <- NA
  show
}

#' @export
xtfrm.Period <- function(x) {
  xtfrm(period_to_seconds(x))
}

#' @export
setMethod("c", signature(x = "Period"), function(x, ...) {
  dots <- list(...)
  nempty <- sapply(dots, length) != 0
  elements <- lapply(dots[nempty], as.period)
  seconds <- c(x@.Data, unlist(lapply(elements, slot, ".Data")))
  years <- c(x@year, unlist(lapply(elements, slot, "year")))
  months <- c(x@month, unlist(lapply(elements, slot, "month")))
  days <- c(x@day, unlist(lapply(elements, slot, "day")))
  hours <- c(x@hour, unlist(lapply(elements, slot, "hour")))
  minutes <- c(x@minute, unlist(lapply(elements, slot, "minute")))
  new("Period", seconds, year = years, month = months, day = days,
    hour = hours, minute = minutes)
})

#' @export
setMethod("rep", signature(x = "Period"), function(x, ...) {
  new("Period", rep(x@.Data, ...), year = rep(x@year, ...),
    month = rep(x@month, ...), day = rep(x@day, ...),
    hour = rep(x@hour, ...), minute = rep(x@minute, ...))
})

#' @export
setMethod("[", signature(x = "Period"),
  function(x, i, j, ..., drop = TRUE) {
    new("Period", x@.Data[i], year = x@year[i], month = x@month[i],
      day = x@day[i], hour = x@hour[i], minute = x@minute[i])
})

#' @export
setMethod("[[", signature(x = "Period"),
          function(x, i, j, ..., exact = TRUE) {
            new("Period", x@.Data[i], year = x@year[i], month = x@month[i],
                day = x@day[i], hour = x@hour[i], minute = x@minute[i])
})

#' @export
setMethod("[<-", signature(x = "Period", value = "Period"),
  function(x, i, j, ..., value) {
    x@.Data[i] <- value@.Data
    x@year[i] <- value@year
    x@month[i] <- value@month
    x@day[i] <- value@day
    x@hour[i] <- value@hour
    x@minute[i] <- value@minute
    x
})

#' @export
setMethod("[[<-", signature(x = "Period", value = "Period"),
          function(x, i, j, ..., value) {
            x@.Data[i] <- value@.Data
            x@year[i] <- value@year
            x@month[i] <- value@month
            x@day[i] <- value@day
            x@hour[i] <- value@hour
            x@minute[i] <- value@minute
            x
})

#' @export
setMethod("$", signature(x = "Period"), function(x, name) {
  if (name == "second") name <- ".Data"
    slot(x, name)
})

#' @export
setMethod("$<-", signature(x = "Period"), function(x, name, value) {
  if (name == "second") name <- ".Data"
    slot(x, name) <- rep_len(value, length(x))
    x
})

#' Create or parse period objects
#'
#' `period()` creates or parses a period object with the specified values.
#'
#' Within a Period object, time units do not have a fixed length (except for
#' seconds) until they are added to a date-time. The length of each time unit
#' will depend on the date-time to which it is added. For example, a year that
#' begins on 2009-01-01 will be 365 days long.  A year that begins on 2012-01-01
#' will be 366 days long. When math is performed with a period object, each unit
#' is applied separately. How the length of a period is distributed among its
#' units is non-trivial. For example, when leap seconds occur 1 minute is longer
#' than 60 seconds.
#'
#' Periods track the change in the "clock time" between two date-times. They
#' are measured in common time related units: years, months, days, hours,
#' minutes, and seconds. Each unit except for seconds must be expressed in
#' integer values.
#'
#' Besides the main constructor and parser [period()], period objects can also
#' be created with the specialized functions [years()], [months()], [weeks()],
#' [days()], [hours()], [minutes()], and [seconds()]. These objects can be added
#' to and subtracted to date-times to create a user interface similar to object
#' oriented programming.
#'
#' Note: Arithmetic with periods can result in undefined behavior when
#' non-existent dates are involved (such as February 29th in non-leap years).
#' Please see [Period-class] for more details and \code{\link{\%m+\%}} and
#' [add_with_rollback()] for alternative operations.
#'
#' @name period
#' @aliases periods
#' @param num a numeric or character vector. A character vector can specify
#'   periods in a convenient shorthand format or ISO 8601 specification. All
#'   unambiguous name units and abbreviations are supported, "m" stands for
#'   months, "M" for minutes unless ISO 8601 "P" modifier is present (see
#'   examples). Fractional units are supported but the fractional part is always
#'   converted to seconds.
#' @param units a character vector that lists the type of units to be used. The
#'   units in units are matched to the values in num according to their
#'   order. When `num` is character, this argument is ignored.
#' @param ... a list of time units to be included in the period and their
#'   amounts. Seconds, minutes,  hours, days, weeks, months, and years are
#'   supported. Normally only one of `num` or `...` are present. If both are
#'   present, the periods are concatenated.
#' @param x Any R object for `is.periods` and a numeric value of the number of
#'   units for elementary constructors. With the exception of seconds(), x must
#'   be an integer.
#' @param abbreviate Ignored. For consistency with S3 generic in base namespace.
#' @seealso [Period-class], [period()], \code{\link{\%m+\%}},
#'   [add_with_rollback()]
#' @return a period object
#' @keywords chron classes
#' @examples
#'
#' ### Separate period and units vectors
#'
#' period(c(90, 5), c("second", "minute"))
#' #  "5M 90S"
#' period(-1, "days")
#' period(c(3, 1, 2, 13, 1), c("second", "minute", "hour", "day", "week"))
#' period(c(1, -60), c("hour", "minute"))
#' period(0, "second")
#'
#' ### Units as arguments
#'
#' period (second = 90, minute = 5)
#' period(day = -1)
#' period(second = 3, minute = 1, hour = 2, day = 13, week = 1)
#' period(hour = 1, minute = -60)
#' period(second = 0)
#' period(c(1, -60), c("hour", "minute"), hour = c(1, 2), minute = c(3, 4))
#'
#' ### Lubridate style parsing
#'
#' period("2M 1sec")
#' period("2hours 2minutes 1second")
#' period("2d 2H 2M 2S")
#' period("2days 2hours 2mins 2secs")
#' period("2 days, 2 hours, 2 mins, 2 secs")
#' # Missing numerals default to 1. Repeated units are added up.
#' duration("day day")
#'
#' ### ISO 8601 parsing
#'
#' period("P10M23DT23H") # M stands for months
#' period("10DT10M") # M stands for minutes
#' period("P3Y6M4DT12H30M5S") # M for both minutes and months
#' period("P23DT60H 20min 100 sec") # mixing ISO and lubridate style parsing
#'
#' ### Comparison with characters (from v1.6.0)
#'
#' duration("day 2 sec") > "day 1sec"
#'
#' ### Elementary Constructors
#'
#' x <- ymd("2009-08-03")
#' x + days(1) + hours(6) + minutes(30)
#' x + days(100) - hours(8)
#'
#' class(as.Date("2009-08-09") + days(1)) # retains Date class
#' as.Date("2009-08-09") + hours(12)
#' class(as.Date("2009-08-09") + hours(12))
#' # converts to POSIXt class to accomodate time units
#'
#' years(1) - months(7)
#' c(1:3) * hours(1)
#' hours(1:3)
#'
#' # sequencing
#' y <- ymd(090101) # "2009-01-01 CST"
#' y + months(0:11)
#'
#' # compare DST handling to durations
#' boundary <- ymd_hms("2009-03-08 01:59:59", tz="America/Chicago")
#' boundary + days(1) # period
#' boundary + ddays(1) # duration
#' @export
period <- function(num = NULL, units = "second", ...) {
  if (is.character(num)) {
    parse_period(num)
  } else {
    out1 <- .period_from_num(num, units)
    out2 <- .period_from_units(list(...))
    if (is.null(out1) && is.null(out2)) new("Period", numeric())
    else if (is.null(out1)) out2
    else if (is.null(out2)) out1
    else c(out1, out2)
  }
}

parse_period <- function(x) {
  out <- .Call(C_parse_period, as.character(x))
  new("Period",
      out[1, ],
      minute = out[2, ],
      hour   = out[3, ],
      day    = out[4, ] + 7L*out[5, ],
      month  = out[6, ],
      year   = out[7, ])
}

.period_from_num <- function(num, units) {
  if (length(num) == 0)
    return(NULL)

  if (!is.numeric(num)) {
    stop(sprintf("First argument to `period()` constructor must be character or numeric. Supplied object of class '%s'", class(num)))
  }

  ## qucik check for common wrongdoings: #462
  if (inherits(num, c("Interval", "Duration")))
    stop("Interval or Durations objects cannot be used as input to 'period()' constructor. Plese use 'as.period()'.")

  if (length(units) %% length(num) != 0)
    stop("Arguments `num` and `units` must have same length")

  .period_from_units(structure(num, names = units))
}

.period_from_units <- function(pieces) {
  if (length(pieces) == 0)
    return(NULL)

  if (!is.numeric(pieces))
    pieces <- lapply(pieces, as.numeric)

  out <- list(second = 0, minute = 0, hour = 0, day = 0,
              week = 0, month = 0, year = 0)

  unit <- standardise_date_names(names(pieces))
  for (i in seq_along(unit)) {
    nm <- unit[[i]]
    out[[nm]] <- out[[nm]] + pieces[[i]]
  }
  out$day <- out$day + 7 * out$week

  new("Period", out$second, year = out$year, month = out$month,
      day = out$day, hour = out$hour, minute = out$minute)
}

#' @rdname period
#' @examples
#' is.period(as.Date("2009-08-03")) # FALSE
#' is.period(period(months= 1, days = 15)) # TRUE
#' @export
is.period <- function(x) is(x, "Period")

#' @export seconds minutes hours days weeks years milliseconds microseconds microseconds nanoseconds picoseconds
#' @rdname period
seconds <- function(x = 1) period(second = x)
#' @rdname period
minutes <- function(x = 1) period(minute = x)
#' @rdname period
hours <- function(x = 1) period(hour = x)
#' @rdname period
days <- function(x = 1) period(day = x)
#' @rdname period
weeks <- function(x = 1) period(week = x)
#' @rdname period
years <- function(x = 1) period(year = x)
#' @rdname period
milliseconds <- function(x = 1) seconds(x/1000)
#' @rdname period
microseconds <- function(x = 1) seconds(x/1000000)
#' @rdname period
nanoseconds <- function(x = 1) seconds(x/1e9)
#' @rdname period
picoseconds <- function(x = 1) seconds(x/1e12)

#' @rdname period
#' @export
months.numeric <- function(x, abbreviate) {
  period(month = x)
}

#' Contrive a period to/from a given number of seconds
#'
#' `period_to_seconds()` approximately converts a period to seconds assuming
#' there are 365.25 days in a calendar year and 365.25/12 days in a month.
#'
#' @param x A numeric object. The number of seconds to coerce into a period.
#' @return A number (period) that roughly equates to the period (seconds) given.
#' @export
period_to_seconds <- function(x) {
  x@.Data +
    60 * x@minute +
    60 * 60 * x@hour +
    60 * 60 * 24 * x@day +
    60 * 60 * 24 * 365.25 / 12 * x@month +
    60 * 60 * 24 * 365.25 * x@year
}

#' @description
#' `seconds_to_period()` create a period that has the maximum number of
#' non-zero elements (days, hours, minutes, seconds). This computation is exact
#' because it doesn't involve years or months.
#' @rdname period_to_seconds
#' @export
seconds_to_period <- function(x) {
  span <- as.double(x)
  remainder <- abs(span)
  newper <- period(second = rep(0, length(x)))

  slot(newper, "day") <- remainder %/% (3600 * 24)
  remainder <- remainder %% (3600 * 24)

  slot(newper, "hour") <- remainder %/% (3600)
  remainder <- remainder %% (3600)

  slot(newper, "minute") <- remainder %/% (60)

  slot(newper, ".Data") <- remainder %% (60)

  newper * sign(span)
}

#' @export
summary.Period <- function(object, ...) {
  nas <- is.na(object)
  object <- object[!nas]
  persecs <- period_to_seconds(object)
  qq <- stats::quantile(persecs)
  qq <- c(qq[1L:3L], mean(persecs), qq[4L:5L])
  qq <- seconds_to_period(qq)
  qq <- as.character(qq)
  names(qq) <- c("Min.", "1st Qu.", "Median", "Mean", "3rd Qu.",
                 "Max.")
  if (any(nas))
    c(qq, `NA's` = sum(nas))
  else qq
}

#' @export
setMethod("Arith", signature(e1 = "Period", e2 = "ANY"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, .Generic)
})

#' @export
setMethod("Arith", signature(e1 = "ANY", e2 = "Period"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, .Generic)
})

## duration is.numeric. So we need these explicits here:
#' @export
setMethod("Arith", signature(e1 = "Duration", e2 = "Period"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, .Generic)
})

#' @export
setMethod("Arith", signature(e1 = "Period", e2 = "Duration"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, .Generic)
})

#' @export
setMethod("Compare", signature(e1 = "Period", e2 = "Period"),
          function(e1, e2) {
            callGeneric(period_to_seconds(e1), period_to_seconds(e2))
          })

#' @export
setMethod("Compare", signature(e1 = "Period", e2 = "character"),
          function(e1, e2) {
            callGeneric(e1, as.period(e2))
          })

#' @export
setMethod("Compare", signature(e1 = "character", e2 = "Period"),
          function(e1, e2) {
            callGeneric(as.period(e1), e2)
          })

#' @export
setMethod("Compare", signature(e1 = "Period", e2 = "Duration"),
          function(e1, e2) {
            callGeneric(as.duration(e1), e2)
          })

#' @export
setMethod("Compare", signature(e1 = "Duration", e2 = "Period"),
          function(e1, e2) {
            callGeneric(e1, as.duration(e2))
          })

#' @export
setMethod("Compare", signature(e1 = "Period", e2 = "numeric"),
          function(e1, e2) {
            callGeneric(as.numeric(e1, "secs"), e2)
          })

#' @export
setMethod("Compare", signature(e1 = "numeric", e2 = "Period"),
          function(e1, e2) {
            callGeneric(e1, as.numeric(e2, "secs"))
          })

#' @export
setMethod("Compare", c(e1 = "Period", e2 = "difftime"),
          function(e1, e2) {
            callGeneric(as.numeric(e1, units = "secs"), as.numeric(e2, units = "secs"))
          })

#' @export
setMethod("Compare", c(e1 = "difftime", e2 = "Period"),
          function(e1, e2) {
            callGeneric(as.numeric(e1, units = "secs"), as.numeric(e2, units = "secs"))
          })

#' @include periods.r
NULL

#' Get/set months component of a date-time
#'
#' Date-time must be a POSIXct, POSIXlt, Date, Period, chron, yearmon, yearqtr, zoo,
#' zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
#'
#' @param x a date-time object
#' @param label logical. TRUE will display the month as a character string such
#'   as "January." FALSE will display the month as a number.
#' @param abbr logical. FALSE will display the month as a character string
#'   label, such as "January". TRUE will display an abbreviated version of the
#'  label, such as "Jan". abbr is disregarded if label = FALSE.
#' @param value a numeric object
#' @param locale for month, locale to use for month names. Default to current locale.
#' @return If `label = FALSE`: month as number (1-12, 1 = January, 12 = December),
#'   otherwise as an ordered factor.
#' @keywords utilities manip chron methods
#' @examples
#' x <- ymd("2012-03-26")
#' month(x)
#' month(x) <- 1
#' month(x) <- 13
#' month(x) > 3
#'
#' month(ymd(080101))
#' month(ymd(080101), label = TRUE)
#' month(ymd(080101), label = TRUE, abbr = FALSE)
#' month(ymd(080101) + months(0:11), label = TRUE)
#' @export
month <- function(x, label = FALSE, abbr = TRUE, locale = Sys.getlocale("LC_TIME"))
  UseMethod("month")

#' @export
month.default <- function(x, label = FALSE, abbr = TRUE, locale = Sys.getlocale("LC_TIME"))
  month(as.POSIXlt(x, tz = tz(x))$mon + 1, label, abbr, locale = locale)

#' @export
month.numeric <- function(x, label = FALSE, abbr = TRUE, locale = Sys.getlocale("LC_TIME")) {
  if (!all(x[!is.na(x)] %in% 1:12))
    stop("Values are not in 1:12")

  if (!label) {
    return(x)
  }

  names <- .get_locale_regs(locale)$month_names
  labels <- if (abbr) names$abr else names$full

  ordered(x, levels = 1:12, labels = labels)
}

#' @export
month.Period <- function(x, label = FALSE, abbr = TRUE, locale = Sys.getlocale("LC_TIME"))
  slot(x, "month")

#' @rdname month
#' @export
"month<-" <- function(x, value) {
  ## FIXME: how to make this localized and preserve backward compatibility? Guesser?
  if (!is.numeric(value)) {
    value <- pmatch(tolower(value),
                    c("january", "february", "march",
                      "june", "july", "august", "september",
                      "october", "november", "december"))
  }
  x <- x + months(value - month(x))
 }

setGeneric("month<-")

#' @export
setMethod("month<-", signature("Period"), function(x, value) {
  slot(x, "month") <- value
  x
})

#' Get the number of days in the month of a date-time
#'
#' Date-time must be a POSIXct, POSIXlt, Date, chron, yearmon, yearqtr,
#' zoo, zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
#'
#' @export
#' @param x a date-time object
#' @return An integer of the number of days in the month component of the date-time object.
days_in_month <- function(x) {
  month_x <- month(x, label = TRUE, locale = "C")
  n_days <- N_DAYS_IN_MONTHS[month_x]
  n_days[month_x == "Feb" & leap_year(x)] <- 29L
  n_days
}

## fixme: integrate with above, this oen is needed internally
.days_in_month <- function(m, y) {
  n_days <- N_DAYS_IN_MONTHS[m]
  n_days[m == 2L & leap_year(y)] <- 29L
  n_days
}

## tothink: export?
days_in_months_so_far <- function(month, leap) {
  ## if month is negative, compute from the end of the year
  cum_days_pos <- c(0, cumsum(N_DAYS_IN_MONTHS)[-12])
  cum_days_neg <- c(0, cumsum(rev(N_DAYS_IN_MONTHS))[-12])
  negative <- month < 0
  positive <- month > 0
  sofar <- integer(length(month))
  sofar[negative] <- cum_days_neg[-month[negative]]
  sofar[positive] <- cum_days_pos[month[positive]]
  adjust <- leap & ((negative & month == -12) | (positive & month > 2))
  sofar[adjust] <- sofar[adjust] + 1L
  sofar
}
## days_in_months_so_far(c(1, 2, 3, -10, -11, -12), rep.int(T, 6))
## [1]   0  31  60 275 306 335
## days_in_months_so_far(c(1, 2, 3, -10, -11, -12), rep.int(F, 6))
## [1]   0  31  59 275 306 334

#' Is x a date-time object?
#'
#' An instant is a specific moment in time. Most common date-time
#' objects (e.g, POSIXct, POSIXlt, and Date objects) are instants.
#'
#' @aliases instant instants
#' @export
#' @param x an R object
#' @return TRUE if x is a POSIXct, POSIXlt, or Date object, FALSE otherwise.
#' @seealso [is.timespan()], [is.POSIXt()], [is.Date()]
#' @keywords logic chron
#' @examples
#' is.instant(as.Date("2009-08-03")) # TRUE
#' is.timepoint(5) # FALSE
is.instant <- function(x) inherits(x, c("POSIXt", "POSIXct", "POSIXlt", "Date"))

#' @export
#' @rdname is.instant
is.timepoint <- is.instant

#' The current day and time
#'
#' @param tzone a character vector specifying which time zone you would like the
#'   current time in. tzone defaults to your computer's system timezone. You can
#'   retrieve the current time in the Universal Coordinated Time (UTC) with
#'   now("UTC").
#' @return `now` - the current datetime as a `POSIXct` object
#'
#' @keywords chron utilities
#' @examples
#' now()
#' now("GMT")
#' now("")
#' now() == now() # would be TRUE if computer processed both at the same instant
#' now() < now() # TRUE
#' now() > now() # FALSE
#' @export
now <- function(tzone = "")
  with_tz(Sys.time(), tzone)


#' @rdname now
#' @examples
#' today()
#' today("GMT")
#' today() == today("GMT") # not always true
#' today() < as.Date("2999-01-01") # TRUE  (so far)
#' @export
today <- function(tzone = "") {
  as_date(now(tzone))
}

#' 1970-01-01 UTC
#'
#' Origin is the date-time for 1970-01-01 UTC in POSIXct format. This date-time
#' is the origin for the numbering system used by POSIXct, POSIXlt, chron, and
#' Date classes.
#'
#' @keywords data chron
#' @examples
#' origin
#' @export origin
origin <- structure(0, class = c("POSIXct", "POSIXt"), tzone = "UTC")

.rep_maybe <- function(x, N) {
  if (N > 1 && length(x) > 1 && length(x) != N) {
    out <- rep_len(x, N)
    ## repl_len doesn't preserve attributes
    if (is.POSIXct(x))
      attributes(out) <- attributes(x)
    out
  } else {
    x
  }
}

##' Efficient creation of date-times from numeric representations
##'
##' `make_datetime()` is a very fast drop-in replacement for
##' [base::ISOdate()] and [base::ISOdatetime()]. `make_date()` produces
##' objects of class `Date`.
##'
##' Input vectors are silently recycled. All inputs except `sec` are
##' silently converted to integer vectors; `sec` can be either integer or
##' double.
##'
##' @param year numeric year
##' @param month numeric month
##' @param day numeric day
##' @param hour numeric hour
##' @param min numeric minute
##' @param sec numeric second
##' @param tz time zone. Defaults to UTC.
##' @export
##' @examples
##' make_datetime(year = 1999, month = 12, day = 22, sec = 10)
##' make_datetime(year = 1999, month = 12, day = 22, sec = c(10, 11))
make_datetime <- function(year = 1970L, month = 1L, day = 1L, hour = 0L, min = 0L, sec = 0, tz = "UTC") {
  lengths <- vapply(list(year, month, day, hour, min, sec), length, 1, USE.NAMES = FALSE)
  if (min(lengths) == 0L) {
    .POSIXct(numeric(), tz = tz)
  } else {
    N <- max(lengths)
    C_update_dt(.rep_maybe(origin, N), year = .rep_maybe(year, N), month = .rep_maybe(month, N),
                yday = integer(), mday = .rep_maybe(day, N), wday = integer(),
                hour = .rep_maybe(hour, N), minute = .rep_maybe(min, N),
                second = .rep_maybe(sec, N), tz = tz)
  }
}

##' @rdname make_datetime
##' @export
make_date <- function(year = 1970L, month = 1L, day = 1L) {
  lengths <- vapply(list(year, month, day), length, 1, USE.NAMES = FALSE)
  if (min(lengths) == 0L) {
    as.Date(integer(), origin = origin)
  } else {
    N <- max(lengths)
    secs <- .Call(C_make_d,
                  rep_len(as.integer(year), N),
                  rep_len(as.integer(month), N),
                  rep_len(as.integer(day), N))
    structure(secs/86400L, class = "Date")
  }
}

#' @include periods.r
NULL

#' Get/set days component of a date-time
#'
#' @details `mday()` and `yday()` return the day of the month and day of the
#'   year respectively. `day()` and `day<-()` are aliases for `mday()` and
#'   `mday<-()`.
#' @param x a POSIXct, POSIXlt, Date, chron, yearmon, yearqtr, zoo, zooreg,
#'   timeDate, xts, its, ti, jul, timeSeries, or fts object.
#' @param label logical. Only available for wday. TRUE will display the day of
#'   the week as an ordered factor of character strings, such as "Sunday." FALSE
#'   will display the day of the week as a number.
#' @param abbr logical. Only available for wday. FALSE will display the day of
#'   the week as an ordered factor of character strings, such as "Sunday." TRUE
#'   will display an abbreviated version of the label, such as "Sun". abbr is
#'   disregarded if label = FALSE.
#' @param value a numeric object
#' @param week_start day on which week starts following ISO conventions - 1
#'   means Monday, 7 means Sunday (default). When `label = TRUE`, this will be
#'   the first level of the returned factor. You can set `lubridate.week.start`
#'   option to control this parameter globally.
#' @param locale locale to use for day names. Default to current locale.
#' @return `wday()` returns the day of the week as a decimal number or an
#'   ordered factor if label is `TRUE`.
#' @keywords utilities manip chron methods
#' @examples
#' x <- as.Date("2009-09-02")
#' wday(x) #4
#'
#' wday(ymd(080101))
#' wday(ymd(080101), label = TRUE, abbr = FALSE)
#' wday(ymd(080101), label = TRUE, abbr = TRUE)
#' wday(ymd(080101) + days(-2:4), label = TRUE, abbr = TRUE)
#'
#' x <- as.Date("2009-09-02")
#' yday(x) #245
#' mday(x) #2
#' yday(x) <- 1  #"2009-01-01"
#' yday(x) <- 366 #"2010-01-01"
#' mday(x) > 3
#' @export day mday
day <- function(x)
  UseMethod("mday")

#' @rdname day
#' @export
mday <- day

#' @rdname day
#' @export
wday <- function(x, label = FALSE, abbr = TRUE,
                 week_start = getOption("lubridate.week.start", 7),
                 locale = Sys.getlocale("LC_TIME"))
  UseMethod("wday")

#' @export
wday.default <- function(x, label = FALSE, abbr = TRUE,
                         week_start = getOption("lubridate.week.start", 7),
                         locale = Sys.getlocale("LC_TIME")) {
  wday(as.POSIXlt(x, tz = tz(x))$wday + 1, label, abbr, locale = locale, week_start = week_start)
}

.shift_wday_names <- function(names, week_start = 7) {
  if (week_start != 7) {
    c(names[(week_start + 1):7], names[1:week_start])
  } else {
    names
  }
}

#' @export
wday.numeric <- function(x, label = FALSE, abbr = TRUE,
                         week_start = getOption("lubridate.week.start", 7),
                         locale = Sys.getlocale("LC_TIME")) {

  start <- as.integer(week_start)

  if (start > 7 || start < 1)
    stop("Invalid 'week_start' argument; must be between 1 and 7")

  if (start != 7) {
    x <- 1 + (x + (6 - start)) %% 7
  }

  if (!label) {
    return(x)
  }

  names <- .get_locale_regs(locale)$wday_names
  labels <- if (abbr) names$abr else names$full
  ordered(x, levels = 1:7, labels = .shift_wday_names(labels, week_start = start))
}

#' @export
mday.default <- function(x)
  as.POSIXlt(x, tz = tz(x))$mday

#' @export
mday.Period <- function(x)
  slot(x, "day")

#' @rdname day
#' @export
qday <- function(x)
  UseMethod("qday")

#' @export
qday.default <- function(x) {
  x <- as_date(x)
  as.integer(x - floor_date(x, "quarter")) + 1
}

#' @rdname day
#' @export
yday <- function(x)
  UseMethod("yday")

#' @export
yday.default <- function(x)
  as.POSIXlt(x, tz = tz(x))$yday + 1

#' @rdname day
#' @export
`day<-` <- function(x, value) standardGeneric("day<-")

#' @rdname day
#' @export
`mday<-` <- function(x, value) {
  day(x) <- value
  x
}

#' @rdname day
#' @export
`qday<-` <- function(x, value) standardGeneric("qday<-")

#' @export
setGeneric("qday<-", useAsDefault = function(x, value)
  x <- x + days(value - qday(x)))

#' @export
setGeneric("day<-", useAsDefault = function(x, value)
  x <- x + days(value - mday(x)))

#' @export
setMethod("day<-", signature("Period"), function(x, value) {
  slot(x, "day") <- value
  x
})

#' @rdname day
#' @export
"wday<-" <- function(x, week_start = getOption("lubridate.week.start", 7), value) {
  if (!is.numeric(value)) {
    ## FIXME: how to make this localized and preserve backward compatibility? Guesser?
    labels <- .shift_wday_names(c("sunday", "monday", "tuesday", "wednesday",
                                  "thursday", "friday", "saturday"),
                                week_start)
    value <- pmatch(tolower(value), labels)
  }
  x <- x + days(value - wday(x, week_start = week_start))
}

#' @rdname day
#' @export
"yday<-" <- function(x, value)
  x <- x + days(value - yday(x))

#' Cyclic encoding of date-times
#'
#' Encode a date-time object into a cyclic coordinate system in which the
#' distances between two pairs of dates separated by the same time duration are
#' the same.
#'
#' Machine learning models don't know that December 31st and January 1st are
#' close in our human calendar sense. `cyclic_encoding` makes it obvious to the
#' machine learner that two calendar dates are close by mapping the dates onto
#' the circle.
#'
#' @param x a date-time object
#' @param periods a character vector of periods. Follows same specification as
#'   [period] and [floor_date] functions.
#' @param encoders names of functions to produce the encoding. Defaults to
#'   "sin" and "cos". Names of any predefined functions accepting a numeric
#'   input are allowed.
#' @inheritParams DateTimeUpdate
#' @return a numeric matrix with number of columns equal `length(periods) *
#'   length(types)`.
#' @examples
#'
#' times <- ymd_hms("2019-01-01 00:00:00") + hours(0:23)
#' cyclic_encoding(times, c("day", "week", "month"))
#' plot(cyclic_encoding(times, "1d"))
#' plot(cyclic_encoding(times, "2d"), xlim = c(-1, 1))
#' plot(cyclic_encoding(times, "4d"), xlim = c(-1, 1))
#' @export
cyclic_encoding <- function(x, periods, encoders = c("sin", "cos"),
                            week_start = getOption("lubridate.week.start", 7)) {
  Ne <- length(encoders)
  Np <- length(periods)
  out <- matrix(0, nrow = length(x), ncol = Ne * Np)
  colnames(out) <- paste(rep.int(encoders, Np),
    rep(periods, each = Ne),
    sep = "."
  )
  for (pix in seq_along(periods)) {
    p <- periods[[pix]]
    beg <- as.numeric(floor_date(x, unit = p, week_start = week_start))
    end <- as.numeric(ceiling_date(x, unit = p, week_start = week_start))
    diff <- (2 * pi) * (as.numeric(x) - beg) / (end - beg)
    diff[end == beg] <- 0
    for (tix in seq_along(encoders)) {
      out[, (pix - 1) * Ne + tix] <- do.call(encoders[[tix]], list(diff))
    }
  }
  out
}

#' @include periods.r
NULL

#' Get/set seconds component of a date-time
#'
#' Date-time must be a  POSIXct, POSIXlt, Date, Period, chron, yearmon, yearqtr, zoo,
#' zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
#'
#' @export
#' @param x a date-time object
#' @param value numeric value to be assigned
#' @return the seconds element of x as a decimal number
#' @keywords utilities manip chron methods
#' @examples
#' x <- ymd("2012-03-26")
#' second(x)
#' second(x) <- 1
#' second(x) <- 61
#' second(x) > 2
second <- function(x) {
  UseMethod("second")
}

#' @export
second.default <- function(x) {
  as.POSIXlt(x, tz = tz(x))$sec
}

#' @export
second.Period <- function(x) {
  slot(x, ".Data")
}

#' @rdname second
#' @export
setGeneric("second<-",
  function (x, value) standardGeneric("second<-"),
  useAsDefault = function(x, value) {
    y <- update_datetime(as.POSIXct(x), seconds = value)
    reclass_date(y, x)
  }
)

#' @export
setMethod("second<-", "Duration", function(x, value) {
  x <- x + seconds(value - second(x))
})

#' @export
setMethod("second<-", signature("Period"), function(x, value) {
  slot(x, ".Data") <- value
  x
})

#' @export
setMethod("second<-", "Interval", function(x, value) {
  x <- x + seconds(value - second(x))
})

#' @export
setMethod("second<-", "POSIXt", function(x, value) {
  update_datetime(x, seconds = value)
})

#' @export
setMethod("second<-", "Date", function(x, value) {
  update_datetime(x, seconds = value)
})

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
#' @include Dates.r
#' @include difftimes.r
#' @include numeric.r
#' @include POSIXt.r
#' @include ops-addition.r
NULL

#' Add and subtract months to a date without exceeding the last day of the new month
#'
#' Adding months frustrates basic arithmetic because consecutive months have
#' different lengths. With other elements, it is helpful for arithmetic to
#' perform automatic roll over. For example, 12:00:00 + 61 seconds becomes
#' 12:01:01. However, people often prefer that this behavior NOT occur with
#' months. For example, we sometimes want January 31 + 1 month = February 28 and
#' not March 3. \code{\%m+\%} performs this type of arithmetic. Date \code{\%m+\%} months(n)
#' always returns a date in the nth month after Date. If the new date would
#' usually spill over into the n + 1th month, \code{\%m+\%} will return the last day of
#' the nth month ([rollback()]). Date \code{\%m-\%} months(n) always returns a
#' date in the nth month before Date.
#'
#' \code{\%m+\%} and \code{\%m-\%} handle periods with components less than a month by first
#' adding/subtracting months and then performing usual arithmetics with smaller
#' units.
#'
#' \code{\%m+\%} and \code{\%m-\%} should be used with caution as they are not one-to-one
#' operations and results for either will be sensitive to the order of
#' operations.
#'
#' @rdname mplus
#' @usage e1 \%m+\% e2
#' @aliases m+ %m+% m- %m-% %m+%,ANY,ANY-method %m-%,ANY,ANY-method %m+%,Period,ANY-method %m+%,ANY,Period-method %m-%,Period,ANY-method %m-%,ANY,Period-method %m+%,Duration,ANY-method %m+%,ANY,Duration-method %m-%,Duration,ANY-method %m-%,ANY,Duration-method %m+%,Interval,ANY-method %m+%,ANY,Interval-method %m-%,Interval,ANY-method %m-%,ANY,Interval-method
#' @param e1 A period or a date-time object of class [POSIXlt], [POSIXct]
#' or [Date].
#' @param e2 A period or a date-time object of class [POSIXlt], [POSIXct]
#' or [Date]. Note that one of e1 and e2 must be a period and the other a
#' date-time object.
#' @return A date-time object of class POSIXlt, POSIXct or Date
#' @examples
#' jan <- ymd_hms("2010-01-31 03:04:05")
#' jan + months(1:3) # Feb 31 and April 31 returned as NA
#' # NA "2010-03-31 03:04:05 UTC" NA
#' jan %m+% months(1:3) # No rollover
#'
#' leap <- ymd("2012-02-29")
#' "2012-02-29 UTC"
#' leap %m+% years(1)
#' leap %m+% years(-1)
#' leap %m-% years(1)
#' @export
"%m+%" <- function(e1, e2) standardGeneric("%m+%")

#' @export
setGeneric("%m+%")

#' @export
setMethod(
  "%m+%", signature(e2 = "Period"),
  function(e1, e2) add_with_rollback(e1, e2)
)

#' @export
setMethod(
  "%m+%", signature(e1 = "Period"),
  function(e1, e2) add_with_rollback(e2, e1)
)

#' @export
setMethod(
  "%m+%", signature(e2 = "ANY"),
  function(e1, e2) {
    stop("%m+% handles only Period objects as second argument")
  }
)

#' @export
"%m-%" <- function(e1, e2) standardGeneric("%m-%")

#' @export
setGeneric("%m-%")

#' @export
setMethod(
  "%m-%", signature(e2 = "Period"),
  function(e1, e2) add_with_rollback(e1, -e2)
)

#' @export
setMethod(
  "%m-%", signature(e1 = "Period"),
  function(e1, e2) add_with_rollback(e2, -e1)
)

#' @export
setMethod(
  "%m-%", signature(e2 = "ANY"),
  function(e1, e2) {
    stop("%m-% handles only Period objects as second argument")
  }
)

#' `add_with_rollback()` is like \code{\%m+\%} and \code{\%m-\%} with more
#' control over the rollback process. It allows the rollback to first day of the
#' month instead of the last day of the previous month and controls whether HMS
#' component of the end date is preserved or not. Please note that the rollback
#' kicks in only when the resulting date lands on the non-existing date.
#' @rdname mplus
#' @param roll_to_first rollback to the first day of the month instead of the
#'   last day of the previous month (passed to [rollback()])
#' @param preserve_hms retains the same hour, minute, and second information? If
#'   FALSE, the new date will be at 00:00:00 (passed to [rollback()])
#' @examples
#'
#' x <- ymd_hms("2019-01-29 01:02:03")
#' add_with_rollback(x, months(1))
#' add_with_rollback(x, months(1), preserve_hms = FALSE)
#' add_with_rollback(x, months(1), roll_to_first = TRUE)
#' add_with_rollback(x, months(1), roll_to_first = TRUE, preserve_hms = FALSE)
#' @export
add_with_rollback <- function(e1, e2, roll_to_first = FALSE, preserve_hms = TRUE) {
  any_HMS <- any(e2@.Data != 0) || any(e2@minute != 0) || any(e2@hour != 0) || any(e2@day != 0)
  any_year <- any(e2@year != 0)
  if (!is.na(any_year) && any_year) {
    e2$month <- 12 * e2@year + e2@month
    e2$year <- 0L
  }

  new <- .quick_month_add(e1, e2@month)
  roll <- day(new) < day(e1)
  roll <- !is.na(roll) & roll
  new[roll] <- rollbackward(new[roll], roll_to_first = roll_to_first, preserve_hms = preserve_hms)

  if (!is.na(any_HMS) && any_HMS) {
    e2$month <- 0L
    new + e2
  } else {
    new
  }
}

.quick_month_add <- function(object, mval) {
  tzs <- tz(object)
  utc <- as.POSIXlt(force_tz(object, tzone = "UTC"))
  utc$mon <- utc$mon + mval
  utc <- as.POSIXct(utc)
  new <- force_tz(utc, tzone = tzs, roll_dst = c("boundary", "post"))
  reclass_date(new, object)
}

#' Roll backward or forward a date the previous, current or next month
#'
#' `rollbackward()` changes a date to the last day of the previous month or to
#' the first day of the month. `rollforward()` rolls to the last day of the
#' current month or to the first day of the next month. Optionally, the new date
#' can retain the same hour, minute, and second information. `rollback()` is a
#' synonym for `rollbackward()`.
#'
#' @export
#' @param dates A POSIXct, POSIXlt or Date class object.
#' @param roll_to_first Rollback to the first day of the month instead of the
#' last day of the month
#' @param preserve_hms Retains the same hour, minute, and second information? If
#' FALSE, the new date will be at 00:00:00.
#' @return A date-time object of class POSIXlt, POSIXct or Date, whose day has
#' been adjusted to the last day of the previous month, or to the first day of
#' the month.
#' @examples
#' date <- ymd("2010-03-03")
#' rollbackward(date)
#'
#' dates <- date + months(0:2)
#' rollbackward(dates)
#'
#' date <- ymd_hms("2010-03-03 12:44:22")
#' rollbackward(date)
#' rollbackward(date, roll_to_first = TRUE)
#' rollbackward(date, preserve_hms = FALSE)
#' rollbackward(date, roll_to_first = TRUE, preserve_hms = FALSE)
rollbackward <- function(dates, roll_to_first = FALSE, preserve_hms = TRUE) {
  .roll(dates, roll_to_first, preserve_hms)
}

#' @rdname rollbackward
#' @export
rollback <- rollbackward

#' @rdname rollbackward
#' @export
rollforward <- function(dates, roll_to_first = FALSE, preserve_hms = TRUE) {
  .roll(dates, roll_to_first, preserve_hms, forward = TRUE)
}

.roll <- function(dates, roll_to_first, preserve_hms, forward = FALSE) {
  if (length(dates) == 0) {
    return(dates)
  }
  day(dates) <- 1
  if (!preserve_hms) {
    hour(dates) <- 0
    minute(dates) <- 0
    second(dates) <- 0
  }
  if (forward) {
    dates <- dates + months(1)
  }
  if (roll_to_first) {
    dates
  } else {
    dates - days(1)
  }
}

test_that("rollback returns correct results", {
  expect_equal(
    rollback(ymd_hms("2010-03-03 12:44:22")),
    ymd_hms("2010-02-28 12:44:22")
  )
  expect_equal(
    rollback(ymd_hms("2010-03-03 12:44:22"), preserve_hms = FALSE),
    ymd_hms("2010-02-28 00:00:00")
  )
  expect_equal(
    rollback(ymd_hms("2010-03-03 12:44:22"), roll_to_first = TRUE),
    ymd_hms("2010-03-01 12:44:22")
  )
  expect_equal(
    rollback(ymd_hms("2010-03-03 12:44:22"), roll_to_first = TRUE, preserve_hms = FALSE),
    ymd_hms("2010-03-01 00:00:00")
  )
})


test_that("rollback works with missing numbers", {
  expect_equal(
    ymd(c("20150131", NA)) %m+% months(1),
    ymd(c("201502-28", NA))
  )
  expect_equal(
    rollbackward(ymd(c("2015-03-31", NA))),
    ymd(c("201502-28", NA))
  )
})


test_that("rollforward returns correct results", {
  expect_equal(
    rollforward(ymd_hms("2010-02-03 12:44:22")),
    ymd_hms("2010-02-28 12:44:22")
  )
  expect_equal(
    rollforward(ymd_hms("2010-02-03 12:44:22"), preserve_hms = FALSE),
    ymd_hms("2010-02-28 00:00:00")
  )
  expect_equal(
    rollforward(ymd_hms("2010-02-03 12:44:22"), roll_to_first = TRUE),
    ymd_hms("2010-03-01 12:44:22")
  )
  expect_equal(
    rollforward(ymd_hms("2010-02-03 12:44:22"), roll_to_first = TRUE, preserve_hms = FALSE),
    ymd_hms("2010-03-01 00:00:00")
  )
})


test_that("rollforward works with missing numbers", {
  expect_equal(
    ymd(c("20150131", NA)) %m+% months(1),
    ymd(c("201502-28", NA))
  )

  expect_equal(
    rollforward(ymd(c("20150203", NA))),
    ymd(c("2015-02-28", NA))
  )
})

context("Rollback")

test_that("rollback returns correct results", {
  expect_equal(rollback(ymd_hms("2010-03-03 12:44:22")),
               ymd_hms("2010-02-28 12:44:22"))
  expect_equal(rollback(ymd_hms("2010-03-03 12:44:22"), preserve_hms = FALSE),
               ymd_hms("2010-02-28 00:00:00"))
  expect_equal(rollback(ymd_hms("2010-03-03 12:44:22"), roll_to_first = TRUE),
               ymd_hms("2010-03-01 12:44:22"))
  expect_equal(rollback(ymd_hms("2010-03-03 12:44:22"), roll_to_first = TRUE, preserve_hms = FALSE),
               ymd_hms("2010-03-01 00:00:00"))
})


test_that("rollback works with missing numbers", {
  expect_equal(ymd(c("20150131", NA)) %m+% months(1),
               ymd(c("201502-28", NA)))
  expect_equal(rollbackward(ymd(c("2015-03-31", NA))),
               ymd(c("201502-28", NA)))
})


test_that("rollforward returns correct results", {
  expect_equal(rollforward(ymd_hms("2010-02-03 12:44:22")),
               ymd_hms("2010-02-28 12:44:22"))
  expect_equal(rollforward(ymd_hms("2010-02-03 12:44:22"), preserve_hms = FALSE),
               ymd_hms("2010-02-28 00:00:00"))
  expect_equal(rollforward(ymd_hms("2010-02-03 12:44:22"), roll_to_first = TRUE),
               ymd_hms("2010-03-01 12:44:22"))
  expect_equal(rollforward(ymd_hms("2010-02-03 12:44:22"), roll_to_first = TRUE, preserve_hms = FALSE),
               ymd_hms("2010-03-01 00:00:00"))
})


test_that("rollforward works with missing numbers", {
  expect_equal(ymd(c("20150131", NA)) %m+% months(1),
               ymd(c("201502-28", NA)))

  expect_equal(rollforward(ymd(c("20150203", NA))),
               ymd(c("2015-02-28", NA)))
})

#' @include timespans.r
NULL

trunc_divide <- function(e1, e2) {
  trunc(e1 / e2)
}

#' @export
setMethod("%/%", signature("Timespan", "Timespan"), trunc_divide)
#' @export
setMethod("%/%", signature("difftime", "Timespan"), trunc_divide)

#' @export
setMethod("%/%", signature("Interval", "Interval"), trunc_divide)
#' @export
setMethod("%/%", signature("Period", "Period"), trunc_divide)

#' @export
setMethod("%/%", signature("Interval", "Duration"), trunc_divide)
#' @export
setMethod("%/%", signature("Duration", "Interval"), trunc_divide)

#' @export
setMethod("%/%", signature("Interval", "Period"), trunc_divide)
#' @export
setMethod("%/%", signature("Period", "Interval"), trunc_divide)

#' @include periods.r
NULL

#' Get/set minutes component of a date-time
#'
#' Date-time must be a  POSIXct, POSIXlt, Date, Period, chron, yearmon, yearqtr, zoo,
#' zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
#'
#' @export
#' @param x a date-time object
#' @param value numeric value to be assigned
#' @keywords utilities manip chron methods
#' @return the minutes element of x as a decimal number
#' @examples
#' x <- ymd("2012-03-26")
#' minute(x)
#' minute(x) <- 1
#' minute(x) <- 61
#' minute(x) > 2
minute <- function(x) {
  UseMethod("minute")
}

#' @export
minute.default <- function(x) {
  as.POSIXlt(x, tz = tz(x))$min
}

#' @export
minute.Period <- function(x) {
  slot(x, "minute")
}

#' @rdname minute
#' @export
setGeneric("minute<-",
  function (x, value) standardGeneric("minute<-"),
  useAsDefault = function(x, value) {
    y <- update_datetime(as.POSIXct(x), minutes = value)
    reclass_date(y, x)
  }
)

#' @export
setMethod("minute<-", "Duration", function(x, value) {
  x <- x + minutes(value - minute(x))
})

#' @export
setMethod("minute<-", signature("Period"), function(x, value) {
  slot(x, "minute") <- value
  x
})

#' @export
setMethod("minute<-", "Interval", function(x, value) {
  x <- x + minutes(value - minute(x))
})

#' @export
setMethod("minute<-", "POSIXt", function(x, value) {
  update_datetime(x, minutes = value)
})

#' @export
setMethod("minute<-", "Date", function(x, value) {
  update_datetime(x, minutes = value)
})

#' @include timespans.r
#' @include parse.r
#' @include util.r
NULL

check_interval <- function(object) {
  errors <- character()
  if (!is.numeric(object@.Data)) {
    msg <- "Span length must be numeric."
    errors <- c(errors, msg)
  }
  if (!is(object@start, "POSIXct")) {
    msg <- "Start date must be in POSIXct format."
    errors <- c(errors, msg)
  }
  if (length(object@.Data) != length(object@start)) {
    msg <- paste("Inconsistent lengths: spans = ", length(object@.Data),
      ", start dates = ", length(object@start),
      sep = ""
    )
    errors <- c(errors, msg)
  }
  if (length(errors) == 0) {
    TRUE
  } else {
    errors
  }
}

.units_within_seconds <- function(secs, unit = "second") {
  ## return a list suitable to pass to new("Period", ...)
  switch(unit,
    second = list(secs),
    minute = list(secs %% 60, minute = secs %/% 60),
    hour =
      c(
        .units_within_seconds(secs %% 3600, "minute"),
        list(hour = secs %/% 3600)
      ),
    day =
      c(
        .units_within_seconds(secs %% 86400, "hour"),
        list(day = secs %/% 86400)
      ),
    stop("Unsuported unit ", unit)
  )
}

#' Interval class
#'
#' Interval is an S4 class that extends the [Timespan-class] class. An
#' Interval object records one or more spans of time. Intervals record these
#' timespans as a sequence of seconds that begin at a specified date. Since
#' intervals are anchored to a precise moment of time, they can accurately be
#' converted to [Period-class] or [Duration-class] class objects. This
#' is because we can observe the length in seconds of each period that begins on a
#' specific date. Contrast this to a generalized period, which may not have a
#' consistent length in seconds (e.g. the number of seconds in a year will change
#' if it is a leap year).
#'
#' Intervals can be both negative and positive. Negative intervals progress
#' backwards from the start date; positive intervals progress forwards.
#'
#' Interval class objects have two slots: .Data, a numeric object equal to the number
#' of seconds in the interval; and start, a POSIXct object that specifies the time
#' when the interval starts.
#'
#' @aliases intervals
#' @export
setClass("Interval",
  contains = c("Timespan", "numeric"),
  slots = c(start = "POSIXct", tzone = "character"), validity = check_interval
)

#' @export
setMethod("show", signature(object = "Interval"), function(object) {
  if (length(object@.Data) == 0) {
    cat("<Interval[0]>\n")
  } else {
    print(format(object), quote = FALSE)
  }
})

#' @export
format.Interval <- function(x, ...) {
  paste(
    format(x@start, tz = x@tzone, usetz = TRUE),
    format(x@start + x@.Data, tz = x@tzone, usetz = TRUE),
    sep = "--"
  )
}

#' @export
c.Interval <- function(...) {
  dots <- list(...)
  nempty <- sapply(dots, length) != 0
  elements <- lapply(dots[nempty], as.interval)
  spans <- unlist(elements@.Data)
  starts <- do.call(c, unname(lapply(elements, int_start)))
  new("Interval", spans, start = starts, tzone = elements[[1]]@tzone)
}

#' @export
setMethod("rep", signature(x = "Interval"), function(x, ...) {
  new("Interval", rep(x@.Data, ...), start = rep(x@start, ...), tzone = x@tzone)
})

#' @export
setMethod(
  "[", signature(x = "Interval"),
  function(x, i, j, ..., drop = TRUE) {
    new("Interval", x@.Data[i], start = x@start[i], tzone = x@tzone)
  }
)

#' @export
setMethod(
  "[[", signature(x = "Interval"),
  function(x, i, j, ..., exact = TRUE) {
    new("Interval", x@.Data[i], start = x@start[i], tzone = x@tzone)
  }
)

#' @export
setMethod("[<-", signature(x = "Interval"), function(x, i, j, ..., value) {
  if (is.interval(value)) {
    x@.Data[i] <- value@.Data
    x@start[i] <- value@start
    new("Interval", x@.Data, start = x@start, tzone = x@tzone)
  } else {
    x@.Data[i] <- value
    new("Interval", x@.Data, start = x@start, tzone = x@tzone)
  }
})

#' @export
setMethod("[[<-", signature(x = "Interval"), function(x, i, j, ..., value) {
  if (is.interval(value)) {
    x@.Data[i] <- value@.Data
    x@start[i] <- value@start
    new("Interval", x@.Data, start = x@start, tzone = x@tzone)
  } else {
    x@.Data[i] <- value
    new("Interval", x@.Data, start = x@start, tzone = x@tzone)
  }
})

#' @export
setMethod("$", signature(x = "Interval"), function(x, name) {
  if (name == "span") name <- ".Data"
  slot(x, name)
})

#' @export
setMethod("$<-", signature(x = "Interval"), function(x, name, value) {
  if (name == "span") name <- ".Data"
  slot(x, name) <- value
  x
})

#' @export
unique.Interval <- function(x, ...) {
  df <- unique.data.frame(data.frame(data = x@.Data, start = x@start), ...)
  new("Interval", df$data, start = df$start, tzone = x@tzone)
}

#' Utilities for creation and manipulation of `Interval` objects
#'
#' `interval()` creates an [Interval-class] object with the specified start and
#' end dates. If the start date occurs before the end date, the interval will be
#' positive. Otherwise, it will be negative. Character vectors in ISO 8601
#' format are supported from v1.7.2.
#'
#' Intervals are time spans bound by two real date-times.  Intervals can be
#' accurately converted to either period or duration objects using
#' [as.period()], [as.duration()]. Since an interval is anchored to a fixed
#' history of time, both the exact number of seconds that passed and the number
#' of variable length time units that occurred during the interval can be
#' calculated.
#'
#' @export
#' @param start,end POSIXt, Date or a character vectors. When `start` is a
#'   character vector and end is `NULL`, ISO 8601 specification is assumed but
#'   with much more permisive lubridate style parsing both for dates and periods
#'   (see examples).
#' @param tzone a recognized timezone to display the interval in
#' @param x an R object
#' @return `interval()` -- [Interval-class] object.
#' @seealso [Interval-class], [as.interval()], \code{\link{\%within\%}}
#' @examples
#' interval(ymd(20090201), ymd(20090101))
#'
#' date1 <- ymd_hms("2009-03-08 01:59:59")
#' date2 <- ymd_hms("2000-02-29 12:00:00")
#' interval(date2, date1)
#' interval(date1, date2)
#' span <- interval(ymd(20090101), ymd(20090201))
#'
#' ### ISO Intervals
#'
#' interval("2007-03-01T13:00:00Z/2008-05-11T15:30:00Z")
#' interval("2007-03-01T13:00:00Z/P1Y2M10DT2H30M")
#' interval("P1Y2M10DT2H30M/2008-05-11T15:30:00Z")
#' interval("2008-05-11/P2H30M")
#'
#' ### More permisive parsing (as long as there are no intermittent / characters)
#' interval("2008 05 11/P2hours 30minutes")
#' interval("08 05 11/P 2h 30m")
#'
#' is.interval(period(months = 1, days = 15)) # FALSE
#' is.interval(interval(ymd(20090801), ymd(20090809))) # TRUE
interval <- function(start = NULL, end = NULL, tzone = tz(start)) {
  # NB: tzone is forced and never called on NULL here

  if (is.character(start) && is.null(end)) {
    return(parse_interval(start, tzone))
  }

  if (length(start) == 0 || length(end) == 0) {
    ## We used to return UTC on NULL
    if (is.null(start) && missing(tzone)) {
      tzone <- "UTC"
    }
    start <- POSIXct(tz = tzone)
    return(new("Interval", numeric(), start = start, tzone = tzone))
  }

  if (is.Date(start)) start <- date_to_posix(start)
  if (is.Date(end)) end <- date_to_posix(end)

  force(tzone)
  start <- as_POSIXct(start, tzone)
  end <- as_POSIXct(end, tzone)

  span <- as.numeric(end) - as.numeric(start)
  starts <- start + rep(0, length(span))
  if (tzone != tz(starts)) starts <- with_tz(starts, tzone)

  new("Interval", span, start = starts, tzone = tzone)
}

parse_interval <- function(x, tz) {

  # create matrix of string parts from x: 1st column is anything before /, 2nd is anything after.
  # replicates without stringr: str_split_fixed(x, "/", 2)
  mat <- matrix(
    c(gsub("(^[^/]+)/(.+$)", "\\1", x), gsub("(^[^/]+)/(.+$)", "\\2", x)),
    ncol = 2
  )
  pstart <- grepl("^P", mat[, 1])
  pend <- grepl("^P", mat[, 2])

  if (any(pstart & pend)) {
    stop(sprintf("Interval specified with period endpoints (%s)", x[pstart & pend][[1]]))
  }

  start <- .POSIXct(rep.int(NA_real_, length(x)), tz = tz)
  end <- .POSIXct(rep.int(NA_real_, length(x)), tz = tz)

  start[!pstart] <- .parse_iso_dt(mat[!pstart, 1], tz)
  end[!pend] <- .parse_iso_dt(mat[!pend, 2], tz = tz)
  end[pend] <- start[pend] + parse_period(mat[pend, 2])
  start[pstart] <- end[pstart] - parse_period(mat[pstart, 1])
  interval(start, end, tz)
}

#'
#' \code{\%--\%} Creates an interval that covers the range spanned by two
#' dates. It replaces the original behavior of \pkg{lubridate}, which created an
#' interval by default whenever two date-times were subtracted.
#'
#' @export
#' @rdname interval
"%--%" <- function(start, end) interval(start, end)

#' @rdname interval
#' @export
is.interval <- function(x) is(x, c("Interval"))


#' @rdname interval
#' @param int an interval object
#' @return `int_start()` and `int_end()` return a POSIXct date object when
#'   used as an accessor. Nothing when used as a setter.
#' @examples
#' int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int_start(int)
#' int_start(int) <- ymd("2001-06-01")
#' int
#'
#' int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int_end(int)
#' int_end(int) <- ymd("2002-06-01")
#' int
#' @export
int_start <- function(int) int@start

#' @rdname interval
#' @param value interval's start/end to be assigned to `int`
#' @export
"int_start<-" <- function(int, value) {
  value <- as.POSIXct(value)
  span <- as.numeric(int@start + int@.Data - value, "secs")
  equal.lengths <- data.frame(span, value)
  int <- new("Interval", span,
    start = equal.lengths$value,
    tzone = int@tzone
  )
}

#' @description `int_start()`/`int_end()` and `int_start<-()`/`int_end<-()` are
#'   "accessors" and "setters" respectively of the start/end date of an
#'   interval.
#'
#' @rdname interval
#' @export
int_end <- function(int) int@start + int@.Data

#' @rdname interval
#' @export
"int_end<-" <- function(int, value) {
  value <- as.POSIXct(value)
  span <- as.numeric(value - int@start, "secs")
  int <- new("Interval", span,
    start = int@start,
    tzone = int@tzone
  )
}

#' @rdname interval
#' @return `int_length()` -- numeric length of the interval in
#'   seconds. A negative number connotes a negative interval.
#' @examples
#' int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int_length(int)
#' @export
int_length <- function(int) int@.Data

#' @description
#' `int_flip()` reverses the order of the start date and end date in an
#' interval. The new interval takes place during the same timespan as the
#' original interval, but has the opposite direction.
#'
#' @rdname interval
#' @return `int_flip()` -- flipped interval object
#' @examples
#' int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int_flip(int)
#' @export
int_flip <- function(int) {
  new("Interval", -int@.Data, start = int@start + int@.Data, tzone = int@tzone)
}


#' @description
#' `int_shift()` shifts the start and end dates of an interval up or down the
#' timeline by a specified amount. Note that this may change the exact length of
#' the interval if the interval is shifted by a Period object. Intervals shifted
#' by a Duration or difftime object will retain their exact length in seconds.
#'
#' @rdname interval
#' @param by A period or duration object to shift by (for `int_shift`)
#' @return `int_shift()` -- an Interval object
#' @examples
#' int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int_shift(int, duration(days = 11))
#' int_shift(int, duration(hours = -1))
#' @export
int_shift <- function(int, by) {
  if (!is.timespan(by)) stop("by is not a recognized timespan object")
  if (is.interval(by)) stop("an interval cannot be shifted by another interval.
    Convert second interval to a period or duration.")
  interval(int@start + by, int_end(int) + by)
}


#' @description
#' `int_overlaps()` tests if two intervals overlap.
#'
#' @rdname interval
#' @param int1 an Interval object (for `int_overlaps()`, `int_aligns()`)
#' @param int2 an Interval object (for `int_overlaps()`, `int_aligns()`)
#' @return `int_overlaps()` -- logical, TRUE if int1 and int2 overlap by at
#'   least one second. FALSE otherwise
#' @examples
#' int1 <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int2 <- interval(ymd("2001-06-01"), ymd("2002-06-01"))
#' int3 <- interval(ymd("2003-01-01"), ymd("2004-01-01"))
#'
#' int_overlaps(int1, int2) # TRUE
#' int_overlaps(int1, int3) # FALSE
#' @export
int_overlaps <- function(int1, int2) {
  stopifnot(c(is.interval(int1), is.interval(int2)))
  int1 <- int_standardize(int1)
  int2 <- int_standardize(int2)
  int1@start <= int2@start + int2@.Data & int2@start <= int1@start + int1@.Data
}


#' @description
#' `int_standardize()` ensures all intervals in an interval object are
#' positive. If an interval is not positive, flip it so that it retains its
#' endpoints but becomes positive.
#'
#' @rdname interval
#' @examples
#' int <- interval(ymd("2002-01-01"), ymd("2001-01-01"))
#' int_standardize(int)
#' @export
int_standardize <- function(int) {
  negs <- !is.na(int@.Data) & int@.Data < 0
  int[negs] <- int_flip(int[negs])
  int
}


#' @description
#' `int_aligns()` tests if two intervals share an endpoint. The direction of
#' each interval is ignored. int_align tests whether the earliest or latest
#' moments of each interval occur at the same time.
#'
#' @rdname interval
#' @return `int_aligns()` -- logical, TRUE if int1 and int2 begin or end on the
#'   same moment. FALSE otherwise
#' @examples
#' int1 <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int2 <- interval(ymd("2001-06-01"), ymd("2002-01-01"))
#' int3 <- interval(ymd("2003-01-01"), ymd("2004-01-01"))
#'
#' int_aligns(int1, int2) # TRUE
#' int_aligns(int1, int3) # FALSE
#' @export
int_aligns <- function(int1, int2) {
  int1 <- int_standardize(int1)
  int2 <- int_standardize(int2)

  int1@start == int2@start | (int1@start + int1@.Data) == (int2@start + int2@.Data)
}

#' @description
#' `int_diff()` returns the intervals that occur between the elements of a
#' vector of date-times. `int_diff()` is similar to the POSIXt and Date
#' methods of [diff()], but returns an [Interval-class] object instead
#' of a difftime object.
#'
#' @rdname interval
#' @param times A vector of POSIXct, POSIXlt or Date class date-times (for
#'   `int_diff()`)
#' @return `int_diff()` -- interval object that contains the n-1 intervals
#'   between the n date-time in times
#' @examples
#' dates <- now() + days(1:10)
#' int_diff(dates)
#' @export
int_diff <- function(times) {
  interval(times[-length(times)], times[-1])
}


#' @importFrom generics intersect
#' @export
generics::intersect

#' @export
intersect.Interval <- function(x, y, ...) {
  int1 <- int_standardize(x)
  int2 <- int_standardize(y)

  starts <- pmax(int1@start, int2@start)
  ends <- pmin(int1@start + int1@.Data, int2@start + int2@.Data)
  spans <- as.numeric(ends) - as.numeric(starts)

  no.int <- ends < starts
  no.int <- is.na(no.int) | no.int
  spans[no.int] <- NA
  starts[no.int] <- NA

  new.int <- new("Interval", spans, start = starts, tzone = x@tzone)
  negix <- !is.na(x@.Data) & (sign(x@.Data) == -1)
  new.int[negix] <- int_flip(new.int[negix])
  new.int
}

#' @importFrom generics union
#' @export
generics::union

#' @export
union.Interval <- function(x, y, ...) {
  int1 <- int_standardize(x)
  int2 <- int_standardize(y)

  starts <- pmin(int1@start, int2@start)
  ends <- pmax(int1@start + int1@.Data, int2@start + int2@.Data)

  spans <- as.numeric(ends) - as.numeric(starts)

  if (any(!int_overlaps(int1, int2)) && is_verbose()) {
    message("Union includes intervening time between intervals.")
  }

  tz(starts) <- x@tzone
  new.int <- new("Interval", spans, start = starts, tzone = x@tzone)
  new.int[sign(x@.Data) == -1] <- int_flip(new.int[sign(x@.Data) == -1])
  new.int
}

#' @importFrom generics setdiff
#' @export
generics::setdiff

#' @export
setdiff.Interval <- function(x, y, ...) {
  if (length(x) != length(y)) {
    xy <- match_lengths(x, y)
    x <- xy[[1]]
    y <- xy[[2]]
  }

  aligned <- int_aligns(x, y)
  inside <- y %within% x
  makes2 <- !aligned & inside

  if (sum(makes2)) {
    stop(paste(
      "Cases", which(makes2),
      "result in discontinuous intervals."
    ))
  }

  int1 <- int_standardize(x)
  int2 <- int_standardize(y)

  first.y <- int_start(int2)
  last.y <- int_end(int2)

  starts <- int_start(int1)

  starts[(last.y + 1) %within% int1] <- last.y[(last.y + 1) %within% int1]

  ends <- int_end(int1)
  ends[(first.y - 1) %within% int1] <- first.y[(first.y - 1) %within% int1]

  spans <- as.numeric(ends) - as.numeric(starts)

  new.int <- new("Interval", spans, start = starts, tzone = x@tzone)
  new.int[sign(x@.Data) == -1] <- int_flip(new.int[sign(x@.Data) == -1])
  new.int
}


#' Does a date (or interval) fall within an interval?
#'
#' Check whether `a` lies within the interval `b`, inclusive of the endpoints.
#'
#' @export
#' @rdname within-interval
#' @aliases %within%,Interval,Interval-method %within%,ANY,Interval-method
#'   %within%,Date,list-method %within%,POSIXt,list-method %within%,Interval,list-method
#' @param a An interval or date-time object.
#' @param b Either an interval vector, or a list of intervals.
#'
#'   If `b` is an interval (or interval vector) it is recycled to the same length as `a`.
#'   If `b` is a list of intervals, `a` is checked if it falls within _any_
#'   of the intervals, i.e. `a %within% list(int1, int2)` is equivalent to
#'   `a %within% int1 | a %within% int2`.
#' @return A logical vector.
#' @examples
#' int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int2 <- interval(ymd("2001-06-01"), ymd("2002-01-01"))
#'
#' ymd("2001-05-03") %within% int # TRUE
#' int2 %within% int # TRUE
#' ymd("1999-01-01") %within% int # FALSE
#'
#' ## recycling (carefully note the difference between using a vector of
#' ## intervals and list of intervals for the second argument)
#' dates <- ymd(c("2014-12-20", "2014-12-30", "2015-01-01", "2015-01-03"))
#' blackout_vector <- c(
#'   interval(ymd("2014-12-30"), ymd("2014-12-31")),
#'   interval(ymd("2014-12-30"), ymd("2015-01-03"))
#' )
#' dates %within% blackout_vector
#'
#' ## within ANY of the intervals of a list
#' dates <- ymd(c("2014-12-20", "2014-12-30", "2015-01-01", "2015-01-03"))
#' lst <- list(
#'   interval(ymd("2014-12-30"), ymd("2014-12-31")),
#'   interval(ymd("2014-12-30"), ymd("2015-01-03"))
#' )
#' dates %within% lst
#'
#' ## interval within a list of intervals
#' int <- interval(
#'   ymd("2014-12-20", "2014-12-30"),
#'   ymd("2015-01-01", "2015-01-03")
#' )
#' int %within% lst
setGeneric("%within%", useAsDefault = function(a, b) {
  stop(sprintf(
    "No %%within%% method with signature a = %s,  b = %s",
    class(a)[[1]], class(b)[[1]]
  ))
})

.within <- function(a, int) {
  int <- int_standardize(int)
  as.numeric(a) - as.numeric(int@start) <= int@.Data & as.numeric(a) - as.numeric(int@start) >= 0
}

setMethod("%within%", signature(b = "Interval"), function(a, b) {
  if (!is.instant(a))
    stop("Argument 1 is not a recognized datetime")
  a <- as.POSIXct(a)
  .within(a, b)
})

.within_interval <- function(a, b) {
  start.in <- as.numeric(a@start) >= as.numeric(b@start)
  end.in <- (as.numeric(a@start) + a@.Data) <= (as.numeric(b@start) + b@.Data)
  start.in & end.in
}

setMethod("%within%", signature(a = "Interval", b = "Interval"), function(a, b) {
  a <- int_standardize(a)
  b <- int_standardize(b)
  .within_interval(a, b)
})

.within_list_instant <- function(a, b) {
  if (!all(sapply(b, is.interval))) {
    stop("When second argument to %within% is a list it must contain interval objects only")
  }
  a <- as.POSIXct(a)
  out <- FALSE
  for (int in b) {
    out <- out | .within(a, int)
  }
  out
}

setMethod("%within%", signature(a = "POSIXt", b = "list"), .within_list_instant)
setMethod("%within%", signature(a = "Date", b = "list"), .within_list_instant)

setMethod("%within%", signature(a = "Interval", b = "list"), function(a, b) {
  if (!all(sapply(b, is.interval))) {
    stop("When second argument to %within% is a list it must contain interval objects only")
  }
  out <- FALSE
  a <- int_standardize(a)
  for (int in b) {
    int <- int_standardize(int)
    out <- out | .within_interval(a, int)
  }
  out
})


#' @export
as.list.Interval <- function(x, ...) {
  lapply(seq_along(x), function(i) x[[i]])
}

#' @export
summary.Interval <- function(object, ...) {
  nas <- is.na(object)
  object <- object[!nas]
  n <- length(object)
  dates <- c(int_start(object), int_end(object))
  earliest <- as.character(min(dates))
  latest <- as.character(max(dates))
  zone <- tz(dates)

  qq <- c(n, earliest, latest, zone)
  names(qq) <- c("Intervals", "Earliest endpoint", "Latest endpoint", "Time zone")
  if (any(nas)) {
    c(qq, `NA's` = sum(nas))
  } else {
    qq
  }
}

#' @rdname time_length
setMethod("time_length", signature("Interval"), function(x, unit = "second") {
  unit <- standardise_period_names(unit)
  if (unit %in% c("year", "month")) {
    periods <- as.period(x, unit = unit)
    int_part <- slot(periods, unit)

    prev_aniv <- add_with_rollback(
      int_start(x), (int_part * period(1, units = unit)),
      roll_to_first = TRUE, preserve_hms = FALSE
    )
    next_aniv <- add_with_rollback(
      int_start(x), ((int_part + ifelse(x@.Data < 0, -1, 1)) * period(1, units = unit)),
      roll_to_first = TRUE, preserve_hms = FALSE
    )

    sofar <- as.duration(int_end(x) - prev_aniv)
    total <- as.duration(next_aniv - prev_aniv)
    int_part + sign(x@.Data) * sofar / total
  } else {
    as.duration(x) / duration(num = 1, units = unit)
  }
})

#' @export
setMethod("Arith", signature(e1 = "Interval", e2 = "ANY"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, .Generic)
})

#' @export
setMethod("Arith", signature(e1 = "ANY", e2 = "Interval"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, .Generic)
})

#' @name hidden_aliases
#' @aliases Arith,Interval,ANY-method Arith,ANY,Interval-method
#'   intersect,Interval,Interval-method union,Interval,Interval-method
#'   setdiff,Interval,Interval-method as.numeric,Interval-method
#'   as.integer,Interval-method show,Interval-method c,Interval-method
#'   rep,Interval-method [,Interval-method [<-,Interval,ANY,ANY,ANY-method
#'   [[,Interval-method [[<-,Interval,ANY,ANY,ANY-method $,Interval-method
#'   $<-,Interval-method as.difftime,Interval-method as.character,Interval-method
#'   day<-,Interval-method hour<-,Interval-method minute<-,Interval-method
#'   month<-,Interval-method second<-,Interval-method year<-,Interval-method
#'   +,Interval,Duration-method +,Interval,Interval-method +,Interval,Period-method
#'   +,Interval,Date-method +,Date,Interval-method +,Interval,difftime-method
#'   +,difftime,Interval-method +,Interval,numeric-method +,numeric,Interval-method
#'   +,Interval,POSIXct-method +,POSIXct,Interval-method +,Interval,POSIXlt-method
#'   +,POSIXlt,Interval-method /,Interval,Duration-method /,Interval,Interval-method
#'   /,Interval,Period-method /,Interval,difftime-method /,difftime,Interval-method
#'   /,Interval,numeric-method /,numeric,Interval-method *,Interval,ANY-method
#'   *,ANY,Interval-method -,Interval,missing-method -,Interval,Interval-method
#'   -,Date,Interval-method -,POSIXct,Interval-method -,POSIXlt,Interval-method
#'   -,numeric,Interval-method -,Interval,Date-method -,Interval,POSIXct-method
#'   -,Interval,POSIXlt-method -,Interval,numeric-method -,Duration,Interval-method
#'   -,Period,Interval-method %%,Interval,Duration-method %%,Interval,Interval-method
#'   %%,Interval,Period-method %%,Interval,Duration %%,Interval,Interval
#'   %%,Interval,Period -,Date,Interval -,Duration,Interval -,Interval,Date
#'   -,Interval,Interval -,Interval,POSIXct -,Interval,POSIXlt -,Interval,numeric
#'   -,POSIXct,Interval -,POSIXlt,Interval -,numeric,Interval -,Interval,Duration-method
#'   -,Interval,Period-method
NULL

# multiplication for everything

test_that("multiplication throws error for instants", {
  x <- as.POSIXct("2010-03-15 00:00:00", tz = "UTC")

  expect_error(3 * x)
})

test_that("multiplication throws error for intervals", {
  time1 <- as.POSIXct("2008-08-03 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  int <- interval(time1, time2)
  diff <- difftime(time2, time1)
  int2 <- interval(time1, time1 + 2 * diff)

  expect_equal(2 * int, int2)
})

test_that("multiplication works as expected for periods", {
  expect_equal(3 * months(1), months(3))
  expect_s4_class(3 * months(1), "Period")
})

test_that("multiplying vectors works for periods", {
  expect_equal(c(2, 3) * months(1), months(2:3))
  expect_s4_class(c(2, 3) * months(1), "Period")
})

test_that("multiplication works as expected for durations", {
  expect_equal(3 * dhours(1), dhours(3))
  expect_s4_class(3 * dhours(1), "Duration")
})

test_that("multiplying vectors works for durations", {
  expect_equal(c(2, 3) * dhours(1), dhours(2:3))
  expect_s4_class(c(2, 3) * dhours(1), "Duration")
})


test_that("make_difftime makes a correct difftime object", {
  x <- as.POSIXct("2008-01-01 12:00:00", tz = "UTC")
  y <- difftime(x + 3600, x)
  attr(y, "tzone") <- NULL

  expect_equal(make_difftime(3600), y)
  expect_s3_class(make_difftime(3600), "difftime")
})

context("Multiplication operations")


# multiplication for everything

test_that("multiplication throws error for instants", {
  x <- as.POSIXct("2010-03-15 00:00:00", tz = "UTC")

  expect_error(3 * x)

})

test_that("multiplication throws error for intervals", {
  time1 <- as.POSIXct("2008-08-03 00:00:00", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 00:00:00", tz = "UTC")
  int <- interval(time1, time2)
  diff <- difftime(time2, time1)
  int2 <- interval(time1, time1 + 2*diff)

  expect_equal(2*int, int2)
})

test_that("multiplication works as expected for periods", {

  expect_equal(3*months(1), months(3))
  expect_is(3*months(1), "Period")

})

test_that("multiplying vectors works for periods", {

  expect_equal(c(2, 3)*months(1), months(2:3))
  expect_is(c(2, 3)*months(1), "Period")

})

test_that("multiplication works as expected for durations", {

  expect_equal(3*dhours(1), dhours(3))
  expect_is(3*dhours(1), "Duration")

})

test_that("multiplying vectors works for durations", {

  expect_equal(c(2, 3)*dhours(1), dhours(2:3))
  expect_is(c(2, 3)*dhours(1), "Duration")

})


test_that("make_difftime makes a correct difftime object", {
  x <- as.POSIXct("2008-01-01 12:00:00", tz = "UTC")
  y <- difftime(x + 3600, x)
  attr(y, "tzone") <- NULL

  expect_equal(make_difftime(3600), y)
  expect_is(make_difftime(3600), "difftime")
})

test_that("interval() returns zero-length vector", {
  x <- interval()
  expect_s4_class(x, "Interval")
  expect_length(x, 0)
  expect_equal(format(x), character())
  expect_output(print(x), "<Interval[0]>", fixed = TRUE)

  x <- interval(POSIXct(), POSIXct())
  expect_s4_class(x, "Interval")
  expect_length(x, 0)

  x <- interval(POSIXct())
  expect_s4_class(x, "Interval")
  expect_length(x, 0)

  x <- interval(NULL, POSIXct())
  expect_s4_class(x, "Interval")
  expect_length(x, 0)
})

test_that("zero-length interval() respects `tzone`", {
  x <- interval(tzone = "America/Los_Angeles")
  expect_equal(x@tzone, "America/Los_Angeles")
  expect_equal(tz(int_start(x)), "America/Los_Angeles")
})

test_that("is.interval works as expected", {
  expect_false(is.interval(234))
  expect_false(is.interval(as.POSIXct("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.interval(as.POSIXlt("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.interval(Sys.Date()))
  expect_false(is.interval(minutes(1)))
  expect_false(is.interval(dminutes(1)))
  expect_true(is.interval(interval(
    as.POSIXct("2008-08-03 13:01:59", tz = "UTC"),
    as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  )))
})

test_that("Parsing of iso 8601 intervals works", {
  expect_equal(
    interval("2007-03-01T13:00:00Z/2008-05-11T15:30:00Z"),
    interval("2007-03-01T13:00:00Z", "2008-05-11T15:30:00Z")
  )
  expect_equal(
    interval("2007-03-01T13:00:00Z/P1Y2M10DT2H30M"),
    interval(
      "2007-03-01T13:00:00Z",
      ymd_hms("2007-03-01T13:00:00Z") + period("P1Y2M10DT2H30M")
    )
  )

  expect_equal(
    interval("P1Y2M10DT2H30M/2008-05-11T15:30:00Z"),
    interval(
      ymd_hms("2008-05-11T15:30:00Z") - period("P1Y2M10DT2H30M"),
      "2008-05-11T15:30:00Z"
    )
  )

  expect_equal(
    interval(c(
      "P1Y2M10DT2H30M/2008-05-11T15:30:00Z",
      "P2H30M/2008-05-11T15:30:00Z"
    )),
    interval(
      ymd_hms("2008-05-11T15:30:00Z") - period(c("P1Y2M10DT2H30M", "P2H30M")),
      "2008-05-11T15:30:00Z"
    )
  )

  expect_equal(
    interval(c(
      "P1Y2M10DT2H30M/2008-05-11T15:30:00Z",
      "2001-05-11T15:30:00Z/P2H30M"
    )),
    interval(
      c(
        ymd_hms("2008-05-11T15:30:00Z") - period("P1Y2M10DT2H30M"),
        ymd_hms("2001-05-11T15:30:00Z")
      ),
      c(
        ymd_hms("2008-05-11T15:30:00Z"),
        ymd_hms("2001-05-11T15:30:00Z") + period("P2H30M")
      )
    )
  )
})

test_that("interval works as expected", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  int <- interval(time1, time2)
  num <- as.numeric(time2) - as.numeric(time1)

  expect_equal(int@.Data, num)
  expect_equal(int@start, time1)
  expect_s4_class(int, "Interval")
})

test_that("as.integer works with intervals", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  int <- interval(time1, time2)
  expect_equal(as.integer(int, units = "minute"), as.integer(as.numeric(int, units = "minute")))
  expect_equal(as.integer(int, units = "hour"), as.integer(as.numeric(int, units = "hour")))
  expect_equal(as.integer(int), as.numeric(int))
  expect_type(as.integer(int), "integer")
})


test_that("interval handles vector input", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  time3 <- as.POSIXct("2009-08-04 13:01:59", tz = "UTC")
  int <- interval(c(time1, time2), time3)
  num <- as.numeric(time3) - as.numeric(c(time1, time2))

  expect_equal(int@.Data, num)
  expect_equal(int@start, c(time1, time2))
  expect_s4_class(int, "Interval")

  int2 <- interval(time1, c(time3, time2))
  num2 <- as.numeric(c(time3, time2)) - as.numeric(time1)
  starts <- structure(c(time1, time1), "tzone" = "UTC")

  expect_equal(int2@.Data, num2)
  expect_equal(int2@start, starts)
  expect_s4_class(int2, "Interval")
})


test_that("format.Interval works as expected", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  int <- interval(time1, time2)

  expect_match(format(int), "2008-08-03 13:01:59 UTC--2009-08-03 13:01:59 UTC")
})

test_that("interval handles character inputs", {
  t1 <- c("2007-01-01")
  t2 <- c("2007-08-01")
  expect_equal(
    interval(t1, t2),
    new("Interval", 18316800,
      start = as.POSIXct("2007-01-01", tz = "UTC"),
      tzone = "UTC"
    )
  )

  t1 <- c("2007-01-01")
  t2 <- c("2007-08-01 00:01:02")
  expect_equal(
    interval(t1, t2),
    new("Interval", 18316862,
      start = as.POSIXct("2007-01-01", tz = "UTC"),
      tzone = "UTC"
    )
  )
})


test_that("interval handles POSIXlt inputs", {
  t1 <- parse_date_time2("2007-01-01", "Ymd", lt = TRUE, tz = "America/Los_Angeles")
  t2 <- parse_date_time2("2007-08-01", "Ymd", lt = TRUE, tz = "America/Los_Angeles")

  expect_equal(
    unclass(interval(t1, t2)),
    unclass(new("Interval", 18313200,
      start = as.POSIXct("2007-01-01", tz = "America/Los_Angeles"),
      tzone = "America/Los_Angeles"
    ))
  )

  t1 <- parse_date_time2("2007-01-01", "Ymd", lt = TRUE, tz = "America/Los_Angeles")
  t2 <- parse_date_time2("2007-08-01 00:01:02", "Ymd HMS", lt = TRUE, tz = "America/Los_Angeles")

  expect_equal(
    unclass(interval(t1, t2)),
    unclass(new("Interval", 18313262,
      start = as.POSIXct("2007-01-01", tz = "America/Los_Angeles"),
      tzone = "America/Los_Angeles"
    ))
  )
})

test_that("as.interval works as expected", {
  a <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  pint <- interval(a, a + days(1))
  dint <- interval(a, a + ddays(1))

  expect_equal(pint, as.interval(days(1), a))
  expect_equal(dint, as.interval(ddays(1), a))
})

test_that("as.interval handles vector input", {
  a <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  b <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  pint <- interval(a, a + days(1:2))
  dint <- interval(a, a + ddays(1:2))
  pint1 <- interval(c(a, b), c(a, b) + days(1:2))
  dint1 <- interval(c(a, b), c(a, b) + ddays(1:2))
  pint2 <- interval(c(a, b), c(a, b) + days(1))
  dint2 <- interval(c(a, b), c(a, b) + ddays(1))

  expect_equal(pint, as.interval(days(1:2), a))
  expect_equal(dint, as.interval(ddays(1:2), a))
  expect_equal(pint1, as.interval(days(1:2), c(a, b)))
  expect_equal(dint1, as.interval(days(1:2), c(a, b)))
  expect_equal(pint2, as.interval(days(1), c(a, b)))
  expect_equal(dint2, as.interval(days(1), c(a, b)))
})

test_that("[<- can subset intervals", {
  ints <- data.frame(spans = c(
    interval(ymd(20090201), ymd(20090101)),
    interval(ymd(20090201), ymd(20090101))
  ))
  my_int <- interval(ymd(18800101), ymd(18810101))
  ints[1, 1] <- 31536000
  ints[2, 1] <- my_int
  int2 <- interval(ymd(20090201), ymd(20100201))

  expect_equal(ints[1, 1], int2)
  expect_equal(ints[2, 1], my_int)
})

test_that("interval handles correctly time zones of Date objects", {
  expect_equal(
    interval(as.Date("2011-01-01"), as.Date("2013-01-01")),
    interval(ymd("2011-01-01"), ymd("2013-01-01"))
  )
})

test_that("summary.Interval creates useful summary", {
  int <- interval(ymd(20090201), ymd(20090101))
  text <- c(3, "2009-01-01", "2009-02-01", "UTC", 1)
  names(text) <- c("Intervals", "Earliest endpoint", "Latest endpoint", "Time zone", "NA's")

  expect_equal(summary(c(int, int, int, as.interval(NA))), text)
})

test_that("as.interval handles NAs", {
  expect_equal(as.interval(NA), interval(NA, NA))
})

test_that("c.interval works as expected", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 15:03:39", tz = "UTC")
  int <- interval(time1, time2)
  ints <- interval(c(time1, time1), c(time2, time2))
  nants <- interval(c(time1, NA), c(time2, NA))

  expect_equal(c(int, int), ints)
  expect_equal(c(int, NA), nants)
})

test_that("c.interval works with named elements", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 15:03:39", tz = "UTC")
  time3 <- as.POSIXct("2009-09-03 15:03:39", tz = "UTC")
  a <- interval(time1, time2)
  b <- interval(time1, time2)
  c <- interval(time2, time3)

  nmd <- c(a = a, b = b)
  umd <- c(a, b)
  names(umd) <- c("a", "b")
  expect_equal(nmd, umd)

  nmd <- c(a = c(a, a), b = b)
  umd <- c(a, a, b)
  names(umd) <- c("a1", "a2", "b")
  expect_equal(nmd, umd)

  ## empty elements
  nmd <- c(a = a, NULL, b = b, c)
  umd <- c(a, b, c)
  names(umd) <- c("a", "b", "")
  expect_equal(nmd, umd)
})

test_that("%--% works as expected", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 15:03:39", tz = "UTC")
  int <- interval(time1, time2)
  attempt <- time1 %--% time2

  expect_equal(attempt, int)
})

test_that("int_start and int_end work as expected", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 15:03:39", tz = "UTC")
  int1 <- interval(time1, time2)
  int2 <- interval(time2, time1)

  expect_equal(int_start(int1), time1)
  expect_equal(int_start(int2), time2)
  expect_equal(int_start(c(int1, int2)), c(time1, time2))

  expect_equal(int_end(int1), time2)
  expect_equal(int_end(int2), time1)
  expect_equal(int_end(c(int2, int1)), c(time1, time2))
})

test_that("int_length works as expected", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- time1 + 1111
  time3 <- time1 - 5555
  int <- interval(time1, time2)
  int2 <- interval(time1, time3)

  expect_equal(int_length(int), 1111)
  expect_equal(int_length(int2), -5555)
  expect_equal(int_length(c(int, int2)), c(1111, -5555))
})

test_that("int_shift works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  int <- interval(time1, time2)
  int2 <- interval(time2, time1)

  dur <- duration(seconds = 11 * 24 * 60 * 60)
  dur2 <- duration(seconds = -3600)

  dint <- interval(time1 + 11 * 24 * 60 * 60, time2 + 11 * 24 * 60 * 60)
  dint2 <- interval(time2 + 11 * 24 * 60 * 60, time1 + 11 * 24 * 60 * 60)

  nint <- interval(time1 - 3600, time2 - 3600)
  nint2 <- interval(time2 - 3600, time1 - 3600)


  expect_equal(int_shift(int, dur), dint)
  expect_equal(int_shift(int, dur2), nint)
  expect_equal(int_shift(int, c(dur, dur2)), c(dint, nint))
  expect_equal(int_shift(int2, dur), dint2)
  expect_equal(int_shift(int2, dur2), nint2)
  expect_equal(int_shift(int2, c(dur, dur2)), c(dint2, nint2))
  expect_equal(int_shift(c(int, int2), dur), c(dint, dint2))
  expect_equal(int_shift(c(int, int2), dur2), c(nint, nint2))
  expect_equal(int_shift(c(int, int2), c(dur, dur2)), c(dint, nint2))
  expect_equal(int_shift(c(int, int2), c(dur2, dur)), c(nint, dint2))
})

test_that("int_overlaps works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  time21 <- as.POSIXct("2001-06-01", tz = "UTC")
  time22 <- as.POSIXct("2002-06-01", tz = "UTC")
  time31 <- as.POSIXct("2003-01-01", tz = "UTC")
  time32 <- as.POSIXct("2004-01-01", tz = "UTC")
  int1 <- interval(time1, time2)
  int2 <- interval(time21, time22)
  int3 <- interval(time31, time32)

  nint1 <- interval(time2, time1)
  nint2 <- interval(time22, time21)
  nint3 <- interval(time32, time31)


  expect_true(int_overlaps(int1, int2))
  expect_false(int_overlaps(int1, int3))
  expect_equal(int_overlaps(int1, c(int2, int3)), c(TRUE, FALSE))
  expect_equal(int_overlaps(c(int1, int3), int2), c(TRUE, FALSE))

  expect_true(int_overlaps(nint1, nint2))
  expect_false(int_overlaps(nint1, nint3))
  expect_equal(int_overlaps(nint1, c(nint2, nint3)), c(TRUE, FALSE))
  expect_equal(int_overlaps(c(nint1, nint3), nint2), c(TRUE, FALSE))

  expect_true(int_overlaps(int1, nint2))
  expect_true(int_overlaps(nint1, int2))
  expect_false(int_overlaps(int1, nint3))
  expect_false(int_overlaps(nint1, int3))
  expect_equal(int_overlaps(int1, c(int2, nint3)), c(TRUE, FALSE))
  expect_equal(int_overlaps(c(int1, nint3), nint2), c(TRUE, FALSE))
})

test_that("int_aligns works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  time21 <- as.POSIXct("2001-06-01", tz = "UTC")
  time22 <- as.POSIXct("2002-01-01", tz = "UTC")
  time31 <- as.POSIXct("2003-01-01", tz = "UTC")
  time32 <- as.POSIXct("2004-01-01", tz = "UTC")
  int1 <- interval(time1, time2)
  int2 <- interval(time21, time22)
  int3 <- interval(time31, time32)

  nint1 <- interval(time2, time1)
  nint2 <- interval(time22, time21)
  nint3 <- interval(time32, time31)

  expect_true(int_aligns(int1, int2))
  expect_false(int_aligns(int1, int3))
  expect_equal(int_aligns(int1, c(int2, int3)), c(TRUE, FALSE))
  expect_equal(int_aligns(c(int1, int3), int2), c(TRUE, FALSE))

  expect_true(int_aligns(nint1, nint2))
  expect_false(int_aligns(nint1, nint3))
  expect_equal(int_aligns(nint1, c(nint2, nint3)), c(TRUE, FALSE))
  expect_equal(int_aligns(c(nint1, nint3), nint2), c(TRUE, FALSE))

  expect_true(int_aligns(int1, nint2))
  expect_true(int_aligns(nint1, int2))
  expect_false(int_aligns(int1, nint3))
  expect_false(int_aligns(nint1, int3))
  expect_equal(int_aligns(int1, c(int2, nint3)), c(TRUE, FALSE))
  expect_equal(int_aligns(c(int1, nint3), nint2), c(TRUE, FALSE))
})

test_that("int_diff works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  time3 <- as.POSIXct("2003-06-01", tz = "UTC")
  time4 <- as.POSIXct("2002-01-01", tz = "UTC")
  time5 <- as.POSIXct("2003-01-01", tz = "UTC")
  time6 <- as.POSIXct("2004-01-01", tz = "UTC")
  times <- c(time1, time2, time3, time4, time5, time6)
  int1 <- interval(time1, time2)
  int2 <- interval(time2, time3)
  int3 <- interval(time3, time4)
  int4 <- interval(time4, time5)
  int5 <- interval(time5, time6)
  ints <- c(int1, int2, int3, int4, int5)

  expect_equal(int_diff(times), ints)
})

test_that("int_standardize works as expected", {
  int <- interval(ymd(20100101), ymd(20090101))
  standard <- interval(ymd(20090101), ymd(20100101))
  ints <- c(int, int)
  ints2 <- c(int, standard)
  standards <- c(standard, standard)

  expect_equal(int_standardize(int), standard)
  expect_equal(int_standardize(standard), standard)
  expect_equal(int_standardize(ints), standards)
  expect_equal(int_standardize(ints2), standards)
  expect_equal(int_standardize(standards), standards)
})

test_that("int_flip works as expected", {
  int <- interval(ymd(20100101), ymd(20090101))
  flip <- interval(ymd(20090101), ymd(20100101))
  ints <- c(int, int)
  ints2 <- c(int, flip)
  flips <- c(flip, flip)
  flips2 <- c(flip, int)

  expect_equal(int_flip(int), flip)
  expect_equal(int_flip(ints), flips)
  expect_equal(int_flip(ints2), flips2)
  expect_equal(int_flip(flip), int)
  expect_equal(int_flip(flips), ints)
  expect_equal(int_flip(flips2), ints2)
})

test_that("intersect.Interval works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  time21 <- as.POSIXct("2001-06-01", tz = "UTC")
  time22 <- as.POSIXct("2002-06-01", tz = "UTC")
  time31 <- as.POSIXct("2003-01-01", tz = "UTC")
  time32 <- as.POSIXct("2004-01-01", tz = "UTC")
  int1 <- interval(time1, time2)
  int2 <- interval(time21, time22)
  int3 <- interval(time31, time32)

  nint1 <- interval(time2, time1)
  nint2 <- interval(time22, time21)
  nint3 <- interval(time32, time31)

  expect_equal(intersect(int1, int2), interval(time21, time2))
  expect_equal(intersect(int1, int3), interval(NA, NA, tzone = "UTC"))
  expect_equal(intersect(int1, c(int2, int3)), interval(c(time21, NA), c(time2, NA)))
  expect_equal(intersect(c(int1, int3), int2), interval(c(time21, NA), c(time2, NA)))
  expect_equal(intersect(c(int1, int3), c(int2, int2)), interval(c(time21, NA), c(time2, NA)))

  expect_equal(intersect(nint1, nint2), interval(time2, time21))
  expect_equal(intersect(nint1, nint3), interval(NA, NA, tzone = "UTC"))
  expect_equal(intersect(nint1, c(nint2, nint3)), interval(c(time2, NA), c(time21, NA)))
  expect_equal(intersect(c(nint1, nint3), nint2), interval(c(time2, NA), c(time21, NA)))
  expect_equal(intersect(c(nint1, nint3), c(nint2, nint2)), interval(c(time2, NA), c(time21, NA)))

  expect_equal(intersect(int1, nint2), interval(time21, time2))
  expect_equal(intersect(nint1, nint2), interval(time2, time21))
  expect_equal(intersect(int1, nint3), interval(NA, NA, tzone = "UTC"))
  expect_equal(intersect(nint1, int3), interval(NA, NA, tzone = "UTC"))
  expect_equal(intersect(int1, c(nint2, int3)), interval(c(time21, NA), c(time2, NA)))
  expect_equal(intersect(nint1, c(int2, int3)), interval(c(time2, NA), c(time21, NA)))
  expect_equal(intersect(c(int1, nint3), nint2), interval(c(time21, NA), c(time2, NA)))
  expect_equal(intersect(c(nint1, int3), nint2), interval(c(time2, NA), c(time21, NA)))
  expect_equal(intersect(c(int1, nint3), int2), interval(c(time21, NA), c(time2, NA)))
  expect_equal(intersect(c(nint1, int3), int2), interval(c(time2, NA), c(time21, NA)))
  expect_equal(intersect(c(int1, nint3), c(nint2, int2)), interval(c(time21, NA), c(time2, NA)))
  expect_equal(intersect(c(nint1, int3), c(int2, nint2)), interval(c(time2, NA), c(time21, NA)))
})

test_that("union.Interval works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  time21 <- as.POSIXct("2001-06-01", tz = "UTC")
  time22 <- as.POSIXct("2002-06-01", tz = "UTC")
  time31 <- as.POSIXct("2003-01-01", tz = "UTC")
  time32 <- as.POSIXct("2004-01-01", tz = "UTC")
  int1 <- interval(time1, time2)
  int2 <- interval(time21, time22)
  int3 <- interval(time31, time32)

  nint1 <- interval(time2, time1)
  nint2 <- interval(time22, time21)
  nint3 <- interval(time32, time31)

  expect_equal(
    suppressMessages(union(int1, int2)),
    interval(time1, time22)
  )
  expect_equal(
    suppressMessages(union(int1, int3)),
    interval(time1, time32)
  )
  expect_equal(
    suppressMessages(union(int3, int1)),
    interval(time1, time32)
  )
  expect_equal(
    suppressMessages(union(int1, c(int2, int3))),
    interval(c(time1, time1), c(time22, time32))
  )
  expect_equal(
    suppressMessages(union(c(int1, int3), int2)),
    interval(c(time1, time21), c(time22, time32))
  )
  expect_equal(
    suppressMessages(union(c(int1, int3), c(int2, int1))),
    interval(c(time1, time1), c(time22, time32))
  )

  expect_equal(
    suppressMessages(union(nint1, nint2)),
    interval(time22, time1)
  )
  expect_equal(
    suppressMessages(union(nint1, nint3)),
    interval(time32, time1)
  )
  expect_equal(
    suppressMessages(union(nint3, nint1)),
    interval(time32, time1)
  )
  expect_equal(
    suppressMessages(union(nint1, c(nint2, nint3))),
    interval(c(time22, time32), c(time1, time1))
  )
  expect_equal(
    suppressMessages(union(c(nint1, nint3), nint2)),
    interval(c(time22, time32), c(time1, time21))
  )
  expect_equal(
    suppressMessages(union(c(nint1, nint3), c(nint2, nint1))),
    interval(c(time22, time32), c(time1, time1))
  )


  expect_equal(
    suppressMessages(union(int1, nint2)),
    interval(time1, time22)
  )
  expect_equal(
    suppressMessages(union(nint1, int2)),
    interval(time22, time1)
  )
  expect_equal(
    suppressMessages(union(int1, nint3)),
    interval(time1, time32)
  )
  expect_equal(
    suppressMessages(union(nint1, int3)),
    interval(time32, time1)
  )
  expect_equal(
    suppressMessages(union(int3, nint1)),
    interval(time1, time32)
  )
  expect_equal(
    suppressMessages(union(nint3, int1)),
    interval(time32, time1)
  )
  expect_equal(
    suppressMessages(union(int1, c(nint2, nint3))),
    interval(c(time1, time1), c(time22, time32))
  )
  expect_equal(
    suppressMessages(union(nint1, c(int2, int3))),
    interval(c(time22, time32), c(time1, time1))
  )
  expect_equal(
    suppressMessages(union(nint1, c(int2, nint3))),
    interval(c(time22, time32), c(time1, time1))
  )
  expect_equal(
    suppressMessages(union(c(int1, int3), nint2)),
    interval(c(time1, time21), c(time22, time32))
  )
  expect_equal(
    suppressMessages(union(c(nint1, nint3), int2)),
    interval(c(time22, time32), c(time1, time21))
  )
  expect_equal(
    suppressMessages(union(c(int1, nint3), nint2)),
    interval(c(time1, time32), c(time22, time21))
  )
  expect_equal(
    suppressMessages(union(c(int1, int3), c(nint2, nint1))),
    interval(c(time1, time1), c(time22, time32))
  )
  expect_equal(
    suppressMessages(union(c(nint1, nint3), c(int2, int1))),
    interval(c(time22, time32), c(time1, time1))
  )
  expect_equal(
    suppressMessages(union(c(int1, nint3), c(nint2, int1))),
    interval(c(time1, time32), c(time22, time1))
  )
})

test_that("setdiff.Interval works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  time21 <- as.POSIXct("2001-06-01", tz = "UTC")
  time22 <- as.POSIXct("2002-06-01", tz = "UTC")
  time31 <- as.POSIXct("2003-01-01", tz = "UTC")
  time32 <- as.POSIXct("2004-01-01", tz = "UTC")
  int1 <- interval(time1, time2)
  int2 <- interval(time21, time22)
  int3 <- interval(time31, time32)

  nint1 <- interval(time2, time1)
  nint2 <- interval(time22, time21)
  nint3 <- interval(time32, time31)

  arrive <- ymd_hms("2011-06-04 12:00:00",
    tz = "Pacific/Auckland"
  )
  leave <- ymd_hms("2011-08-10 14:00:00",
    tz = "Pacific/Auckland"
  )
  auckland <- interval(arrive, leave)
  jsm2 <- interval(
    ymd(20110720, tz = "Pacific/Auckland"),
    ymd(20110731, tz = "Pacific/Auckland")
  )

  expect_equal(setdiff(int1, int2), interval(time1, time21))
  expect_equal(setdiff(int1, int3), int1)
  expect_equal(setdiff(int1, c(int2, int3)), c(interval(time1, time21), int1))
  expect_equal(setdiff(c(int1, int3), int2), c(interval(time1, time21), int3))
  expect_equal(setdiff(c(int1, int3), c(int2, int2)), c(interval(time1, time21), int3))

  expect_equal(setdiff(nint1, nint2), interval(time21, time1))
  expect_equal(setdiff(nint1, nint3), nint1)
  expect_equal(setdiff(nint1, c(nint2, nint3)), c(interval(time21, time1), nint1))
  expect_equal(setdiff(c(nint1, nint3), nint2), c(interval(time21, time1), nint3))
  expect_equal(setdiff(c(nint1, nint3), c(nint2, nint2)), c(interval(time21, time1), nint3))

  expect_equal(setdiff(int1, nint2), interval(time1, time21))
  expect_equal(setdiff(nint1, int2), interval(time21, time1))
  expect_equal(setdiff(int1, nint3), int1)
  expect_equal(setdiff(nint1, int3), nint1)
  expect_equal(setdiff(int1, c(nint2, nint3)), c(interval(time1, time21), int1))
  expect_equal(setdiff(nint1, c(int2, int3)), c(interval(time21, time1), nint1))
  expect_equal(setdiff(int1, c(nint2, int3)), c(interval(time1, time21), int1))
  expect_equal(setdiff(nint1, c(int2, nint3)), c(interval(time21, time1), nint1))
  expect_equal(setdiff(c(int1, int3), nint2), c(interval(time1, time21), int3))
  expect_equal(setdiff(c(nint1, nint3), int2), c(interval(time21, time1), nint3))
  expect_equal(setdiff(c(nint1, int3), nint2), c(interval(time21, time1), int3))
  expect_equal(setdiff(c(int1, int3), c(nint2, nint2)), c(interval(time1, time21), int3))
  expect_equal(setdiff(c(nint1, nint3), c(int2, int2)), c(interval(time21, time1), nint3))
  expect_equal(setdiff(c(int1, nint3), c(nint2, int2)), c(interval(time1, time21), nint3))
  expect_error(
    setdiff(auckland, jsm2),
    "Cases 1 result in discontinuous intervals."
  )
})

test_that("%within% fails meaningfully when applied on wrong types", {
  expect_error(now() %within% now(), "No %within% method")
})


test_that("%within% works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2003-01-01", tz = "UTC")
  time3 <- as.POSIXct("2001-06-01", tz = "UTC")
  time4 <- as.POSIXct("2002-06-01", tz = "UTC")
  time5 <- as.POSIXct("2003-01-01", tz = "UTC")
  time6 <- as.POSIXct("2004-01-01", tz = "UTC")
  time7 <- as.POSIXct("2003-01-02", tz = "UTC")
  base <- interval(time1, time2)
  ins <- interval(time3, time4)
  bord <- interval(time5, time6)
  olap <- interval(time4, time6)
  outs <- interval(time7, time6)

  nbase <- interval(time2, time1)
  nins <- interval(time4, time3)
  nbord <- interval(time6, time5)
  nolap <- interval(time6, time4)
  nouts <- interval(time6, time7)

  expect_true(ins %within% base)
  expect_false(outs %within% base)
  expect_false(bord %within% base)
  expect_false(olap %within% base)

  expect_true(nins %within% nbase)
  expect_false(nouts %within% nbase)
  expect_false(nbord %within% nbase)
  expect_false(nolap %within% nbase)

  expect_true(ins %within% nbase)
  expect_false(outs %within% nbase)
  expect_false(bord %within% nbase)
  expect_false(olap %within% nbase)

  expect_true(nins %within% base)
  expect_false(nouts %within% base)
  expect_false(nbord %within% base)
  expect_false(nolap %within% base)
})

test_that("%within% works with inverted intervals", {
  int <- interval(as.Date("2021-01-01"), as.Date("2021-12-31"))
  x <- as.Date("2021-06-15")
  expect_true(x %within% int)
  flip <- int_flip(int)
  expect_true(x %within% flip)
  int2 <- interval(as.Date("2021-02-01"), as.Date("2021-10-31"))
  expect_true(int2 %within% int)
  expect_true(int_flip(int2) %within% int)
  expect_true(int2 %within% int_flip(int))
})


test_that("%with% recycles both arguments", {
  blackouts <- c(
    interval(ymd("2014-12-30"), ymd("2014-12-31")),
    interval(ymd("2014-12-30"), ymd("2015-01-03"))
  )
  testdates <- c(ymd("2014-12-20", ymd("2014-12-30"), ymd("2015-01-01"), ymd("2015-01-03")))
  expect_equal(testdates %within% blackouts, c(F, T, F, T))
})

test_that("%within% works with list of intervals", {
  testdates <- ymd(c("2014-12-20", "2014-12-30", "2015-01-01", "2015-01-03"))
  blackouts <- list(
    interval(ymd("2014-12-30"), ymd("2014-12-31")),
    interval(ymd("2014-12-30"), ymd("2015-01-03"))
  )
  expect_equal(testdates %within% blackouts, c(F, T, T, T))
  testdates <- c(ymd(c("2014-12-20", "2014-12-30", "2015-01-01", "2015-01-03"), tz = "UTC"))
  blackouts <- list(
    interval(ymd("2014-12-30"), ymd("2014-12-31")),
    interval(ymd("2014-12-30"), ymd("2015-01-03"))
  )
  expect_equal(testdates %within% blackouts, c(F, T, T, T))
})

test_that("%within% works with interval LHS", {
  lst <- list(
    interval(ymd("2014-12-30"), ymd("2014-12-31")),
    interval(ymd("2014-12-30"), ymd("2015-01-03"))
  )

  int <- interval(
    ymd("2014-12-20", "2014-12-30"),
    ymd("2015-01-01", "2015-01-03")
  )

  expect_equal(int %within% lst, c(F, T))
  expect_false(int[[1]] %within% lst)
  expect_true(int[[2]] %within% lst)

  expect_equal(int[[1]] %within% int, c(T, F))
  expect_equal(int[[2]] %within% int, c(F, T))
  expect_equal(int %within% c(int, int), c(T, T, T, T))
})

test_that("%% on interval uses m+ arithmetic", {

  ## From https://github.com/tidyverse/lubridate/issues/633
  start <- c(
    "2016-04-29 12:00:00 GMT", "2013-10-31 12:00:00 GMT", "2012-05-31 12:00:00 GMT", "2010-06-29 12:00:00 GMT",
    "2014-12-31 12:00:00 GMT", "2015-08-31 12:00:00 GMT", "2013-03-29 12:00:00 GMT", "2014-07-31 12:00:00 GMT"
  )
  end <- c("2017-03-25", "2014-03-16", "2012-12-15", "2011-03-25", "2015-10-16", "2016-03-16", "2014-03-28", "2015-07-22")

  x <- as.POSIXct(start)
  int <- interval(x, end)
  n <- int %/% months(1)
  expect_true(!any(is.na(int %% months(1))))
})

test_that("summary.Interval creates useful summary", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2003-01-01", tz = "UTC")
  int <- interval(time1, time2)

  text <- structure(c("1", "2001-01-01", "2003-01-01", "UTC", "1"),
    .Names = c(
      "Intervals", "Earliest endpoint", "Latest endpoint", "Time zone",
      "NA's"
    )
  )

  expect_equal(summary(c(int, NA)), text)
})

test_that("intersect on intervals propagates NAs (#226)", {
  expect_equal(
    intersect(
      c(
        interval("2014-01-15", "2014-01-16"),
        interval("2014-01-15", "2014-01-16")
      ),
      c(interval(NA, NA), interval(NA, NA))
    ),
    c(interval(NA, NA), interval(NA, NA))
  )
  expect_equal(
    intersect(
      c(
        interval("2014-01-15", "2014-01-16"),
        interval("2014-01-15", "2014-01-16"),
        interval("2014-01-01", "2014-01-30")
      ),
      c(
        interval(NA, NA), interval(NA, NA),
        interval("2014-01-15", "2014-02-16")
      )
    ),
    c(interval(NA, NA), interval(NA, NA), interval("2014-01-15", "2014-01-30"))
  )
})

test_that("Reduce works with intervals (#348)", {
  ints <- c(
    interval(ymd("2001-01-01"), ymd("2002-01-01")),
    interval(ymd("2001-01-01"), ymd("2004-01-01")),
    interval(ymd("2001-02-01"), ymd("2002-01-01"))
  )

  expect_equal(
    Reduce(union, ints),
    interval(ymd("2001-01-01"), ymd("2004-01-01"))
  )
})

test_that("Intervals handles missing numbers", {
  int <- new("Interval",
    .Data = c(NA, NA, NA, 123552000, 71020800, 82425600),
    start = structure(c(NA, NA, NA, 1174953600, 1364428800, 1183334400),
      class = c("POSIXct", "POSIXt"), tzone = "UTC"
    ),
    tzone = "UTC"
  )

  out <- new("Interval",
    .Data = c(NA, NA, NA, 31622400, NA, 31622400),
    start = structure(c(NA, NA, NA, 1199145600, NA, 1199145600),
      tzone = "UTC",
      class = c("POSIXct", "POSIXt")
    ),
    tzone = "UTC"
  )

  expect_equal(intersect(int, interval("2008-01-01", "2009-01-01")), out)

  expect_equal(intersect(int, int), int)
})

context("Intervals")

test_that("interval() returns zero-length vector", {
  x <- interval()
  expect_s4_class(x, "Interval")
  expect_length(x, 0)
  expect_equal(format(x), character())
  expect_output(print(x), "<Interval[0]>", fixed = TRUE)

  x <- interval(POSIXct(), POSIXct())
  expect_s4_class(x, "Interval")
  expect_length(x, 0)

  x <- interval(POSIXct())
  expect_s4_class(x, "Interval")
  expect_length(x, 0)

  x <- interval(NULL, POSIXct())
  expect_s4_class(x, "Interval")
  expect_length(x, 0)
})

test_that("zero-length interval() respects `tzone`", {
  x <- interval(tzone = "America/Los_Angeles")
  expect_equal(x@tzone, "America/Los_Angeles")
  expect_equal(tz(int_start(x)), "America/Los_Angeles")
})

test_that("is.interval works as expected", {
  expect_false(is.interval(234))
  expect_false(is.interval(as.POSIXct("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.interval(as.POSIXlt("2008-08-03 13:01:59", tz = "UTC")))
  expect_false(is.interval(Sys.Date()))
  expect_false(is.interval(minutes(1)))
  expect_false(is.interval(dminutes(1)))
  expect_true(is.interval(interval(
    as.POSIXct("2008-08-03 13:01:59", tz = "UTC"),
    as.POSIXct("2009-08-03 13:01:59", tz = "UTC"))))
})

test_that("Parsing of iso 8601 intervals works", {
  expect_equal(interval("2007-03-01T13:00:00Z/2008-05-11T15:30:00Z"),
               interval("2007-03-01T13:00:00Z", "2008-05-11T15:30:00Z"))
  expect_equal(interval("2007-03-01T13:00:00Z/P1Y2M10DT2H30M"),
               interval("2007-03-01T13:00:00Z",
                        ymd_hms("2007-03-01T13:00:00Z") + period("P1Y2M10DT2H30M")))

  expect_equal(interval("P1Y2M10DT2H30M/2008-05-11T15:30:00Z"),
               interval(ymd_hms("2008-05-11T15:30:00Z") - period("P1Y2M10DT2H30M"),
                        "2008-05-11T15:30:00Z"))

  expect_equal(interval(c("P1Y2M10DT2H30M/2008-05-11T15:30:00Z",
                          "P2H30M/2008-05-11T15:30:00Z")),
               interval(ymd_hms("2008-05-11T15:30:00Z") - period(c("P1Y2M10DT2H30M", "P2H30M")),
                        "2008-05-11T15:30:00Z"))

  expect_equal(interval(c("P1Y2M10DT2H30M/2008-05-11T15:30:00Z",
                          "2001-05-11T15:30:00Z/P2H30M")),
               interval(c(ymd_hms("2008-05-11T15:30:00Z") - period("P1Y2M10DT2H30M"),
                          ymd_hms("2001-05-11T15:30:00Z")),
                        c(ymd_hms("2008-05-11T15:30:00Z"),
                          ymd_hms("2001-05-11T15:30:00Z") + period("P2H30M"))))

})

test_that("interval works as expected", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  int <- interval(time1, time2)
  num <- as.numeric(time2) - as.numeric(time1)

  expect_equal(int@.Data, num)
  expect_equal(int@start, time1)
  expect_is(int, "Interval")

})

test_that("interval handles vector input", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  time3 <- as.POSIXct("2009-08-04 13:01:59", tz = "UTC")
  int <- interval(c(time1, time2), time3)
  num <- as.numeric(time3) -  as.numeric(c(time1, time2))

  expect_equal(int@.Data, num)
  expect_equal(int@start, c(time1, time2))
  expect_is(int, "Interval")

  int2 <- interval(time1, c(time3, time2))
  num2 <- as.numeric(c(time3, time2)) - as.numeric(time1)
  starts <- structure(c(time1, time1), "tzone" = "UTC")

  expect_equal(int2@.Data, num2)
  expect_equal(int2@start, starts)
  expect_is(int2, "Interval")
})


test_that("format.Interval works as expected", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
  int <- interval(time1, time2)

  expect_match(format(int), "2008-08-03 13:01:59 UTC--2009-08-03 13:01:59 UTC")
})

test_that("interval handles character inputs", {
  t1 <- c("2007-01-01")
  t2 <- c("2007-08-01")
  expect_equal(interval(t1, t2),
               new("Interval", 18316800,
                   start = as.POSIXct("2007-01-01", tz = "UTC"),
                   tzone = "UTC"))

  t1 <- c("2007-01-01")
  t2 <- c("2007-08-01 00:01:02")
  expect_equal(interval(t1, t2),
               new("Interval", 18316862,
                   start = as.POSIXct("2007-01-01", tz = "UTC"),
                   tzone = "UTC"))
})


test_that("interval handles POSIXlt inputs", {

  t1 <- parse_date_time2("2007-01-01", "Ymd", lt = TRUE, tz = "America/Los_Angeles")
  t2 <- parse_date_time2("2007-08-01", "Ymd", lt = TRUE, tz = "America/Los_Angeles")

  expect_equal(unclass(interval(t1, t2)),
               unclass(new("Interval", 18313200,
                           start = as.POSIXct("2007-01-01", tz = "America/Los_Angeles"),
                           tzone = "America/Los_Angeles")))

  t1 <- parse_date_time2("2007-01-01", "Ymd", lt = TRUE, tz = "America/Los_Angeles")
  t2 <- parse_date_time2("2007-08-01 00:01:02", "Ymd HMS", lt = TRUE, tz = "America/Los_Angeles")

  expect_equal(unclass(interval(t1, t2)),
               unclass(new("Interval", 18313262,
                           start = as.POSIXct("2007-01-01", tz = "America/Los_Angeles"),
                           tzone = "America/Los_Angeles")))

})

test_that("as.interval works as expected", {
  a <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  pint <- interval(a, a + days(1))
  dint <- interval(a, a + ddays(1))

  expect_equal(pint, as.interval(days(1), a))
  expect_equal(dint, as.interval(ddays(1), a))
})

test_that("as.interval handles vector input", {
    a <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
    b <- as.POSIXct("2009-08-03 13:01:59", tz = "UTC")
    pint <- interval(a, a + days(1:2))
    dint <- interval(a, a + ddays(1:2))
    pint1 <- interval(c(a, b), c(a, b) + days(1:2))
    dint1 <- interval(c(a, b), c(a, b) + ddays(1:2))
    pint2 <- interval(c(a, b), c(a, b) + days(1))
    dint2 <- interval(c(a, b), c(a, b) + ddays(1))

  expect_equal(pint, as.interval(days(1:2), a))
  expect_equal(dint, as.interval(ddays(1:2), a))
  expect_equal(pint1, as.interval(days(1:2), c(a, b)))
  expect_equal(dint1, as.interval(days(1:2), c(a, b)))
  expect_equal(pint2, as.interval(days(1), c(a, b)))
  expect_equal(dint2, as.interval(days(1), c(a, b)))

})

test_that("[<- can subset intervals", {
  ints <- data.frame(spans = c(interval(ymd(20090201), ymd(20090101)),
                               interval(ymd(20090201), ymd(20090101))))
  my_int <- interval(ymd(18800101), ymd(18810101))
  ints[1, 1] <- 31536000
  ints[2, 1] <- my_int
  int2 <- interval(ymd(20090201), ymd(20100201))

  expect_equal(ints[1, 1], int2)
  expect_equal(ints[2, 1], my_int)

})

test_that("interval handles correctly time zones of Date objects", {
  expect_equal(interval(as.Date("2011-01-01"), as.Date("2013-01-01")),
               interval(ymd("2011-01-01"), ymd("2013-01-01")))
})

test_that("summary.Interval creates useful summary", {
  int <- interval(ymd(20090201), ymd(20090101))
  text <- c(3, "2009-01-01", "2009-02-01", "UTC", 1)
  names(text) <- c("Intervals", "Earliest endpoint", "Latest endpoint", "Time zone", "NA's")

  expect_equal(summary(c(int, int, int, as.interval(NA))), text)
})

test_that("as.interval handles NAs", {
  expect_equal(as.interval(NA), interval(NA, NA))
})

test_that("c.interval works as expected", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 15:03:39", tz = "UTC")
  int <- interval(time1, time2)
  ints <- interval(c(time1, time1), c(time2, time2))
  nants <- interval(c(time1, NA), c(time2, NA))

  expect_equal(c(int, int), ints)
  expect_equal(c(int, NA), nants)
})

test_that("%--% works as expected", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 15:03:39", tz = "UTC")
  int <- interval(time1, time2)
  attempt <- time1 %--% time2

  expect_equal(attempt, int)
})

test_that("int_start and int_end work as expected", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- as.POSIXct("2009-08-03 15:03:39", tz = "UTC")
  int1 <- interval(time1, time2)
  int2 <- interval(time2, time1)

  expect_equal(int_start(int1), time1)
  expect_equal(int_start(int2), time2)
  expect_equal(int_start(c(int1, int2)), c(time1, time2))

  expect_equal(int_end(int1), time2)
  expect_equal(int_end(int2), time1)
  expect_equal(int_end(c(int2, int1)), c(time1, time2))
})

test_that("int_length works as expected", {
  time1 <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  time2 <- time1 + 1111
  time3 <- time1 - 5555
  int <- interval(time1, time2)
  int2 <- interval(time1, time3)

  expect_equal(int_length(int), 1111)
  expect_equal(int_length(int2), -5555)
  expect_equal(int_length(c(int, int2)), c(1111, -5555))
})

test_that("int_shift works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  int <- interval(time1, time2)
  int2 <- interval(time2, time1)

  dur <- duration(seconds = 11*24*60*60)
  dur2 <- duration(seconds = -3600)

  dint <- interval(time1 + 11*24*60*60, time2 + 11*24*60*60)
  dint2 <- interval(time2 + 11*24*60*60, time1 + 11*24*60*60)

  nint <- interval(time1 - 3600, time2 - 3600)
  nint2 <- interval(time2 - 3600, time1 - 3600)


  expect_equal(int_shift(int, dur), dint)
  expect_equal(int_shift(int, dur2), nint)
  expect_equal(int_shift(int, c(dur, dur2)), c(dint, nint))
  expect_equal(int_shift(int2, dur), dint2)
  expect_equal(int_shift(int2, dur2), nint2)
  expect_equal(int_shift(int2, c(dur, dur2)), c(dint2, nint2))
  expect_equal(int_shift(c(int, int2), dur), c(dint, dint2))
  expect_equal(int_shift(c(int, int2), dur2), c(nint, nint2))
  expect_equal(int_shift(c(int, int2), c(dur, dur2)), c(dint, nint2))
  expect_equal(int_shift(c(int, int2), c(dur2, dur)), c(nint, dint2))
})

test_that("int_overlaps works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  time21 <- as.POSIXct("2001-06-01", tz = "UTC")
  time22 <- as.POSIXct("2002-06-01", tz = "UTC")
  time31 <- as.POSIXct("2003-01-01", tz = "UTC")
  time32 <- as.POSIXct("2004-01-01", tz = "UTC")
  int1 <- interval(time1, time2)
  int2 <- interval(time21, time22)
  int3 <- interval(time31, time32)

  nint1 <- interval(time2, time1)
  nint2 <- interval(time22, time21)
  nint3 <- interval(time32, time31)


  expect_true(int_overlaps(int1, int2))
  expect_false(int_overlaps(int1, int3))
  expect_equal(int_overlaps(int1, c(int2, int3)), c(TRUE, FALSE))
  expect_equal(int_overlaps(c(int1, int3), int2), c(TRUE, FALSE))

  expect_true(int_overlaps(nint1, nint2))
  expect_false(int_overlaps(nint1, nint3))
  expect_equal(int_overlaps(nint1, c(nint2, nint3)), c(TRUE, FALSE))
  expect_equal(int_overlaps(c(nint1, nint3), nint2), c(TRUE, FALSE))

  expect_true(int_overlaps(int1, nint2))
  expect_true(int_overlaps(nint1, int2))
  expect_false(int_overlaps(int1, nint3))
  expect_false(int_overlaps(nint1, int3))
  expect_equal(int_overlaps(int1, c(int2, nint3)), c(TRUE, FALSE))
  expect_equal(int_overlaps(c(int1, nint3), nint2), c(TRUE, FALSE))

})

test_that("int_aligns works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  time21 <- as.POSIXct("2001-06-01", tz = "UTC")
  time22 <- as.POSIXct("2002-01-01", tz = "UTC")
  time31 <- as.POSIXct("2003-01-01", tz = "UTC")
  time32 <- as.POSIXct("2004-01-01", tz = "UTC")
  int1 <- interval(time1, time2)
  int2 <- interval(time21, time22)
  int3 <- interval(time31, time32)

  nint1 <- interval(time2, time1)
  nint2 <- interval(time22, time21)
  nint3 <- interval(time32, time31)

  expect_true(int_aligns(int1, int2))
  expect_false(int_aligns(int1, int3))
  expect_equal(int_aligns(int1, c(int2, int3)), c(TRUE, FALSE))
  expect_equal(int_aligns(c(int1, int3), int2), c(TRUE, FALSE))

  expect_true(int_aligns(nint1, nint2))
  expect_false(int_aligns(nint1, nint3))
  expect_equal(int_aligns(nint1, c(nint2, nint3)), c(TRUE, FALSE))
  expect_equal(int_aligns(c(nint1, nint3), nint2), c(TRUE, FALSE))

  expect_true(int_aligns(int1, nint2))
  expect_true(int_aligns(nint1, int2))
  expect_false(int_aligns(int1, nint3))
  expect_false(int_aligns(nint1, int3))
  expect_equal(int_aligns(int1, c(int2, nint3)), c(TRUE, FALSE))
  expect_equal(int_aligns(c(int1, nint3), nint2), c(TRUE, FALSE))
})

test_that("int_diff works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  time3 <- as.POSIXct("2003-06-01", tz = "UTC")
  time4 <- as.POSIXct("2002-01-01", tz = "UTC")
  time5 <- as.POSIXct("2003-01-01", tz = "UTC")
  time6 <- as.POSIXct("2004-01-01", tz = "UTC")
  times <- c(time1, time2, time3, time4, time5, time6)
  int1 <- interval(time1, time2)
  int2 <- interval(time2, time3)
  int3 <- interval(time3, time4)
  int4 <- interval(time4, time5)
  int5 <- interval(time5, time6)
  ints <- c(int1, int2, int3, int4, int5)

  expect_equal(int_diff(times), ints)
})

test_that("int_standardize works as expected", {
   int <- interval(ymd(20100101), ymd(20090101))
   standard <- interval(ymd(20090101), ymd(20100101))
   ints <- c(int, int)
   ints2 <- c(int, standard)
   standards <- c(standard, standard)

   expect_equal(int_standardize(int), standard)
   expect_equal(int_standardize(standard), standard)
   expect_equal(int_standardize(ints), standards)
   expect_equal(int_standardize(ints2), standards)
   expect_equal(int_standardize(standards), standards)
})

test_that("int_flip works as expected", {
  int <- interval(ymd(20100101), ymd(20090101))
  flip <- interval(ymd(20090101), ymd(20100101))
  ints <- c(int, int)
  ints2 <- c(int, flip)
  flips <- c(flip, flip)
  flips2 <- c(flip, int)

  expect_equal(int_flip(int), flip)
  expect_equal(int_flip(ints), flips)
  expect_equal(int_flip(ints2), flips2)
  expect_equal(int_flip(flip), int)
  expect_equal(int_flip(flips), ints)
  expect_equal(int_flip(flips2), ints2)
})

test_that("intersect.Interval works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  time21 <- as.POSIXct("2001-06-01", tz = "UTC")
  time22 <- as.POSIXct("2002-06-01", tz = "UTC")
  time31 <- as.POSIXct("2003-01-01", tz = "UTC")
  time32 <- as.POSIXct("2004-01-01", tz = "UTC")
  int1 <- interval(time1, time2)
  int2 <- interval(time21, time22)
  int3 <- interval(time31, time32)

  nint1 <- interval(time2, time1)
  nint2 <- interval(time22, time21)
  nint3 <- interval(time32, time31)

  expect_equal(intersect(int1, int2), interval(time21, time2))
  expect_equal(intersect(int1, int3), interval(NA, NA, tzone = "UTC"))
  expect_equal(intersect(int1, c(int2, int3)), interval(c(time21, NA), c(time2, NA)))
  expect_equal(intersect(c(int1, int3), int2), interval(c(time21, NA), c(time2, NA)))
  expect_equal(intersect(c(int1, int3), c(int2, int2)), interval(c(time21, NA), c(time2, NA)))

  expect_equal(intersect(nint1, nint2), interval(time2, time21))
  expect_equal(intersect(nint1, nint3), interval(NA, NA, tzone = "UTC"))
  expect_equal(intersect(nint1, c(nint2, nint3)), interval(c(time2, NA), c(time21, NA)))
  expect_equal(intersect(c(nint1, nint3), nint2), interval(c(time2, NA), c(time21, NA)))
  expect_equal(intersect(c(nint1, nint3), c(nint2, nint2)), interval(c(time2, NA), c(time21, NA)))

  expect_equal(intersect(int1, nint2), interval(time21, time2))
  expect_equal(intersect(nint1, nint2), interval(time2, time21))
  expect_equal(intersect(int1, nint3), interval(NA, NA, tzone = "UTC"))
  expect_equal(intersect(nint1, int3), interval(NA, NA, tzone = "UTC"))
  expect_equal(intersect(int1, c(nint2, int3)), interval(c(time21, NA), c(time2, NA)))
  expect_equal(intersect(nint1, c(int2, int3)), interval(c(time2, NA), c(time21, NA)))
  expect_equal(intersect(c(int1, nint3), nint2), interval(c(time21, NA), c(time2, NA)))
  expect_equal(intersect(c(nint1, int3), nint2), interval(c(time2, NA), c(time21, NA)))
  expect_equal(intersect(c(int1, nint3), int2), interval(c(time21, NA), c(time2, NA)))
  expect_equal(intersect(c(nint1, int3), int2), interval(c(time2, NA), c(time21, NA)))
  expect_equal(intersect(c(int1, nint3), c(nint2, int2)), interval(c(time21, NA), c(time2, NA)))
  expect_equal(intersect(c(nint1, int3), c(int2, nint2)), interval(c(time2, NA), c(time21, NA)))
})

test_that("union.Interval works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  time21 <- as.POSIXct("2001-06-01", tz = "UTC")
  time22 <- as.POSIXct("2002-06-01", tz = "UTC")
  time31 <- as.POSIXct("2003-01-01", tz = "UTC")
  time32 <- as.POSIXct("2004-01-01", tz = "UTC")
  int1 <- interval(time1, time2)
  int2 <- interval(time21, time22)
  int3 <- interval(time31, time32)

  nint1 <- interval(time2, time1)
  nint2 <- interval(time22, time21)
  nint3 <- interval(time32, time31)

  expect_equal(suppressMessages(union(int1, int2)),
               interval(time1, time22))
  expect_equal(suppressMessages(union(int1, int3)),
               interval(time1, time32))
  expect_equal(suppressMessages(union(int3, int1)),
               interval(time1, time32))
  expect_equal(suppressMessages(union(int1, c(int2, int3))),
               interval(c(time1, time1), c(time22, time32)))
  expect_equal(suppressMessages(union(c(int1, int3), int2)),
               interval(c(time1, time21), c(time22, time32)))
  expect_equal(suppressMessages(union(c(int1, int3), c(int2, int1))),
               interval(c(time1, time1), c(time22, time32)))

  expect_equal(suppressMessages(union(nint1, nint2)),
               interval(time22, time1))
  expect_equal(suppressMessages(union(nint1, nint3)),
               interval(time32, time1))
  expect_equal(suppressMessages(union(nint3, nint1)),
               interval(time32, time1))
  expect_equal(suppressMessages(union(nint1, c(nint2, nint3))),
               interval(c(time22, time32), c(time1, time1)))
  expect_equal(suppressMessages(union(c(nint1, nint3), nint2)),
               interval(c(time22, time32), c(time1, time21)))
  expect_equal(suppressMessages(union(c(nint1, nint3), c(nint2, nint1))),
               interval(c(time22, time32), c(time1, time1)))


  expect_equal(suppressMessages(union(int1, nint2)),
               interval(time1, time22))
  expect_equal(suppressMessages(union(nint1, int2)),
               interval(time22, time1))
  expect_equal(suppressMessages(union(int1, nint3)),
               interval(time1, time32))
  expect_equal(suppressMessages(union(nint1, int3)),
               interval(time32, time1))
  expect_equal(suppressMessages(union(int3, nint1)),
               interval(time1, time32))
  expect_equal(suppressMessages(union(nint3, int1)),
               interval(time32, time1))
  expect_equal(suppressMessages(union(int1, c(nint2, nint3))),
               interval(c(time1, time1), c(time22, time32)))
  expect_equal(suppressMessages(union(nint1, c(int2, int3))),
               interval(c(time22, time32), c(time1, time1)))
  expect_equal(suppressMessages(union(nint1, c(int2, nint3))),
               interval(c(time22, time32), c(time1, time1)))
  expect_equal(suppressMessages(union(c(int1, int3), nint2)),
               interval(c(time1, time21), c(time22, time32)))
  expect_equal(suppressMessages(union(c(nint1, nint3), int2)),
               interval(c(time22, time32), c(time1, time21)))
  expect_equal(suppressMessages(union(c(int1, nint3), nint2)),
               interval(c(time1, time32), c(time22, time21)))
  expect_equal(suppressMessages(union(c(int1, int3), c(nint2, nint1))),
               interval(c(time1, time1), c(time22, time32)))
  expect_equal(suppressMessages(union(c(nint1, nint3), c(int2, int1))),
               interval(c(time22, time32), c(time1, time1)))
  expect_equal(suppressMessages(union(c(int1, nint3), c(nint2, int1))),
               interval(c(time1, time32), c(time22, time1)))
})

test_that("setdiff.Interval works as expected", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2002-01-01", tz = "UTC")
  time21 <- as.POSIXct("2001-06-01", tz = "UTC")
  time22 <- as.POSIXct("2002-06-01", tz = "UTC")
  time31 <- as.POSIXct("2003-01-01", tz = "UTC")
  time32 <- as.POSIXct("2004-01-01", tz = "UTC")
  int1 <- interval(time1, time2)
  int2 <- interval(time21, time22)
  int3 <- interval(time31, time32)

  nint1 <- interval(time2, time1)
  nint2 <- interval(time22, time21)
  nint3 <- interval(time32, time31)

  arrive <- ymd_hms("2011-06-04 12:00:00",
    tz = "Pacific/Auckland")
  leave <- ymd_hms("2011-08-10 14:00:00",
    tz = "Pacific/Auckland")
  auckland <- interval(arrive, leave)
  jsm2 <- interval(ymd(20110720, tz = "Pacific/Auckland"),
    ymd(20110731, tz = "Pacific/Auckland"))

  expect_equal(setdiff(int1, int2), interval(time1, time21))
  expect_equal(setdiff(int1, int3), int1)
  expect_equal(setdiff(int1, c(int2, int3)), c(interval(time1, time21), int1))
  expect_equal(setdiff(c(int1, int3), int2), c(interval(time1, time21), int3))
  expect_equal(setdiff(c(int1, int3), c(int2, int2)), c(interval(time1, time21), int3))

  expect_equal(setdiff(nint1, nint2), interval(time21, time1))
  expect_equal(setdiff(nint1, nint3), nint1)
  expect_equal(setdiff(nint1, c(nint2, nint3)), c(interval(time21, time1), nint1))
  expect_equal(setdiff(c(nint1, nint3), nint2), c(interval(time21, time1), nint3))
  expect_equal(setdiff(c(nint1, nint3), c(nint2, nint2)), c(interval(time21, time1), nint3))

  expect_equal(setdiff(int1, nint2), interval(time1, time21))
  expect_equal(setdiff(nint1, int2), interval(time21, time1))
  expect_equal(setdiff(int1, nint3), int1)
  expect_equal(setdiff(nint1, int3), nint1)
  expect_equal(setdiff(int1, c(nint2, nint3)), c(interval(time1, time21), int1))
  expect_equal(setdiff(nint1, c(int2, int3)), c(interval(time21, time1), nint1))
  expect_equal(setdiff(int1, c(nint2, int3)), c(interval(time1, time21), int1))
  expect_equal(setdiff(nint1, c(int2, nint3)), c(interval(time21, time1), nint1))
  expect_equal(setdiff(c(int1, int3), nint2), c(interval(time1, time21), int3))
  expect_equal(setdiff(c(nint1, nint3), int2), c(interval(time21, time1), nint3))
  expect_equal(setdiff(c(nint1, int3), nint2), c(interval(time21, time1), int3))
  expect_equal(setdiff(c(int1, int3), c(nint2, nint2)), c(interval(time1, time21), int3))
  expect_equal(setdiff(c(nint1, nint3), c(int2, int2)), c(interval(time21, time1), nint3))
  expect_equal(setdiff(c(int1, nint3), c(nint2, int2)), c(interval(time1, time21), nint3))
  expect_error(setdiff(auckland, jsm2),
    "Cases 1 result in discontinuous intervals.")

})

test_that("%within% fails meaningfully when applied on wrong types", {
  expect_error(now() %within% now(), "No %within% method")
})

test_that("%within% works as expected", {
    time1 <- as.POSIXct("2001-01-01", tz = "UTC")
    time2 <- as.POSIXct("2003-01-01", tz = "UTC")
    time3 <- as.POSIXct("2001-06-01", tz = "UTC")
    time4 <- as.POSIXct("2002-06-01", tz = "UTC")
    time5 <- as.POSIXct("2003-01-01", tz = "UTC")
    time6 <- as.POSIXct("2004-01-01", tz = "UTC")
    time7 <- as.POSIXct("2003-01-02", tz = "UTC")
    base <- interval(time1, time2)
    ins <- interval(time3, time4)
    bord <- interval(time5, time6)
    olap <- interval(time4, time6)
    outs <- interval(time7, time6)

    nbase <- interval(time2, time1)
    nins <- interval(time4, time3)
    nbord <- interval(time6, time5)
    nolap <- interval(time6, time4)
    nouts <- interval(time6, time7)

    expect_true(ins %within% base)
    expect_false(outs %within% base)
    expect_false(bord %within% base)
    expect_false(olap %within% base)

    expect_true(nins %within% nbase)
    expect_false(nouts %within% nbase)
    expect_false(nbord %within% nbase)
    expect_false(nolap %within% nbase)

    expect_true(ins %within% nbase)
    expect_false(outs %within% nbase)
    expect_false(bord %within% nbase)
    expect_false(olap %within% nbase)

    expect_true(nins %within% base)
    expect_false(nouts %within% base)
    expect_false(nbord %within% base)
    expect_false(nolap %within% base)

})


test_that("%with% recycles both arguments", {
  blackouts<- c(interval(ymd("2014-12-30"), ymd("2014-12-31")),
                interval(ymd("2014-12-30"), ymd("2015-01-03")))
  testdates <-c(ymd("2014-12-20", ymd("2014-12-30"), ymd("2015-01-01"), ymd("2015-01-03")))
  expect_equal(testdates %within% blackouts, c(F, T, F, T))
})

test_that("%with% works with list of intervals", {

  testdates <-ymd(c("2014-12-20", "2014-12-30", "2015-01-01", "2015-01-03"))
  blackouts<- list(interval(ymd("2014-12-30"), ymd("2014-12-31")),
                   interval(ymd("2014-12-30"), ymd("2015-01-03")))
  expect_equal(testdates %within% blackouts, c(F, T, T, T))
  testdates <-c(ymd(c("2014-12-20", "2014-12-30", "2015-01-01", "2015-01-03"), tz = "UTC"))
  blackouts<- list(interval(ymd("2014-12-30"), ymd("2014-12-31")),
                   interval(ymd("2014-12-30"), ymd("2015-01-03")))
  expect_equal(testdates %within% blackouts, c(F, T, T, T))

})

test_that("%% on interval uses m+ arithmetic", {

  ## From https://github.com/tidyverse/lubridate/issues/633
  start <- c("2016-04-29 12:00:00 GMT", "2013-10-31 12:00:00 GMT", "2012-05-31 12:00:00 GMT", "2010-06-29 12:00:00 GMT",
             "2014-12-31 12:00:00 GMT", "2015-08-31 12:00:00 GMT", "2013-03-29 12:00:00 GMT", "2014-07-31 12:00:00 GMT")
  end <- c("2017-03-25", "2014-03-16", "2012-12-15", "2011-03-25", "2015-10-16", "2016-03-16", "2014-03-28", "2015-07-22")

  x <- as.POSIXct(start)
  int <- interval(x, end)
  n <- int %/% months(1)
  expect_true(!any(is.na(int %% months(1))))

})

test_that("summary.Interval creates useful summary", {
  time1 <- as.POSIXct("2001-01-01", tz = "UTC")
  time2 <- as.POSIXct("2003-01-01", tz = "UTC")
  int <- interval(time1, time2)

  text <- structure(c("1", "2001-01-01", "2003-01-01", "UTC", "1"),
    .Names = c("Intervals", "Earliest endpoint", "Latest endpoint", "Time zone",
               "NA's"))

  expect_equal(summary(c(int, NA)), text)
})

test_that("intersect on intervals propagates NAs (#226)", {
  expect_equal(intersect(c(interval("2014-01-15", "2014-01-16"),
                           interval("2014-01-15", "2014-01-16")),
                         c(interval(NA, NA), interval(NA, NA))),
               c(interval(NA, NA), interval(NA, NA)))
  expect_equal(intersect(c(interval("2014-01-15", "2014-01-16"),
                           interval("2014-01-15", "2014-01-16"),
                           interval("2014-01-01", "2014-01-30")),
                         c(interval(NA, NA), interval(NA, NA),
                           interval("2014-01-15", "2014-02-16"))),
               c(interval(NA, NA), interval(NA, NA), interval("2014-01-15", "2014-01-30")))
})

test_that("Reduce works with intervals (#348)", {

  ints <- c(
    interval(ymd("2001-01-01"), ymd("2002-01-01")),
    interval(ymd("2001-01-01"), ymd("2004-01-01")),
    interval(ymd("2001-02-01"), ymd("2002-01-01"))
  )

  expect_equal(Reduce(union, ints),
               interval(ymd("2001-01-01"), ymd("2004-01-01")))

})

test_that("Intervals handles missing numbers", {

  int <- new("Interval"
           , .Data = c(NA, NA, NA, 123552000, 71020800, 82425600)
           , start = structure(c(NA, NA, NA, 1174953600, 1364428800, 1183334400),
                               class = c("POSIXct", "POSIXt"), tzone = "UTC")
           , tzone = "UTC")

  out <- new("Interval"
           , .Data = c(NA, NA, NA, 31622400, NA, 31622400)
           , start = structure(c(NA, NA, NA, 1199145600, NA, 1199145600), tzone = "UTC",
                               class = c("POSIXct", "POSIXt"))
           , tzone = "UTC")

  expect_equal(intersect(int, interval("2008-01-01", "2009-01-01")), out)

  expect_equal(intersect(int, int), int)

})

#' Cyclic encoding of date-times
#'
#' Encode a date-time object into a cyclic coordinate system in which the
#' distances between two pairs of dates separated by the same time duration are
#' the same.
#'
#' Machine learning models don't know that December 31st and January 1st are
#' close in our human calendar sense. `cyclic_encoding` makes it obvious to the
#' machine learner that two calendar dates are close by mapping the dates onto
#' the circle.
#'
#' @param x a date-time object
#' @param periods a character vector of periods. Follows same specification as
#'   [period] and [floor_date] functions.
#' @param encoders names of functions to produce the encoding. Defaults to
#'   "sin" and "cos". Names of any predefined functions accepting a numeric
#'   input are allowed.
#' @param week_start day starting a week (used for weekly periods).
#' @return a numeric matrix with number of columns equal `length(periods) *
#'   length(types)`.
#' @examples
#'
#' times <- ymd_hms("2019-01-01 00:00:00") + hours(0:23)
#' cyclic_encoding(times, c("day", "week", "month"))
#' plot(cyclic_encoding(times, "1d"))
#' plot(cyclic_encoding(times, "2d"), xlim = c(-1, 1))
#' plot(cyclic_encoding(times, "4d"), xlim = c(-1, 1))
#'
#' @export
cyclic_encoding <- function(x, periods, encoders = c("sin", "cos"),
                             week_start = getOption("lubridate.week.start", 7)) {
  Ne <- length(encoders)
  Np <- length(periods)
  out <- matrix(0, nrow = length(x), ncol = Ne * Np)
  colnames(out) <- paste(rep.int(encoders, Np),
                         rep(periods, each = Ne),
                         sep = ".")
  for (pix in seq_along(periods)) {
    p <- periods[[pix]]
    beg <- as.numeric(floor_date(x, unit = p, week_start = week_start))
    end <- as.numeric(ceiling_date(x, unit = p, week_start = week_start))
    diff <- (2*pi)*(as.numeric(x) - beg)/(end - beg)
    diff[end == beg] <- 0
    for (tix in seq_along(encoders)) {
      out[, (pix - 1)*Ne + tix] <- do.call(encoders[[tix]], list(diff))
    }
  }
  out
}

#' @include periods.r
NULL

#' Get/set seconds component of a date-time
#'
#' Date-time must be a  POSIXct, POSIXlt, Date, Period, chron, yearmon, yearqtr, zoo,
#' zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
#'
#' @export
#' @param x a date-time object
#' @return the seconds element of x as a decimal number
#' @keywords utilities manip chron methods
#' @examples
#' x <- ymd("2012-03-26")
#' second(x)
#' second(x) <- 1
#' second(x) <- 61
#' second(x) > 2
second <- function(x)
  UseMethod("second")

#' @export
second.default <- function(x)
  as.POSIXlt(x, tz = tz(x))$sec

#' @export
second.Period <- function(x)
  slot(x, ".Data")

#' @rdname second
#' @param value numeric value to be assigned
#' @export
"second<-" <- function(x, value)
  x <- x + seconds(value - second(x))


setGeneric("second<-")

#' @export
setMethod("second<-", signature("Period"), function(x, value) {
  slot(x, ".Data") <- value
  x
})

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
#' @include Dates.r
#' @include difftimes.r
#' @include numeric.r
#' @include POSIXt.r
#' @include ops-addition.r
NULL

#' Add and subtract months to a date without exceeding the last day of the new month
#'
#' Adding months frustrates basic arithmetic because consecutive months have
#' different lengths. With other elements, it is helpful for arithmetic to
#' perform automatic roll over. For example, 12:00:00 + 61 seconds becomes
#' 12:01:01. However, people often prefer that this behavior NOT occur with
#' months. For example, we sometimes want January 31 + 1 month = February 28 and
#' not March 3. \code{\%m+\%} performs this type of arithmetic. Date \code{\%m+\%} months(n)
#' always returns a date in the nth month after Date. If the new date would
#' usually spill over into the n + 1th month, \code{\%m+\%} will return the last day of
#' the nth month ([rollback()]). Date \code{\%m-\%} months(n) always returns a
#' date in the nth month before Date.
#'
#' \code{\%m+\%} and \code{\%m-\%} handle periods with components less than a month by first
#' adding/subtracting months and then performing usual arithmetics with smaller
#' units.
#'
#' \code{\%m+\%} and \code{\%m-\%} should be used with caution as they are not one-to-one
#' operations and results for either will be sensitive to the order of
#' operations.
#'
#' @rdname mplus
#' @usage e1 \%m+\% e2
#' @aliases m+ %m+% m- %m-% %m+%,ANY,ANY-method %m-%,ANY,ANY-method %m+%,Period,ANY-method %m+%,ANY,Period-method %m-%,Period,ANY-method %m-%,ANY,Period-method %m+%,Duration,ANY-method %m+%,ANY,Duration-method %m-%,Duration,ANY-method %m-%,ANY,Duration-method %m+%,Interval,ANY-method %m+%,ANY,Interval-method %m-%,Interval,ANY-method %m-%,ANY,Interval-method
#' @param e1 A period or a date-time object of class [POSIXlt], [POSIXct]
#' or [Date].
#' @param e2 A period or a date-time object of class [POSIXlt], [POSIXct]
#' or [Date]. Note that one of e1 and e2 must be a period and the other a
#' date-time object.
#' @return A date-time object of class POSIXlt, POSIXct or Date
#' @examples
#' jan <- ymd_hms("2010-01-31 03:04:05")
#' jan + months(1:3) # Feb 31 and April 31 returned as NA
#' # NA "2010-03-31 03:04:05 UTC" NA
#' jan %m+% months(1:3) # No rollover
#'
#' leap <- ymd("2012-02-29")
#' "2012-02-29 UTC"
#' leap %m+% years(1)
#' leap %m+% years(-1)
#' leap %m-% years(1)
#' @export
"%m+%" <- function(e1, e2) standardGeneric("%m+%")

#' @export
setGeneric("%m+%")

#' @export
setMethod("%m+%", signature(e2 = "Period"),
          function(e1, e2) add_with_rollback(e1, e2))

#' @export
setMethod("%m+%", signature(e1 = "Period"),
          function(e1, e2) add_with_rollback(e2, e1))

#' @export
setMethod("%m+%", signature(e2 = "ANY"),
          function(e1, e2)
            stop("%m+% handles only Period objects as second argument"))

#' @export
"%m-%" <- function(e1, e2) standardGeneric("%m-%")

#' @export
setGeneric("%m-%")

#' @export
setMethod("%m-%", signature(e2 = "Period"),
          function(e1, e2) add_with_rollback(e1, -e2))

#' @export
setMethod("%m-%", signature(e1 = "Period"),
          function(e1, e2) add_with_rollback(e2, -e1))

#' @export
setMethod("%m-%", signature(e2 = "ANY"),
          function(e1, e2)
            stop("%m-% handles only Period objects as second argument"))

#' `add_with_rollback()` is like \code{\%m+\%} and \code{\%m-\%} with more
#' control over the rollback process. It allows the rollback to first day of the
#' month instead of the last day of the previous month and controls whether HMS
#' component of the end date is preserved or not. Please note that the rollback
#' kicks in only when the resulting date lands on the non-existing date.
#' @rdname mplus
#' @param roll_to_first rollback to the first day of the month instead of the
#'   last day of the previous month (passed to [rollback()])
#' @param preserve_hms retains the same hour, minute, and second information? If
#'   FALSE, the new date will be at 00:00:00 (passed to [rollback()])
#' @examples
#'
#' x <- ymd_hms("2019-01-29 01:02:03")
#' add_with_rollback(x, months(1))
#' add_with_rollback(x, months(1), preserve_hms = FALSE)
#' add_with_rollback(x, months(1), roll_to_first = TRUE)
#' add_with_rollback(x, months(1), roll_to_first = TRUE, preserve_hms = FALSE)
#' @export
add_with_rollback <- function(e1, e2, roll_to_first = FALSE, preserve_hms = TRUE) {

  any_HMS <- any(e2@.Data != 0) || any(e2@minute != 0) || any(e2@hour != 0) || any(e2@day != 0)
  any_year <- any(e2@year != 0)
  if (!is.na(any_year) && any_year) {
    e2$month <- 12 * e2@year + e2@month
    e2$year <- 0L
  }

  new <- .quick_month_add(e1, e2@month)
  roll <- day(new) < day(e1)
  roll <- !is.na(roll) & roll
  new[roll] <- rollback(new[roll], roll_to_first = roll_to_first, preserve_hms = preserve_hms)

  if (!is.na(any_HMS) && any_HMS) {
    e2$month <- 0L
    new + e2
  } else {
    new
  }
}

.quick_month_add <- function(object, mval) {
  tzs <- tz(object)
  utc <- as.POSIXlt(force_tz(object, tzone = "UTC"))
  utc$mon <- utc$mon + mval
  utc <- as.POSIXct(utc)
  new <- force_tz(utc, tzone = tzs, roll = TRUE)
  reclass_date(new, object)
}

#' Roll backward or forward a date the previous, current or next month
#'
#' `rollbackward()` changes a date to the last day of the previous month or to
#' the first day of the month. `rollforward()` rolls to the last day of the
#' current month or to the first day of the next month. Optionally, the new date
#' can retain the same hour, minute, and second information. `rollback()` is a
#' synonym for `rollbackward()`.
#'
#' @export
#' @param dates A POSIXct, POSIXlt or Date class object.
#' @param roll_to_first Rollback to the first day of the month instead of the
#' last day of the month
#' @param preserve_hms Retains the same hour, minute, and second information? If
#' FALSE, the new date will be at 00:00:00.
#' @return A date-time object of class POSIXlt, POSIXct or Date, whose day has
#' been adjusted to the last day of the previous month, or to the first day of
#' the month.
#' @examples
#' date <- ymd("2010-03-03")
#' rollbackward(date)
#'
#' dates <- date + months(0:2)
#' rollbackward(dates)
#'
#' date <- ymd_hms("2010-03-03 12:44:22")
#' rollbackward(date)
#' rollbackward(date, roll_to_first = TRUE)
#' rollbackward(date, preserve_hms = FALSE)
#' rollbackward(date, roll_to_first = TRUE, preserve_hms = FALSE)
rollbackward <- function(dates, roll_to_first = FALSE, preserve_hms = TRUE) {
  .roll(dates, roll_to_first, preserve_hms)
}

#' @rdname rollbackward
#' @export
rollback <- rollbackward

#' @rdname rollbackward
#' @export
rollforward <- function(dates, roll_to_first = FALSE, preserve_hms = TRUE) {
  .roll(dates, roll_to_first, preserve_hms, forward = TRUE)
}

.roll <- function(dates, roll_to_first, preserve_hms, forward = FALSE) {
  if (length(dates) == 0)
    return(dates)
  day(dates) <- 1
  if (!preserve_hms) {
    hour(dates) <- 0
    minute(dates) <- 0
    second(dates) <- 0
  }
  if (forward) {
    dates <- dates + months(1)
  }
  if (roll_to_first) {
    dates
  } else {
    dates - days(1)
  }
}

#' @include accessors-month.r
NULL

#' Get the fiscal quarter and semester of a date-time
#'
#' Quarters divide the year into fourths. Semesters divide the year into halfs.
#'
#' @param x a date-time object of class POSIXct, POSIXlt, Date, chron, yearmon,
#'   yearqtr, zoo, zooreg, timeDate, xts, its, ti, jul, timeSeries, fts or
#'   anything else that can be converted with as.POSIXlt
#' @param type the format to be returned for the quarter. Can be one one of
#'   "quarter" - return numeric quarter (default), "year.quarter" return
#'   fractional numeric year.quarter, "date_first" or "date_last" which return
#'   the date at the quarter's start end end.
#' @param fiscal_start numeric indicating the starting month of a fiscal year.
#' @param with_year logical indicating whether or not to include the quarter or
#'   semester's year (deprecated; use the `type` parameter instead).
#' @return numeric or a vector of class POSIXct if `type` argument is
#'   `date_first` or `date_last`
#' @examples
#' x <- ymd(c("2012-03-26", "2012-05-04", "2012-09-23", "2012-12-31"))
#' quarter(x)
#' quarter(x, type = "year.quarter")
#' quarter(x, type = "year.quarter", fiscal_start = 11)
#' quarter(x, type = "date_first", fiscal_start = 11)
#' quarter(x, type = "date_last", fiscal_start = 11)
#' semester(x)
#' semester(x, with_year = TRUE)
#' @export
quarter <- function(x, type = "quarter", fiscal_start = 1, with_year = identical(type, "year.quarter")) {
  if (length(fiscal_start) > 1) {
    stop("`fiscal_start` must be a singleton", call. = FALSE)
  }
  fs <- (fiscal_start - 1) %% 12
  shifted <- seq(fs, 11 + fs) %% 12 + 1
  m <- month(x)
  quarters <- rep(1:4, each = 3)
  s <- match(m, shifted)
  q <- quarters[s]

  ## Doing this to handle positional calls where previously `with_year` was the
  ## second param, and also now to handle soft-deprecation of `with_year`.
  if (is.logical(type)) {
    type <- if (type) "year.quarter" else "quarter"
  }
  if (with_year == TRUE) {
    type <- "year.quarter"
  }

  switch(type,
    "quarter" = q,
    "year.quarter" = {
      nxt_year_months <- if (fs != 0) (fs + 1):12
      year(x) + (m %in% nxt_year_months) + (q / 10)
    },
    "date_first" = ,
    "date_last" = {
      starting_months <- shifted[seq(1, length(shifted), 3)]
      final_years <- year(x) - (starting_months[q] > m)
      quarter_starting_dates <-
        make_date(year = final_years, month = starting_months[q], day = 1L)
      if (type == "date_first") {
        quarter_starting_dates
      } else if (type == "date_last") {
        add_with_rollback(quarter_starting_dates, months(3)) - days(1)
      }
    },
    stop("Unsuported type ", type)
  )
}

#' @rdname quarter
#' @export
semester <- function(x, with_year = FALSE) {
  m <- month(x)
  semesters <- rep(1:2, each = 6)
  s <- semesters[m]
  if (with_year) {
    year(x) + s / 10
  } else {
    s
  }
}

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
#' @include difftimes.r
#' @include numeric.r
NULL


divide_duration_by_duration <- function(dur1, dur2) {
  dur1@.Data / dur2@.Data
}

divide_duration_by_difftime <- function(dur, diff) {
  dur@.Data / as.numeric(diff, units = "secs")
}

divide_duration_by_number <- function(dur, num) {
  new("Duration", dur@.Data / num)
}

divide_interval_by_duration <- function(int, dur) {
  int@.Data / dur@.Data
}

adjust_estimate <- function(est, int, per) {
  start <- int_start(int)
  end <- int_end(int)
  est <- ceiling(est)

  up_date <- add_with_rollback(start, est * per)
  while (length(which <- which(up_date < end))) {
    est[which] <- est[which] + 1
    up_date[which] <- add_with_rollback(up_date[which], est[which] * per[which])
  }

  low_date <- up_date
  while (length(which <- which(low_date > end))) {
    est[which] <- est[which] - 1
    up_date[which] <- low_date[which]
    low_date[which] <- add_with_rollback(start[which], est[which] * per[which])
  }

  frac <-
    as.numeric(difftime(end, low_date, units = "secs")) /
      as.numeric(difftime(up_date, low_date, units = "secs"))

  frac[low_date == up_date] <- 0

  est + frac
}

divide_interval_by_period <- function(int, per) {
  if (length(int) == 0 || length(per) == 0) {
    return(numeric())
  }

  estimate <- int / as.duration(per)
  not_nas <- !is.na(estimate)
  timespans <- match_lengths(int, per)
  if (all(not_nas)) {
    adjust_estimate(estimate, timespans[[1]], timespans[[2]])
  } else {
    int2 <- timespans[[1]][not_nas]
    per2 <- timespans[[2]][not_nas]
    estimate[not_nas] <- adjust_estimate(estimate[not_nas], int2, per2)
    estimate
  }
}

divide_interval_by_difftime <- function(int, diff) {
  int@.Data / as.double(diff, units = "secs")
}

divide_interval_by_number <- function(int, num) {
  starts <- int@start + rep(0, length(num))
  new("Interval", int@.Data / num, start = starts, tzone = int@tzone)
}


divisible_period <- function(per, anchor) {
  per@month <- per@month + 12 * per@year
  per@year <- rep(0, length(per@year))

  secs.in.months <- as.numeric(anchor + months(per@month)) - as.numeric(anchor)
  days.in.months <- round(secs.in.months / 86400)

  per@day <- per@day + days.in.months
  per@month <- 0

  per
}


divide_period_by_period <- function(per1, per2) {
  period_to_seconds(per1) / period_to_seconds(per2)
}

divide_period_by_number <- function(per, num) {
  new("Period", per@.Data / num,
    year = per@year / num,
    month = per@month / num, day = per@day / num, hour = per@hour / num,
    minute = per@minute / num
  )
}



divide_difftime_by_duration <- function(dif, dur) {
  as.numeric(dif, units = "secs") / dur@.Data
}






#' @export
setMethod(
  "/", signature(e1 = "Duration", e2 = "Duration"),
  function(e1, e2) divide_duration_by_duration(e1, e2)
)

#' @export
setMethod("/", signature(e1 = "Duration", e2 = "Interval"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod("/", signature(e1 = "Duration", e2 = "Period"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod(
  "/", signature(e1 = "Duration", e2 = "difftime"),
  function(e1, e2) divide_duration_by_difftime(e1, e2)
)

#' @export
setMethod(
  "/", signature(e1 = "Duration", e2 = "numeric"),
  function(e1, e2) divide_duration_by_number(e1, e2)
)


#' @export
setMethod(
  "/", signature(e1 = "Interval", e2 = "Duration"),
  function(e1, e2) divide_interval_by_duration(e1, e2)
)

#' @export
setMethod("/", signature(e1 = "Interval", e2 = "Interval"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod(
  "/", signature(e1 = "Interval", e2 = "Period"),
  function(e1, e2) divide_interval_by_period(e1, e2)
)

#' @export
setMethod(
  "/", signature(e1 = "Interval", e2 = "difftime"),
  function(e1, e2) divide_interval_by_difftime(e1, e2)
)

#' @export
setMethod(
  "/", signature(e1 = "Interval", e2 = "numeric"),
  function(e1, e2) divide_interval_by_number(e1, e2)
)


#' @export
setMethod("/", signature(e1 = "Period", e2 = "Duration"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod("/", signature(e1 = "Period", e2 = "Interval"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod(
  "/", signature(e1 = "Period", e2 = "Period"),
  function(e1, e2) divide_period_by_period(e1, e2)
)

#' @export
setMethod("/", signature(e1 = "Period", e2 = "difftime"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod(
  "/", signature(e1 = "Period", e2 = "numeric"),
  function(e1, e2) divide_period_by_number(e1, e2)
)



#' @export
setMethod(
  "/", signature(e1 = "difftime", e2 = "Duration"),
  function(e1, e2) divide_difftime_by_duration(e1, e2)
)

#' @export
setMethod("/", signature(e1 = "difftime", e2 = "Interval"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod("/", signature(e1 = "difftime", e2 = "Period"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})



#' @export
setMethod(
  "/", signature(e1 = "numeric", e2 = "Duration"),
  function(e1, e2) stop_incompatible_classes(e1, e2, "/")
)
#' @export
setMethod(
  "/", signature(e1 = "numeric", e2 = "Interval"),
  function(e1, e2) stop_incompatible_classes(e1, e2, "/")
)
#' @export
setMethod(
  "/", signature(e1 = "numeric", e2 = "Period"),
  function(e1, e2) stop_incompatible_classes(e1, e2, "/")
)

##' Deprecated functions in the lubridate package
##' @param x numeric value to be converted into duration
##' @param ... arguments to be passed to the functions
##'     (obscured to enforce the usage of new functions)
##' @rdname Deprecated
##' @keywords internal
##' @name Deprecated-lubridate
NULL

## modified ggplot2:::gg_dep
.deprecated_fun <- function(replacement, version, n_call = 1) {
  name <- as.character(sys.call(-(n_call))[[1]])
  name <- sprintf("'%s'", name[[length(name)]])
  replacement <- sprintf("; use '%s' instead", replacement)
  .deprecated(name, replacement, version)
}

.deprecated_arg <- function(arg, version, n_call = NULL) {
  if (!is.null(n_call)) {
    name <- as.character(sys.call(n_call)[[1]])
    mes <- sprintf("Argument '%s' of '%s'", arg, name)
  } else {
    mes <- sprintf("Argument '%s'", arg)
  }
  .deprecated(mes, "", version)
}

.deprecated <- function(name, replacement = NULL, version) {
  v <- as.package_version(version)
  cv <- packageVersion("lubridate")

  replacement <- if (is.null(replacement)) "" else replacement

  ## If current major number is greater than last-good major number, or if
  ## current minor number is more than 2 greater than last-good minor number,
  ## give error.
  if (cv[[1, 1]] > v[[1, 1]] || cv[[1, 2]] > v[[1, 2]] + 2) {
    stop(sprintf(
      "%s is defunct%s. Deprecated in version '%s'.",
      name, replacement, version
    ),
    call. = FALSE
    )
  } else if (cv[[1, 2]] > v[[1, 2]]) {
    ## If minor number differs by one, give warning
    warning(sprintf(
      "%s is deprecated%s. Deprecated in version '%s'.",
      name, replacement, version
    ),
    call. = FALSE
    )
  } else if (cv[[1, 3]] >= v[[1, 3]]) {
    ## If only subminor number is greater, give message
    message(sprintf(
      "%s is deprecated%s. Deprecated in version '%s'.",
      name, replacement, version
    ))
  }
  invisible()
}

##' Parse dates with **y**ear, **m**onth, and **d**ay components
##'
##' Transforms dates stored in character and numeric vectors to Date or POSIXct
##' objects (see `tz` argument). These functions recognize arbitrary
##' non-digit separators as well as no separator. As long as the order of
##' formats is correct, these functions will parse dates correctly even when the
##' input vectors contain differently formatted dates. See examples.
##'
##' In case of heterogeneous date formats, the `ymd()` family guesses formats based
##' on a subset of the input vector. If the input vector contains many missing
##' values or non-date strings, the subset might not contain meaningful dates
##' and the date-time format won't be guessed resulting in
##' `All formats failed to parse` error. In such cases please see
##' [parse_date_time()] for a more flexible parsing interface.
##'
##' If the `truncated` parameter is non-zero, the `ymd()` functions also check for
##' truncated formats. For example, `ymd()` with `truncated = 2` will also
##' parse incomplete dates like `2012-06` and `2012`.
##'
##' NOTE: The `ymd()` family of functions is based on `parse_date_time()` and thus
##' directly drop to the internal C parser for numeric months, but uses
##' [base::strptime()] for alphabetic months. This implies that some of [base::strptime()]'s
##' limitations are inherited by \pkg{lubridate}'s parser. For example, truncated
##' formats (like `%Y-%b`) will not be parsed. Numeric truncated formats (like
##' `%Y-%m`) are handled correctly by \pkg{lubridate}'s C parser.
##'
##' As of version 1.3.0, \pkg{lubridate}'s parse functions no longer return a
##' message that displays which format they used to parse their input. You can
##' change this by setting the `lubridate.verbose` option to `TRUE` with
##' `options(lubridate.verbose = TRUE)`.
##'
##' @export
##' @param ... a character or numeric vector of suspected dates
##' @param quiet logical. If `TRUE`, function evaluates without displaying
##'   customary messages.
##' @param tz Time zone indicator. If `NULL` (default), a Date object is
##'   returned. Otherwise a POSIXct with time zone attribute set to `tz`.
##' @param locale locale to be used, see [locales]. On Linux systems you
##'   can use `system("locale -a")` to list all the installed locales.
##' @param truncated integer. Number of formats that can be truncated.
##' @return a vector of class POSIXct if `tz` argument is non-`NULL` or Date if tz
##'   is `NULL` (default)
##' @seealso [parse_date_time()] for an even more flexible low level
##'   mechanism.
##' @keywords chron
##' @examples
##' x <- c("09-01-01", "09-01-02", "09-01-03")
##' ymd(x)
##' x <- c("2009-01-01", "2009-01-02", "2009-01-03")
##' ymd(x)
##' ymd(090101, 90102)
##' now() > ymd(20090101)
##' ## TRUE
##' dmy(010210)
##' mdy(010210)
##'
##' yq('2014.2')
##'
##' ## heterogeneous formats in a single vector:
##' x <- c(20090101, "2009-01-02", "2009 01 03", "2009-1-4",
##'        "2009-1, 5", "Created on 2009 1 6", "200901 !!! 07")
##' ymd(x)
##'
##' ## What lubridate might not handle:
##'
##' ## Extremely weird cases when one of the separators is "" and some of the
##' ## formats are not in double digits might not be parsed correctly:
##' \dontrun{ymd("201002-01", "201002-1", "20102-1")
##' dmy("0312-2010", "312-2010")}
ymd <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx(..., orders = "ymd", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd
ydm <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx(..., orders = "ydm", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd
mdy <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx(..., orders = "mdy", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd
myd <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx(..., orders = "myd", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd
dmy <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx(..., orders = "dmy", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd
dym <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx(..., orders = "dym", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd
yq <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME")) {
  .parse_xxx(..., orders = "yq", quiet = quiet, tz = tz, locale = locale, truncated = 0)
}

#' @export
#' @rdname ymd
ym <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME")) {
  .parse_xxx(..., orders = "ym", quiet = quiet, tz = tz, locale = locale, truncated = 0)
}

#' @export
#' @rdname ymd
my <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME")) {
  .parse_xxx(..., orders = "my", quiet = quiet, tz = tz, locale = locale, truncated = 0)
}


##' Parse date-times with **y**ear, **m**onth, and **d**ay, **h**our,
##' **m**inute, and **s**econd components.
##'
##' Transform dates stored as character or numeric vectors to POSIXct
##' objects. The `ymd_hms()` family of functions recognizes all non-alphanumeric
##' separators (with the exception of "." if `frac = TRUE`) and correctly
##' handles heterogeneous date-time representations. For more flexibility in
##' treatment of heterogeneous formats, see low level parser
##' [parse_date_time()].
##'
##' The `ymd_hms()` functions automatically assign the Universal Coordinated Time
##' Zone (UTC) to the parsed date. This time zone can be changed with
##' [force_tz()].
##'
##' The most common type of irregularity in date-time data is the truncation
##' due to rounding or unavailability of the time stamp. If the `truncated`
##' parameter is non-zero, the `ymd_hms()` functions also check for truncated
##' formats. For example, `ymd_hms()` with `truncated = 3` will also parse
##' incomplete dates like `2012-06-01 12:23`, `2012-06-01 12` and
##' `2012-06-01`. NOTE: The `ymd()` family of functions is based on
##' [base::strptime()] which currently fails to parse `%y-%m` formats.
##'
##' In case of heterogeneous date formats the `ymd_hms()` family guesses formats
##' based on a subset of the input vector. If the input vector contains many
##' missing values or non-date strings, the subset might not contain meaningful
##' dates and the date-time format won't be guessed resulting in
##' `All formats failed to parse` error. In such cases please see
##' [parse_date_time()] for a more flexible parsing interface.
##'
##' As of version 1.3.0, \pkg{lubridate}'s parse functions no longer return a
##' message that displays which format they used to parse their input. You can
##' change this by setting the `lubridate.verbose` option to `TRUE` with
##' `options(lubridate.verbose = TRUE)`.
##'
##' @export
##' @param ... a character vector of dates in year, month, day, hour, minute,
##'   second format
##' @param quiet logical. If `TRUE`, function evaluates without displaying customary messages.
##' @param tz a character string that specifies which time zone to parse the date with. The string
##' must be a time zone that is recognized by the user's OS.
##' @param locale locale to be used, see \link{locales}. On Linux systems you
##' can use `system("locale -a")` to list all the installed locales.
##' @param truncated integer, indicating how many formats can be missing. See details.
##' @return a vector of [POSIXct] date-time objects
##' @seealso
##' - [ymd()], [hms()]
##' - [parse_date_time()] for the underlying mechanism
##' @keywords POSIXt parse
##' @examples
##'
##' x <- c("2010-04-14-04-35-59", "2010-04-01-12-00-00")
##' ymd_hms(x)
##' x <- c("2011-12-31 12:59:59", "2010-01-01 12:00:00")
##' ymd_hms(x)
##'
##'
##' ## ** heterogeneous formats **
##' x <- c(20100101120101, "2009-01-02 12-01-02", "2009.01.03 12:01:03",
##'        "2009-1-4 12-1-4",
##'        "2009-1, 5 12:1, 5",
##'        "200901-08 1201-08",
##'        "2009 arbitrary 1 non-decimal 6 chars 12 in between 1 !!! 6",
##'        "OR collapsed formats: 20090107 120107 (as long as prefixed with zeros)",
##'        "Automatic wday, Thu, detection, 10-01-10 10:01:10 and p format: AM",
##'        "Created on 10-01-11 at 10:01:11 PM")
##' ymd_hms(x)
##'
##' ## ** fractional seconds **
##' op <- options(digits.secs=3)
##' dmy_hms("20/2/06 11:16:16.683")
##' options(op)
##'
##' ## ** different formats for ISO8601 timezone offset **
##' ymd_hms(c("2013-01-24 19:39:07.880-0600",
##' "2013-01-24 19:39:07.880", "2013-01-24 19:39:07.880-06:00",
##' "2013-01-24 19:39:07.880-06", "2013-01-24 19:39:07.880Z"))
##'
##' ## ** internationalization **
##' \dontrun{
##' x_RO <- "Ma 2012 august 14 11:28:30 "
##'   ymd_hms(x_RO, locale = "ro_RO.utf8")
##' }
##'
##' ## ** truncated time-dates **
##' x <- c("2011-12-31 12:59:59", "2010-01-01 12:11", "2010-01-01 12", "2010-01-01")
##' ymd_hms(x, truncated = 3)
##' x <- c("2011-12-31 12:59", "2010-01-01 12", "2010-01-01")
##' ymd_hm(x, truncated = 2)
##' ## ** What lubridate might not handle **
##' ## Extremely weird cases when one of the separators is "" and some of the
##' ## formats are not in double digits might not be parsed correctly:
##' \dontrun{
##' ymd_hm("20100201 07-01", "20100201 07-1", "20100201 7-01")}
##'
ymd_hms <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx_hms(..., orders = c("ymdTz", "ymdT"), quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd_hms
ymd_hm <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx_hms(..., orders = "ymdR", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd_hms
ymd_h <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx_hms(..., orders = "ymdr", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd_hms
dmy_hms <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx_hms(..., orders = c("dmyTz", "dmyT"), quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd_hms
dmy_hm <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx_hms(..., orders = "dmyR", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd_hms
dmy_h <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx_hms(..., orders = "dmyr", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd_hms
mdy_hms <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx_hms(..., orders = c("mdyTz", "mdyT"), quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd_hms
mdy_hm <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx_hms(..., orders = "mdyR", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd_hms
mdy_h <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx_hms(..., orders = "mdyr", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd_hms
ydm_hms <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx_hms(..., orders = c("ydmTz", "ydmT"), quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd_hms
ydm_hm <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx_hms(..., orders = "ydmR", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}

#' @export
#' @rdname ymd_hms
ydm_h <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"), truncated = 0) {
  .parse_xxx_hms(..., orders = "ydmr", quiet = quiet, tz = tz, locale = locale, truncated = truncated)
}


##' @rdname hms
##' @examples
##' ms(c("09:10", "09:02", "1:10"))
##' ms("7 6")
##' ms("6,5")
##' @export
ms <- function(..., quiet = FALSE, roll = FALSE) {
  out <- .parse_hms(..., order = "MS", quiet = quiet)
  if (roll) {
    hms <- .roll_hms(min = out["M", ], sec = out["S", ])
    period(hour = hms$hour, minute = hms$min, second = hms$sec)
  } else {
    period(minute = out["M", ], second = out["S", ])
  }
}

##' @rdname hms
##' @examples
##' hm(c("09:10", "09:02", "1:10"))
##' hm("7 6")
##' hm("6,5")
##' @export
hm <- function(..., quiet = FALSE, roll = FALSE) {
  out <- .parse_hms(..., order = "HM", quiet = quiet)
  if (roll) {
    hms <- .roll_hms(hour = out["H", ], min = out["M", ])
    period(hour = hms$hour, minute = hms$min, second = hms$sec)
  } else {
    period(hour = out["H", ], minute = out["M", ])
  }
}

##' Parse periods with **h**our, **m**inute, and **s**econd components
##'
##' Transforms a character or numeric vector into a period object with the
##' specified number of hours, minutes, and seconds. `hms()` recognizes all
##' non-numeric characters except '-' as separators ('-' is used for negative
##' `durations`). After hours, minutes and seconds have been parsed, the
##' remaining input is ignored.
##'
##' @param ... a character vector of hour minute second triples
##' @param quiet logical. If `TRUE`, function evaluates without displaying
##'   customary messages.
##' @param roll logical. If `TRUE`, smaller units are rolled over to higher units
##'   if they exceed the conventional limit. For example,
##'   `hms("01:59:120", roll = TRUE)` produces period "2H 1M 0S".
##' @return a vector of period objects
##' @seealso [hm()], [ms()]
##' @keywords period
##' @examples
##'
##' x <- c("09:10:01", "09:10:02", "09:10:03")
##' hms(x)
##'
##' hms("7 6 5", "3:23:::2", "2 : 23 : 33", "Finished in 9 hours, 20 min and 4 seconds")
##' @export
hms <- function(..., quiet = FALSE, roll = FALSE) {
  out <- .parse_hms(..., order = "HMS", quiet = quiet)
  if (roll) {
    hms <- .roll_hms(out["H", ], out["M", ], out["S", ])
    period(hour = hms$hour, minute = hms$min, second = hms$sec)
  } else {
    period(hour = out["H", ], minute = out["M", ], second = out["S", ])
  }
}

.roll_hms <- function(hour = 0, min = 0, sec = 0) {
  min <- min + sec %/% 60
  sec <- sec %% 60
  hour <- hour + min %/% 60
  min <- min %% 60
  list(hour = hour, min = min, sec = sec)
}

.parse_hms <- function(..., order, quiet = FALSE) {
  ## wrapper for C level parse_hms
  hms <- unlist(lapply(list(...), .num_to_date), use.names = FALSE)
  out <- matrix(.Call(C_parse_hms, hms, order),
    nrow = 3L, dimnames = list(c("H", "M", "S"), NULL)
  )
  if (!quiet) {
    ## fixme: this warning should be dropped to C and thrown only when there are
    ## real parsing errors #530
    if (any(is.na(out[substr(order, ln <- nchar(order), ln), ]))) {
      warning("Some strings failed to parse, or all strings are NAs")
    }
  }
  out
}

##' User friendly date-time parsing functions
##'
##' `parse_date_time()` parses an input vector into POSIXct date-time
##' object. It differs from [base::strptime()] in two respects. First,
##' it allows specification of the order in which the formats occur without the
##' need to include separators and the `%` prefix. Such a formatting argument is
##' referred to as "order". Second, it allows the user to specify several
##' format-orders to handle heterogeneous date-time character
##' representations.
##'
##' When several format-orders are specified, `parse_date_time()` selects
##' (guesses) format-orders based on a training subset of the input
##' strings. After guessing the formats are ordered according to the performance
##' on the training set and applied recursively on the entire input vector. You
##' can disable training with `train = FALSE`.
##'
##' `parse_date_time()`, and all derived functions, such as `ymd_hms()`,
##' `ymd()`, etc., will drop into `fast_strptime()` instead of
##' [base::strptime()] whenever the guessed from the input data formats are all
##' numeric.
##'
##' The list below contains formats recognized by \pkg{lubridate}. For numeric
##' formats leading 0s are optional. As compared to [base::strptime()], some of
##' the formats are new or have been extended for efficiency reasons. These
##' formats are marked with "(*)" below. Fast parsers `parse_date_time2()` and
##' `fast_strptime()` accept only formats marked with "(!)".
##'
##'
##' \describe{ \item{`a`}{Abbreviated weekday name in the current
##' locale. (Also matches full name)}
##'
##' \item{`A`}{Full weekday name in the current locale.  (Also matches
##' abbreviated name).
##'
##' You don't need to specify `a` and `A` formats explicitly. Wday is
##' automatically handled if `preproc_wday = TRUE`}
##'
##' \item{`b` (!)}{Abbreviated or full month name in the current locale. The C
##' parser currently understands only English month names.}
##'
##' \item{`B` (!)}{Same as b.}
##'
##' \item{`d` (!)}{Day of the month as decimal number (01--31 or 0--31)}
##'
##' \item{`H` (!)}{Hours as decimal number (00--24 or 0--24).}
##'
##' \item{`I` (!)}{Hours as decimal number (01--12 or 1--12).}
##'
##' \item{`j`}{Day of year as decimal number (001--366 or 1--366).}
##'
##' \item{`q` (!*)}{Quarter (1--4). The quarter month is added to the parsed
##'   month if `m` element is present.}
##'
##' \item{`m` (!*)}{Month as decimal number (01--12 or 1--12). For
##'   `parse_date_time` also matches abbreviated and full months names as `b`
##'   and `B` formats. C parser understands only English month names.}
##'
##' \item{`M` (!)}{Minute as decimal number (00--59 or 0--59).}
##'
##' \item{`p` (!)}{AM/PM indicator in the locale. Commonly used in conjunction
##'  with `I` and \bold{not} with `H`.  But \pkg{lubridate}'s C parser accepts H
##'  format as long as hour is not greater than 12. C parser understands only
##'  English locale AM/PM indicator.}
##'
##' \item{`S` (!)}{Second as decimal number (00--61 or 0--61), allowing for up
##' to two leap-seconds (but POSIX-compliant implementations will ignore leap
##' seconds).}
##'
##' \item{`OS`}{Fractional second.}
##'
##' \item{`U`}{Week of the year as decimal number (00--53 or 0--53) using
##' Sunday as the first day 1 of the week (and typically with the first Sunday
##' of the year as day 1 of week 1).  The US convention.}
##'
##' \item{`w`}{Weekday as decimal number (0--6, Sunday is 0).}
##'
##' \item{`W`}{Week of the year as decimal number (00--53 or 0--53) using
##' Monday as the first day of week (and typically with the first Monday of the
##' year as day 1 of week 1).  The UK convention.}
##'
##' \item{`y` (!*)}{Year without century (00--99 or 0--99).  In
##' `parse_date_time()` also matches year with century (Y format).}
##'
##' \item{`Y` (!)}{Year with century.}
##'
##' \item{`z` (!*)}{ISO8601 signed offset in hours and minutes from UTC. For
##' example `-0800`, `-08:00` or `-08`, all represent 8 hours behind UTC. This
##' format also matches the Z (Zulu) UTC indicator. Because [base::strptime()]
##' doesn't fully support ISO8601 this format is implemented as an union of 4
##' formats: Ou (Z), Oz (-0800), OO (-08:00) and Oo (-08). You can use these
##' formats as any other but it is rarely necessary. `parse_date_time2()` and
##' `fast_strptime()` support all of these formats.}
##'
##' \item{`Om` (!*)}{Matches numeric month and English alphabetic months
##'                    (Both, long and abbreviated forms).}
##'
##' \item{`Op` (!*)}{Matches AM/PM English indicator.}
##'
##' \item{`r` (*)}{Matches `Ip` and `H` orders.}
##'
##' \item{`R` (*)}{Matches `HM` and`IMp` orders.}
##'
##' \item{`T` (*)}{Matches `IMSp`, `HMS`, and `HMOS` orders.}
##' }
##'
##'
##' @export
##' @param x a character or numeric vector of dates
##' @param orders a character vector of date-time formats. Each order string is
##'   a series of formatting characters as listed in [base::strptime()] but
##'   might not include the `"%"` prefix. For example, "ymd" will match all the
##'   possible dates in year, month, day order. Formatting orders might include
##'   arbitrary separators. These are discarded. See details for the implemented
##'   formats. If multiple order strings are supplied, they are applied in turn
##'   for `parse_date_time2()` and `fast_strptime()`. For `parse_date_time()`
##'   the order of applied formats is determined by `select_formats` parameter.
##' @param tz a character string that specifies the time zone with which to
##'   parse the dates
##' @param truncated integer, number of formats that can be missing. The most
##'   common type of irregularity in date-time data is the truncation due to
##'   rounding or unavailability of the time stamp. If the `truncated` parameter
##'   is non-zero `parse_date_time()` also checks for truncated formats. For
##'   example,  if the format order is "ymdHMS" and `truncated = 3`,
##'   `parse_date_time()` will correctly parse incomplete date-times like
##'   `2012-06-01 12:23`, `2012-06-01 12` and `2012-06-01`. \bold{NOTE:} The
##'   `ymd()` family of functions is based on [base::strptime()] which currently
##'   fails to parse `%Y-%m` formats.
##' @param quiet logical. If `TRUE`, progress messages are not printed, and `No
##'   formats found` error is suppressed and the function simply returns a
##'   vector of NAs.  This mirrors the behavior of base R functions
##'   [base::strptime()] and [base::as.POSIXct()].
##' @param locale locale to be used, see \link{locales}. On Linux systems you
##'   can use `system("locale -a")` to list all the installed locales.
##' @param select_formats A function to select actual formats for parsing from a
##'   set of formats which matched a training subset of `x`. It receives a named
##'   integer vector and returns a character vector of selected formats. Names
##'   of the input vector are formats (not orders) that matched the training
##'   set. Numeric values are the number of dates (in the training set) that
##'   matched the corresponding format. You should use this argument if the
##'   default selection method fails to select the formats in the right
##'   order. By default the formats with most formatting tokens (`%`) are
##'   selected and `%Y` counts as 2.5 tokens (so that it has a priority over
##'   `%y%m`). See examples.
##' @param exact logical. If `TRUE`, the `orders` parameter is interpreted as an
##'   exact [base::strptime()] format and no training or guessing are performed
##'   (i.e. `train`, `drop` parameters are ignored).
##' @param train logical, default `TRUE`. Whether to train formats on a subset of the
##'   input vector. As a result the supplied orders are sorted according to performance
##'   on this training set, which commonly results in increased performance. Please note
##'   that even when `train = FALSE` (and `exact = FALSE`) guessing of the actual
##'   formats is still performed on the training set (a pseudo-random subset of the
##'   original input vector). This might result in `All formats failed to parse`
##'   error. See notes below.
##' @param drop logical, default `FALSE`. Whether to drop formats that didn't
##'   match on the training set. If `FALSE`, unmatched on the training set
##'   formats are tried as a last resort at the end of the parsing
##'   queue. Applies only when `train = TRUE`. Setting this parameter to `TRUE`
##'   might slightly speed up parsing in situations involving many
##'   formats. Prior to v1.7.0 this parameter was implicitly `TRUE`, which
##'   resulted in occasional surprising behavior when rare patterns where not
##'   present in the training set.
##' @return a vector of POSIXct date-time objects
##' @seealso [base::strptime()], [ymd()], [ymd_hms()]
##' @keywords chron
##' @note `parse_date_time()` (and the derivatives `ymd()`, `ymd_hms()`, etc.)
##'   relies on a sparse guesser that takes at most 501 elements from the
##'   supplied character vector in order to identify appropriate formats from
##'   the supplied orders. If you get the error `All formats failed to parse`
##'   and you are confident that your vector contains valid dates, you should
##'   either set `exact` argument to `TRUE` or use functions that don't perform
##'   format guessing (`fast_strptime()`, `parse_date_time2()` or
##'   [base::strptime()]).
##' @note For performance reasons, when timezone is not UTC,
##'   `parse_date_time2()` and `fast_strptime()` perform no validity checks for
##'   daylight savings time. Thus, if your input string contains an invalid date
##'   time which falls into DST gap and `lt = TRUE` you will get an `POSIXlt`
##'   object with a non-existent time. If `lt = FALSE` your time instant will be
##'   adjusted to a valid time by adding an hour. See examples. If you want to
##'   get NA for invalid date-times use [fit_to_timeline()] explicitly.
##'
##' @examples
##'
##' ## ** orders are much easier to write **
##' x <- c("09-01-01", "09-01-02", "09-01-03")
##' parse_date_time(x, "ymd")
##' parse_date_time(x, "y m d")
##' parse_date_time(x, "%y%m%d")
##' #  "2009-01-01 UTC" "2009-01-02 UTC" "2009-01-03 UTC"
##'
##' ## ** heterogeneous date-times **
##' x <- c("09-01-01", "090102", "09-01 03", "09-01-03 12:02")
##' parse_date_time(x, c("ymd", "ymd HM"))
##'
##' ## ** different ymd orders **
##' x <- c("2009-01-01", "02022010", "02-02-2010")
##' parse_date_time(x, c("dmY", "ymd"))
##' ##  "2009-01-01 UTC" "2010-02-02 UTC" "2010-02-02 UTC"
##'
##' ## ** truncated time-dates **
##' x <- c("2011-12-31 12:59:59", "2010-01-01 12:11", "2010-01-01 12", "2010-01-01")
##' parse_date_time(x, "Ymd HMS", truncated = 3)
##'
##' ## ** specifying exact formats and avoiding training and guessing **
##' parse_date_time(x, c("%m-%d-%y", "%m%d%y", "%m-%d-%y %H:%M"), exact = TRUE)
##' parse_date_time(c('12/17/1996 04:00:00','4/18/1950 0130'),
##'                 c('%m/%d/%Y %I:%M:%S','%m/%d/%Y %H%M'), exact = TRUE)
##'
##' ## ** quarters and partial dates **
##' parse_date_time(c("2016.2", "2016-04"), orders = "Yq")
##' parse_date_time(c("2016", "2016-04"), orders = c("Y", "Ym"))
##'
##' ## ** fast parsing **
##' \dontrun{
##'   options(digits.secs = 3)
##'   ## random times between 1400 and 3000
##'   tt <- as.character(.POSIXct(runif(1000, -17987443200, 32503680000)))
##'   tt <- rep.int(tt, 1000)
##'
##'   system.time(out <- as.POSIXct(tt, tz = "UTC"))
##'   system.time(out1 <- ymd_hms(tt)) # constant overhead on long vectors
##'   system.time(out2 <- parse_date_time2(tt, "YmdHMOS"))
##'   system.time(out3 <- fast_strptime(tt, "%Y-%m-%d %H:%M:%OS"))
##'
##'   all.equal(out, out1)
##'   all.equal(out, out2)
##'   all.equal(out, out3)
##' }
##'
##' ## ** how to use `select_formats` argument **
##' ## By default %Y has precedence:
##' parse_date_time(c("27-09-13", "27-09-2013"), "dmy")
##'
##' ## to give priority to %y format, define your own select_format function:
##'
##' my_select <-   function(trained, drop=FALSE, ...){
##'    n_fmts <- nchar(gsub("[^%]", "", names(trained))) + grepl("%y", names(trained))*1.5
##'    names(trained[ which.max(n_fmts) ])
##' }
##'
##' parse_date_time(c("27-09-13", "27-09-2013"), "dmy", select_formats = my_select)
##'
##' ## ** invalid times with "fast" parsing **
##' parse_date_time("2010-03-14 02:05:06",  "YmdHMS", tz = "America/New_York")
##' parse_date_time2("2010-03-14 02:05:06",  "YmdHMS", tz = "America/New_York")
##' parse_date_time2("2010-03-14 02:05:06",  "YmdHMS", tz = "America/New_York", lt = TRUE)
parse_date_time <- function(x, orders, tz = "UTC", truncated = 0, quiet = FALSE,
                            locale = Sys.getlocale("LC_TIME"), select_formats = .select_formats,
                            exact = FALSE, train = TRUE, drop = FALSE) {

  ## backward compatible hack
  if (is.null(tz)) tz <- ""
  if (length(tz) != 1 || is.na(tz)) {
    stop("`tz` argument must be a character of length one")
  }

  orig_locale <- Sys.getlocale("LC_TIME")
  Sys.setlocale("LC_TIME", locale)
  on.exit(Sys.setlocale("LC_TIME", orig_locale))

  x <- as.character(.num_to_date(x))
  if (truncated != 0) {
    orders <- .add_truncated(orders, truncated)
  }

  .local_parse <- function(x, first = FALSE) {
    formats <-
      if (exact) {
        orders
      } else {
        trainset <- .get_train_set(x)
        .best_formats(trainset, orders, locale = locale, select_formats, drop = drop, train = train)
      }
    if (length(formats) > 0) {
      out <- .parse_date_time(x, formats, tz = tz, quiet = quiet, locale = locale)
      new_na <- is.na(out)
      if (any(new_na)) {
        x <- x[new_na]
        if (length(x) == length(out)) {
          # don't recur if failed for all
          failed <<- length(x)
        } else {
          out[new_na] <- .local_parse(x)
        }
      }
      out
    } else {
      if (first && !quiet) {
        warning("All formats failed to parse. No formats found.", call. = FALSE)
        warned <<- TRUE
      }
      failed <<- length(x)
      NA
    }
  }

  failed <- 0L
  warned <- FALSE
  to_parse <- which(!is.na(x) & nzchar(x)) ## missing data might be ""
  ## prepare an NA vector
  out <- .POSIXct(rep.int(NA_real_, length(x)), tz = tz)

  if (length(to_parse)) {
    out[to_parse] <- .local_parse(x[to_parse], TRUE)
    if (failed > 0 && !quiet && !warned) {
      warning(" ", failed, " failed to parse.", call. = FALSE)
    }
  }

  out
}

parse_dt <- function(x, orders, is_format = FALSE, return_lt = FALSE, cutoff_2000 = 68L) {
  .Call(C_parse_dt, x, orders, as.logical(is_format), as.logical(return_lt), as.integer(cutoff_2000))
}

##' @description `parse_date_time2()` is a fast C parser of numeric orders.
##'
##' @rdname parse_date_time
##' @export
##' @param lt logical. If `TRUE`, returned object is of class POSIXlt, and POSIXct
##'   otherwise. For compatibility with [base::strptime()] the default is `TRUE`
##'   for `fast_strptime()` and `FALSE` for `parse_date_time2()`.
##' @param cutoff_2000 integer. For `y` format,  two-digit numbers smaller or equal
##'    to `cutoff_2000` are parsed as though starting with `20`, otherwise parsed
##'    as though starting with `19`. Available only for functions relying on
##'    `lubridate`s internal parser.
parse_date_time2 <- function(x, orders, tz = "UTC", exact = FALSE, lt = FALSE, cutoff_2000 = 68L) {
  if (length(tz) != 1 || is.na(tz)) {
    stop("`tz` argument must be a character of length one")
  }
  parse1 <- function(x, order) {
    if (!exact) {
      order <- gsub("[^[:alpha:]]+", "", as.character(order))
    }
    if (lt) {
      .mklt(parse_dt(x, order, exact, TRUE, cutoff_2000), tz)
    } else {
      if (is_utc(tz)) {
        .POSIXct(parse_dt(x, order, exact, FALSE, cutoff_2000), tz = "UTC")
      } else {
        as.POSIXct(.mklt(parse_dt(x, order, exact, TRUE, cutoff_2000), tz))
      }
    }
  }
  nnas <- !is.na(x)
  out <- parse1(x, orders[[1]])
  for (order in orders[-1]) {
    nnas <- nnas & is.na(out)
    out[nnas] <- parse1(x[nnas], order)
  }
  out
}

##' @description `fast_strptime()` is a fast C parser of numeric formats only
##'   that accepts explicit format arguments, just like [base::strptime()].
##' @rdname parse_date_time
##' @export
##' @param format a vector of formats. If multiple formats supplied they are
##'   applied in turn till success. The formats should include all the
##'   separators and each format letter must be prefixed with %, just as in the
##'   format argument of [base::strptime()].
fast_strptime <- function(x, format, tz = "UTC", lt = TRUE, cutoff_2000 = 68L) {
  parse1 <- function(x, fmt) {
    if (lt) {
      .mklt(parse_dt(x, fmt, TRUE, TRUE, cutoff_2000), tz)
    } else {
      if (is_utc(tz)) {
        .POSIXct(parse_dt(x, fmt, TRUE, FALSE, cutoff_2000), "UTC")
      } else {
        as.POSIXct(.mklt(parse_dt(x, fmt, TRUE, TRUE, cutoff_2000), tz))
      }
    }
  }
  nnas <- !is.na(x)
  out <- parse1(x, format[[1]])
  for (fmt in format[-1]) {
    nnas <- nnas & is.na(out)
    out[nnas] <- parse1(x[nnas], fmt)
  }
  out
}



### INTERNAL
.mklt <- function(dtlist, tz) {
  n <- length(dtlist$sec)

  na_fill <- rep_len(NA_integer_, n)
  dtlist[["wday"]] <- na_fill
  dtlist[["yday"]] <- na_fill

  dst_fill <- rep_len(-1L, n)
  dtlist[["isdst"]] <- dst_fill

  .POSIXlt(dtlist, tz = tz)
}

.parse_date_time <- function(x, formats, tz, quiet, locale) {

  out <- .strptime(x, formats[[1]], tz = tz, quiet = quiet, locale = locale)
  na <- is.na(out)
  newx <- x[na]

  if (is_verbose()) {
    message(" ", sum(!na), " parsed with ", formats[[1]])
  }

  ## recursive parsing
  if (length(formats) > 1 && length(newx) > 0) {
    out[na] <- .parse_date_time(newx, formats[-1], tz = tz, quiet = quiet, locale = locale)
  }

  out
}

.strptime <- function(x, fmt, tz = "UTC", quiet = FALSE, locale = NULL) {

  ## Depending on fmt we might need to preprocess x.
  ## ISO8601 and internal parser are the only cases so far.
  ## %Ou: "2013-04-16T04:59:59Z"
  ## %Oo: "2013-04-16T04:59:59+01"
  ## %Oz: "2013-04-16T04:59:59+0100"
  ## %OO: "2013-04-16T04:59:59+01:00"

  ## is_posix <-  0 < regexpr("^[^%]*%Y[^%]+%m[^%]+%d[^%]+(%H[^%](%M[^%](%S)?)?)?[^%Z]*$", fmt)

  c_parser <- 0 < regexpr("^[^%0-9]*(%([YymdqIHMSz]|O[SzuoOpmb])[^%0-9Z]*)+$", fmt)
  zpos <- regexpr("%O((?<z>z)|(?<u>u)|(?<o>o)|(?<O>O))", fmt, perl = TRUE)

  if (c_parser) {
    ## C PARSER:
    out <- fast_strptime(x, fmt, tz = "UTC", lt = FALSE)

    if (!is_utc(tz)) {
      out <-
        if (zpos > 0) {
          if (!quiet) {
            message("Date in ISO8601 format; converting timezone from UTC to \"", tz, "\".")
          }
          with_tz(out, tzone = tz)
        } else {
          ## force_tz is very slow
          force_tz(out, tzone = tz)
        }
    }

    out
  } else {
    ## STRPTIME PARSER:

    ## strptime doesn't accept 'locale' argument; need a hard reset
    if (!is.null(locale)) {
      old_lc_time <- Sys.getlocale("LC_TIME")
      if (old_lc_time != locale) {
        Sys.setlocale("LC_TIME", locale)
        on.exit(Sys.setlocale("LC_TIME", old_lc_time))
      }
    }

    if (zpos > 0) {
      ## If ISO8601 -> pre-process x and fmt
      capt <- attr(zpos, "capture.names")[attr(zpos, "capture.start") > 0][[2]] ## <- second subexp
      repl <- switch(capt,
        z = "%z",
        u = "Z",
        ## substitute +aa with +aa00
        o = {
          x <- sub("([+-]\\d{2}(?=\\D+)?$)", "\\100", x, perl = TRUE)
          "%z"
        },
        ## substitute +aa:bb with +aabb
        O = {
          x <- sub("([+-]\\d{2}):(?=[^:]+$)", "\\1", x, perl = TRUE)
          "%z"
        },
        stop("Unrecognised capture detected; please report this bug")
      )

      fmt <- .str_sub(fmt, zpos, zpos + attr(zpos, "match.length") - 1, repl)

      ## user has supplied tz argument -> convert to tz
      if (!is_utc(tz)) {
        if (!quiet) {
          message("Date in ISO8601 format; converting timezone from UTC to \"", tz, "\".")
        }
        return(with_tz(strptime(.enclose(x), .enclose(fmt), "UTC"), tzone = tz))
      }
    }

    strptime(.enclose(x), .enclose(fmt), tz)
  }
}


## Expand format strings to also include truncated formats
## Get locations of letters as vector
## Choose the number at the n - truncated place in the vector
## return the substring created by 1 to tat number.
.add_truncated <- function(orders, truncated) {
  out <- orders

  if (truncated > 0) {
    trunc_one <- function(order) {
      alphas <- gregexpr("[a-zA-Z]", order)[[1]]
      start <- max(0, length(alphas) - truncated)
      cut_points <- alphas[start:(length(alphas) - 1)]

      truncs <- c()
      for (j in seq_along(cut_points)) {
        truncs <- c(truncs, substr(order, 1, cut_points[j]))
      }
      truncs
    }
  } else {
    trunc_one <- function(order) {
      alphas <- gregexpr("[a-zA-Z]", order)[[1]][-1]
      end <- max(1, abs(truncated) - 1)
      cut_points <- alphas[1:end]

      truncs <- c()
      for (j in seq_along(cut_points)) {
        truncs <- c(truncs, substr(order, cut_points[j], nchar(order)))
      }
      truncs
    }
  }

  for (i in seq_along(orders)) {
    out <- c(out, trunc_one(orders[i]))
  }

  out
}


.xxx_hms_truncations <- list(T = c("R", "r", ""), R = c("r", ""), r = "")

.parse_xxx_hms <- function(..., orders, truncated, quiet, tz, locale) {
  ## !!! NOTE: truncated operates on first element in ORDERS !
  if (truncated > 0) {
    ## Take first 3 formats and append formats from .xxx_hms_truncations
    ## co responding to the 4th format letter in order[[1]] -- T, R or r.
    xxx <- substr(orders[[1]], 1, 3) ##
    add <- paste(xxx, .xxx_hms_truncations[[substr(orders[[1]], 4, 4)]], sep = "")
    rest <- length(add) - truncated
    if (rest < 0) {
      orders <- c(orders, add, .add_truncated(xxx, abs(rest)))
    } else {
      orders <- c(orders, add[1:truncated])
    }
  }
  dates <- unlist(lapply(list(...), .num_to_date), use.names = FALSE)
  parse_date_time(dates, orders, tz = tz, quiet = quiet, locale = locale, )
}

.parse_xxx <- function(..., orders, quiet, tz, locale, truncated) {
  dates <- unlist(lapply(list(...), .num_to_date), use.names = FALSE)
  if (is.null(tz)) {
    as.Date.POSIXct(parse_date_time(dates, orders,
      quiet = quiet, tz = "UTC",
      locale = locale, truncated = truncated
    ))
  } else {
    parse_date_time(dates, orders,
      quiet = quiet, tz = tz,
      locale = locale, truncated = truncated
    )
  }
}

.num_to_date <- function(x) {
  if (is.numeric(x)) {
    prefix0 <- any(x > 1900, na.rm = TRUE)
    out <- rep.int(as.character(NA), length(x))
    nnas <- !is.na(x)
    x <- format(x[nnas], scientific = FALSE, trim = TRUE)
    # #1002 lower numbers are normally something else than full dates
    if (prefix0) {
      x <- paste(ifelse(nchar(x) %% 2 == 1, "0", ""), x, sep = "")
    }
    out[nnas] <- x
    out
  } else {
    as.character(x)
  }
}

.parse_iso_dt <- function(x, tz) {
  parse_date_time(x, orders = c("ymdTz", "ymdT", "ymd"), tz = tz, train = FALSE)
}

as_POSIXct <- function(x, tz = tz(x)) {
  if (is.character(x)) {
    .parse_iso_dt(x, tz = tz)
  } else if (is.Date(x)) {
    ## as.POSIXct.Date assues UTC in computation but returns POSIXct with system TZ
    ## same as as_datetime Date method
    .POSIXct(as.numeric(x) * 86400, tz = "UTC")
  } else if (!is.POSIXct(x)) {
    as.POSIXct(x, tz = tz)
  } else {
    x
  }
}
## parse.r ends here

## This is for interactive tests mainly, (test-parser.R and test-stamp.R are ussing gueesser, so it should be enough)


## x <- c('February 20th 1973',
##        "february  14, 2004",
##        "Sunday, May 1, 2000",
##        "Sunday, May 1, 2000",
##        "february  14, 04",
##        'Feb 20th 73',
##        "January 5 1999 at 7 pm",
##        "January 5 1999 at7pm",
##        "jan 3 2010",
##        "Jan 1, 1999",
##        "jan 3   10",
##        "01 3 2010",
##        "1 3 10",
##        '1 13 89',
##        "5/27/1979",
##        "12/31/99",
##        "DOB:12/11/00",
##        "-----------",
##        'Thu, 1 July 2004 22:30:00',
##        'Thu, 1st of July 2004 at 22:30:00',
##        'Thu, 1July 2004 at 22:30:00',
##        'Thu, 1July2004 22:30:00',
##        "21 Aug 2011, 11:15:34 pm",
##        "-----------",
##        "1979-05-27 05:00:59",
##        "1979-05-27",
##        "-----------",
##        "3 jan 2000",
##        "17 april 85",
##        "27/5/1979",
##        '20 01 89',
##        '00/13/10',
##        "-------",
##        "14 12 00",
##        "03:23:22 pm")

## guess_formats(x, "BdY", print = T)
## guess_formats(x, "Bdy")
## guess_formats(x, "bdY")
## guess_formats(x, "bdy")
## guess_formats(x, "mdy", print = T)

## guess_formats(x, "T", print = T)

## guess_formats(x, c("mdY", "BdY", "Bdy", "bdY", "bdy"))
## guess_formats(x, c("dby", "dbY", "dBy", "dBY"))

## guess_formats(x, c("mdY", "mdy"))
## guess_formats(x, c("dmY", "dmy"))

## guess_formats(x, c("BdY H", "dBY HMS", "dbY HMS"))
## guess_formats(x, c("dBy HMS"))

## guess_formats(x, c("Ymd HMS"), print_matches=T)
## guess_formats(x, c("dmy HMS"), print_matches=T)
## guess_formats(x, c("mdY r"), print_matches=T)


## stamp(x, c("BdY", "bdY", "bd"))(Sys.time())
## system.time(stamp(x[1], "BdY", quiet = T))
## stamp(x, quiet = F)


## y <- rep(x, 3)
## f <- rep(c("BdY", "bdY", "bdy"), 10)
## system.time( for (i in 1:1) substitute_formats(y, f))

## reg <- c("\\b(?<Q>[0-2]?\\d)\\D*?(?<p>(AM|PM))",
##          "\\b(?<T>[0-2]?\\d)\\D*?(?<D>(DD|BB))")

## reg <- c("\\b((?<fixed>(?<Q_f>[0-2]?\\d)\\D*?(?<p_f>(AM|PM)))|(?<flex>(?<Q>[0-2]?\\d)\\D*?(?<p>(DD|BB))))")
## x <- c("aaa 12 DD aaa",
##        "bbb 22 AM bbb",
##        "ccc 14 PM ccc")
## ## gsub(reg , "####",x , ignore.case=T, perl = TRUE)
## regexpr(reg[[1]], x, perl=T)


test_that(".get_train_set can find non NA dates", {
  x <- suppressWarnings(suppressMessages(ymd(c(rep(NA, 199), 20130213))))
  expect_equal(x, as.Date(as.POSIXct(c(rep(NA, 199), "2013-02-13"), tz = "UTC")))
})

test_that("guess_format works with missing entries", {
  expect_equal(
    guess_formats("01-01-10", "mdy"),
    guess_formats(c("01-01-10", NA), "mdy")
  )
  expect_null(guess_formats(NA, "mdy"))
})

test_that("guess_format consistently adds 0b for B and b", {
  # Abbreviated month names:
  expect_true("Obdy" %in% names(guess_formats("jan 3 2010", "bdy")))
  expect_true("Obdy" %in% names(guess_formats("jan 3 2010", "Bdy")))
  # Full month names:
  expect_true("Obdy" %in% names(guess_formats("January 5 1999", "Bdy")))
  expect_true("Obdy" %in% names(guess_formats("January 5 1999", "Bdy")))
})

test_that("b/B give complete and equivalent results for trained parsing", {
  full <- c("february  14, 2004", "January 5 1999")
  mixed <- c("jan 3 2010", "January 5 1999")
  short <- c("jan 3 2010", "Jan 1, 1999")
  expect_equal(
    parse_date_time(full, "bdy"),
    parse_date_time(full, "Bdy")
  )
  expect_equal(
    parse_date_time(mixed, "bdy"),
    parse_date_time(mixed, "Bdy")
  )
  expect_equal(
    parse_date_time(short, "bdy"),
    parse_date_time(short, "Bdy")
  )

  expect_false(any(is.na(parse_date_time(full, "bdy"))))
  expect_false(any(is.na(parse_date_time(mixed, "bdy"))))
  expect_false(any(is.na(parse_date_time(short, "bdy"))))
})

#' @include timespans.r
#' @include parse.r
#' @include util.r
NULL

check_interval <- function(object) {
  errors <- character()
  if (!is.numeric(object@.Data)) {
    msg <- "Span length must be numeric."
    errors <- c(errors, msg)
  }
  if (!is(object@start, "POSIXct")) {
    msg <- "Start date must be in POSIXct format."
    errors <- c(errors, msg)
  }
  if (length(object@.Data) != length(object@start)) {
    msg <- paste("Inconsistent lengths: spans = ", length(object@.Data),
      ", start dates = ", length(object@start), sep = "")
    errors <- c(errors, msg)
  }
  if (length(errors) == 0)
    TRUE
  else
    errors
}

.units_within_seconds <- function(secs, unit = "second") {
  ## return a list suitable to pass to new("Period", ...)
  switch(unit,
         second = list(secs),
         minute = list(secs %% 60, minute = secs %/% 60),
         hour =
           c(.units_within_seconds(secs %% 3600, "minute"),
             list(hour = secs %/% 3600)),
         day =
           c(.units_within_seconds(secs %% 86400, "hour"),
             list(day = secs %/% 86400)),
         stop("Unsuported unit ", unit))
}

#' Interval class
#'
#' Interval is an S4 class that extends the [Timespan-class] class. An
#' Interval object records one or more spans of time. Intervals record these
#' timespans as a sequence of seconds that begin at a specified date. Since
#' intervals are anchored to a precise moment of time, they can accurately be
#' converted to [Period-class] or [Duration-class] class objects. This
#' is because we can observe the length in seconds of each period that begins on a
#' specific date. Contrast this to a generalized period, which may not have a
#' consistent length in seconds (e.g. the number of seconds in a year will change
#' if it is a leap year).
#'
#' Intervals can be both negative and positive. Negative intervals progress
#' backwards from the start date; positive intervals progress forwards.
#'
#' Interval class objects have two slots: .Data, a numeric object equal to the number
#' of seconds in the interval; and start, a POSIXct object that specifies the time
#' when the interval starts.
#'
#' @aliases intervals
#' @export
setClass("Interval", contains = c("Timespan", "numeric"),
  slots = c(start = "POSIXct",   tzone = "character"), validity = check_interval)

#' @export
setMethod("show", signature(object = "Interval"), function(object) {
  if (length(object@.Data) == 0) {
    cat("<Interval[0]>\n")
  } else {
    print(format(object), quote = FALSE)
  }
})

#' @export
format.Interval <- function(x, ...) {
  paste(
    format(x@start, tz = x@tzone, usetz = TRUE),
    format(x@start + x@.Data, tz = x@tzone, usetz = TRUE),
    sep = "--"
  )
}

#' @export
setMethod("c", signature(x = "Interval"), function(x, ...) {
  elements <- lapply(list(...), as.interval)
  spans <- c(x@.Data, unlist(elements@.Data))
  starts <- c(x@start, lapply(elements, int_start))
  new("Interval", spans, start = starts, tzone = x@tzone)
})

#' @export
setMethod("rep", signature(x = "Interval"), function(x, ...) {
  new("Interval", rep(x@.Data, ...), start = rep(x@start, ...), tzone = x@tzone)
})

#' @export
setMethod("[", signature(x = "Interval"),
  function(x, i, j, ..., drop = TRUE) {
      new("Interval", x@.Data[i], start = x@start[i], tzone = x@tzone)
  }
)

#' @export
setMethod("[[", signature(x = "Interval"),
  function(x, i, j, ..., exact = TRUE) {
    new("Interval", x@.Data[i], start = x@start[i], tzone = x@tzone)
  }
)

#' @export
setMethod("[<-", signature(x = "Interval"), function(x, i, j, ..., value) {
    if (is.interval(value)) {
      x@.Data[i] <- value@.Data
      x@start[i] <- value@start
      new("Interval", x@.Data, start = x@start, tzone = x@tzone)
    }
    else {
      x@.Data[i] <- value
    new("Interval", x@.Data, start = x@start, tzone = x@tzone)
  }
})

#' @export
setMethod("[[<-", signature(x = "Interval"), function(x, i, j, ..., value) {
  if (is.interval(value)) {
    x@.Data[i] <- value@.Data
    x@start[i] <- value@start
    new("Interval", x@.Data, start = x@start, tzone = x@tzone)
  }
  else {
    x@.Data[i] <- value
    new("Interval", x@.Data, start = x@start, tzone = x@tzone)
  }
})

#' @export
setMethod("$", signature(x = "Interval"), function(x, name) {
  if (name == "span") name <- ".Data"
  slot(x, name)
})

#' @export
setMethod("$<-", signature(x = "Interval"), function(x, name, value) {
  if (name == "span") name <- ".Data"
  slot(x, name) <- value
  x
})

#' @export
unique.Interval <- function(x, ...) {
  df <- unique.data.frame(data.frame(data = x@.Data, start = x@start), ...)
  new("Interval", df$data, start = df$start, tzone = x@tzone)
}

#' Utilities for creation and manipulation of `Interval` objects
#'
#' `interval()` creates an [Interval-class] object with the specified start and
#' end dates. If the start date occurs before the end date, the interval will be
#' positive. Otherwise, it will be negative. Character vectors in ISO 8601
#' format are supported from v1.7.2.
#'
#' Intervals are time spans bound by two real date-times.  Intervals can be
#' accurately converted to either period or duration objects using
#' [as.period()], [as.duration()]. Since an interval is anchored to a fixed
#' history of time, both the exact number of seconds that passed and the number
#' of variable length time units that occurred during the interval can be
#' calculated.
#'
#' @export
#' @param start,end POSIXt, Date or a character vectors. When `start` is a
#'   character vector and end is `NULL`, ISO 8601 specification is assumed but
#'   with much more permisive lubridate style parsing both for dates and periods
#'   (see examples).
#' @param tzone a recognized timezone to display the interval in
#' @param x an R object
#' @return `interval()` -- [Interval-class] object.
#' @seealso [Interval-class], [as.interval()], \code{\link{\%within\%}}
#' @examples
#' interval(ymd(20090201), ymd(20090101))
#'
#' date1 <- ymd_hms("2009-03-08 01:59:59")
#' date2 <- ymd_hms("2000-02-29 12:00:00")
#' interval(date2, date1)
#' interval(date1, date2)
#' span <- interval(ymd(20090101), ymd(20090201))
#'
#' ### ISO Intervals
#'
#' interval("2007-03-01T13:00:00Z/2008-05-11T15:30:00Z")
#' interval("2007-03-01T13:00:00Z/P1Y2M10DT2H30M")
#' interval("P1Y2M10DT2H30M/2008-05-11T15:30:00Z")
#' interval("2008-05-11/P2H30M")
#'
#' ### More permisive parsing (as long as there are no intermittent / characters)
#' interval("2008 05 11/P2hours 30minutes")
#' interval("08 05 11/P 2h 30m")
#'
#' is.interval(period(months= 1, days = 15)) # FALSE
#' is.interval(interval(ymd(20090801), ymd(20090809))) # TRUE
interval <- function(start = NULL, end = NULL, tzone = tz(start)) {
  # NB: tzone is forced and never called on NULL here

  if (is.character(start) && is.null(end)) {
    return(parse_interval(start, tzone))
  }

  if (length(start) == 0 || length(end) == 0) {
    ## We used to return UTC on NULL
    if (is.null(start) && missing(tzone)) {
      tzone <- "UTC"
    }
    start <- POSIXct(tz = tzone)
    return(new("Interval", numeric(), start = start, tzone = tzone))
  }

  if (is.Date(start)) start <- date_to_posix(start)
  if (is.Date(end)) end <- date_to_posix(end)

  force(tzone)
  start <- as_POSIXct(start, tzone)
  end <- as_POSIXct(end, tzone)

  span <- as.numeric(end) - as.numeric(start)
  starts <- start + rep(0, length(span))
  if (tzone != tz(starts)) starts <- with_tz(starts, tzone)

  new("Interval", span, start = starts, tzone = tzone)
}

parse_interval <- function(x, tz) {

  # create matrix of string parts from x: 1st column is anything before /, 2nd is anything after.
  # replicates without stringr: str_split_fixed(x, "/", 2)
  mat <- matrix(
    c(gsub('(^[^/]+)/(.+$)', '\\1', x), gsub('(^[^/]+)/(.+$)', '\\2', x)),
    ncol = 2
  )
  pstart <- grepl("^P", mat[, 1])
  pend <- grepl("^P",  mat[, 2])

  if (any(pstart & pend)) {
    stop(sprintf("Interval specified with period endpoints (%s)", x[pstart & pend][[1]]))
  }

  start <- .POSIXct(rep.int(NA_real_, length(x)), tz = tz)
  end <- .POSIXct(rep.int(NA_real_, length(x)), tz = tz)

  start[!pstart] <- .parse_iso_dt(mat[!pstart, 1], tz)
  end[!pend] <- .parse_iso_dt(mat[!pend, 2], tz = tz)
  end[pend] <- start[pend] + parse_period(mat[pend, 2])
  start[pstart] <- end[pstart] - parse_period(mat[pstart, 1])
  interval(start, end, tz)
}

#'
#' \code{\%--\%} Creates an interval that covers the range spanned by two
#' dates. It replaces the original behavior of \pkg{lubridate}, which created an
#' interval by default whenever two date-times were subtracted.
#'
#' @export
#' @rdname interval
"%--%" <- function(start, end) interval(start, end)

#' @rdname interval
#' @export
is.interval <- function(x) is(x, c("Interval"))


#' @rdname interval
#' @param int an interval object
#' @return `int_start()` and `int_end()` return a POSIXct date object when
#'   used as an accessor. Nothing when used as a setter.
#' @examples
#' int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int_start(int)
#' int_start(int) <- ymd("2001-06-01")
#' int
#'
#' int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int_end(int)
#' int_end(int) <- ymd("2002-06-01")
#' int
#' @export
int_start <- function(int) int@start

#' @rdname interval
#' @param value interval's start/end to be assigned to `int`
#' @export
"int_start<-" <- function(int, value) {
  value <- as.POSIXct(value)
  span <- as.numeric(int@start + int@.Data - value, "secs")
  equal.lengths <- data.frame(span, value)
  int <- new("Interval", span, start = equal.lengths$value,
    tzone = int@tzone)
}

#' @description `int_start()`/`int_end()` and `int_start<-()`/`int_end<-()` are
#'   "accessors" and "setters" respectively of the start/end date of an
#'   interval.
#'
#' @rdname interval
#' @export
int_end <- function(int) int@start + int@.Data

#' @rdname interval
#' @export
"int_end<-" <- function(int, value) {
  value <- as.POSIXct(value)
  span <- as.numeric(value - int@start, "secs")
  int <- new("Interval", span, start = int@start,
    tzone = int@tzone)
}

#' @rdname interval
#' @return `int_length()` -- numeric length of the interval in
#'   seconds. A negative number connotes a negative interval.
#' @examples
#' int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int_length(int)
#' @export
int_length <- function(int) int@.Data

#' @description
#' `int_flip()` reverses the order of the start date and end date in an
#' interval. The new interval takes place during the same timespan as the
#' original interval, but has the opposite direction.
#'
#' @rdname interval
#' @return `int_flip()` -- flipped interval object
#' @examples
#' int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int_flip(int)
#' @export
int_flip <- function(int) {
  new("Interval", -int@.Data, start = int@start + int@.Data, tzone = int@tzone)
}


#' @description
#' `int_shift()` shifts the start and end dates of an interval up or down the
#' timeline by a specified amount. Note that this may change the exact length of
#' the interval if the interval is shifted by a Period object. Intervals shifted
#' by a Duration or difftime object will retain their exact length in seconds.
#'
#' @rdname interval
#' @param by A period or duration object to shift by (for `int_shift`)
#' @return `int_shift()` -- an Interval object
#' @examples
#' int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int_shift(int, duration(days = 11))
#' int_shift(int, duration(hours = -1))
#' @export
int_shift <- function(int, by) {
  if (!is.timespan(by)) stop("by is not a recognized timespan object")
  if (is.interval(by)) stop("an interval cannot be shifted by another interval.
    Convert second interval to a period or duration.")
  interval(int@start + by, int_end(int) + by)
}


#' @description
#' `int_overlaps()` tests if two intervals overlap.
#'
#' @rdname interval
#' @param int1 an Interval object (for `int_overlaps()`, `int_aligns()`)
#' @param int2 an Interval object (for `int_overlaps()`, `int_aligns()`)
#' @return `int_overlaps()` -- logical, TRUE if int1 and int2 overlap by at
#'   least one second. FALSE otherwise
#' @examples
#' int1 <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int2 <- interval(ymd("2001-06-01"), ymd("2002-06-01"))
#' int3 <- interval(ymd("2003-01-01"), ymd("2004-01-01"))
#'
#' int_overlaps(int1, int2) # TRUE
#' int_overlaps(int1, int3) # FALSE
#' @export
int_overlaps <- function(int1, int2) {
  stopifnot(c(is.interval(int1), is.interval(int2)))
  int1 <- int_standardize(int1)
  int2 <- int_standardize(int2)
  int1@start <= int2@start + int2@.Data & int2@start <= int1@start + int1@.Data
}


#' @description
#' `int_standardize()` ensures all intervals in an interval object are
#' positive. If an interval is not positive, flip it so that it retains its
#' endpoints but becomes positive.
#'
#' @rdname interval
#' @examples
#' int <- interval(ymd("2002-01-01"), ymd("2001-01-01"))
#' int_standardize(int)
#' @export
int_standardize <- function(int) {
  negs <- !is.na(int@.Data) & int@.Data < 0
  int[negs] <- int_flip(int[negs])
  int
}


#' @description
#' `int_aligns()` tests if two intervals share an endpoint. The direction of
#' each interval is ignored. int_align tests whether the earliest or latest
#' moments of each interval occur at the same time.
#'
#' @rdname interval
#' @return `int_aligns()` -- logical, TRUE if int1 and int2 begin or end on the
#'   same moment. FALSE otherwise
#' @examples
#' int1 <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int2 <- interval(ymd("2001-06-01"), ymd("2002-01-01"))
#' int3 <- interval(ymd("2003-01-01"), ymd("2004-01-01"))
#'
#' int_aligns(int1, int2) # TRUE
#' int_aligns(int1, int3) # FALSE
#' @export
int_aligns <- function(int1, int2) {
  int1 <- int_standardize(int1)
  int2 <- int_standardize(int2)

  int1@start == int2@start | (int1@start + int1@.Data) == (int2@start + int2@.Data)
}

#' @description
#' `int_diff()` returns the intervals that occur between the elements of a
#' vector of date-times. `int_diff()` is similar to the POSIXt and Date
#' methods of [diff()], but returns an [Interval-class] object instead
#' of a difftime object.
#'
#' @rdname interval
#' @param times A vector of POSIXct, POSIXlt or Date class date-times (for
#'   `int_diff()`)
#' @return `int_diff()` -- interval object that contains the n-1 intervals
#'   between the n date-time in times
#' @examples
#' dates <- now() + days(1:10)
#' int_diff(dates)
#' @export
int_diff <- function(times) {
  interval(times[-length(times)], times[-1])
}


#' @importFrom generics intersect
#' @export
generics::intersect

#' @export
intersect.Interval <- function(x, y, ...) {
  int1 <- int_standardize(x)
  int2 <- int_standardize(y)

  starts <- pmax(int1@start, int2@start)
  ends <- pmin(int1@start + int1@.Data, int2@start + int2@.Data)
  spans <- as.numeric(ends) - as.numeric(starts)

  no.int <- ends < starts
  no.int <- is.na(no.int) | no.int
  spans[no.int] <- NA
  starts[no.int] <- NA

  new.int <- new("Interval", spans, start = starts, tzone = x@tzone)
  negix <- !is.na(x@.Data) & (sign(x@.Data) == -1)
  new.int[negix] <- int_flip(new.int[negix])
  new.int
}

#' @importFrom generics union
#' @export
generics::union

#' @export
union.Interval <- function(x, y, ...) {
  int1 <- int_standardize(x)
  int2 <- int_standardize(y)

  starts <- pmin(int1@start, int2@start)
  ends <- pmax(int1@start + int1@.Data, int2@start + int2@.Data)

  spans <- as.numeric(ends) - as.numeric(starts)

  if (any(!int_overlaps(int1, int2)) && is_verbose()) {
    message("Union includes intervening time between intervals.")
  }

  tz(starts) <- x@tzone
  new.int <- new("Interval", spans, start = starts, tzone = x@tzone)
  new.int[sign(x@.Data) == -1] <- int_flip(new.int[sign(x@.Data) == -1])
  new.int
}

#' @importFrom generics setdiff
#' @export
generics::setdiff

#' @export
setdiff.Interval <- function(x, y, ...) {

  if (length(x) != length(y)) {
    xy <- match_lengths(x, y)
    x <- xy[[1]]
    y <- xy[[2]]
  }

  aligned <- int_aligns(x, y)
  inside <- y %within% x
  makes2 <- !aligned & inside

  if (sum(makes2)) {
    stop(paste("Cases", which(makes2),
      "result in discontinuous intervals."))
  }

  int1 <- int_standardize(x)
  int2 <- int_standardize(y)

  first.y <- int_start(int2)
  last.y <- int_end(int2)

  starts <- int_start(int1)

  starts[(last.y + 1) %within% int1] <- last.y[(last.y + 1) %within% int1]

  ends <- int_end(int1)
  ends[(first.y - 1) %within% int1] <- first.y[(first.y - 1) %within% int1]

  spans <- as.numeric(ends) - as.numeric(starts)

  new.int <- new("Interval", spans, start = starts, tzone = x@tzone)
  new.int[sign(x@.Data) == -1] <- int_flip(new.int[sign(x@.Data) == -1])
  new.int
}


#' Does a date (or interval) fall within an interval?
#'
#' Check whether `a` lies within the interval `b`, inclusive of the endpoints.
#'
#' @export
#' @rdname within-interval
#' @aliases %within%,Interval,Interval-method %within%,ANY,Interval-method
#'   %within%,Date,list-method %within%,POSIXt,list-method
#' @param a An interval or date-time object.
#' @param b Either an interval vector, or a list of intervals.
#'
#'   If `b` is an interval it is recycled to the same length as `a`.
#'   If `b` is a list of intervals, `a` is checked if it falls within _any_
#'   of the intervals, i.e. `a %within% list(int1, int2)` is equivalent to
#'   `a %within% int1 | a %within% int2`.
#' @return A logical vector.
#' @examples
#' int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
#' int2 <- interval(ymd("2001-06-01"), ymd("2002-01-01"))
#'
#' ymd("2001-05-03") %within% int # TRUE
#' int2 %within% int # TRUE
#' ymd("1999-01-01") %within% int # FALSE
#'
#' ## recycling
#' dates <- ymd(c("2014-12-20", "2014-12-30", "2015-01-01", "2015-01-03"))
#' blackouts<- c(interval(ymd("2014-12-30"), ymd("2014-12-31")),
#'               interval(ymd("2014-12-30"), ymd("2015-01-03")))
#' dates %within% blackouts
#'
#' ## within ANY of the intervals of a list
#' dates <- ymd(c("2014-12-20", "2014-12-30", "2015-01-01", "2015-01-03"))
#' blackouts<- list(interval(ymd("2014-12-30"), ymd("2014-12-31")),
#'                  interval(ymd("2014-12-30"), ymd("2015-01-03")))
#' dates %within% blackouts
setGeneric("%within%", useAsDefault = function(a, b) {
  stop(sprintf("No %%within%% method with signature a = %s,  b = %s",
               class(a)[[1]], class(b)[[1]]))
})

.within <- function(a, int) {
  as.numeric(a) - as.numeric(int@start) <= int@.Data & as.numeric(a) - as.numeric(int@start) >= 0
}

setMethod("%within%", signature(b = "Interval"), function(a, b) {
  if (!is.instant(a)) stop("Argument 1 is not a recognized date-time")
  a <- as.POSIXct(a)
  .within(a, b)
})

setMethod("%within%", signature(a = "Interval", b = "Interval"), function(a, b) {
  a <- int_standardize(a)
  b <- int_standardize(b)
  start.in <- as.numeric(a@start) >= as.numeric(b@start)
  end.in <- (as.numeric(a@start) + a@.Data) <= (as.numeric(b@start) + b@.Data)
  start.in & end.in
})

setMethod("%within%", signature(a = "Interval", b = "Interval"), function(a, b) {
  a <- int_standardize(a)
  b <- int_standardize(b)
  start.in <- as.numeric(a@start) >= as.numeric(b@start)
  end.in <- (as.numeric(a@start) + a@.Data) <= (as.numeric(b@start) + b@.Data)
  start.in & end.in
})

.within_instant <- function(a, b) {
  if (!all(sapply(b, is.interval)))
    stop("When second argument to %within% is a list it must contain interval objects only")
  a <- as.POSIXct(a)
  out <- FALSE
  for (int in b) {
    out <- out | .within(a, int)
  }
  out
}

setMethod("%within%", signature(a = "POSIXt", b = "list"), .within_instant)
setMethod("%within%", signature(a = "Date", b = "list"), .within_instant)

#' @export
as.list.Interval <- function(x, ...) {
  lapply(seq_along(x), function(i) x[[i]])
}

#' @export
summary.Interval <- function(object, ...) {
  nas <- is.na(object)
  object <- object[!nas]
  n <- length(object)
  dates <- c(int_start(object), int_end(object))
  earliest <- as.character(min(dates))
  latest <- as.character(max(dates))
  zone <- tz(dates)

  qq <- c(n, earliest, latest, zone)
  names(qq) <- c("Intervals", "Earliest endpoint", "Latest endpoint", "Time zone")
  if (any(nas))
    c(qq, `NA's` = sum(nas))
  else qq
}

#' @rdname time_length
setMethod("time_length", signature("Interval"), function(x, unit = "second") {
  unit <- standardise_period_names(unit)
  if (unit %in% c("year", "month")) {

    periods <- as.period(x, unit = unit)
    int_part <- slot(periods, unit)

    prev_aniv <- add_with_rollback(
      int_start(x), (int_part * period(1, units = unit)),
      roll_to_first = TRUE, preserve_hms = FALSE)
    next_aniv <- add_with_rollback(
      int_start(x), ((int_part + ifelse(x@.Data < 0, -1, 1)) * period(1, units = unit)),
      roll_to_first = TRUE, preserve_hms = FALSE)

    sofar <- as.duration(int_end(x) - prev_aniv)
    total <- as.duration(next_aniv - prev_aniv)
    int_part + sign(x@.Data) * sofar / total

  } else {
    as.duration(x) / duration(num = 1, units = unit)
  }
})

#' @export
setMethod("Arith", signature(e1 = "Interval", e2 = "ANY"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, .Generic)
})

#' @export
setMethod("Arith", signature(e1 = "ANY", e2 = "Interval"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, .Generic)
})

#' @name hidden_aliases
#' @aliases Arith,Interval,ANY-method Arith,ANY,Interval-method
#'   intersect,Interval,Interval-method union,Interval,Interval-method
#'   setdiff,Interval,Interval-method as.numeric,Interval-method
#'   show,Interval-method c,Interval-method rep,Interval-method
#'   [,Interval-method [<-,Interval,ANY,ANY,ANY-method [[,Interval-method
#'   [[<-,Interval,ANY,ANY,ANY-method $,Interval-method $<-,Interval-method
#'   as.difftime,Interval-method as.character,Interval-method
#'   +,Interval,Duration-method +,Interval,Interval-method
#'   +,Interval,Period-method +,Interval,Date-method +,Date,Interval-method
#'   +,Interval,difftime-method +,difftime,Interval-method
#'   +,Interval,numeric-method +,numeric,Interval-method
#'   +,Interval,POSIXct-method +,POSIXct,Interval-method
#'   +,Interval,POSIXlt-method +,POSIXlt,Interval-method
#'   /,Interval,Duration-method /,Interval,Interval-method
#'   /,Interval,Period-method /,Interval,difftime-method
#'   /,difftime,Interval-method /,Interval,numeric-method
#'   /,numeric,Interval-method *,Interval,ANY-method *,ANY,Interval-method
#'   -,Interval,missing-method -,Interval,Interval-method -,Date,Interval-method
#'   -,POSIXct,Interval-method -,POSIXlt,Interval-method
#'   -,numeric,Interval-method -,Interval,Date-method -,Interval,POSIXct-method
#'   -,Interval,POSIXlt-method -,Interval,numeric-method
#'   -,Duration,Interval-method -,Period,Interval-method
#'   %%,Interval,Duration-method %%,Interval,Interval-method
#'   %%,Interval,Period-method %%,Interval,Duration %%,Interval,Interval
#'   %%,Interval,Period -,Date,Interval -,Duration,Interval -,Interval,Date
#'   -,Interval,Interval -,Interval,POSIXct -,Interval,POSIXlt
#'   -,Interval,numeric -,POSIXct,Interval -,POSIXlt,Interval -,numeric,Interval
NULL

#' Dates and times made easy with lubridate
#'
#' Lubridate provides tools that make it easier to parse and
#' manipulate dates. These tools are grouped below by common
#' purpose. More information about each function can be found in
#' its help documentation.
#'
#' @section Parsing dates:
#'
#' Lubridate's parsing functions read strings into R as POSIXct
#' date-time objects. Users should choose the function whose name
#' models the order in which the year ('y'), month ('m') and day
#' ('d') elements appear the string to be parsed:
#' [dmy()], [myd()], [ymd()],
#' [ydm()], [dym()], [mdy()],
#' [ymd_hms()]). A very flexible and user friendly parser
#' is provided by [parse_date_time()].
#'
#' Lubridate can also parse partial dates from strings into
#' [Period-class] objects with the functions
#' [hm()], [hms()] and [ms()].
#'
#' Lubridate has an inbuilt very fast POSIX parser. Most of the [strptime()]
#' formats and various extensions are supported for English locales. See
#' [parse_date_time()] for more details.
#'
#' @section Manipulating dates:
#'
#' Lubridate distinguishes between moments in time (known as
#' [instants()]) and spans of time (known as time spans, see
#' [Timespan-class]). Time spans are further separated into
#' [Duration-class], [Period-class] and
#' [Interval-class] objects.
#'
#' @section Instants:
#'
#' Instants are specific moments of time. Date, POSIXct, and
#' POSIXlt are the three object classes Base R recognizes as
#' instants. [is.Date()] tests whether an object
#' inherits from the Date class. [is.POSIXt()] tests
#' whether an object inherits from the POSIXlt or POSIXct classes.
#' [is.instant()] tests whether an object inherits from
#' any of the three classes.
#'
#' [now()] returns the current system time as a POSIXct
#' object. [today()] returns the current system date.
#' For convenience, 1970-01-01 00:00:00 is saved to
#' [origin]. This is the instant from which POSIXct
#' times are calculated. Try `unclass(now())` to see the numeric structure that
#' underlies POSIXct objects. Each POSIXct object is saved as the number of seconds
#' it occurred after 1970-01-01 00:00:00.
#'
#' Conceptually, instants are a combination of measurements on different units
#' (i.e, years, months, days, etc.). The individual values for
#' these units can be extracted from an instant and set with the
#' accessor functions [second()], [minute()],
#' [hour()], [day()], [yday()],
#' [mday()], [wday()], [week()],
#' [month()], [year()], [tz()],
#' and [dst()].
#' Note: the accessor functions are named after the singular form
#' of an element. They shouldn't be confused with the period
#' helper functions that have the plural form of the units as a
#' name (e.g, [seconds()]).
#'
#' @section Rounding dates:
#'
#' Instants can be rounded to a convenient unit using the
#' functions [ceiling_date()], [floor_date()]
#' and [round_date()].
#'
#' @section Time zones:
#'
#' Lubridate provides two helper functions for working with time
#' zones. [with_tz()] changes the time zone in which an
#' instant is displayed. The clock time displayed for the instant
#' changes, but the moment of time described remains the same.
#' [force_tz()] changes only the time zone element of an
#' instant. The clock time displayed remains the same, but the
#' resulting instant describes a new moment of time.
#'
#' @section Timespans:
#'
#' A timespan is a length of time that may or may not be connected to
#' a particular instant. For example, three months is a timespan. So is an hour and
#' a half. Base R uses difftime class objects to record timespans.
#' However, people are not always consistent in how they expect time to behave.
#' Sometimes the passage of time is a monotone progression of instants that should
#' be as mathematically reliable as the number line. On other occasions time must
#' follow complex conventions and rules so that the clock times we see reflect what
#' we expect to observe in terms of daylight, season, and congruence with the
#' atomic clock. To better navigate the nuances of time, \pkg{lubridate} creates three
#' additional timespan classes, each with its own specific and consistent behavior:
#' [Interval-class], [Period-class] and
#' [Duration-class].
#'
#' [is.difftime()] tests whether an object
#' inherits from the difftime class. [is.timespan()]
#' tests whether an object inherits from any of the four timespan
#' classes.
#'
#' @section Durations:
#'
#' Durations measure the exact amount of time that occurs between two
#' instants. This can create unexpected results in relation to clock times if a
#' leap second, leap year, or change in daylight savings time (DST) occurs in
#' the interval.
#'
#' Functions for working with durations include [is.duration()],
#' [as.duration()] and [duration()]. [dseconds()],
#' [dminutes()], [dhours()],  [ddays()],
#' [dweeks()] and [dyears()] convenient lengths.
#'
#' @section Periods:
#'
#' Periods measure the change in clock time that occurs between two
#' instants. Periods provide robust predictions of clock time in the presence of
#' leap seconds, leap years, and changes in DST.
#'
#' Functions for working with periods include
#' [is.period()], [as.period()] and
#' [period()]. [seconds()],
#' [minutes()], [hours()], [days()],
#' [weeks()], [months()] and
#' [years()] quickly create periods of convenient
#' lengths.
#'
#' @section Intervals:
#'
#' Intervals are timespans that begin at a specific instant and
#' end at a specific instant. Intervals retain complete information about a
#' timespan. They provide the only reliable way to convert between
#' periods and durations.
#'
#' Functions for working with intervals include
#' [is.interval()], [as.interval()],
#' [interval()], [int_shift()],
#' [int_flip()], [int_aligns()],
#' [int_overlaps()], and
#' \code{\link{\%within\%}}. Intervals can also be manipulated with
#' intersect, union, and setdiff().
#'
#' @section Miscellaneous:
#'
#' [decimal_date()] converts an instant to a decimal of
#' its year.
#' [leap_year()] tests whether an instant occurs during
#' a leap year.
#' [pretty_dates()] provides a method of making pretty
#' breaks for date-times.
#' [lakers] is a data set that contains information
#' about the Los Angeles Lakers 2008-2009 basketball season.
#'
#' @references Garrett Grolemund, Hadley Wickham (2011). Dates and Times Made
#'   Easy with lubridate. Journal of Statistical Software, 40(3), 1-25.
#'   \url{https://www.jstatsoft.org/v40/i03/}.
#' @importFrom methods setClass setGeneric new show allNames callGeneric is slot slot<- slotNames validObject Compare Arith initialize coerce<-
#' @importFrom utils packageVersion read.delim
#' @importFrom stats na.omit setNames update
#' @importFrom timechange time_update time_add time_get time_force_tz
#' @useDynLib lubridate, .registration=TRUE
#' @keywords internal
"_PACKAGE"


## NOTE ON EXPORTS: `useDynLib` from above exports all registered functions (see
## ../src/cpp11.cpp). _lubridate_C_xyz in there are C++/cpp11 functions. All
## other `C_xyz` functions are from plain .C files and are registered
## automatically by cpp11 when there are .Call() calls present that reference
## those functions (cpp11 does this through the usage of
## tools::package_native_routine_registration_skeleton(".", character_only = FALSE)).

##' @include util.r
NULL

## #' @export
## setMethod("<", signature(e1 = "POSIXct", e2 = "Date"),
##           function(e1, e2) {
##             cat("here")
##             callNextMethod(e1, .POSIXct(e2, tz = attr(e1, "tzone")))
##           })

## #' @export
## setMethod("<", signature(e1 = "Date", e2 = "POSIXct"),
##           function(e1, e2) {
##             cat("here")
##            callNextMethod(.POSIXct(e1, tz = attr(e2, "tzone")), e2)
##           })

comp_posix_date <- function(e1, e2) {
  if (nargs() == 1) {
    stop(gettextf("unary '%s' not defined for \"%s\" objects", .Generic, class(e1)), domain = NA)
  }

  if (inherits(e1, "POSIXlt")) {
    e1 <- as.POSIXct(e1)
  }
  if (inherits(e2, "POSIXlt")) {
    e2 <- as.POSIXct(e2)
  }

  if (is.character(e1)) {
    e1 <-
      if (is.Date(e2)) {
        as.Date(e1)
      } else {
        as.POSIXct(e1)
      }
  }

  if (is.character(e2)) {
    e2 <-
      if (is.Date(e1)) {
        as.Date(e2)
      } else {
        as.POSIXct(e2)
      }
  }

  if (is.POSIXct(e1) && is.Date(e2)) {
    e2 <- date_to_posix(e2, tz(e1))
    base::check_tzones(e1, e2)
  } else if (is.Date(e1) && is.POSIXct(e2)) {
    e1 <- date_to_posix(e1, tz = tz(e2))
    base::check_tzones(e1, e2)
  }
  NextMethod(.Generic)
}

## Nothing else seems to work, only this sneaky trick.
evalqOnLoad({
  for (op in c("<", ">", "==", ">=", "<=", "!=")) {
    registerS3method(op, "Date", comp_posix_date)
    registerS3method(op, "POSIXt", comp_posix_date)
  }
})

cpp_update_dt <- function(dt,
                          year,
                          month,
                          yday,
                          mday,
                          wday,
                          hour,
                          minute,
                          second,
                          tz,
                          roll,
                          week_start) {
  # Catch rare integer POSIXct, retaining attributes
  storage.mode(dt) <- "double"

  year <- as.integer(year)
  month <- as.integer(month)
  yday <- as.integer(yday)
  mday <- as.integer(mday)
  wday <- as.integer(wday)
  hour <- as.integer(hour)
  minute <- as.integer(minute)

  second <- as.double(second)

  roll <- as.logical(roll)
  week_start <- as.integer(week_start)

  C_update_dt(
    dt = dt,
    year = year,
    month = month,
    yday = yday,
    mday = mday,
    wday = wday,
    hour = hour,
    minute = minute,
    second = second,
    tz = tz,
    roll = roll,
    week_start = week_start
  )
}

cpp_force_tz <- function(dt, tz, roll) {
  # Catch rare integer POSIXct, retaining attributes
  storage.mode(dt) <- "double"

  roll <- as.logical(roll)

  # Must be a character vector, never `NULL`
  tz <- as.character(tz)

  C_force_tz(
    dt = dt,
    tz = tz,
    roll = roll
  )
}

cpp_force_tzs <- function(dt, tzs, tz_out, roll) {
  # Catch rare integer POSIXct, retaining attributes
  storage.mode(dt) <- "double"

  # Must be a character vector, never `NULL`
  tzs <- as.character(tzs)
  tz_out <- as.character(tz_out)

  roll <- as.logical(roll)

  C_force_tzs(
    dt = dt,
    tzs = tzs,
    tz_out = tz_out,
    roll = roll
  )
}

cpp_local_time <- function(dt, tzs) {
  # Catch rare integer POSIXct, retaining attributes
  storage.mode(dt) <- "double"

  # Must be a character vector, never `NULL`
  tzs <- as.character(tzs)

  C_local_time(
    dt = dt,
    tzs = tzs
  )
}

#' @include accessors-month.r
NULL

#' Get the fiscal quarter and semester of a date-time
#'
#' Quarters divide the year into fourths. Semesters divide the year into halfs.
#'
#' @param x a date-time object of class POSIXct, POSIXlt, Date, chron, yearmon,
#'   yearqtr, zoo, zooreg, timeDate, xts, its, ti, jul, timeSeries, fts or
#'   anything else that can be converted with as.POSIXlt
#' @param with_year logical indicating whether or not to include the quarter's
#'   year.
#' @param fiscal_start numeric indicating the starting month of a fiscal year
#' @return numeric
#' @examples
#' x <- ymd(c("2012-03-26", "2012-05-04", "2012-09-23", "2012-12-31"))
#' quarter(x)
#' quarter(x, with_year = TRUE)
#' quarter(x, with_year = TRUE, fiscal_start = 11)
#' semester(x)
#' semester(x, with_year = TRUE)
#' @export
quarter <- function(x, with_year = FALSE, fiscal_start = 1) {
  fs <- (fiscal_start - 1) %% 12
  shifted <- seq(fs, 11 + fs) %% 12 + 1
  m <- month(x)
  quarters <- rep(1:4, each = 3)
  s <- match(m, shifted)
  q <- quarters[s]
  if (with_year) {
    this_year <- if (fs != 0) (fs + 1):12
    year(x) + (m %in% this_year) + q/10
  } else {
    q
  }
}

#' @rdname quarter
#' @export
semester <- function(x, with_year = FALSE) {
  m <- month(x)
  semesters <- rep(1:2, each = 6)
  s <- semesters[m]
  if (with_year) year(x) + s/10
  else  s
}

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
#' @include difftimes.r
#' @include numeric.r
NULL


divide_duration_by_duration <- function(dur1, dur2)
  dur1@.Data / dur2@.Data

divide_duration_by_difftime <- function(dur, diff)
  dur@.Data / as.numeric(diff, units = "secs")

divide_duration_by_number <- function(dur, num)
  new("Duration", dur@.Data / num)

divide_interval_by_duration <- function(int, dur) {
  int@.Data / dur@.Data
}

adjust_estimate <- function(est, int, per) {
  start <- int_start(int)
  end <- int_end(int)
  est <- ceiling(est)

  up_date <- add_with_rollback(start, est * per)
  while (length(which <- which(up_date < end))) {
    up_date[which] <- add_with_rollback(up_date[which], per[which])
    est[which] <- est[which] + 1
  }

  low_date <- up_date ## add_with_rollback(up_date, -per)
  while (length(which <- which(low_date > end))) {
    up_date[which] <- low_date[which]
    low_date[which] <- add_with_rollback(low_date[which], -per[which])
    est[which] <- est[which] - 1
  }

  frac <-
    as.numeric(difftime(end, low_date, units = "secs"))/
    as.numeric(difftime(up_date, low_date, units = "secs"))

  frac[low_date == up_date] <- 0

  est + frac
}

divide_interval_by_period <- function(int, per) {
  if (length(int) == 0 || length(per) == 0)
    return(numeric())

  estimate <- int/as.duration(per)
  not_nas <- !is.na(estimate)
  timespans <- match_lengths(int, per)
  if (all(not_nas)) {
    adjust_estimate(estimate, timespans[[1]], timespans[[2]])
  } else {
    int2 <- timespans[[1]][not_nas]
    per2 <- timespans[[2]][not_nas]
    estimate[not_nas] <- adjust_estimate(estimate[not_nas], int2, per2)
    estimate
  }
}

divide_interval_by_difftime <- function(int, diff) {
  int@.Data / as.double(diff, units = "secs")
}

divide_interval_by_number <- function(int, num) {
  starts <- int@start + rep(0, length(num))
  new("Interval", int@.Data / num, start = starts, tzone = int@tzone)
}


divisible_period <- function(per, anchor) {
  per@month <- per@month + 12*per@year
  per@year <- rep(0, length(per@year))

  secs.in.months <- as.numeric(anchor + months(per@month)) - as.numeric(anchor)
  days.in.months <- round(secs.in.months/86400)

  per@day <- per@day + days.in.months
  per@month <- 0

  per
}


divide_period_by_period <- function(per1, per2) {
  period_to_seconds(per1) / period_to_seconds(per2)
}

divide_period_by_number <- function(per, num) {
  new("Period", per@.Data / num, year = per@year / num,
      month = per@month / num, day = per@day / num, hour = per@hour / num,
      minute = per@minute / num)
}



divide_difftime_by_duration <- function(dif, dur)
  as.numeric(dif, units = "secs") / dur@.Data






#' @export
setMethod("/", signature(e1 = "Duration", e2 = "Duration"),
          function(e1, e2) divide_duration_by_duration(e1, e2))

#' @export
setMethod("/", signature(e1 = "Duration", e2 = "Interval"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod("/", signature(e1 = "Duration", e2 = "Period"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod("/", signature(e1 = "Duration", e2 = "difftime"),
          function(e1, e2) divide_duration_by_difftime(e1, e2))

#' @export
setMethod("/", signature(e1 = "Duration", e2 = "numeric"),
          function(e1, e2) divide_duration_by_number(e1, e2))


#' @export
setMethod("/", signature(e1 = "Interval", e2 = "Duration"),
          function(e1, e2) divide_interval_by_duration(e1, e2))

#' @export
setMethod("/", signature(e1 = "Interval", e2 = "Interval"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod("/", signature(e1 = "Interval", e2 = "Period"),
          function(e1, e2) divide_interval_by_period(e1, e2))

#' @export
setMethod("/", signature(e1 = "Interval", e2 = "difftime"),
          function(e1, e2) divide_interval_by_difftime(e1, e2))

#' @export
setMethod("/", signature(e1 = "Interval", e2 = "numeric"),
          function(e1, e2) divide_interval_by_number(e1, e2))


#' @export
setMethod("/", signature(e1 = "Period", e2 = "Duration"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod("/", signature(e1 = "Period", e2 = "Interval"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod("/", signature(e1 = "Period", e2 = "Period"),
          function(e1, e2) divide_period_by_period(e1, e2))

#' @export
setMethod("/", signature(e1 = "Period", e2 = "difftime"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod("/", signature(e1 = "Period", e2 = "numeric"),
          function(e1, e2) divide_period_by_number(e1, e2))



#' @export
setMethod("/", signature(e1 = "difftime", e2 = "Duration"),
          function(e1, e2) divide_difftime_by_duration(e1, e2))

#' @export
setMethod("/", signature(e1 = "difftime", e2 = "Interval"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})

#' @export
setMethod("/", signature(e1 = "difftime", e2 = "Period"), function(e1, e2) {
  stop_incompatible_classes(e1, e2, "/")
})



#' @export
setMethod("/", signature(e1 = "numeric", e2 = "Duration"),
          function(e1, e2) stop_incompatible_classes(e1, e2, "/"))
#' @export
setMethod("/", signature(e1 = "numeric", e2 = "Interval"),
          function(e1, e2) stop_incompatible_classes(e1, e2, "/"))
#' @export
setMethod("/", signature(e1 = "numeric", e2 = "Period"),
          function(e1, e2) stop_incompatible_classes(e1, e2, "/"))

#' @include periods.r
NULL

#' Get/set weeks component of a date-time
#'
#' @description
#' `week()` returns the number of complete seven day periods that have
#' occurred between the date and January 1st, plus one.

#' @param x a date-time object. Must be a POSIXct, POSIXlt, Date, chron,
#'   yearmon, yearqtr, zoo, zooreg, timeDate, xts, its, ti, jul, timeSeries, or
#'   fts object.
#' @param value a numeric object
#' @return the weeks element of x as an integer number
#' @keywords utilities manip chron
#' @references
#'    \url{https://en.wikipedia.org/wiki/ISO_week_date}
#' @seealso [isoyear()]
#' @examples
#' x <- ymd("2012-03-26")
#' week(x)
#' week(x) <- 1
#' week(x) <- 54
#' week(x) > 3
#' @export
week <- function(x) {
  (yday(x) - 1) %/% 7 + 1
}

#' @rdname week
#' @export
"week<-" <- function(x, value) {
  x <- x + days((value - week(x)) * 7)
}

.other_week <- function(x, week_start) {
  x <- as.POSIXlt(x)
  date <- make_date(year(x), month(x), day(x))
  wday <- wday(x, week_start = week_start)
  date <- date + (4 - wday)
  jan1 <- as.numeric(make_date(year(date), 1, 1))
  1L + (as.numeric(date) - jan1) %/% 7L
}

#' @description `isoweek()` returns the week as it would appear in the ISO 8601
#'   system, which uses a reoccurring leap week.
#' @rdname week
#' @export
isoweek <- function(x) {
  .other_week(x, 1)
}

#' @description `epiweek()` is the US CDC version of epidemiological week. It
#'   follows same rules as `isoweek()` but starts on Sunday. In other parts of
#'   the world the convention is to start epidemiological weeks on Monday,
#'   which is the same as `isoweek`.
#'
#' @rdname week
#' @export
epiweek <- function(x) {
  .other_week(x, 7)
}

##' Deprecated functions in the lubridate package
##' @param x numeric value to be converted into duration
##' @param ... arguments to be passed to the functions
##'     (obscured to enforce the usage of new functions)
##' @rdname Deprecated
##' @keywords internal
##' @name Deprecated-lubridate
NULL

## modified ggplot2:::gg_dep
.deprecated_fun <- function(replacement, version, n_call = 1) {
  name <- as.character(sys.call(-(n_call))[[1]])
  name <- sprintf("'%s'", name[[length(name)]])
  replacement <- sprintf("; use '%s' instead", replacement)
  .deprecated(name, replacement, version)
}

.deprecated_arg <- function(arg, version, n_call = 1) {
  name <- as.character(sys.call(-n_call)[[1]])
  mes <- sprintf("Argument '%s' of '%s'", arg, name)
  .deprecated(mes, "", version)
}

.deprecated <- function(name, replacement = NULL, version) {

  v <- as.package_version(version)
  cv <- packageVersion("lubridate")

  replacement <- if (is.null(replacement)) "" else replacement

  ## If current major number is greater than last-good major number, or if
  ## current minor number is more than 2 greater than last-good minor number,
  ## give error.
  if (cv[[1, 1]] > v[[1, 1]]  ||  cv[[1, 2]] > v[[1, 2]] + 2) {
    stop(sprintf("%s is defunct%s. Deprecated in version '%s'.",
                 name, replacement, version),
         call. = FALSE)
  } else if (cv[[1, 2]] > v[[1, 2]]) {
    ## If minor number differs by one, give warning
    warning(sprintf("%s is deprecated%s. Deprecated in version '%s'.",
                    name, replacement, version),
            call. = FALSE)
  } else if (cv[[1, 3]] >=  v[[1, 3]]) {
    ## If only subminor number is greater, give message
    message(sprintf("%s is deprecated%s. Deprecated in version '%s'.",
                    name, replacement, version))
  }
  invisible()

}

#' Round, floor and ceiling methods for date-time objects
#'
#' @description
#' `round_date()` takes a date-time object and time unit, and rounds it to the nearest value
#' of the specified time unit. For rounding date-times which are exactly halfway
#' between two consecutive units, the convention is to round up. Note that this
#' is in line with the behavior of R's [base::round.POSIXt()] function
#' but does not follow the convention of the base [base::round()] function
#' which "rounds to the even digit", as per IEC 60559.
#'
#' Rounding to the nearest unit or multiple of a unit is supported. All
#' meaningful specifications in the English language are supported - secs, min,
#' mins, 2 minutes, 3 years etc.
#'
#' Rounding to fractional seconds is also supported. Please note that rounding to
#' fractions smaller than 1 second can lead to large precision errors due to the
#' floating point representation of the POSIXct objects. See examples.
#'
#'
#' @details In \pkg{lubridate}, functions that round date-time objects try to
#'   preserve the class of the input object whenever possible. This is done by
#'   first rounding to an instant, and then converting to the original class as per
#'   usual R conventions.
#'
#'
#' @section Rounding Up Date Objects:
#'
#' By default, rounding up `Date` objects follows 3 steps:
#'
#' 1. Convert to an instant representing lower bound of the Date:
#'    `2000-01-01` --> `2000-01-01 00:00:00`
#'
#' 2. Round up to the \strong{next} closest rounding unit boundary. For example,
#'    if the rounding unit is `month` then next closest boundary of `2000-01-01`
#'    is `2000-02-01 00:00:00`.
#'
#'    The motivation for this is that the "partial" `2000-01-01` is conceptually
#'    an interval (`2000-01-01 00:00:00` -- `2000-01-02 00:00:00`) and the day
#'    hasn't started clocking yet at the exact boundary `00:00:00`. Thus, it
#'    seems wrong to round a day to its lower boundary.
#'
#'    Behavior on the boundary can be changed by setting
#'    `change_on_boundary` to `TRUE` or `FALSE`.
#'
#' 3. If the rounding unit is smaller than a day, return the instant from step 2
#'     (`POSIXct`), otherwise convert to and return a `Date` object.
#'
#' @rdname round_date
#' @param x a vector of date-time objects
#' @param unit a string, `Period` object or a date-time object. When a singleton string,
#'   it specifies a time unit or a multiple of a unit to be rounded to. Valid base units
#'   are `second`, `minute`, `hour`, `day`, `week`, `month`, `bimonth`, `quarter`,
#'   `season`, `halfyear` and `year`. Arbitrary unique English abbreviations as in the
#'   [period()] constructor are allowed. Rounding to multiples of units (except weeks)
#'   is supported. When `unit` is a `Period` object, units of the period objects are
#'   used. This is equivalent to converting the period object to its string
#'   representation and passing as `unit` argument.
#'
#'   When `unit` is a date-time object rounding is done to the nearest of the
#'   elements in `unit`. If range of `unit` vector does not cover the range of
#'   `x` `ceiling_date()` and `floor_date()` round to the `max(x)` and `min(x)`
#'   for elements that fall outside of `range(unit)`.
#' @param change_on_boundary if this is `NULL` (the default), instants on the
#'   boundary remain unchanged, but `Date` objects are rounded up to the next
#'   boundary. If this is `TRUE`, instants on the boundary are rounded up to the
#'   next boundary. If this is `FALSE`, nothing on the boundary is rounded up at
#'   all. This was the default for \pkg{lubridate} prior to `v1.6.0`. See
#'   section `Rounding Up Date Objects` below for more details.
#'
#' @inheritParams DateTimeUpdate
#' @return When `unit` is a string, return a Date object if `x` is a Date and
#'   `unit` is larger or equal than "day", otherwise a POSIXct object. When
#'   `unit` is a date-time object, return a date-time object of the same class
#'   and same time zone as `unit`.
#' @keywords manip chron
#' @seealso [base::round()]
#' @examples
#'
#' ## print fractional seconds
#' options(digits.secs = 6)
#'
#' x <- ymd_hms("2009-08-03 12:01:59.23")
#' round_date(x, ".5s")
#' round_date(x, "sec")
#' round_date(x, "second")
#' round_date(x, "minute")
#' round_date(x, "5 mins")
#' round_date(x, "hour")
#' round_date(x, "2 hours")
#' round_date(x, "day")
#' round_date(x, "week")
#' round_date(x, "month")
#' round_date(x, "bimonth")
#' round_date(x, "quarter") == round_date(x, "3 months")
#' round_date(x, "halfyear")
#' round_date(x, "year")
#'
#' x <- ymd_hms("2009-08-03 12:01:59.23")
#' floor_date(x, ".1s")
#' floor_date(x, "second")
#' floor_date(x, "minute")
#' floor_date(x, "hour")
#' floor_date(x, "day")
#' floor_date(x, "week")
#' floor_date(x, "month")
#' floor_date(x, "bimonth")
#' floor_date(x, "quarter")
#' floor_date(x, "season")
#' floor_date(x, "halfyear")
#' floor_date(x, "year")
#'
#' x <- ymd_hms("2009-08-03 12:01:59.23")
#' ceiling_date(x, ".1 sec") # imprecise representation at 0.1 sec !!!
#' ceiling_date(x, "second")
#' ceiling_date(x, "minute")
#' ceiling_date(x, "5 mins")
#' ceiling_date(x, "hour")
#' ceiling_date(x, "day")
#' ceiling_date(x, "week")
#' ceiling_date(x, "month")
#' ceiling_date(x, "bimonth") == ceiling_date(x, "2 months")
#' ceiling_date(x, "quarter")
#' ceiling_date(x, "season")
#' ceiling_date(x, "halfyear")
#' ceiling_date(x, "year")
#'
#' ## Period unit argument
#' floor_date(x, days(2))
#' floor_date(x, years(1))
#'
#' ## As of R 3.4.2 POSIXct printing of fractional numbers is wrong
#' as.POSIXct("2009-08-03 12:01:59.3") ## -> "2009-08-03 12:01:59.2 CEST"
#' ceiling_date(x, ".1 sec") ## -> "2009-08-03 12:01:59.2 CEST"
#'
#' ## behaviour of `change_on_boundary`
#' ## As per default behaviour `NULL`, instants on the boundary remain the
#' ## same but dates are rounded up
#' ceiling_date(ymd_hms("2000-01-01 00:00:00"), "month")
#' ceiling_date(ymd("2000-01-01"), "month")
#'
#' ## If `TRUE`, both instants and dates on the boundary are rounded up
#' ceiling_date(ymd_hms("2000-01-01 00:00:00"), "month", change_on_boundary = TRUE)
#' ceiling_date(ymd("2000-01-01"), "month")
#'
#' ## If `FALSE`, both instants and dates on the boundary remain the same
#' ceiling_date(ymd_hms("2000-01-01 00:00:00"), "month", change_on_boundary = FALSE)
#' ceiling_date(ymd("2000-01-01"), "month")
#' @export
round_date <- function(x, unit = "second", week_start = getOption("lubridate.week.start", 7)) {
  if (length(x) == 0) {
    return(x)
  }

  if (is.POSIXt(unit) || is.Date(unit)) {

    .round <- function(below, above) {
      mid <- unclass(xx)
      below <- unclass(below)
      above <- unclass(above)
      wabove <- (above - mid) <= (mid - below)
      wabove <- !is.na(wabove) & wabove
      new <- below
      new[wabove] <- above[wabove]
      .POSIXct(new, tz = tz(x))
    }

    xx <- as.POSIXct(as_datetime(x, tz = tz(x)))
    u <- as_datetime(unit, tz = tz(unit))
    below <- floor_to_posix(xx, u, TRUE)
    above <- ceil_to_posix(xx, u, TRUE)
    reclass_date(.round(below, above), unit)

  } else {

    unit <- as.character(unit)
    ## FIXME: remove when multi-unit week is implemented in timechange
    unit <- .normalize_multi_week_unit(unit)
    timechange::time_round(x, unit = unit, week_start = as_week_start(week_start))

  }
}

#' @description
#' `floor_date()` takes a date-time object and rounds it down to the nearest
#' boundary of the specified time unit.
#' @rdname round_date
#' @export
floor_date <- function(x, unit = "seconds",
                       week_start = getOption("lubridate.week.start", 7)) {
  if (length(x) == 0) {
    return(x)
  }

  if (is.POSIXct(unit) || is.Date(unit)) {
    xx <- as_datetime(x, tz = tz(x))
    new <- floor_to_posix(xx, as_datetime(unit, tz = tz(unit)))
    return(reclass_date(new, unit))
  }

  unit <- as.character(unit)
  ## FIXME: remove when multi-unit week is implemented in timechange
  unit <- .normalize_multi_week_unit(unit)
  timechange::time_floor(x, unit = unit, week_start = as_week_start(week_start))
}

.normalize_multi_week_unit <- function(unit) {
  ## temp workaround not to throw an error on multi-unit weeks
  unit <- as.character(unit)
  parsed <- timechange:::parse_units(unit)
  if (timechange:::standardise_unit_name(parsed$unit) == "week" && parsed$n > 1) {
    warning("Multi-week unit is not supported. Ignoring.")
    unit <- "week"
  }
  unit
}

#' @description
#' `ceiling_date()` takes a date-time object and rounds it up to the nearest
#' boundary of the specified time unit.
#' @rdname round_date
#' @export
#' @examples
#'
#' x <- ymd_hms("2000-01-01 00:00:00")
#' ceiling_date(x, "month")
#' ceiling_date(x, "month", change_on_boundary = TRUE)
#'
#' ## For Date objects first day of the month is not on the
#' ## "boundary". change_on_boundary applies to instants only.
#' x <- ymd("2000-01-01")
#' ceiling_date(x, "month")
#' ceiling_date(x, "month", change_on_boundary = TRUE)
ceiling_date <- function(x, unit = "seconds", change_on_boundary = NULL,
                         week_start = getOption("lubridate.week.start", 7)) {
  if (length(x) == 0) {
    return(x)
  }

  if (is.null(change_on_boundary)) {
    change_on_boundary <- inherits(x, "Date")
  }

  if (is.POSIXct(unit) || is.Date(unit)) {
    xx <- as_datetime(x, tz = tz(x))
    new <- ceil_to_posix(xx, as_datetime(unit, tz = tz(unit)))
    return(reclass_date(new, unit))
  }

  unit <- as.character(unit)
  ## FIXME: remove when multi-unit week is implemented in timechange
  unit <- .normalize_multi_week_unit(unit)

  timechange::time_ceiling(x, unit = unit,
    change_on_boundary = as.logical(change_on_boundary), # FIXME: remove this protection with devel timechange
    week_start = as_week_start(week_start))
}

check_round_endpoints <- function(unit) {
  if (length(unit) == 0) {
    stop("Zero length `unit` argument", call. = FALSE)
  }
  if (any(is.na(unit))) {
    stop("NAs in `unit` argument`", call. = FALSE)
  }
}

floor_to_posix <- function(x, endpoints, tight = FALSE) {
  check_round_endpoints(endpoints)
  if (all(is.na(x))) {
    return(x)
  }
  endpoints <- sort(endpoints)
  ux <- unclass(x)
  min <- min(ux, na.rm = TRUE)
  breaks <- unclass(endpoints)
  emin <- breaks[[1]]
  if (min < emin) {
    beg <- .POSIXct(if (tight) emin else min, tz = tz(endpoints))
    endpoints <- c(beg, endpoints)
    breaks <- c(min, breaks)
  }
  breaks <- c(breaks, Inf)
  bins <- .bincode(ux, breaks, right = FALSE, include.lowest = TRUE)
  endpoints[bins]
}

ceil_to_posix <- function(x, endpoints, tight = FALSE) {
  check_round_endpoints(endpoints)
  if (all(is.na(x))) {
    return(x)
  }
  endpoints <- sort(endpoints)
  ux <- unclass(x)
  max <- max(ux, na.rm = TRUE)
  breaks <- unclass(endpoints)
  emax <- breaks[[length(breaks)]]
  if (emax < max) {
    end <- .POSIXct(if (tight) emax else max, tz = tz(endpoints))
    endpoints <- c(endpoints, end)
    breaks <- c(breaks, max)
  }
  breaks <- c(-Inf, breaks)
  bins <- .bincode(ux, breaks, right = TRUE, include.lowest = TRUE)
  out <- endpoints[bins]
}

test_that("floor_date works as expected with POSIX units", {
  set.seed(11111)
  x <- ymd_hms(c("2020-01-03 00:01:03", "2020-02-04 00:01:03", "2020-03-05 00:00:00"), tz = "America/New_York")
  min <- min(x)
  b <- ymd_hms(c("2020-01-03 00:00:01", "2020-02-05 00:00:01"))
  expect_equal(floor_date(x, b), b[c(1, 1, 2)])
  expect_equal(floor_date(x, sample(b)), b[c(1, 1, 2)])
  expect_equal(floor_date(x, rev(b)), b[c(1, 1, 2)])
  expect_equal(floor_date(b, x), with_tz(c(b[1], x[2]), "America/New_York"))
  expect_equal(floor_date(b, sample(x)), with_tz(c(b[1], x[2]), "America/New_York"))
  expect_equal(floor_date(rev(b), x), with_tz(rev(c(b[1], x[2])), "America/New_York"))
  expect_equal(floor_date(rev(x), b), rev(b[c(1, 1, 2)]))
  expect_equal(floor_date(x, b[[2]]), c(with_tz(c(min, min), "UTC"), b[[2]]))
  expect_equal(floor_date(ymd_hms("2030-01-01 00:03:00"), x), x[[3]])
  expect_equal(floor_date(y <- ymd_hms("2010-01-01 00:03:00"), x), with_tz(y, "America/New_York"))
  expect_equal(floor_date(ymd_hms("2030-01-01 00:03:00"), b), b[2])
})

test_that("floor_date works as expected with Date units", {
  set.seed(11111)
  x <- ymd(c("2020-01-03", "2020-02-04", "2020-03-05"))
  min <- min(x)
  b <- ymd(c("2020-01-03", "2020-02-05"))
  expect_equal(floor_date(x, b), b[c(1, 1, 2)])
  expect_equal(floor_date(x, sample(b)), b[c(1, 1, 2)])
  expect_equal(floor_date(x, rev(b)), b[c(1, 1, 2)])
  expect_equal(floor_date(b, x), c(b[1], x[2]))
  expect_equal(floor_date(b, sample(x)), c(b[1], x[2]))
  expect_equal(floor_date(rev(b), x), rev(c(b[1], x[2])))
  expect_equal(floor_date(rev(x), b), rev(b[c(1, 1, 2)]))
  expect_equal(floor_date(x, b[[2]]), c(c(min, min), b[[2]]))
  expect_equal(floor_date(ymd("2030-01-01"), x), x[[3]])
  expect_equal(floor_date(y <- ymd("2010-01-01"), x), y)
  expect_equal(floor_date(ymd("2030-01-01"), b), b[2])
})

test_that("ceiling_date works as expected with POSIX units", {
  set.seed(11111)
  x <- ymd_hms(c("2020-01-03 00:01:03", "2020-02-04 00:01:03", "2020-03-05 00:00:00"), tz = "America/New_York")
  na <- x
  na[[2]] <- NA
  max <- max(x)
  b <- ymd_hms(c("2020-01-03 00:00:01", "2020-02-05 00:00:01"))
  expect_equal(ceiling_date(x, b), c(b[c(2, 2)], max))
  expect_equal(ceiling_date(na, b), c(b[2], NA, max))
  expect_equal(ceiling_date(rev(x), b), rev(c(b[c(2, 2)], max)))
  expect_equal(ceiling_date(x, sample(b)), c(b[c(2, 2)], max))
  expect_equal(ceiling_date(x, rev(b)), c(b[c(2, 2)], max))
  expect_equal(ceiling_date(b, x), x[c(1, 3)])
  expect_equal(ceiling_date(b, sample(x)), x[c(1, 3)])
  expect_equal(ceiling_date(rev(b), x), rev(x[c(1, 3)]))
  expect_equal(ceiling_date(x, b[[2]]), c(b[c(2, 2)], max))
  expect_equal(ceiling_date(y <- ymd_hms("2030-01-01 00:03:00"), x), with_tz(y, "America/New_York"))
  expect_equal(ceiling_date(y <- ymd_hms("2010-01-01 00:03:00"), x), x[[1]])
})

test_that("ceiling_date works as expected with Date units", {
  set.seed(11111)
  x <- ymd(c("2020-01-04", "2020-02-04", "2020-03-05"))
  na <- x
  na[[2]] <- NA
  max <- max(x)
  b <- ymd(c("2020-01-03", "2020-02-05"))
  expect_equal(ceiling_date(x, b), c(b[c(2, 2)], max))
  expect_equal(ceiling_date(x, rep(b, 3)), c(b[c(2, 2)], max))
  expect_equal(ceiling_date(na, b), c(b[2], NA, max))
  expect_equal(ceiling_date(rev(x), b), rev(c(b[c(2, 2)], max)))
  expect_equal(ceiling_date(x, sample(b)), c(b[c(2, 2)], max))
  expect_equal(ceiling_date(x, rev(b)), c(b[c(2, 2)], max))
  expect_equal(ceiling_date(b, x), x[c(1, 3)])
  expect_equal(ceiling_date(b, sample(rep(x, 8))), x[c(1, 3)])
  expect_equal(ceiling_date(rev(b), x), rev(x[c(1, 3)]))
  expect_equal(ceiling_date(x, b[[2]]), c(b[c(2, 2)], max))
  expect_equal(ceiling_date(y <- ymd("2030-01-01"), x), y)
  expect_equal(ceiling_date(y <- ymd("2010-01-01"), x), x[[1]])
})

test_that("round_date works as expected with Date units", {
  set.seed(11111)
  x <- ymd(c("2020-01-04", "2020-02-04", "2020-03-05"))
  na <- x
  na[[2]] <- NA
  max <- max(x)
  b <- ymd(c("2020-01-03", "2020-02-05"))
  expect_equal(round_date(x, b), b[c(1, 2, 2)])
  expect_equal(round_date(na, b), c(b[1], NA, b[2]))
  expect_equal(round_date(rev(x), b), b[c(2, 2, 1)])
  expect_equal(round_date(x, sample(b)), b[c(1, 2, 2)])
  expect_equal(round_date(x, rev(b)), b[c(1, 2, 2)])
  expect_equal(round_date(x, rep(b, 10)), b[c(1, 2, 2)])
  expect_equal(round_date(b, x), x[c(1, 2)])
  expect_equal(round_date(b, sample(x)), x[c(1, 2)])
  expect_equal(round_date(rev(b), x), rev(x[c(1, 2)]))
  expect_equal(round_date(y <- ymd("2030-01-01"), x), x[3])
  expect_equal(round_date(y <- ymd("2010-01-01"), x), x[1])
})

test_that("round_date works on POSIXlt objects", {
  x <- as.POSIXlt(ymd_hms(c("2012-11-01 23:41:00", "2012-11-01 00:44:10"), tz = "America/New_York"))
  expect_equal(
    round_date(x, "5 mins"),
    as.POSIXlt(ymd_hms(c("2012-11-01 23:40:00", "2012-11-01 00:45:00"), tz = "America/New_York"))
  )
})

test_that("round_date fails with invalid date-time units", {
  expect_error(round_date(now(), .POSIXct(double())), "Zero")
  expect_error(floor_date(now(), .POSIXct(double())), "Zero")
  expect_error(ceiling_date(now(), .POSIXct(double())), "Zero")
  expect_error(round_date(now(), .POSIXct(NA_real_)), "NAs")
  expect_error(floor_date(now(), .POSIXct(NA_real_)), "NAs")
  expect_error(ceiling_date(now(), .POSIXct(NA_real_)), "NAs")
})

test_that("floor_date works for each time element", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_identical(floor_date(x, "second"), as.POSIXct("2009-08-03 12:01:59", tz = "UTC"))
  expect_identical(floor_date(x, "minute"), as.POSIXct("2009-08-03 12:01:00", tz = "UTC"))
  expect_identical(floor_date(x, "hour"), as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "day"), as.POSIXct("2009-08-03 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "week"), as.POSIXct("2009-08-02 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "month"), as.POSIXct("2009-08-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "bimonth"), as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "quarter"), as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "halfyear"), as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "year"), as.POSIXct("2009-01-01 00:00:00", tz = "UTC"))
})

test_that("floor_date and round_date throw on invalid multi-unit spec", {
  x <- ymd_hms("2009-08-03 12:01:57.11")

  expect_error(floor_date(x, "120 sec"))
  expect_error(floor_date(x, "120 min"))
  expect_error(floor_date(x, "25 h"))
  expect_error(floor_date(x, "32 days"))
})

test_that("floor_date works for multi-units", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_identical(floor_date(x, "2 secs"), as.POSIXct("2009-08-03 12:01:58", tz = "UTC"))
  expect_identical(floor_date(x, "2s"), as.POSIXct("2009-08-03 12:01:58", tz = "UTC"))
  expect_identical(floor_date(x, "2 mins"), as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2M"), as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2 hours"), as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2h"), as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2 days"), as.POSIXct("2009-08-03 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "3 days"), as.POSIXct("2009-08-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "10 days"), as.POSIXct("2009-08-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "10d"), as.POSIXct("2009-08-01 00:00:00", tz = "UTC"))
  ## expect_identical(floor_date(x, "2 week"), as.POSIXct("2009-08-xx 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2 month"), as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2m"), as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2 bimonth"), floor_date(x, "4 months"))
  expect_identical(floor_date(x, "2 quarter"), floor_date(x, "6 months"))
  expect_identical(floor_date(x, "2 halfyear"), floor_date(x, "year"))
  expect_identical(floor_date(x, "2 year"), as.POSIXct("2008-01-01 00:00:00", tz = "UTC"))
})

test_that("floor_date works for fractional multi-units", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_identical(floor_date(x, ".2 secs"), as.POSIXct("2009-08-03 12:01:59.2", tz = "UTC"))
  expect_identical(floor_date(x, ".1s"), as.POSIXct("2009-08-03 12:01:59.2", tz = "UTC"))
  expect_identical(floor_date(x, ".05s"), as.POSIXct("2009-08-03 12:01:59.2", tz = "UTC"))
  expect_identical(floor_date(x, ".01s"), as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC"))
  expect_identical(floor_date(x, ".005s"), as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC"))
  expect_identical(floor_date(x, ".5 mins"), as.POSIXct("2009-08-03 12:01:30", tz = "UTC"))
})

test_that("multi-unit rounding works the same for POSIX and Date objects", {
  px <- ymd("2009-08-01", tz = "UTC")
  dt <- ymd("2009-08-01")
  expect_identical(floor_date(px, "5 mins"), floor_date(dt, "5 mins"))
  expect_identical(floor_date(px, "5 mins"), floor_date(dt, "5 mins"))
  expect_identical(ceiling_date(px + 0.0001, "5 mins"), ceiling_date(dt, "5 mins"))
  expect_identical(
    ceiling_date(px, "5 mins", change_on_boundary = T),
    ceiling_date(dt, "5 mins", change_on_boundary = T)
  )
  expect_identical(ceiling_date(px + 0.001, "5 hours"), ceiling_date(dt, "5 hours"))
  expect_identical(ceiling_date(px + 0.0001, "5 hours", change_on_boundary = T), ceiling_date(dt, "5 hours", change_on_boundary = T))
  expect_identical(ceiling_date(px + 0.001, "2 hours"), ceiling_date(dt, "2 hours"))
  expect_identical(as_date(floor_date(px, "2 days")), floor_date(dt, "2 days"))
  expect_identical(as_date(ceiling_date(px + 0.001, "2 days")), ceiling_date(dt, "2 days"))
  expect_identical(as_date(floor_date(px, "5 days")), floor_date(dt, "5 days"))
  expect_identical(as_date(ceiling_date(px + 0.001, "5 days")), ceiling_date(dt, "5 days"))
  expect_identical(as_date(floor_date(px, "2 months")), floor_date(dt, "2 months"))
  expect_identical(as_date(ceiling_date(px, "2 months")), ceiling_date(dt, "2 months"))
  expect_identical(as_date(floor_date(px, "5 months")), floor_date(dt, "5 months"))
  expect_identical(as_date(ceiling_date(px, "5 months")), ceiling_date(dt, "5 months"))
})


test_that("ceiling_date works for multi-units", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  y <- as.POSIXct("2009-08-03 12:01:30.23", tz = "UTC")
  z <- as.POSIXct("2009-08-24 12:01:30.23", tz = "UTC")
  expect_identical(ceiling_date(x, "2 secs"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "3 secs"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "5 secs"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_identical(ceiling_date(y, "2 secs"), as.POSIXct("2009-08-03 12:01:32", tz = "UTC"))
  expect_identical(ceiling_date(y, "3 secs"), as.POSIXct("2009-08-03 12:01:33", tz = "UTC"))
  expect_identical(ceiling_date(y, "5 secs"), as.POSIXct("2009-08-03 12:01:35", tz = "UTC"))
  expect_identical(ceiling_date(x, "2 mins"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "3 mins"), as.POSIXct("2009-08-03 12:03:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "5 mins"), as.POSIXct("2009-08-03 12:05:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "2 hours"), as.POSIXct("2009-08-03 14:00:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "3 hours"), as.POSIXct("2009-08-03 15:00:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "2 days"), as.POSIXct("2009-08-05 00:00:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "3 days"), as.POSIXct("2009-08-04 00:00:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "10 days"), as.POSIXct("2009-08-11 00:00:00", tz = "UTC"))
  expect_identical(ceiling_date(z, "5 days"), as.POSIXct("2009-08-26 00:00:00", tz = "UTC"))
  expect_identical(ceiling_date(z, "10 days"), as.POSIXct("2009-08-31 00:00:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "2 month"), as.POSIXct("2009-09-01 00:00:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "2 bimonth"), ceiling_date(x, "4 months"))
  expect_identical(ceiling_date(x, "2 quarter"), ceiling_date(x, "6 months"))
  expect_identical(ceiling_date(x, "2 halfyear"), ceiling_date(x, "year"))
  expect_identical(ceiling_date(x, "2 year"), as.POSIXct("2010-01-01 00:00:00", tz = "UTC"))
})

test_that("ceiling_date works for fractional multi-units", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_identical(ceiling_date(x, ".2 secs"), as.POSIXct("2009-08-03 12:01:59.4", tz = "UTC"))
  expect_identical(ceiling_date(x, ".1s"), as.POSIXct("2009-08-03 12:01:59.3", tz = "UTC"))
  expect_identical(ceiling_date(x, ".05s"), as.POSIXct("2009-08-03 12:01:59.25", tz = "UTC"))
  expect_identical(ceiling_date(x, ".5 mins"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
})

test_that("round_date works for each time element", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_equal(round_date(x, "second"), as.POSIXct("2009-08-03 12:01:59", tz = "UTC"))
  expect_equal(round_date(x, "minute"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_equal(round_date(x, "hour"), as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_equal(round_date(x, "day"), as.POSIXct("2009-08-04 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "week"), as.POSIXct("2009-08-02 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "month"), as.POSIXct("2009-08-01 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "bimonth"), as.POSIXct("2009-09-01 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "quarter"), as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "halfyear"), as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "year"), as.POSIXct("2010-01-01 00:00:00", tz = "UTC"))
})

test_that("floor_date and ceiling_date work with leap years", {
  expect_equal(
    floor_date(ymd_hms(c("2016-02-29 1:2:3", "2016-03-01 10:20:30")), "year"),
    ymd_hms(c("2016-01-01 0:0:0", "2016-01-01 0:0:0"))
  )
  expect_equal(
    floor_date(ymd_hms(c("2016-02-29 1:2:3", "2016-03-01 10:20:30"), tz = "America/New_York"), "year"),
    ymd_hms(c("2016-01-01 0:0:0", "2016-01-01 0:0:0"), tz = "America/New_York")
  )
  expect_equal(
    ceiling_date(ymd_hms(c("2016-02-29 1:2:3", "2016-03-01 10:20:30")), "year"),
    ymd_hms(c("2017-01-01 0:0:0", "2017-01-01 0:0:0"))
  )
  expect_equal(
    ceiling_date(ymd_hms(c("2016-02-29 1:2:3", "2016-03-01 10:20:30"), tz = "America/New_York"), "year"),
    ymd_hms(c("2017-01-01 0:0:0", "2017-01-01 0:0:0"), tz = "America/New_York")
  )
})

test_that("round_date works for multi-units", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_equal(round_date(x, "2 second"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_equal(round_date(x, "2 minute"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_equal(round_date(x, "3 mins"), as.POSIXct("2009-08-03 12:03:00", tz = "UTC"))
  expect_equal(round_date(x, "5 mins"), as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_equal(round_date(x, "2 hour"), as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_equal(round_date(x, "5 hour"), as.POSIXct("2009-08-03 10:00:00", tz = "UTC"))
  expect_equal(round_date(x, "2 days"), as.POSIXct("2009-08-03 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "2 months"), as.POSIXct("2009-09-01 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "bimonth"), round_date(x, "2 months"))
  expect_equal(round_date(x, "bimonth"), round_date(x, "4 months"))
  expect_equal(round_date(x, "quarter"), round_date(x, "3 months"))
  expect_equal(round_date(x, "halfyear"), round_date(x, "6 months"))
  expect_equal(round_date(x, "3 years"), as.POSIXct("2010-01-01 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "4 years"), as.POSIXct("2008-01-01 00:00:00", tz = "UTC"))
})

test_that("round_date works for fractional multi-units", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_identical(round_date(x, ".2 secs"), as.POSIXct("2009-08-03 12:01:59.2", tz = "UTC"))
  expect_identical(round_date(x, ".1s"), as.POSIXct("2009-08-03 12:01:59.2", tz = "UTC"))
  expect_identical(round_date(x, ".05s"), as.POSIXct("2009-08-03 12:01:59.25", tz = "UTC"))
  expect_identical(round_date(x, ".5 mins"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
})

test_that("floor_date handles vectors", {
  x <- as.POSIXct(c(
    "2009-08-03 12:01:59.23",
    "2010-08-03 12:01:59.23"
  ), tz = "UTC")
  expect_identical(
    floor_date(x, "second"),
    as.POSIXct(c("2009-08-03 12:01:59", "2010-08-03 12:01:59"), tz = "UTC")
  )
  expect_identical(
    floor_date(x, "minute"),
    as.POSIXct(c("2009-08-03 12:01:00", "2010-08-03 12:01:00"), tz = "UTC")
  )
  expect_identical(
    floor_date(x, "hour"),
    as.POSIXct(c("2009-08-03 12:00:00", "2010-08-03 12:00:00"), tz = "UTC")
  )
  expect_identical(
    floor_date(x, "day"),
    as.POSIXct(c("2009-08-03 00:00:00", "2010-08-03 00:00:00"), tz = "UTC")
  )
  expect_identical(
    floor_date(x, "week"),
    as.POSIXct(c("2009-08-02 00:00:00", "2010-08-01 00:00:00"), tz = "UTC")
  )
  expect_identical(
    floor_date(x, "month"),
    as.POSIXct(c("2009-08-01 00:00:00", "2010-08-01 00:00:00"), tz = "UTC")
  )
  expect_identical(
    floor_date(x, "year"),
    as.POSIXct(c("2009-01-01 00:00:00", "2010-01-01 00:00:00"), tz = "UTC")
  )
})

test_that("ceiling_date handles vectors", {
  x <- as.POSIXct(c("2009-08-03 12:01:59.23", "2010-08-03 12:01:59.23"), tz = "UTC")
  expect_identical(
    ceiling_date(x, "second"),
    as.POSIXct(c("2009-08-03 12:02:00", "2010-08-03 12:02:00"), tz = "UTC")
  )
  expect_identical(
    ceiling_date(x, "minute"),
    as.POSIXct(c("2009-08-03 12:02:00", "2010-08-03 12:02:00"), tz = "UTC")
  )
  expect_identical(
    ceiling_date(x, "hour"),
    as.POSIXct(c("2009-08-03 13:00:00", "2010-08-03 13:00:00"), tz = "UTC")
  )
  expect_identical(
    ceiling_date(x, "day"),
    as.POSIXct(c("2009-08-04 00:00:00", "2010-08-04 00:00:00"), tz = "UTC")
  )
  expect_identical(
    ceiling_date(x, "week"),
    as.POSIXct(c("2009-08-09 00:00:00", "2010-08-08 00:00:00"), tz = "UTC")
  )
  expect_identical(
    ceiling_date(x, "month"),
    as.POSIXct(c("2009-09-01 00:00:00", "2010-09-01 00:00:00"), tz = "UTC")
  )
  expect_identical(
    ceiling_date(x, "year"),
    as.POSIXct(c("2010-01-01 00:00:00", "2011-01-01 00:00:00"), tz = "UTC")
  )
})


test_that("round_date handles vectors", {
  x <- as.POSIXct(c("2009-08-03 12:01:59.23", "2010-08-03 12:01:59.23"), tz = "UTC")
  expect_identical(
    round_date(x, "second"),
    as.POSIXct(c(
      "2009-08-03 12:01:59",
      "2010-08-03 12:01:59"
    ), tz = "UTC")
  )
  expect_identical(
    round_date(x, "minute"),
    as.POSIXct(c(
      "2009-08-03 12:02:00",
      "2010-08-03 12:02:00"
    ), tz = "UTC")
  )
  expect_identical(
    round_date(x, "hour"),
    as.POSIXct(c(
      "2009-08-03 12:00:00",
      "2010-08-03 12:00:00"
    ), tz = "UTC")
  )
  expect_identical(
    round_date(x, "day"),
    as.POSIXct(c(
      "2009-08-04 00:00:00",
      "2010-08-04 00:00:00"
    ), tz = "UTC")
  )
  expect_identical(
    round_date(x, "week"),
    as.POSIXct(c(
      "2009-08-02 00:00:00",
      "2010-08-01 00:00:00"
    ), tz = "UTC")
  )
  expect_identical(
    round_date(x, "month"),
    as.POSIXct(c(
      "2009-08-01 00:00:00",
      "2010-08-01 00:00:00"
    ), tz = "UTC")
  )
  expect_identical(
    round_date(x, "year"),
    as.POSIXct(c(
      "2010-01-01 00:00:00",
      "2011-01-01 00:00:00"
    ), tz = "UTC")
  )
})

test_that("floor_date works for a variety of formats", {
  x <- as.POSIXct("2009-08-03 12:01:59", tz = "UTC")

  expect_equal(
    floor_date(x, "minute"),
    as.POSIXct("2009-08-03 12:01:00", tz = "UTC")
  )
  expect_equal(
    floor_date(as.Date(x), "month"),
    as.Date("2009-08-01")
  )
  expect_equal(
    floor_date(as.Date(x), "bimonth"),
    ymd("2009-07-01")
  )
  expect_equal(
    floor_date(as.Date(x), "quarter"),
    ymd("2009-07-01")
  )
  expect_equal(
    floor_date(as.Date(x), "halfyear"),
    ymd("2009-07-01")
  )
  expect_equal(
    floor_date(as.POSIXlt(x), "minute"),
    as.POSIXlt(as.POSIXct("2009-08-03 12:01:00", tz = "UTC"))
  )
})

test_that("ceiling_date works for a variety of formats", {
  x <- as.POSIXct("2009-08-03 12:01:59", tz = "UTC")

  expect_equal(
    ceiling_date(x, "minute"),
    as.POSIXct("2009-08-03 12:02:00", tz = "UTC")
  )
  expect_equal(
    ceiling_date(as.Date(x), "month"),
    as.Date("2009-09-01")
  )
  expect_equal(
    ceiling_date(as.Date(x), "bimonth"),
    ymd("2009-09-01")
  )
  expect_equal(
    ceiling_date(as.Date(x), "quarter"),
    ymd("2009-10-01")
  )
  expect_equal(
    ceiling_date(as.Date(x), "halfyear"),
    ymd("2010-01-01")
  )
  expect_equal(
    ceiling_date(as.POSIXlt(x), "minute"),
    as.POSIXlt(as.POSIXct("2009-08-03 12:02:00",
      tz =
        "UTC"
    ))
  )
})

test_that("round_date works for a variety of formats", {
  x <- as.POSIXct("2009-08-03 12:01:59", tz = "UTC")
  expect_equal(round_date(x, "minute"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_equal(round_date(as.Date(x), "month"), as.Date("2009-08-01"))

  expect_equal(
    # Workaround for https://github.com/r-lib/testthat/issues/1710
    structure(round_date(as.POSIXlt(x), "minute"), balanced = NULL),
    structure(as.POSIXlt(as.POSIXct("2009-08-03 12:02:00", tz = "UTC")), balanced = NULL))

  ## https://github.com/tidyverse/lubridate/issues/1008
  y <- as.Date("2021-01-31")
  expect_equal(round_date(y, "month"), as.Date("2021-02-01"))
})


test_that("rounding works across DST", {
  ## https://github.com/tidyverse/lubridate/issues/399
  tt <- ymd("2016-03-27", tz = "Europe/Helsinki")
  expect_equal(ceiling_date(tt, "month"), as.POSIXct("2016-04-01", tz = "Europe/Helsinki"))
  expect_equal(ceiling_date(tt, "day"), as.POSIXct("2016-03-27", tz = "Europe/Helsinki"))
  tt <- ymd("2016-03-28", tz = "Europe/Helsinki")
  expect_equal(floor_date(tt, "month"), as.POSIXct("2016-03-01", tz = "Europe/Helsinki"))
  tt <- ymd_hms("2016-03-27 05:00:00", tz = "Europe/Helsinki")
  expect_equal(floor_date(tt, "day"), as.POSIXct("2016-03-27", tz = "Europe/Helsinki"))
  ## https://github.com/tidyverse/lubridate/issues/605
  x <- ymd_hms("2017-11-05 23:59:03", tz = "America/New_York")
  expect_equal(ceiling_date(x, "day"), as.POSIXct("2017-11-06", tz = "America/New_York"))
})

test_that("Ceiling for partials (Date) rounds up on boundary", {
  expect_identical(ceiling_date(as.Date("2012-09-27"), "day"), ymd("2012-09-28"))
  expect_identical(ceiling_date(as.Date("2012-09-01"), "day"), ymd("2012-09-02"))
  expect_identical(ceiling_date(as.Date("2012-09-01"), "2 days"), ymd("2012-09-03"))
})

test_that("Ceiling for Date returns date when unit level is higher than day", {
  expect_true(is.Date(ceiling_date(ymd("20160927"), "year")))
  expect_true(is.Date(ceiling_date(ymd("20160927"), "halfyear")))
  expect_true(is.Date(ceiling_date(ymd("20160927"), "quarter")))
  expect_true(is.Date(ceiling_date(ymd("20160927"), "bimonth")))
  expect_true(is.Date(ceiling_date(ymd("20160927"), "month")))
  expect_true(is.Date(ceiling_date(ymd("20160927"), "week")))
  expect_true(is.Date(ceiling_date(ymd("20160927"), "day")))
  expect_true(is.POSIXct(ceiling_date(ymd("20160927"), "hour")))
  expect_true(is.POSIXct(ceiling_date(ymd("20160927"), "minute")))
  expect_true(is.POSIXct(ceiling_date(ymd("20160927"), "second")))
})

test_that("Ceiling for POSIXct always returns POSIXct", {
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "year")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "halfyear")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "quarter")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "bimonth")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "month")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "week")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "day")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "hour")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "minute")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "second")))
})

test_that("ceiling_date does not round up dates that are already on a boundary", {
  expect_equal(ceiling_date(ymd_hms("2012-09-01 00:00:00"), "month"), as.POSIXct("2012-09-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "year"), as.POSIXct("2012-01-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "2 year"), as.POSIXct("2012-01-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "3 year"), as.POSIXct("2013-01-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "5 year"), as.POSIXct("2015-01-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 01:00:00"), "second"), ymd_hms("2012-01-01 01:00:00"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 01:00:00"), "2 second"), ymd_hms("2012-01-01 01:00:00"))
  expect_equal(
    ceiling_date(ymd_hms("2012-01-01 01:00:00"), "2 second", change_on_boundary = T),
    ymd_hms("2012-01-01 01:00:02")
  )
  expect_equal(ceiling_date(ymd_hms("2012-01-01 01:00:00"), "5 second"), ymd_hms("2012-01-01 01:00:00"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "bimonth"), ymd_hms("2012-01-01 00:00:00"))
})

test_that("ceiling_date does round up dates on a boundary with change_on_boundary=TRUE", {
  expect_equal(ceiling_date(as.Date("2012-09-27"), "day", TRUE), as.Date("2012-09-28"))
  expect_equal(ceiling_date(as.Date("2012-09-01"), "month", TRUE), as.Date("2012-10-01"))
  expect_equal(ceiling_date(ymd_hms("2012-09-01 00:00:00"), "month", TRUE), ymd("2012-10-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-09-01 00:00:00"), "bimonth", TRUE), ymd("2012-11-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "year", TRUE), as.POSIXct("2013-01-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 01:00:00"), "hour", TRUE), ymd_hms("2012-01-01 02:00:00"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "day", TRUE), ymd("2012-01-02", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 01:00:00"), "second", TRUE), ymd_hms("2012-01-01 01:00:01"))
})

test_that("floor_date does not round down dates that are already on a boundary", {
  expect_equal(floor_date(as.Date("2012-09-27"), "day"), as.Date("2012-09-27"))
})

test_that("round_date does not round dates that are already on a boundary", {
  expect_equal(round_date(as.Date("2012-09-27"), "day"), as.Date("2012-09-27"))
})

test_that("ceiling_date returns input of length zero when given input of length zero", {
  x <- structure(vector(mode = "numeric"), class = c("POSIXct", "POSIXt"))
  expect_equal(ceiling_date(x), x)
})

test_that("floor_date returns input of length zero when given input of length zero", {
  x <- structure(vector(mode = "numeric"), class = c("POSIXct", "POSIXt"))
  expect_equal(floor_date(x), x)
})

test_that("round_date returns input of length zero when given input of length zero", {
  x <- structure(vector(mode = "numeric"), class = c("POSIXct", "POSIXt"))
  expect_equal(round_date(x), x)
})

test_that("round_date behaves correctly on 60th second", {
  ## (bug #217)
  x <- ymd_hms("2013-12-01 23:59:59.9999")
  expect_equal(
    round_date(x, unit = "second"),
    ymd("2013-12-02", tz = "UTC")
  )
  second(x) <- 60
  expect_equal(x, ymd("2013-12-02", tz = "UTC"))
})

test_that("round_date and ceiling_date skip day time gap", {

  ##  (#240)
  tz <- "Europe/Amsterdam"
  times <- ymd_hms("2013-03-31 01:00:00 CET", "2013-03-31 01:15:00 CEST",
    "2013-03-31 01:30:00 CEST", "2013-03-31 01:45:00 CEST",
    "2013-03-31 03:00:00 CEST", "2013-03-31 03:15:00 CEST",
    tz = tz
  )

  round <- ymd_hms("2013-03-31 01:00:00 CET",
    "2013-03-31 01:00:00 CEST", "2013-03-31 03:00:00 CEST",
    "2013-03-31 03:00:00 CEST", "2013-03-31 03:00:00 CEST",
    "2013-03-31 03:00:00 CEST",
    tz = tz
  )
  expect_equal(round_date(times, "hour"), round)

  ceiling <- ymd_hms("2013-03-31 01:00:00 CET",
    "2013-03-31 03:00:00 CEST", "2013-03-31 03:00:00 CEST",
    "2013-03-31 03:00:00 CEST", "2013-03-31 03:00:00 CEST",
    "2013-03-31 04:00:00 CEST",
    tz = tz
  )
  expect_equal(ceiling_date(times, "hour"), ceiling)


  tz <- "America/Chicago"

  x <- ymd_hms(c(
    "2014-03-09 00:00:00", "2014-03-09 00:29:59", "2014-03-09 00:30:00",
    "2014-03-09 00:59:59", "2014-03-09 01:35:00", "2014-03-09 03:15:00",
    "2014-11-02 00:30:00", "2014-11-02 00:59:59", "2014-11-02 01:35:00",
    "2014-11-02 02:15:00", "2014-11-02 02:45:00"
  ),
  tz = tz
  )

  y <- as.POSIXct(c(
    "2014-03-09 00:00:00", "2014-03-09 00:00:00", "2014-03-09 01:00:00",
    "2014-03-09 01:00:00", "2014-03-09 03:00:00", "2014-03-09 03:00:00",
    "2014-11-02 01:00:00", "2014-11-02 01:00:00", "2014-11-02 02:00:00",
    "2014-11-02 02:00:00", "2014-11-02 03:00:00"
  ),
  tz = tz
  )

  ## ymd_hms("2014-11-02 01:35:00", tz = tz)
  ## force_tz(ymd_hms("2014-11-02 01:35:00"), tzone = "America/Chicago")
  ## round_date(ymd_hms("2014-11-02 01:35:00", tz = "America/Chicago"), "hour")
  expect_equal(round_date(x, "hour"), y)


})

test_that("rounding works within repeated DST", {
  ## "2014-11-02 00:00:00 CDT"
  ref <- ymd_hms("2014-11-02 00:00:00", tz = "America/Chicago")

  expect_equal(round_date(ref + dminutes(25), "hour"), ref)
  expect_equal(round_date(ref + dminutes(35), "hour"), ref + dhours(1))
  expect_equal(round_date(ref + dhours(1) + dminutes(25), "hour"), ref + dhours(1))

  expect_equal(round_date(ref + dhours(1) + dminutes(35), "hour"), ref + dhours(2))

  expect_equal(round_date(ref + dhours(2) + dminutes(20), "hour"), ref + dhours(2))
  expect_equal(round_date(ref + dhours(2) + dminutes(35), "hour"), ref + dhours(3))

  expect_equal(round_date(ref + duration("1H 35M 2S"), "minute"), ref + duration("1H 35M"))
  expect_equal(round_date(ref + duration("1H 35M 35S"), "minute"), ref + duration("1H 36M"))
  expect_equal(round_date(ref + duration("2H 35M 2S"), "minute"), ref + duration("2H 35M"))
})

test_that("ceiling_date, round_date and floor_date behave correctly with NA", {
  ## (bug #486)
  x <- ymd_hms("2009-08-03 12:01:59.23", tz = "UTC") + (0:1) * days()
  x[2] <- NA
  expect_equal(ceiling_date(x, unit = "day"), ymd(c("2009-08-04", NA), tz = "UTC"))
  expect_equal(ceiling_date(x, unit = "seconds"), ymd_hms(c("2009-08-03 12:02:00", NA), tz = "UTC"))
  expect_equal(ceiling_date(x, unit = "months"), ymd(c("2009-09-01", NA), tz = "UTC"))
  expect_equal(floor_date(x, unit = "day"), ymd(c("2009-08-03", NA), tz = "UTC"))
  expect_equal(floor_date(x, unit = "months"), ymd(c("2009-08-01", NA), tz = "UTC"))
  expect_equal(round_date(x, unit = "minute"), ymd_hms(c("2009-08-03 12:02:00", NA), tz = "UTC"))
})


test_that("floor_date works for seasons", {
  dts <- ymd_hms(sprintf("2017-%d-02 0:34:3", 1:12))
  expect_equal(month(floor_date(dts, "season")), c(12, 12, 3, 3, 3, 6, 6, 6, 9, 9, 9, 12))
  dts <- force_tz(dts, "America/New_York")
  expect_equal(month(floor_date(dts, "season")), c(12, 12, 3, 3, 3, 6, 6, 6, 9, 9, 9, 12))
})

test_that("ceiling_date works for seasons", {
  dts <- ymd_hms(sprintf("2017-%d-02 0:34:3", 1:12))
  expect_equal(month(ceiling_date(dts, "season")), c(3, 3, 6, 6, 6, 9, 9, 9, 12, 12, 12, 3))
  dts <- force_tz(dts, "America/New_York")
  expect_equal(month(ceiling_date(dts, "season")), c(3, 3, 6, 6, 6, 9, 9, 9, 12, 12, 12, 3))
})


test_that("round on week respects week_start", {
  date <- ymd("2017-05-07") ## sunday
  ct <- as.POSIXct("2010-02-03 13:45:59", tz = "America/New_York", format = "%Y-%m-%d %H:%M:%S") ## Wednesday

  expect_equal(wday(floor_date(ct, "week", week_start = 1)), 2)
  expect_equal(wday(floor_date(ct, "week", week_start = 2)), 3)
  expect_equal(wday(floor_date(ct, "week", week_start = 5)), 6)
  expect_equal(wday(floor_date(ct, "week", week_start = 7)), 1)
  expect_equal(wday(floor_date(date, "week", week_start = 1)), 2)
  expect_equal(wday(floor_date(date, "week", week_start = 2)), 3)
  expect_equal(wday(floor_date(date, "week", week_start = 5)), 6)
  expect_equal(wday(floor_date(date, "week", week_start = 7)), 1)

  expect_equal(wday(floor_date(ct, "week", week_start = 1), week_start = 1), 1)
  expect_equal(wday(floor_date(ct, "week", week_start = 2), week_start = 2), 1)
  expect_equal(wday(floor_date(ct, "week", week_start = 5), week_start = 5), 1)
  expect_equal(wday(floor_date(ct, "week", week_start = 7), week_start = 7), 1)
  expect_equal(wday(floor_date(date, "week", week_start = 1), week_start = 1), 1)
  expect_equal(wday(floor_date(date, "week", week_start = 2), week_start = 2), 1)
  expect_equal(wday(floor_date(date, "week", week_start = 5), week_start = 5), 1)
  expect_equal(wday(floor_date(date, "week", week_start = 7), week_start = 7), 1)

  expect_equal(wday(ceiling_date(ct, "week", week_start = 1)), 2)
  expect_equal(wday(ceiling_date(ct, "week", week_start = 2)), 3)
  expect_equal(wday(ceiling_date(ct, "week", week_start = 5)), 6)
  expect_equal(wday(ceiling_date(ct, "week", week_start = 7)), 1)
  expect_equal(wday(ceiling_date(date, "week", week_start = 1)), 2)
  expect_equal(wday(ceiling_date(date, "week", week_start = 2)), 3)
  expect_equal(wday(ceiling_date(date, "week", week_start = 5)), 6)
  expect_equal(wday(ceiling_date(date, "week", week_start = 7)), 1)

  expect_equal(wday(ceiling_date(ct, "week", week_start = 1), week_start = 1), 1)
  expect_equal(wday(ceiling_date(ct, "week", week_start = 2), week_start = 2), 1)
  expect_equal(wday(ceiling_date(ct, "week", week_start = 5), week_start = 5), 1)
  expect_equal(wday(ceiling_date(ct, "week", week_start = 7), week_start = 7), 1)
  expect_equal(wday(ceiling_date(date, "week", week_start = 1), week_start = 1), 1)
  expect_equal(wday(ceiling_date(date, "week", week_start = 2), week_start = 2), 1)
  expect_equal(wday(ceiling_date(date, "week", week_start = 5), week_start = 5), 1)
  expect_equal(wday(ceiling_date(date, "week", week_start = 7), week_start = 7), 1)

  expect_equal(wday(round_date(ct, "week", week_start = 1)), 2)
  expect_equal(wday(round_date(ct, "week", week_start = 2)), 3)
  expect_equal(wday(round_date(ct, "week", week_start = 5)), 6)
  expect_equal(wday(round_date(ct, "week", week_start = 7)), 1)
  expect_equal(wday(round_date(date, "week", week_start = 1)), 2)
  expect_equal(wday(round_date(date, "week", week_start = 2)), 3)
  expect_equal(wday(round_date(date, "week", week_start = 5)), 6)
  expect_equal(wday(round_date(date, "week", week_start = 7)), 1)

  expect_equal(wday(round_date(ct, "week", week_start = 1), week_start = 1), 1)
  expect_equal(wday(round_date(ct, "week", week_start = 2), week_start = 2), 1)
  expect_equal(wday(round_date(ct, "week", week_start = 5), week_start = 5), 1)
  expect_equal(wday(round_date(ct, "week", week_start = 7), week_start = 7), 1)
  expect_equal(wday(round_date(date, "week", week_start = 1), week_start = 1), 1)
  expect_equal(wday(round_date(date, "week", week_start = 2), week_start = 2), 1)
  expect_equal(wday(round_date(date, "week", week_start = 5), week_start = 5), 1)
  expect_equal(wday(round_date(date, "week", week_start = 7), week_start = 7), 1)
})

test_that("rounding works when week_start = name of week day", {
  date <- ymd("2017-05-07") ## sunday
  ct <- as.POSIXct("2010-02-03 13:45:59", tz = "America/New_York", format = "%Y-%m-%d %H:%M:%S") ## Wednesday

  expect_identical(floor_date(ct, "week", week_start = "Monday"), floor_date(ct, "week", week_start = 1))
  expect_identical(floor_date(ct, "week", week_start = "Tuesday"), floor_date(ct, "week", week_start = 2))

  expect_identical(ceiling_date(ct, "week", week_start = "Monday"), ceiling_date(ct, "week", week_start = 1))
  expect_identical(ceiling_date(ct, "week", week_start = "Friday"), ceiling_date(ct, "week", week_start = 5))

  expect_identical(ceiling_date(ct, "week", week_start = "Mond"), ceiling_date(ct, "week", week_start = 1))
  expect_identical(ceiling_date(ct, "week", week_start = "Fri"), ceiling_date(ct, "week", week_start = 5))

  expect_identical(round_date(ct, "week", week_start = "Monday"), round_date(ct, "week", week_start = 1))
  expect_identical(round_date(ct, "week", week_start = "Sunday"), round_date(ct, "week", week_start = 7))
  expect_identical(round_date(ct, "week", week_start = "Mon"), round_date(ct, "week", week_start = 1))
  expect_identical(round_date(ct, "week", week_start = "Sund"), round_date(ct, "week", week_start = 7))
})


test_that("wday extract and update: week_start can be a week day", {

  d <- ymd("2017-01-02") # Monday
  p <- as_datetime(d)

  expect_identical(wday(d, week_start = "Tue"), 7)
  expect_identical(wday(p, week_start = "Tue"), 7)
  expect_identical(wday(d, week_start = "Monday"), 1)
  expect_identical(wday(p, week_start = "Monday"), 1)
  expect_identical(wday(2, week_start = "Monday"), 1)
  expect_identical(wday(2, week_start = "Monday"), wday(2, week_start = 1))

  wday(d) <- "Tue"
  expect_identical(wday(d), 3)

  # ignore week_start when value is a character
  wday(d, week_start = 1) <- "Tue"
  expect_identical(wday(d), 3)
  wday(p, week_start = 2) <- "Sunday"
  expect_identical(wday(p), 1)

  wday(p) <- "Wednesday"
  expect_identical(wday(p), 4)

  wday(d) <- 4
  expect_identical(wday(d), 4)
  wday(d, week_start = 2) <- 4
  expect_identical(wday(d, week_start = 2), 4)
  wday(p) <- 5
  expect_identical(wday(p), 5)

})

test_that("rounding works for dates < 1970", {
  expect_equal(round_date(ymd_hms("1957-12-01 14:00:00 UTC"), "month"),
               ymd("1957-12-01", tz = "UTC"))
  expect_equal(round_date(ymd_hms("1958-01-31 09:59:59 UTC"), "month"),
               ymd("1958-02-01", tz = "UTC"))
})

test_that("rounding of fractional second works around origin", {
  ## #789
  seq <- seq(0, 500, 33)/1000

  time <- lubridate::origin - seq
  rnd <- round_date(time, ".033s")
  expect_true(all(unclass(rnd) > -.52))

  time <- lubridate::origin + seq
  rnd <- round_date(time, ".033s")
  expect_true(all(unclass(rnd) < .5))

  ## plot(time, rnd)
})


test_that("ceiling of dates works correctly before 1970", {
  #1107
  dates <- ymd('1969-01-01') + years(c(0:2))
  expect_equal(ceiling_date(dates, 'month', change_on_boundary = FALSE), dates)
  expect_equal(ceiling_date(dates, unit = 'month', change_on_boundary = TRUE), dates + months(1))
})

context("Rounding")

test_that("floor_date works for each time element", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_identical(floor_date(x, "second"), as.POSIXct("2009-08-03 12:01:59", tz = "UTC"))
  expect_identical(floor_date(x, "minute"), as.POSIXct("2009-08-03 12:01:00", tz = "UTC"))
  expect_identical(floor_date(x, "hour"), as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "day"), as.POSIXct("2009-08-03 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "week"), as.POSIXct("2009-08-02 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "month"), as.POSIXct("2009-08-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "bimonth"), as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "quarter"), as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "halfyear"), as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "year"), as.POSIXct("2009-01-01 00:00:00", tz = "UTC"))
})

test_that("floor_date and round_date throw on invalid multi-unit spec", {
  x <- ymd_hms("2009-08-03 12:01:57.11")

  expect_error(floor_date(x, "120 sec"))
  expect_error(floor_date(x, "120 min"))
  expect_error(floor_date(x, "25 h"))
  expect_error(floor_date(x, "32 days"))
})

test_that("floor_date works for multi-units", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_identical(floor_date(x, "2 secs"),   as.POSIXct("2009-08-03 12:01:58", tz = "UTC"))
  expect_identical(floor_date(x, "2s"),   as.POSIXct("2009-08-03 12:01:58", tz = "UTC"))
  expect_identical(floor_date(x, "2 mins"),   as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2M"),   as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2 hours"),     as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2h"),     as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2 days"),      as.POSIXct("2009-08-03 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "3 days"),      as.POSIXct("2009-08-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "10 days"),      as.POSIXct("2009-08-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "10d"),      as.POSIXct("2009-08-01 00:00:00", tz = "UTC"))
  ## expect_identical(floor_date(x, "2 week"), as.POSIXct("2009-08-xx 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2 month"),    as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2m"),    as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_identical(floor_date(x, "2 bimonth"),  floor_date(x, "4 months"))
  expect_identical(floor_date(x, "2 quarter"),  floor_date(x, "6 months"))
  expect_identical(floor_date(x, "2 halfyear"), floor_date(x, "year"))
  expect_identical(floor_date(x, "2 year"),     as.POSIXct("2008-01-01 00:00:00", tz = "UTC"))
})

test_that("floor_date works for fractional multi-units", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_identical(floor_date(x, ".2 secs"),   as.POSIXct("2009-08-03 12:01:59.2", tz = "UTC"))
  expect_identical(floor_date(x, ".1s"),   as.POSIXct("2009-08-03 12:01:59.2", tz = "UTC"))
  expect_identical(floor_date(x, ".05s"),   as.POSIXct("2009-08-03 12:01:59.2", tz = "UTC"))
  expect_identical(floor_date(x, ".01s"),   as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC"))
  expect_identical(floor_date(x, ".005s"),   as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC"))
  expect_identical(floor_date(x, ".5 mins"),   as.POSIXct("2009-08-03 12:01:30", tz = "UTC"))
})

test_that("multi-unit rounding works the same for POSIX and Date objects", {
  px <- ymd("2009-08-01", tz = "UTC")
  dt <- ymd("2009-08-01")
  expect_identical(floor_date(px, "5 mins"), floor_date(dt, "5 mins"))
  expect_identical(floor_date(px, "5 mins"), floor_date(dt, "5 mins"))
  expect_identical(ceiling_date(px + 0.0001, "5 mins"), ceiling_date(dt, "5 mins"))
  expect_identical(ceiling_date(px, "5 mins", change_on_boundary = T),
                   ceiling_date(dt, "5 mins", change_on_boundary = T))
  expect_identical(ceiling_date(px + 0.001, "5 hours"), ceiling_date(dt, "5 hours"))
  expect_identical(ceiling_date(px + 0.0001, "5 hours", change_on_boundary = T), ceiling_date(dt, "5 hours", change_on_boundary = T))
  expect_identical(ceiling_date(px + 0.001, "2 hours"), ceiling_date(dt, "2 hours"))
  expect_identical(as_date(floor_date(px, "2 days")), floor_date(dt, "2 days"))
  expect_identical(as_date(ceiling_date(px + 0.001, "2 days")), ceiling_date(dt, "2 days"))
  expect_identical(as_date(floor_date(px, "5 days")), floor_date(dt, "5 days"))
  expect_identical(as_date(ceiling_date(px + 0.001, "5 days")), ceiling_date(dt, "5 days"))
  expect_identical(as_date(floor_date(px, "2 months")), floor_date(dt, "2 months"))
  expect_identical(as_date(ceiling_date(px, "2 months")), ceiling_date(dt, "2 months"))
  expect_identical(as_date(floor_date(px, "5 months")), floor_date(dt, "5 months"))
  expect_identical(as_date(ceiling_date(px, "5 months")), ceiling_date(dt, "5 months"))
})


test_that("ceiling_date works for multi-units", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  y <- as.POSIXct("2009-08-03 12:01:30.23", tz = "UTC")
  z <- as.POSIXct("2009-08-24 12:01:30.23", tz = "UTC")
  expect_identical(ceiling_date(x, "2 secs"),   as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "3 secs"),   as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "5 secs"),   as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_identical(ceiling_date(y, "2 secs"),   as.POSIXct("2009-08-03 12:01:32", tz = "UTC"))
  expect_identical(ceiling_date(y, "3 secs"),   as.POSIXct("2009-08-03 12:01:33", tz = "UTC"))
  expect_identical(ceiling_date(y, "5 secs"),   as.POSIXct("2009-08-03 12:01:35", tz = "UTC"))
  expect_identical(ceiling_date(x, "2 mins"),   as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "3 mins"),   as.POSIXct("2009-08-03 12:03:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "5 mins"),   as.POSIXct("2009-08-03 12:05:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "2 hours"),     as.POSIXct("2009-08-03 14:00:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "3 hours"),     as.POSIXct("2009-08-03 15:00:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "2 days"),      as.POSIXct("2009-08-05 00:00:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "3 days"),      as.POSIXct("2009-08-04 00:00:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "10 days"),      as.POSIXct("2009-08-11 00:00:00", tz = "UTC"))
  expect_identical(ceiling_date(z, "5 days"),      as.POSIXct("2009-08-26 00:00:00", tz = "UTC"))
  expect_identical(ceiling_date(z, "10 days"),      as.POSIXct("2009-08-31 00:00:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "2 month"),    as.POSIXct("2009-09-01 00:00:00", tz = "UTC"))
  expect_identical(ceiling_date(x, "2 bimonth"),  ceiling_date(x, "4 months"))
  expect_identical(ceiling_date(x, "2 quarter"),  ceiling_date(x, "6 months"))
  expect_identical(ceiling_date(x, "2 halfyear"), ceiling_date(x, "year"))
  expect_identical(ceiling_date(x, "2 year"),     as.POSIXct("2010-01-01 00:00:00", tz = "UTC"))
})

test_that("ceiling_date works for fractional multi-units", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_identical(ceiling_date(x, ".2 secs"),   as.POSIXct("2009-08-03 12:01:59.4", tz = "UTC"))
  expect_identical(ceiling_date(x, ".1s"),   as.POSIXct("2009-08-03 12:01:59.3", tz = "UTC"))
  expect_identical(ceiling_date(x, ".05s"),   as.POSIXct("2009-08-03 12:01:59.25", tz = "UTC"))
  expect_identical(ceiling_date(x, ".5 mins"),   as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
})

test_that("round_date works for each time element", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_equal(round_date(x, "second"), as.POSIXct("2009-08-03 12:01:59", tz = "UTC"))
  expect_equal(round_date(x, "minute"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_equal(round_date(x, "hour"), as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_equal(round_date(x, "day"), as.POSIXct("2009-08-04 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "week"), as.POSIXct("2009-08-02 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "month"), as.POSIXct("2009-08-01 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "bimonth"), as.POSIXct("2009-09-01 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "quarter"), as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "halfyear"), as.POSIXct("2009-07-01 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "year"), as.POSIXct("2010-01-01 00:00:00", tz = "UTC"))
})

test_that("floor_date and ceiling_date work with leap years", {
  expect_equal(floor_date(ymd_hms(c("2016-02-29 1:2:3", "2016-03-01 10:20:30")), "year"),
               ymd_hms(c("2016-01-01 0:0:0", "2016-01-01 0:0:0")))
  expect_equal(floor_date(ymd_hms(c("2016-02-29 1:2:3", "2016-03-01 10:20:30"), tz = "America/New_York"), "year"),
               ymd_hms(c("2016-01-01 0:0:0", "2016-01-01 0:0:0"), tz = "America/New_York"))
  expect_equal(ceiling_date(ymd_hms(c("2016-02-29 1:2:3", "2016-03-01 10:20:30")), "year"),
               ymd_hms(c("2017-01-01 0:0:0", "2017-01-01 0:0:0")))
  expect_equal(ceiling_date(ymd_hms(c("2016-02-29 1:2:3", "2016-03-01 10:20:30"), tz = "America/New_York"), "year"),
               ymd_hms(c("2017-01-01 0:0:0", "2017-01-01 0:0:0"), tz = "America/New_York"))
})

test_that("round_date works for multi-units", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_equal(round_date(x, "2 second"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_equal(round_date(x, "2 minute"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_equal(round_date(x, "3 mins"), as.POSIXct("2009-08-03 12:03:00", tz = "UTC"))
  expect_equal(round_date(x, "5 mins"), as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_equal(round_date(x, "2 hour"), as.POSIXct("2009-08-03 12:00:00", tz = "UTC"))
  expect_equal(round_date(x, "5 hour"), as.POSIXct("2009-08-03 10:00:00", tz = "UTC"))
  expect_equal(round_date(x, "2 days"), as.POSIXct("2009-08-03 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "2 months"), as.POSIXct("2009-09-01 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "bimonth"), round_date(x, "2 months"))
  expect_equal(round_date(x, "bimonth"), round_date(x, "4 months"))
  expect_equal(round_date(x, "quarter"), round_date(x, "3 months"))
  expect_equal(round_date(x, "halfyear"), round_date(x, "6 months"))
  expect_equal(round_date(x, "3 years"), as.POSIXct("2010-01-01 00:00:00", tz = "UTC"))
  expect_equal(round_date(x, "4 years"), as.POSIXct("2008-01-01 00:00:00", tz = "UTC"))
})

test_that("round_date works for fractional multi-units", {
  x <- as.POSIXct("2009-08-03 12:01:59.23", tz = "UTC")
  expect_identical(round_date(x, ".2 secs"),   as.POSIXct("2009-08-03 12:01:59.2", tz = "UTC"))
  expect_identical(round_date(x, ".1s"),   as.POSIXct("2009-08-03 12:01:59.2", tz = "UTC"))
  expect_identical(round_date(x, ".05s"),   as.POSIXct("2009-08-03 12:01:59.25", tz = "UTC"))
  expect_identical(round_date(x, ".5 mins"),   as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
})

test_that("floor_date handles vectors", {
  x <- as.POSIXct(c("2009-08-03 12:01:59.23",
        "2010-08-03 12:01:59.23"), tz = "UTC")
  expect_identical(floor_date(x, "second"),
                   as.POSIXct(c("2009-08-03 12:01:59", "2010-08-03 12:01:59"), tz = "UTC"))
  expect_identical(floor_date(x, "minute"),
                   as.POSIXct(c("2009-08-03 12:01:00", "2010-08-03 12:01:00"), tz = "UTC"))
  expect_identical(floor_date(x, "hour"),
                   as.POSIXct(c("2009-08-03 12:00:00", "2010-08-03 12:00:00"), tz = "UTC"))
  expect_identical(floor_date(x, "day"),
                   as.POSIXct(c("2009-08-03 00:00:00", "2010-08-03 00:00:00"), tz = "UTC"))
  expect_identical(floor_date(x, "week"),
                   as.POSIXct(c("2009-08-02 00:00:00", "2010-08-01 00:00:00"), tz = "UTC"))
  expect_identical(floor_date(x, "month"),
                   as.POSIXct(c("2009-08-01 00:00:00", "2010-08-01 00:00:00"), tz = "UTC"))
  expect_identical(floor_date(x, "year"),
                   as.POSIXct(c("2009-01-01 00:00:00", "2010-01-01 00:00:00"), tz = "UTC"))
})

test_that("ceiling_date handles vectors", {
  x <- as.POSIXct(c("2009-08-03 12:01:59.23", "2010-08-03 12:01:59.23"), tz = "UTC")
  expect_identical(ceiling_date(x, "second"),
                   as.POSIXct(c("2009-08-03 12:02:00", "2010-08-03 12:02:00"), tz = "UTC"))
  expect_identical(ceiling_date(x, "minute"),
                   as.POSIXct(c("2009-08-03 12:02:00", "2010-08-03 12:02:00"), tz = "UTC"))
  expect_identical(ceiling_date(x, "hour"),
                   as.POSIXct(c("2009-08-03 13:00:00", "2010-08-03 13:00:00"), tz = "UTC"))
  expect_identical(ceiling_date(x, "day"),
                   as.POSIXct(c("2009-08-04 00:00:00", "2010-08-04 00:00:00"), tz = "UTC"))
  expect_identical(ceiling_date(x, "week"),
                   as.POSIXct(c("2009-08-09 00:00:00", "2010-08-08 00:00:00"), tz = "UTC"))
  expect_identical(ceiling_date(x, "month"),
                   as.POSIXct(c("2009-09-01 00:00:00", "2010-09-01 00:00:00"), tz = "UTC"))
  expect_identical(ceiling_date(x, "year"),
                   as.POSIXct(c("2010-01-01 00:00:00", "2011-01-01 00:00:00"), tz = "UTC"))
})


test_that("round_date handles vectors", {
  x <- as.POSIXct(c("2009-08-03 12:01:59.23", "2010-08-03 12:01:59.23"), tz = "UTC")
  expect_identical(round_date(x, "second"),
    as.POSIXct(c("2009-08-03 12:01:59",
      "2010-08-03 12:01:59"), tz = "UTC"))
  expect_identical(round_date(x, "minute"),
    as.POSIXct(c("2009-08-03 12:02:00",
      "2010-08-03 12:02:00"), tz = "UTC"))
  expect_identical(round_date(x, "hour"),
    as.POSIXct(c("2009-08-03 12:00:00",
      "2010-08-03 12:00:00"), tz = "UTC"))
  expect_identical(round_date(x, "day"),
    as.POSIXct(c("2009-08-04 00:00:00",
      "2010-08-04 00:00:00"), tz = "UTC"))
  expect_identical(round_date(x, "week"),
    as.POSIXct(c("2009-08-02 00:00:00",
      "2010-08-01 00:00:00"), tz = "UTC"))
  expect_identical(round_date(x, "month"),
    as.POSIXct(c("2009-08-01 00:00:00",
      "2010-08-01 00:00:00"), tz = "UTC"))
  expect_identical(round_date(x, "year"),
    as.POSIXct(c("2010-01-01 00:00:00",
      "2011-01-01 00:00:00"), tz = "UTC"))
})

test_that("floor_date works for a variety of formats", {
  x <- as.POSIXct("2009-08-03 12:01:59", tz = "UTC")

  expect_equal(floor_date(x, "minute"),
    as.POSIXct("2009-08-03 12:01:00", tz = "UTC"))
  expect_equal(floor_date(as.Date(x), "month"),
    as.Date("2009-08-01"))
  expect_equal(floor_date(as.Date(x), "bimonth"),
    ymd("2009-07-01"))
  expect_equal(floor_date(as.Date(x), "quarter"),
    ymd("2009-07-01"))
  expect_equal(floor_date(as.Date(x), "halfyear"),
    ymd("2009-07-01"))
  expect_equal(floor_date(as.POSIXlt(x), "minute"),
    as.POSIXlt(as.POSIXct("2009-08-03 12:01:00", tz = "UTC")))
})

test_that("ceiling_date works for a variety of formats", {
  x <- as.POSIXct("2009-08-03 12:01:59", tz = "UTC")

  expect_equal(ceiling_date(x, "minute"),
    as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_equal(ceiling_date(as.Date(x), "month"),
    as.Date("2009-09-01"))
  expect_equal(ceiling_date(as.Date(x), "bimonth"),
    ymd("2009-09-01"))
  expect_equal(ceiling_date(as.Date(x), "quarter"),
    ymd("2009-10-01"))
  expect_equal(ceiling_date(as.Date(x), "halfyear"),
    ymd("2010-01-01"))
  expect_equal(ceiling_date(as.POSIXlt(x), "minute"),
    as.POSIXlt(as.POSIXct("2009-08-03 12:02:00", tz =
    "UTC")))
})

test_that("round_date works for a variety of formats", {
  x <- as.POSIXct("2009-08-03 12:01:59", tz = "UTC")
  expect_equal(round_date(x, "minute"), as.POSIXct("2009-08-03 12:02:00", tz = "UTC"))
  expect_equal(round_date(as.Date(x), "month"), as.Date("2009-08-01"))
  expect_equal(round_date(as.POSIXlt(x), "minute"), as.POSIXlt(as.POSIXct("2009-08-03 12:02:00", tz = "UTC")))
})


test_that("rounding works across DST", {
  ## https://github.com/tidyverse/lubridate/issues/399
  tt <- ymd("2016-03-27", tz = "Europe/Helsinki");
  expect_equal(ceiling_date(tt, "month"), as.POSIXct("2016-04-01", tz = "Europe/Helsinki"))
  expect_equal(ceiling_date(tt, "day"), as.POSIXct("2016-03-27", tz = "Europe/Helsinki"))
  tt <- ymd("2016-03-28", tz = "Europe/Helsinki");
  expect_equal(floor_date(tt, "month"), as.POSIXct("2016-03-01", tz = "Europe/Helsinki"))
  tt <- ymd_hms("2016-03-27 05:00:00", tz = "Europe/Helsinki");
  expect_equal(floor_date(tt, "day"), as.POSIXct("2016-03-27", tz = "Europe/Helsinki"))
  ## https://github.com/tidyverse/lubridate/issues/605
  x <- ymd_hms("2017-11-05 23:59:03", tz = 'America/New_York')
  expect_equal(ceiling_date(x, "day"), as.POSIXct("2017-11-06", tz = "America/New_York"))
})

test_that("Ceiling for partials (Date) rounds up on boundary", {
  expect_identical(ceiling_date(as.Date("2012-09-27"), "day"), ymd("2012-09-28"))
  expect_identical(ceiling_date(as.Date("2012-09-01"), "day"), ymd("2012-09-02"))
  expect_identical(ceiling_date(as.Date("2012-09-01"), "2 days"), ymd("2012-09-03"))
})

test_that("Ceiling for Date returns date when unit level is higher than day", {
  expect_true(is.Date(ceiling_date(ymd("20160927"), "year")))
  expect_true(is.Date(ceiling_date(ymd("20160927"), "halfyear")))
  expect_true(is.Date(ceiling_date(ymd("20160927"), "quarter")))
  expect_true(is.Date(ceiling_date(ymd("20160927"), "bimonth")))
  expect_true(is.Date(ceiling_date(ymd("20160927"), "month")))
  expect_true(is.Date(ceiling_date(ymd("20160927"), "week")))
  expect_true(is.Date(ceiling_date(ymd("20160927"), "day")))
  expect_true(is.POSIXct(ceiling_date(ymd("20160927"), "hour")))
  expect_true(is.POSIXct(ceiling_date(ymd("20160927"), "minute")))
  expect_true(is.POSIXct(ceiling_date(ymd("20160927"), "second")))
})

test_that("Ceiling for POSIXct always returns POSIXct", {
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "year")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "halfyear")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "quarter")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "bimonth")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "month")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "week")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "day")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "hour")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "minute")))
  expect_true(is.POSIXct(ceiling_date(ymd_hms("20160927 00:00:00"), "second")))
})

test_that("ceiling_date does not round up dates that are already on a boundary", {
  expect_equal(ceiling_date(ymd_hms("2012-09-01 00:00:00"), "month"), as.POSIXct("2012-09-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "year"), as.POSIXct("2012-01-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "2 year"), as.POSIXct("2012-01-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "3 year"), as.POSIXct("2013-01-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "5 year"), as.POSIXct("2015-01-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 01:00:00"), "second"), ymd_hms("2012-01-01 01:00:00"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 01:00:00"), "2 second"), ymd_hms("2012-01-01 01:00:00"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 01:00:00"), "2 second", change_on_boundary = T),
               ymd_hms("2012-01-01 01:00:02"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 01:00:00"), "5 second"), ymd_hms("2012-01-01 01:00:00"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "bimonth"), ymd_hms("2012-01-01 00:00:00"))
})

test_that("ceiling_date does round up dates on a boundary with change_on_boundary=TRUE", {
  expect_equal(ceiling_date(as.Date("2012-09-27"), "day", TRUE), as.Date("2012-09-28"))
  expect_equal(ceiling_date(as.Date("2012-09-01"), "month", TRUE), as.Date("2012-10-01"))
  expect_equal(ceiling_date(ymd_hms("2012-09-01 00:00:00"), "month", TRUE), ymd("2012-10-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-09-01 00:00:00"), "bimonth", TRUE), ymd("2012-11-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "year", TRUE), as.POSIXct("2013-01-01", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 01:00:00"), "hour", TRUE), ymd_hms("2012-01-01 02:00:00"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 00:00:00"), "day", TRUE), ymd("2012-01-02", tz = "UTC"))
  expect_equal(ceiling_date(ymd_hms("2012-01-01 01:00:00"), "second", TRUE), ymd_hms("2012-01-01 01:00:01"))
})

test_that("floor_date does not round down dates that are already on a boundary", {
  expect_equal(floor_date(as.Date("2012-09-27"), "day"), as.Date("2012-09-27"))
})

test_that("round_date does not round dates that are already on a boundary", {
  expect_equal(round_date(as.Date("2012-09-27"), "day"), as.Date("2012-09-27"))
})

test_that("ceiling_date returns input of length zero when given input of length zero", {
  x <- structure(vector(mode = "numeric"), class = c("POSIXct", "POSIXt"))
  expect_equal(ceiling_date(x), x)
})

test_that("floor_date returns input of length zero when given input of length zero", {
  x <- structure(vector(mode = "numeric"), class = c("POSIXct", "POSIXt"))
  expect_equal(floor_date(x), x)
})

test_that("round_date returns input of length zero when given input of length zero", {
  x <- structure(vector(mode = "numeric"), class = c("POSIXct", "POSIXt"))
  expect_equal(round_date(x), x)
})

test_that("round_date behaves correctly on 60th second", {
  ## (bug #217)
  x <- ymd_hms("2013-12-01 23:59:59.9999")
  expect_equal(round_date(x, unit = "second"),
               ymd("2013-12-02", tz = "UTC"))
  second(x) <-  60
  expect_equal(x, ymd("2013-12-02", tz = "UTC"))
})

test_that("round_date and ceiling_date skip day time gap", {
  ##  (#240)
  tz <- "Europe/Amsterdam"
  times <- ymd_hms("2013-03-31 01:00:00 CET", "2013-03-31 01:15:00 CEST",
                   "2013-03-31 01:30:00 CEST", "2013-03-31 01:45:00 CEST",
                   "2013-03-31 03:00:00 CEST", "2013-03-31 03:15:00 CEST",
                   tz = tz)

  round <- ymd_hms("2013-03-31 01:00:00 CET",
                   "2013-03-31 01:00:00 CEST", "2013-03-31 03:00:00 CEST",
                   "2013-03-31 03:00:00 CEST", "2013-03-31 03:00:00 CEST",
                   "2013-03-31 03:00:00 CEST",
                   tz = tz)
  expect_equal(round_date(times, "hour"), round)

  ceiling <- ymd_hms("2013-03-31 01:00:00 CET",
                     "2013-03-31 03:00:00 CEST", "2013-03-31 03:00:00 CEST",
                     "2013-03-31 03:00:00 CEST", "2013-03-31 03:00:00 CEST",
                     "2013-03-31 04:00:00 CEST",
                     tz = tz)
  expect_equal(ceiling_date(times, "hour"), ceiling)


  tz <- "America/Chicago"

  x <- ymd_hms(c("2014-03-09 00:00:00", "2014-03-09 00:29:59", "2014-03-09 00:30:00",
                 "2014-03-09 00:59:59", "2014-03-09 01:35:00", "2014-03-09 03:15:00",
                 "2014-11-02 00:30:00", "2014-11-02 00:59:59", "2014-11-02 01:35:00",
                 "2014-11-02 02:15:00", "2014-11-02 02:45:00"),
               tz = tz)

  y <- as.POSIXct(c("2014-03-09 00:00:00", "2014-03-09 00:00:00", "2014-03-09 01:00:00",
                    "2014-03-09 01:00:00", "2014-03-09 03:00:00", "2014-03-09 03:00:00",
                    "2014-11-02 01:00:00", "2014-11-02 01:00:00", "2014-11-02 02:00:00",
                    "2014-11-02 02:00:00", "2014-11-02 03:00:00"),
                  tz = tz)

  expect_equal(round_date(x, "hour"),  y)
})

test_that("ceiling_date, round_date and floor_date behave correctly with NA", {
  ## (bug #486)
  x <- ymd_hms("2009-08-03 12:01:59.23", tz = "UTC") + (0:1)*days()
  x[2] <- NA
  expect_equal(ceiling_date(x, unit = "day"), ymd(c("2009-08-04", NA), tz = "UTC"))
  expect_equal(ceiling_date(x, unit = "seconds"), ymd_hms(c("2009-08-03 12:02:00", NA), tz = "UTC"))
  expect_equal(ceiling_date(x, unit = "months"), ymd(c("2009-09-01", NA), tz = "UTC"))
  expect_equal(floor_date(x, unit = "day"), ymd(c("2009-08-03", NA), tz = "UTC"))
  expect_equal(floor_date(x, unit = "months"), ymd(c("2009-08-01", NA), tz = "UTC"))
  expect_equal(round_date(x, unit = "minute"), ymd_hms(c("2009-08-03 12:02:00", NA), tz = "UTC"))
})


test_that("floor_date works for seasons", {
  dts <- ymd_hms(sprintf("2017-%d-02 0:34:3", 1:12))
  expect_equal(month(floor_date(dts, "season")), c(12, 12, 3, 3, 3, 6, 6, 6, 9, 9, 9, 12))
  dts <- force_tz(dts, "America/New_York")
  expect_equal(month(floor_date(dts, "season")), c(12, 12, 3, 3, 3, 6, 6, 6, 9, 9, 9, 12))
})

test_that("ceiling_date works for seasons", {
  dts <- ymd_hms(sprintf("2017-%d-02 0:34:3", 1:12))
  expect_equal(month(ceiling_date(dts, "season")), c(3, 3, 6, 6, 6, 9, 9, 9, 12, 12, 12, 3))
  dts <- force_tz(dts, "America/New_York")
  expect_equal(month(ceiling_date(dts, "season")), c(3, 3, 6, 6, 6, 9, 9, 9, 12, 12, 12, 3))
})


test_that("round on week respects week_start", {
  date <- ymd("2017-05-07") ## sunday
  ct <- as.POSIXct("2010-02-03 13:45:59", tz = "America/New_York", format = "%Y-%m-%d %H:%M:%S") ## Wednesday

  expect_equal(wday(floor_date(ct, "week", week_start = 1)), 2)
  expect_equal(wday(floor_date(ct, "week", week_start = 2)), 3)
  expect_equal(wday(floor_date(ct, "week", week_start = 5)), 6)
  expect_equal(wday(floor_date(ct, "week", week_start = 7)), 1)
  expect_equal(wday(floor_date(date, "week", week_start = 1)), 2)
  expect_equal(wday(floor_date(date, "week", week_start = 2)), 3)
  expect_equal(wday(floor_date(date, "week", week_start = 5)), 6)
  expect_equal(wday(floor_date(date, "week", week_start = 7)), 1)

  expect_equal(wday(floor_date(ct, "week", week_start = 1), week_start = 1), 1)
  expect_equal(wday(floor_date(ct, "week", week_start = 2), week_start = 2), 1)
  expect_equal(wday(floor_date(ct, "week", week_start = 5), week_start = 5), 1)
  expect_equal(wday(floor_date(ct, "week", week_start = 7), week_start = 7), 1)
  expect_equal(wday(floor_date(date, "week", week_start = 1), week_start = 1), 1)
  expect_equal(wday(floor_date(date, "week", week_start = 2), week_start = 2), 1)
  expect_equal(wday(floor_date(date, "week", week_start = 5), week_start = 5), 1)
  expect_equal(wday(floor_date(date, "week", week_start = 7), week_start = 7), 1)

  expect_equal(wday(ceiling_date(ct, "week", week_start = 1)), 2)
  expect_equal(wday(ceiling_date(ct, "week", week_start = 2)), 3)
  expect_equal(wday(ceiling_date(ct, "week", week_start = 5)), 6)
  expect_equal(wday(ceiling_date(ct, "week", week_start = 7)), 1)
  expect_equal(wday(ceiling_date(date, "week", week_start = 1)), 2)
  expect_equal(wday(ceiling_date(date, "week", week_start = 2)), 3)
  expect_equal(wday(ceiling_date(date, "week", week_start = 5)), 6)
  expect_equal(wday(ceiling_date(date, "week", week_start = 7)), 1)

  expect_equal(wday(ceiling_date(ct, "week", week_start = 1), week_start = 1), 1)
  expect_equal(wday(ceiling_date(ct, "week", week_start = 2), week_start = 2), 1)
  expect_equal(wday(ceiling_date(ct, "week", week_start = 5), week_start = 5), 1)
  expect_equal(wday(ceiling_date(ct, "week", week_start = 7), week_start = 7), 1)
  expect_equal(wday(ceiling_date(date, "week", week_start = 1), week_start = 1), 1)
  expect_equal(wday(ceiling_date(date, "week", week_start = 2), week_start = 2), 1)
  expect_equal(wday(ceiling_date(date, "week", week_start = 5), week_start = 5), 1)
  expect_equal(wday(ceiling_date(date, "week", week_start = 7), week_start = 7), 1)

  expect_equal(wday(round_date(ct, "week", week_start = 1)), 2)
  expect_equal(wday(round_date(ct, "week", week_start = 2)), 3)
  expect_equal(wday(round_date(ct, "week", week_start = 5)), 6)
  expect_equal(wday(round_date(ct, "week", week_start = 7)), 1)
  expect_equal(wday(round_date(date, "week", week_start = 1)), 2)
  expect_equal(wday(round_date(date, "week", week_start = 2)), 3)
  expect_equal(wday(round_date(date, "week", week_start = 5)), 6)
  expect_equal(wday(round_date(date, "week", week_start = 7)), 1)

  expect_equal(wday(round_date(ct, "week", week_start = 1), week_start = 1), 1)
  expect_equal(wday(round_date(ct, "week", week_start = 2), week_start = 2), 1)
  expect_equal(wday(round_date(ct, "week", week_start = 5), week_start = 5), 1)
  expect_equal(wday(round_date(ct, "week", week_start = 7), week_start = 7), 1)
  expect_equal(wday(round_date(date, "week", week_start = 1), week_start = 1), 1)
  expect_equal(wday(round_date(date, "week", week_start = 2), week_start = 2), 1)
  expect_equal(wday(round_date(date, "week", week_start = 5), week_start = 5), 1)
  expect_equal(wday(round_date(date, "week", week_start = 7), week_start = 7), 1)

})

#' @include timespans.r
#' @include difftimes.r

check_duration <- function(object) {
  if (is.numeric(object@.Data)) {
    TRUE
  } else {
    "Duration value is not a number. Should be numeric."
  }
}


#' Duration class
#'
#' Duration is an S4 class that extends the [Timespan-class] class.
#' Durations record the exact number of seconds in a time span. They measure the
#' exact passage of time but do not always align with measurements
#' made in larger units of time such as hours, months and years.
#' This is because the exact length of larger time units can be affected
#' by conventions such as leap years
#' and Daylight Savings Time.
#'
#' Durations provide a method for measuring generalized timespans when we wish to
#' treat time as a mathematical quantity that increases in a uniform, monotone manner
#' along a continuous numberline. They allow exact comparisons with other durations.
#' See [Period-class] for an alternative way to measure timespans that better
#' preserves clock times.
#'
#' Durations class objects have one slot: .Data, a numeric object equal to the number
#' of seconds in the duration.
#'
#' @aliases durations
#' @export
setClass("Duration", contains = c("Timespan", "numeric"), validity = check_duration)

#' @name hidden_aliases
#' @aliases Compare,Duration,ANY-method Compare,Duration,Duration-method
#'   Compare,difftime,Duration-method Compare,ANY,Duration-method
#'   Compare,Duration,Period-method Compare,Duration,difftime-method
#'   Compare,character,Duration-method Compare,Duration,character-method
#'   Compare,numeric,Duration-method Compare,Duration,numeric-method
#'   as.numeric,Duration-method as.integer,Duration-method show,Duration-method
#'   c,Duration-method rep,Duration-method [,Duration-method
#'   [<-,Duration,ANY,ANY,ANY-method [[,Duration-method [[<-,Duration,ANY,ANY,ANY-method
#'   $,Duration-method $<-,Duration-method as.difftime,Duration-method
#'   day<-,Duration-method hour<-,Duration-method minute<-,Duration-method
#'   month<-,Duration-method second<-,Duration-method year<-,Duration-method
#'   as.character,Duration-method +,Duration,Duration-method +,Duration,Interval-method
#'   +,Duration,Period-method +,Duration,Date-method +,Date,Duration-method
#'   +,Duration,difftime-method +,difftime,Duration-method +,Duration,numeric-method
#'   +,numeric,Duration-method +,Duration,POSIXct-method +,POSIXct,Duration-method
#'   +,Duration,POSIXlt-method +,POSIXlt,Duration-method -,Duration,ANY-method
#'   /,Duration,Duration-method /,Duration,Interval-method /,Duration,Period-method
#'   /,Duration,difftime-method /,difftime,Duration-method /,Duration,numeric-method
#'   /,numeric,Duration-method *,Duration,ANY-method *,ANY,Duration-method
#'   %%,Duration,Duration-method %%,Duration,Interval-method %%,Duration,Period-method
#'   -,Duration,missing-method -,ANY,Duration-method -,Duration,Duration-method
#'   -,Duration,Date-method -,Duration,POSIXct-method -,Duration,POSIXlt-method
#'   -,Duration,Period-method -,Duration,numeric-method -,POSIXct,Duration-method
#'   -,numeric,Duration-method -,POSIXlt,Duration-method -,Date,Duration-method
NULL

SECONDS_IN_ONE <- c(
  second = 1,
  minute = 60,
  hour   = 3600,
  day    = 86400,
  week   = 604800,
  year   = 31557600
)

.readable_duration <- function(x, unit) {
  if (unit == "second") {
    paste0(x, "s")
  } else {
    x2 <- round(x / SECONDS_IN_ONE[[unit]], 2)
    sprintf("%ss (~%s %ss)", x, x2, unit)
  }
}

.next_unit <- structure(as.list(c(names(SECONDS_IN_ONE[-1]), list(NULL))),
  names = names(SECONDS_IN_ONE)
)

compute_estimate <- function(secs, unit = "second") {
  next_unit <- .next_unit[[unit]]
  if (is.null(next_unit)) {
    return(.readable_duration(secs, "year"))
  }
  out <- character(length(secs))
  tt <- abs(secs) < SECONDS_IN_ONE[[next_unit]]
  if (any(tt)) {
    out[tt] <- .readable_duration(secs[tt], unit)
  }
  wnext <- which(!tt)
  if (length(wnext)) {
    out[wnext] <- compute_estimate(secs[wnext], next_unit)
  }
  out
}

#' @export
setMethod("show", signature(object = "Duration"), function(object) {
  if (length(object@.Data) == 0) {
    cat("<Duration[0]>\n")
  } else {
    print(format(object), quote = TRUE)
  }
})

#' @export
format.Duration <- function(x, ...) {
  out <- character(length(x@.Data))
  nnas <- !is.na(x@.Data)
  out[nnas] <- compute_estimate(x@.Data[nnas])
  out[!nnas] <- NA_character_
  out
}

#' @export
c.Duration <- function(...) {
  dots <- list(...)
  new("Duration", unlist(dots))
}

#' @export
setMethod("rep", signature(x = "Duration"), function(x, ...) {
  new("Duration", rep(as.numeric(x), ...))
})

#' @export
setMethod(
  "[", signature(x = "Duration"),
  function(x, i, j, ..., drop = TRUE) {
    new("Duration", x@.Data[i])
  }
)

#' @export
setMethod(
  "[[", signature(x = "Duration"),
  function(x, i, j, ..., exact = TRUE) {
    new("Duration", x@.Data[i])
  }
)

#' @export
setMethod(
  "[<-", signature(x = "Duration"),
  function(x, i, j, ..., value) {
    x@.Data[i] <- value
    new("Duration", x@.Data)
  }
)

#' @export
setMethod(
  "[[<-", signature(x = "Duration"),
  function(x, i, j, ..., value) {
    x@.Data[i] <- as.numeric(value)
    new("Duration", x@.Data)
  }
)


#' Create a duration object.
#'
#' `duration()` creates a duration object with the specified values. Entries
#' for different units are cumulative. durations display as the number of
#' seconds in a time span. When this number is large, durations also display an
#' estimate in larger units, however, the underlying object is always recorded
#' as a fixed number of seconds. For display and creation purposes, units are
#' converted to seconds using their most common lengths in seconds. Minutes = 60
#' seconds, hours = 3600 seconds, days = 86400 seconds, weeks = 604800. Units
#' larger than weeks are not used due to their variability.
#'
#' Durations record the exact number of seconds in a time span. They measure the
#' exact passage of time but do not always align with measurements
#' made in larger units of time such as hours, months and years.
#' This is because the length of larger time units can be affected
#' by conventions such as leap years
#' and Daylight Savings Time. Base R provides a second class for measuring
#' durations, the difftime class.
#'
#' Duration objects can be easily created with the helper functions [dweeks()],
#' [ddays()], [dminutes()], [dseconds()]. These objects can be added to and
#' subtracted to date- times to create a user interface similar to object
#' oriented programming.
#'
#' @param num the number or a character vector of time units. In string
#'   representation all unambiguous name units and abbreviations and ISO 8601
#'   formats are supported; 'm' stands for month and 'M' for minutes unless ISO
#'   8601 "P" modifier is present (see examples). Fractional units are
#'   supported.
#' @param units a character string that specifies the type of units that `num`
#'   refers to. When `num` is character, this argument is ignored.
#' @param ... a list of time units to be included in the duration and their
#'   amounts. Seconds, minutes, hours, days, weeks, months and years are
#'   supported. Durations of months and years assume that year consists of
#'   365.25 days.
#' @param x numeric value of the number of units to be contained in the
#'   duration.
#' @return a duration object
#' @seealso [as.duration()] [Duration-class]
#' @keywords chron classes
#' @examples
#'
#' ### Separate period and units vectors
#'
#' duration(90, "seconds")
#' duration(1.5, "minutes")
#' duration(-1, "days")
#'
#' ### Units as arguments
#'
#' duration(day = -1)
#' duration(second = 90)
#' duration(minute = 1.5)
#' duration(mins = 1.5)
#' duration(second = 3, minute = 1.5, hour = 2, day = 6, week = 1)
#' duration(hour = 1, minute = -60)
#'
#' ### Parsing
#'
#' duration("2M 1sec")
#' duration("2hours 2minutes 1second")
#' duration("2d 2H 2M 2S")
#' duration("2days 2hours 2mins 2secs")
#' # Missing numerals default to 1. Repeated units are added up.
#' duration("day day")
#'
#' ### ISO 8601 parsing
#'
#' duration("P3Y6M4DT12H30M5S")
#' duration("P23DT23H") # M stands for months
#' duration("10DT10M") # M stands for minutes
#' duration("P23DT60H 20min 100 sec") # mixing ISO and lubridate style parsing
#'
#' # Comparison with characters (from v1.6.0)
#'
#' duration("day 2 sec") > "day 1sec"
#'
#'
#' ## ELEMENTARY CONSTRUCTORS:
#'
#' dseconds(1)
#' dminutes(3.5)
#'
#' x <- ymd("2009-08-03", tz = "America/Chicago")
#' x + ddays(1) + dhours(6) + dminutes(30)
#' x + ddays(100) - dhours(8)
#'
#' class(as.Date("2009-08-09") + ddays(1)) # retains Date class
#' as.Date("2009-08-09") + dhours(12)
#' class(as.Date("2009-08-09") + dhours(12))
#' # converts to POSIXt class to accomodate time units
#'
#' dweeks(1) - ddays(7)
#' c(1:3) * dhours(1)
#'
#' # compare DST handling to durations
#' boundary <- ymd_hms("2009-03-08 01:59:59", tz = "America/Chicago")
#' boundary + days(1) # period
#' boundary + ddays(1) # duration
#' @export
duration <- function(num = NULL, units = "seconds", ...) {
  if (is.character(num)) {
    as.duration(parse_period(num))
  } else {
    out1 <- .duration_from_num(num, units)
    out2 <- .duration_from_pieces(list(...))
    if (is.null(out1) && is.null(out2)) {
      new("Duration", numeric())
    } else if (is.null(out1)) {
      out2
    } else if (is.null(out2)) {
      out1
    } else {
      c(out1, out2)
    }
  }
}

average_durations <- c(
  second = 1, minute = 60, hour = 3600, mday = 86400,
  wday = 86400, yday = 86400, day = 86400, week = 604800,
  month = 60 * 60 * 24 * 365.25 / 12, year = 60 * 60 * 24 * 365.25
)

.duration_from_num <- function(num, units) {
  if (length(num) == 0) {
    return(NULL)
  }

  if (!is.numeric(num)) {
    stop(sprintf("First argument to `duration()` constructor must be character or numeric. Supplied object of class '%s'", class(num)))
  }

  ## qucik check for common wrongdoings: #462
  if (inherits(num, c("Interval", "Period"))) {
    stop("Interval or Period objects cannot be used as input to `duration()` constructor. Use `as.duration()` instead.", call. = FALSE)
  }

  unit <- standardise_date_names(units)
  new("Duration", num * unname(average_durations[unit]))
}

.duration_from_pieces <- function(pieces) {
  if (length(pieces) == 0) {
    return(NULL)
  }
  names(pieces) <- standardise_date_names(names(pieces))
  out <- 0
  for (nm in names(pieces)) {
    out <- out + pieces[[nm]] * average_durations[[nm]]
  }
  new("Duration", out)
}

#' @export dseconds dminutes dmonths dhours ddays dweeks dyears dmilliseconds dmicroseconds dnanoseconds dpicoseconds
#' @rdname duration
dseconds <- function(x = 1) new("Duration", x)
#' @rdname duration
dminutes <- function(x = 1) new("Duration", x * 60)
#' @rdname duration
dhours <- function(x = 1) new("Duration", x * 3600)
#' @rdname duration
ddays <- function(x = 1) new("Duration", x * 86400)
#' @rdname duration
dweeks <- function(x = 1) new("Duration", x * average_durations[["week"]])
#' @rdname duration
dmonths <- function(x = 1) new("Duration", x * average_durations[["month"]])
#' @rdname duration
dyears <- function(x = 1) new("Duration", x * average_durations[["year"]])
#' @rdname duration
dmilliseconds <- function(x = 1) new("Duration", x / 1000)
#' @rdname duration
dmicroseconds <- function(x = 1) new("Duration", x / 1000 / 1000)
#' @rdname duration
dnanoseconds <- function(x = 1) new("Duration", x / 1000 / 1000 / 1000)
#' @rdname duration
dpicoseconds <- function(x = 1) new("Duration", x / 1000 / 1000 / 1000 / 1000)

#' @rdname duration
#' @export
#' @examples
#' is.duration(as.Date("2009-08-03")) # FALSE
#' is.duration(duration(days = 12.4)) # TRUE
is.duration <- function(x) is(x, "Duration")

#' @export
summary.Duration <- function(object, ...) {
  nas <- is.na(object)
  object <- object[!nas]
  nums <- as.numeric(object)
  qq <- stats::quantile(nums)
  qq <- c(qq[1L:3L], mean(nums), qq[4L:5L])
  qq <- dseconds(qq)
  qq <- as.character(qq)
  names(qq) <- c(
    "Min.", "1st Qu.", "Median", "Mean", "3rd Qu.",
    "Max."
  )
  if (any(nas)) {
    c(qq, `NA's` = sum(nas))
  } else {
    qq
  }
}

#' @export
setMethod(
  "Compare", c(e1 = "Duration", e2 = "ANY"),
  function(e1, e2) stop_incompatible_classes(e1, e2, .Generic)
)

#' @export
setMethod(
  "Compare", c(e1 = "ANY", e2 = "Duration"),
  function(e1, e2) {
    stop(sprintf(
      "Incompatible duration classes (%s, %s). Please coerce with `as.duration()`.",
      class(e1), class(e2)
    ),
    call. = FALSE
    )
  }
)

#' @export
setMethod(
  "Compare", signature(e1 = "Duration", e2 = "numeric"),
  function(e1, e2) {
    callGeneric(e1@.Data, e2)
  }
)

#' @export
setMethod(
  "Compare", signature(e1 = "numeric", e2 = "Duration"),
  function(e1, e2) {
    callGeneric(e1, e2@.Data)
  }
)

#' @export
setMethod(
  "Compare", signature(e1 = "Duration", e2 = "character"),
  function(e1, e2) {
    callGeneric(e1, as.duration(e2))
  }
)

#' @export
setMethod(
  "Compare", signature(e1 = "character", e2 = "Duration"),
  function(e1, e2) {
    callGeneric(as.duration(e1), e2)
  }
)

#' @export
setMethod(
  "Compare", c(e1 = "difftime", e2 = "Duration"),
  function(e1, e2) {
    callGeneric(as.numeric(e1, "secs"), e2@.Data)
  }
)

#' @export
setMethod(
  "Compare", c(e1 = "Duration", e2 = "difftime"),
  function(e1, e2) {
    callGeneric(e1@.Data, as.numeric(e2, "secs"))
  }
)

#' @export
setMethod(
  "Compare", c(e1 = "Duration", e2 = "Duration"),
  function(e1, e2) {
    callGeneric(e1@.Data, e2@.Data)
  }
)

test_that("Comparison operators work with POSIX and Date objects", {
  expect_true(ymd_hms("2016-01-03 00:00:00", tz = "") == "2016-01-03 00:00:00")
  expect_true(ymd_hms("2016-01-03 00:00:00", tz = "") == "2016-01-03")
  expect_true(ymd_hms("2016-01-03 00:00:01", tz = "") > "2016-01-03 00:00:00")
  expect_true(ymd("2016-01-03") == "2016-01-03")
  expect_true(ymd("2016-01-03") > "2016-01-02")
  expect_true(ymd("2016-01-03") < "2016-01-04")
  expect_true(ymd("2016-01-03") == ymd_hms("2016-01-03 00:00:00"))
  expect_true(ymd("2016-01-03") < ymd_hms("2016-01-03 00:00:01"))
  expect_true(ymd("2016-01-03") != ymd_hms("2016-01-03 00:00:01"))
  expect_true(ymd("2016-01-03") > ymd_hms("2016-01-02 23:59:59"))
  expect_true(ymd("2016-01-03") > ymd_hms("2016-01-02 23:59:59", tz = "Europe/Paris"))
  expect_true(ymd("2016-01-03") < ymd_hms("2016-01-03 00:00:01", tz = "Europe/Paris"))
  expect_true(ymd("2016-01-03") < ymd_hms("2016-01-03 00:00:01", tz = "Europe/Paris"))
  expect_true(ymd("2016-01-03") < as.POSIXlt(ymd_hms("2016-01-03 00:00:01", tz = "Europe/Paris")))
  expect_true(ymd("2016-01-03") < as.POSIXlt(ymd_hms("2016-01-03 00:00:01", tz = "Europe/Paris")))
})

test_that("Duration and periods are comparable", {
  expect_true(dyears(1) == years(1))
})

test_that("character comparison with periods works as expected", {
  expect_true(period(days = 1) == "day")
  expect_true(period(days = 2, seconds = 2) == "2 days, 2 secs")
  expect_true("day" == period(days = 1))
  expect_true("2 days, 2 secs" == period(days = 2, seconds = 2))
  expect_true(period("day 1s") > period(days = 1))
  expect_true("day 1s" > period(days = 1))
  expect_true("day 1S 2H" == period(days = 1, seconds = 1, hours = 2))
  expect_false("day 1S 2H" < period(days = 1, hours = 2))
  expect_true("day 1S 2H 3m 2y" > period(days = 1, months = 3, years = 2, hours = 2))
})

test_that("duration comparison with periods works as expected", {
  expect_true(period(days = 1) == duration(days = 1))
  expect_true(duration(days = 1) == period(days = 1))
  expect_true(period(days = 2, seconds = 2) == duration("2 days, 2 secs"))
  expect_true(duration("day") == period(days = 1))
  expect_true(ddays() == days())
  expect_true(duration("day 1s") > period(days = 1))
  expect_true(duration("day 1s") > period(days = 1))
  expect_true(duration("day 1S 2H") == period(days = 1, seconds = 1, hours = 2))
  expect_false(duration("day 1S 2H") < period(days = 1, hours = 2))
  expect_true(duration("day 1S 2H 3m 2y") > period(days = 1, months = 3, years = 2, hours = 2))
})

test_that("character comparison with durations works as expected", {
  expect_true("day" == duration(days = 1))
  expect_true("2 days, 2 secs" == duration(days = 2, seconds = 2))
  expect_true(duration("day 1s") > duration(days = 1))
  expect_true("day 1s" > duration(days = 1))
  expect_true("day 1S 2H" == duration(days = 1, seconds = 1, hours = 2))
  expect_false("day 1S 2H" < duration(days = 1, hours = 2))
})

test_that("numeric comparison with durations works as expected", {
  expect_true(1 == duration(secs = 1))
  expect_true(172802 == duration(days = 2, seconds = 2))
  expect_true(86401 > duration(days = 1))
  expect_true(93601 == duration(days = 1, seconds = 1, hours = 2))
  expect_true(86400 < duration(days = 1, hours = 2))
  expect_true(10 > as.duration(1))
})

test_that("difftime comparison with periods works", {
  t <- ymd_hms("2019-03-01 12:30:50")
  expect_true(days(1) == (t + days(1)) - t)
  expect_true((t + days(1)) - t == days(1))
})

test_that("difftime comparison with durations works", {
  t <- ymd_hms("2019-03-01 12:30:50")
  expect_true(ddays(1) == (t + ddays(1)) - t)
  expect_true((t + ddays(1)) - t == ddays(1))
})


test_that("Comparison operators work duration and difftime objects (#323)", {
  t1 <- ymd_hms("2019-03-01 12:30:50")
  t2 <- t1 + dhours(1)
  t3 <- t1 + dseconds(1)

  expect_true((t2 - t1) > dseconds(60))
  expect_false((t2 - t1) > dseconds(3600))
  expect_true((t2 - t1) < dseconds(3601))

  expect_true(dhours(1) > dminutes(59))
  expect_true(dhours(1) == dseconds(3600))

  expect_true(dhours(1) == 3600)
  expect_false(dhours(1) == 1)
})

context("Comparisons operations")

test_that("Comparison operators work with POSIX and Date objects", {
  expect_true(ymd_hms("2016-01-03 00:00:00", tz = "") == "2016-01-03 00:00:00")
  expect_true(ymd_hms("2016-01-03 00:00:00", tz = "") == "2016-01-03")
  expect_true(ymd_hms("2016-01-03 00:00:01", tz = "") > "2016-01-03 00:00:00")
  expect_true(ymd("2016-01-03") == "2016-01-03")
  expect_true(ymd("2016-01-03") >  "2016-01-02")
  expect_true(ymd("2016-01-03") <  "2016-01-04")
  expect_true(ymd("2016-01-03") == ymd_hms("2016-01-03 00:00:00"))
  expect_true(ymd("2016-01-03") < ymd_hms("2016-01-03 00:00:01"))
  expect_true(ymd("2016-01-03") != ymd_hms("2016-01-03 00:00:01"))
  expect_true(ymd("2016-01-03") > ymd_hms("2016-01-02 23:59:59"))
  expect_true(ymd("2016-01-03") > ymd_hms("2016-01-02 23:59:59", tz = "Europe/Paris"))
  expect_true(ymd("2016-01-03") < ymd_hms("2016-01-03 00:00:01", tz = "Europe/Paris"))
  expect_true(ymd("2016-01-03") < ymd_hms("2016-01-03 00:00:01", tz = "Europe/Paris"))
  expect_true(ymd("2016-01-03") < as.POSIXlt(ymd_hms("2016-01-03 00:00:01", tz = "Europe/Paris")))
  expect_true(ymd("2016-01-03") < as.POSIXlt(ymd_hms("2016-01-03 00:00:01", tz = "Europe/Paris")))
})

test_that("Duration and periods are comparable", {
  expect_true(dyears(1) == years(1))
})

test_that("character comparison with periods works as expected", {
  expect_true(period(days = 1) == "day")
  expect_true(period(days = 2, seconds = 2) == "2 days, 2 secs")
  expect_true("day" == period(days = 1))
  expect_true("2 days, 2 secs" == period(days = 2, seconds = 2))
  expect_true(period("day 1s") >  period(days = 1))
  expect_true("day 1s" >  period(days = 1))
  expect_true("day 1S 2H" == period(days = 1, seconds = 1, hours = 2))
  expect_false("day 1S 2H" < period(days = 1, hours = 2))
  expect_true("day 1S 2H 3m 2y" >  period(days = 1, months = 3, years = 2, hours = 2))
})

test_that("duration comparison with periods works as expected", {
  expect_true(period(days = 1) == duration(days = 1))
  expect_true(duration(days = 1) == period(days = 1))
  expect_true(period(days = 2, seconds = 2) == duration("2 days, 2 secs"))
  expect_true(duration("day") == period(days = 1))
  expect_true(ddays() == days())
  expect_true(duration("day 1s") >  period(days = 1))
  expect_true(duration("day 1s") >  period(days = 1))
  expect_true(duration("day 1S 2H") == period(days = 1, seconds = 1, hours = 2))
  expect_false(duration("day 1S 2H") < period(days = 1, hours = 2))
  expect_true(duration("day 1S 2H 3m 2y") >  period(days = 1, months = 3, years = 2, hours = 2))
})

test_that("character comparison with durations works as expected", {
  expect_true("day" == duration(days = 1))
  expect_true("2 days, 2 secs" == duration(days = 2, seconds = 2))
  expect_true(duration("day 1s") >  duration(days = 1))
  expect_true("day 1s" >  duration(days = 1))
  expect_true("day 1S 2H" == duration(days = 1, seconds = 1, hours = 2))
  expect_false("day 1S 2H" < duration(days = 1, hours = 2))
})

test_that("numeric comparison with durations works as expected", {
  expect_true(1 == duration(secs = 1))
  expect_true(172802 == duration(days = 2, seconds = 2))
  expect_true(86401 > duration(days = 1))
  expect_true(93601 == duration(days = 1, seconds = 1, hours = 2))
  expect_true(86400 < duration(days = 1, hours = 2))
  expect_true(10 > as.duration(1))
})

test_that("difftime comparison with periods works", {
  t <- ymd_hms("2019-03-01 12:30:50")
  expect_true(days(1) == (t + days(1)) - t)
  expect_true((t + days(1)) - t == days(1))
})

test_that("difftime comparison with durations works", {
  t <- ymd_hms("2019-03-01 12:30:50")
  expect_true(ddays(1) == (t + ddays(1)) - t)
  expect_true((t + ddays(1)) - t == ddays(1))
})


test_that("Comparison operators work duration and difftime objects (#323)", {
  t1 <- ymd_hms("2019-03-01 12:30:50")
  t2 <- t1 + dhours(1)
  t3 <- t1 + dseconds(1)

  expect_true((t2 - t1) >  dseconds(60))
  expect_false((t2 - t1) >  dseconds(3600))
  expect_true((t2 - t1) < dseconds(3601))

  expect_true(dhours(1) > dminutes(59))
  expect_true(dhours(1) == dseconds(3600))

  expect_true(dhours(1) == 3600)
  expect_false(dhours(1) == 1)
})

context("Durations")

test_that("duration() returns zero-length vector", {
  x <- duration()
  expect_s4_class(x, "Duration")
  expect_length(x, 0)
  expect_equal(format(x), character())
  expect_output(print(x), "<Duration[0]>", fixed = TRUE)
})

test_that("duration(...) returns zero-length with on 0-length inputs", {
  x <- duration(character())
  expect_s4_class(x, "Duration")
  expect_length(x, 0)

  x <- duration(hour = numeric())
  expect_s4_class(x, "Duration")
  expect_length(x, 0)

  x <- duration(numeric(), hour = numeric())
  expect_s4_class(x, "Duration")
  expect_length(x, 0)

  expect_equal(duration(c(30, 20)),
               duration(c(30, 20), days = numeric()))

  expect_equal(duration(numeric(), days = 10),
               duration(days = 10))
})

test_that("duration constructor doesn't accept non-numeric or non-character inputs", {
  expect_error(duration(interval(ymd("2014-01-01"), ymd("2015-01-01"))))
})

test_that("make_difftime works as expected", {
  x <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  y <- difftime(x + 5 + 30*60 + 60*60 + 14*24*60*60, x, tz = "UTC")
  attr(y, "tzone") <- NULL
  diff <- make_difftime(seconds = 5, minutes = 30, days = 0, hour = 1, weeks = 2)
  expect_equal(diff, y)
})

test_that("Duration parsing works", {
  expect_equal(duration("1min 2sec 2secs 1H 2M 1d"),
               duration(seconds = 4, minutes = 3, hours = 1, days = 1))
  expect_equal(duration("day day"),
               duration(days = 2))
  expect_equal(duration("S M H d w"),
               duration(seconds = 1, minutes = 1, hours = 1, days = 1, weeks = 1))
  expect_equal(duration("1S 2M 3H 4d 5w"),
               duration(seconds = 1, minutes = 2, hours = 3, days = 4, weeks = 5))
})

test_that("fractional parsing works as expected", {
  expect_equal(duration("1.1min 2.3sec 2.3secs 1.0H 2.2M 1.5d"),
               duration(seconds = 43222.6, minutes = 3, hours = 1, days = 1))
  expect_equal(duration("day 1.2days"),
               duration(days = 2, seconds = 17280))
})

test_that("sub-unit fractional parsing works as expected", {

  expect_identical(duration(".1min .3sec .3secs .0H .2M .5d"),
                   duration("0.1min 0.3sec 0.3secs 0.0H 0.2M 0.5d"))

  expect_identical(duration(".1min .3sec .3secs .0H .2M .5d"),
                   duration(seconds = 6.6, minutes = .2, hours = 12))
})

test_that("parsing with 0 units works as expected", {
  expect_equal(as.numeric(duration("2d 0H 0M 1s")), 2*24*3600 + 1)
  expect_equal(as.numeric(duration("0d 0H 0M 1s")), 1)
  expect_equal(period("2d 0H 0M 1s"), days(2) + seconds(1))
  expect_equal(period("y 0m 2d 0H 0M 1s"), years(1) + days(2) + seconds(1))
})

test_that("make_difftime handles vectors", {
  x <- as.POSIXct(c("2008-08-03 13:01:59", "2008-08-03 13:01:59"), tz = "UTC")
  y <- difftime(x + c(5 + 30*60 + 60*60 + 14*24*60*60,
    1 + 3*24*60*60 + 60*60), x, tz = "UTC")
  attr(y, "tzone") <- NULL
  z <- difftime(x + c(5 + 30*60 + 60*60 + 14*24*60*60,
                      5 + 30*60 + 60*60 + 14*24*60*60 + 3*24*60*60),
                x, tz = "UTC")
  attr(z, "tzone") <- NULL

  expect_equal(make_difftime(seconds = c(5, 1), minutes = c(30,0),
                             days = c(0, 3), hour = c(1, 1), weeks = c(2, 0)),
               y)

  expect_equal(make_difftime(seconds = 5, minutes = 30,
                             days = c(0, 3), hour = 1, weeks = 2),
               z)

})

test_that("duration works as expected", {
  dur <- duration(seconds = 5, minutes = 30, days = 0,
    hour = 1, weeks = 2)

  expect_equal(dur@.Data, 1215005)
  expect_is(dur, "Duration")
})

test_that("duration handles vectors", {
  dur1 <- duration(seconds = c(5, 1), minutes = c(30, 0),
        days = c(0, 3), hour = c(1, 1), weeks = c(2, 0))
  dur2 <- duration(seconds = 5, minutes = 30, days =
    c(0, 3), hour = 1, weeks = 2)

  expect_equal(dur1@.Data, c(1215005, 262801))
  expect_equal(dur2@.Data, c(1215005, 1474205))
  expect_is(dur1, "Duration")
  expect_is(dur2, "Duration")

})

test_that("as.duration handles vectors", {
  expect_equal(as.duration(minutes(1:3)), dminutes(1:3))
})

test_that("as.duration handles periods", {

  expect_equal(as.duration(seconds(1)), dseconds(1))
  expect_equal(as.duration(minutes(2)), dminutes(2))
  expect_equal(as.duration(hours(3)), dhours(3))
  expect_equal(as.duration(days(4)), ddays(4))
  expect_equal(as.duration(weeks(5)), dweeks(5))
  expect_equal(as.duration(months(1)), dseconds(60*60*24*365.25/12))
  expect_equal(as.duration(years(1)), dseconds(60*60*24*365.25))
  expect_equal(as.duration(seconds(1) + minutes(4)), dseconds(1) + dminutes(4))
})

test_that("as.duration handles intervals", {
        time1 <- as.POSIXct("2009-01-02 12:24:03", tz = "UTC")
        time2 <- as.POSIXct("2010-02-03 14:31:42", tz = "UTC")
        dur <- as.duration(interval(time1, time2))
    y <- as.numeric(time2 - time1, units = "secs")

  expect_equal(dur@.Data, y)
  expect_is(dur, "Duration")
})

test_that("as.duration handles difftimes", {
  x <- difftime(as.POSIXct("2010-02-03 14:31:42", tz = "UTC"),
    as.POSIXct("2009-01-02 12:24:03", tz = "UTC"))
  dur <- as.duration(x)
  y <- as.numeric(x, units = "secs")

  expect_equal(dur@.Data, y)
  expect_is(dur, "Duration")
})

test_that("eobjects handle vectors", {
  dur <- dseconds(c(1, 3, 4))

  expect_equal(dur@.Data, c(1, 3, 4))
  expect_is(dur, "Duration")
})

test_that("is.duration works as expected", {
  ct_time <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  lt_time <- as.POSIXlt("2009-08-03 13:01:59", tz = "UTC")

  expect_false(is.duration(234))
  expect_false(is.duration(ct_time))
  expect_false(is.duration(lt_time))
  expect_false(is.duration(Sys.Date()))
  expect_false(is.duration(minutes(1)))
  expect_true(is.duration(dminutes(1)))
  expect_false(is.duration(make_difftime(1000)))
  expect_false(is.duration(interval(lt_time, ct_time)))
})

test_that("format.Duration works as expected", {
  dur <- duration(seconds = c(5, NA, 10, -10, 1000, -1000))
  expect_equivalent(format(dur),
                    c("5s", NA, "10s", "-10s",
                      "1000s (~16.67 minutes)",
                      "-1000s (~-16.67 minutes)"))
})

test_that("summary.Duration creates useful summary", {
  dur <- dminutes(5)
  text <- c(rep("300s (~5 minutes)", 6), 1)
  names(text) <- c("Min.", "1st Qu.", "Median", "Mean", "3rd Qu.", "Max.", "NA's")
  expect_equal(summary(c(dur, NA)), text)
})

test_that("format.Duration works with NA values", {
  expect_equal(format(duration(c(NA, 1, 100, 10000, 100000, 100000000, 75.25, 75.5001))),
               c(NA, "1s", "100s (~1.67 minutes)", "10000s (~2.78 hours)", "1e+05s (~1.16 days)",
                 "1e+08s (~3.17 years)",
                 "75.25s (~1.25 minutes)", "75.5001s (~1.26 minutes)"))
  expect_equal(format(duration(c(75.25, 75.5001))),
               c("75.25s (~1.25 minutes)", "75.5001s (~1.26 minutes)"))
})

test_that("as.duration handles NA interval objects", {
  one_missing_date <- as.POSIXct(NA_real_, origin = origin)
  one_missing_interval <- interval(one_missing_date,
                                       one_missing_date)
  several_missing_dates <- rep(as.POSIXct(NA_real_, origin = origin), 2)
  several_missing_intervals <- interval(several_missing_dates,
                                            several_missing_dates)
  start_missing_intervals <- interval(several_missing_dates, origin)
  end_missing_intervals <- interval(origin, several_missing_dates)
  na.dur <- dseconds(NA)

  expect_equal(as.duration(one_missing_interval), na.dur)
  expect_equal(as.duration(several_missing_intervals), c(na.dur, na.dur))
  expect_equal(as.duration(start_missing_intervals), c(na.dur, na.dur))
  expect_equal(as.duration(end_missing_intervals), c(na.dur, na.dur))
})

test_that("as.duration handles NA period objects", {
  na.dur <- dseconds(NA)
  expect_equal(as.duration(years(NA)), na.dur)
  expect_equal(as.duration(years(c(NA, NA))), c(na.dur, na.dur))
  expect_equal(as.duration(years(c(1, NA))), c(dyears(1), na.dur))
})

test_that("as.duration handles NA objects", {
  na.dur <- dseconds(NA)
  expect_equal(as.duration(NA), na.dur)
})

#' Get **d**aylight **s**avings **t**ime indicator of a date-time
#'
#' Date-time must be a POSIXct, POSIXlt, Date, chron, yearmon, yearqtr, zoo,
#' zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
#'
#' A date-time's daylight savings flag can not be set because it depends on the
#' date-time's year, month, day, and hour values.
#'
#' @export
#' @param x a date-time object
#' @return A logical. TRUE if DST is in force, FALSE if not, NA if unknown.
#' @keywords utilities chron methods
#' @examples
#' x <- ymd("2012-03-26")
#' dst(x)
dst <- function(x) {
  UseMethod("dst")
}

#' @export
dst.default <- function(x) {
  c(NA, FALSE, TRUE)[as.POSIXlt(x)$isdst + 2]
}

##' Parse dates with **y**ear, **m**onth, and **d**ay components
##'
##' Transforms dates stored in character and numeric vectors to Date or POSIXct
##' objects (see `tz` argument). These functions recognize arbitrary
##' non-digit separators as well as no separator. As long as the order of
##' formats is correct, these functions will parse dates correctly even when the
##' input vectors contain differently formatted dates. See examples.
##'
##' In case of heterogeneous date formats, the `ymd()` family guesses formats based
##' on a subset of the input vector. If the input vector contains many missing
##' values or non-date strings, the subset might not contain meaningful dates
##' and the date-time format won't be guessed resulting in
##' `All formats failed to parse` error. In such cases please see
##' [parse_date_time()] for a more flexible parsing interface.
##'
##' If the `truncated` parameter is non-zero, the `ymd()` functions also check for
##' truncated formats. For example, `ymd()` with `truncated = 2` will also
##' parse incomplete dates like `2012-06` and `2012`.
##'
##' NOTE: The `ymd()` family of functions is based on `parse_date_time()` and thus
##' directly drop to the internal C parser for numeric months, but uses
##' [base::strptime()] for alphabetic months. This implies that some of [base::strptime()]'s
##' limitations are inherited by \pkg{lubridate}'s parser. For example, truncated
##' formats (like `%Y-%b`) will not be parsed. Numeric truncated formats (like
##' `%Y-%m`) are handled correctly by \pkg{lubridate}'s C parser.
##'
##' As of version 1.3.0, \pkg{lubridate}'s parse functions no longer return a
##' message that displays which format they used to parse their input. You can
##' change this by setting the `lubridate.verbose` option to `TRUE` with
##' `options(lubridate.verbose = TRUE)`.
##'
##' @export
##' @param ... a character or numeric vector of suspected dates
##' @param quiet logical. If `TRUE`, function evaluates without displaying
##'   customary messages.
##' @param tz Time zone indicator. If `NULL` (default), a Date object is
##'   returned. Otherwise a POSIXct with time zone attribute set to `tz`.
##' @param locale locale to be used, see [locales]. On Linux systems you
##'   can use `system("locale -a")` to list all the installed locales.
##' @param truncated integer. Number of formats that can be truncated.
##' @return a vector of class POSIXct if `tz` argument is non-`NULL` or Date if tz
##'   is `NULL` (default)
##' @seealso [parse_date_time()] for an even more flexible low level
##'   mechanism.
##' @keywords chron
##' @examples
##' x <- c("09-01-01", "09-01-02", "09-01-03")
##' ymd(x)
##' x <- c("2009-01-01", "2009-01-02", "2009-01-03")
##' ymd(x)
##' ymd(090101, 90102)
##' now() > ymd(20090101)
##' ## TRUE
##' dmy(010210)
##' mdy(010210)
##'
##' yq('2014.2')
##'
##' ## heterogeneous formats in a single vector:
##' x <- c(20090101, "2009-01-02", "2009 01 03", "2009-1-4",
##'        "2009-1, 5", "Created on 2009 1 6", "200901 !!! 07")
##' ymd(x)
##'
##' ## What lubridate might not handle:
##'
##' ## Extremely weird cases when one of the separators is "" and some of the
##' ## formats are not in double digits might not be parsed correctly:
##' \dontrun{ymd("201002-01", "201002-1", "20102-1")
##' dmy("0312-2010", "312-2010")}
ymd <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx(..., orders = "ymd", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd
ydm <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx(..., orders = "ydm", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd
mdy <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx(..., orders = "mdy", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd
myd <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx(..., orders = "myd", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd
dmy <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx(..., orders = "dmy", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd
dym <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx(..., orders = "dym", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd
yq <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"))
  .parse_xxx(..., orders = "yq", quiet = quiet, tz = tz, locale = locale, truncated = 0)

#' @export
#' @rdname ymd
ym <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"))
  .parse_xxx(..., orders = "ym", quiet = quiet, tz = tz, locale = locale, truncated = 0)

#' @export
#' @rdname ymd
my <- function(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"))
  .parse_xxx(..., orders = "my", quiet = quiet, tz = tz, locale = locale, truncated = 0)


##' Parse date-times with **y**ear, **m**onth, and **d**ay, **h**our,
##' **m**inute, and **s**econd components.
##'
##' Transform dates stored as character or numeric vectors to POSIXct
##' objects. The `ymd_hms()` family of functions recognizes all non-alphanumeric
##' separators (with the exception of "." if `frac = TRUE`) and correctly
##' handles heterogeneous date-time representations. For more flexibility in
##' treatment of heterogeneous formats, see low level parser
##' [parse_date_time()].
##'
##' The `ymd_hms()` functions automatically assign the Universal Coordinated Time
##' Zone (UTC) to the parsed date. This time zone can be changed with
##' [force_tz()].
##'
##' The most common type of irregularity in date-time data is the truncation
##' due to rounding or unavailability of the time stamp. If the `truncated`
##' parameter is non-zero, the `ymd_hms()` functions also check for truncated
##' formats. For example, `ymd_hms()` with `truncated = 3` will also parse
##' incomplete dates like `2012-06-01 12:23`, `2012-06-01 12` and
##' `2012-06-01`. NOTE: The `ymd()` family of functions is based on
##' [base::strptime()] which currently fails to parse `%y-%m` formats.
##'
##' In case of heterogeneous date formats the `ymd_hms()` family guesses formats
##' based on a subset of the input vector. If the input vector contains many
##' missing values or non-date strings, the subset might not contain meaningful
##' dates and the date-time format won't be guessed resulting in
##' `All formats failed to parse` error. In such cases please see
##' [parse_date_time()] for a more flexible parsing interface.
##'
##' As of version 1.3.0, \pkg{lubridate}'s parse functions no longer return a
##' message that displays which format they used to parse their input. You can
##' change this by setting the `lubridate.verbose` option to `TRUE` with
##' `options(lubridate.verbose = TRUE)`.
##'
##' @export
##' @param ... a character vector of dates in year, month, day, hour, minute,
##'   second format
##' @param quiet logical. If `TRUE`, function evaluates without displaying customary messages.
##' @param tz a character string that specifies which time zone to parse the date with. The string
##' must be a time zone that is recognized by the user's OS.
##' @param locale locale to be used, see \link{locales}. On Linux systems you
##' can use `system("locale -a")` to list all the installed locales.
##' @param truncated integer, indicating how many formats can be missing. See details.
##' @return a vector of [POSIXct] date-time objects
##' @seealso
##' - [ymd()], [hms()]
##' - [parse_date_time()] for the underlying mechanism
##' @keywords POSIXt parse
##' @examples
##'
##' x <- c("2010-04-14-04-35-59", "2010-04-01-12-00-00")
##' ymd_hms(x)
##' x <- c("2011-12-31 12:59:59", "2010-01-01 12:00:00")
##' ymd_hms(x)
##'
##'
##' ## ** heterogeneous formats **
##' x <- c(20100101120101, "2009-01-02 12-01-02", "2009.01.03 12:01:03",
##'        "2009-1-4 12-1-4",
##'        "2009-1, 5 12:1, 5",
##'        "200901-08 1201-08",
##'        "2009 arbitrary 1 non-decimal 6 chars 12 in between 1 !!! 6",
##'        "OR collapsed formats: 20090107 120107 (as long as prefixed with zeros)",
##'        "Automatic wday, Thu, detection, 10-01-10 10:01:10 and p format: AM",
##'        "Created on 10-01-11 at 10:01:11 PM")
##' ymd_hms(x)
##'
##' ## ** fractional seconds **
##' op <- options(digits.secs=3)
##' dmy_hms("20/2/06 11:16:16.683")
##' options(op)
##'
##' ## ** different formats for ISO8601 timezone offset **
##' ymd_hms(c("2013-01-24 19:39:07.880-0600",
##' "2013-01-24 19:39:07.880", "2013-01-24 19:39:07.880-06:00",
##' "2013-01-24 19:39:07.880-06", "2013-01-24 19:39:07.880Z"))
##'
##' ## ** internationalization **
##' \dontrun{
##' x_RO <- "Ma 2012 august 14 11:28:30 "
##'   ymd_hms(x_RO, locale = "ro_RO.utf8")
##' }
##'
##' ## ** truncated time-dates **
##' x <- c("2011-12-31 12:59:59", "2010-01-01 12:11", "2010-01-01 12", "2010-01-01")
##' ymd_hms(x, truncated = 3)
##' x <- c("2011-12-31 12:59", "2010-01-01 12", "2010-01-01")
##' ymd_hm(x, truncated = 2)
##' ## ** What lubridate might not handle **
##' ## Extremely weird cases when one of the separators is "" and some of the
##' ## formats are not in double digits might not be parsed correctly:
##' \dontrun{
##' ymd_hm("20100201 07-01", "20100201 07-1", "20100201 7-01")}
##'
ymd_hms <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"),  truncated = 0){
  .parse_xxx_hms(..., orders = c("ymdTz", "ymdT"), quiet = quiet, tz = tz, locale = locale,  truncated = truncated)
}

#' @export
#' @rdname ymd_hms
ymd_hm <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx_hms(..., orders =  "ymdR", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd_hms
ymd_h <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx_hms(..., orders = "ymdr", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd_hms
dmy_hms <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx_hms(..., orders = c("dmyTz", "dmyT"), quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd_hms
dmy_hm <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx_hms(..., orders = "dmyR", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd_hms
dmy_h <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx_hms(..., orders = "dmyr", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd_hms
mdy_hms <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx_hms(..., orders = c("mdyTz", "mdyT"), quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd_hms
mdy_hm <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx_hms(..., orders = "mdyR", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd_hms
mdy_h <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx_hms(..., orders = "mdyr", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd_hms
ydm_hms <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx_hms(..., orders = c("ydmTz", "ydmT"), quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd_hms
ydm_hm <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx_hms(..., orders = "ydmR", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)

#' @export
#' @rdname ymd_hms
ydm_h <- function(..., quiet = FALSE, tz = "UTC", locale = Sys.getlocale("LC_TIME"),  truncated = 0)
  .parse_xxx_hms(..., orders = "ydmr", quiet = quiet, tz = tz, locale = locale,  truncated = truncated)


##' @rdname hms
##' @examples
##' ms(c("09:10", "09:02", "1:10"))
##' ms("7 6")
##' ms("6,5")
##' @export
ms <- function(..., quiet = FALSE, roll = FALSE) {
  out <- .parse_hms(..., order = "MS", quiet = quiet)
  if (roll) {
    hms <- .roll_hms(min = out["M", ], sec = out["S", ])
    period(hour = hms$hour, minute = hms$min, second = hms$sec)
  } else {
    period(minute = out["M", ], second = out["S", ])
  }
}

##' @rdname hms
##' @examples
##' hm(c("09:10", "09:02", "1:10"))
##' hm("7 6")
##' hm("6,5")
##' @export
hm <- function(..., quiet = FALSE, roll = FALSE) {
  out <- .parse_hms(..., order = "HM", quiet = quiet)
  if (roll) {
    hms <- .roll_hms(hour = out["H", ], min = out["M", ])
    period(hour = hms$hour, minute = hms$min, second = hms$sec)
  } else {
    period(hour = out["H", ], minute = out["M", ])
  }
}

##' Parse periods with **h**our, **m**inute, and **s**econd components
##'
##' Transforms a character or numeric vector into a period object with the
##' specified number of hours, minutes, and seconds. `hms()` recognizes all
##' non-numeric characters except '-' as separators ('-' is used for negative
##' `durations`). After hours, minutes and seconds have been parsed, the
##' remaining input is ignored.
##'
##' @param ... a character vector of hour minute second triples
##' @param quiet logical. If `TRUE`, function evaluates without displaying
##'   customary messages.
##' @param roll logical. If `TRUE`, smaller units are rolled over to higher units
##'   if they exceed the conventional limit. For example,
##'   `hms("01:59:120", roll = TRUE)` produces period "2H 1M 0S".
##' @return a vector of period objects
##' @seealso [hm()], [ms()]
##' @keywords period
##' @examples
##'
##' x <- c("09:10:01", "09:10:02", "09:10:03")
##' hms(x)
##'
##' hms("7 6 5", "3:23:::2", "2 : 23 : 33", "Finished in 9 hours, 20 min and 4 seconds")
##' @export
hms <- function(..., quiet = FALSE, roll = FALSE) {
  out <- .parse_hms(..., order = "HMS", quiet = quiet)
  if (roll) {
    hms <- .roll_hms(out["H", ], out["M", ], out["S", ])
    period(hour = hms$hour, minute = hms$min, second = hms$sec)
  } else {
    period(hour = out["H", ], minute = out["M", ], second = out["S", ])
  }
}

.roll_hms <- function(hour = 0, min = 0, sec = 0){
  min <- min + sec %/% 60
  sec <- sec %% 60
  hour <- hour + min %/% 60
  min <- min %% 60
  list(hour = hour, min = min, sec = sec)
}

.parse_hms <- function(..., order, quiet = FALSE) {
  ## wrapper for C level parse_hms
  hms <- unlist(lapply(list(...), .num_to_date), use.names= FALSE)
  out <- matrix(.Call(C_parse_hms, hms, order),
                nrow = 3L, dimnames = list(c("H", "M", "S"), NULL))
  if (!quiet) {
    ## fixme: this warning should be dropped to C and thrown only when there are
    ## real parsing errors #530
    if (any(is.na(out[substr(order, ln <- nchar(order), ln), ])))
      warning("Some strings failed to parse, or all strings are NAs")
  }
  out
}

##' User friendly date-time parsing functions
##'
##' `parse_date_time()` parses an input vector into POSIXct date-time
##' object. It differs from [base::strptime()] in two respects. First,
##' it allows specification of the order in which the formats occur without the
##' need to include separators and the `%` prefix. Such a formatting argument is
##' referred to as "order". Second, it allows the user to specify several
##' format-orders to handle heterogeneous date-time character
##' representations.
##'
##' When several format-orders are specified, `parse_date_time()` selects
##' (guesses) format-orders based on a training subset of the input
##' strings. After guessing the formats are ordered according to the performance
##' on the training set and applied recursively on the entire input vector. You
##' can disable training with `train = FALSE`.
##'
##' `parse_date_time()`, and all derived functions, such as `ymd_hms()`,
##' `ymd()`, etc., will drop into `fast_strptime()` instead of
##' [base::strptime()] whenever the guessed from the input data formats are all
##' numeric.
##'
##' The list below contains formats recognized by \pkg{lubridate}. For numeric formats
##' leading 0s are optional. As compared to [base::strptime()], some of the
##' formats are new or have been extended for efficiency reasons. These formats
##' are marked with "(*)". The fast parsers `parse_date_time2()` and
##' `fast_strptime()` accept only formats marked with "(!)".
##'
##'
##' \describe{ \item{`a`}{Abbreviated weekday name in the current
##' locale. (Also matches full name)}
##'
##' \item{`A`}{Full weekday name in the current locale.  (Also matches
##' abbreviated name).
##'
##' You don't need to specify `a` and `A` formats explicitly. Wday is
##' automatically handled if `preproc_wday = TRUE`}
##'
##' \item{`b` (!)}{Abbreviated or full month name in the current locale. The C
##' parser currently understands only English month names.}
##'
##' \item{`B` (!)}{Same as b.}
##'
##' \item{`d` (!)}{Day of the month as decimal number (01--31 or 0--31)}
##'
##' \item{`H` (!)}{Hours as decimal number (00--24 or 0--24).}
##'
##' \item{`I` (!)}{Hours as decimal number (01--12 or 1--12).}
##'
##' \item{`j`}{Day of year as decimal number (001--366 or 1--366).}
##'
##' \item{`q` (!*)}{Quarter (1--4). The quarter month is added to the parsed month
##' if `m` format is present.}
##'
##' \item{`m` (!*)}{Month as decimal number (01--12 or 1--12). For
##'                   `parse_date_time`. As a \pkg{lubridate} extension, also
##'                   matches abbreviated and full months names as `b` and
##'                   `B` formats. C parser understands only English month
##'                   names.}
##'
##' \item{`M` (!)}{Minute as decimal number (00--59 or 0--59).}
##'
##' \item{`p` (!)}{AM/PM indicator in the locale. Normally used in conjunction
##'                   with `I` and \bold{not} with `H`.  But the \pkg{lubridate}
##'                   C parser accepts H format as long as hour is not greater
##'                   than 12. C parser understands only English locale AM/PM
##'                   indicator.}
##'
##' \item{`S` (!)}{Second as decimal number (00--61 or 0--61), allowing for up
##' to two leap-seconds (but POSIX-compliant implementations will ignore leap
##' seconds).}
##'
##' \item{`OS`}{Fractional second.}
##'
##' \item{`U`}{Week of the year as decimal number (00--53 or 0--53) using
##' Sunday as the first day 1 of the week (and typically with the first Sunday
##' of the year as day 1 of week 1).  The US convention.}
##'
##' \item{`w`}{Weekday as decimal number (0--6, Sunday is 0).}
##'
##' \item{`W`}{Week of the year as decimal number (00--53 or 0--53) using
##' Monday as the first day of week (and typically with the first Monday of the
##' year as day 1 of week 1).  The UK convention.}
##'
##' \item{`y` (!*)}{Year without century (00--99 or 0--99).  In
##' `parse_date_time()` also matches year with century (Y format).}
##'
##' \item{`Y` (!)}{Year with century.}
##'
##' \item{`z` (!*)}{ISO8601 signed offset in hours and minutes from UTC. For
##' example `-0800`, `-08:00` or `-08`, all represent 8 hours
##' behind UTC. This format also matches the Z (Zulu) UTC indicator. Because
##' [base::strptime()] doesn't fully support ISO8601 this format is implemented as an
##' union of 4 orders: Ou (Z), Oz (-0800), OO (-08:00) and Oo (-08). You can use
##' these four orders as any other but it is rarely
##' necessary. `parse_date_time2()` and `fast_strptime()` support all of
##' the timezone formats.}
##'
##' \item{`Om` (!*)}{Matches numeric month and English alphabetic months
##'                    (Both, long and abbreviated forms).}
##'
##' \item{`Op` (!*)}{Matches AM/PM English indicator.}
##'
##' \item{`r` (*)}{Matches `Ip` and `H` orders.}
##'
##' \item{`R` (*)}{Matches `HM` and`IMp` orders.}
##'
##' \item{`T` (*)}{Matches `IMSp`, `HMS`, and `HMOS` orders.}
##' }
##'
##'
##' @export
##' @param x a character or numeric vector of dates
##' @param orders a character vector of date-time formats. Each order string is a
##'   series of formatting characters as listed in [base::strptime()] but might not
##'   include the `"%"` prefix. For example, "ymd" will match all the possible
##'   dates in year, month, day order. Formatting orders might include
##'   arbitrary separators. These are discarded. See details for implemented
##'   formats.
##' @param tz a character string that specifies the time zone with which to
##'   parse the dates
##' @param truncated integer, number of formats that can be missing. The most
##'   common type of irregularity in date-time data is the truncation due to
##'   rounding or unavailability of the time stamp. If the `truncated` parameter
##'   is non-zero `parse_date_time()` also checks for truncated formats. For
##'   example,  if the format order is "ymdHMS" and `truncated = 3`,
##'   `parse_date_time()` will correctly parse incomplete date-times like
##'   `2012-06-01 12:23`, `2012-06-01 12` and `2012-06-01`. \bold{NOTE:} The `ymd()`
##'   family of functions is based on [base::strptime()] which currently fails to parse
##'   `%Y-%m` formats.
##' @param quiet logical. If `TRUE`, progress messages are not printed, and
##'   `No formats found` error is suppressed and the function simply returns a
##'   vector of NAs.  This mirrors the behavior of base R functions [base::strptime()]
##'   and [base::as.POSIXct()].
##' @param locale locale to be used, see \link{locales}. On Linux systems you
##'   can use `system("locale -a")` to list all the installed locales.
##' @param select_formats A function to select actual formats for parsing from a
##'   set of formats which matched a training subset of `x`. It receives a named
##'   integer vector and returns a character vector of selected formats. Names
##'   of the input vector are formats (not orders) that matched the training
##'   set. Numeric values are the number of dates (in the training set) that
##'   matched the corresponding format. You should use this argument if the
##'   default selection method fails to select the formats in the right
##'   order. By default the formats with most formatting tokens (`%`) are
##'   selected and `%Y` counts as 2.5 tokens (so that it has a priority over
##'   `%y%m`). See examples.
##' @param exact logical. If `TRUE`, the `orders` parameter is interpreted as an
##'   exact [base::strptime()] format and no training or guessing are performed
##'   (i.e. `train`, `drop` parameters are irrelevant).
##' @param train logical, default `TRUE`. Whether to train formats on a subset of
##'   the input vector. The result of this is that supplied orders are sorted
##'   according to performance on this training set, which commonly results in
##'   increased performance. Please note that even when `train = FALSE` (and
##'   `exact = FALSE`) guessing of the actual formats is still performed on a
##'   pseudo-random subset of the original input vector. This might result in
##'   `All formats failed to parse` error. See notes below.
##' @param drop logical, default `FALSE`. Whether to drop formats that didn't
##'   match on the training set. If `FALSE`, unmatched on the training set formats
##'   are tried as a last resort at the end of the parsing queue. Applies only
##'   when `train = TRUE`. Setting this parameter to `TRUE` might slightly speed up
##'   parsing in situations involving many formats. Prior to v1.7.0 this
##'   parameter was implicitly `TRUE`, which resulted in occasional surprising
##'   behavior when rare patterns where not present in the training set.
##' @return a vector of POSIXct date-time objects
##' @seealso [base::strptime()], [ymd()], [ymd_hms()]
##' @keywords chron
##' @note `parse_date_time()` (and the derivatives `ymd()`, `ymd_hms()`, etc.)
##'   relies on a sparse guesser that takes at most 501 elements from the supplied
##'   character vector in order to identify appropriate formats from the
##'   supplied orders. If you get the error `All formats failed to parse` and
##'   you are confident that your vector contains valid dates, you should either
##'   set `exact` argument to `TRUE` or use functions that don't perform format
##'   guessing (`fast_strptime()`, `parse_date_time2()` or [base::strptime()]).
##' @note For performance reasons, when timezone is not UTC,
##'   `parse_date_time2()` and `fast_strptime()` perform no validity checks for
##'   daylight savings time. Thus, if your input string contains an invalid date
##'   time which falls into DST gap and `lt = TRUE` you will get an `POSIXlt`
##'   object with a non-existent time. If `lt = FALSE` your time instant will be
##'   adjusted to a valid time by adding an hour. See examples. If you want to
##'   get NA for invalid date-times use [fit_to_timeline()] explicitly.
##'
##' @examples
##'
##' ## ** orders are much easier to write **
##' x <- c("09-01-01", "09-01-02", "09-01-03")
##' parse_date_time(x, "ymd")
##' parse_date_time(x, "y m d")
##' parse_date_time(x, "%y%m%d")
##' #  "2009-01-01 UTC" "2009-01-02 UTC" "2009-01-03 UTC"
##'
##' ## ** heterogeneous date-times **
##' x <- c("09-01-01", "090102", "09-01 03", "09-01-03 12:02")
##' parse_date_time(x, c("ymd", "ymd HM"))
##'
##' ## ** different ymd orders **
##' x <- c("2009-01-01", "02022010", "02-02-2010")
##' parse_date_time(x, c("dmY", "ymd"))
##' ##  "2009-01-01 UTC" "2010-02-02 UTC" "2010-02-02 UTC"
##'
##' ## ** truncated time-dates **
##' x <- c("2011-12-31 12:59:59", "2010-01-01 12:11", "2010-01-01 12", "2010-01-01")
##' parse_date_time(x, "Ymd HMS", truncated = 3)
##'
##' ## ** specifying exact formats and avoiding training and guessing **
##' parse_date_time(x, c("%m-%d-%y", "%m%d%y", "%m-%d-%y %H:%M"), exact = TRUE)
##' parse_date_time(c('12/17/1996 04:00:00','4/18/1950 0130'),
##'                 c('%m/%d/%Y %I:%M:%S','%m/%d/%Y %H%M'), exact = TRUE)
##'
##' ## ** quarters and partial dates **
##' parse_date_time(c("2016.2", "2016-04"), orders = "Yq")
##' parse_date_time(c("2016", "2016-04"), orders = c("Y", "Ym"))
##'
##' ## ** fast parsing **
##' \dontrun{
##'   options(digits.secs = 3)
##'   ## random times between 1400 and 3000
##'   tt <- as.character(.POSIXct(runif(1000, -17987443200, 32503680000)))
##'   tt <- rep.int(tt, 1000)
##'
##'   system.time(out <- as.POSIXct(tt, tz = "UTC"))
##'   system.time(out1 <- ymd_hms(tt)) # constant overhead on long vectors
##'   system.time(out2 <- parse_date_time2(tt, "YmdHMOS"))
##'   system.time(out3 <- fast_strptime(tt, "%Y-%m-%d %H:%M:%OS"))
##'
##'   all.equal(out, out1)
##'   all.equal(out, out2)
##'   all.equal(out, out3)
##' }
##'
##' ## ** how to use `select_formats` argument **
##' ## By default %Y has precedence:
##' parse_date_time(c("27-09-13", "27-09-2013"), "dmy")
##'
##' ## to give priority to %y format, define your own select_format function:
##'
##' my_select <-   function(trained, drop=FALSE, ...){
##'    n_fmts <- nchar(gsub("[^%]", "", names(trained))) + grepl("%y", names(trained))*1.5
##'    names(trained[ which.max(n_fmts) ])
##' }
##'
##' parse_date_time(c("27-09-13", "27-09-2013"), "dmy", select_formats = my_select)
##'
##' ## ** invalid times with "fast" parsing **
##' parse_date_time("2010-03-14 02:05:06",  "YmdHMS", tz = "America/New_York")
##' parse_date_time2("2010-03-14 02:05:06",  "YmdHMS", tz = "America/New_York")
##' parse_date_time2("2010-03-14 02:05:06",  "YmdHMS", tz = "America/New_York", lt = TRUE)
parse_date_time <- function(x, orders, tz = "UTC", truncated = 0, quiet = FALSE,
                            locale = Sys.getlocale("LC_TIME"), select_formats = .select_formats,
                            exact = FALSE, train = TRUE, drop = FALSE) {

  ## backward compatible hack
  if (is.null(tz)) tz <- ""
  if (length(tz) != 1 || is.na(tz))
    stop("`tz` argument must be a character of length one")

  orig_locale <- Sys.getlocale("LC_TIME")
  Sys.setlocale("LC_TIME", locale)
  on.exit(Sys.setlocale("LC_TIME", orig_locale))

  x <- as.character(.num_to_date(x))
  if (truncated != 0)
    orders <- .add_truncated(orders, truncated)

  .local_parse <- function(x, first = FALSE) {
    formats <-
      if (exact) {
        orders
      } else {
        train <- .get_train_set(x)
        .best_formats(train, orders, locale = locale, select_formats, drop = drop)
      }
    if (length(formats) > 0) {
      out <- .parse_date_time(x, formats, tz = tz, quiet = quiet, locale = locale)
      new_na <- is.na(out)
      if (any(new_na)) {
        x <- x[new_na]
        if (length(x) == length(out)) {
          # don't recur if failed for all
          failed <<- length(x)
        } else {
          out[new_na] <- .local_parse(x)
        }
      }
      out
    } else {
      if (first && !quiet) {
        warning("All formats failed to parse. No formats found.", call. = FALSE)
        warned <<- TRUE
      }
      failed <<- length(x)
      NA
    }
  }

  failed <- 0L
  warned <- FALSE
  to_parse <- which(!is.na(x) & nzchar(x)) ## missing data might be ""
  ## prepare an NA vector
  out <- .POSIXct(rep.int(NA_real_, length(x)), tz = tz)

  if (length(to_parse)) {
    out[to_parse] <- .local_parse(x[to_parse], TRUE)
    if (failed > 0 && !quiet && !warned)
      warning(" ", failed, " failed to parse.", call. = FALSE)
  }

  out
}

parse_dt <- function(x, orders, is_format = FALSE, return_lt = FALSE, cutoff_2000 = 68L) {
  .Call(C_parse_dt, x, orders, as.logical(is_format), as.logical(return_lt), as.integer(cutoff_2000))
}

##' @description
##' `parse_date_time2()` is a fast C parser of numeric
##' orders.
##'
##' @rdname parse_date_time
##' @export
##' @param lt logical. If `TRUE`, returned object is of class POSIXlt, and POSIXct
##'   otherwise. For compatibility with [base::strptime()] the default is `TRUE`
##'   for `fast_strptime()` and `FALSE` for `parse_date_time2()`.
##' @param cutoff_2000 integer. For `y` format,  two-digit numbers smaller or equal to
##'   `cutoff_2000` are parsed as 20th's century, 19th's otherwise. Available only
##'   for functions relying on `lubridate`s internal parser.
parse_date_time2 <- function(x, orders, tz = "UTC", exact = FALSE, lt = FALSE, cutoff_2000 = 68L){
  if (length(tz) != 1 || is.na(tz))
    stop("`tz` argument must be a character of length one")
  if (length(orders) > 1)
    warning("Multiple orders supplied. Only first order is used.")
  if (!exact)
    orders <- gsub("[^[:alpha:]]+", "", as.character(orders[[1]])) ## remove all separators
  if (lt) {
    .mklt(parse_dt(x, orders, exact, TRUE, cutoff_2000), tz)
  } else {
    if (is_utc(tz)){
      .POSIXct(parse_dt(x, orders, exact, FALSE, cutoff_2000), tz = "UTC")
    } else {
      as.POSIXct(.mklt(parse_dt(x, orders, exact, TRUE, cutoff_2000), tz))
    }
  }
}

##' @description
##' `fast_strptime()` is a fast C parser of numeric formats only
##' that accepts explicit format arguments, just like
##' [base::strptime()].
##' @rdname parse_date_time
##' @export
##' @param format a character string of formats. It should include all the
##'   separators and each format must be prefixed with %, just as in the format
##'   argument of [base::strptime()].
fast_strptime <- function(x, format, tz = "UTC", lt = TRUE, cutoff_2000 = 68L) {
  if (length(format) > 1)
    warning("Multiple formats supplied. Only first format is used.")
  format <- as.character(format[[1]])
  if (lt) {
    .mklt(parse_dt(x, format, TRUE, TRUE, cutoff_2000), tz)
  } else{
    if (is_utc(tz)) {
      .POSIXct(parse_dt(x, format, TRUE, FALSE, cutoff_2000), "UTC")
    } else {
      as.POSIXct(.mklt(parse_dt(x, format, TRUE, TRUE, cutoff_2000), tz))
    }
  }
}



### INTERNAL
.mklt <- function(dtlist, tz){
  na_fill <- rep_len(NA_integer_, length(dtlist$sec))
  dtlist[["wday"]] <- na_fill
  dtlist[["yday"]] <- na_fill
  dtlist[["isdst"]] <- -1L
  .POSIXlt(dtlist, tz = tz)
}

.parse_date_time <- function(x, formats, tz, quiet, locale){

  ## print(formats) # for debugging

  out <- .strptime(x, formats[[1]], tz = tz, quiet = quiet, locale = locale)
  na <- is.na(out)
  newx <- x[na]

  if (is_verbose())
    message(" ", sum(!na) , " parsed with ", gsub("^@|@$", "", formats[[1]]))

  ## recursive parsing
  if (length(formats) > 1 && length(newx) > 0)
    out[na] <- .parse_date_time(newx, formats[-1], tz = tz, quiet = quiet, locale = locale)

  ## return POSIXlt
  out
}

.strptime <- function(x, fmt, tz = "UTC", quiet = FALSE, locale = NULL){

  ## Depending on fmt we might need to preprocess x.
  ## ISO8601 and internal parser are the only cases so far.
  ## %Ou: "2013-04-16T04:59:59Z"
  ## %Oo: "2013-04-16T04:59:59+01"
  ## %Oz: "2013-04-16T04:59:59+0100"
  ## %OO: "2013-04-16T04:59:59+01:00"

  ## is_posix <-  0 < regexpr("^[^%]*%Y[^%]+%m[^%]+%d[^%]+(%H[^%](%M[^%](%S)?)?)?[^%Z]*$", fmt)

  c_parser <- 0 < regexpr("^[^%0-9]*(%([YymdqIHMSz]|O[SzuoOpmb])[^%0-9Z]*)+$", fmt)
  zpos <- regexpr("%O((?<z>z)|(?<u>u)|(?<o>o)|(?<O>O))", fmt, perl = TRUE)

  if (c_parser) {
    ## C PARSER:
    out <- fast_strptime(x, fmt, tz = "UTC", lt = FALSE)

    if (!is_utc(tz)) {
      out <-
        if (zpos > 0){
          if (!quiet)
            message("Date in ISO8601 format; converting timezone from UTC to \"", tz,  "\".")
          with_tz(out, tzone = tz)
        } else {
          ## force_tz is very slow
          force_tz(out, tzone = tz)
        }
    }

    out

  } else {
    ## STRPTIME PARSER:

    ## strptime doesn't accept 'locale' argument; need a hard reset
    if (!is.null(locale)) {
      old_lc_time <- Sys.getlocale("LC_TIME")
      if (old_lc_time != locale){
        Sys.setlocale("LC_TIME", locale)
        on.exit(Sys.setlocale("LC_TIME", old_lc_time))
      }
    }

    if (zpos > 0){
      ## If ISO8601 -> pre-process x and fmt
      capt <- attr(zpos, "capture.names")[attr(zpos, "capture.start") > 0][[2]] ## <- second subexp
      repl <- switch(capt,
                     z = "%z",
                     u ="Z",
                     ## substitute +aa with +aa00
                     o = { x <- sub("([+-]\\d{2}(?=\\D+)?$)", "\\100", x, perl = TRUE)
                           "%z"},
                     ## substitute +aa:bb with +aabb
                     O = { x <- sub("([+-]\\d{2}):(?=[^:]+$)", "\\1", x, perl = TRUE)
                           "%z"},
                     stop("Unrecognised capture detected; please report this bug"))

      fmt <- .str_sub(fmt, zpos, zpos + attr(zpos, "match.length") - 1, repl)

      ## user has supplied tz argument -> convert to tz
      if (!is_utc(tz)){
        if (!quiet)
          message("Date in ISO8601 format; converting timezone from UTC to \"", tz,  "\".")
        return(with_tz(strptime(.enclose(x), .enclose(fmt), "UTC"), tzone = tz))
      }
    }

    strptime(.enclose(x), .enclose(fmt), tz)
  }
}


## Expand format strings to also include truncated formats
## Get locations of letters as vector
## Choose the number at the n - truncated place in the vector
## return the substring created by 1 to tat number.
.add_truncated <- function(orders, truncated){
  out <- orders

  if (truncated > 0) {
    trunc_one <- function(order) {
      alphas <- gregexpr("[a-zA-Z]", order)[[1]]
      start <- max(0, length(alphas) - truncated)
      cut_points <- alphas[start:(length(alphas)-1)]

      truncs <- c()
      for (j in seq_along(cut_points))
        truncs <- c(truncs, substr(order, 1, cut_points[j]))
      truncs
    }

  } else {
    trunc_one <- function(order) {
      alphas <- gregexpr("[a-zA-Z]", order)[[1]][-1]
      end <- max(1, abs(truncated) - 1)
      cut_points <- alphas[1:end]

      truncs <- c()
      for (j in seq_along(cut_points))
        truncs <- c(truncs, substr(order, cut_points[j], nchar(order)))
      truncs
    }
  }

  for (i in seq_along(orders)) {
    out <- c(out, trunc_one(orders[i]))
  }

  out
}


.xxx_hms_truncations <- list(T = c("R", "r", ""), R = c("r", ""), r = "")

.parse_xxx_hms <- function(..., orders, truncated, quiet, tz, locale){
  ## !!! NOTE: truncated operates on first element in ORDERS !
  if (truncated > 0){
    ## Take first 3 formats and append formats from .xxx_hms_truncations
    ## co responding to the 4th format letter in order[[1]] -- T, R or r.
    xxx <- substr(orders[[1]], 1, 3) ##
    add <- paste(xxx, .xxx_hms_truncations[[substr(orders[[1]], 4, 4)]], sep = "")
    rest <- length(add) - truncated
    if (rest  < 0)
      orders <- c(orders, add, .add_truncated(xxx, abs(rest)))
    else
      orders <- c(orders, add[1:truncated])
  }
  dates <- unlist(lapply(list(...), .num_to_date), use.names = FALSE)
  parse_date_time(dates, orders, tz = tz, quiet = quiet, locale = locale)
}

.parse_xxx <- function(..., orders, quiet, tz, locale,  truncated){
  dates <- unlist(lapply(list(...), .num_to_date), use.names = FALSE)
  if (is.null(tz)) {
    as.Date.POSIXct(parse_date_time(dates, orders, quiet = quiet, tz = "UTC",
                                    locale = locale, truncated = truncated))
  } else {
    parse_date_time(dates, orders, quiet = quiet, tz = tz,
                    locale = locale, truncated = truncated)
  }
}

.num_to_date <- function(x) {
  if (is.numeric(x)) {
    out <- rep.int(as.character(NA), length(x))
    nnas <- !is.na(x)
    x <- format(x[nnas], scientific = FALSE, trim = TRUE)
    x <- paste(ifelse(nchar(x) %% 2 == 1, "0", ""), x, sep = "")
    out[nnas] <- x
    out
  }else as.character(x)
}

.parse_iso_dt <- function(x, tz) {
  parse_date_time(x, orders = c("ymdTz", "ymdT", "ymd"), tz = tz, train = FALSE)
}

as_POSIXct <- function(x, tz) {
  if (is.character(x))
    .parse_iso_dt(x, tz = tz)
  else if (!is.POSIXct(x))
    as.POSIXct(x, tz = tz)
  else x
}
## parse.r ends here

#' Changes the components of a date object
#'
#' `update.Date()` and `update.POSIXt()` return a date with the specified
#' elements updated.  Elements not specified will be left unaltered. update.Date
#' and update.POSIXt do not add the specified values to the existing date, they
#' substitute them for the appropriate parts of the existing date.
#'
#'
#' @name DateTimeUpdate
#' @param object a date-time object
#' @param ... named arguments: years, months, ydays, wdays, mdays, days, hours, minutes,
#'   seconds, tzs (time zone component)
#' @param week_start week start day (Default is 7, Sunday. Set `lubridate.week.start` to
#'   override). Full or abbreviated names of the days of the week can be in English or
#'   as provided by the current locale.
#' @param simple,roll deprecated
#' @return a date object with the requested elements updated. The object will retain its
#'   original class unless an element is updated which the original class does not
#'   support. In this case, the date returned will be a POSIXlt date object.
#' @inheritParams timechange::time_add
#' @keywords manip chron
#' @examples
#' date <- ymd("2009-02-10")
#' update(date, year = 2010, month = 1, mday = 1)
#'
#' update(date, year = 2010, month = 13, mday = 1)
#'
#' update(date, minute = 10, second = 3)
#' @export
update.POSIXt <- function(object, ...,
                          roll_dst = c("NA", "post"),
                          week_start = getOption("lubridate.week.start", 7),
                          roll = NULL, simple = NULL) {
  if (!is.null(simple)) roll <- simple
  do.call(update_datetime, c(
    list(object, week_start = week_start, roll_dst = roll_dst, roll = roll),
    list(...)
  ))
}

update_datetime <- function(object, years = NULL, months = NULL,
                            days = NULL, mdays = NULL, ydays = NULL, wdays = NULL,
                            hours = NULL, minutes = NULL, seconds = NULL, tzs = NULL,
                            roll_month = "full",
                            roll_dst = c("NA", "post"),
                            roll = NULL,
                            week_start = 7,
                            exact = FALSE) {
  roll_dst <- normalize_roll_dst(roll_dst, roll, 1)
  week_start <- as_week_start(week_start)

  if (!is.null(days)) {
    if (!is.null(mdays)) {
      stop("Only one of `days` and `mdays` must be supplied")
    }
    mdays <- days
  }

  updates <- list(
    year = years, month = months,
    mday = mdays, wday = wdays, yday = ydays,
    hour = hours, minute = minutes, second = seconds
  )
  timechange::time_update(
    object, updates = normalize_units_length(updates), tz = tzs,
    roll_dst = roll_dst, roll_month = roll_month,
    week_start = as_week_start(week_start),
    exact = exact
  )
}

normalize_units_length <- function(units) {
  if (length(units) == 0)
    return(units)
  maxlen <- max(unlist(lapply(units, length)))
  if (maxlen > 1) {
    for (nm in names(units)) {
      len <- length(units[[nm]])
      if (len != maxlen && len > 1)
        units[[nm]] <- rep_len(units[[nm]], maxlen)
    }
  }
  units
}

as_week_start <- function(x) {
  if (is.numeric(x)) {
    if (x > 7 || x < 1) {
      stop("Invalid 'week_start' argument; must be between 1 and 7")
    }
    return(x)
  }

  if (length(x) != 1L)
    stop("Invalid `week_start` argument; must be of length 1")

  if (!is.character(x))
    stop("Invalid `week_start` argument; must be a number in 1-7 (Monday based) or string day of the week", call. = FALSE)

  # Use pmatch to also capture abbreviations.
  english_idx <- pmatch(x, c("Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"))
  if (!is.na(english_idx)) {
    return(english_idx)
  }

  # Other locales might use abbreviations that are not so pmatch-friendly; try full first, then partial.
  native_full_idx <- match(x, .get_locale_regs()$wday_names$full)
  if (!is.na(native_full_idx)) {
    return(native_full_idx)
  }

  native_abbr_idx <- match(x, .get_locale_regs()$wday_names$abr)
  if (!is.na(native_abbr_idx)) {
    return(native_abbr_idx)
  }
}

## prior to v1.7.0
update_posixt_old <- function(object, ..., simple = FALSE) {
  if (!length(object)) {
    return(object)
  }
  date <- as.POSIXlt(object)

  ## adjudicate units input
  units <- list(...)
  names(units) <- standardise_lt_names(names(units))

  if (!is.null(new_tz <- units$tz)) {
    units$tz <- NULL
  }

  day.units <- c("day", "wday", "mday", "yday")
  wunit <- day.units %in% names(units)

  if (n <- sum(wunit)) {
    if (n > 1) stop("conflicting days input")
    uname <- day.units[wunit]
    if (uname != "mday") {
      ## we compute everything with mdays (operating with ydays doesn't work)
      if (uname != "day") {
        if (uname == "yday" & !is.null(units$year)) {
          warning("Updating on both 'year' and 'yday' can lead to wrong results. See bug #319.", call. = F)
        }
        diff <- units[[uname]] - date[[uname]] - 1
        units[[uname]] <- diff + date$mday
      }
      names(units)[names(units) == uname] <- "mday"
    }
  }

  if (!is.null(units$mon)) units$mon <- units$mon - 1
  if (!is.null(units$year)) units$year <- units$year - 1900

  ## make new date-times
  date <- unclass(date)

  date[names(units)] <- units
  date[c("wday", "yday")] <- list(wday = NA, yday = NA)

  ## unbalanced POSIXlt often results in R crashes
  maxlen <- max(unlist(lapply(date, length)))

  if (maxlen > 1) {
    for (nm in names(date)) {
      if (length(date[[nm]]) != maxlen) {
        date[[nm]] <- rep_len(date[[nm]], maxlen)
      }
    }
  }

  class(date) <- c("POSIXlt", "POSIXt")
  if (!is.null(new_tz)) {
    attr(date, "tzone") <- new_tz
  }

  ## fit to timeline
  ## POSIXct format avoids negative and NA elements in POSIXlt format
  fit_to_timeline(date, class(object)[[1]], simple = simple)
}

is_zero_hms <- function(hours = NULL, minutes = NULL, seconds = NULL, tzs = NULL, ...) {
  is.null(tzs) &&
    (is.null(hours) || sum(hours, na.rm = TRUE) == 0) &&
    (is.null(minutes) || sum(minutes, na.rm = TRUE) == 0) &&
    (is.null(seconds) || sum(seconds, na.rm = TRUE) == 0)
}

#' @export
update.Date <- function(object, ...) {
  out <- update_datetime(object, ...)
  ## lubridate's missing hms component checks for actual zeros. Not a great idea as the
  ## return type is unpredictable but we are stuck with it.
  if (is.POSIXct(out) && is_zero_hms(...))
    as_date(out)
  else
    out
}

#' Fit a POSIXlt date-time to the timeline
#'
#' The POSIXlt format allows you to create instants that do not exist in real
#' life due to daylight savings time and other conventions. fit_to_timeline
#' matches POSIXlt date-times to a real times. If an instant does not exist, fit
#' to timeline will replace it with an NA. If an instant does exist, but has
#' been paired with an incorrect timezone/daylight savings time combination,
#' fit_to_timeline returns the instant with the correct combination.
#'
#'
#' @param lt a POSIXlt date-time object.
#' @param class a character string that describes what type of object to return,
#'   POSIXlt or POSIXct. Defaults to POSIXct. This is an optimization to avoid
#'   needless conversions.
#' @param simple if TRUE, \pkg{lubridate} makes no attempt to detect
#'   meaningless time-dates or to correct time zones. No NAs are produced and
#'   the most meaningful valid dates are returned instead. See examples.
#' @return a POSIXct or POSIXlt object that contains no illusory date-times
#'
#' @examples
#' \dontrun{
#'
#' tricky <- structure(list(
#'   sec = c(5, 0, 0, -1),
#'   min = c(0L, 5L, 5L, 0L),
#'   hour = c(2L, 0L, 2L, 2L),
#'   mday = c(4L, 4L, 14L, 4L),
#'   mon = c(10L, 10L, 2L, 10L),
#'   year = c(112L, 112L, 110L, 112L),
#'   wday = c(0L, 0L, 0L, 0L),
#'   yday = c(308L, 308L, 72L, 308L),
#'   isdst = c(1L, 0L, 0L, 1L)
#' ),
#' .Names = c(
#'   "sec", "min", "hour", "mday", "mon",
#'   "year", "wday", "yday", "isdst"
#' ),
#' class = c("POSIXlt", "POSIXt"),
#' tzone = c("America/Chicago", "CST", "CDT")
#' )
#'
#' tricky
#' ## [1] "2012-11-04 02:00:00 CDT" Doesn't exist because clocks "fall back" to 1:00 CST
#' ## [2] "2012-11-04 00:05:00 CST" Times are still CDT, not CST at this instant
#' ## [3] "2010-03-14 02:00:00 CDT" DST gap
#' ## [4] "2012-11-04 01:59:59 CDT" Does exist, but has deceptive internal structure
#'
#' fit_to_timeline(tricky)
#' ## Returns:
#' ## [1] "2012-11-04 02:00:00 CST" instant paired with correct tz & DST combination
#' ## [2] "2012-11-04 00:05:00 CDT" instant paired with correct tz & DST combination
#' ## [3] NA - fake time changed to NA (compare to as.POSIXct(tricky))
#' ## [4] "2012-11-04 01:59:59 CDT" -real instant, left as is
#'
#' fit_to_timeline(tricky, simple = TRUE)
#' ## Returns valid time-dates by extrapolating CDT and CST zones:
#' ## [1] "2012-11-04 01:00:05 CST" "2012-11-04 01:05:00 CDT"
#' ## [3] "2010-03-14 03:05:00 CDT" "2012-11-04 01:59:59 CDT"
#' }
#' @export
fit_to_timeline <- function(lt, class = "POSIXct", simple = FALSE) {
  if (class != "POSIXlt" && class != "POSIXct") {
    stop("class argument must be POSIXlt or POSIXct")
  }

  if (simple) {
    if (class == "POSIXct") {
      as.POSIXct(lt)
    } else {
      as.POSIXlt(as.POSIXct(lt))
    }
  } else {

    ## fall break - DST only changes if it has to
    ct <- as.POSIXct(lt)
    lt2 <- as.POSIXlt(ct)

    dstdiff <- !is.na(ct) & (lt$isdst != lt2$isdst)
    if (any(dstdiff)) {
      dlt <- lt[dstdiff]
      dlt2 <- lt2[dstdiff]
      dlt$isdst <- dlt2$isdst
      dlt$zone <- dlt2$zone
      dlt$gmtoff <- dlt2$gmtoff
      dct <- as.POSIXct(dlt) # should directly match if not in gap

      if (class == "POSIXct") {
        ct[dstdiff] <- dct
      } else {
        lt2[dstdiff] <- dlt
      }

      chours <- format.POSIXlt(as.POSIXlt(dct), "%H", usetz = FALSE)
      lhours <- format.POSIXlt(dlt, "%H", usetz = FALSE)

      any <- any(hdiff <- chours != lhours)
      if (!is.na(any) && any) {
        if (class == "POSIXct") {
          ct[dstdiff][hdiff] <- NA
        } else {
          lt2[dstdiff][hdiff] <- NA
        }
      }
    }
    if (class == "POSIXct") ct else lt2
  }
}

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
#' @include Dates.r
#' @include difftimes.r
#' @include numeric.r
#' @include POSIXt.r
#' @include time-zones.r
NULL


#' Convert a variety of date-time classes to POSIXlt and POSIXct
#' @name DateCoercion
#' @keywords internal
#'
NULL

#' @export
as.POSIXct.fts <- function(x, tz = "", ...) as.POSIXct(zoo::index(x))
#' @export
as.POSIXlt.fts <- function(x, tz = "", ...) as.POSIXlt(zoo::index(x))

## #' @export
## as.POSIXlt.its <- function(x, tz = "", ...) as.POSIXlt(attr(x, "dates"))
## #' @export
## as.POSIXct.its <- function(x, tz = "", ...) as.POSIXct(attr(x, "dates"))

#' @export
as.POSIXlt.timeSeries <- function(x, tz = "", ...) {
  as.POSIXlt(timeDate::timeDate(x@positions,
    zone = x@FinCenter, FinCenter = x@FinCenter
  ))
}
#' @export
as.POSIXct.timeSeries <- function(x, tz = "", ...) {
  as.POSIXct(timeDate::timeDate(x@positions,
    zone = x@FinCenter, FinCenter = x@FinCenter
  ))
}

#' @export
as.POSIXlt.irts <- function(x, tz = "", ...) as.POSIXlt(x$time)
#' @export
as.POSIXct.irts <- function(x, tz = "", ...) as.POSIXct(x$time)

#' @export
as.POSIXlt.xts <- function(x, tz = "", ...) as.POSIXlt(zoo::index(x))
#' @export
as.POSIXct.xts <- function(x, tz = "", ...) as.POSIXct(zoo::index(x))
#' @export
as.POSIXlt.zoo <- function(x, tz = "", ...) as.POSIXlt(zoo::index(x))
#' @export
as.POSIXct.zoo <- function(x, tz = "", ...) as.POSIXct(zoo::index(x))

#' @export
as.POSIXlt.tis <- function(x, tz = "", ...) as.Date(x)

#' Convenience method to reclass dates post-modification.
#' @keywords internal
#'
#' @export
reclass_date <- function(new, orig) UseMethod("reclass_date", orig)
#' @export
reclass_date.POSIXlt <- function(new, orig) {
  as.POSIXlt(new)
}
#' @export
reclass_date.POSIXct <- function(new, orig) {
  as.POSIXct(new)
}
#' @export
reclass_date.chron <- function(new, orig) {
  chron::as.chron(new)
}
#' @export
reclass_date.timeDate <- function(new, orig) {
  timeDate::as.timeDate(new)
}
## #' @export
## reclass_date.its <- function(new, orig) {
##   its::its(new, format = "%Y-%m-%d %X")
## }
#' @export
reclass_date.ti <- function(new, orig) {
  tis::as.ti(new, tis::tifName(orig))
}
#' @export
reclass_date.Date <- function(new, orig) {
  as_date(new)
}
#' @export
reclass_date.IDate <- function(new, orig) {
  data.table::as.IDate(new)
}
#' @export
reclass_date.ITime <- function(new, orig) {
  data.table::as.ITime(new)
}


period_to_difftime <- function(per) {
  as.difftime(per)
}

#' Convenience method to reclass timespans post-modification.
#' @keywords internal
#'
#' @aliases reclass_timespan,ANY,Duration-method reclass_timespan,ANY,Interval-method
#' reclass_timespan,ANY,Period-method reclass_timespan,ANY,difftime-method
#' @export
reclass_timespan <- function(new, orig) standardGeneric("reclass_timespan")

#' @export
setGeneric("reclass_timespan",
  useAsDefault = function(new, orig) {
    stop(sprintf("reclass_timespan is not defined for class '%s'", class(x)))
  }
)

#' @export
setMethod("reclass_timespan", signature(orig = "difftime"), function(new, orig) {
  if (is.period(new)) {
    as.difftime(new)
  } else {
    make_difftime(as.numeric(new))
  }
})

#' @export
setMethod("reclass_timespan", signature(orig = "Duration"), function(new, orig) {
  as.duration(new)
})

#' @export
setMethod("reclass_timespan", signature(orig = "Interval"), function(new, orig) {
  as.duration(new)
})

#' @export
setMethod("reclass_timespan", signature(orig = "Period"), function(new, orig) {
  as.period(new)
})


#' Change an object to a duration
#'
#' as.duration changes Interval, Period and numeric class objects to
#' Duration objects. Numeric objects are changed to Duration objects
#' with the seconds unit equal to the numeric value.
#'
#' Durations are exact time measurements, whereas periods are relative time
#' measurements. See [Period-class]. The length of a period depends
#' on when it occurs. Hence, a one to one mapping does not exist between
#' durations and periods. When used with a period object, as.duration provides
#' an inexact estimate of the length of the period; each time unit is assigned
#' its most common number of seconds. A period of one month is converted to
#' 2628000 seconds (approximately 30.42 days). This ensures that 12 months will
#' sum to 365 days, or one normal year. For an exact transformation, first
#' transform the period to an interval with [as.interval()].
#'
#' @param x Object to be coerced to a duration
#' @param ... Parameters passed to other methods. Currently unused.
#' @return A duration object
#' @seealso [Duration-class], [duration()]
#' @keywords classes manip methods chron
#' @examples
#' span <- interval(ymd("2009-01-01"), ymd("2009-08-01")) # interval
#' as.duration(span)
#' as.duration(10) # numeric
#' dur <- duration(hours = 10, minutes = 6)
#' as.numeric(dur, "hours")
#' as.numeric(dur, "minutes")
#' @aliases as.duration,numeric-method as.duration,logical-method
#'   as.duration,difftime-method as.duration,Interval-method
#'   as.duration,Duration-method as.duration,Period-method
#'   as.duration,character-method
#' @export
setGeneric("as.duration",
  function(x, ...) standardGeneric("as.duration"),
  useAsDefault = function(x, ...) {
    stop(sprintf("as.duration is not defined for class '%s'", class(x)))
  }
)

setMethod("as.duration", signature(x = "character"), function(x) {
  as.duration(as.period(x))
})

setMethod("as.duration", signature(x = "numeric"), function(x) {
  new("Duration", x)
})

setMethod("as.duration", signature(x = "logical"), function(x) {
  new("Duration", as.numeric(x))
})

as_duration_from_difftime <- function(x) {
  new("Duration", as.numeric(x, "secs"))
}
setMethod("as.duration", signature(x = "difftime"), as_duration_from_difftime)

setMethod("as.duration", signature(x = "Interval"), function(x) {
  new("Duration", x@.Data)
})

setMethod("as.duration", signature(x = "Duration"), function(x) {
  x
})

setMethod("as.duration", signature(x = "Period"), function(x) {
  new("Duration", period_to_seconds(x))
})


#' Change an object to an `interval`
#'
#' as.interval changes difftime, Duration, Period and numeric class objects to
#' intervals that begin at the specified date-time. Numeric objects are first
#' coerced to timespans equal to the numeric value in seconds.
#'
#' as.interval can be used to create accurate transformations between Period
#' objects, which measure time spans in variable length units, and Duration objects,
#' which measure timespans as an exact number of seconds. A start date-
#' time must be supplied to make the conversion. Lubridate uses
#' this start date to look up how many seconds each variable
#' length unit (e.g. month, year) lasted for during the time span
#' described. See
#' [as.duration()], [as.period()].
#'
#' @param x a duration, difftime, period, or numeric object that describes the length of the interval
#' @param start a POSIXt or Date object that describes when the interval begins
#' @param ... additional arguments to pass to as.interval
#' @return an interval object
#' @seealso [interval()]
#' @keywords classes manip methods chron
#' @examples
#' diff <- make_difftime(days = 31) # difftime
#' as.interval(diff, ymd("2009-01-01"))
#' as.interval(diff, ymd("2009-02-01"))
#'
#' dur <- duration(days = 31) # duration
#' as.interval(dur, ymd("2009-01-01"))
#' as.interval(dur, ymd("2009-02-01"))
#'
#' per <- period(months = 1) # period
#' as.interval(per, ymd("2009-01-01"))
#' as.interval(per, ymd("2009-02-01"))
#'
#' as.interval(3600, ymd("2009-01-01")) # numeric
#' @aliases as.interval,numeric-method as.interval,difftime-method as.interval,Interval-method as.interval,Duration-method as.interval,Period-method as.interval,POSIXt-method as.interval,logical-method
#' @export
as.interval <- function(x, start, ...) standardGeneric("as.interval")

#' @export
setGeneric("as.interval",
  function(x, start, ...) standardGeneric("as.interval"),
  useAsDefault = function(x, start, ...) {
    stop(sprintf("as.interval is not defined for class '%s'", class(x)))
  }
)

setMethod("as.interval", signature(x = "numeric"), function(x, start, ...) {
  .number_to_interval(x, start, ...)
})

setMethod("as.interval", signature(x = "difftime"), function(x, start, ...) {
  .number_to_interval(x, start, ...)
})

setMethod("as.interval", signature(x = "Interval"), function(x, start, ...) {
  x
})

setMethod("as.interval", signature(x = "POSIXt"), function(x, start, ...) {
  .number_to_interval(x, start, ...)
})

#' @export
setMethod("as.interval", signature("logical"), function(x, start, ...) {
  .number_to_interval(as.numeric(x), start, ...)
})

.number_to_interval <- function(x, start, ...) {
  if (missing(start) & all(is.na(x))) {
    start <- .POSIXct(NA_real_, tz = "UTC")
  } else {
    stopifnot(is.instant(start))
  }

  if (is.instant(x)) {
    return(interval(x, start))
  } else {
    interval(start, start + x)
  }
}


#' Change an object to a period
#'
#' as.period changes Interval, Duration, difftime and numeric class objects
#' to Period class objects with the specified units.
#'
#' Users must specify which time units to measure the period in. The exact length of
#' each time unit in a period will depend on when it occurs. See
#' [Period-class] and [period()].
#' The choice of units is not trivial; units that are
#' normally equal may differ in length depending on when the time period
#' occurs. For example, when a leap second occurs one minute is longer than 60
#' seconds.
#'
#' Because periods do not have a fixed length, they can not be accurately
#' converted to and from Duration objects. Duration objects measure time spans
#' in exact numbers of seconds, see [Duration-class]. Hence, a one to one
#' mapping does not exist between durations and periods. When used with a
#' Duration object, as.period provides an inexact estimate; the duration is
#' broken into time units based on the most common lengths of time units, in
#' seconds. Because the length of months are particularly variable, a period
#' with a months unit can not be coerced from a duration object. For an exact
#' transformation, first transform the duration to an interval with
#' [as.interval()].
#'
#' Coercing an interval to a period may cause surprising behavior if you request
#' periods with small units. A leap year is 366 days long, but one year long. Such
#' an interval will convert to 366 days when unit is set to days and 1 year when
#' unit is set to years. Adding 366 days to a date will often give a different
#' result than adding one year. Daylight savings is the one exception where this
#' does not apply. Interval lengths are calculated on the UTC timeline, which does
#' not use daylight savings. Hence, periods converted with seconds or minutes will not
#' reflect the actual variation in seconds and minutes that occurs due to daylight
#' savings. These periods will show the "naive" change in seconds and minutes that is
#' suggested by the differences in clock time. See the examples below.
#'
#' @param x an interval, difftime, or numeric object
#' @param unit A character string that specifies which time units to build period in.
#' unit is only implemented for the as.period.numeric method and the as.period.interval method.
#' For as.period.interval, as.period will convert intervals to units no larger than the specified
#' unit.
#' @param ... additional arguments to pass to as.period
#' @return a period object
#' @seealso [Period-class], [period()]
#' @keywords classes manip methods chron
#' @examples
#' span <- interval(ymd_hms("2009-01-01 00:00:00"), ymd_hms("2010-02-02 01:01:01")) # interval
#' as.period(span)
#' as.period(span, unit = "day")
#' "397d 1H 1M 1S"
#' leap <- interval(ymd("2016-01-01"), ymd("2017-01-01"))
#' as.period(leap, unit = "days")
#' as.period(leap, unit = "years")
#' dst <- interval(
#'   ymd("2016-11-06", tz = "America/Chicago"),
#'   ymd("2016-11-07", tz = "America/Chicago")
#' )
#' # as.period(dst, unit = "seconds")
#' as.period(dst, unit = "hours")
#' per <- period(hours = 10, minutes = 6)
#' as.numeric(per, "hours")
#' as.numeric(per, "minutes")
#' @aliases as.period,numeric-method as.period,difftime-method
#'   as.period,Interval-method as.period,Duration-method as.period,Period-method
#'   as.period,logical-method as.period,character-method
#' @export
setGeneric("as.period",
  function(x, unit, ...) standardGeneric("as.period"),
  useAsDefault = function(x, unit, ...) {
    stop(sprintf("as.period is not defined for class '%s'", class(x)))
  }
)

setMethod("as.period", signature(x = "character"), function(x, ...) {
  parse_period(x)
})

setMethod("as.period", signature(x = "numeric"), function(x, unit = "second", ...) {
  x <- as.numeric(x)
  if (missing(unit)) unit <- "second"
  unit <- standardise_date_names(unit[[1]])
  f <- get(paste(unit, "s", sep = ""),
    envir = asNamespace("lubridate"),
    mode = "function", inherits = FALSE
  )
  f(x)
})

setMethod("as.period", signature(x = "difftime"), function(x, unit = NULL, ...) {
  seconds_to_period(as.double(x, "secs"))
})

setMethod("as.period", signature(x = "Interval"), function(x, unit = NULL, ...) {
  ## fixme: document this in the manual

  ## SEMANTICS: for postitive intervals all units of the period will be
  ## positive, and the oposite for negatve intervals.

  ## Periods are not symetric in the sense that as.period(int) might not be the
  ## same as -as.period(int_flip(int)). See
  ## https://github.com/tidyverse/lubridate/issues/285 for motivation.

  unit <-
    if (missing(unit)) {
      "year"
    } else {
      standardise_period_names(unit)
    }

  switch(unit,
    year = .int_to_period(x),
    month = {
      pers <- .int_to_period(x)
      month(pers) <- month(pers) + year(pers) * 12
      year(pers) <- 0
      pers
    },
    ## fixme: add note to the docs that unit <= days results in much faster conversion
    ## fixme: add week
    day = ,
    hour = ,
    minute = ,
    second = {
      secs <- x@.Data
      negs <- secs < 0 & !is.na(secs)
      units <- .units_within_seconds(abs(secs), unit)
      pers <- do.call("new", c("Period", units))
      pers[negs] <- -pers[negs]
      pers
    },
    stop("Unsuported unit ", unit)
  )
})

.int_to_period <- function(x) {
  ## this function is called only for conversion with units > day
  start <- as.POSIXlt(x@start)
  end <- unclass(as.POSIXlt(start + x@.Data))
  start <- unclass(start)

  negs <- x@.Data < 0 & !is.na(x@.Data)

  per <- list()

  for (nm in c("sec", "min", "hour", "mday", "mon", "year")) {
    per[[nm]] <- ifelse(negs, start[[nm]] - end[[nm]], end[[nm]] - start[[nm]])
  }

  names(per) <- c("second", "minute", "hour", "day", "month", "year")

  ## Remove negative ...

  ## seconds
  nsecs <- per$second < 0 & !is.na(per$second)
  per$second[nsecs] <- 60 + per$second[nsecs]
  per$minute[nsecs] <- per$minute[nsecs] - 1
  per$second[negs] <- -per$second[negs]

  ## minutes
  nmins <- per$minute < 0 & !is.na(per$minute)
  per$minute[nmins] <- 60 + per$minute[nmins]
  per$hour[nmins] <- per$hour[nmins] - 1
  per$minute[negs] <- -per$minute[negs]

  ## hours
  nhous <- per$hour < 0 & !is.na(per$hour)
  per$hour[nhous] <- 24 + per$hour[nhous]
  per$hour[negs] <- -per$hour[negs]

  ## days

  ### postivie periods
  ndays <- !negs & per$day < 0 & !is.na(per$day)
  if (any(ndays)) {

    ## compute nr days in previous month
    add_months <- rep.int(-1, sum(ndays))

    pmonth <- end$mon[ndays]
    pmonth[pmonth == 0] <- 1 # dec == jan == 31 days
    prev_month_days <- .days_in_month(pmonth, end$year[ndays])

    ## difference in days:
    ## /need pmax to capture as.period(interval(ymd("1985-01-31"), ymd("1986-03-28")))/
    per$day[ndays] <- pmax(prev_month_days - start$mday[ndays], 0) + end$mday[ndays]
    per$month[ndays] <- per$month[ndays] + add_months
  }

  ## negative periods
  ndays <- negs & per$day < 0 & !is.na(per$day)
  if (any(ndays)) {
    add_months <- rep.int(1, sum(ndays))
    this_month_days <- .days_in_month(end$mon[ndays] + 1, end$year[ndays])

    ## Compute nr of days:
    ## /need pmax to capture as.period(interval(ymd("1985-01-31"), ymd("1986-03-28")))/
    per$day[ndays] <- pmax(this_month_days - end$mday[ndays], 0) + start$mday[ndays]
    per$month[ndays] <- per$month[ndays] - add_months
  }

  ## substract only after the day computation to capture intervals like:
  ## as.period(interval(ymd_hms("1985-12-31 5:0:0"), ymd_hms("1986-02-01 3:0:0")))
  per$day[nhous] <- per$day[nhous] - 1
  per$day[negs] <- -per$day[negs]

  ## months
  nmons <- per$month < 0 & !is.na(per$month)
  per$month[nmons] <- 12 + per$month[nmons]
  per$year[nmons] <- per$year[nmons] - 1
  per$month[negs] <- -per$month[negs]

  per$year[negs] <- -per$year[negs]

  new("Period", per$second,
    year = per$year, month = per$month,
    day = per$day, hour = per$hour, minute = per$minute
  )
}

setMethod("as.period", signature(x = "Duration"), function(x, unit = NULL, ...) {
  span <- x@.Data
  remainder <- abs(span)
  newper <- period(second = rep(0, length(x)))

  slot(newper, "year") <- remainder %/% average_durations[["year"]]
  remainder <- remainder %% average_durations[["year"]]

  slot(newper, "day") <- remainder %/% (3600 * 24)
  remainder <- remainder %% (3600 * 24)

  slot(newper, "hour") <- remainder %/% (3600)
  remainder <- remainder %% (3600)

  slot(newper, "minute") <- remainder %/% (60)
  newper$second <- remainder %% (60)

  newper * sign(span)
})

setMethod(
  "as.period", signature("Period"),
  function(x, unit = NULL, ...) {
    if (missing(unit) || is.null(unit)) {
      x
    } else {
      unit <- standardise_period_names(unit)
      switch(unit,
        year = x,
        month = {
          month(x) <- month(x) + year(x) * 12
          year(x) <- 0
          x
        },
        day = ,
        hour = ,
        minute = ,
        second = {
          N <- .units_within_seconds(period_to_seconds(x), unit)
          do.call("new", c("Period", N))
        },
        stop("Unsuported unit ", unit)
      )
    }
  }
)

setMethod("as.period", signature("logical"), function(x, unit = NULL, ...) {
  as.period(as.numeric(x), unit, ...)
})

#' @importFrom generics as.difftime
#' @export
generics::as.difftime

#' @export
setMethod("as.difftime", signature(tim = "Interval"), function(tim, format = "%X", units = "secs") {
  as.difftime(as.numeric(tim, units), format, units)
})

as_difftime_from_duration <- function(tim, format = "%X", units = "secs") {
  as.difftime(tim@.Data, format, units)
}
#' @export
setMethod("as.difftime", signature(tim = "Duration"), as_difftime_from_duration)

#' @export
setMethod("as.difftime", signature(tim = "Period"), function(tim, format = "%X", units = "secs") {
  as.difftime(period_to_seconds(tim), format, units)
})

seconds_to_unit <- function(secs, unit = "second") {
  switch(unit,
    second = secs,
    minute = secs / 60,
    hour   = secs / 3600,
    day    = secs / 86400,
    month  = secs / (86400 * 365.25 / 12),
    week   = secs / (86400 * 7),
    year   = secs / (86400 * 365.25),
    stop("invalid unit ", unit)
  )
}

setGeneric("as.numeric")

#' @export
setMethod("as.numeric", signature("Duration"), function(x, units = "secs", ...) {
  unit <- standardise_period_names(units)
  as.numeric(seconds_to_unit(x@.Data, unit), ...)
})

#' @export
setMethod("as.numeric", signature(x = "Interval"), function(x, units = "secs", ...) {
  as.numeric(as.duration(x), units, ...)
})

#' @export
setMethod("as.numeric", signature(x = "Period"), function(x, units = "second", ...) {
  unit <- standardise_period_names(units)
  as.numeric(seconds_to_unit(period_to_seconds(x), unit = unit), ...)
})

setGeneric("as.integer")

#' @export
setMethod("as.integer", signature("Duration"), function(x, units = "secs", ...) {
  unit <- standardise_period_names(units)
  as.integer(seconds_to_unit(x@.Data, unit), ...)
})

#' @export
setMethod("as.integer", signature(x = "Interval"), function(x, units = "secs", ...) {
  as.integer(as.duration(x), units, ...)
})

#' @export
setMethod("as.integer", signature(x = "Period"), function(x, units = "second", ...) {
  unit <- standardise_period_names(units)
  as.integer(seconds_to_unit(period_to_seconds(x), unit = unit), ...)
})

as.POSIXt <- function(x) as.POSIXlt(x)

## TODO{MM}:  Do not confuse  as.character()  and  format()
## -------   {originally were used interchangably in R for Date and POSIX?t}

#' @export
setMethod("as.character", signature(x = "Period"), function(x, ...) {
  format(x)
})

#' @export
setMethod("as.character", signature(x = "Duration"), function(x, ...) {
  format(x)
})

#' @export
setMethod("as.character", signature(x = "Interval"), function(x, ...) {
  format(x)
})


#' Convert an object to a date or date-time
#'
#'
#' @section Compare to base R:
#'
#' These are drop in replacements for [as.Date()] and [as.POSIXct()], with a few
#' tweaks to make them work more intuitively.
#'
#' \itemize{
#'
#' \item Called on a `POSIXct` object, `as_date()` uses the tzone attribute of
#'   the object to return the same date as indicated by the printed representation
#'   of the object. This differs from as.Date, which ignores the attribute and
#'   uses only the tz argument to `as.Date()` ("UTC" by default).
#'
#' \item Both functions provide a default origin argument for numeric vectors.
#'
#' \item Both functions will generate NAs for invalid date format. Valid formats are
#'   those described by ISO8601 standard. A warning message will provide a count of the
#'   elements that were not converted.
#'
#' \item `as_datetime()` defaults to using UTC.
#'
#' }
#'
#' @param x a vector of [POSIXt], numeric or character objects
#' @param origin a Date object, or something which can be coerced by `as.Date(origin,
#'   ...)` to such an object (default: the Unix epoch of "1970-01-01"). Note that in
#'   this instance, `x` is assumed to reflect the number of days since `origin` at
#'   `"UTC"`.
#' @param tz a time zone name (default: time zone of the POSIXt object `x`). See
#'   [OlsonNames()].
#' @param format format argument for character methods. When supplied parsing is
#'   performed by `parse_date_time(x, orders = formats, exact = TRUE)`. Thus, multiple
#'   formats are supported and are tried in turn.
#' @param ... further arguments to be passed to specific methods (see above).
#' @return a vector of [Date] objects corresponding to `x`.
#' @examples
#' dt_utc <- ymd_hms("2010-08-03 00:50:50")
#' dt_europe <- ymd_hms("2010-08-03 00:50:50", tz = "Europe/London")
#' c(as_date(dt_utc), as.Date(dt_utc))
#' c(as_date(dt_europe), as.Date(dt_europe))
#' ## need not supply origin
#' as_date(10)
#' ## Will replace invalid date format with NA
#' dt_wrong <- c("2009-09-29", "2012-11-29", "2015-29-12")
#' as_date(dt_wrong)
#' @export
setGeneric("as_date",
  def = function(x, ...) standardGeneric("as_date")
)

#' @rdname as_date
#' @export
setMethod("as_date", "ANY",
  function(x, ...) {
    ## From: Kurt Hornik <Kurt.Hornik@wu.ac.at>
    ## Date: Tue, 3 Apr 2018 18:53:19
    ##
    ## `zoo` has its own as.Date for which it registers its yearmon
    ## method (and base::as.Date as the default S3 method).  In fact,
    ## zoo also exports as.Date.yearmon etc, but the above
    ##
    ##    lubridate::as_date(zoo::as.yearmon("2011-01-07"))
    ##
    ## does not attach the zoo exports, hence does not find
    ## as.Date.yearmon on the search path.
    if (inherits(x, c("yearmon", "yearqtr"))) {
      zoo::as.Date(x, ...)
    } else {
      base::as.Date(x, ...)
    }
  }
)

#' @rdname as_date
#' @export
setMethod("as_date", "POSIXt",
  function(x, tz = NULL) {
    tz <- if (is.null(tz)) tz(x) else tz
    as.Date(x, tz = tz)
  }
)

#' @rdname as_date
setMethod("as_date", "numeric",
  function(x, origin = lubridate::origin) {
    as.Date(x, origin = origin)
  }
)

#' @rdname as_date
#' @export
setMethod("as_date", "character",
  function(x, tz = NULL, format = NULL) {
    if (!is.null(tz)) {
      warning("`tz` argument is ignored by `as_date()`", call. = FALSE)
    }
    if (is.null(format)) {
      as_date(.parse_iso_dt(x, tz = "UTC"))
    } else {
      as_date(parse_date_time(x, format, tz = "UTC", exact = TRUE))
    }
  }
)

#' @rdname as_date
#' @export
setGeneric("as_datetime",
  function(x, ...) standardGeneric("as_datetime")
)

#' @rdname as_date
#' @export
setMethod("as_datetime", "ANY",
  function(x, tz = lubridate::tz(x)) {
    with_tz(as.POSIXct(x, tz = tz), tzone = tz)
  }
)

#' @rdname as_date
#' @export
setMethod("as_datetime", "POSIXt",
  function(x, tz = lubridate::tz(x)) {
    with_tz(x, tz)
  }
)

#' @rdname as_date
#' @export
setMethod("as_datetime", "numeric",
  function(x, origin = lubridate::origin, tz = "UTC") {
    as.POSIXct(x, origin = origin, tz = tz)
  }
)

#' @rdname as_date
#' @export
setMethod("as_datetime", "character",
  function(x, tz = "UTC", format = NULL) {
    if (is.null(format)) {
      .parse_iso_dt(x, tz)
    } else {
      parse_date_time(x, format, tz = tz, exact = TRUE)
    }
  }
)

#' @rdname as_date
#' @export
setMethod("as_datetime", "Date",
  function(x, tz = "UTC") {
    dt <- .POSIXct(as.numeric(x) * 86400, tz = "UTC")
    if (is_utc(tz)) {
      return(dt)
    } else {
      force_tz(dt, tzone = tz)
    }
  }
)

#' Dates and times made easy with lubridate
#'
#' Lubridate provides tools that make it easier to parse and
#' manipulate dates. These tools are grouped below by common
#' purpose. More information about each function can be found in
#' its help documentation.
#'
#' @section Parsing dates:
#'
#' Lubridate's parsing functions read strings into R as POSIXct
#' date-time objects. Users should choose the function whose name
#' models the order in which the year ('y'), month ('m') and day
#' ('d') elements appear the string to be parsed:
#' [dmy()], [myd()], [ymd()],
#' [ydm()], [dym()], [mdy()],
#' [ymd_hms()]). A very flexible and user friendly parser
#' is provided by [parse_date_time()].
#'
#' Lubridate can also parse partial dates from strings into
#' [Period-class] objects with the functions
#' [hm()], [hms()] and [ms()].
#'
#' Lubridate has an inbuilt very fast POSIX parser. Most of the [strptime()]
#' formats and various extensions are supported for English locales. See
#' [parse_date_time()] for more details.
#'
#' @section Manipulating dates:
#'
#' Lubridate distinguishes between moments in time (known as
#' [instants()]) and spans of time (known as time spans, see
#' [Timespan-class]). Time spans are further separated into
#' [Duration-class], [Period-class] and
#' [Interval-class] objects.
#'
#' @section Instants:
#'
#' Instants are specific moments of time. Date, POSIXct, and
#' POSIXlt are the three object classes Base R recognizes as
#' instants. [is.Date()] tests whether an object
#' inherits from the Date class. [is.POSIXt()] tests
#' whether an object inherits from the POSIXlt or POSIXct classes.
#' [is.instant()] tests whether an object inherits from
#' any of the three classes.
#'
#' [now()] returns the current system time as a POSIXct
#' object. [today()] returns the current system date.
#' For convenience, 1970-01-01 00:00:00 is saved to
#' [origin]. This is the instant from which POSIXct
#' times are calculated. Try `unclass(now())` to see the numeric structure that
#' underlies POSIXct objects. Each POSIXct object is saved as the number of seconds
#' it occurred after 1970-01-01 00:00:00.
#'
#' Conceptually, instants are a combination of measurements on different units
#' (i.e, years, months, days, etc.). The individual values for
#' these units can be extracted from an instant and set with the
#' accessor functions [second()], [minute()],
#' [hour()], [day()], [yday()],
#' [mday()], [wday()], [week()],
#' [month()], [year()], [tz()],
#' and [dst()].
#' Note: the accessor functions are named after the singular form
#' of an element. They shouldn't be confused with the period
#' helper functions that have the plural form of the units as a
#' name (e.g, [seconds()]).
#'
#' @section Rounding dates:
#'
#' Instants can be rounded to a convenient unit using the
#' functions [ceiling_date()], [floor_date()]
#' and [round_date()].
#'
#' @section Time zones:
#'
#' Lubridate provides two helper functions for working with time
#' zones. [with_tz()] changes the time zone in which an
#' instant is displayed. The clock time displayed for the instant
#' changes, but the moment of time described remains the same.
#' [force_tz()] changes only the time zone element of an
#' instant. The clock time displayed remains the same, but the
#' resulting instant describes a new moment of time.
#'
#' @section Timespans:
#'
#' A timespan is a length of time that may or may not be connected to
#' a particular instant. For example, three months is a timespan. So is an hour and
#' a half. Base R uses difftime class objects to record timespans.
#' However, people are not always consistent in how they expect time to behave.
#' Sometimes the passage of time is a monotone progression of instants that should
#' be as mathematically reliable as the number line. On other occasions time must
#' follow complex conventions and rules so that the clock times we see reflect what
#' we expect to observe in terms of daylight, season, and congruence with the
#' atomic clock. To better navigate the nuances of time, \pkg{lubridate} creates three
#' additional timespan classes, each with its own specific and consistent behavior:
#' [Interval-class], [Period-class] and
#' [Duration-class].
#'
#' [is.difftime()] tests whether an object
#' inherits from the difftime class. [is.timespan()]
#' tests whether an object inherits from any of the four timespan
#' classes.
#'
#' @section Durations:
#'
#' Durations measure the exact amount of time that occurs between two
#' instants. This can create unexpected results in relation to clock times if a
#' leap second, leap year, or change in daylight savings time (DST) occurs in
#' the interval.
#'
#' Functions for working with durations include [is.duration()],
#' [as.duration()] and [duration()]. [dseconds()],
#' [dminutes()], [dhours()],  [ddays()],
#' [dweeks()] and [dyears()] convenient lengths.
#'
#' @section Periods:
#'
#' Periods measure the change in clock time that occurs between two
#' instants. Periods provide robust predictions of clock time in the presence of
#' leap seconds, leap years, and changes in DST.
#'
#' Functions for working with periods include
#' [is.period()], [as.period()] and
#' [period()]. [seconds()],
#' [minutes()], [hours()], [days()],
#' [weeks()], [months()] and
#' [years()] quickly create periods of convenient
#' lengths.
#'
#' @section Intervals:
#'
#' Intervals are timespans that begin at a specific instant and
#' end at a specific instant. Intervals retain complete information about a
#' timespan. They provide the only reliable way to convert between
#' periods and durations.
#'
#' Functions for working with intervals include
#' [is.interval()], [as.interval()],
#' [interval()], [int_shift()],
#' [int_flip()], [int_aligns()],
#' [int_overlaps()], and
#' \code{\link{\%within\%}}. Intervals can also be manipulated with
#' intersect, union, and setdiff().
#'
#' @section Miscellaneous:
#'
#' [decimal_date()] converts an instant to a decimal of
#' its year.
#' [leap_year()] tests whether an instant occurs during
#' a leap year.
#' [pretty_dates()] provides a method of making pretty
#' breaks for date-times.
#' [lakers] is a data set that contains information
#' about the Los Angeles Lakers 2008-2009 basketball season.
#'
#' @references Garrett Grolemund, Hadley Wickham (2011). Dates and Times Made
#'   Easy with lubridate. Journal of Statistical Software, 40(3), 1-25.
#'   \url{https://www.jstatsoft.org/v40/i03/}.
#' @importFrom methods setClass setGeneric new show allNames callGeneric is slot slot<- slotNames validObject Compare Arith initialize coerce<-
#' @importFrom utils packageVersion read.delim
#' @importFrom stats na.omit setNames update
#' @importFrom Rcpp sourceCpp
#' @useDynLib lubridate, .registration=TRUE
#' @keywords internal
"_PACKAGE"


## NOTE ON EXPORTS: `useDynLib` from above exports all registered functions (see
## ../src/RcppExports.cpp). _lubridate_C_xyz in there are Rcpp functions. All
## other `C_xyz` functions, are from plain .C files and are added manually with
## the help of tools::package_native_routine_registration_skeleton(".", character_only = FALSE).

##' @include util.r
NULL

## #' @export
## setMethod("<", signature(e1 = "POSIXct", e2 = "Date"),
##           function(e1, e2) {
##             cat("here")
##             callNextMethod(e1, .POSIXct(e2, tz = attr(e1, "tzone")))
##           })

## #' @export
## setMethod("<", signature(e1 = "Date", e2 = "POSIXct"),
##           function(e1, e2) {
##             cat("here")
##            callNextMethod(.POSIXct(e1, tz = attr(e2, "tzone")), e2)
##           })

comp_posix_date <- function(e1, e2) {
  if (nargs() == 1)
    stop(gettextf("unary '%s' not defined for \"%s\" objects",  .Generic, class(e1)), domain = NA)

  if (inherits(e1, "POSIXlt"))
    e1 <- as.POSIXct(e1)
  if (inherits(e2, "POSIXlt"))
    e2 <- as.POSIXct(e2)

  if (is.character(e1)) {
    e1 <-
      if (is.Date(e2)) as.Date(e1)
      else as.POSIXct(e1)
  }

  if (is.character(e2)) {
    e2 <-
      if (is.Date(e1)) as.Date(e2)
      else as.POSIXct(e2)
  }

  if (is.POSIXct(e1) && is.Date(e2)) {
    e2 <- date_to_posix(e2, tz(e1))
    base::check_tzones(e1, e2)
  } else if (is.Date(e1) && is.POSIXct(e2)) {
    e1 <- date_to_posix(e1, tz = tz(e2))
    base::check_tzones(e1, e2)
  }
  NextMethod(.Generic)
}

## Nothing else seems to work, only this sneaky trick.
evalqOnLoad({

  for (op in c("<", ">", "==", ">=", "<=", "!=")) {
    registerS3method(op, "Date", comp_posix_date)
    registerS3method(op, "POSIXt", comp_posix_date)
  }

})

#' Is a year a leap year?
#'
#' If x is a recognized date-time object, leap_year will return whether x
#' occurs during a leap year. If x is a number, leap_year returns whether it
#' would be a leap year under the Gregorian calendar.
#'
#' @export leap_year
#' @param date a date-time object or a year
#' @return TRUE if x is a leap year, FALSE otherwise
#' @keywords logic chron
#' @examples
#' x <- as.Date("2009-08-02")
#' leap_year(x) # FALSE
#' leap_year(2009) # FALSE
#' leap_year(2008) # TRUE
#' leap_year(1900) # FALSE
#' leap_year(2000) # TRUE
leap_year <- function(date) {
  if (is.numeric(date)) {
    year <- date
  } else {
    year <- year(date)
  }
  (year %% 4 == 0) & ((year %% 100 != 0) | (year %% 400 == 0))
}

##' Guess possible date-times formats from a character vector
##'
##' Guess possible date-times formats from a character vector.
##'
##' @param x input vector of date-times.
##' @param orders format orders to look for. See examples.
##' @param locale locale to use. Defaults to the current locale.
##' @param preproc_wday whether to preprocess weekday names. Internal
##' optimization used by `ymd_hms()` family of functions. If `TRUE`, weekdays are
##' substituted with %a or %A accordingly, so that there is no need to supply
##' this format explicitly.
##' @param print_matches for development purposes mainly. If `TRUE`, prints a matrix
##' of matched templates.
##' @return a vector of matched formats
##' @export
##' @examples
##'
##' x <- c('February 20th 1973',
##'        "february  14, 2004",
##'        "Sunday, May 1, 2000",
##'        "Sunday, May 1, 2000",
##'        "february  14, 04",
##'        'Feb 20th 73',
##'        "January 5 1999 at 7pm",
##'        "jan 3 2010",
##'        "Jan 1, 1999",
##'        "jan 3   10",
##'        "01 3 2010",
##'        "1 3 10",
##'        '1 13 89',
##'        "5/27/1979",
##'        "12/31/99",
##'        "DOB:12/11/00",
##'        "-----------",
##'        'Thu, 1 July 2004 22:30:00',
##'        'Thu, 1st of July 2004 at 22:30:00',
##'        'Thu, 1July 2004 at 22:30:00',
##'        'Thu, 1July2004 22:30:00',
##'        'Thu, 1July04 22:30:00',
##'        "21 Aug 2011, 11:15:34 pm",
##'        "-----------",
##'        "1979-05-27 05:00:59",
##'        "1979-05-27",
##'        "-----------",
##'        "3 jan 2000",
##'        "17 april 85",
##'        "27/5/1979",
##'        '20 01 89',
##'        '00/13/10',
##'        "-------",
##'        "14 12 00",
##'        "03:23:22 pm")
##'
##' guess_formats(x, "BdY")
##' guess_formats(x, "Bdy")
##' ## m also matches b and B; y also matches Y
##' guess_formats(x, "mdy", print_matches = TRUE)
##'
##' ## T also matches IMSp order
##' guess_formats(x, "T", print_matches = TRUE)
##'
##' ## b and B are equivalent and match, both, abreviated and full names
##' guess_formats(x, c("mdY", "BdY", "Bdy", "bdY", "bdy"), print_matches = TRUE)
##' guess_formats(x, c("dmy", "dbY", "dBy", "dBY"), print_matches = TRUE)
##'
##'
##' guess_formats(x, c("dBY HMS", "dbY HMS", "dmyHMS", "BdY H"), print_matches = TRUE)
##'
##' guess_formats(x, c("ymd HMS"), print_matches = TRUE)
##'
guess_formats <- function(x, orders, locale = Sys.getlocale("LC_TIME"),
                          preproc_wday = TRUE, print_matches = FALSE) {

  ## remove all separators
  orders <- gsub("[^[:alpha:]]+", "", orders)

  if (any(grepl("hms|hm|ms", orders))) {
    .deprecated("hms, hm and ms usage", ", please use HMS, HM or MS instead", "1.5.6")
    orders <- gsub("hms", "HMS", orders, ignore.case = TRUE)
    orders <- gsub("hm", "HM", orders, ignore.case = TRUE)
    orders <- gsub("ms", "MS", orders, ignore.case = TRUE)
  }

  ## redirect some formats to C parser (using perl's lookbehind)
  if (any(wp <- grepl("(?<!O)p", orders, perl = T))) {
    orders <- c(sub("p", "Op", orders[wp], fixed = T), orders)
  }
  if (any(wm <- grepl("(?<!O)m", orders, perl = T))) {
    orders <- c(sub("m", "Om", orders[wm]), orders)
  }
  if (any(wm <- grepl("(?<!O)[bB]", orders, perl = T))) {
    orders <- c(sub("[Bb]", "Ob", orders[wm]), orders)
  }
  if (any(wT <- grepl("T", orders, fixed = T))) {
    orders <- c(sub("T", "HMSOp", orders[wT], fixed = T), orders)
  }
  if (any(wR <- grepl("R", orders, fixed = T))) {
    orders <- c(sub("R", "HMOp", orders[wR], fixed = T), orders)
  }
  if (any(wr <- grepl("r", orders, fixed = T))) {
    orders <- c(sub("r", "HOp", orders[wr], fixed = T), orders)
  }

  ## We split into characterst first and then paste together formats that start
  ## with O.
  osplits <- strsplit(orders, "", fixed = TRUE)
  osplits <- lapply(
    osplits,
    function(ospt) {
      if (length(which_O <- which(ospt == "O")) > 0) {
        ospt[which_O + 1] <- paste("O", ospt[which_O + 1], sep = "")
        ospt[-which_O]
      } else {
        ospt
      }
    }
  )

  reg <- .get_locale_regs(locale)

  flex_regs <- c(reg$alpha_flex, reg$num_flex, .c_parser_reg_flex)
  exact_regs <- c(reg$alpha_exact, reg$num_exact, .c_parser_reg_exact)

  all_format_names <- c(
    names(reg$alpha_flex),
    names(reg$num_flex),
    names(.c_parser_reg_exact)
  )

  REGS <- unlist(lapply(osplits, function(fnames) {
    ## fnames are names of smalest valid formats, like a, A, b, z, OS, OZ ...
    invalid <- !fnames %in% all_format_names
    if (any(invalid)) {
      stop("Unknown formats supplied: ", paste(fnames[invalid], sep = ", "))
    }

    ## restriction: no numbers before or after
    ## fixme: using \\D*? because \\D+ doesn't work in flex match, why?
    paste("^\\D*?\\b((",
      paste(unlist(flex_regs[fnames]), collapse = "\\D*?"),
      ")|(",
      paste(unlist(exact_regs[fnames]), collapse = "\\D*?"),
      "))\\D*$",
      sep = ""
    )
  }))

  ## print debugging info
  if (print_matches) {
    subs <- lapply(REGS, .substitute_formats, x, fmts_only = FALSE)
    names(subs) <- orders
    print(do.call(cbind, c(list(x), subs)))
  }

  .build_formats <- function(regs, orders, x) {
    out <- mapply(
      function(reg, name) {
        out <- .substitute_formats(reg, x)
        if (!is.null(out)) {
          names(out) <- rep.int(name, length(out))
          out
        }
      }, REGS, orders,
      SIMPLIFY = F, USE.NAMES = F
    )
    names(out) <- NULL
    unlist(out)
  }

  if (preproc_wday && !any(grepl("[aA]", orders))) {
    ## replace short/long weak days in current locale
    x2 <- gsub(reg$alpha_exact[["A"]], "%A", x, ignore.case = T, perl = T)
    x2 <- gsub(reg$alpha_exact[["a"]], "%a", x2, ignore.case = T, perl = T)
    formats <- .build_formats(REGS, orders, x2)

    ## In some locales abreviated day (italian "gio") is part of month name
    ## ("maggio"). So check if %a or %A is present and append wday-less formats.
    if (any(grepl("%[aA]", formats))) {
      c(formats, .build_formats(REGS, orders, x))
    } else {
      formats
    }
  } else {
    .build_formats(REGS, orders, x)
  }
}

.substitute_formats <- function(reg, x, fmts_only = TRUE) {
  ## Take date X and substitute year with %Y/%y, month with %B/%b etc.
  ## Return the formatted string if REG matched, or null otherwise.
  ## REG should be with captures as build by .build_locale_regs
  ## Captures should start with the strptyme formats. For example Y_e, B_m_e
  ## Traling _e, _m_e are removed.
  ## fmts_only == FALSE is for debugging purpose only

  m <- regexpr(reg, x, ignore.case = TRUE, perl = TRUE)
  ## print(regs[[1]])
  matched <- which(m > 0)

  if (length(matched) > 0) {
    nms <- attr(m, "capture.names")
    nms <- nms[nzchar(nms)]
    ## e <- grepl("_e", nms, fixed = TRUE)
    ## nms_e <- nms[e]
    ## nms <- nms[!e]

    start <- attr(m, "capture.start")[matched, , drop = FALSE]
    end <- start + attr(m, "capture.length")[matched, , drop = FALSE] - 1L

    lout <- x[matched]
    for (n in rev(nms)) { ## start from the end
      w <- end[, n] > 0 ## -1 if unmatched  subpatern
      lout[w] <- .str_sub(
        lout[w], start[w, n], end[w, n],
        paste("%", gsub("_.*$", "", n), sep = "")
      )
    }
    if (fmts_only) {
      lout
    } else {
      ## developer
      out <- character(length(x)) ## dev only
      out[matched] <- lout
      out
    }
  } else if (fmts_only) {
    NULL
  } else {
    character(length(x))
  }
}

.enclose <- function(fmts) {
  paste("@", fmts, "@", sep = "")
}

.enclosed.na <- function(x) {
  x == "@NA@"
}

.get_train_set <- function(x) {
  ## Use first 501 primes to retrieve a subset of the original vector for training of
  ## which formats should come first.
  len <- length(x)
  if (len < 100) {
    x
  } else if (len < 3571) {
    x[.primes[.primes <= length(x)]]
  } else {
    x[.primes * (length(x) %/% 3571)]
  } # 501 primes
}

.train_formats <- function(x, formats, locale) {
  ## return a numeric vector of size length(formats), with each element giving
  ## the number of matched elements in X
  ## can return NULL if formats is NULL

  trials <- lapply(formats, function(fmt) .strptime(x, fmt, locale = locale))
  successes <- unlist(lapply(trials, function(x) sum(!is.na(x))), use.names = FALSE)
  names(successes) <- formats
  sort(successes, decreasing = TRUE)
}

.best_formats <- function(x, orders, locale, .select_formats = .select_formats, drop = FALSE, train = TRUE) {
  ## return a vector of formats that matched X at least once.
  ## Can be zero length vector, if none matched

  fmts <- unique(guess_formats(x, orders, locale = locale, preproc_wday = TRUE))
  if (train && length(fmts)) {
    trained <- .train_formats(x, fmts, locale = locale)

    if (drop) {
      trained <- trained[trained > 0]
    }
    fmts <- .select_formats(trained, drop)
  }
  if (length(fmts)) {
    fmts
  }
}

.select_formats <- function(trained, drop = FALSE) {
  nms <- names(trained)

  score <-
    nchar(gsub("[^%]", "", nms)) + ## longer formats have priority
    grepl("%Y", nms, fixed = T) * 1.5 +
    grepl("%y(?!%)", nms, perl = T) * 1.6 + ## y has priority over Y, but only when not followed by %
    grepl("%[Bb]", nms) * .31 + ## B/b format has priority over %Om
    ## C parser formats have higher priority
    grepl("%Om", nms) * .3 +
    grepl("%Op", nms) * .3 +
    grepl("%Ob", nms) * .32 ## Ob has higher priority than B/b

  ## ties are broken by `trained`
  n0 <- trained != 0
  if (drop) {
    score <- score[n0]
    trained <- trained[n0]
  } else {
    score[!n0] <- -100
  }

  ## print(rbind(trained, score))
  names(trained)[order(score, trained, decreasing = T)]
}

## These are formats that are effectively matched by c parser. But we must get
## through the format guesser first for ymd_hms family.
.c_parser_reg_flex <- list(
  Op = "(?<Op>(AM|PM))(?![[:alpha:]])",
  Om = "((?<Om>1[0-2]|0?[1-9](?!\\d))|(((?<Om_b>Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)|(?<Om_B>January|February|March|April|May|June|July|August|September|October|November|December))(?![[:alpha:]])))",
  Ob = "(((?<Ob_b>Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)|(?<Ob_B>January|February|March|April|May|June|July|August|September|October|November|December))(?![[:alpha:]]))"
)

.c_parser_reg_exact <- list(
  Op = "(?<Op_e>AM|PM)(?![[:alpha:]])",
  Om = "((?<Om_e>1[0-2]|0[1-9])|(((?<Om_b_e>Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)|(?<Om_B_e>January|February|March|April|May|June|July|August|September|October|November|December))(?![[:alpha:]])))",
  Ob = "(((?<Ob_b_e>Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)|(?<Ob_B_e>January|February|March|April|May|June|July|August|September|October|November|December))(?![[:alpha:]]))"
)

.locale_reg_cache <- new.env(hash = FALSE)

.get_locale_regs <- function(locale = Sys.getlocale("LC_TIME")) {
  ## build locale specific regexps for all posible orders

  if (exists(locale, envir = .locale_reg_cache, inherits = FALSE)) {
    return(get(locale, envir = .locale_reg_cache))
  }

  orig_opt <- options(warn = 5)
  on.exit({
    options(orig_opt)
    Sys.setlocale("LC_TIME", orig_locale)
  })
  orig_locale <- Sys.getlocale("LC_TIME")
  Sys.setlocale("LC_TIME", locale)
  options(orig_opt)

  format <- "%a@%A@%b@%B@%p@"
  L <- enc2utf8(unique(format(.date_template, format = format)))
  mat <- do.call(rbind, strsplit(L, "@", fixed = TRUE))
  mat[] <- gsub("([].|(){^$*+?[])", "\\\\\\1", mat) # escaping meta chars
  names <- colnames(mat) <- strsplit(format, "[%@]+")[[1]][-1L]

  ## Captures should be unique. Thus we build captures with the following
  ## rule. Capture starts with the name of strptime format (B, b, y etc) It ends
  ## with _e if the expression is an exact match (as oposed to flex), see below.

  ## It can contain _x where x is a main format in which this format
  ## occurs. For example <B_b_e> is an exact capture in the strptime format B
  ## but that also matches b in lubridate. Try lubridate:::.get_loc_regs()
  ## todo: elaborate

  ## ALPHABETIC FORMATS
  alpha <- list()
  alpha["b"] <-
    sprintf(
      "((?<b_b>%s)|(?<B_b>%s))(?![[:alpha:]])",
      paste(unique(mat[, "b"]), collapse = "|"),
      paste(unique(mat[, "B"]), collapse = "|")
    )
  alpha["B"] <-
    sprintf(
      "(?<B_B>%s)(?![[:alpha:]])",
      paste(unique(mat[, "B"]), collapse = "|")
    )

  alpha["a"] <-
    sprintf(
      "((?<a_a>%s)|(?<A_a>%s))(?![[:alpha:]])",
      paste(unique(mat[, "a"]), collapse = "|"),
      paste(unique(mat[, "A"]), collapse = "|")
    )

  alpha["A"] <-
    sprintf(
      "(?<A_A>%s)(?![[:alpha:]])",
      paste(unique(mat[, "A"]), collapse = "|")
    )

  ## just match Z in ISO8601, (UTC zulu format)
  alpha["Ou"] <- "(?<Ou_Ou>Z)(?![[:alpha:]])"

  p <- unique(mat[, "p"])
  p <- p[nzchar(p)]
  no_p <- length(p) == 0L
  alpha["p"] <-
    if (no_p) {
      ""
    } else {
      sprintf("(?<p>%s)(?![[:alpha:]])", paste(p, collapse = "|"))
    }

  alpha <- unlist(alpha)

  ##  NUMERIC FORMATS
  num <- num_flex <- num_exact <- c(
    d = "(?<d>[012]?[1-9]|3[01]|[12]0)",
    q = "(?<q>[0]?[1-4])",
    H = "(?<H>2[0-4]|[01]?\\d)",
    h = "(?<H>2[0-4]|[01]?\\d)",
    I = "(?<I>1[0-2]|0?[1-9])",
    j = "(?<j>[0-3]?\\d?\\d)",
    M = "(?<M>[0-5]?\\d)",
    S = "((?<OS_S>[0-5]?\\d\\.\\d+)|(?<S>[0-6]?\\d))",
    s = "((?<OS_S>[0-5]?\\d\\.\\d+)|(?<S>[0-6]?\\d))",
    U = "(?<U>[0-5]?\\d)",
    w = "(?<w>[0-6])", # merge with a, A??
    u = "(?<u>[1-7])",
    W = "(?<W>[0-5]?\\d)",
    ## x = "(?<x>\\d{2}/[01]?\\d/[0-3]?\\d)",
    ## X = "(?<X>[012]?\\d:[0-5]?\\d:[0-6]?\\d)",
    Y = "(?<Y>\\d{4})",
    y = "((?<Y_y>\\d{4})|(?<y>\\d{2}))",
    Oz = "(?<Oz_Oz>[-+]\\d{4})", ## sptrtime implements only this format (4 digits)
    ## F = "(?<F>\\d{4)-\\d{2}-\\d{2})",
    OO = "(?<OO>[-+]\\d{2}:\\d{2})",
    Oo = "(?<Oo>[-+]\\d{2})"
  )

  ## special lubridate formats
  nms <- c("T", "R", "r")

  if (no_p) {
    num <- c(num,
      T = sprintf("(%s\\D+%s\\D+%s)", num[["H"]], num[["M"]], num[["S"]]),
      R = sprintf("(%s\\D+%s)", num[["H"]], num[["M"]]),
      r = sprintf("(%s)",   num[["H"]])
    )
  } else {
    num <- c(num,
      T = sprintf(
        "((%s\\D+%s\\D+%s\\D*%s)|(%s\\D+%s\\D+%s))",
        num[["I"]], num[["M"]], num[["S"]], alpha[["p"]],
        num[["H"]], num[["M"]], num[["S"]]
      ),
      R = sprintf(
        "((%s\\D+%s\\D*%s)|(%s\\D+%s))",
        num[["I"]], num[["M"]], alpha[["p"]],
        num[["H"]], num[["M"]]
      ),
      r = sprintf(
        "((%s\\D*%s)|%s)",
        num[["I"]], alpha[["p"]],
        num[["H"]]
      )
    )

    num[nms] <- sub(
      "<M", "<M_T",
      sub(
        "<S", "<S_T",
        sub("<OS", "<OS_T", num[nms])
      )
    )
  }

  ## don't let special nms be confused with standard formats
  num[nms] <- gsub("(<[IMSpHS]|<OS)", "\\1_s", num[nms])


  ## The difference between flex and exact regexp is that flex should follow
  ## by non-digit, thus flexible strings like 12-1-2 can be matched. In exact
  ## regexp a number may not be followed by a non-number. In this case the
  ## number of digits in a number should be precisely specified. For example
  ## 120102 is matched by ymd.
  ##
  ## *Note*: the exact regexps is build from flex regex by gsubing below. So,
  ## pay attention when you modify the regexp above.

  alpha_flex <- alpha
  alpha_exact <- gsub(">", "_e>", alpha, fixed = TRUE)

  num_exact <- num_flex <- num
  num_flex[] <- sprintf("%s(?!\\d)", num)

  num_exact[] <- gsub("(?<!\\()\\?(?!<)", "",
    perl = T, # remove ?
    gsub("+", "*",
      fixed = T,
      gsub(">", "_e>", num)
    )
  ) # append _e to avoid duplicates

  num_flex["m"] <- sprintf("((?<m>1[0-2]|0?[1-9](?!\\d))|(%s))", gsub("_[bB]", "\\1_m", alpha[["b"]]))
  num_exact["m"] <- sprintf("((?<m_e>1[0-2]|0[1-9])|(%s))", gsub("_[bB]", "\\1_m_e", alpha[["b"]]))

  ## canoot be in num above because gsub("+", "*") messes it up
  num_flex["OS"] <- "(?<OS_f>[0-5]\\d\\.\\d+)"
  num_exact["OS"] <- "(?<OS_e>[0-5]\\d\\.\\d+)"
  num_flex["z"] <- sprintf("(%s|%s|%s|%s)", alpha_flex[["Ou"]], num_flex[["Oz"]], num_flex[["OO"]], num_flex[["Oo"]])
  num_exact["z"] <- sprintf("(%s|%s|%s|%s)", alpha_exact[["Ou"]], num_exact[["Oz"]], num_exact[["OO"]], num_exact[["Oo"]])

  wday_order <- order(wday(.date_template, week_start = 7))
  wday_names <- list(
    abr = unique(mat[, "a"][wday_order]),
    full = unique(mat[, "A"][wday_order])
  )

  month_order <- order(month(.date_template))
  # remove trailing dot in some locales (#781)
  month_names <- list(
    abr = sub("\\.", "", unique(mat[, "b"][month_order]), fixed = T),
    full = unique(mat[, "B"][month_order])
  )

  out <- list(
    alpha_flex = alpha_flex, num_flex = num_flex,
    alpha_exact = alpha_exact, num_exact = num_exact,
    wday_names = wday_names, month_names = month_names
  )

  assign(locale, out, envir = .locale_reg_cache)
  out
}

#' @include periods.r
NULL

#' Get/set weeks component of a date-time
#'
#' @description
#' `week()` returns the number of complete seven day periods that have
#' occurred between the date and January 1st, plus one.

#' @param x a date-time object. Must be a POSIXct, POSIXlt, Date, chron,
#'   yearmon, yearqtr, zoo, zooreg, timeDate, xts, its, ti, jul, timeSeries, or
#'   fts object.
#' @param value a numeric object
#' @return the weeks element of x as an integer number
#' @keywords utilities manip chron
#' @references
#'    \url{https://en.wikipedia.org/wiki/ISO_week_date}
#' @seealso [isoyear()]
#' @examples
#' x <- ymd("2012-03-26")
#' week(x)
#' week(x) <- 1
#' week(x) <- 54
#' week(x) > 3
#' @export
week <- function(x)
  (yday(x) - 1) %/% 7 + 1

#' @rdname week
#' @export
"week<-" <- function(x, value)
  x <- x + days((value - week(x)) * 7)

.other_week <- function(x, week_start) {
  x <- as.POSIXlt(x)
  date <- make_date(year(x), month(x), day(x))
  wday <- wday(x, week_start = week_start)
  date <- date + (4 - wday)
  jan1 <- as.numeric(make_date(year(date), 1, 1))
  1L + (as.numeric(date) - jan1) %/% 7L
}

#' @description `isoweek()` returns the week as it would appear in the ISO 8601
#'   system, which uses a reoccurring leap week.
#' @rdname week
#' @export
isoweek <- function(x) {
  .other_week(x, 1)
}

#' @description `epiweek()` is the US CDC version of epidemiological week. It
#'   follows same rules as `isoweek()` but starts on Sunday. In other parts of
#'   the world the convention is to start epidemiological weeks on Monday,
#'   which is the same as `isoweek`.
#'
#' @rdname week
#' @export
epiweek <- function(x) {
  .other_week(x, 7)
}

#' Lakers 2008-2009 basketball data set
#'
#' This data set contains play by play statistics of each Los
#' Angeles Lakers basketball game in the 2008-2009 season. Data
#' includes the date, opponent, and type of each game (home or
#' away). Each play is described by the time on the game clock
#' when the play was made, the period in which the play was
#' attempted, the type of play, the player and team who made the
#' play, the result of the play, and the location on the court
#' where each play was made.
#'
#' @name lakers
#' @docType data
#' @references Originally taken from www.basketballgeek.com/data/.
#' @keywords data
NULL

#' Round, floor and ceiling methods for date-time objects
#'
#' @description
#' `round_date()` takes a date-time object and time unit, and rounds it to the nearest value
#' of the specified time unit. For rounding date-times which are exactly halfway
#' between two consecutive units, the convention is to round up. Note that this
#' is in line with the behavior of R's [base::round.POSIXt()] function
#' but does not follow the convention of the base [base::round()] function
#' which "rounds to the even digit", as per IEC 60559.
#'
#' Rounding to the nearest unit or multiple of a unit is supported. All
#' meaningful specifications in the English language are supported - secs, min,
#' mins, 2 minutes, 3 years etc.
#'
#' Rounding to fractional seconds is also supported. Please note that rounding to
#' fractions smaller than 1 second can lead to large precision errors due to the
#' floating point representation of the POSIXct objects. See examples.
#'
#'
#' @details In \pkg{lubridate}, functions that round date-time objects try to
#'   preserve the class of the input object whenever possible. This is done by
#'   first rounding to an instant, and then converting to the original class as per
#'   usual R conventions.
#'
#'
#' @section Rounding Up Date Objects:
#'
#' By default, rounding up `Date` objects follows 3 steps:
#'
#' 1. Convert to an instant representing lower bound of the Date:
#'    `2000-01-01` --> `2000-01-01 00:00:00`
#'
#' 2. Round up to the \strong{next} closest rounding unit boundary. For example,
#'    if the rounding unit is `month` then next closest boundary of `2000-01-01`
#'    is `2000-02-01 00:00:00`.
#'
#'    The motivation for this is that the "partial" `2000-01-01` is conceptually
#'    an interval (`2000-01-01 00:00:00` -- `2000-01-02 00:00:00`) and the day
#'    hasn't started clocking yet at the exact boundary `00:00:00`. Thus, it
#'    seems wrong to round a day to its lower boundary.
#'
#'    Behavior on the boundary can be changed by setting
#'    `change_on_boundary` to `TRUE` or `FALSE`.
#'
#' 3. If the rounding unit is smaller than a day, return the instant from step 2
#'     (`POSIXct`), otherwise convert to and return a `Date` object.
#'
#' @rdname round_date
#' @param x a vector of date-time objects
#' @param unit a character string specifying a time unit or a multiple of a unit
#'   to be rounded to. Valid base units are `second`, `minute`, `hour`, `day`,
#'   `week`, `month`, `bimonth`, `quarter`, `season`, `halfyear` and
#'   `year`. Arbitrary unique English abbreviations as in the [period()]
#'   constructor are allowed. Rounding to multiples of units (except weeks) is
#'   supported.
#' @param change_on_boundary if this is `NULL` (the default), instants on the boundary
#'   remain unchanged, but `Date` objects are rounded up to the next boundary.
#'   If this is `TRUE`, instants on the boundary are rounded up to the next boundary.
#'   If this is `FALSE`, nothing on the boundary is rounded up at all. This was the
#'   default for \pkg{lubridate} prior to `v1.6.0`.
#'   See section `Rounding Up Date Objects` below for more details.
#'
#' @param week_start when unit is `weeks`, specify the reference day.
#'   7 represents Sunday and 1 represents Monday.
#' @keywords manip chron
#' @seealso [base::round()]
#' @examples
#'
#' ## print fractional seconds
#' options(digits.secs=6)
#'
#' x <- ymd_hms("2009-08-03 12:01:59.23")
#' round_date(x, ".5s")
#' round_date(x, "sec")
#' round_date(x, "second")
#' round_date(x, "minute")
#' round_date(x, "5 mins")
#' round_date(x, "hour")
#' round_date(x, "2 hours")
#' round_date(x, "day")
#' round_date(x, "week")
#' round_date(x, "month")
#' round_date(x, "bimonth")
#' round_date(x, "quarter") == round_date(x, "3 months")
#' round_date(x, "halfyear")
#' round_date(x, "year")
#'
#' x <- ymd_hms("2009-08-03 12:01:59.23")
#' floor_date(x, ".1s")
#' floor_date(x, "second")
#' floor_date(x, "minute")
#' floor_date(x, "hour")
#' floor_date(x, "day")
#' floor_date(x, "week")
#' floor_date(x, "month")
#' floor_date(x, "bimonth")
#' floor_date(x, "quarter")
#' floor_date(x, "season")
#' floor_date(x, "halfyear")
#' floor_date(x, "year")
#'
#' x <- ymd_hms("2009-08-03 12:01:59.23")
#' ceiling_date(x, ".1 sec") # imprecise representation at 0.1 sec !!!
#' ceiling_date(x, "second")
#' ceiling_date(x, "minute")
#' ceiling_date(x, "5 mins")
#' ceiling_date(x, "hour")
#' ceiling_date(x, "day")
#' ceiling_date(x, "week")
#' ceiling_date(x, "month")
#' ceiling_date(x, "bimonth") == ceiling_date(x, "2 months")
#' ceiling_date(x, "quarter")
#' ceiling_date(x, "season")
#' ceiling_date(x, "halfyear")
#' ceiling_date(x, "year")
#'
#' ## As of R 3.4.2 POSIXct printing of fractional numbers is wrong
#' as.POSIXct("2009-08-03 12:01:59.3") ## -> "2009-08-03 12:01:59.2 CEST"
#' ceiling_date(x, ".1 sec") ## -> "2009-08-03 12:01:59.2 CEST"
#'
#' ## behaviour of `change_on_boundary`
#' ## As per default behaviour `NULL`, instants on the boundary remain the
#' ## same but dates are rounded up
#' ceiling_date(ymd_hms("2000-01-01 00:00:00"), "month")
#' ceiling_date(ymd("2000-01-01"), "month")
#'
#' ## If `TRUE`, both instants and dates on the boundary are rounded up
#' ceiling_date(ymd_hms("2000-01-01 00:00:00"), "month", change_on_boundary = TRUE)
#' ceiling_date(ymd("2000-01-01"), "month")
#'
#' ## If `FALSE`, both instants and dates on the boundary remain the same
#' ceiling_date(ymd_hms("2000-01-01 00:00:00"), "month", change_on_boundary = FALSE)
#' ceiling_date(ymd("2000-01-01"), "month")
#'

#' @export
round_date <- function(x, unit = "second", week_start = getOption("lubridate.week.start", 7)) {

  if (!length(x)) return(x)

  parsed_unit <- parse_period_unit(unit)
  n <- parsed_unit$n
  basic_unit <- standardise_period_names(parsed_unit$unit)

  new <-
    if (n == 1 && basic_unit %in% c("second", "minute", "hour", "day")) {
      ## special case for fast rounding
      round.POSIXt(x, units = lub2base_units[[basic_unit]])
    } else {
      above <- unclass(as.POSIXct(ceiling_date(x, unit = unit, week_start = week_start)))
      mid <- unclass(as.POSIXct(x))
      below <- unclass(as.POSIXct(floor_date(x, unit = unit, week_start = week_start)))
      wabove <- (above - mid) <= (mid - below)
      wabove <- !is.na(wabove) & wabove
      new <- below
      new[wabove] <- above[wabove]
      .POSIXct(new, tz = tz(x))
    }

  reclass_date(new, x)
}

reclass_date_maybe <- function(new, orig, unit) {
  if (is.Date(orig) && !unit %in% c("day", "week", "month", "year")) as.POSIXct(new)
  else reclass_date(new, orig)
}

#' @description
#' `floor_date()` takes a date-time object and rounds it down to the nearest
#' boundary of the specified time unit.
#' @rdname round_date
#' @export
floor_date <- function(x, unit = "seconds", week_start = getOption("lubridate.week.start", 7)) {
  if (!length(x)) return(x)

  parsed_unit <- parse_period_unit(unit)
  n <- parsed_unit$n
  unit <- standardise_period_names(parsed_unit$unit)

  if (unit %in% c("second", "minute", "hour", "day")) {

    out <- trunc_multi_unit(x, unit, n)
    reclass_date_maybe(out, x, unit)

  } else {

    if (n > 1 && unit == "week") {
      ## fixme:
      warning("Multi-unit not supported for weeks. Ignoring.")
    }

    if (unit %in% c("bimonth", "quarter", "halfyear", "season") ||
       (n > 1 && unit == "month")) {
      new_months <-
        switch(unit,
               month    = floor_multi_unit1(month(x), n),
               bimonth  = floor_multi_unit1(month(x), 2 * n),
               quarter  = floor_multi_unit1(month(x), 3 * n),
               halfyear = floor_multi_unit1(month(x), 6 * n),
               season   = floor_multi_unit(month(x), 3 * n))
      n <- Inf
      unit <- "month"
    }

    switch(unit,
           week     = update(x, wdays = 1, hours = 0, minutes = 0, seconds = 0, week_start = week_start),
           month    = {
             if (n > 1) update(x, months = new_months, mdays = 1, hours = 0, minutes = 0, seconds = 0)
             else      update(x, mdays = 1, hours = 0, minutes = 0, seconds = 0)
           },
           year     = {
             ## due to bug https://github.com/tidyverse/lubridate/issues/319 we
             ## need to do it in two steps
             if (n > 1) {
               y <- update(x, ydays = 1, hours = 0, minutes = 0, seconds = 0)
               update(y, years = floor_multi_unit(year(y), n))
             } else {
               update(x, ydays = 1, hours = 0, minutes = 0, seconds = 0)
             }
           })
  }
}

#' @description
#' `ceiling_date()` takes a date-time object and rounds it up to the nearest
#' boundary of the specified time unit.
#' @rdname round_date
#' @export
#' @examples
#'
#'  x <- ymd_hms("2000-01-01 00:00:00")
#'  ceiling_date(x, "month")
#'  ceiling_date(x, "month", change_on_boundary = TRUE)
#'
#'  ## For Date objects first day of the month is not on the
#'  ## "boundary". change_on_boundary applies to instants only.
#'  x <- ymd("2000-01-01")
#'  ceiling_date(x, "month")
#'  ceiling_date(x, "month", change_on_boundary = TRUE)
ceiling_date <- function(x, unit = "seconds", change_on_boundary = NULL, week_start = getOption("lubridate.week.start", 7)) {

  if (!length(x))
    return(x)

  parsed_unit <- parse_period_unit(unit)
  n <- parsed_unit$n
  unit <- standardise_period_names(parsed_unit$unit)

  if (is.null(change_on_boundary)) {
    change_on_boundary <- is.Date(x)
  }

  if (unit == "second") {

    sec <- second(x)
    csec <- ceil_multi_unit(sec, n)
    if (!change_on_boundary) {
      wsec <- which(csec - n ==  sec)
      if (length(wsec))
        csec[wsec] <- sec[wsec]
    }
    update(x, seconds = csec, simple = T)

  } else if (unit %in% c("minute", "hour")) {

    ## cannot use this for minute/hour for Date class; local tz interferes with
    ## the computation
    new <- as_datetime(x, tz = tz(x))
    delta <- switch(unit, minute = 60, hour = 3600, day = 86400) * n
    new <-
      if (change_on_boundary) {
        trunc_multi_unit(new, unit, n) + delta
      } else {
        new1 <- trunc_multi_unit(new, unit, n)
        not_same <- which(new1 != new)
        new1[not_same] <- new1[not_same] + delta
        new1
      }
    reclass_date_maybe(new, x, unit)

  } else {

    if (n > 1 && unit == "week") {
      warning("Multi-unit not supported for weeks. Ignoring.")
    }

    ## need this to accomodate the case when date is on a boundary
    new <-
      if (change_on_boundary) x
      else update(x, seconds = second(x) - 0.00001, simple = T)

    if (unit %in% c("month", "bimonth", "quarter", "halfyear", "season")) {
      new_month <-
        switch(unit,
               month    = ceil_multi_unit1(month(new), n),
               bimonth  = ceil_multi_unit1(month(new), 2 * n),
               quarter  = ceil_multi_unit1(month(new), 3 * n),
               halfyear = ceil_multi_unit1(month(new), 6 * n),
               season   = ceil_multi_unit(month(new), 3 * n))
      unit <- "month"
    }

    new <- switch(unit,
                  minute = update(new, minutes = ceil_multi_unit(minute(new), n), seconds = 0, simple = T),
                  hour   = update(new, hours = ceil_multi_unit(hour(new), n), minutes = 0, seconds = 0, simple = T),
                  day    = update(new, days = ceil_multi_unit1(day(new), n), hours = 0, minutes = 0, seconds = 0),
                  week   = update(new, wdays = 8, hours = 0, minutes = 0, seconds = 0, week_start = week_start),
                  month  = update(new, months = new_month, mdays = 1, hours = 0, minutes = 0, seconds = 0),
                  year   = update(new, years = ceil_multi_unit(year(new), n), months = 1, mdays = 1, hours = 0, minutes = 0, seconds = 0))

    reclass_date_maybe(new, x, unit)
  }
}

trunc_multi_limits <- c(second = 60L, minute = 60L, hour = 24, day = 31)

trunc_multi_unit <- function(x, unit, n) {
  x <- as.POSIXlt(x)
  if (n > trunc_multi_limits[[unit]])
    stop(sprintf("Rounding with %s > %d is not supported", unit, trunc_multi_limits[[unit]]))

  switch(unit,
         second = {
           x$sec <- if (n == 1) trunc(x$sec) else floor_multi_unit(x$sec, n)
         },
         minute = {
           x$sec[] <- 0
           x$min <- floor_multi_unit(x$min, n)
         },
         hour = {
           x$sec[] <- 0
           x$min[] <- 0L
           x$hour <- floor_multi_unit(x$hour, n)
         },
         day = {
           x$sec[] <- 0
           x$min[] <- 0L
           x$hour[] <- 0L
           x$isdst[] <- -1L
           x$mday <- floor_multi_unit1(x$mday, n)
         },
         stop("Invalid unit ", unit))
  x
}

floor_multi_unit <- function(x, n) {
  (x %/% n) * n
}

floor_multi_unit1 <- function(x, n) {
  (((x - 1) %/% n) * n) + 1L
}

ceil_multi_unit <- function(x, n) {
  (x %/% n) * n + n
}

ceil_multi_unit1 <- function(x, n) {
  (((x - 1) %/% n) *  n) + n + 1L
}

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r


setOldClass("difftime")

#' Create a difftime object.
#'
#' `make_difftime()` creates a difftime object with the specified number of
#' units. Entries for different units are cumulative. difftime displays
#' durations in various units,  but these units are estimates given for
#' convenience. The underlying object is always recorded as a fixed number of
#' seconds.
#'
#' Conceptually, difftime objects are a type of duration. They measure the
#' exact passage of time but do not always align with measurements
#' made in larger units of time such as hours, months and years.
#' This is because the length of larger time units can be affected
#' by conventions such as leap years
#' and Daylight Savings Time. \pkg{lubridate} provides a second class for measuring durations, the Duration class.
#' @param num Optional number of seconds
#' @param units a character vector that lists the type of units to use for the
#'   display of the return value (see examples). If `units` is "auto" (the
#'   default) the display units are computed automatically. This might create
#'   undesirable effects when converting `difftime` objects to numeric
#'   values in data processing.
#' @param ... a list of time units to be included in the difftime and their amounts. Seconds,
#'   minutes, hours, days, and weeks are supported. Normally only one of `num` or `...` are present. If
#'   both are present, the `difftime` objects are concatenated.
#' @return a difftime object
#' @seealso [duration()], [as.duration()]
#' @keywords chron classes
#' @export
#' @examples
#' make_difftime(1)
#' make_difftime(60)
#' make_difftime(3600)
#' make_difftime(3600, units = "minute")
#' # Time difference of 60 mins
#' make_difftime(second = 90)
#' # Time difference of 1.5 mins
#' make_difftime(minute = 1.5)
#' # Time difference of 1.5 mins
#' make_difftime(second = 3, minute = 1.5, hour = 2, day = 6, week = 1)
#' # Time difference of 13.08441 days
#' make_difftime(hour = 1, minute = -60)
#' # Time difference of 0 secs
#' make_difftime(day = -1)
#' # Time difference of -1 days
#' make_difftime(120, day = -1, units = "minute")
#' # Time differences in mins
make_difftime <- function(num = NULL, units = "auto", ...) {
  pieces <- list(...)
  if (!is.null(num) && length(pieces) > 0) {
    .difftime_from_num(c(num, .difftime_from_pieces(pieces)), units)
  } else if (!is.null(num)) {
    .difftime_from_num(num, units)
  } else if (length(pieces)) {
    .difftime_from_num(.difftime_from_pieces(pieces), units)
  } else {
    stop("No valid values have been passed to 'make_difftime' constructor")
  }
}

difftime_lengths <- c(secs = 1, mins = 60, hours = 3600, days = 86400, weeks = 7 * 86400)

.difftime_from_num <- function(num, units = "auto") {
  seconds <- abs(na.omit(num))
  units <- units[[1]]
  if (units == "auto") {
    if (any(seconds < 60)) {
      units <- "secs"
    } else if (any(seconds < 3600)) {
      units <- "mins"
    } else if (any(seconds < 86400)) {
      units <- "hours"
    } else {
      units <- "days"
    }
  } else {
    units <- standardise_difftime_names(units)
  }
  structure(num / difftime_lengths[[units]], units = units, class = "difftime")
}

.difftime_from_pieces <- function(pieces) {
  names(pieces) <- standardise_difftime_names(names(pieces))
  out <- 0
  for (nm in names(pieces)) {
    out <- out + pieces[[nm]] * difftime_lengths[[nm]]
  }
  out
}

#' Is x a difftime object?
#'
#' @export is.difftime
#' @param x an R object
#' @return TRUE if x is a difftime object, FALSE otherwise.
#' @seealso [is.instant()], [is.timespan()], [is.interval()],
#'   [is.period()].
#' @keywords logic chron
#' @examples
#' is.difftime(as.Date("2009-08-03")) # FALSE
#' is.difftime(make_difftime(days = 12.4)) # TRUE
is.difftime <- function(x) is(x, "difftime")

test_that("is.difftime works as expected", {
  ct_time <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  lt_time <- as.POSIXlt("2009-08-03 13:01:59", tz = "UTC")

  expect_true(!is.difftime(234))
  expect_true(!is.difftime(ct_time))
  expect_true(!is.difftime(lt_time))
  expect_true(!is.difftime(Sys.Date()))
  expect_true(!is.difftime(minutes(1)))
  expect_true(!is.difftime(dminutes(1)))
  expect_true(!is.difftime(interval(ct_time, lt_time)))

  expect_true(is.difftime(make_difftime(1000)))
  expect_true(is.difftime(ct_time - lt_time))
})

test_that("is.difftime handle vectors", {
  expect_true(is.difftime(make_difftime(1:3)))
})

context("Difftimes")

test_that("is.difftime works as expected", {
  ct_time <- as.POSIXct("2008-08-03 13:01:59", tz = "UTC")
  lt_time <- as.POSIXlt("2009-08-03 13:01:59", tz = "UTC")

  expect_true(!is.difftime(234))
  expect_true(!is.difftime(ct_time))
  expect_true(!is.difftime(lt_time))
  expect_true(!is.difftime(Sys.Date()))
  expect_true(!is.difftime(minutes(1)))
  expect_true(!is.difftime(dminutes(1)))
  expect_true(!is.difftime(interval(ct_time, lt_time)))

  expect_true(is.difftime(make_difftime(1000)))
  expect_true(is.difftime(ct_time - lt_time))

})

test_that("is.difftime handle vectors", {
  expect_true(is.difftime(make_difftime(1:3)))
})

#' @include timespans.r
#' @include difftimes.r

check_duration <- function(object) {
  if (is.numeric(object@.Data))
    TRUE
  else
    "Duration value is not a number. Should be numeric."
}


#' Duration class
#'
#' Duration is an S4 class that extends the [Timespan-class] class.
#' Durations record the exact number of seconds in a time span. They measure the
#' exact passage of time but do not always align with measurements
#' made in larger units of time such as hours, months and years.
#' This is because the exact length of larger time units can be affected
#' by conventions such as leap years
#' and Daylight Savings Time.
#'
#' Durations provide a method for measuring generalized timespans when we wish to
#' treat time as a mathematical quantity that increases in a uniform, monotone manner
#' along a continuous numberline. They allow exact comparisons with other durations.
#' See [Period-class] for an alternative way to measure timespans that better
#' preserves clock times.
#'
#' Durations class objects have one slot: .Data, a numeric object equal to the number
#' of seconds in the duration.
#'
#' @aliases durations
#' @export
setClass("Duration", contains = c("Timespan", "numeric"), validity = check_duration)

#' @name hidden_aliases
#' @aliases Compare,Duration,ANY-method Compare,Duration,Duration-method
#'   Compare,difftime,Duration-method Compare,ANY,Duration-method
#'   Compare,Duration,Period-method Compare,Duration,difftime-method
#'   Compare,character,Duration-method Compare,Duration,character-method
#'   Compare,numeric,Duration-method Compare,Duration,numeric-method
#'   as.numeric,Duration-method show,Duration-method c,Duration-method
#'   rep,Duration-method [,Duration-method [<-,Duration,ANY,ANY,ANY-method
#'   [[,Duration-method [[<-,Duration,ANY,ANY,ANY-method $,Duration-method
#'   $<-,Duration-method as.difftime,Duration-method
#'   as.character,Duration-method +,Duration,Duration-method
#'   +,Duration,Interval-method +,Duration,Period-method +,Duration,Date-method
#'   +,Date,Duration-method +,Duration,difftime-method
#'   +,difftime,Duration-method +,Duration,numeric-method
#'   +,numeric,Duration-method +,Duration,POSIXct-method
#'   +,POSIXct,Duration-method +,Duration,POSIXlt-method
#'   +,POSIXlt,Duration-method -,Duration,ANY-method /,Duration,Duration-method
#'   /,Duration,Interval-method /,Duration,Period-method
#'   /,Duration,difftime-method /,difftime,Duration-method
#'   /,Duration,numeric-method /,numeric,Duration-method *,Duration,ANY-method
#'   *,ANY,Duration-method %%,Duration,Duration-method
#'   %%,Duration,Interval-method %%,Duration,Period-method
#'   -,Duration,missing-method -,ANY,Duration-method
NULL

SECONDS_IN_ONE <- c(
  second = 1,
  minute = 60,
  hour   = 3600,
  day    = 86400,
  week   = 604800,
  year   = 31557600)

.readable_duration <- function(x, unit) {
  if (unit == "second")
    paste0(x, "s")
  else {
    x2 <- round(x / SECONDS_IN_ONE[[unit]], 2)
    sprintf("%ss (~%s %ss)", x, x2, unit)
  }
}

.next_unit <- structure(as.list(c(names(SECONDS_IN_ONE[-1]), list(NULL))),
                        names = names(SECONDS_IN_ONE))

compute_estimate <- function (secs, unit = "second") {
  next_unit <- .next_unit[[unit]]
  if (is.null(next_unit))
    return(.readable_duration(secs, "year"))
  out <- character(length(secs))
  tt <- abs(secs) < SECONDS_IN_ONE[[next_unit]]
  if (any(tt))
    out[tt] <- .readable_duration(secs[tt], unit)
  wnext <- which(!tt)
  if (length(wnext))
    out[wnext] <- compute_estimate(secs[wnext], next_unit)
  out
}

#' @export
setMethod("show", signature(object = "Duration"), function(object) {
  if (length(object@.Data) == 0) {
    cat("<Duration[0]>\n")
  } else {
    print(format(object), quote = TRUE)
  }
})

#' @export
format.Duration <- function(x, ...) {
  out <- character(length(x@.Data))
  nnas <- !is.na(x@.Data)
  out[nnas] <- compute_estimate(x@.Data[nnas])
  out[!nnas] <- NA_character_
  out
}

#' @export
setMethod("c", signature(x = "Duration"), function(x, ...) {
  durs <- c(x@.Data, unlist(list(...)))
  new("Duration", durs)
})

#' @export
setMethod("rep", signature(x = "Duration"), function(x, ...) {
  new("Duration", rep(as.numeric(x), ...))
})

#' @export
setMethod("[", signature(x = "Duration"),
  function(x, i, j, ..., drop = TRUE) {
    new("Duration", x@.Data[i])
})

#' @export
setMethod("[[", signature(x = "Duration"),
  function(x, i, j, ..., exact = TRUE) {
    new("Duration", x@.Data[i])
})

#' @export
setMethod("[<-", signature(x = "Duration"),
  function(x, i, j, ..., value) {
    x@.Data[i] <- value
    new("Duration", x@.Data)
})

#' @export
setMethod("[[<-", signature(x = "Duration"),
  function(x, i, j, ..., value) {
    x@.Data[i] <- as.numeric(value)
    new("Duration", x@.Data)
})


#' Create a duration object.
#'
#' `duration()` creates a duration object with the specified values. Entries
#' for different units are cumulative. durations display as the number of
#' seconds in a time span. When this number is large, durations also display an
#' estimate in larger units,; however, the underlying object is always recorded
#' as a fixed number of seconds. For display and creation purposes, units are
#' converted to seconds using their most common lengths in seconds. Minutes = 60
#' seconds, hours = 3600 seconds, days = 86400 seconds, weeks = 604800. Units
#' larger than weeks are not used due to their variability.
#'
#' Durations record the exact number of seconds in a time span. They measure the
#' exact passage of time but do not always align with measurements
#' made in larger units of time such as hours, months and years.
#' This is because the length of larger time units can be affected
#' by conventions such as leap years
#' and Daylight Savings Time. Base R provides a second class for measuring
#' durations, the difftime class.
#'
#' Duration objects can be easily created with the helper functions [dweeks()],
#' [ddays()], [dminutes()], [dseconds()]. These objects can be added to and
#' subtracted to date- times to create a user interface similar to object
#' oriented programming.
#'
#' @param num the number or a character vector of time units. In string
#'   representation all unambiguous name units and abbreviations and ISO 8601
#'   formats are supported; 'm' stands for month and 'M' for minutes unless ISO
#'   8601 "P" modifier is present (see examples). Fractional units are
#'   supported.
#' @param units a character string that specifies the type of units that `num`
#'   refers to. When `num` is character, this argument is ignored.
#' @param ... a list of time units to be included in the duration and their
#'   amounts. Seconds, minutes, hours, days, weeks, months and years are
#'   supported. Durations of months and years assume that year consists of
#'   365.25 days.
#' @param x numeric value of the number of units to be contained in the
#'   duration.
#' @return a duration object
#' @seealso [as.duration()] [Duration-class]
#' @keywords chron classes
#' @examples
#'
#' ### Separate period and units vectors
#'
#' duration(90, "seconds")
#' duration(1.5, "minutes")
#' duration(-1, "days")
#'
#' ### Units as arguments
#'
#' duration(day = -1)
#' duration(second = 90)
#' duration(minute = 1.5)
#' duration(mins = 1.5)
#' duration(second = 3, minute = 1.5, hour = 2, day = 6, week = 1)
#' duration(hour = 1, minute = -60)
#'
#' ### Parsing
#'
#' duration("2M 1sec")
#' duration("2hours 2minutes 1second")
#' duration("2d 2H 2M 2S")
#' duration("2days 2hours 2mins 2secs")
#' # Missing numerals default to 1. Repeated units are added up.
#' duration("day day")
#'
#' ### ISO 8601 parsing
#'
#' duration("P3Y6M4DT12H30M5S")
#' duration("P23DT23H") # M stands for months
#' duration("10DT10M") # M stands for minutes
#' duration("P23DT60H 20min 100 sec") # mixing ISO and lubridate style parsing
#'
#' # Comparison with characters (from v1.6.0)
#'
#' duration("day 2 sec") > "day 1sec"
#'
#'
#' ## ELEMENTARY CONSTRUCTORS:
#'
#' dseconds(1)
#' dminutes(3.5)
#'
#' x <- ymd_hms("2009-08-03", tz="America/Chicago")
#' x + ddays(1) + dhours(6) + dminutes(30)
#' x + ddays(100) - dhours(8)
#'
#' class(as.Date("2009-08-09") + ddays(1)) # retains Date class
#' as.Date("2009-08-09") + dhours(12)
#' class(as.Date("2009-08-09") + dhours(12))
#' # converts to POSIXt class to accomodate time units
#'
#' dweeks(1) - ddays(7)
#' c(1:3) * dhours(1)
#'
#' # compare DST handling to durations
#' boundary <- ymd_hms("2009-03-08 01:59:59", tz="America/Chicago")
#' boundary + days(1) # period
#' boundary + ddays(1) # duration
#' @export
duration <- function(num = NULL, units = "seconds", ...) {
  if (is.character(num)) {
    as.duration(parse_period(num))
  } else {
    out1 <- .duration_from_num(num, units)
    out2 <- .duration_from_pieces(list(...))
    if (is.null(out1) && is.null(out2)) new("Duration", numeric())
    else if (is.null(out1)) out2
    else if (is.null(out2)) out1
    else c(out1, out2)
  }
}

average_durations <- c(second = 1, minute = 60, hour = 3600, mday = 86400,
                       wday = 86400, yday = 86400, day = 86400, week = 604800,
                       month = 60 * 60 * 24 * 365.25 / 12, year = 60 * 60 * 24 * 365.25)

.duration_from_num <- function(num, units) {
  if (length(num) == 0)
    return(NULL)

  if (!is.numeric(num)) {
    stop(sprintf("First argument to `duration()` constructor must be character or numeric. Supplied object of class '%s'", class(num)))
  }

  ## qucik check for common wrongdoings: #462
  if (inherits(num, c("Interval", "Period")))
    stop("Interval or Period objects cannot be used as input to `duration()` constructor. Use `as.duration()` instead.", call. = FALSE)

  unit <- standardise_date_names(units)
  new("Duration", num * unname(average_durations[unit]))
}

.duration_from_pieces <- function(pieces) {
  if (length(pieces) == 0)
    return(NULL)
  names(pieces) <- standardise_date_names(names(pieces))
  out <- 0
  for (nm in names(pieces))
    out <- out + pieces[[nm]] * average_durations[[nm]]
  new("Duration", out)
}

#' @export dseconds dminutes dmonths dhours ddays dweeks dyears dmilliseconds dmicroseconds dnanoseconds dpicoseconds
#' @rdname duration
dseconds <- function(x = 1) new("Duration", x)
#' @rdname duration
dminutes <- function(x = 1) new("Duration", x * 60)
#' @rdname duration
dhours <- function(x = 1) new("Duration", x * 3600)
#' @rdname duration
ddays <- function(x = 1) new("Duration", x * 86400)
#' @rdname duration
dweeks <- function(x = 1) new("Duration", x * average_durations[["week"]])
#' @rdname duration
dmonths <- function(x = 1) new("Duration", x * average_durations[["month"]])
#' @rdname duration
dyears <- function(x = 1) new("Duration", x * average_durations[["year"]])
#' @rdname duration
dmilliseconds <- function(x = 1) new("Duration", x / 1000)
#' @rdname duration
dmicroseconds <- function(x = 1) new("Duration", x / 1000 / 1000)
#' @rdname duration
dnanoseconds <- function(x = 1) new("Duration", x / 1000 / 1000 / 1000)
#' @rdname duration
dpicoseconds <- function(x = 1) new("Duration", x / 1000 / 1000 / 1000 / 1000)

#' @rdname duration
#' @export
#' @examples
#' is.duration(as.Date("2009-08-03")) # FALSE
#' is.duration(duration(days = 12.4)) # TRUE
is.duration <- function(x) is(x, "Duration")

#' @export
summary.Duration <- function(object, ...) {
  nas <- is.na(object)
  object <- object[!nas]
  nums <- as.numeric(object)
  qq <- stats::quantile(nums)
  qq <- c(qq[1L:3L], mean(nums), qq[4L:5L])
  qq <- dseconds(qq)
  qq <- as.character(qq)
  names(qq) <- c("Min.", "1st Qu.", "Median", "Mean", "3rd Qu.",
                 "Max.")
  if (any(nas))
    c(qq, `NA's` = sum(nas))
  else qq
}

#' @export
setMethod("Compare", c(e1 = "Duration", e2 = "ANY"),
  function(e1, e2) stop_incompatible_classes(e1, e2, .Generic)
)

#' @export
setMethod("Compare", c(e1 = "ANY", e2 = "Duration"),
          function(e1, e2) {
            stop(sprintf("Incompatible duration classes (%s, %s). Please coerce with `as.duration()`.",
                         class(e1), class(e2)),
                 call. = FALSE)
          })

#' @export
setMethod("Compare", signature(e1 = "Duration", e2 = "numeric"),
          function(e1, e2) {
            callGeneric(e1@.Data, e2)
          })

#' @export
setMethod("Compare", signature(e1 = "numeric", e2 = "Duration"),
          function(e1, e2) {
            callGeneric(e1, e2@.Data)
          })

#' @export
setMethod("Compare", signature(e1 = "Duration", e2 = "character"),
          function(e1, e2) {
            callGeneric(e1, as.duration(e2))
          })

#' @export
setMethod("Compare", signature(e1 = "character", e2 = "Duration"),
          function(e1, e2) {
            callGeneric(as.duration(e1), e2)
          })

#' @export
setMethod("Compare", c(e1 = "difftime", e2 = "Duration"),
          function(e1, e2) {
            callGeneric(as.numeric(e1, "secs"), e2@.Data)
          })

#' @export
setMethod("Compare", c(e1 = "Duration", e2 = "difftime"),
          function(e1, e2) {
            callGeneric(e1@.Data, as.numeric(e2, "secs"))
          })

#' @export
setMethod("Compare", c(e1 = "Duration", e2 = "Duration"),
          function(e1, e2) {
            callGeneric(e1@.Data, e2@.Data)
          })

#' @include periods.r
NULL

#' Get/set years component of a date-time
#'
#' Date-time must be a POSIXct, POSIXlt, Date, Period or any other object
#' convertible to POSIXlt.
#'
#' year does not yet support years before 0 C.E.
#'
#' @param x a date-time object
#' @param value a numeric object
#' @return the years element of x as a decimal number
#' @keywords utilities manip chron methods
#' @references
#'    \url{https://en.wikipedia.org/wiki/ISO_week_date}
#' @examples
#' x <- ymd("2012-03-26")
#' year(x)
#' year(x) <- 2001
#' year(x) > 1995
#' @export
year <- function(x) {
  UseMethod("year")
}

#' @export
year.default <- function(x) {
  as.POSIXlt(x, tz = tz(x))$year + 1900
}

#' @export
year.Period <- function(x) {
  slot(x, "year")
}

#' @rdname year
#' @export
setGeneric("year<-",
  function (x, value) standardGeneric("year<-"),
  useAsDefault = function(x, value) {
    y <- update_datetime(as.POSIXct(x), years = value, roll_month = "NAym")
    reclass_date(y, x)
  }
)

#' @export
setMethod("year<-", "Duration", function(x, value) {
  x <- x + years(value - year(x))
})

#' @export
setMethod("year<-", signature("Period"), function(x, value) {
  slot(x, "year") <- value
  x
})

#' @export
setMethod("year<-", signature("Interval"), function(x, value) {
  x <- x + years(value - year(x))
})

#' @export
setMethod("year<-", "POSIXt", function(x, value) {
  update_datetime(x, years = value, roll_month = "NAym")
})

#' @export
setMethod("year<-", "Date", function(x, value) {
  update_datetime(x, years = value, roll_month = "NAym")
})

.other_year <- function(x, week_start = 1) {
  x <- as.POSIXlt(x)
  date <- make_date(year(x), month(x), day(x))
  isodate <- date + ddays(4 - wday(date, week_start = week_start))
  year(isodate)
}

#' @rdname year
#' @description
#' `isoyear()` returns years according to the ISO 8601 week calendar.
#' @export
isoyear <- function(x) {
  .other_year(x, 1)
}

#' @rdname year
#' @description
#' `epiyear()` returns years according to the epidemiological week calendars.
#' @export
epiyear <- function(x) {
  .other_year(x, 7)
}

#' Get **d**aylight **s**avings **t**ime indicator of a date-time
#'
#' Date-time must be a POSIXct, POSIXlt, Date, chron, yearmon, yearqtr, zoo,
#' zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
#'
#' A date-time's daylight savings flag can not be set because it depends on the
#' date-time's year, month, day, and hour values.
#'
#' @export
#' @param x a date-time object
#' @return A logical. TRUE if DST is in force, FALSE if not, NA if unknown.
#' @keywords utilities chron methods
#' @examples
#' x <- ymd("2012-03-26")
#' dst(x)
dst <- function(x)
  UseMethod("dst")

#' @export
dst.default <- function(x)
  c(NA, FALSE, TRUE)[as.POSIXlt(x)$isdst + 2]

setOldClass("POSIXt")
setOldClass("POSIXct")
setOldClass("POSIXlt")

#' Various POSIX utilities
#'
#' [POSIXct()] mirrors primitive contructors in base R ([double()],
#' [character()] etc.)
#'
#' @rdname posix_utils
#' @param x an R object
#' @param length A non-negative number specifying the desired length. Supplying
#'   an argument of length other than one is an error.
#' @param tz a timezone (defaults to "utc")
#' @return TRUE if x is a POSIXct or POSIXlt object, FALSE otherwise.
#' @seealso [is.instant()], [is.timespan()], [is.Date()]
#' @keywords logic chron
#' @examples
#' is.POSIXt(as.Date("2009-08-03"))
#' is.POSIXt(as.POSIXct("2009-08-03"))
#' @export
is.POSIXt <- function(x) inherits(x, "POSIXt")

#' @rdname posix_utils
#' @export
is.POSIXlt <- function(x) inherits(x, "POSIXlt")

#' @rdname posix_utils
#' @export
is.POSIXct <- function(x) inherits(x, "POSIXct")

#' @rdname posix_utils
#' @export
POSIXct <- function(length = 0L, tz = "UTC") {
  t0 <- if (is_utc(tz)) origin else force_tz(origin, tz)
  .POSIXct(rep.int(t0, length), tz = tz)
}

#' @rdname posix_utils
#' @export
NA_POSIXct_ <- .POSIXct(NA_real_, tz = "UTC")

.recursive_posixct_unclass <- function(x, tz = "UTC") {
  if (length(x) == 0) {
    .POSIXct(numeric(), tz = tz)
  } else if (is.POSIXlt(x)) {
    with_tz(as.POSIXct(x), tz)
  } else if (is.recursive(x)) {
    lapply(x, .recursive_posixct_unclass, tz = tz)
  } else {
    as_datetime(x, tz = tz)
  }
}

#' @method c POSIXct
c.POSIXct <- function(..., recursive = FALSE) {
  dots <- list(...)
  tz <- tz(dots[[1]])
  .POSIXct(c(unlist(lapply(dots, .recursive_posixct_unclass, tz = tz))),
    tz = tz
  )
}

#' @method c POSIXlt
c.POSIXlt <- function(..., recursive = FALSE) {
  as.POSIXlt(do.call("c.POSIXct", lapply(list(...), as.POSIXct)))
}

evalqOnLoad({
  registerS3method("c", "POSIXct", c.POSIXct)
  ## registerS3method("c", "POSIXlt", c.POSIXlt)
})

#' @name hidden_aliases
#' @aliases day<-,POSIXt-method hour<-,POSIXt-method minute<-,POSIXt-method
#'   month<-,POSIXt-method second<-,POSIXt-method year<-,POSIXt-method
NULL

#' @include periods.r
NULL

#' Get/set hours component of a date-time
#'
#' Date-time must be a POSIXct, POSIXlt, Date, Period, chron, yearmon, yearqtr, zoo,
#' zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
#'
#' @export
#' @param x a date-time object
#' @param value numeric value to be assigned to the `hour` component
#' @keywords utilities manip chron methods
#' @return the hours element of x as a decimal number
#' @examples
#' x <- ymd("2012-03-26")
#' hour(x)
#' hour(x) <- 1
#' hour(x) <- 25
#' hour(x) > 2
hour <- function(x) {
  UseMethod("hour")
}

#' @export
hour.default <- function(x) {
  as.POSIXlt(x, tz = tz(x))$hour
}

#' @export
hour.Period <- function(x) {
  slot(x, "hour")
}


#' @rdname hour
#' @export
setGeneric("hour<-",
  function (x, value) standardGeneric("hour<-"),
  useAsDefault = function(x, value) {
    y <- update_datetime(as.POSIXct(x), hours = value)
    reclass_date(y, x)
  }
)

#' @export
setMethod("hour<-", "Duration", function(x, value) {
  x <- x + hours(value - hour(x))
})

#' @export
setMethod("hour<-", signature("Period"), function(x, value) {
  slot(x, "hour") <- value
  x
})

#' @export
setMethod("hour<-", "Interval", function(x, value) {
  x <- x + hours(value - hour(x))
})

#' @export
setMethod("hour<-", "POSIXt", function(x, value) {
  update_datetime(x, hours = value)
})

#' @export
setMethod("hour<-", "Date", function(x, value) {
  update_datetime(x, hours = value)
})

##' Format dates and times based on human-friendly templates
##'
##' Stamps are just like [format()], but based on human-friendly
##' templates like "Recorded at 10 am, September 2002" or "Meeting, Sunday May
##' 1, 2000, at 10:20 pm".
##'
##' `stamp()` is a stamping function date-time templates mainly, though it
##' correctly handles all date and time formats as long as they are
##' unambiguous. `stamp_date()`, and `stamp_time()` are the specialized
##' stamps for dates and times (MHS). These function might be useful when the
##' input template is unambiguous and matches both a time and a date format.
##'
##' Lubridate tries hard to guess the formats, but often a given format can be
##' interpreted in multiple ways. One way to deal with such cases is to provide
##' unambiguous formats like 22/05/81 instead of 10/05/81 for d/m/y
##' format. Another way is to use a more specialized [`stamp_date`] and
##' [`stamp_time`]. The core function `stamp()` prioritizes longer date-time
##' formats.
##'
##' If `x` is a vector of values \pkg{lubridate} will choose the format which
##' "fits" `x` the best. Note that longer formats are preferred. If you have
##' "22:23:00 PM" then "HMSp" format will be given priority to shorter "HMS"
##' order which also fits the supplied string.
##'
##' Finally, you can give desired format order directly as `orders`
##' argument.
##'
##' @param x a character vector of templates.
##' @param orders orders are sequences of formatting characters which might be
##' used for disambiguation. For example "ymd hms", "aym" etc. See
##' [guess_formats()] for a list of available formats.
##' @param locale locale in which `x` is encoded. On Linux-like systems use
##' `locale -a` in the terminal to list available locales.
##' @param quiet whether to output informative messages.
##' @param exact logical. If `TRUE`, the `orders` parameter is interpreted as an
##'   exact [base::strptime()] format and no format guessing is performed.
##' @return a function to be applied on a vector of dates
##' @seealso [guess_formats()], [parse_date_time()], [strptime()]
##' @export
##' @examples
##' D <- ymd("2010-04-05") - days(1:5)
##' stamp("March 1, 1999")(D)
##' sf <- stamp("Created on Sunday, Jan 1, 1999 3:34 pm")
##' sf(D)
##' stamp("Jan 01")(D)
##' stamp("Sunday, May 1, 2000", locale = "C")(D)
##' stamp("Sun Aug 5")(D) #=> "Sun Aug 04" "Sat Aug 04" "Fri Aug 04" "Thu Aug 04" "Wed Aug 03"
##' stamp("12/31/99")(D)              #=> "06/09/11"
##' stamp("Sunday, May 1, 2000 22:10", locale = "C")(D)
##' stamp("2013-01-01T06:00:00Z")(D)
##' stamp("2013-01-01T00:00:00-06")(D)
##' stamp("2013-01-01T00:00:00-08:00")(force_tz(D, "America/Chicago"))
stamp <- function(x, orders = lubridate_formats,
                  locale = Sys.getlocale("LC_TIME"),
                  quiet = FALSE, exact = FALSE) {
  if (exact) {
    stopifnot(length(orders) > 0)
    fmts <- orders
  } else {
    fmts <- unique(guess_formats(x, orders, locale))
    ## remove internal lubridate formats except of the ISO8601 time zone formats
    fmts <- grep("%O[^oOzudHImMUVwWy]", fmts, value = TRUE, invert = TRUE)
    if (length(fmts) == 0)
      stop(sprintf("Couldn't guess formats for: %s. Wrong locale? Consider using `exact = TRUE`.", x[[1]]))
  }
  if (length(fmts) == 1L) {
    FMT <- fmts[[1]]
  } else {
    trained <- .train_formats(x, fmts, locale = locale)
    formats <- .select_formats(trained)
    FMT <- formats[[1]]
    if (!quiet && length(trained) > 1) {
      message(
        "Multiple formats matched: ",
        paste("\"", names(trained), "\"(", trained, ")",
          sep = "",
          collapse = ", "
        )
      )
    }
  }

  if (!quiet) {
    message("Using: \"", FMT, "\"")
  }


  ## format doesn't accept 'locale' argument; need a hard reset
  reset_local_expr <-
    quote({
      old_lc_time <- Sys.getlocale("LC_TIME")
      if (old_lc_time != locale) {
        on.exit(Sys.setlocale("LC_TIME", old_lc_time))
        Sys.setlocale("LC_TIME", locale)
      }
    })

  ## ISO8601
  ## %Ou: "2013-04-16T04:59:59Z"
  ## %Oo: "2013-04-16T04:59:59+01"
  ## %Oz: "2013-04-16T04:59:59+0100"
  ## %OO: "2013-04-16T04:59:59+01:00"

  ## Is a timezone format?
  if (grepl("%O[oOzu]|%z", FMT)) {
    ## We need to post-process x in the case of %Oo, %OO and %Oz formats
    ## because standard %z output format ignores timezone.

    ## Post-process only when at the end of the string, otherwise don't bother
    ## and just output with %z format after a conversion to UTC.

    # replicate str_extract(FMT, "%O[oOz]$") without stringr dependence.
    # must return NA if no match.
    oOz_end <- ifelse(grepl("%O[oOz]$", FMT), gsub("^.*(%O[oOz]$)", "\\1", FMT), rep(NA, length(FMT)))

    if (is.na(oOz_end)) {
      FMT <- sub(
        "%O[oOz]", "%z",
        sub("%Ou", "Z", FMT, fixed = TRUE)
      )

      out <- eval(bquote(
        function(x, locale = .(locale)) {
          ## %z ignores timezone
          if (!is_utc(tz(x[[1L]]))) {
            x <- with_tz(x, tzone = "UTC")
          }
          .(reset_local_expr)
          format(x, format = .(FMT))
        }
      ))
    } else {
      FMT <- sub("%O[oOz]$", "", FMT)

      out <- eval(bquote(
        function(x, locale = .(locale)) {
          .(reset_local_expr)
          paste0(
            format(x, format = .(FMT)),
            .format_offset(x, fmt = .(oOz_end))
          )
        }
      ))
    }
  } else {
    ## most common case
    out <- eval(bquote(function(x, locale = .(locale)) {
      .(reset_local_expr)
      format(x, format = .(FMT))
    }))
  }

  attr(out, "srcref") <- NULL
  out
}


.format_offset <- function(x, fmt = "%Oz") {
  ## .format_offset
  ##
  ## function to format the offset of a time from UTC
  ##
  ## This is an internal function, used in conjunction with [stamp()].
  ## There are three available formats:
  ##
  ## \itemize{
  ##   \item \code{\%Oo} +01
  ##   \item \code{\%Oz} +0100
  ##   \item \code{\%OO} +01:00
  ## }
  ##
  ## If the \code{\%Oo} format is used for a half-hour timezone, a warning
  ## is issued, and the format is changed to \code{\%Oz}
  ##
  ## @param x      POSIXct for which offset-string is sought
  ## @param fmt    string describing format of offset, default: \code{\%Oz}
  ##
  ## @return string
  ##

  ## "%Oo"  +01
  ## "%Oz"  +0100
  ## "%OO"  +01:00

  ## calculate offset by forcing this time as utc
  dtm_utc <- force_tz(x, tzone = "UTC")

  ## the offset is the duration represented by the difference in time
  offset_duration <- difftime(dtm_utc, x, units = "secs")

  ## determine sign
  .sgn <- ifelse(offset_duration >= as.duration(0), "+", "-")

  ## remove sign
  offset_duration <- abs(offset_duration)

  ## determine hour
  .hr <- floor(offset_duration / dhours(1))

  ## determine minutes
  .min <- floor((offset_duration - dhours(.hr)) / dminutes(1))

  ## warning if we need minutes, but are using format without minutes
  if (any(.min > 0) & fmt == "%Oo") {
    warning("timezone offset-minutes are non-zero - changing format to %Oz")
    fmt <- "%Oz"
  }

  result <- switch(fmt,
    "%Oo" = sprintf("%s%02d", .sgn, .hr),
    "%Oz" = sprintf("%s%02d%02d", .sgn, .hr, .min),
    "%OO" = sprintf("%s%02d:%02d", .sgn, .hr, .min)
  )

  return(result)
}

##' @rdname stamp
##' @export
stamp_date <- function(x, locale = Sys.getlocale("LC_TIME"), quiet = FALSE) {
  stamp(x,
    orders = c("ymd", "dmy", "mdy", "ydm", "dym", "myd", "my", "ym", "md", "dm", "m", "d", "y"),
    locale = locale, quiet = quiet
  )
}

##' @rdname stamp
##' @export
stamp_time <- function(x, locale = Sys.getlocale("LC_TIME"), quiet = FALSE) {
  stamp(x,
    orders = c("hms", "hm", "ms", "h", "m", "s"),
    locale = locale, quiet = quiet
  )
}


lubridate_formats <- local({
  xxx <- c("ymd", "ydm", "mdy", "myd", "dmy", "dym")
  names(xxx) <- xxx
  out <- character()
  for (D in xxx) {
    out[[paste(D, "_hms", sep = "")]] <- paste(xxx[[D]], "T", sep = "")
    out[[paste(D, "_hm", sep = "")]] <- paste(xxx[[D]], "R", sep = "")
    out[[paste(D, "_h", sep = "")]] <- paste(xxx[[D]], "r", sep = "")
  }

  out <- c(out, xxx,
    my = "my", ym = "ym", md = "md", dm = "dm",
    hms = "T", hm = "R", ms = "MS", h = "r", m = "m", y = "y"
  )

  ## adding ISO8601
  out <- c(ymd_hmsz = "ymdTz", out)

  out
})

#' Changes the components of a date object
#'
#' `update.Date()` and `update.POSIXt()` return a date with the specified
#' elements updated.  Elements not specified will be left unaltered. update.Date
#' and update.POSIXt do not add the specified values to the existing date, they
#' substitute them for the appropriate parts of the existing date.
#'
#'
#' @name DateTimeUpdate
#' @param object a date-time object
#' @param ... named arguments: years, months, ydays, wdays, mdays, days, hours,
#'   minutes, seconds, tzs (time zone component)
#' @param roll logical. If `TRUE`, and the resulting date-time lands on a
#'   non-existent civil time instant (DST, 29th February, etc.) roll the date
#'   till next valid point. When `FALSE`, the default, produce NA for non
#'   existing date-times.
#' @param week_start week starting day (Default is 7, Sunday). Set
#'   `lubridate.week.start` option to control this.
#' @param simple logical. Deprecated. Same as `roll`.
#' @return a date object with the requested elements updated. The object will
#'   retain its original class unless an element is updated which the original
#'   class does not support. In this case, the date returned will be a POSIXlt
#'   date object.
#' @keywords manip chron
#' @examples
#' date <- ymd("2009-02-10")
#' update(date, year = 2010, month = 1, mday = 1)
#'
#' update(date, year =2010, month = 13, mday = 1)
#'
#' update(date, minute = 10, second = 3)
#' @export
update.POSIXt <- function(object, ..., roll = FALSE,
                          week_start = getOption("lubridate.week.start", 7),
                          simple = NULL) {
  if (!is.null(simple)) roll <- simple
  do.call(update_date_time, c(list(object, roll = roll, week_start = week_start),
                              list(...)))
}

update_date_time <- function(object, years = integer(), months = integer(),
                             days = integer(), mdays = integer(), ydays = integer(), wdays = integer(),
                             hours = integer(), minutes = integer(), seconds = double(), tzs = NULL,
                             roll = FALSE, week_start = 7) {

  if (!length(object)) return(object)

  if (length(days) > 0) mdays = days;
  updates <- list(year = years, month = months,
                  yday = ydays, mday = mdays, wday = wdays,
                  hour = hours, minute = minutes, second = seconds)
  maxlen <- max(unlist(lapply(updates, length)))

  if (maxlen > 1) {
    for (nm in names(updates)) {
      len <- length(updates[[nm]])
      ## len == 1 is treated at C_level
      if (len != maxlen && len > 1)
        updates[[nm]] <- rep_len(updates[[nm]], maxlen)
    }
  }

  if (is.null(tzs))
    tzs <- tz(object)

  ## todo: check if the following lines make any unnecessary copies
  updates[["dt"]] <- as.POSIXct(object)
  updates[["roll"]] <- roll
  updates[["tz"]] <- tzs
  updates[["week_start"]] <- week_start
  reclass_date(do.call(C_update_dt, updates), object)
}

## prior to v1.7.0
update_posixt_old <- function(object, ..., simple = FALSE) {

  if (!length(object)) return(object)
  date <- as.POSIXlt(object)

  ## adjudicate units input
  units <- list(...)
  names(units) <- standardise_lt_names(names(units))

  if (!is.null(new_tz <- units$tz)) {
    units$tz <- NULL
  }

  day.units <- c("day", "wday", "mday", "yday")
  wunit <- day.units %in% names(units)

  if (n <- sum(wunit)) {
    if (n > 1) stop("conflicting days input")
    uname <- day.units[wunit]
    if (uname != "mday") {
      ## we compute everything with mdays (operating with ydays doesn't work)
      if (uname != "day") {
        if (uname == "yday" & !is.null(units$year))
          warning("Updating on both 'year' and 'yday' can lead to wrong results. See bug #319.", call. = F)
        diff <- units[[uname]] - date[[uname]] - 1
        units[[uname]] <- diff + date$mday
      }
      names(units)[names(units) == uname] <- "mday"
    }
  }

  if (!is.null(units$mon)) units$mon <- units$mon - 1
  if (!is.null(units$year)) units$year <- units$year - 1900

  ## make new date-times
  date <- unclass(date)

  date[names(units)] <- units
  date[c("wday", "yday")] <- list(wday = NA, yday = NA)

  ## unbalanced POSIXlt often results in R crashes
  maxlen <- max(unlist(lapply(date, length)))

  if (maxlen > 1) {
    for (nm in names(date))
 if (length(date[[nm]]) != maxlen)
        date[[nm]] <- rep_len(date[[nm]], maxlen)
  }

  class(date) <- c("POSIXlt", "POSIXt")
  if (!is.null(new_tz))
    attr(date, "tzone") <- new_tz

  ## fit to timeline
  ## POSIXct format avoids negative and NA elements in POSIXlt format
  fit_to_timeline(date, class(object)[[1]], simple = simple)
}

#' @export
update.Date <- function(object, ...) {

  ct <- as_datetime(object, tz = "UTC")
  new <- update(ct, ...)
  ## fixme: figure out a way to avoid this, or write specialized update for Date
  new_lt <- as.POSIXlt(new, tz = "UTC")
  if (sum(c(new_lt$hour, new_lt$min, new_lt$sec), na.rm = TRUE)) {
    new
  } else {
    make_date(new_lt$year + 1900, new_lt$mon + 1, new_lt$mday)
  }
}

#' Fit a POSIXlt date-time to the timeline
#'
#' The POSIXlt format allows you to create instants that do not exist in real
#' life due to daylight savings time and other conventions. fit_to_timeline
#' matches POSIXlt date-times to a real times. If an instant does not exist, fit
#' to timeline will replace it with an NA. If an instant does exist, but has
#' been paired with an incorrect timezone/daylight savings time combination,
#' fit_to_timeline returns the instant with the correct combination.
#'
#'
#' @param lt a POSIXlt date-time object.
#' @param class a character string that describes what type of object to return,
#'   POSIXlt or POSIXct. Defaults to POSIXct. This is an optimization to avoid
#'   needless conversions.
#' @param simple if TRUE, \pkg{lubridate} makes no attempt to detect
#'   meaningless time-dates or to correct time zones. No NAs are produced and
#'   the most meaningful valid dates are returned instead. See examples.
#' @return a POSIXct or POSIXlt object that contains no illusory date-times
#'
#' @examples
#' \dontrun{
#'
#' tricky <- structure(list(sec   = c(5,    0,    0,    -1),
#'                          min   = c(0L,   5L,   5L,   0L),
#'                          hour  = c(2L,   0L,   2L,   2L),
#'                          mday  = c(4L,   4L,   14L,  4L),
#'                          mon   = c(10L,  10L,  2L,   10L),
#'                          year  = c(112L, 112L, 110L, 112L),
#'                          wday  = c(0L,   0L,   0L,   0L),
#'                          yday  = c(308L, 308L, 72L,  308L),
#'                          isdst = c(1L,   0L,   0L,   1L)),
#'                     .Names = c("sec", "min", "hour", "mday", "mon",
#'                                "year", "wday", "yday",  "isdst"),
#'                     class = c("POSIXlt", "POSIXt"),
#'                     tzone = c("America/Chicago", "CST", "CDT"))
#'
#' tricky
#' ## [1] "2012-11-04 02:00:00 CDT" Doesn't exist because clocks "fall back" to 1:00 CST
#' ## [2] "2012-11-04 00:05:00 CST" Times are still CDT, not CST at this instant
#' ## [3] "2010-03-14 02:00:00 CDT" DST gap
#' ## [4] "2012-11-04 01:59:59 CDT" Does exist, but has deceptive internal structure
#'
#' fit_to_timeline(tricky)
#' ## Returns:
#' ## [1] "2012-11-04 02:00:00 CST" instant paired with correct tz & DST combination
#' ## [2] "2012-11-04 00:05:00 CDT" instant paired with correct tz & DST combination
#' ## [3] NA - fake time changed to NA (compare to as.POSIXct(tricky))
#' ## [4] "2012-11-04 01:59:59 CDT" -real instant, left as is
#'
#' fit_to_timeline(tricky, simple = TRUE)
#' ## Returns valid time-dates by extrapolating CDT and CST zones:
#' ## [1] "2012-11-04 01:00:05 CST" "2012-11-04 01:05:00 CDT"
#' ## [3] "2010-03-14 03:05:00 CDT" "2012-11-04 01:59:59 CDT"
#' }
#' @export
fit_to_timeline <- function(lt, class = "POSIXct", simple = FALSE) {
  if (class != "POSIXlt" && class != "POSIXct")
    stop("class argument must be POSIXlt or POSIXct")

  if (simple) {

    if (class == "POSIXct") as.POSIXct(lt)
    else as.POSIXlt(as.POSIXct(lt))

  } else {

    ## fall break - DST only changes if it has to
    ct <- as.POSIXct(lt)
    lt2 <- as.POSIXlt(ct)

    dstdiff <- !is.na(ct) & (lt$isdst != lt2$isdst)
    if (any(dstdiff)) {

      dlt <- lt[dstdiff]
      dlt2 <- lt2[dstdiff]
      dlt$isdst <- dlt2$isdst
      dlt$zone <- dlt2$zone
      dlt$gmtoff <- dlt2$gmtoff
      dct <- as.POSIXct(dlt) # should directly match if not in gap

      if (class == "POSIXct")
        ct[dstdiff] <- dct
      else
        lt2[dstdiff] <- dlt

      chours <- format.POSIXlt(as.POSIXlt(dct), "%H", usetz = FALSE)
      lhours <- format.POSIXlt(dlt, "%H", usetz = FALSE)

      any <- any(hdiff <- chours != lhours)
      if (!is.na(any) && any) {
        if (class == "POSIXct")
          ct[dstdiff][hdiff] <- NA
        else
          lt2[dstdiff][hdiff] <- NA
      }
    }
    if (class == "POSIXct") ct else lt2
  }
}

#' Computes attractive axis breaks for date-time data
#'
#' pretty.dates indentifies which unit of time the sub-intervals should be
#' measured in to provide approximately n breaks. It then chooses a "pretty"
#' length for the sub-intervals and sets start and endpoints that 1) span the
#' entire range of the data, and 2) allow the breaks to occur on important
#' date-times (i.e. on the hour, on the first of the month, etc.)
#'
#' @param x a vector of POSIXct, POSIXlt, Date, or chron date-time objects
#' @param n integer value of the desired number of breaks
#' @param ... additional arguments to pass to function
#' @return a vector of date-times that can be used as axis tick marks or bin breaks
#' @keywords dplot utilities chron
#' @export
#' @examples
#' x <- seq.Date(as.Date("2009-08-02"), by = "year", length.out = 2)
#' pretty_dates(x, 12)
pretty_dates <- function(x, n, ...) {
  otz <- Sys.getenv("TZ")
  if (Sys.getenv("TZ") == "") otz <- "unset"
  Sys.setenv(TZ = tz(x[1]))
  on.exit(if (otz == "unset") {
    Sys.unsetenv("TZ")
  } else {
    Sys.setenv(TZ = otz)
  })

  rng <- range(x)
  diff <- difftime(rng[2], rng[1], units = "secs")

  binunits <- pretty_unit(diff / n)

  f <- get(paste("pretty", binunits, sep = "_"), mode = "function")
  binlength <- f(diff, n)

  start <- pretty_point(min(rng), binunits, binlength)
  end <- pretty_point(max(rng), binunits, binlength, start = FALSE)

  breaks <- seq.POSIXt(start, end, paste(binlength, binunits))
  breaks
}

pretty_unit <- function(x, ...) {
  if (x > 3600 * 24 * 365) {
    return("year")
  }
  if (x > 3600 * 24 * 30) {
    return("month")
  }
  if (x > 3600 * 24) {
    return("day")
  }
  if (x > 3600) {
    return("hour")
  }
  if (x > 60) {
    return("min")
  } else {
    return("sec")
  }
}

pretty_sec <- function(x, n, ...) {
  lengths <- c(1, 2, 5, 10, 15, 30, 60)
  fit <- abs(x - lengths * n)
  lengths[which.min(fit)]
}

pretty_min <- function(x, n, ...) {
  span <- x / 60
  lengths <- c(1, 2, 5, 10, 15, 30, 60)
  fit <- abs(span - lengths * n)
  lengths[which.min(fit)]
}

pretty_hour <- function(x, n, ...) {
  span <- x / 3600
  lengths <- c(1, 2, 3, 4, 6, 8, 12, 24)
  fit <- abs(span - lengths * n)
  lengths[which.min(fit)]
}

pretty_day <- function(x, n, ...) {
  span <- x / (3600 * 24)
  pretty(1:span, n = n)[2]
}

pretty_month <- function(x, n, ...) {
  span <- x / (3600 * 24 * 30)
  lengths <- c(1, 2, 3, 4, 6, 12)
  fit <- abs(span - lengths * n)
  lengths[which.min(fit)]
}

pretty_year <- function(x, n, ...) {
  span <- x / (3600 * 24 * 365)
  pretty(1:span, n = n)[2]
}

pretty_point <- function(x, units, length, start = TRUE, ...) {
  x <- as.POSIXct(x)

  if (units %in% c("day", "year")) {
    if (start) {
      return(floor_date(x, units))
    } else {
      return(ceiling_date(x, units))
    }
  } else {
    floors <- c("sec", "min", "hour", "day", "month", "year")
    floorto <- floors[match(units, floors) + 1L]
    lower <- floor_date(x, floorto)
    upper <- ceiling_date(x, floorto)

    points <- seq.POSIXt(lower, upper, paste(length, units))

    if (start) {
      points <- points[points <= x]
    } else {
      points <- points[points >= x]
    }

    fit <- as.duration(x - points)
    fit <- abs(as.double(fit, "secs"))
    points[which.min(fit)]
  }
}

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
#' @include Dates.r
#' @include difftimes.r
#' @include numeric.r
#' @include POSIXt.r
#' @include time-zones.r
NULL


#' Convert a variety of date-time classes to POSIXlt and POSIXct
#' @name DateCoercion
#' @keywords internal
#'
NULL

#' @export
as.POSIXct.fts <- function(x, tz = "", ...) as.POSIXct(zoo::index(x))
#' @export
as.POSIXlt.fts <- function(x, tz = "", ...) as.POSIXlt(zoo::index(x))

## #' @export
## as.POSIXlt.its <- function(x, tz = "", ...) as.POSIXlt(attr(x, "dates"))
## #' @export
## as.POSIXct.its <- function(x, tz = "", ...) as.POSIXct(attr(x, "dates"))

#' @export
as.POSIXlt.timeSeries <- function(x, tz = "", ...) {
  as.POSIXlt(timeDate::timeDate(x@positions,
    zone = x@FinCenter, FinCenter = x@FinCenter))
}
#' @export
as.POSIXct.timeSeries <- function(x, tz = "", ...) {
  as.POSIXct(timeDate::timeDate(x@positions,
    zone = x@FinCenter, FinCenter = x@FinCenter))
}

#' @export
as.POSIXlt.irts <- function(x, tz = "", ...) as.POSIXlt(x$time)
#' @export
as.POSIXct.irts <- function(x, tz = "", ...) as.POSIXct(x$time)

#' @export
as.POSIXlt.xts <- function(x, tz = "", ...) as.POSIXlt(zoo::index(x))
#' @export
as.POSIXct.xts <- function(x, tz = "", ...) as.POSIXct(zoo::index(x))
#' @export
as.POSIXlt.zoo <- function(x, tz = "", ...) as.POSIXlt(zoo::index(x))
#' @export
as.POSIXct.zoo <- function(x, tz = "", ...) as.POSIXct(zoo::index(x))

#' @export
as.POSIXlt.tis <- function(x, tz = "", ...) as.Date(x)

#' Convenience method to reclass dates post-modification.
#' @keywords internal
#'
#' @export
reclass_date <- function(new, orig) UseMethod("reclass_date", orig)
#' @export
reclass_date.POSIXlt <- function(new, orig) {
  as.POSIXlt(new)
}
#' @export
reclass_date.POSIXct <- function(new, orig) {
  as.POSIXct(new)
}
#' @export
reclass_date.chron <- function(new, orig) {
  chron::as.chron(new)
}
#' @export
reclass_date.timeDate <- function(new, orig) {
  timeDate::as.timeDate(new)
}
## #' @export
## reclass_date.its <- function(new, orig) {
##   its::its(new, format = "%Y-%m-%d %X")
## }
#' @export
reclass_date.ti <- function(new, orig) {
  tis::as.ti(new, tis::tifName(orig))
}
#' @export
reclass_date.Date <- function(new, orig) {
  as.Date(new)
}


period_to_difftime <- function(per) {
  as.difftime(per)
}

#' Convenience method to reclass timespans post-modification.
#' @keywords internal
#'
#' @aliases reclass_timespan,ANY,Duration-method reclass_timespan,ANY,Interval-method
#' reclass_timespan,ANY,Period-method reclass_timespan,ANY,difftime-method
#' @export
reclass_timespan <- function(new, orig) standardGeneric("reclass_timespan")

#' @export
setGeneric("reclass_timespan")

#' @export
setMethod("reclass_timespan", signature(orig = "difftime"), function(new, orig) {
  if (is.period(new))
    as.difftime(new)
  else
    make_difftime(as.numeric(new))
})

#' @export
setMethod("reclass_timespan", signature(orig = "Duration"), function(new, orig) {
  as.duration(new)
})

#' @export
setMethod("reclass_timespan", signature(orig = "Interval"), function(new, orig) {
  as.duration(new)
})

#' @export
setMethod("reclass_timespan", signature(orig = "Period"), function(new, orig) {
  as.period(new)
})


#' Change an object to a duration
#'
#' as.duration changes Interval, Period and numeric class objects to
#' Duration objects. Numeric objects are changed to Duration objects
#' with the seconds unit equal to the numeric value.
#'
#' Durations are exact time measurements, whereas periods are relative time
#' measurements. See [Period-class]. The length of a period depends
#' on when it occurs. Hence, a one to one mapping does not exist between
#' durations and periods. When used with a period object, as.duration provides
#' an inexact estimate of the length of the period; each time unit is assigned
#' its most common number of seconds. A period of one month is converted to
#' 2628000 seconds (approximately 30.42 days). This ensures that 12 months will
#' sum to 365 days, or one normal year. For an exact transformation, first
#' transform the period to an interval with [as.interval()].
#'
#' @param x Object to be coerced to a duration
#' @param ... Parameters passed to other methods. Currently unused.
#' @return A duration object
#' @seealso [Duration-class], [duration()]
#' @keywords classes manip methods chron
#' @examples
#' span <- interval(ymd("2009-01-01"), ymd("2009-08-01")) #interval
#' as.duration(span)
#' as.duration(10) # numeric
#' dur <- duration(hours = 10, minutes = 6)
#' as.numeric(dur, "hours")
#' as.numeric(dur, "minutes")
#'
#' @aliases as.duration,numeric-method as.duration,logical-method
#'   as.duration,difftime-method as.duration,Interval-method
#'   as.duration,Duration-method as.duration,Period-method
#'   as.duration,character-method
#' @export
setGeneric("as.duration",
           function(x, ...) standardGeneric("as.duration"),
           useAsDefault = function(x, ...) {
             stop(sprintf("as.duration is not defined for class '%s'", class(x)))
           })

setMethod("as.duration", signature(x = "character"), function(x) {
  as.duration(as.period(x))
})

setMethod("as.duration", signature(x = "numeric"), function(x) {
  new("Duration", x)
})

setMethod("as.duration", signature(x = "logical"), function(x) {
  new("Duration", as.numeric(x))
})

as_duration_from_difftime <- function(x) {
  new("Duration", as.numeric(x, "secs"))
}
setMethod("as.duration", signature(x = "difftime"), as_duration_from_difftime)

setMethod("as.duration", signature(x = "Interval"), function(x) {
  new("Duration", x@.Data)
})

setMethod("as.duration", signature(x = "Duration"), function(x) {
  x
})

setMethod("as.duration", signature(x = "Period"), function(x) {
  new("Duration", period_to_seconds(x))
})


#' Change an object to an `interval`
#'
#' as.interval changes difftime, Duration, Period and numeric class objects to
#' intervals that begin at the specified date-time. Numeric objects are first
#' coerced to timespans equal to the numeric value in seconds.
#'
#' as.interval can be used to create accurate transformations between Period
#' objects, which measure time spans in variable length units, and Duration objects,
#' which measure timespans as an exact number of seconds. A start date-
#' time must be supplied to make the conversion. Lubridate uses
#' this start date to look up how many seconds each variable
#' length unit (e.g. month, year) lasted for during the time span
#' described. See
#' [as.duration()], [as.period()].
#'
#' @param x a duration, difftime, period, or numeric object that describes the length of the interval
#' @param start a POSIXt or Date object that describes when the interval begins
#' @param ... additional arguments to pass to as.interval
#' @return an interval object
#' @seealso [interval()]
#' @keywords classes manip methods chron
#' @examples
#' diff <- make_difftime(days = 31) #difftime
#' as.interval(diff, ymd("2009-01-01"))
#' as.interval(diff, ymd("2009-02-01"))
#'
#' dur <- duration(days = 31) #duration
#' as.interval(dur, ymd("2009-01-01"))
#' as.interval(dur, ymd("2009-02-01"))
#'
#' per <- period(months = 1) #period
#' as.interval(per, ymd("2009-01-01"))
#' as.interval(per, ymd("2009-02-01"))
#'
#' as.interval(3600, ymd("2009-01-01")) #numeric
#' @aliases as.interval,numeric-method as.interval,difftime-method as.interval,Interval-method as.interval,Duration-method as.interval,Period-method as.interval,POSIXt-method as.interval,logical-method
#' @export
as.interval <- function(x, start, ...) standardGeneric("as.interval")

#' @export
setGeneric("as.interval")

setMethod("as.interval", signature(x = "numeric"), function(x, start, ...) {
  .number_to_interval(x, start, ...)
})

setMethod("as.interval", signature(x = "difftime"), function(x, start, ...) {
  .number_to_interval(x, start, ...)
})

setMethod("as.interval", signature(x = "Interval"), function(x, start, ...) {
  x
})

setMethod("as.interval", signature(x = "POSIXt"), function(x, start, ...) {
  .number_to_interval(x, start, ...)
})

#' @export
setMethod("as.interval", signature("logical"), function(x, start, ...) {
  .number_to_interval(as.numeric(x), start, ...)
})

.number_to_interval <- function(x, start, ...) {
  if (missing(start) & all(is.na(x)))
    start <- .POSIXct(NA_real_, tz = "UTC")
  else stopifnot(is.instant(start))

  if (is.instant(x))
    return(interval(x, start))
  else
    interval(start, start + x)
}


#' Change an object to a period
#'
#' as.period changes Interval, Duration, difftime and numeric class objects
#' to Period class objects with the specified units.
#'
#' Users must specify which time units to measure the period in. The exact length of
#' each time unit in a period will depend on when it occurs. See
#' [Period-class] and [period()].
#' The choice of units is not trivial; units that are
#' normally equal may differ in length depending on when the time period
#' occurs. For example, when a leap second occurs one minute is longer than 60
#' seconds.
#'
#' Because periods do not have a fixed length, they can not be accurately
#' converted to and from Duration objects. Duration objects measure time spans
#' in exact numbers of seconds, see [Duration-class]. Hence, a one to one
#' mapping does not exist between durations and periods. When used with a
#' Duration object, as.period provides an inexact estimate; the duration is
#' broken into time units based on the most common lengths of time units, in
#' seconds. Because the length of months are particularly variable, a period
#' with a months unit can not be coerced from a duration object. For an exact
#' transformation, first transform the duration to an interval with
#' [as.interval()].
#'
#' Coercing an interval to a period may cause surprising behavior if you request
#' periods with small units. A leap year is 366 days long, but one year long. Such
#' an interval will convert to 366 days when unit is set to days and 1 year when
#' unit is set to years. Adding 366 days to a date will often give a different
#' result than adding one year. Daylight savings is the one exception where this
#' does not apply. Interval lengths are calculated on the UTC timeline, which does
#' not use daylight savings. Hence, periods converted with seconds or minutes will not
#' reflect the actual variation in seconds and minutes that occurs due to daylight
#' savings. These periods will show the "naive" change in seconds and minutes that is
#' suggested by the differences in clock time. See the examples below.
#'
#' @param x an interval, difftime, or numeric object
#' @param unit A character string that specifies which time units to build period in.
#' unit is only implemented for the as.period.numeric method and the as.period.interval method.
#' For as.period.interval, as.period will convert intervals to units no larger than the specified
#' unit.
#' @param ... additional arguments to pass to as.period
#' @return a period object
#' @seealso [Period-class], [period()]
#' @keywords classes manip methods chron
#' @examples
#' span <- interval(ymd_hms("2009-01-01 00:00:00"), ymd_hms("2010-02-02 01:01:01")) #interval
#' as.period(span)
#' as.period(span, unit = "day")
#' "397d 1H 1M 1S"
#' leap <- interval(ymd("2016-01-01"), ymd("2017-01-01"))
#' as.period(leap, unit = "days")
#' as.period(leap, unit = "years")
#' dst <- interval(ymd("2016-11-06", tz = "America/Chicago"),
#' ymd("2016-11-07", tz = "America/Chicago"))
#' # as.period(dst, unit = "seconds")
#' as.period(dst, unit = "hours")
#' per <- period(hours = 10, minutes = 6)
#' as.numeric(per, "hours")
#' as.numeric(per, "minutes")
#'
#' @aliases as.period,numeric-method as.period,difftime-method
#'   as.period,Interval-method as.period,Duration-method as.period,Period-method
#'   as.period,logical-method as.period,character-method
#' @export
setGeneric("as.period",
           function(x, unit, ...) standardGeneric("as.period"),
           useAsDefault = function(x, unit, ...) {
             stop(sprintf("as.period is not defined for class '%s'", class(x)))
           })

setMethod("as.period", signature(x = "character"), function(x, ...) {
  parse_period(x)
})

setMethod("as.period", signature(x = "numeric"), function(x, unit = "second", ...) {
  x <- as.numeric(x)
  if (missing(unit)) unit <- "second"
  unit <- standardise_date_names(unit[[1]])
  f <- get(paste(unit, "s", sep = ""),
           envir = asNamespace("lubridate"),
           mode = "function", inherits = FALSE)
  f(x)
})

setMethod("as.period", signature(x = "difftime"), function(x, unit = NULL, ...) {
  seconds_to_period(as.double(x, "secs"))
})

setMethod("as.period", signature(x = "Interval"), function(x, unit = NULL, ...) {
  ## fixme: document this in the manual

  ## SEMANTICS: for postitive intervals all units of the period will be
  ## positive, and the oposite for negatve intervals.

  ## Periods are not symetric in the sense that as.period(int) might not be the
  ## same as -as.period(int_flip(int)). See
  ## https://github.com/tidyverse/lubridate/issues/285 for motivation.

  unit <-
    if (missing(unit)) "year"
    else standardise_period_names(unit)

  switch(unit,
         year = .int_to_period(x),
         month = {
           pers <- .int_to_period(x)
           month(pers) <- month(pers) + year(pers)*12
           year(pers) <- 0
           pers
         },
         ## fixme: add note to the docs that unit <= days results in much faster conversion
         ## fixme: add week
         day = , hour = , minute = , second = {
           secs <- x@.Data
           negs <- secs < 0 & !is.na(secs)
           units <- .units_within_seconds(abs(secs), unit)
           pers <- do.call("new", c("Period", units))
           pers[negs] <- -pers[negs]
           pers
         },
         stop("Unsuported unit ", unit))
})

.int_to_period <- function(x) {
  ## this function is called only for conversion with units > day
  start <- as.POSIXlt(x@start)
  end <- unclass(as.POSIXlt(start + x@.Data))
  start <- unclass(start)

  negs <- x@.Data < 0 & !is.na(x@.Data)

  per <- list()

  for (nm in c("sec", "min", "hour", "mday", "mon", "year")) {
    per[[nm]] <- ifelse(negs, start[[nm]] - end[[nm]], end[[nm]] - start[[nm]])
  }

  names(per) <- c("second", "minute", "hour", "day", "month", "year")

  ## Remove negative ...

  ## seconds
  nsecs <- per$second < 0 & !is.na(per$second)
  per$second[nsecs] <- 60 + per$second[nsecs]
  per$minute[nsecs] <- per$minute[nsecs] - 1
  per$second[negs] <- -per$second[negs]

  ## minutes
  nmins <- per$minute < 0 & !is.na(per$minute)
  per$minute[nmins] <- 60 + per$minute[nmins]
  per$hour[nmins] <- per$hour[nmins] - 1
  per$minute[negs] <- -per$minute[negs]

  ## hours
  nhous <- per$hour < 0 & !is.na(per$hour)
  per$hour[nhous] <- 24 + per$hour[nhous]
  per$hour[negs] <- -per$hour[negs]

  ## days

  ### postivie periods
  ndays <- !negs & per$day < 0 & !is.na(per$day)
  if (any(ndays)) {

    ## compute nr days in previous month
    add_months <- rep.int(-1, sum(ndays))

    pmonth <- end$mon[ndays]
    pmonth[pmonth == 0] <- 1 # dec == jan == 31 days
    prev_month_days <- .days_in_month(pmonth, end$year[ndays])

    ## difference in days:
    ## /need pmax to capture as.period(interval(ymd("1985-01-31"), ymd("1986-03-28")))/
    per$day[ndays] <- pmax(prev_month_days - start$mday[ndays], 0) + end$mday[ndays]
    per$month[ndays] <- per$month[ndays] + add_months
  }

  ## negative periods
  ndays <- negs & per$day < 0 & !is.na(per$day)
  if (any(ndays)) {

    add_months <- rep.int(1, sum(ndays))
    this_month_days <- .days_in_month(end$mon[ndays] + 1, end$year[ndays])

    ## Compute nr of days:
    ## /need pmax to capture as.period(interval(ymd("1985-01-31"), ymd("1986-03-28")))/
    per$day[ndays] <- pmax(this_month_days - end$mday[ndays], 0) + start$mday[ndays]
    per$month[ndays] <- per$month[ndays] - add_months
  }

  ## substract only after the day computation to capture intervals like:
  ## as.period(interval(ymd_hms("1985-12-31 5:0:0"), ymd_hms("1986-02-01 3:0:0")))
  per$day[nhous] <- per$day[nhous] - 1
  per$day[negs] <- -per$day[negs]

  ## months
  nmons <- per$month < 0 & !is.na(per$month)
  per$month[nmons] <- 12 + per$month[nmons]
  per$year[nmons] <- per$year[nmons] - 1
  per$month[negs] <- -per$month[negs]

  per$year[negs] <- -per$year[negs]

  new("Period", per$second, year = per$year, month = per$month,
    day = per$day, hour = per$hour, minute = per$minute)
}

setMethod("as.period", signature(x = "Duration"), function(x, unit = NULL, ...) {
  span <- x@.Data
  remainder <- abs(span)
  newper <- period(second = rep(0, length(x)))

  slot(newper, "year") <- remainder %/% average_durations[["year"]]
  remainder <- remainder %% average_durations[["year"]]

  slot(newper, "day") <- remainder %/% (3600 * 24)
  remainder <- remainder %% (3600 * 24)

  slot(newper, "hour") <- remainder %/% (3600)
  remainder <- remainder %% (3600)

  slot(newper, "minute") <- remainder %/% (60)
  newper$second <- remainder %% (60)

  newper * sign(span)
})

setMethod("as.period", signature("Period"),
          function(x, unit = NULL, ...) {
            if (missing(unit) || is.null(unit)) {
              x
            } else {
              unit <- standardise_period_names(unit)
              switch(unit,
                     year = x,
                     month = {
                       month(x) <- month(x) + year(x)*12
                       year(x) <- 0
                       x
                     },
                     day = , hour = , minute = , second = {
                       N <- .units_within_seconds(period_to_seconds(x), unit)
                       do.call("new", c("Period", N))
                     },
                     stop("Unsuported unit ", unit))
            }
          })

setMethod("as.period", signature("logical"), function(x, unit = NULL, ...) {
  as.period(as.numeric(x), unit, ...)
})

setGeneric("as.difftime")

#' @export
setMethod("as.difftime", signature(tim = "Interval"), function(tim, format = "%X", units = "secs") {
  as.difftime(as.numeric(tim, units), format, units)
})

as_difftime_from_duration <- function(tim, format = "%X", units = "secs") {
  as.difftime(tim@.Data, format, units)
}
#' @export
setMethod("as.difftime", signature(tim = "Duration"), as_difftime_from_duration)

#' @export
setMethod("as.difftime", signature(tim = "Period"), function(tim, format = "%X", units = "secs") {
  as.difftime(period_to_seconds(tim), format, units)
})

setGeneric("as.numeric")

seconds_to_unit <- function(secs, unit = "second") {
  switch(unit,
         second = secs,
         minute = secs / 60,
         hour   = secs / 3600,
         day    = secs / 86400,
         month  = secs / (86400 * 365.25 / 12),
         week   = secs / (86400 * 7),
         year   = secs / (86400 * 365.25),
         stop("invalid unit ", unit))
}

#' @export
setMethod("as.numeric", signature("Duration"), function(x, units = "secs", ...) {
  unit <- standardise_period_names(units)
  as.numeric(seconds_to_unit(x@.Data, unit), ...)
})

#' @export
setMethod("as.numeric", signature(x = "Interval"), function(x, units = "secs", ...) {
  as.numeric(as.duration(x), units, ...)
})

#' @export
setMethod("as.numeric", signature(x = "Period"), function(x, units = "second", ...) {
  unit <- standardise_period_names(units)
  as.numeric(seconds_to_unit(period_to_seconds(x), unit = unit), ...)
})

as.POSIXt <- function(x) as.POSIXlt(x)

#' @export
setMethod("as.character", signature(x = "Period"), function(x, ...) {
  format(x)
})

#' @export
setMethod("as.character", signature(x = "Duration"), function(x, ...) {
  format(x)
})

#' @export
setMethod("as.character", signature(x = "Interval"), function(x, ...) {
  format(x)
})


#' Convert an object to a date or date-time
#'
#'
#' @section Compare to base R:
#'
#' These are drop in replacements for [as.Date()] and [as.POSIXct()], with a few
#' tweaks to make them work more intuitively.
#'
#' \itemize{
#'   \item `as_date()` ignores the timezone attribute, resulting in
#'      a more intuitive conversion (see examples)
#'   \item Both functions provide a default origin argument for numeric
#'      vectors.
#'   \item Both functions will generate NAs for invalid date format. A warning message will provide a count of the elements that were not converted
#'   \item `as_datetime()` defaults to using UTC.
#' }
#'
#' @param x a vector of [POSIXt], numeric or character objects
#' @param origin a Date object, or something which can be coerced by
#'   `as.Date(origin, ...)` to such an object (default: the Unix epoch of
#'   "1970-01-01"). Note that in this instance, `x` is assumed to reflect the
#'   number of days since `origin` at `"UTC"`.
#' @param tz a time zone name (default: time zone of the POSIXt object `x`). See
#'   [OlsonNames()].
#' @param format format argument for character methods. When supplied parsing is
#'   performed by [strptime()]. For this reason consider using specialized
#'   parsing functions in lubridate.
#' @param ... further arguments to be passed to specific methods (see above).
#' @return a vector of [Date] objects corresponding to `x`.
#' @examples
#' dt_utc <- ymd_hms("2010-08-03 00:50:50")
#' dt_europe <- ymd_hms("2010-08-03 00:50:50", tz="Europe/London")
#' c(as_date(dt_utc), as.Date(dt_utc))
#' c(as_date(dt_europe), as.Date(dt_europe))
#' ## need not supply origin
#' as_date(10)
#' ## Will replace invalid date format with NA
#' dt_wrong <- c("2009-09-29", "2012-11-29", "2015-29-12")
#' as_date(dt_wrong)
#' @export
setGeneric(name = "as_date",
           def = function(x, ...) standardGeneric("as_date"))

#' @rdname as_date
#' @export
setMethod("as_date", "ANY",
          function(x, ...) {
            ## From: Kurt Hornik <Kurt.Hornik@wu.ac.at>
            ## Date: Tue, 3 Apr 2018 18:53:19
            ##
            ## `zoo` has its own as.Date for which it registers its yearmon
            ## method (and base::as.Date as the default S3 method).  In fact,
            ## zoo also exports as.Date.yearmon etc, but the above
            ##
            ##    lubridate::as_date(zoo::as.yearmon("2011-01-07"))
            ##
            ## does not attach the zoo exports, hence does not find
            ## as.Date.yearmon on the search path.
            if (inherits(x, c("yearmon", "yearqtr")))
              zoo::as.Date(x, ...)
            else
              base::as.Date(x, ...)
          })

#' @rdname as_date
#' @export
setMethod(f = "as_date", signature = "POSIXt",
          function(x, tz = NULL) {
            tz <- if (is.null(tz)) tz(x) else tz
            as.Date(x, tz = tz)
          })

#' @rdname as_date
setMethod(f = "as_date", signature = "numeric",
          function(x, origin = lubridate::origin) {
            as.Date(x, origin = origin)
          })


#' @rdname as_date
#' @export
setMethod("as_date", "character",
          function(x, tz = NULL, format = NULL) {
            if (!is.null(tz)) {
              warning("`tz` argument is ignored by `as_date()`", call. = FALSE)
            }

            if (is.null(format))
              as_date(.parse_iso_dt(x, tz = "UTC"))
            else
              as_date(strptime(x, format, tz = "UTC"))
          })

#' @rdname as_date
#' @export
setGeneric("as_datetime",
           function(x, ...) {
             standardGeneric("as_datetime")
           })

#' @rdname as_date
#' @export
setMethod("as_datetime", "POSIXt",
  function(x, tz = "UTC") {
    with_tz(x, tz)
  })

#' @rdname as_date
#' @export
setMethod("as_datetime", "numeric",
  function(x, origin = lubridate::origin, tz = "UTC") {
    as.POSIXct(x, origin = origin, tz = tz)
  })


#' @rdname as_date
#' @export
setMethod("as_datetime", "character",
          function(x, tz = "UTC", format = NULL) {
            if (is.null(format))
              .parse_iso_dt(x, tz)
            else
              as.POSIXct(strptime(x, format = format, tz = tz))
          })

#' @rdname as_date
#' @export
setMethod("as_datetime", "ANY",
  function(x, tz = "UTC") {
    with_tz(as.POSIXct(x), tzone = tz)
  })

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
NULL


multiply_duration_by_number <- function(dur, num) {
  new("Duration", dur@.Data * num)
}

multiply_interval_by_number <- function(int, num) {
  new("Interval", int@.Data * num, start = int@start, tzone = int@tzone)
}

multiply_period_by_number <- function(per, num) {
  new("Period", per@.Data * num,
    year = per@year * num,
    month = per@month * num,
    day = per@day * num,
    hour = per@hour * num,
    minute = per@minute * num
  )
}


#' @export
setMethod(
  "*", signature(e1 = "Timespan", e2 = "Timespan"),
  function(e1, e2) stop_incompatible_classes(e1, e2, "*")
)

#' @export
setMethod("*", signature(e1 = "Duration"), function(e1, e2) multiply_duration_by_number(e1, e2))

#' @export
setMethod("*", signature(e2 = "Duration"), function(e1, e2) multiply_duration_by_number(e2, e1))

#' @export
setMethod("*", signature(e1 = "Interval"), function(e1, e2) multiply_interval_by_number(e1, e2))

#' @export
setMethod("*", signature(e2 = "Interval"), function(e1, e2) multiply_interval_by_number(e2, e1))

#' @export
setMethod("*", signature(e1 = "Period"), function(e1, e2) multiply_period_by_number(e1, e2))

#' @export
setMethod("*", signature(e2 = "Period"), function(e1, e2) multiply_period_by_number(e2, e1))

##' Guess possible date-times formats from a character vector
##'
##' Guess possible date-times formats from a character vector.
##'
##' @param x input vector of date-times.
##' @param orders format orders to look for. See examples.
##' @param locale locale to use. Defaults to the current locale.
##' @param preproc_wday whether to preprocess weekday names. Internal
##' optimization used by `ymd_hms()` family of functions. If `TRUE`, weekdays are
##' substituted with %a or %A accordingly, so that there is no need to supply
##' this format explicitly.
##' @param print_matches for development purposes mainly. If `TRUE`, prints a matrix
##' of matched templates.
##' @return a vector of matched formats
##' @export
##' @examples
##'
##' x <- c('February 20th 1973',
##'        "february  14, 2004",
##'        "Sunday, May 1, 2000",
##'        "Sunday, May 1, 2000",
##'        "february  14, 04",
##'        'Feb 20th 73',
##'        "January 5 1999 at 7pm",
##'        "jan 3 2010",
##'        "Jan 1, 1999",
##'        "jan 3   10",
##'        "01 3 2010",
##'        "1 3 10",
##'        '1 13 89',
##'        "5/27/1979",
##'        "12/31/99",
##'        "DOB:12/11/00",
##'        "-----------",
##'        'Thu, 1 July 2004 22:30:00',
##'        'Thu, 1st of July 2004 at 22:30:00',
##'        'Thu, 1July 2004 at 22:30:00',
##'        'Thu, 1July2004 22:30:00',
##'        'Thu, 1July04 22:30:00',
##'        "21 Aug 2011, 11:15:34 pm",
##'        "-----------",
##'        "1979-05-27 05:00:59",
##'        "1979-05-27",
##'        "-----------",
##'        "3 jan 2000",
##'        "17 april 85",
##'        "27/5/1979",
##'        '20 01 89',
##'        '00/13/10',
##'        "-------",
##'        "14 12 00",
##'        "03:23:22 pm")
##'
##' guess_formats(x, "BdY")
##' guess_formats(x, "Bdy")
##' ## m also matches b and B; y also matches Y
##' guess_formats(x, "mdy", print_matches = TRUE)
##'
##' ## T also matches IMSp order
##' guess_formats(x, "T", print_matches = TRUE)
##'
##' ## b and B are equivalent and match, both, abreviated and full names
##' guess_formats(x, c("mdY", "BdY", "Bdy", "bdY", "bdy"), print_matches = TRUE)
##' guess_formats(x, c("dmy", "dbY", "dBy", "dBY"), print_matches = TRUE)
##'
##'
##' guess_formats(x, c("dBY HMS", "dbY HMS", "dmyHMS", "BdY H"), print_matches = TRUE)
##'
##' guess_formats(x, c("ymd HMS"), print_matches = TRUE)
##'
guess_formats <- function(x, orders, locale = Sys.getlocale("LC_TIME"),
                          preproc_wday = TRUE, print_matches = FALSE) {

  ## remove all separators
  orders <- gsub("[^[:alpha:]]+", "", orders)

  if (any(grepl("hms|hm|ms", orders))) {
    .deprecated("hms, hm and ms usage", ", please use HMS, HM or MS instead", "1.5.6")
    orders <- gsub("hms", "HMS", orders, ignore.case = TRUE)
    orders <- gsub("hm", "HM", orders, ignore.case = TRUE)
    orders <- gsub("ms", "MS", orders, ignore.case = TRUE)
  }

  ## redirect some formats to C parser (using perl's lookbehind)
  if (length(wp <- grepl("(?<!O)p", orders, perl = T)))
    orders <- c(sub("p", "Op", orders[wp], fixed = T), orders)
  if (length(wm <- grepl("(?<!O)m", orders, perl = T)))
    orders <- c(sub("m", "Om", orders[wm]), orders)
  if (length(wm <- grepl("(?<!O)[bB]", orders, perl = T)))
    orders <- c(sub("b", "Ob", orders[wm]), orders)
  if (length(wT <- grepl("T", orders, fixed = T)))
    orders <- c(sub("T", "HMSOp", orders[wT], fixed = T), orders)
  if (length(wR <- grepl("R", orders, fixed = T)))
    orders <- c(sub("R", "HMOp", orders[wR], fixed = T), orders)
  if (length(wr <- grepl("r", orders, fixed = T)))
    orders <- c(sub("r", "HOp", orders[wR], fixed = T), orders)

  ## We split into characterst first and then paste together formats that start
  ## with O. If perl style lookahead would have worked we wouldn't need this,
  ## but it doesn't.
  osplits <- strsplit(orders, "", fixed = TRUE)
  osplits <- lapply(osplits,
                    function(ospt) {
                      if (length(which_O <- which(ospt == "O")) > 0) {
                        ospt[which_O + 1] <- paste("O", ospt[which_O + 1], sep = "")
                        ospt[-which_O]
                      } else
                        ospt
                    })

  reg <- .get_locale_regs(locale)

  flex_regs <- c(reg$alpha_flex, reg$num_flex, .c_parser_reg_flex)
  exact_regs <- c(reg$alpha_exact, reg$num_exact, .c_parser_reg_exact)

  REGS <- unlist(lapply(osplits, function(fnames) {
    ## fnames are names of smalest valid formats, like a, A, b, z, OS, OZ ...
    which <- !fnames %in% c(names(reg$alpha_flex),
                             names(reg$num_flex),
                             names(.c_parser_reg_exact))
    if (any(which))
      stop("Unknown formats supplied: ", paste(fnames[ which ], sep = ", "))

    ## restriction: no numbers before or after
    ## fixme: using \\D*? because \\D+ doesn't work in flex match, why?
    paste("^\\D*?\\b((",
          paste(unlist(flex_regs[fnames]), collapse = "\\D*?"),
          ")|(",
          paste(unlist(exact_regs[fnames]), collapse = "\\D*?"),
          "))\\D*$", sep = "")
  }))

  ## print debugging info
  if (print_matches) {
    subs <- lapply(REGS, .substitute_formats, x, fmts_only = FALSE)
    names(subs) <- orders
    print(do.call(cbind, c(list(x), subs)))
  }

  .build_formats <- function(regs, orders, x) {
    out <- mapply(
      function(reg, name) {
        out <- .substitute_formats(reg, x)
        if (!is.null(out)) names(out) <- rep.int(name, length(out))
        out
      }, REGS, orders, SIMPLIFY = F, USE.NAMES = F)
    names(out) <- NULL
    unlist(out)
  }

  if (preproc_wday && !any(grepl("[aA]", orders))) {
    ## replace short/long weak days in current locale
    x2 <- gsub(reg$alpha_exact[["A"]], "%A", x, ignore.case = T, perl = T)
    x2 <- gsub(reg$alpha_exact[["a"]], "%a", x2, ignore.case =  T, perl = T)
    formats <- .build_formats(REGS, orders, x2)

    ## In some locales abreviated day (italian "gio") is part of month name
    ## ("maggio"). So check if %a or %A is present and append wday-less formats.
    if (any(grepl("%[aA]", formats)))
      c(formats, .build_formats(REGS, orders, x))
    else
      formats
  } else {
    .build_formats(REGS, orders, x)
  }

}

.substitute_formats <- function(reg, x, fmts_only = TRUE) {
  ## Take date X and substitute year with %Y/%y, month with %B/%b etc.
  ## Return the formatted string if REG matched, or null otherwise.
  ## REG should be with captures as build by .build_locale_regs
  ## Captures should start with the strptyme formats. For example Y_e, B_m_e
  ## Traling _e, _m_e are removed.
  ## fmts_only == FALSE is for debugging purpose only

  m <- regexpr(reg, x, ignore.case = TRUE, perl = TRUE)
  ## print(regs[[1]])
  matched <- which(m > 0)

  if (length(matched) > 0) {
    nms <- attr(m, "capture.names")
    nms <- nms[nzchar(nms)]
    ## e <- grepl("_e", nms, fixed = TRUE)
    ## nms_e <- nms[e]
    ## nms <- nms[!e]

    start <- attr(m, "capture.start")[matched, , drop = FALSE]
    end <- start + attr(m, "capture.length")[matched, , drop = FALSE] - 1L

    lout <- x[matched]
    for (n in rev(nms)) {  ## start from the end
      w <- end[, n] > 0 ## -1 if unmatched  subpatern
      lout[w] <- .str_sub(
        lout[w], start[w, n], end[w, n],
        paste("%", gsub("_.*$", "", n), sep = "")
      )
    }
    if (fmts_only)
      lout
    else {
      ## developer
      out <- character(length(x)) ## dev only
      out[matched] <- lout
      out
    }
  } else if (fmts_only)
    NULL
  else character(length(x))
}

.enclose <- function(fmts)
  paste("@", fmts, "@", sep = "")

.enclosed.na <- function(x)
  x == "@NA@"

.get_train_set <- function(x) {
  ## the best irregular guesser I could come up with
  x <- x[!.enclosed.na(x)]
  len <- length(x)
  if (len < 100)
    x
  else if (len < 3571)
    x[.primes[.primes  <= length(x) ] ]
  else
    x[ .primes * (length(x) %/% 3571) ] # 501 primes
}

.train_formats <- function(x, formats, locale) {
  ## return a numeric vector of size length(formats), with each element giving
  ## the number of matched elements in X
  ## can return NULL if formats is NULL

  trials <- lapply(formats, function(fmt) .strptime(x, fmt, locale = locale))
  successes <- unlist(lapply(trials, function(x) sum(!is.na(x))), use.names = FALSE)
  names(successes) <- formats
  sort(successes, decreasing = TRUE)
}

.best_formats <- function(x, orders, locale, .select_formats = .select_formats, drop = FALSE) {
  ## return a vector of formats that matched X at least once.
  ## Can be zero length vector, if none matched

  fmts <- unique(guess_formats(x, orders, locale = locale, preproc_wday = TRUE)) # orders as names
  if (length(fmts)) {
    trained <- .train_formats(x, fmts, locale = locale)

    if (drop)
      trained <- trained[ trained > 0 ]
    .select_formats(trained, drop)
  }
}

.select_formats <- function(trained, drop = FALSE) {
  nms <- names(trained)

  score <-
    nchar(gsub("[^%]", "", nms)) + ## longer formats have priority
    grepl("%Y", nms, fixed = T)*1.5 +
    grepl("%y(?!%)", nms, perl = T)*1.6 + ## y has priority over Y, but only when not followed by %
    grepl("%[Bb]", nms)*.31 + ## B/b format has priority over %Om
    ## C parser formats have higher priority
    grepl("%Om", nms)*.3 +
    grepl("%Op", nms)*.3 +
    grepl("%Ob", nms)*.32  ## Ob has higher priority than B/b

  ## ties are broken by `trained`
  n0 <- trained != 0
  if (drop) {
    score <- score[n0]
    trained <- trained[n0]
  } else {
    score[!n0] <- -100
  }

  ## print(rbind(trained, score))
  names(trained)[order(score, trained, decreasing = T)]
}

## These are formats that are effectively matched by c parser. But we must get
## through the format guesser first for ymd_hms family.
.c_parser_reg_flex <- list(Op = "(?<Op>(AM|PM))(?![[:alpha:]])",
                           Om = "((?<Om>1[0-2]|0?[1-9](?!\\d))|(((?<Om_b>Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)|(?<Om_B>January|February|March|April|May|June|July|August|September|October|November|December))(?![[:alpha:]])))",
                           Ob = "(((?<Ob_b>Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)|(?<Ob_B>January|February|March|April|May|June|July|August|September|October|November|December))(?![[:alpha:]]))")

.c_parser_reg_exact <- list(Op = "(?<Op_e>AM|PM)(?![[:alpha:]])",
                            Om = "((?<Om_e>1[0-2]|0[1-9])|(((?<Om_b_e>Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)|(?<Om_B_e>January|February|March|April|May|June|July|August|September|October|November|December))(?![[:alpha:]])))",
                            Ob = "(((?<Ob_b_e>Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)|(?<Ob_B_e>January|February|March|April|May|June|July|August|September|October|November|December))(?![[:alpha:]]))")

.locale_reg_cache <- new.env(hash = FALSE)

.get_locale_regs <- function(locale = Sys.getlocale("LC_TIME")) {
  ## build locale specific regexps for all posible orders

  if (exists(locale, envir = .locale_reg_cache, inherits = FALSE))
    return(get(locale, envir = .locale_reg_cache))

  orig_opt <- options(warn = 5)
  on.exit({
    options(orig_opt);
    Sys.setlocale("LC_TIME", orig_locale)
  })
  orig_locale <- Sys.getlocale("LC_TIME")
  Sys.setlocale("LC_TIME", locale)
  options(orig_opt)

  format <- "%a@%A@%b@%B@%p@"
  L <- enc2utf8(unique(format(.date_template, format = format)))
  mat <- do.call(rbind, strsplit(L, "@", fixed = TRUE))
  mat[] <- gsub("([].|(){^$*+?[])", "\\\\\\1", mat) # escaping meta chars
  names <- colnames(mat) <-  strsplit(format, "[%@]+")[[1]][-1L]

  ## Captures should be unique. Thus we build captures with the following
  ## rule. Capture starts with the name of strptime format (B, b, y etc) It ends
  ## with _e if the expression is an exact match (as oposed to flex), see below.

  ## It can contain _x where x is a main format in which this format
  ## occurs. For example <B_b_e> is an exact capture in the strptime format B
  ## but that also matches b in lubridate. Try lubridate:::.get_loc_regs()
  ## todo: elaborate

  ## ALPHABETIC FORMATS
  alpha <- list()
  alpha["b"] <-
    sprintf("((?<b_b>%s)|(?<B_b>%s))(?![[:alpha:]])",
            paste(unique(mat[, "b"]), collapse = "|"),
            paste(unique(mat[, "B"]), collapse = "|"))
  alpha["B"] <-
    sprintf("(?<B_B>%s)(?![[:alpha:]])",
            paste(unique(mat[, "B"]), collapse = "|"))

  alpha["a"] <-
    sprintf("((?<a_a>%s)|(?<A_a>%s))(?![[:alpha:]])",
            paste(unique(mat[, "a"]), collapse = "|"),
            paste(unique(mat[, "A"]), collapse = "|"))

  alpha["A"] <-
    sprintf("(?<A_A>%s)(?![[:alpha:]])",
            paste(unique(mat[, "A"]), collapse = "|"))

  ## just match Z in ISO8601, (UTC zulu format)
  alpha["Ou"] <- "(?<Ou_Ou>Z)(?![[:alpha:]])"

  p <- unique(mat[, "p"])
  p <- p[nzchar(p)]
  alpha["p"] <-
    if (length(p) == 0L) ""
    else sprintf("(?<p>%s)(?![[:alpha:]])", paste(p, collapse = "|"))

  alpha <- unlist(alpha)

  ##  NUMERIC FORMATS
  num <- num_flex <- num_exact <- c(
      d = "(?<d>[012]?[1-9]|3[01]|[12]0)",
      q = "(?<q>[0]?[1-4])",
      H = "(?<H>2[0-4]|[01]?\\d)",
      h = "(?<H>2[0-4]|[01]?\\d)",
      I = "(?<I>1[0-2]|0?[1-9])",
      j = "(?<j>[0-3]?\\d?\\d)",
      M = "(?<M>[0-5]?\\d)",
      S = "((?<OS_S>[0-5]?\\d\\.\\d+)|(?<S>[0-6]?\\d))",
      s = "((?<OS_S>[0-5]?\\d\\.\\d+)|(?<S>[0-6]?\\d))",
      U = "(?<U>[0-5]?\\d)",
      w = "(?<w>[0-6])", # merge with a, A??
      u = "(?<u>[1-7])",
      W = "(?<W>[0-5]?\\d)",
      ## x = "(?<x>\\d{2}/[01]?\\d/[0-3]?\\d)",
      ## X = "(?<X>[012]?\\d:[0-5]?\\d:[0-6]?\\d)",
      Y = "(?<Y>\\d{4})",
      y = "((?<Y_y>\\d{4})|(?<y>\\d{2}))",
      Oz = "(?<Oz_Oz>[-+]\\d{4})", ## sptrtime implements only this format (4 digits)
      ## F = "(?<F>\\d{4)-\\d{2}-\\d{2})",
      OO = "(?<OO>[-+]\\d{2}:\\d{2})",
      Oo = "(?<Oo>[-+]\\d{2})")

  ## special lubridate formats
  nms <- c("T", "R", "r")

  ## in some locales %p = ""
  if (length(p) == 0L) {
    num <- c(num,
             T = sprintf("(%s\\D+%s\\D+%s)", num[["H"]], num[["M"]], num[["S"]]),
             R = sprintf("(%s\\D+%s)", num[["H"]], num[["M"]]),
             r = sprintf("(%s\\D+)", num[["H"]]))
  } else {
    num <- c(num,
             T = sprintf("((%s\\D+%s\\D+%s\\D*%s)|(%s\\D+%s\\D+%s))",
                         num[["I"]], num[["M"]], num[["S"]], alpha[["p"]], num[["H"]], num[["M"]], num[["S"]]),
             R = sprintf("((%s\\D+%s\\D*%s)|(%s\\D+%s))",
                         num[["I"]], num[["M"]], alpha[["p"]], num[["H"]], num[["M"]]),
             r = sprintf("((%s\\D*%s)|%s)",
                         num[["I"]], alpha[["p"]], num[["H"]]))

    num[nms] <- sub("<M", "<M_T",
                    sub("<S", "<S_T",
                        sub("<OS", "<OS_T", num[nms])))
  }

  ## don't let special nms be confused with standard formats
  num[nms] <- gsub("(<[IMSpHS]|<OS)", "\\1_s", num[nms])


  ## The difference between flex and exact regexp is that flex should follow
  ## by non-digit, thus flexible strings like 12-1-2 can be matched. In exact
  ## regexp a number may not be followed by a non-number. In this case the
  ## number of digits in a number should be precisely specified. For example
  ## 120102 is matched by ymd.
  ##
  ## *Note*: the exact regexps is build from flex regex by gsubing below. So,
  ## pay attention when you modify the regexp above.

  alpha_flex <- alpha
  alpha_exact <- gsub(">", "_e>", alpha, fixed = TRUE)

  num_exact <- num_flex <- num
  num_flex[] <- sprintf("%s(?!\\d)", num)

  num_exact[] <- gsub("(?<!\\()\\?(?!<)", "", perl = T, # remove ?
                      gsub("+", "*",  fixed = T,
                           gsub(">", "_e>", num))) # append _e to avoid duplicates

  num_flex["m"] <- sprintf("((?<m>1[0-2]|0?[1-9](?!\\d))|(%s))", gsub("_[bB]", "\\1_m", alpha[["b"]]))
  num_exact["m"] <- sprintf("((?<m_e>1[0-2]|0[1-9])|(%s))", gsub("_[bB]", "\\1_m_e", alpha[["b"]]))

  ## canoot be in num above because gsub("+", "*") messes it up
  num_flex["OS"] <- "(?<OS_f>[0-5]\\d\\.\\d+)"
  num_exact["OS"] <- "(?<OS_e>[0-5]\\d\\.\\d+)"
  num_flex["z"] <- sprintf("(%s|%s|%s|%s)", alpha_flex[["Ou"]], num_flex[["Oz"]], num_flex[["OO"]], num_flex[["Oo"]])
  num_exact["z"] <- sprintf("(%s|%s|%s|%s)", alpha_exact[["Ou"]], num_exact[["Oz"]], num_exact[["OO"]], num_exact[["Oo"]])

  wday_order <- order(wday(.date_template, week_start = 7))
  wday_names <- list(abr = unique(mat[, "a"][wday_order]),
                     full = unique(mat[, "A"][wday_order]))

  month_order <- order(month(.date_template))
  # remove trailing dot in some locales (#781)
  month_names <- list(abr = sub("\\.", "", unique(mat[, "b"][month_order]), fixed = T),
                      full = unique(mat[, "B"][month_order]))

  out <- list(alpha_flex = alpha_flex, num_flex = num_flex,
              alpha_exact = alpha_exact, num_exact = num_exact,
              wday_names = wday_names, month_names = month_names)

  assign(locale, out, envir = .locale_reg_cache)
  out
}

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r


setOldClass("difftime")

#' Create a difftime object.
#'
#' `make_difftime()` creates a difftime object with the specified number of
#' units. Entries for different units are cumulative. difftime displays
#' durations in various units,  but these units are estimates given for
#' convenience. The underlying object is always recorded as a fixed number of
#' seconds.
#'
#' Conceptually, difftime objects are a type of duration. They measure the
#' exact passage of time but do not always align with measurements
#' made in larger units of time such as hours, months and years.
#' This is because the length of larger time units can be affected
#' by conventions such as leap years
#' and Daylight Savings Time. \pkg{lubridate} provides a second class for measuring durations, the Duration class.
#' @param num Optional number of seconds
#' @param units a character vector that lists the type of units to use for the
#'   display of the return value (see examples). If `units` is "auto" (the
#'   default) the display units are computed automatically. This might create
#'   undesirable effects when converting `difftime` objects to numeric
#'   values in data processing.
#' @param ... a list of time units to be included in the difftime and their amounts. Seconds,
#'   minutes, hours, days, and weeks are supported. Normally only one of `num` or `...` are present. If
#'   both are present, the `difftime` objects are concatenated.
#' @return a difftime object
#' @seealso [duration()], [as.duration()]
#' @keywords chron classes
#' @export
#' @examples
#' make_difftime(1)
#' make_difftime(60)
#' make_difftime(3600)
#' make_difftime(3600, units = "minute")
#' # Time difference of 60 mins
#' make_difftime(second = 90)
#' # Time difference of 1.5 mins
#' make_difftime(minute = 1.5)
#' # Time difference of 1.5 mins
#' make_difftime(second = 3, minute = 1.5, hour = 2, day = 6, week = 1)
#' # Time difference of 13.08441 days
#' make_difftime(hour = 1, minute = -60)
#' # Time difference of 0 secs
#' make_difftime(day = -1)
#' # Time difference of -1 days
#' make_difftime(120, day = -1, units = "minute")
#' # Time differences in mins
make_difftime <- function(num = NULL, units = "auto", ...) {
  pieces <- list(...)
  if (!is.null(num) && length(pieces) > 0) {
    .difftime_from_num(c(num, .difftime_from_pieces(pieces)), units)
  } else if (!is.null(num)) {
    .difftime_from_num(num, units)
  } else if (length(pieces)) {
    .difftime_from_num(.difftime_from_pieces(pieces), units)
  } else {
    stop("No valid values have been passed to 'make_difftime' constructor")
  }
}

difftime_lengths <- c(secs = 1, mins = 60, hours = 3600, days = 86400, weeks = 7 * 86400)

.difftime_from_num <- function(num, units = "auto") {
  seconds <- abs(na.omit(num))
  units <- units[[1]]
  if (units == "auto") {
    if (any(seconds < 60))
      units <- "secs"
    else if (any(seconds < 3600))
      units <- "mins"
    else if (any(seconds < 86400))
      units <- "hours"
    else
      units <- "days"
  } else {
    units <- standardise_difftime_names(units)
  }
  structure(num/difftime_lengths[[units]], units = units, class = "difftime")
}

.difftime_from_pieces <- function(pieces) {
  names(pieces) <- standardise_difftime_names(names(pieces))
  out <- 0
  for (nm in names(pieces))
    out <- out + pieces[[nm]] * difftime_lengths[[nm]]
  out
}

#' Is x a difftime object?
#'
#' @export is.difftime
#' @param x an R object
#' @return TRUE if x is a difftime object, FALSE otherwise.
#' @seealso [is.instant()], [is.timespan()], [is.interval()],
#'   [is.period()].
#' @keywords logic chron
#' @examples
#' is.difftime(as.Date("2009-08-03")) # FALSE
#' is.difftime(make_difftime(days = 12.4)) # TRUE
is.difftime <- function(x) is(x, "difftime")

#' @include periods.r
NULL

#' Get/set date component of a date-time
#'
#' Date-time must be a POSIXct, POSIXlt, Date, chron, yearmon, yearqtr, zoo,
#' zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
#'
#' `date()` does not yet support years before 0 C.E.
#' Also `date()` is not defined for Period objects.
#'
#' @section Base compatibility:
#' `date()` can be called without any arguments to return a string representing
#' the current date-time. This provides compatiblity with `base:date()` which
#' it overrides.
#'
#' @param x a date-time object
#' @param value an object for which the `date()` function is defined
#' @return the date of x as a Date
#' @keywords utilities manip chron methods
#' @examples
#' x <- ymd_hms("2012-03-26 23:12:13", tz = "America/New_York")
#' date(x)
#' as.Date(x) # by default as.Date assumes you want to know the date in UTC
#' as.Date(x, tz = "America/New_York")
#' date(x) <- as.Date("2000-01-02")
#' x
#' @export
date <- function(x) {
  UseMethod("date")
}

#' @export
date.default <- function(x) {
  if (missing(x)) {
    base::date()
  } else {
    x <- as.POSIXlt(x, tz = tz(x))
    year <- x$year + 1900
    month <- x$mon + 1
    day <- x$mday
    as.Date(make_datetime(year, month, day))
  }
}
#' @export
date.Period <- function(x) {
  stop("date is undefined for Period objects")
}

#' @rdname date
#' @export
"date<-" <- function(x, value) {
  update(x,
    years = year(value),
    months = month(value),
    days = day(value))
}

setGeneric("date<-")

#' @export
setMethod("date<-", signature("Period"), function(x, value) {
  stop("date<- is undefined for Period objects")
})

#' @include periods.r
NULL

#' Get/set years component of a date-time
#'
#' Date-time must be a POSIXct, POSIXlt, Date, Period or any other object
#' convertible to POSIXlt.
#'
#' year does not yet support years before 0 C.E.
#'
#' @param x a date-time object
#' @param value a numeric object
#' @return the years element of x as a decimal number
#' @keywords utilities manip chron methods
#' @references
#'    \url{https://en.wikipedia.org/wiki/ISO_week_date}
#' @examples
#' x <- ymd("2012-03-26")
#' year(x)
#' year(x) <- 2001
#' year(x) > 1995
#' @export
year <- function(x)
  UseMethod("year")

#' @export
year.default <- function(x)
    as.POSIXlt(x, tz = tz(x))$year + 1900

#' @export
year.Period <- function(x)
  slot(x, "year")

#' @rdname year
#' @export
"year<-" <- function(x, value) {
  x <- x + years(value - year(x))
}

setGeneric("year<-")

#' @export
setMethod("year<-", signature("Period"), function(x, value) {
  slot(x, "year") <- value
  x
})

.other_year <- function(x, week_start = 1) {
  x <- as.POSIXlt(x)
  date <- make_date(year(x), month(x), day(x))
  isodate <- date + ddays(4 - wday(date, week_start = week_start))
  year(isodate)
}

#' @rdname year
#' @description
#' `isoyear()` returns years according to the ISO 8601 week calendar.
#' @export
isoyear <- function(x) {
  .other_year(x, 1)
}

#' @rdname year
#' @description
#' `epiyear()` returns years according to the epidemilogical week calendars.
#' @export
epiyear <- function(x) {
  .other_year(x, 7)
}

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
#' @include Dates.r
#' @include difftimes.r
#' @include POSIXt.r
NULL


subtract_interval_from_date <- function(int, date) {
  end <- int@start + int@.Data
  dur <- as.duration(int)
  add_duration_to_date(-dur, date)
}

#' @export
setMethod(
  "-", signature(e1 = "Duration", e2 = "missing"),
  function(e1, e2) multiply_duration_by_number(e1, -1)
)

#' @export
setMethod(
  "-", signature(e1 = "Interval", e2 = "missing"),
  function(e1, e2) multiply_interval_by_number(e1, -1)
)

#' @export
setMethod(
  "-", signature(e1 = "Period", e2 = "missing"),
  function(e1, e2) multiply_period_by_number(e1, -1)
)

base <- c("Date", "POSIXct", "POSIXlt", "numeric")
classes <- c("Duration", "Period", "Interval", base)
for (class_i in classes) {
  for (class_j in classes) {
    if (class_i %in% base && class_j %in% base) {
      break
    }
    setMethod("-", signature(class_i, class_j), function(e1, e2) e1 + (-e2))
  }
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

C_local_tz <- function() {
    .Call(`_lubridate_C_local_tz`)
}

C_valid_tz <- function(tz_name) {
    .Call(`_lubridate_C_valid_tz`, tz_name)
}

C_update_dt <- function(dt, year, month, yday, mday, wday, hour, minute, second, tz = NULL, roll = FALSE, week_start = 7L) {
    .Call(`_lubridate_C_update_dt`, dt, year, month, yday, mday, wday, hour, minute, second, tz, roll, week_start)
}

C_force_tz <- function(dt, tz, roll = FALSE) {
    .Call(`_lubridate_C_force_tz`, dt, tz, roll)
}

C_force_tzs <- function(dt, tzs, tz_out, roll = FALSE) {
    .Call(`_lubridate_C_force_tzs`, dt, tzs, tz_out, roll)
}

C_local_time <- function(dt, tzs) {
    .Call(`_lubridate_C_local_time`, dt, tzs)
}

setOldClass("POSIXt")
setOldClass("POSIXct")
setOldClass("POSIXlt")

#' Various POSIX utilities
#'
#' [POSIXct()] mirrors primitive contructors in base R ([double()],
#' [character()] etc.)
#'
#' @rdname posix_utils
#' @param x an R object
#' @param length A non-negative number specifying the desired length. Supplying
#'   an argument of length other than one is an error.
#' @param tz a timezone (defaults to "utc")
#' @return TRUE if x is a POSIXct or POSIXlt object, FALSE otherwise.
#' @seealso [is.instant()], [is.timespan()], [is.Date()]
#' @keywords logic chron
#' @examples
#' is.POSIXt(as.Date("2009-08-03"))
#' is.POSIXt(as.POSIXct("2009-08-03"))
#' @export
is.POSIXt <- function(x) inherits(x, "POSIXt")

#' @rdname posix_utils
#' @export
is.POSIXlt <- function(x) inherits(x, "POSIXlt")

#' @rdname posix_utils
#' @export
is.POSIXct <- function(x) inherits(x, "POSIXct")

#' @rdname posix_utils
#' @export
POSIXct <- function(length = 0L, tz = "UTC") {
  t0 <- if (is_utc(tz)) origin else force_tz(origin, tz)
  .POSIXct(rep.int(t0, length), tz = tz)
}

#' @rdname posix_utils
#' @export
NA_POSIXct_ <- .POSIXct(NA_real_, tz = "UTC")

#' @method c POSIXct
c.POSIXct <- function(..., recursive = FALSE) {
  dots <- list(...)
  .POSIXct(c(unlist(lapply(dots, unclass))), tz = tz(dots[[1]]))
}

#' @method c POSIXlt
c.POSIXlt <- function(..., recursive = FALSE) {
  as.POSIXlt(do.call("c.POSIXct", lapply(list(...), as.POSIXct)))
}

evalqOnLoad({
    registerS3method("c", "POSIXct", c.POSIXct)
    ## registerS3method("c", "POSIXlt", c.POSIXlt)
})

#' @include periods.r
NULL

#' Get/set hours component of a date-time
#'
#' Date-time must be a POSIXct, POSIXlt, Date, Period, chron, yearmon, yearqtr, zoo,
#' zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
#'
#' @export
#' @param x a date-time object
#' @keywords utilities manip chron methods
#' @return the hours element of x as a decimal number
#' @examples
#' x <- ymd("2012-03-26")
#' hour(x)
#' hour(x) <- 1
#' hour(x) <- 25
#' hour(x) > 2
hour <- function(x)
  UseMethod("hour")

#' @export
hour.default <- function(x)
    as.POSIXlt(x, tz = tz(x))$hour

#' @export
hour.Period <- function(x)
  slot(x, "hour")

#' @export
#' @param value numeric value to be assigned to the `hour` component
#' @rdname hour
"hour<-" <- function(x, value)
  x <- x + hours(value - hour(x))

setGeneric("hour<-")

#' @export
setMethod("hour<-", signature("Period"), function(x, value) {
  slot(x, "hour") <- value
  x
})

##' Format dates and times based on human-friendly templates
##'
##' Stamps are just like [format()], but based on human-friendly
##' templates like "Recorded at 10 am, September 2002" or "Meeting, Sunday May
##' 1, 2000, at 10:20 pm".
##'
##' `stamp()` is a stamping function date-time templates mainly, though it
##' correctly handles all date and time formats as long as they are
##' unambiguous. `stamp_date()`, and `stamp_time()` are the specialized
##' stamps for dates and times (MHS). These function might be useful when the
##' input template is unambiguous and matches both a time and a date format.
##'
##' Lubridate tries hard to guess the formats, but often a given format can be
##' interpreted in multiple ways. One way to deal with such cases is to provide
##' unambiguous formats like 22/05/81 instead of 10/05/81 for d/m/y
##' format. Another way is to use a more specialized [`stamp_date`] and
##' [`stamp_time`]. The core function `stamp()` prioritizes longer date-time
##' formats.
##'
##' If `x` is a vector of values \pkg{lubridate} will choose the format which
##' "fits" `x` the best. Note that longer formats are preferred. If you have
##' "22:23:00 PM" then "HMSp" format will be given priority to shorter "HMS"
##' order which also fits the supplied string.
##'
##' Finally, you can give desired format order directly as `orders`
##' argument.
##'
##' @param x a character vector of templates.
##' @param orders orders are sequences of formatting characters which might be
##' used for disambiguation. For example "ymd hms", "aym" etc. See
##' [guess_formats()] for a list of available formats.
##' @param locale locale in which `x` is encoded. On Linux-like systems use
##' `locale -a` in the terminal to list available locales.
##' @param quiet whether to output informative messages.
##' @return a function to be applied on a vector of dates
##' @seealso [guess_formats()], [parse_date_time()], [strptime()]
##' @export
##' @examples
##' D <- ymd("2010-04-05") - days(1:5)
##' stamp("March 1, 1999")(D)
##' sf <- stamp("Created on Sunday, Jan 1, 1999 3:34 pm")
##' sf(D)
##' stamp("Jan 01")(D)
##' stamp("Sunday, May 1, 2000", locale = "C")(D)
##' stamp("Sun Aug 5")(D) #=> "Sun Aug 04" "Sat Aug 04" "Fri Aug 04" "Thu Aug 04" "Wed Aug 03"
##' stamp("12/31/99")(D)              #=> "06/09/11"
##' stamp("Sunday, May 1, 2000 22:10", locale = "C")(D)
##' stamp("2013-01-01T06:00:00Z")(D)
##' stamp("2013-01-01T00:00:00-06")(D)
##' stamp("2013-01-01T00:00:00-08:00")(force_tz(D, "America/Chicago"))
stamp <- function(x, orders = lubridate_formats,
                  locale = Sys.getlocale("LC_TIME"), quiet = FALSE) {

  fmts <- unique(guess_formats(x, orders, locale))
  if (is.null(fmts)) stop("Couldn't guess formats of: ", x)
  if (length(fmts) == 1L) {
    FMT <- fmts[[1]]
  } else {
    trained <- .train_formats(x, fmts, locale = locale)
    formats <- .select_formats(trained)
    FMT <- formats[[1]]
    if (!quiet && length(trained) > 1) {
      message("Multiple formats matched: ",
              paste("\"", names(trained), "\"(", trained, ")", sep = "",
                    collapse = ", "))
    }
  }

  if (!quiet)
      message("Using: \"", FMT, "\"")


  ## format doesn't accept 'locale' argument; need a hard reset
  reset_local_expr <-
    quote(
    {
      old_lc_time <- Sys.getlocale("LC_TIME")
      if (old_lc_time != locale) {
        on.exit(Sys.setlocale("LC_TIME", old_lc_time))
        Sys.setlocale("LC_TIME", locale)
      }
    })

  ## ISO8601
  ## %Ou: "2013-04-16T04:59:59Z"
  ## %Oo: "2013-04-16T04:59:59+01"
  ## %Oz: "2013-04-16T04:59:59+0100"
  ## %OO: "2013-04-16T04:59:59+01:00"

  ## Is a timezone format?
  if (grepl("%O[oOzu]|%z", FMT)) {
    ## We need to post-process x in the case of %Oo, %OO and %Oz formats
    ## because standard %z output format ignores timezone.

    ## Post-process only when at the end of the string, otherwise don't bother
    ## and just output with %z format after a conversion to UTC.

    # replicate str_extract(FMT, "%O[oOz]$") without stringr dependence.
    # must return NA if no match.
    oOz_end <- ifelse(grepl('%O[oOz]$', FMT), gsub("^.*(%O[oOz]$)", "\\1", FMT), rep(NA, length(FMT)))

    if (is.na(oOz_end)) {
      FMT <- sub("%O[oOz]", "%z",
                 sub("%Ou", "Z", FMT, fixed = TRUE))

      eval(bquote(
        function(x, locale = .(locale)) {
          ## %z ignores timezone
          if (!is_utc(tz(x[[1L]])))
            x <- with_tz(x, tzone = "UTC")
          .(reset_local_expr)
          format(x, format = .(FMT))
        }))

    } else {
      FMT <- sub("%O[oOz]$", "", FMT)

      eval(bquote(
        function(x, locale = .(locale)) {
          .(reset_local_expr)
          paste0(format(x, format = .(FMT)),
                 .format_offset(x, fmt = .(oOz_end)))
        }))
    }

  } else {
    ## most common case
    eval(bquote(function(x, locale = .(locale)) {
      .(reset_local_expr)
      format(x, format = .(FMT))
    }))
  }
}


.format_offset <- function(x, fmt="%Oz") {
  ## .format_offset
  ##
  ## function to format the offset of a time from UTC
  ##
  ## This is an internal function, used in conjunction with [stamp()].
  ## There are three available formats:
  ##
  ## \itemize{
  ##   \item \code{\%Oo} +01
  ##   \item \code{\%Oz} +0100
  ##   \item \code{\%OO} +01:00
  ## }
  ##
  ## If the \code{\%Oo} format is used for a half-hour timezone, a warning
  ## is issued, and the format is changed to \code{\%Oz}
  ##
  ## @param x      POSIXct for which offset-string is sought
  ## @param fmt    string describing format of offset, default: \code{\%Oz}
  ##
  ## @return string
  ##

  ## "%Oo"  +01
  ## "%Oz"  +0100
  ## "%OO"  +01:00

  ## calculate offset by forcing this time as utc
  dtm_utc <- force_tz(x, tzone = "UTC")

  ## the offset is the duration represented by the difference in time
  offset_duration <- difftime(dtm_utc, x, units = "secs")

  ## determine sign
  .sgn <- ifelse(offset_duration >= as.duration(0), "+", "-")

  ## remove sign
  offset_duration <- abs(offset_duration)

  ## determine hour
  .hr <- floor(offset_duration/dhours(1))

  ## determine minutes
  .min <- floor((offset_duration - dhours(.hr))/dminutes(1))

  ## warning if we need minutes, but are using format without minutes
  if (any(.min > 0) & fmt == "%Oo") {
    warning("timezone offset-minutes are non-zero - changing format to %Oz")
    fmt <- "%Oz"
  }

  result <- switch(
    fmt,
    "%Oo" = sprintf("%s%02d", .sgn, .hr),
    "%Oz" = sprintf("%s%02d%02d", .sgn, .hr, .min),
    "%OO" = sprintf("%s%02d:%02d", .sgn, .hr, .min)
  )

  return(result)
}

##' @rdname stamp
##' @export
stamp_date <- function(x, locale = Sys.getlocale("LC_TIME"))
  stamp(x, orders = c("ymd", "dmy", "mdy", "ydm", "dym", "myd", "my", "ym", "md", "dm", "m", "d", "y"),
        locale = locale)

##' @rdname stamp
##' @export
stamp_time <- function(x, locale = Sys.getlocale("LC_TIME"))
  stamp(x, orders = c("hms", "hm", "ms", "h", "m", "s"), locale = locale)


lubridate_formats <- local({
  xxx <- c("ymd", "ydm", "mdy", "myd", "dmy", "dym")
  names(xxx) <- xxx
  out <- character()
  for (D in xxx) {
    out[[paste(D, "_hms", sep = "")]] <- paste(xxx[[D]], "T", sep = "")
    out[[paste(D, "_hm", sep = "")]] <- paste(xxx[[D]], "R", sep = "")
    out[[paste(D, "_h", sep = "")]] <- paste(xxx[[D]], "r", sep = "")
  }

  out <- c(out, xxx, my = "my", ym = "ym", md = "md", dm = "dm",
           hms = "T", hm = "R", ms = "MS", h = "r", m = "m", y = "y")

  ## adding ISO8601
  out <- c(ymd_hmsz = "ymdTz", out)

  out
})

#' Computes attractive axis breaks for date-time data
#'
#' pretty.dates indentifies which unit of time the sub-intervals should be
#' measured in to provide approximately n breaks. It then chooses a "pretty"
#' length for the sub-intervals and sets start and endpoints that 1) span the
#' entire range of the data, and 2) allow the breaks to occur on important
#' date-times (i.e. on the hour, on the first of the month, etc.)
#'
#' @param x a vector of POSIXct, POSIXlt, Date, or chron date-time objects
#' @param n integer value of the desired number of breaks
#' @param ... additional arguments to pass to function
#' @return a vector of date-times that can be used as axis tick marks or bin breaks
#' @keywords dplot utilities chron
#' @export
#' @examples
#' x <- seq.Date(as.Date("2009-08-02"), by = "year", length.out = 2)
#' pretty_dates(x, 12)
pretty_dates <- function(x, n, ...) {

  otz <- Sys.getenv("TZ")
  if (Sys.getenv("TZ") == "") otz <- "unset"
  Sys.setenv(TZ = tz(x[1]))
  on.exit(if (otz == "unset") Sys.unsetenv("TZ")
          else  Sys.setenv(TZ = otz))

  rng <- range(x)
  diff <- difftime(rng[2], rng[1], units = "secs")

  binunits <- pretty_unit(diff/n)

  f <- get(paste("pretty", binunits, sep = "_"), mode = "function")
  binlength <- f(diff, n)

  start <- pretty_point(min(rng), binunits, binlength)
  end <- pretty_point(max(rng), binunits, binlength, start = FALSE)

  breaks <- seq.POSIXt(start, end, paste(binlength, binunits))
  breaks
}

pretty_unit <- function(x, ...) {
  if (x > 3600*24*365)
    return("year")
  if (x > 3600*24*30)
    return("month")
  if (x > 3600*24)
    return("day")
  if (x > 3600)
    return("hour")
  if (x > 60)
    return("min")
  else
    return("sec")
}

pretty_sec <- function(x, n, ...) {
  lengths <- c(1, 2, 5, 10, 15, 30, 60)
  fit <- abs(x - lengths*n)
  lengths[which.min(fit)]
}

pretty_min <- function(x, n, ...) {
  span <- x/60
  lengths <- c(1, 2, 5, 10, 15, 30, 60)
  fit <- abs(span - lengths*n)
  lengths[which.min(fit)]
}

pretty_hour <- function(x, n, ...) {
  span <- x / 3600
  lengths <- c(1, 2, 3, 4, 6, 8, 12, 24)
  fit <- abs(span - lengths*n)
  lengths[which.min(fit)]
}

pretty_day <- function(x, n, ...) {
  span <- x / (3600 * 24)
  pretty(1:span, n = n)[2]
}

pretty_month <- function(x, n, ...) {
  span <- x / (3600 * 24 * 30)
  lengths <- c(1, 2, 3, 4, 6, 12)
  fit <- abs(span - lengths*n)
  lengths[which.min(fit)]
}

pretty_year <- function(x, n, ...) {
  span <- x / (3600 * 24 * 365)
  pretty(1:span, n = n)[2]
}

pretty_point <- function(x, units, length, start = TRUE, ...) {
  x <- as.POSIXct(x)

  if (units %in% c("day", "year")) {

    if (start) return(floor_date(x, units))
    else return(ceiling_date(x, units))

  } else {

    floors <- c("sec", "min", "hour", "day", "month", "year")
    floorto <- floors[match(units, floors) + 1L]
    lower <- floor_date(x, floorto)
    upper <- ceiling_date(x, floorto)

    points <- seq.POSIXt(lower, upper, paste(length, units))

    if (start)
      points <- points[points <= x]
    else
      points <- points[points >= x]

    fit <- as.duration(x - points)
    fit <- abs(as.double(fit, "secs"))
    points[which.min(fit)]
  }
}

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
NULL


multiply_duration_by_number <- function(dur, num)
  new("Duration", dur@.Data * num)

multiply_interval_by_number <- function(int, num) {
  new("Interval", int@.Data * num, start = int@start, tzone = int@tzone)
}

multiply_period_by_number <- function(per, num) {
  new("Period", per@.Data * num,
      year = per@year * num,
      month = per@month * num,
      day = per@day * num,
      hour = per@hour * num,
      minute = per@minute * num)
}


#' @export
setMethod("*", signature(e1 = "Timespan", e2 = "Timespan"),
  function(e1, e2) stop_incompatible_classes(e1, e2, "*"))

#' @export
setMethod("*", signature(e1 = "Duration"), function(e1, e2) multiply_duration_by_number(e1, e2))

#' @export
setMethod("*", signature(e2 = "Duration"), function(e1, e2) multiply_duration_by_number(e2, e1))

#' @export
setMethod("*", signature(e1 = "Interval"), function(e1, e2) multiply_interval_by_number(e1, e2))

#' @export
setMethod("*", signature(e2 = "Interval"), function(e1, e2) multiply_interval_by_number(e2, e1))

#' @export
setMethod("*", signature(e1 = "Period"), function(e1, e2) multiply_period_by_number(e1, e2))

#' @export
setMethod("*", signature(e2 = "Period"), function(e1, e2) multiply_period_by_number(e2, e1))

#' @include periods.r
NULL

#' Get/set date component of a date-time
#'
#' Date-time must be a POSIXct, POSIXlt, Date, chron, yearmon, yearqtr, zoo,
#' zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
#'
#' `date()` does not yet support years before 0 C.E.
#' Also `date()` is not defined for Period objects.
#'
#' @section Base compatibility:
#' `date()` can be called without any arguments to return a string representing
#' the current date-time. This provides compatiblity with `base:date()` which
#' it overrides.
#'
#' @param x a date-time object
#' @param value an object for which the `date()` function is defined
#' @return the date of x as a Date
#' @keywords utilities manip chron methods
#' @examples
#' x <- ymd_hms("2012-03-26 23:12:13", tz = "America/New_York")
#' date(x)
#' as.Date(x) # by default as.Date assumes you want to know the date in UTC
#' as.Date(x, tz = "America/New_York")
#' date(x) <- as.Date("2000-01-02")
#' x
#' @export
date <- function(x)
  UseMethod("date")

#' @export
date.default <- function(x) {
  if (missing(x)) {
    base::date()
  } else {
    x <- as.POSIXlt(x, tz = tz(x))
    year <- x$year + 1900
    month <- x$mon + 1
    day <- x$mday
    as.Date(make_datetime(year, month, day))
  }
}
#' @export
date.Period <- function(x)
  stop("date is undefined for Period objects")

#' @rdname date
#' @export
"date<-" <- function(x, value)
  x <- x + days(date(value) - date(x))

setGeneric("date<-")

#' @export
setMethod("date<-", signature("Period"), function(x, value) {
  stop("date<- is undefined for Period objects")
})

#' @include timespans.r
#' @include durations.r
#' @include intervals.r
#' @include periods.r
#' @include Dates.r
#' @include difftimes.r
#' @include POSIXt.r
NULL


subtract_interval_from_date <- function(int, date) {
  end <- int@start + int@.Data
  dur <- as.duration(int)
  add_duration_to_date(-dur, date)
}

#' @export
setMethod("-", signature(e1 = "Duration", e2 = "missing"),
          function(e1, e2) multiply_duration_by_number(e1, -1))

#' @export
setMethod("-", signature(e1 = "Interval", e2 = "missing"),
          function(e1, e2) multiply_interval_by_number(e1, -1))

#' @export
setMethod("-", signature(e1 = "Period", e2 = "missing"),
          function(e1, e2) multiply_period_by_number(e1, -1))

#' @export
setMethod("-", signature(e1 = "ANY", e2 = "Duration"),
          function(e1, e2) e1  + (-e2))

#' @export
setMethod("-", signature(e1 = "Duration", e2 = "ANY"),
          function(e1, e2) e1  + (-e2))

#' @export
setMethod("-", signature(e1 = "ANY", e2 = "Period"),
          function(e1, e2) e1  + (-e2))

#' @export
setMethod("-", signature(e1 = "Period", e2 = "ANY"),
          function(e1, e2) e1  + (-e2))

test_that("am and pm correctly identify time of day", {
  x <- as.POSIXct(c("2009-08-03 13:01:59", "2008-08-03 10:01:59"), tz = "UTC")
  expect_equal(am(x), c(FALSE, TRUE))
  expect_equal(pm(x), c(TRUE, FALSE))
})

test_that("am and pm handle various classes of date-time object", {
  x <- as.POSIXct(c("2009-08-03 13:01:59", "2008-08-03 10:01:59"), tz = "UTC")
  expect_equal(am(x), c(FALSE, TRUE))
  expect_equal(am(as.POSIXlt(x)), c(FALSE, TRUE))
  expect_equal(am(as.Date(x)), c(TRUE, TRUE))
})

context("AM/PM")

test_that("am and pm correctly identify time of day", {
  x <- as.POSIXct(c("2009-08-03 13:01:59", "2008-08-03 10:01:59"), tz = "UTC")
  expect_equal(am(x), c(FALSE, TRUE))
  expect_equal(pm(x), c(TRUE, FALSE))
})

test_that("am and pm handle various classes of date-time object", {
  x <- as.POSIXct(c("2009-08-03 13:01:59", "2008-08-03 10:01:59"), tz = "UTC")
  expect_equal(am(x), c(FALSE, TRUE))
  expect_equal(am(as.POSIXlt(x)), c(FALSE, TRUE))
  expect_equal(am(as.Date(x)), c(TRUE, TRUE))
})

#!/bin/bash

cctz=../src/cctz/

rm -rf $cctz
mkdir $cctz
git clone https://github.com/google/cctz.git cctz_tmp
cp -r cctz_tmp/include $cctz/include

LIBCCTZ="tzfile time_zone_fixed time_zone_if time_zone_impl time_zone_info time_zone_libc time_zone_lookup time_zone_posix"
mkdir $cctz/src/

for f in $LIBCCTZ
do
    cp ./cctz_tmp/src/$f* $cctz/src/
done

rm -rf cctz_tmp

---
title: "Do more with dates and times in R"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Do more with dates and times in R}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(comment = "#>", collapse = TRUE)
```

(This vignette is an updated version of the blog post first published at [r-statistics](https://www.r-statistics.com/2012/03/do-more-with-dates-and-times-in-r-with-lubridate-1-1-0/)_

Lubridate is an R package that makes it easier to work with dates and times. Below is a concise tour of some of the things lubridate can do for you. Lubridate was created by Garrett Grolemund and Hadley Wickham, and is now maintained by Vitalie Spinu.

## Parsing dates and times

Getting R to agree that your data contains the dates and times you think it does can be tricky. Lubridate simplifies that. Identify the order in which the year, month, and day appears in your dates. Now arrange "y", "m", and "d" in the same order. This is the name of the function in lubridate that will parse your dates. For example,

```{r}
library(lubridate)
ymd("20110604")
mdy("06-04-2011")
dmy("04/06/2011")
```

Lubridate's parse functions handle a wide variety of formats and separators, which simplifies the parsing process.

If your date includes time information, add h, m, and/or s to the name of the function. `ymd_hms` is probably the most common date time format. To read the dates in with a certain time zone, supply the official name of that time zone in the `tz` argument.

```{r}
arrive <- ymd_hms("2011-06-04 12:00:00", tz = "Pacific/Auckland")
arrive
leave <- ymd_hms("2011-08-10 14:00:00", tz = "Pacific/Auckland")
leave
```

## Setting and Extracting information

Extract information from date times with the functions `second`, `minute`, `hour`, `day`, `wday`, `yday`, `week`, `month`, `year`, and `tz`. You can also use each of these to set (i.e, change) the given information. Notice that this will alter the date time. `wday` and `month` have an optional `label` argument, which replaces their numeric output with the name of the weekday or month.

```{r}
second(arrive)
second(arrive) <- 25
arrive
second(arrive) <- 0

wday(arrive)
wday(arrive, label = TRUE)
```

## Time Zones

There are two very useful things to do with dates and time zones. First, display the same moment in a different time zone. Second, create a new moment by combining an existing clock time with a new time zone. These are accomplished by `with_tz` and `force_tz`.

For example, a while ago I was in Auckland, New Zealand. I arranged to meet the co-author of lubridate, Hadley, over skype at 9:00 in the morning Auckland time. What time was that for Hadley who was back in Houston, TX?

```{r}
meeting <- ymd_hms("2011-07-01 09:00:00", tz = "Pacific/Auckland")
with_tz(meeting, "America/Chicago")
```

So the meetings occurred at 4:00 Hadley's time (and the day before no less). Of course, this was the same actual moment of time as 9:00 in New Zealand. It just appears to be a different day due to the curvature of the Earth.

What if Hadley made a mistake and signed on at 9:00 his time? What time would it then be my time?

```{r}
mistake <- force_tz(meeting, "America/Chicago")
with_tz(mistake, "Pacific/Auckland")
```

His call would arrive at 2:00 am my time! Luckily he never did that.

## Time Intervals

You can save an interval of time as an Interval class object with lubridate. This is quite useful! For example, my stay in Auckland lasted from June 4, 2011 to August 10, 2011 (which we've already saved as arrive and leave). We can create this interval in one of two ways:

```{r}
auckland <- interval(arrive, leave)
auckland
auckland <- arrive %--% leave
auckland
```

My mentor at the University of Auckland, Chris, traveled to various conferences that year including the Joint Statistical Meetings (JSM). This took him out of the country from July 20 until the end of August.

```{r}
jsm <- interval(ymd(20110720, tz = "Pacific/Auckland"), ymd(20110831, tz = "Pacific/Auckland"))
jsm
```

Will my visit overlap with and his travels? Yes.

```{r}
int_overlaps(jsm, auckland)
```

Then I better make hay while the sun shines! For what part of my visit will Chris be there?

```{r}
setdiff(auckland, jsm)
```

Other functions that work with intervals include `int_start`, `int_end`, `int_flip`, `int_shift`, `int_aligns`, `union`, `intersect`, `setdiff`, and `%within%`.

## Arithmetic with date times

Intervals are specific time spans (because they are tied to specific dates), but lubridate also supplies two general time span classes: Durations and Periods. Helper functions for creating periods are named after the units of time (plural). Helper functions for creating durations follow the same format but begin with a "d" (for duration) or, if you prefer, and "e" (for exact).

```{r}
minutes(2) ## period
dminutes(2) ## duration
```

Why two classes? Because the timeline is not as reliable as the number line. The Duration class will always supply mathematically precise results. A duration year will always equal 365 days. Periods, on the other hand, fluctuate the same way the timeline does to give intuitive results. This makes them useful for modeling clock times. For example, durations will be honest in the face of a leap year, but periods may return what you want:

```{r}
leap_year(2011) ## regular year
ymd(20110101) + dyears(1)
ymd(20110101) + years(1)

leap_year(2012) ## leap year
ymd(20120101) + dyears(1)
ymd(20120101) + years(1)
```

You can use periods and durations to do basic arithmetic with date times. For example, if I wanted to set up a reoccuring weekly skype meeting with Hadley, it would occur on:

```{r}
meetings <- meeting + weeks(0:5)
```

Hadley travelled to conferences at the same time as Chris. Which of these meetings would be affected? The last two.

```{r}
meetings %within% jsm
```

How long was my stay in Auckland?

```{r}
auckland / ddays(1)
auckland / ddays(2)
auckland / dminutes(1)
```

And so on. Alternatively, we can do modulo and integer division. Sometimes this is more sensible than division - it is not obvious how to express a remainder as a fraction of a month because the length of a month constantly changes.

```{r}
auckland %/% months(1)
auckland %% months(1)
```

Modulo with an timespan returns the remainder as a new (smaller) interval. You can turn this or any interval into a generalized time span with `as.period`.

```{r}
as.period(auckland %% months(1))
as.period(auckland)
```

### If anyone drove a time machine, they would crash

The length of months and years change so often that doing arithmetic with them can be unintuitive. Consider a simple operation, `January 31st + one month`. Should the answer be

1. `February 31st` (which doesn't exist)
2. `March 4th` (31 days after January 31), or
3. `February 28th` (assuming its not a leap year)

A basic property of arithmetic is that `a + b - b = a`. Only solution 1 obeys this property, but it is an invalid date. I've tried to make lubridate as consistent as possible by invoking the following rule *if adding or subtracting a month or a year creates an invalid date, lubridate will return an NA*. This is new with version 1.3.0, so if you're an old hand with lubridate be sure to remember this!

If you thought solution 2 or 3 was more useful, no problem. You can still get those results with clever arithmetic, or by using the special `%m+%` and `%m-%` operators. `%m+%` and `%m-%` automatically roll dates back to the last day of the month, should that be necessary.

```{r}
jan31 <- ymd("2013-01-31")
jan31 + months(0:11)
floor_date(jan31, "month") + months(0:11) + days(31)
jan31 %m+% months(0:11)
```

Notice that this will only affect arithmetic with months (and arithmetic with years if your start date it Feb 29).

## Vectorization

The code in lubridate is vectorized and ready to be used in both interactive settings and within functions. As an example, I offer a function for advancing a date to the last day of the month

```{r}
last_day <- function(date) {
  ceiling_date(date, "month") - days(1)
}
```


## Further Resources

To learn more about lubridate, including the specifics of periods and durations, please read the [original lubridate paper](https://www.jstatsoft.org/v40/i03/). Questions about lubridate can be addressed to the lubridate google group. Bugs and feature requests should be submitted to the [lubridate development page](https://github.com/tidyverse/lubridate) on github.

# vctrs methods have informative errors

    Code
      # # no common type when mixing Period/Duration/Interval
      vec_ptype2(period(), duration())
    Error <vctrs_error_ptype2>
      Can't combine `period()` <Period> and `duration()` <Duration>.
    Code
      vec_ptype2(duration(), period())
    Error <vctrs_error_ptype2>
      Can't combine `duration()` <Duration> and `period()` <Period>.
    Code
      vec_ptype2(period(), interval())
    Error <vctrs_error_ptype2>
      Can't combine `period()` <Period> and `interval()` <Interval>.
    Code
      vec_ptype2(interval(), period())
    Error <vctrs_error_ptype2>
      Can't combine `interval()` <Interval> and `period()` <Period>.
    Code
      vec_ptype2(duration(), interval())
    Error <vctrs_error_ptype2>
      Can't combine `duration()` <Duration> and `interval()` <Interval>.
    Code
      vec_ptype2(interval(), duration())
    Error <vctrs_error_ptype2>
      Can't combine `interval()` <Interval> and `duration()` <Duration>.
    Code
      # # can't cast between Period/Duration/Interval
      vec_cast(period(), duration())
    Error <vctrs_error_cast>
      Can't convert `period()` <Period> to <Duration>.
    Code
      vec_cast(duration(), period())
    Error <vctrs_error_cast>
      Can't convert `duration()` <Duration> to <Period>.
    Code
      vec_cast(period(), interval())
    Error <vctrs_error_cast>
      Can't convert `period()` <Period> to <Interval>.
    Code
      vec_cast(interval(), period())
    Error <vctrs_error_cast>
      Can't convert `interval()` <Interval> to <Period>.
    Code
      vec_cast(duration(), interval())
    Error <vctrs_error_cast>
      Can't convert `duration()` <Duration> to <Interval>.
    Code
      vec_cast(interval(), duration())
    Error <vctrs_error_cast>
      Can't convert `interval()` <Interval> to <Duration>.
    Code
      # # Period default ptype2 method falls through to `vec_default_ptype2()`
      vec_ptype2(period(), 1)
    Error <vctrs_error_ptype2>
      Can't combine `period()` <Period> and `1` <double>.
    Code
      vec_ptype2(1, period())
    Error <vctrs_error_ptype2>
      Can't combine `1` <double> and `period()` <Period>.
    Code
      # # Period default cast method falls through to `vec_default_cast()`
      vec_cast(period(), 1)
    Error <vctrs_error_cast>
      Can't convert `period()` <Period> to <double>.
    Code
      vec_cast(1, period())
    Error <vctrs_error_cast>
      Can't convert `1` <double> to <Period>.
    Code
      # # Duration default ptype2 method falls through to `vec_default_ptype2()`
      vec_ptype2(duration(), 1)
    Error <vctrs_error_ptype2>
      Can't combine `duration()` <Duration> and `1` <double>.
    Code
      vec_ptype2(1, duration())
    Error <vctrs_error_ptype2>
      Can't combine `1` <double> and `duration()` <Duration>.
    Code
      # # Duration default cast method falls through to `vec_default_cast()`
      vec_cast(duration(), 1)
    Error <vctrs_error_cast>
      Can't convert `duration()` <Duration> to <double>.
    Code
      vec_cast(1, duration())
    Error <vctrs_error_cast>
      Can't convert `1` <double> to <Duration>.
    Code
      # # Interval default ptype2 method falls through to `vec_default_ptype2()`
      vec_ptype2(interval(), 1)
    Error <vctrs_error_ptype2>
      Can't combine `interval()` <Interval> and `1` <double>.
    Code
      vec_ptype2(1, interval())
    Error <vctrs_error_ptype2>
      Can't combine `1` <double> and `interval()` <Interval>.
    Code
      # # Interval default cast method falls through to `vec_default_cast()`
      vec_cast(interval(), 1)
    Error <vctrs_error_cast>
      Can't convert `interval()` <Interval> to <double>.
    Code
      vec_cast(1, interval())
    Error <vctrs_error_cast>
      Can't convert `1` <double> to <Interval>.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/am-pm.r
\name{am}
\alias{am}
\alias{pm}
\title{Does date time occur in the am or pm?}
\usage{
am(x)

pm(x)
}
\arguments{
\item{x}{a date-time object}
}
\value{
TRUE or FALSE depending on whether x occurs in the am or pm
}
\description{
Does date time occur in the am or pm?
}
\examples{
x <- ymd("2012-03-26")
am(x)
pm(x)
}
\keyword{chron}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-zones.r
\name{with_tz}
\alias{with_tz}
\title{Get date-time in a different time zone}
\usage{
with_tz(time, tzone = "")
}
\arguments{
\item{time}{a POSIXct, POSIXlt, Date, chron date-time object or a data.frame
object. When a data.frame all POSIXt elements of a data.frame are processed
with \code{with_tz()} and new data.frame is returned.}

\item{tzone}{a character string containing the time zone to convert to. R
must recognize the name contained in the string as a time zone on your
system.}
}
\value{
a POSIXct object in the updated time zone
}
\description{
with_tz returns a date-time as it would appear in a different time zone.
The actual moment of time measured does not change, just the time zone it is
measured in. with_tz defaults to the Universal Coordinated time zone (UTC)
when an unrecognized time zone is inputted. See \code{\link[=Sys.timezone]{Sys.timezone()}}
for more information on how R recognizes time zones.
}
\examples{
x <- ymd_hms("2009-08-07 00:00:01", tz = "America/New_York")
with_tz(x, "GMT")
}
\seealso{
\code{\link[=force_tz]{force_tz()}}
}
\keyword{chron}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-zones.r
\name{with_tz}
\alias{with_tz}
\alias{with_tz.default}
\title{Get date-time in a different time zone}
\usage{
with_tz(time, tzone = "", ...)

\method{with_tz}{default}(time, tzone = "", ...)
}
\arguments{
\item{time}{a POSIXct, POSIXlt, Date, chron date-time object or a data.frame
object. When a data.frame all POSIXt elements of a data.frame are processed
with \code{with_tz()} and new data.frame is returned.}

\item{tzone}{a character string containing the time zone to convert to. R
must recognize the name contained in the string as a time zone on your
system.}

\item{...}{Parameters passed to other methods.}
}
\value{
a POSIXct object in the updated time zone
}
\description{
with_tz returns a date-time as it would appear in a different time zone.
The actual moment of time measured does not change, just the time zone it is
measured in. with_tz defaults to the Universal Coordinated time zone (UTC)
when an unrecognized time zone is inputted. See \code{\link[=Sys.timezone]{Sys.timezone()}}
for more information on how R recognizes time zones.
}
\examples{
x <- ymd_hms("2009-08-07 00:00:01", tz = "America/New_York")
with_tz(x, "GMT")
}
\seealso{
\code{\link[=force_tz]{force_tz()}}
}
\keyword{chron}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercion.r
\name{as.duration}
\alias{as.duration}
\alias{as.duration,numeric-method}
\alias{as.duration,logical-method}
\alias{as.duration,difftime-method}
\alias{as.duration,Interval-method}
\alias{as.duration,Duration-method}
\alias{as.duration,Period-method}
\alias{as.duration,character-method}
\title{Change an object to a duration}
\usage{
as.duration(x, ...)
}
\arguments{
\item{x}{Object to be coerced to a duration}

\item{...}{Parameters passed to other methods. Currently unused.}
}
\value{
A duration object
}
\description{
as.duration changes Interval, Period and numeric class objects to
Duration objects. Numeric objects are changed to Duration objects
with the seconds unit equal to the numeric value.
}
\details{
Durations are exact time measurements, whereas periods are relative time
measurements. See \linkS4class{Period}. The length of a period depends
on when it occurs. Hence, a one to one mapping does not exist between
durations and periods. When used with a period object, as.duration provides
an inexact estimate of the length of the period; each time unit is assigned
its most common number of seconds. A period of one month is converted to
2628000 seconds (approximately 30.42 days). This ensures that 12 months will
sum to 365 days, or one normal year. For an exact transformation, first
transform the period to an interval with \code{\link[=as.interval]{as.interval()}}.
}
\examples{
span <- interval(ymd("2009-01-01"), ymd("2009-08-01")) # interval
as.duration(span)
as.duration(10) # numeric
dur <- duration(hours = 10, minutes = 6)
as.numeric(dur, "hours")
as.numeric(dur, "minutes")
}
\seealso{
\linkS4class{Duration}, \code{\link[=duration]{duration()}}
}
\keyword{chron}
\keyword{classes}
\keyword{manip}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercion.r
\name{as.duration}
\alias{as.duration}
\alias{as.duration,numeric-method}
\alias{as.duration,logical-method}
\alias{as.duration,difftime-method}
\alias{as.duration,Interval-method}
\alias{as.duration,Duration-method}
\alias{as.duration,Period-method}
\alias{as.duration,character-method}
\title{Change an object to a duration}
\usage{
as.duration(x, ...)
}
\arguments{
\item{x}{Object to be coerced to a duration}

\item{...}{Parameters passed to other methods. Currently unused.}
}
\value{
A duration object
}
\description{
as.duration changes Interval, Period and numeric class objects to
Duration objects. Numeric objects are changed to Duration objects
with the seconds unit equal to the numeric value.
}
\details{
Durations are exact time measurements, whereas periods are relative time
measurements. See \linkS4class{Period}. The length of a period depends
on when it occurs. Hence, a one to one mapping does not exist between
durations and periods. When used with a period object, as.duration provides
an inexact estimate of the length of the period; each time unit is assigned
its most common number of seconds. A period of one month is converted to
2628000 seconds (approximately 30.42 days). This ensures that 12 months will
sum to 365 days, or one normal year. For an exact transformation, first
transform the period to an interval with \code{\link[=as.interval]{as.interval()}}.
}
\examples{
span <- interval(ymd("2009-01-01"), ymd("2009-08-01")) #interval
as.duration(span)
as.duration(10) # numeric
dur <- duration(hours = 10, minutes = 6)
as.numeric(dur, "hours")
as.numeric(dur, "minutes")

}
\seealso{
\linkS4class{Duration}, \code{\link[=duration]{duration()}}
}
\keyword{chron}
\keyword{classes}
\keyword{manip}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/instants.r
\name{is.instant}
\alias{is.instant}
\alias{instant}
\alias{instants}
\alias{is.timepoint}
\title{Is x a date-time object?}
\usage{
is.instant(x)

is.timepoint(x)
}
\arguments{
\item{x}{an R object}
}
\value{
TRUE if x is a POSIXct, POSIXlt, or Date object, FALSE otherwise.
}
\description{
An instant is a specific moment in time. Most common date-time
objects (e.g, POSIXct, POSIXlt, and Date objects) are instants.
}
\examples{
is.instant(as.Date("2009-08-03")) # TRUE
is.timepoint(5) # FALSE
}
\seealso{
\code{\link[=is.timespan]{is.timespan()}}, \code{\link[=is.POSIXt]{is.POSIXt()}}, \code{\link[=is.Date]{is.Date()}}
}
\keyword{chron}
\keyword{logic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{lakers}
\alias{lakers}
\title{Lakers 2008-2009 basketball data set}
\description{
This data set contains play by play statistics of each Los
Angeles Lakers basketball game in the 2008-2009 season. Data
includes the date, opponent, and type of each game (home or
away). Each play is described by the time on the game clock
when the play was made, the period in which the play was
attempted, the type of play, the player and team who made the
play, the result of the play, and the location on the court
where each play was made.
}
\references{
Originally taken from www.basketballgeek.com/data/.
}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-tz.r
\name{tz}
\alias{tz}
\alias{tz<-}
\title{Get/set time zone component of a date-time}
\usage{
tz(x)

tz(x) <- value
}
\arguments{
\item{x}{A date-time vector, usually of class \code{POSIXct} or \code{POSIXlt}.}

\item{value}{New value of time zone.}
}
\value{
A character vector of length 1. An empty string (\code{""}) represents
your current time zone.

For backward compatibility, the time zone of a date, \code{NA}, or
character vector is \code{"UTC"}.
}
\description{
Conveniently get and set the time zone of a date-time.

\verb{tz<-} is an alias for \code{\link[=force_tz]{force_tz()}}, which preserves the local time,
creating a different instant in time. Use \code{\link[=with_tz]{with_tz()}} if you want keep
the instant the same, but change the printed representation.
}
\section{Valid time zones}{

Time zones are stored in system specific database, so are not guaranteed
to be the same on every system (however, they are usually pretty similar
unless your system is very out of date). You can see a complete list with
\code{\link[=OlsonNames]{OlsonNames()}}.
}

\examples{
x <- y <- ymd_hms("2012-03-26 10:10:00", tz = "UTC")
tz(x)

# Note that setting tz() preserved the clock time, which implies
# that the actual instant in time is changing
tz(y) <- "Pacific/Auckland"
y
x - y

# This is the same as force_tz()
force_tz(x, "Pacific/Auckland")

# Use with_tz() if you want to change the time zone, leave
# the instant in time the same
with_tz(x, "Pacific/Auckland")
}
\seealso{
See \link{DateTimeClasses} for a description of the underlying
\code{tzone} attribute..
}
\keyword{chron}
\keyword{manip}
\keyword{methods}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/POSIXt.r
\docType{data}
\name{is.POSIXt}
\alias{is.POSIXt}
\alias{is.POSIXlt}
\alias{is.POSIXct}
\alias{POSIXct}
\alias{NA_POSIXct_}
\title{Various POSIX utilities}
\format{
An object of class \code{POSIXct} (inherits from \code{POSIXt}) of length 1.
}
\usage{
is.POSIXt(x)

is.POSIXlt(x)

is.POSIXct(x)

POSIXct(length = 0L, tz = "UTC")

NA_POSIXct_
}
\arguments{
\item{x}{an R object}

\item{length}{A non-negative number specifying the desired length. Supplying
an argument of length other than one is an error.}

\item{tz}{a timezone (defaults to "utc")}
}
\value{
TRUE if x is a POSIXct or POSIXlt object, FALSE otherwise.
}
\description{
\code{\link[=POSIXct]{POSIXct()}} mirrors primitive contructors in base R (\code{\link[=double]{double()}},
\code{\link[=character]{character()}} etc.)
}
\examples{
is.POSIXt(as.Date("2009-08-03"))
is.POSIXt(as.POSIXct("2009-08-03"))
}
\seealso{
\code{\link[=is.instant]{is.instant()}}, \code{\link[=is.timespan]{is.timespan()}}, \code{\link[=is.Date]{is.Date()}}
}
\keyword{chron}
\keyword{datasets}
\keyword{logic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timespans.r
\docType{class}
\name{Timespan-class}
\alias{Timespan-class}
\alias{*,Timespan,Timespan-method}
\alias{\%/\%,Timespan,Timespan-method}
\alias{\%/\%,difftime,Timespan-method}
\alias{\%/\%,Interval,Interval-method}
\alias{\%/\%,Period,Period-method}
\alias{\%/\%,Interval,Duration-method}
\alias{\%/\%,Duration,Interval-method}
\alias{\%/\%,Interval,Period-method}
\alias{\%/\%,Period,Interval-method}
\title{Timespan class}
\description{
Timespan is an S4 class with no slots. It is extended by the
\linkS4class{Interval}, \linkS4class{Period}, and \linkS4class{Duration}
classes.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format_ISO8601.r
\name{format_ISO8601_precision_check}
\alias{format_ISO8601_precision_check}
\title{Provide a format for ISO8601 dates and times with the requested precision.}
\usage{
format_ISO8601_precision_check(precision, max_precision, usetz = FALSE)
}
\arguments{
\item{precision}{The amount of precision to represent with
substrings of "ymdhms", as "y"ear, "m"onth, "d"ay, "h"our,
"m"inute, and "s"econd. (e.g. "ymdhm" would show precision
through minutes.  When \code{NULL}, full precision for the object
is shown.}

\item{max_precision}{The maximum precision allowed to be output.}

\item{usetz}{Include the time zone in the formatting.  If
\code{usetz} is \code{TRUE}, the time zone is included. If
\code{usetz} is \code{"Z"}, the time is converted to "UTC" and
the time zone is indicated with "Z" ISO8601 notation.}
}
\description{
Provide a format for ISO8601 dates and times with the requested precision.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format_ISO8601.r
\name{format_ISO8601_precision_check}
\alias{format_ISO8601_precision_check}
\title{Provide a format for ISO8601 dates and times with the requested precision.}
\usage{
format_ISO8601_precision_check(precision, max_precision, usetz = FALSE)
}
\arguments{
\item{precision}{The amount of precision to represent with substrings of
"ymdhms", as "y"ear, "m"onth, "d"ay, "h"our, "m"inute, and
"s"econd. (e.g. "ymdhm" would show precision through minutes. When NULL,
\code{max_precision} is returned.  When \code{precision} is more precise than
\code{max_precision}, a warning is given and \code{max_precision} is returned.}

\item{max_precision}{The maximum precision allowed to be output.}

\item{usetz}{Include the timezone in the output format}
}
\description{
Provide a format for ISO8601 dates and times with the requested precision.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-quarter.r
\name{quarter}
\alias{quarter}
\alias{semester}
\title{Get the fiscal quarter and semester of a date-time}
\usage{
quarter(x, with_year = FALSE, fiscal_start = 1)

semester(x, with_year = FALSE)
}
\arguments{
\item{x}{a date-time object of class POSIXct, POSIXlt, Date, chron, yearmon,
yearqtr, zoo, zooreg, timeDate, xts, its, ti, jul, timeSeries, fts or
anything else that can be converted with as.POSIXlt}

\item{with_year}{logical indicating whether or not to include the quarter's
year.}

\item{fiscal_start}{numeric indicating the starting month of a fiscal year}
}
\value{
numeric
}
\description{
Quarters divide the year into fourths. Semesters divide the year into halfs.
}
\examples{
x <- ymd(c("2012-03-26", "2012-05-04", "2012-09-23", "2012-12-31"))
quarter(x)
quarter(x, with_year = TRUE)
quarter(x, with_year = TRUE, fiscal_start = 11)
semester(x)
semester(x, with_year = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-quarter.r
\name{quarter}
\alias{quarter}
\alias{semester}
\title{Get the fiscal quarter and semester of a date-time}
\usage{
quarter(
  x,
  type = "quarter",
  fiscal_start = 1,
  with_year = identical(type, "year.quarter")
)

semester(x, with_year = FALSE)
}
\arguments{
\item{x}{a date-time object of class POSIXct, POSIXlt, Date, chron, yearmon,
yearqtr, zoo, zooreg, timeDate, xts, its, ti, jul, timeSeries, fts or
anything else that can be converted with as.POSIXlt}

\item{type}{the format to be returned for the quarter. Can be one one of
"quarter" - return numeric quarter (default), "year.quarter" return
fractional numeric year.quarter, "date_first" or "date_last" which return
the date at the quarter's start end end.}

\item{fiscal_start}{numeric indicating the starting month of a fiscal year.}

\item{with_year}{logical indicating whether or not to include the quarter or
semester's year (deprecated; use the \code{type} parameter instead).}
}
\value{
numeric or a vector of class POSIXct if \code{type} argument is
\code{date_first} or \code{date_last}
}
\description{
Quarters divide the year into fourths. Semesters divide the year into halfs.
}
\examples{
x <- ymd(c("2012-03-26", "2012-05-04", "2012-09-23", "2012-12-31"))
quarter(x)
quarter(x, type = "year.quarter")
quarter(x, type = "year.quarter", fiscal_start = 11)
quarter(x, type = "date_first", fiscal_start = 11)
quarter(x, type = "date_last", fiscal_start = 11)
semester(x)
semester(x, with_year = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-month.r
\name{month}
\alias{month}
\alias{month<-}
\title{Get/set months component of a date-time}
\usage{
month(x, label = FALSE, abbr = TRUE, locale = Sys.getlocale("LC_TIME"))

month(x) <- value
}
\arguments{
\item{x}{a date-time object}

\item{label}{logical. TRUE will display the month as a character string such
as "January." FALSE will display the month as a number.}

\item{abbr}{logical. FALSE will display the month as a character string
label, such as "January". TRUE will display an abbreviated version of the
label, such as "Jan". abbr is disregarded if label = FALSE.}

\item{locale}{for month, locale to use for month names. Default to current locale.}

\item{value}{a numeric object}
}
\value{
If \code{label = FALSE}: month as number (1-12, 1 = January, 12 = December),
otherwise as an ordered factor.
}
\description{
Date-time must be a POSIXct, POSIXlt, Date, Period, chron, yearmon, yearqtr, zoo,
zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
}
\examples{
x <- ymd("2012-03-26")
month(x)
month(x) <- 1
month(x) <- 13
month(x) > 3

month(ymd(080101))
month(ymd(080101), label = TRUE)
month(ymd(080101), label = TRUE, abbr = FALSE)
month(ymd(080101) + months(0:11), label = TRUE)
}
\keyword{chron}
\keyword{manip}
\keyword{methods}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-year.r
\name{year}
\alias{year}
\alias{year<-}
\alias{isoyear}
\alias{epiyear}
\title{Get/set years component of a date-time}
\usage{
year(x)

year(x) <- value

isoyear(x)

epiyear(x)
}
\arguments{
\item{x}{a date-time object}

\item{value}{a numeric object}
}
\value{
the years element of x as a decimal number
}
\description{
Date-time must be a POSIXct, POSIXlt, Date, Period or any other object
convertible to POSIXlt.

\code{isoyear()} returns years according to the ISO 8601 week calendar.

\code{epiyear()} returns years according to the epidemilogical week calendars.
}
\details{
year does not yet support years before 0 C.E.
}
\examples{
x <- ymd("2012-03-26")
year(x)
year(x) <- 2001
year(x) > 1995
}
\references{
\url{https://en.wikipedia.org/wiki/ISO_week_date}
}
\keyword{chron}
\keyword{manip}
\keyword{methods}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-year.r
\name{year}
\alias{year}
\alias{year<-}
\alias{isoyear}
\alias{epiyear}
\title{Get/set years component of a date-time}
\usage{
year(x)

year(x) <- value

isoyear(x)

epiyear(x)
}
\arguments{
\item{x}{a date-time object}

\item{value}{a numeric object}
}
\value{
the years element of x as a decimal number
}
\description{
Date-time must be a POSIXct, POSIXlt, Date, Period or any other object
convertible to POSIXlt.

\code{isoyear()} returns years according to the ISO 8601 week calendar.

\code{epiyear()} returns years according to the epidemiological week calendars.
}
\details{
year does not yet support years before 0 C.E.
}
\examples{
x <- ymd("2012-03-26")
year(x)
year(x) <- 2001
year(x) > 1995
}
\references{
\url{https://en.wikipedia.org/wiki/ISO_week_date}
}
\keyword{chron}
\keyword{manip}
\keyword{methods}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/difftimes.r
\name{make_difftime}
\alias{make_difftime}
\title{Create a difftime object.}
\usage{
make_difftime(num = NULL, units = "auto", ...)
}
\arguments{
\item{num}{Optional number of seconds}

\item{units}{a character vector that lists the type of units to use for the
display of the return value (see examples). If \code{units} is "auto" (the
default) the display units are computed automatically. This might create
undesirable effects when converting \code{difftime} objects to numeric
values in data processing.}

\item{...}{a list of time units to be included in the difftime and their amounts. Seconds,
minutes, hours, days, and weeks are supported. Normally only one of \code{num} or \code{...} are present. If
both are present, the \code{difftime} objects are concatenated.}
}
\value{
a difftime object
}
\description{
\code{make_difftime()} creates a difftime object with the specified number of
units. Entries for different units are cumulative. difftime displays
durations in various units,  but these units are estimates given for
convenience. The underlying object is always recorded as a fixed number of
seconds.
}
\details{
Conceptually, difftime objects are a type of duration. They measure the
exact passage of time but do not always align with measurements
made in larger units of time such as hours, months and years.
This is because the length of larger time units can be affected
by conventions such as leap years
and Daylight Savings Time. \pkg{lubridate} provides a second class for measuring durations, the Duration class.
}
\examples{
make_difftime(1)
make_difftime(60)
make_difftime(3600)
make_difftime(3600, units = "minute")
# Time difference of 60 mins
make_difftime(second = 90)
# Time difference of 1.5 mins
make_difftime(minute = 1.5)
# Time difference of 1.5 mins
make_difftime(second = 3, minute = 1.5, hour = 2, day = 6, week = 1)
# Time difference of 13.08441 days
make_difftime(hour = 1, minute = -60)
# Time difference of 0 secs
make_difftime(day = -1)
# Time difference of -1 days
make_difftime(120, day = -1, units = "minute")
# Time differences in mins
}
\seealso{
\code{\link[=duration]{duration()}}, \code{\link[=as.duration]{as.duration()}}
}
\keyword{chron}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-zones.r
\name{local_time}
\alias{local_time}
\title{Get local time from a date-time vector.}
\usage{
local_time(dt, tz = NULL, units = "secs")
}
\arguments{
\item{dt}{a date-time object.}

\item{tz}{a character vector of timezones for which to compute the local time.}

\item{units}{passed directly to \code{\link[=as.difftime]{as.difftime()}}.}
}
\description{
\code{local_time} retrieves day clock time in specified time zones. Computation is
vectorized over both \code{dt} and \code{tz} arguments, the shortest is recycled in
accordance with standard R rules.
}
\examples{

x <- ymd_hms(c("2009-08-07 01:02:03", "2009-08-07 10:20:30"))
local_time(x, units = "secs")
local_time(x, units = "hours")
local_time(x, "Europe/Amsterdam")
local_time(x, "Europe/Amsterdam") == local_time(with_tz(x, "Europe/Amsterdam"))

x <- ymd_hms("2009-08-07 01:02:03")
local_time(x, c("America/New_York", "Europe/Amsterdam", "Asia/Shanghai"), unit = "hours")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-minute.r
\name{minute}
\alias{minute}
\alias{minute<-}
\title{Get/set minutes component of a date-time}
\usage{
minute(x)

minute(x) <- value
}
\arguments{
\item{x}{a date-time object}

\item{value}{numeric value to be assigned}
}
\value{
the minutes element of x as a decimal number
}
\description{
Date-time must be a  POSIXct, POSIXlt, Date, Period, chron, yearmon, yearqtr, zoo,
zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
}
\examples{
x <- ymd("2012-03-26")
minute(x)
minute(x) <- 1
minute(x) <- 61
minute(x) > 2
}
\keyword{chron}
\keyword{manip}
\keyword{methods}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/intervals.r
\name{interval}
\alias{interval}
\alias{\%--\%}
\alias{is.interval}
\alias{int_start}
\alias{int_start<-}
\alias{int_end}
\alias{int_end<-}
\alias{int_length}
\alias{int_flip}
\alias{int_shift}
\alias{int_overlaps}
\alias{int_standardize}
\alias{int_aligns}
\alias{int_diff}
\title{Utilities for creation and manipulation of \code{Interval} objects}
\usage{
interval(start = NULL, end = NULL, tzone = tz(start))

start \%--\% end

is.interval(x)

int_start(int)

int_start(int) <- value

int_end(int)

int_end(int) <- value

int_length(int)

int_flip(int)

int_shift(int, by)

int_overlaps(int1, int2)

int_standardize(int)

int_aligns(int1, int2)

int_diff(times)
}
\arguments{
\item{start, end}{POSIXt, Date or a character vectors. When \code{start} is a
character vector and end is \code{NULL}, ISO 8601 specification is assumed but
with much more permisive lubridate style parsing both for dates and periods
(see examples).}

\item{tzone}{a recognized timezone to display the interval in}

\item{x}{an R object}

\item{int}{an interval object}

\item{value}{interval's start/end to be assigned to \code{int}}

\item{by}{A period or duration object to shift by (for \code{int_shift})}

\item{int1}{an Interval object (for \code{int_overlaps()}, \code{int_aligns()})}

\item{int2}{an Interval object (for \code{int_overlaps()}, \code{int_aligns()})}

\item{times}{A vector of POSIXct, POSIXlt or Date class date-times (for
\code{int_diff()})}
}
\value{
\code{interval()} -- \linkS4class{Interval} object.

\code{int_start()} and \code{int_end()} return a POSIXct date object when
used as an accessor. Nothing when used as a setter.

\code{int_length()} -- numeric length of the interval in
seconds. A negative number connotes a negative interval.

\code{int_flip()} -- flipped interval object

\code{int_shift()} -- an Interval object

\code{int_overlaps()} -- logical, TRUE if int1 and int2 overlap by at
least one second. FALSE otherwise

\code{int_aligns()} -- logical, TRUE if int1 and int2 begin or end on the
same moment. FALSE otherwise

\code{int_diff()} -- interval object that contains the n-1 intervals
between the n date-time in times
}
\description{
\code{interval()} creates an \linkS4class{Interval} object with the specified start and
end dates. If the start date occurs before the end date, the interval will be
positive. Otherwise, it will be negative. Character vectors in ISO 8601
format are supported from v1.7.2.

\code{int_start()}/\code{int_end()} and \verb{int_start<-()}/\verb{int_end<-()} are
"accessors" and "setters" respectively of the start/end date of an
interval.

\code{int_flip()} reverses the order of the start date and end date in an
interval. The new interval takes place during the same timespan as the
original interval, but has the opposite direction.

\code{int_shift()} shifts the start and end dates of an interval up or down the
timeline by a specified amount. Note that this may change the exact length of
the interval if the interval is shifted by a Period object. Intervals shifted
by a Duration or difftime object will retain their exact length in seconds.

\code{int_overlaps()} tests if two intervals overlap.

\code{int_standardize()} ensures all intervals in an interval object are
positive. If an interval is not positive, flip it so that it retains its
endpoints but becomes positive.

\code{int_aligns()} tests if two intervals share an endpoint. The direction of
each interval is ignored. int_align tests whether the earliest or latest
moments of each interval occur at the same time.

\code{int_diff()} returns the intervals that occur between the elements of a
vector of date-times. \code{int_diff()} is similar to the POSIXt and Date
methods of \code{\link[=diff]{diff()}}, but returns an \linkS4class{Interval} object instead
of a difftime object.
}
\details{
Intervals are time spans bound by two real date-times.  Intervals can be
accurately converted to either period or duration objects using
\code{\link[=as.period]{as.period()}}, \code{\link[=as.duration]{as.duration()}}. Since an interval is anchored to a fixed
history of time, both the exact number of seconds that passed and the number
of variable length time units that occurred during the interval can be
calculated.
}
\examples{
interval(ymd(20090201), ymd(20090101))

date1 <- ymd_hms("2009-03-08 01:59:59")
date2 <- ymd_hms("2000-02-29 12:00:00")
interval(date2, date1)
interval(date1, date2)
span <- interval(ymd(20090101), ymd(20090201))

### ISO Intervals

interval("2007-03-01T13:00:00Z/2008-05-11T15:30:00Z")
interval("2007-03-01T13:00:00Z/P1Y2M10DT2H30M")
interval("P1Y2M10DT2H30M/2008-05-11T15:30:00Z")
interval("2008-05-11/P2H30M")

### More permisive parsing (as long as there are no intermittent / characters)
interval("2008 05 11/P2hours 30minutes")
interval("08 05 11/P 2h 30m")

is.interval(period(months = 1, days = 15)) # FALSE
is.interval(interval(ymd(20090801), ymd(20090809))) # TRUE
int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int_start(int)
int_start(int) <- ymd("2001-06-01")
int

int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int_end(int)
int_end(int) <- ymd("2002-06-01")
int
int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int_length(int)
int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int_flip(int)
int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int_shift(int, duration(days = 11))
int_shift(int, duration(hours = -1))
int1 <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int2 <- interval(ymd("2001-06-01"), ymd("2002-06-01"))
int3 <- interval(ymd("2003-01-01"), ymd("2004-01-01"))

int_overlaps(int1, int2) # TRUE
int_overlaps(int1, int3) # FALSE
int <- interval(ymd("2002-01-01"), ymd("2001-01-01"))
int_standardize(int)
int1 <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int2 <- interval(ymd("2001-06-01"), ymd("2002-01-01"))
int3 <- interval(ymd("2003-01-01"), ymd("2004-01-01"))

int_aligns(int1, int2) # TRUE
int_aligns(int1, int3) # FALSE
dates <- now() + days(1:10)
int_diff(dates)
}
\seealso{
\linkS4class{Interval}, \code{\link[=as.interval]{as.interval()}}, \code{\link{\%within\%}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/intervals.r
\name{interval}
\alias{interval}
\alias{\%--\%}
\alias{is.interval}
\alias{int_start}
\alias{int_start<-}
\alias{int_end}
\alias{int_end<-}
\alias{int_length}
\alias{int_flip}
\alias{int_shift}
\alias{int_overlaps}
\alias{int_standardize}
\alias{int_aligns}
\alias{int_diff}
\title{Utilities for creation and manipulation of \code{Interval} objects}
\usage{
interval(start = NULL, end = NULL, tzone = tz(start))

start \%--\% end

is.interval(x)

int_start(int)

int_start(int) <- value

int_end(int)

int_end(int) <- value

int_length(int)

int_flip(int)

int_shift(int, by)

int_overlaps(int1, int2)

int_standardize(int)

int_aligns(int1, int2)

int_diff(times)
}
\arguments{
\item{start, end}{POSIXt, Date or a character vectors. When \code{start} is a
character vector and end is \code{NULL}, ISO 8601 specification is assumed but
with much more permisive lubridate style parsing both for dates and periods
(see examples).}

\item{tzone}{a recognized timezone to display the interval in}

\item{x}{an R object}

\item{int}{an interval object}

\item{value}{interval's start/end to be assigned to \code{int}}

\item{by}{A period or duration object to shift by (for \code{int_shift})}

\item{int1}{an Interval object (for \code{int_overlaps()}, \code{int_aligns()})}

\item{int2}{an Interval object (for \code{int_overlaps()}, \code{int_aligns()})}

\item{times}{A vector of POSIXct, POSIXlt or Date class date-times (for
\code{int_diff()})}
}
\value{
\code{interval()} -- \linkS4class{Interval} object.

\code{int_start()} and \code{int_end()} return a POSIXct date object when
used as an accessor. Nothing when used as a setter.

\code{int_length()} -- numeric length of the interval in
seconds. A negative number connotes a negative interval.

\code{int_flip()} -- flipped interval object

\code{int_shift()} -- an Interval object

\code{int_overlaps()} -- logical, TRUE if int1 and int2 overlap by at
least one second. FALSE otherwise

\code{int_aligns()} -- logical, TRUE if int1 and int2 begin or end on the
same moment. FALSE otherwise

\code{int_diff()} -- interval object that contains the n-1 intervals
between the n date-time in times
}
\description{
\code{interval()} creates an \linkS4class{Interval} object with the specified start and
end dates. If the start date occurs before the end date, the interval will be
positive. Otherwise, it will be negative. Character vectors in ISO 8601
format are supported from v1.7.2.

\code{int_start()}/\code{int_end()} and \verb{int_start<-()}/\verb{int_end<-()} are
"accessors" and "setters" respectively of the start/end date of an
interval.

\code{int_flip()} reverses the order of the start date and end date in an
interval. The new interval takes place during the same timespan as the
original interval, but has the opposite direction.

\code{int_shift()} shifts the start and end dates of an interval up or down the
timeline by a specified amount. Note that this may change the exact length of
the interval if the interval is shifted by a Period object. Intervals shifted
by a Duration or difftime object will retain their exact length in seconds.

\code{int_overlaps()} tests if two intervals overlap.

\code{int_standardize()} ensures all intervals in an interval object are
positive. If an interval is not positive, flip it so that it retains its
endpoints but becomes positive.

\code{int_aligns()} tests if two intervals share an endpoint. The direction of
each interval is ignored. int_align tests whether the earliest or latest
moments of each interval occur at the same time.

\code{int_diff()} returns the intervals that occur between the elements of a
vector of date-times. \code{int_diff()} is similar to the POSIXt and Date
methods of \code{\link[=diff]{diff()}}, but returns an \linkS4class{Interval} object instead
of a difftime object.
}
\details{
Intervals are time spans bound by two real date-times.  Intervals can be
accurately converted to either period or duration objects using
\code{\link[=as.period]{as.period()}}, \code{\link[=as.duration]{as.duration()}}. Since an interval is anchored to a fixed
history of time, both the exact number of seconds that passed and the number
of variable length time units that occurred during the interval can be
calculated.
}
\examples{
interval(ymd(20090201), ymd(20090101))

date1 <- ymd_hms("2009-03-08 01:59:59")
date2 <- ymd_hms("2000-02-29 12:00:00")
interval(date2, date1)
interval(date1, date2)
span <- interval(ymd(20090101), ymd(20090201))

### ISO Intervals

interval("2007-03-01T13:00:00Z/2008-05-11T15:30:00Z")
interval("2007-03-01T13:00:00Z/P1Y2M10DT2H30M")
interval("P1Y2M10DT2H30M/2008-05-11T15:30:00Z")
interval("2008-05-11/P2H30M")

### More permisive parsing (as long as there are no intermittent / characters)
interval("2008 05 11/P2hours 30minutes")
interval("08 05 11/P 2h 30m")

is.interval(period(months= 1, days = 15)) # FALSE
is.interval(interval(ymd(20090801), ymd(20090809))) # TRUE
int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int_start(int)
int_start(int) <- ymd("2001-06-01")
int

int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int_end(int)
int_end(int) <- ymd("2002-06-01")
int
int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int_length(int)
int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int_flip(int)
int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int_shift(int, duration(days = 11))
int_shift(int, duration(hours = -1))
int1 <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int2 <- interval(ymd("2001-06-01"), ymd("2002-06-01"))
int3 <- interval(ymd("2003-01-01"), ymd("2004-01-01"))

int_overlaps(int1, int2) # TRUE
int_overlaps(int1, int3) # FALSE
int <- interval(ymd("2002-01-01"), ymd("2001-01-01"))
int_standardize(int)
int1 <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int2 <- interval(ymd("2001-06-01"), ymd("2002-01-01"))
int3 <- interval(ymd("2003-01-01"), ymd("2004-01-01"))

int_aligns(int1, int2) # TRUE
int_aligns(int1, int3) # FALSE
dates <- now() + days(1:10)
int_diff(dates)
}
\seealso{
\linkS4class{Interval}, \code{\link[=as.interval]{as.interval()}}, \code{\link{\%within\%}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/difftimes.r
\name{is.difftime}
\alias{is.difftime}
\title{Is x a difftime object?}
\usage{
is.difftime(x)
}
\arguments{
\item{x}{an R object}
}
\value{
TRUE if x is a difftime object, FALSE otherwise.
}
\description{
Is x a difftime object?
}
\examples{
is.difftime(as.Date("2009-08-03")) # FALSE
is.difftime(make_difftime(days = 12.4)) # TRUE
}
\seealso{
\code{\link[=is.instant]{is.instant()}}, \code{\link[=is.timespan]{is.timespan()}}, \code{\link[=is.interval]{is.interval()}},
\code{\link[=is.period]{is.period()}}.
}
\keyword{chron}
\keyword{logic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-dst.r
\name{dst}
\alias{dst}
\title{Get \strong{d}aylight \strong{s}avings \strong{t}ime indicator of a date-time}
\usage{
dst(x)
}
\arguments{
\item{x}{a date-time object}
}
\value{
A logical. TRUE if DST is in force, FALSE if not, NA if unknown.
}
\description{
Date-time must be a POSIXct, POSIXlt, Date, chron, yearmon, yearqtr, zoo,
zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
}
\details{
A date-time's daylight savings flag can not be set because it depends on the
date-time's year, month, day, and hour values.
}
\examples{
x <- ymd("2012-03-26")
dst(x)
}
\keyword{chron}
\keyword{methods}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse.r
\name{parse_date_time}
\alias{parse_date_time}
\alias{parse_date_time2}
\alias{fast_strptime}
\title{User friendly date-time parsing functions}
\usage{
parse_date_time(
  x,
  orders,
  tz = "UTC",
  truncated = 0,
  quiet = FALSE,
  locale = Sys.getlocale("LC_TIME"),
  select_formats = .select_formats,
  exact = FALSE,
  train = TRUE,
  drop = FALSE
)

parse_date_time2(
  x,
  orders,
  tz = "UTC",
  exact = FALSE,
  lt = FALSE,
  cutoff_2000 = 68L
)

fast_strptime(x, format, tz = "UTC", lt = TRUE, cutoff_2000 = 68L)
}
\arguments{
\item{x}{a character or numeric vector of dates}

\item{orders}{a character vector of date-time formats. Each order string is a
series of formatting characters as listed in \code{\link[base:strptime]{base::strptime()}} but might not
include the \code{"\%"} prefix. For example, "ymd" will match all the possible
dates in year, month, day order. Formatting orders might include
arbitrary separators. These are discarded. See details for implemented
formats.}

\item{tz}{a character string that specifies the time zone with which to
parse the dates}

\item{truncated}{integer, number of formats that can be missing. The most
common type of irregularity in date-time data is the truncation due to
rounding or unavailability of the time stamp. If the \code{truncated} parameter
is non-zero \code{parse_date_time()} also checks for truncated formats. For
example,  if the format order is "ymdHMS" and \code{truncated = 3},
\code{parse_date_time()} will correctly parse incomplete date-times like
\verb{2012-06-01 12:23}, \verb{2012-06-01 12} and \code{2012-06-01}. \bold{NOTE:} The \code{ymd()}
family of functions is based on \code{\link[base:strptime]{base::strptime()}} which currently fails to parse
\verb{\%Y-\%m} formats.}

\item{quiet}{logical. If \code{TRUE}, progress messages are not printed, and
\verb{No formats found} error is suppressed and the function simply returns a
vector of NAs.  This mirrors the behavior of base R functions \code{\link[base:strptime]{base::strptime()}}
and \code{\link[base:as.POSIXlt]{base::as.POSIXct()}}.}

\item{locale}{locale to be used, see \link{locales}. On Linux systems you
can use \code{system("locale -a")} to list all the installed locales.}

\item{select_formats}{A function to select actual formats for parsing from a
set of formats which matched a training subset of \code{x}. It receives a named
integer vector and returns a character vector of selected formats. Names
of the input vector are formats (not orders) that matched the training
set. Numeric values are the number of dates (in the training set) that
matched the corresponding format. You should use this argument if the
default selection method fails to select the formats in the right
order. By default the formats with most formatting tokens (\verb{\%}) are
selected and \verb{\%Y} counts as 2.5 tokens (so that it has a priority over
\verb{\%y\%m}). See examples.}

\item{exact}{logical. If \code{TRUE}, the \code{orders} parameter is interpreted as an
exact \code{\link[base:strptime]{base::strptime()}} format and no training or guessing are performed
(i.e. \code{train}, \code{drop} parameters are irrelevant).}

\item{train}{logical, default \code{TRUE}. Whether to train formats on a subset of
the input vector. The result of this is that supplied orders are sorted
according to performance on this training set, which commonly results in
increased performance. Please note that even when \code{train = FALSE} (and
\code{exact = FALSE}) guessing of the actual formats is still performed on a
pseudo-random subset of the original input vector. This might result in
\verb{All formats failed to parse} error. See notes below.}

\item{drop}{logical, default \code{FALSE}. Whether to drop formats that didn't
match on the training set. If \code{FALSE}, unmatched on the training set formats
are tried as a last resort at the end of the parsing queue. Applies only
when \code{train = TRUE}. Setting this parameter to \code{TRUE} might slightly speed up
parsing in situations involving many formats. Prior to v1.7.0 this
parameter was implicitly \code{TRUE}, which resulted in occasional surprising
behavior when rare patterns where not present in the training set.}

\item{lt}{logical. If \code{TRUE}, returned object is of class POSIXlt, and POSIXct
otherwise. For compatibility with \code{\link[base:strptime]{base::strptime()}} the default is \code{TRUE}
for \code{fast_strptime()} and \code{FALSE} for \code{parse_date_time2()}.}

\item{cutoff_2000}{integer. For \code{y} format,  two-digit numbers smaller or equal to
\code{cutoff_2000} are parsed as 20th's century, 19th's otherwise. Available only
for functions relying on \code{lubridate}s internal parser.}

\item{format}{a character string of formats. It should include all the
separators and each format must be prefixed with \%, just as in the format
argument of \code{\link[base:strptime]{base::strptime()}}.}
}
\value{
a vector of POSIXct date-time objects
}
\description{
\code{parse_date_time()} parses an input vector into POSIXct date-time
object. It differs from \code{\link[base:strptime]{base::strptime()}} in two respects. First,
it allows specification of the order in which the formats occur without the
need to include separators and the \verb{\%} prefix. Such a formatting argument is
referred to as "order". Second, it allows the user to specify several
format-orders to handle heterogeneous date-time character
representations.

\code{parse_date_time2()} is a fast C parser of numeric
orders.

\code{fast_strptime()} is a fast C parser of numeric formats only
that accepts explicit format arguments, just like
\code{\link[base:strptime]{base::strptime()}}.
}
\details{
When several format-orders are specified, \code{parse_date_time()} selects
(guesses) format-orders based on a training subset of the input
strings. After guessing the formats are ordered according to the performance
on the training set and applied recursively on the entire input vector. You
can disable training with \code{train = FALSE}.

\code{parse_date_time()}, and all derived functions, such as \code{ymd_hms()},
\code{ymd()}, etc., will drop into \code{fast_strptime()} instead of
\code{\link[base:strptime]{base::strptime()}} whenever the guessed from the input data formats are all
numeric.

The list below contains formats recognized by \pkg{lubridate}. For numeric formats
leading 0s are optional. As compared to \code{\link[base:strptime]{base::strptime()}}, some of the
formats are new or have been extended for efficiency reasons. These formats
are marked with "(*)". The fast parsers \code{parse_date_time2()} and
\code{fast_strptime()} accept only formats marked with "(!)".

\describe{ \item{\code{a}}{Abbreviated weekday name in the current
locale. (Also matches full name)}

\item{\code{A}}{Full weekday name in the current locale.  (Also matches
abbreviated name).

You don't need to specify \code{a} and \code{A} formats explicitly. Wday is
automatically handled if \code{preproc_wday = TRUE}}

\item{\code{b} (!)}{Abbreviated or full month name in the current locale. The C
parser currently understands only English month names.}

\item{\code{B} (!)}{Same as b.}

\item{\code{d} (!)}{Day of the month as decimal number (01--31 or 0--31)}

\item{\code{H} (!)}{Hours as decimal number (00--24 or 0--24).}

\item{\code{I} (!)}{Hours as decimal number (01--12 or 1--12).}

\item{\code{j}}{Day of year as decimal number (001--366 or 1--366).}

\item{\code{q} (!*)}{Quarter (1--4). The quarter month is added to the parsed month
if \code{m} format is present.}

\item{\code{m} (!*)}{Month as decimal number (01--12 or 1--12). For
\code{parse_date_time}. As a \pkg{lubridate} extension, also
matches abbreviated and full months names as \code{b} and
\code{B} formats. C parser understands only English month
names.}

\item{\code{M} (!)}{Minute as decimal number (00--59 or 0--59).}

\item{\code{p} (!)}{AM/PM indicator in the locale. Normally used in conjunction
with \code{I} and \bold{not} with \code{H}.  But the \pkg{lubridate}
C parser accepts H format as long as hour is not greater
than 12. C parser understands only English locale AM/PM
indicator.}

\item{\code{S} (!)}{Second as decimal number (00--61 or 0--61), allowing for up
to two leap-seconds (but POSIX-compliant implementations will ignore leap
seconds).}

\item{\code{OS}}{Fractional second.}

\item{\code{U}}{Week of the year as decimal number (00--53 or 0--53) using
Sunday as the first day 1 of the week (and typically with the first Sunday
of the year as day 1 of week 1).  The US convention.}

\item{\code{w}}{Weekday as decimal number (0--6, Sunday is 0).}

\item{\code{W}}{Week of the year as decimal number (00--53 or 0--53) using
Monday as the first day of week (and typically with the first Monday of the
year as day 1 of week 1).  The UK convention.}

\item{\code{y} (!*)}{Year without century (00--99 or 0--99).  In
\code{parse_date_time()} also matches year with century (Y format).}

\item{\code{Y} (!)}{Year with century.}

\item{\code{z} (!*)}{ISO8601 signed offset in hours and minutes from UTC. For
example \code{-0800}, \code{-08:00} or \code{-08}, all represent 8 hours
behind UTC. This format also matches the Z (Zulu) UTC indicator. Because
\code{\link[base:strptime]{base::strptime()}} doesn't fully support ISO8601 this format is implemented as an
union of 4 orders: Ou (Z), Oz (-0800), OO (-08:00) and Oo (-08). You can use
these four orders as any other but it is rarely
necessary. \code{parse_date_time2()} and \code{fast_strptime()} support all of
the timezone formats.}

\item{\code{Om} (!*)}{Matches numeric month and English alphabetic months
(Both, long and abbreviated forms).}

\item{\code{Op} (!*)}{Matches AM/PM English indicator.}

\item{\code{r} (*)}{Matches \code{Ip} and \code{H} orders.}

\item{\code{R} (*)}{Matches \code{HM} and\code{IMp} orders.}

\item{\code{T} (*)}{Matches \code{IMSp}, \code{HMS}, and \code{HMOS} orders.}
}
}
\note{
\code{parse_date_time()} (and the derivatives \code{ymd()}, \code{ymd_hms()}, etc.)
relies on a sparse guesser that takes at most 501 elements from the supplied
character vector in order to identify appropriate formats from the
supplied orders. If you get the error \verb{All formats failed to parse} and
you are confident that your vector contains valid dates, you should either
set \code{exact} argument to \code{TRUE} or use functions that don't perform format
guessing (\code{fast_strptime()}, \code{parse_date_time2()} or \code{\link[base:strptime]{base::strptime()}}).

For performance reasons, when timezone is not UTC,
\code{parse_date_time2()} and \code{fast_strptime()} perform no validity checks for
daylight savings time. Thus, if your input string contains an invalid date
time which falls into DST gap and \code{lt = TRUE} you will get an \code{POSIXlt}
object with a non-existent time. If \code{lt = FALSE} your time instant will be
adjusted to a valid time by adding an hour. See examples. If you want to
get NA for invalid date-times use \code{\link[=fit_to_timeline]{fit_to_timeline()}} explicitly.
}
\examples{

## ** orders are much easier to write **
x <- c("09-01-01", "09-01-02", "09-01-03")
parse_date_time(x, "ymd")
parse_date_time(x, "y m d")
parse_date_time(x, "\%y\%m\%d")
#  "2009-01-01 UTC" "2009-01-02 UTC" "2009-01-03 UTC"

## ** heterogeneous date-times **
x <- c("09-01-01", "090102", "09-01 03", "09-01-03 12:02")
parse_date_time(x, c("ymd", "ymd HM"))

## ** different ymd orders **
x <- c("2009-01-01", "02022010", "02-02-2010")
parse_date_time(x, c("dmY", "ymd"))
##  "2009-01-01 UTC" "2010-02-02 UTC" "2010-02-02 UTC"

## ** truncated time-dates **
x <- c("2011-12-31 12:59:59", "2010-01-01 12:11", "2010-01-01 12", "2010-01-01")
parse_date_time(x, "Ymd HMS", truncated = 3)

## ** specifying exact formats and avoiding training and guessing **
parse_date_time(x, c("\%m-\%d-\%y", "\%m\%d\%y", "\%m-\%d-\%y \%H:\%M"), exact = TRUE)
parse_date_time(c('12/17/1996 04:00:00','4/18/1950 0130'),
                c('\%m/\%d/\%Y \%I:\%M:\%S','\%m/\%d/\%Y \%H\%M'), exact = TRUE)

## ** quarters and partial dates **
parse_date_time(c("2016.2", "2016-04"), orders = "Yq")
parse_date_time(c("2016", "2016-04"), orders = c("Y", "Ym"))

## ** fast parsing **
\dontrun{
  options(digits.secs = 3)
  ## random times between 1400 and 3000
  tt <- as.character(.POSIXct(runif(1000, -17987443200, 32503680000)))
  tt <- rep.int(tt, 1000)

  system.time(out <- as.POSIXct(tt, tz = "UTC"))
  system.time(out1 <- ymd_hms(tt)) # constant overhead on long vectors
  system.time(out2 <- parse_date_time2(tt, "YmdHMOS"))
  system.time(out3 <- fast_strptime(tt, "\%Y-\%m-\%d \%H:\%M:\%OS"))

  all.equal(out, out1)
  all.equal(out, out2)
  all.equal(out, out3)
}

## ** how to use `select_formats` argument **
## By default \%Y has precedence:
parse_date_time(c("27-09-13", "27-09-2013"), "dmy")

## to give priority to \%y format, define your own select_format function:

my_select <-   function(trained, drop=FALSE, ...){
   n_fmts <- nchar(gsub("[^\%]", "", names(trained))) + grepl("\%y", names(trained))*1.5
   names(trained[ which.max(n_fmts) ])
}

parse_date_time(c("27-09-13", "27-09-2013"), "dmy", select_formats = my_select)

## ** invalid times with "fast" parsing **
parse_date_time("2010-03-14 02:05:06",  "YmdHMS", tz = "America/New_York")
parse_date_time2("2010-03-14 02:05:06",  "YmdHMS", tz = "America/New_York")
parse_date_time2("2010-03-14 02:05:06",  "YmdHMS", tz = "America/New_York", lt = TRUE)
}
\seealso{
\code{\link[base:strptime]{base::strptime()}}, \code{\link[=ymd]{ymd()}}, \code{\link[=ymd_hms]{ymd_hms()}}
}
\keyword{chron}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse.r
\name{parse_date_time}
\alias{parse_date_time}
\alias{parse_date_time2}
\alias{fast_strptime}
\title{User friendly date-time parsing functions}
\usage{
parse_date_time(
  x,
  orders,
  tz = "UTC",
  truncated = 0,
  quiet = FALSE,
  locale = Sys.getlocale("LC_TIME"),
  select_formats = .select_formats,
  exact = FALSE,
  train = TRUE,
  drop = FALSE
)

parse_date_time2(
  x,
  orders,
  tz = "UTC",
  exact = FALSE,
  lt = FALSE,
  cutoff_2000 = 68L
)

fast_strptime(x, format, tz = "UTC", lt = TRUE, cutoff_2000 = 68L)
}
\arguments{
\item{x}{a character or numeric vector of dates}

\item{orders}{a character vector of date-time formats. Each order string is
a series of formatting characters as listed in \code{\link[base:strptime]{base::strptime()}} but
might not include the \code{"\%"} prefix. For example, "ymd" will match all the
possible dates in year, month, day order. Formatting orders might include
arbitrary separators. These are discarded. See details for the implemented
formats. If multiple order strings are supplied, they are applied in turn
for \code{parse_date_time2()} and \code{fast_strptime()}. For \code{parse_date_time()}
the order of applied formats is determined by \code{select_formats} parameter.}

\item{tz}{a character string that specifies the time zone with which to
parse the dates}

\item{truncated}{integer, number of formats that can be missing. The most
common type of irregularity in date-time data is the truncation due to
rounding or unavailability of the time stamp. If the \code{truncated} parameter
is non-zero \code{parse_date_time()} also checks for truncated formats. For
example,  if the format order is "ymdHMS" and \code{truncated = 3},
\code{parse_date_time()} will correctly parse incomplete date-times like
\verb{2012-06-01 12:23}, \verb{2012-06-01 12} and \code{2012-06-01}. \bold{NOTE:} The
\code{ymd()} family of functions is based on \code{\link[base:strptime]{base::strptime()}} which currently
fails to parse \verb{\%Y-\%m} formats.}

\item{quiet}{logical. If \code{TRUE}, progress messages are not printed, and \verb{No formats found} error is suppressed and the function simply returns a
vector of NAs.  This mirrors the behavior of base R functions
\code{\link[base:strptime]{base::strptime()}} and \code{\link[base:as.POSIXlt]{base::as.POSIXct()}}.}

\item{locale}{locale to be used, see \link{locales}. On Linux systems you
can use \code{system("locale -a")} to list all the installed locales.}

\item{select_formats}{A function to select actual formats for parsing from a
set of formats which matched a training subset of \code{x}. It receives a named
integer vector and returns a character vector of selected formats. Names
of the input vector are formats (not orders) that matched the training
set. Numeric values are the number of dates (in the training set) that
matched the corresponding format. You should use this argument if the
default selection method fails to select the formats in the right
order. By default the formats with most formatting tokens (\verb{\%}) are
selected and \verb{\%Y} counts as 2.5 tokens (so that it has a priority over
\verb{\%y\%m}). See examples.}

\item{exact}{logical. If \code{TRUE}, the \code{orders} parameter is interpreted as an
exact \code{\link[base:strptime]{base::strptime()}} format and no training or guessing are performed
(i.e. \code{train}, \code{drop} parameters are ignored).}

\item{train}{logical, default \code{TRUE}. Whether to train formats on a subset of the
input vector. As a result the supplied orders are sorted according to performance
on this training set, which commonly results in increased performance. Please note
that even when \code{train = FALSE} (and \code{exact = FALSE}) guessing of the actual
formats is still performed on the training set (a pseudo-random subset of the
original input vector). This might result in \verb{All formats failed to parse}
error. See notes below.}

\item{drop}{logical, default \code{FALSE}. Whether to drop formats that didn't
match on the training set. If \code{FALSE}, unmatched on the training set
formats are tried as a last resort at the end of the parsing
queue. Applies only when \code{train = TRUE}. Setting this parameter to \code{TRUE}
might slightly speed up parsing in situations involving many
formats. Prior to v1.7.0 this parameter was implicitly \code{TRUE}, which
resulted in occasional surprising behavior when rare patterns where not
present in the training set.}

\item{lt}{logical. If \code{TRUE}, returned object is of class POSIXlt, and POSIXct
otherwise. For compatibility with \code{\link[base:strptime]{base::strptime()}} the default is \code{TRUE}
for \code{fast_strptime()} and \code{FALSE} for \code{parse_date_time2()}.}

\item{cutoff_2000}{integer. For \code{y} format,  two-digit numbers smaller or equal
to \code{cutoff_2000} are parsed as though starting with \code{20}, otherwise parsed
as though starting with \code{19}. Available only for functions relying on
\code{lubridate}s internal parser.}

\item{format}{a vector of formats. If multiple formats supplied they are
applied in turn till success. The formats should include all the
separators and each format letter must be prefixed with \%, just as in the
format argument of \code{\link[base:strptime]{base::strptime()}}.}
}
\value{
a vector of POSIXct date-time objects
}
\description{
\code{parse_date_time()} parses an input vector into POSIXct date-time
object. It differs from \code{\link[base:strptime]{base::strptime()}} in two respects. First,
it allows specification of the order in which the formats occur without the
need to include separators and the \verb{\%} prefix. Such a formatting argument is
referred to as "order". Second, it allows the user to specify several
format-orders to handle heterogeneous date-time character
representations.

\code{parse_date_time2()} is a fast C parser of numeric orders.

\code{fast_strptime()} is a fast C parser of numeric formats only
that accepts explicit format arguments, just like \code{\link[base:strptime]{base::strptime()}}.
}
\details{
When several format-orders are specified, \code{parse_date_time()} selects
(guesses) format-orders based on a training subset of the input
strings. After guessing the formats are ordered according to the performance
on the training set and applied recursively on the entire input vector. You
can disable training with \code{train = FALSE}.

\code{parse_date_time()}, and all derived functions, such as \code{ymd_hms()},
\code{ymd()}, etc., will drop into \code{fast_strptime()} instead of
\code{\link[base:strptime]{base::strptime()}} whenever the guessed from the input data formats are all
numeric.

The list below contains formats recognized by \pkg{lubridate}. For numeric
formats leading 0s are optional. As compared to \code{\link[base:strptime]{base::strptime()}}, some of
the formats are new or have been extended for efficiency reasons. These
formats are marked with "(*)" below. Fast parsers \code{parse_date_time2()} and
\code{fast_strptime()} accept only formats marked with "(!)".

\describe{ \item{\code{a}}{Abbreviated weekday name in the current
locale. (Also matches full name)}

\item{\code{A}}{Full weekday name in the current locale.  (Also matches
abbreviated name).

You don't need to specify \code{a} and \code{A} formats explicitly. Wday is
automatically handled if \code{preproc_wday = TRUE}}

\item{\code{b} (!)}{Abbreviated or full month name in the current locale. The C
parser currently understands only English month names.}

\item{\code{B} (!)}{Same as b.}

\item{\code{d} (!)}{Day of the month as decimal number (01--31 or 0--31)}

\item{\code{H} (!)}{Hours as decimal number (00--24 or 0--24).}

\item{\code{I} (!)}{Hours as decimal number (01--12 or 1--12).}

\item{\code{j}}{Day of year as decimal number (001--366 or 1--366).}

\item{\code{q} (!*)}{Quarter (1--4). The quarter month is added to the parsed
month if \code{m} element is present.}

\item{\code{m} (!*)}{Month as decimal number (01--12 or 1--12). For
\code{parse_date_time} also matches abbreviated and full months names as \code{b}
and \code{B} formats. C parser understands only English month names.}

\item{\code{M} (!)}{Minute as decimal number (00--59 or 0--59).}

\item{\code{p} (!)}{AM/PM indicator in the locale. Commonly used in conjunction
with \code{I} and \bold{not} with \code{H}.  But \pkg{lubridate}'s C parser accepts H
format as long as hour is not greater than 12. C parser understands only
English locale AM/PM indicator.}

\item{\code{S} (!)}{Second as decimal number (00--61 or 0--61), allowing for up
to two leap-seconds (but POSIX-compliant implementations will ignore leap
seconds).}

\item{\code{OS}}{Fractional second.}

\item{\code{U}}{Week of the year as decimal number (00--53 or 0--53) using
Sunday as the first day 1 of the week (and typically with the first Sunday
of the year as day 1 of week 1).  The US convention.}

\item{\code{w}}{Weekday as decimal number (0--6, Sunday is 0).}

\item{\code{W}}{Week of the year as decimal number (00--53 or 0--53) using
Monday as the first day of week (and typically with the first Monday of the
year as day 1 of week 1).  The UK convention.}

\item{\code{y} (!*)}{Year without century (00--99 or 0--99).  In
\code{parse_date_time()} also matches year with century (Y format).}

\item{\code{Y} (!)}{Year with century.}

\item{\code{z} (!*)}{ISO8601 signed offset in hours and minutes from UTC. For
example \code{-0800}, \code{-08:00} or \code{-08}, all represent 8 hours behind UTC. This
format also matches the Z (Zulu) UTC indicator. Because \code{\link[base:strptime]{base::strptime()}}
doesn't fully support ISO8601 this format is implemented as an union of 4
formats: Ou (Z), Oz (-0800), OO (-08:00) and Oo (-08). You can use these
formats as any other but it is rarely necessary. \code{parse_date_time2()} and
\code{fast_strptime()} support all of these formats.}

\item{\code{Om} (!*)}{Matches numeric month and English alphabetic months
(Both, long and abbreviated forms).}

\item{\code{Op} (!*)}{Matches AM/PM English indicator.}

\item{\code{r} (*)}{Matches \code{Ip} and \code{H} orders.}

\item{\code{R} (*)}{Matches \code{HM} and\code{IMp} orders.}

\item{\code{T} (*)}{Matches \code{IMSp}, \code{HMS}, and \code{HMOS} orders.}
}
}
\note{
\code{parse_date_time()} (and the derivatives \code{ymd()}, \code{ymd_hms()}, etc.)
relies on a sparse guesser that takes at most 501 elements from the
supplied character vector in order to identify appropriate formats from
the supplied orders. If you get the error \verb{All formats failed to parse}
and you are confident that your vector contains valid dates, you should
either set \code{exact} argument to \code{TRUE} or use functions that don't perform
format guessing (\code{fast_strptime()}, \code{parse_date_time2()} or
\code{\link[base:strptime]{base::strptime()}}).

For performance reasons, when timezone is not UTC,
\code{parse_date_time2()} and \code{fast_strptime()} perform no validity checks for
daylight savings time. Thus, if your input string contains an invalid date
time which falls into DST gap and \code{lt = TRUE} you will get an \code{POSIXlt}
object with a non-existent time. If \code{lt = FALSE} your time instant will be
adjusted to a valid time by adding an hour. See examples. If you want to
get NA for invalid date-times use \code{\link[=fit_to_timeline]{fit_to_timeline()}} explicitly.
}
\examples{

## ** orders are much easier to write **
x <- c("09-01-01", "09-01-02", "09-01-03")
parse_date_time(x, "ymd")
parse_date_time(x, "y m d")
parse_date_time(x, "\%y\%m\%d")
#  "2009-01-01 UTC" "2009-01-02 UTC" "2009-01-03 UTC"

## ** heterogeneous date-times **
x <- c("09-01-01", "090102", "09-01 03", "09-01-03 12:02")
parse_date_time(x, c("ymd", "ymd HM"))

## ** different ymd orders **
x <- c("2009-01-01", "02022010", "02-02-2010")
parse_date_time(x, c("dmY", "ymd"))
##  "2009-01-01 UTC" "2010-02-02 UTC" "2010-02-02 UTC"

## ** truncated time-dates **
x <- c("2011-12-31 12:59:59", "2010-01-01 12:11", "2010-01-01 12", "2010-01-01")
parse_date_time(x, "Ymd HMS", truncated = 3)

## ** specifying exact formats and avoiding training and guessing **
parse_date_time(x, c("\%m-\%d-\%y", "\%m\%d\%y", "\%m-\%d-\%y \%H:\%M"), exact = TRUE)
parse_date_time(c('12/17/1996 04:00:00','4/18/1950 0130'),
                c('\%m/\%d/\%Y \%I:\%M:\%S','\%m/\%d/\%Y \%H\%M'), exact = TRUE)

## ** quarters and partial dates **
parse_date_time(c("2016.2", "2016-04"), orders = "Yq")
parse_date_time(c("2016", "2016-04"), orders = c("Y", "Ym"))

## ** fast parsing **
\dontrun{
  options(digits.secs = 3)
  ## random times between 1400 and 3000
  tt <- as.character(.POSIXct(runif(1000, -17987443200, 32503680000)))
  tt <- rep.int(tt, 1000)

  system.time(out <- as.POSIXct(tt, tz = "UTC"))
  system.time(out1 <- ymd_hms(tt)) # constant overhead on long vectors
  system.time(out2 <- parse_date_time2(tt, "YmdHMOS"))
  system.time(out3 <- fast_strptime(tt, "\%Y-\%m-\%d \%H:\%M:\%OS"))

  all.equal(out, out1)
  all.equal(out, out2)
  all.equal(out, out3)
}

## ** how to use `select_formats` argument **
## By default \%Y has precedence:
parse_date_time(c("27-09-13", "27-09-2013"), "dmy")

## to give priority to \%y format, define your own select_format function:

my_select <-   function(trained, drop=FALSE, ...){
   n_fmts <- nchar(gsub("[^\%]", "", names(trained))) + grepl("\%y", names(trained))*1.5
   names(trained[ which.max(n_fmts) ])
}

parse_date_time(c("27-09-13", "27-09-2013"), "dmy", select_formats = my_select)

## ** invalid times with "fast" parsing **
parse_date_time("2010-03-14 02:05:06",  "YmdHMS", tz = "America/New_York")
parse_date_time2("2010-03-14 02:05:06",  "YmdHMS", tz = "America/New_York")
parse_date_time2("2010-03-14 02:05:06",  "YmdHMS", tz = "America/New_York", lt = TRUE)
}
\seealso{
\code{\link[base:strptime]{base::strptime()}}, \code{\link[=ymd]{ymd()}}, \code{\link[=ymd_hms]{ymd_hms()}}
}
\keyword{chron}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format_ISO8601.r
\name{format_ISO8601}
\alias{format_ISO8601}
\alias{format_ISO8601,Date-method}
\alias{format_ISO8601,POSIXt-method}
\alias{format_ISO8601,Interval-method}
\alias{format_ISO8601,Duration-method}
\alias{format_ISO8601,Period-method}
\title{Format in ISO8601 character format}
\usage{
format_ISO8601(x, usetz = FALSE, precision = NULL, ...)
}
\arguments{
\item{x}{An object to convert to ISO8601 character format.}

\item{usetz}{Include the time zone in the formatting (of outputs including
time; date outputs never include time zone information).}

\item{precision}{The amount of precision to represent with substrings of
"ymdhms", as "y"ear, "m"onth, "d"ay, "h"our, "m"inute, and "s"econd. (e.g.
"ymdhm" would show precision through minutes.  When \code{NULL}, full
precision for the object is shown.}

\item{...}{Additional arguments to methods.}
}
\value{
A character vector of ISO8601-formatted text.
}
\description{
Format in ISO8601 character format
}
\examples{
format_ISO8601(as.Date("02-01-2018", format="\%m-\%d-\%Y"))
format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz="EST"), usetz=TRUE)
format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz="EST"), precision="ymdhm")
}
\references{
\url{https://en.wikipedia.org/wiki/ISO_8601}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format_ISO8601.r
\name{format_ISO8601}
\alias{format_ISO8601}
\alias{format_ISO8601,Date-method}
\alias{format_ISO8601,POSIXt-method}
\alias{format_ISO8601,Interval-method}
\alias{format_ISO8601,Duration-method}
\alias{format_ISO8601,Period-method}
\title{Format in ISO8601 character format}
\usage{
format_ISO8601(x, usetz = FALSE, precision = NULL, ...)
}
\arguments{
\item{x}{An object to convert to ISO8601 character format.}

\item{usetz}{Include the time zone in the formatting.  If
\code{usetz} is \code{TRUE}, the time zone is included. If
\code{usetz} is \code{"Z"}, the time is converted to "UTC" and
the time zone is indicated with "Z" ISO8601 notation.}

\item{precision}{The amount of precision to represent with
substrings of "ymdhms", as "y"ear, "m"onth, "d"ay, "h"our,
"m"inute, and "s"econd. (e.g. "ymdhm" would show precision
through minutes.  When \code{NULL}, full precision for the object
is shown.}

\item{...}{Additional arguments to methods.}
}
\value{
A character vector of ISO8601-formatted text.
}
\description{
Format in ISO8601 character format
}
\examples{
format_ISO8601(as.Date("02-01-2018", format = "\%m-\%d-\%Y"))
format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz = "EST"), usetz = TRUE)
format_ISO8601(as.POSIXct("2018-02-01 03:04:05", tz = "EST"), precision = "ymdhm")
}
\references{
\url{https://en.wikipedia.org/wiki/ISO_8601}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-month.r
\name{days_in_month}
\alias{days_in_month}
\title{Get the number of days in the month of a date-time}
\usage{
days_in_month(x)
}
\arguments{
\item{x}{a date-time object}
}
\value{
An integer of the number of days in the month component of the date-time object.
}
\description{
Date-time must be a POSIXct, POSIXlt, Date, chron, yearmon, yearqtr,
zoo, zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercion.r
\name{as_date}
\alias{as_date}
\alias{as_date,ANY-method}
\alias{as_date,POSIXt-method}
\alias{as_date,numeric-method}
\alias{as_date,character-method}
\alias{as_datetime}
\alias{as_datetime,POSIXt-method}
\alias{as_datetime,numeric-method}
\alias{as_datetime,character-method}
\alias{as_datetime,ANY-method}
\title{Convert an object to a date or date-time}
\usage{
as_date(x, ...)

\S4method{as_date}{ANY}(x, ...)

\S4method{as_date}{POSIXt}(x, tz = NULL)

\S4method{as_date}{numeric}(x, origin = lubridate::origin)

\S4method{as_date}{character}(x, tz = NULL, format = NULL)

as_datetime(x, ...)

\S4method{as_datetime}{POSIXt}(x, tz = "UTC")

\S4method{as_datetime}{numeric}(x, origin = lubridate::origin, tz = "UTC")

\S4method{as_datetime}{character}(x, tz = "UTC", format = NULL)

\S4method{as_datetime}{ANY}(x, tz = "UTC")
}
\arguments{
\item{x}{a vector of \link{POSIXt}, numeric or character objects}

\item{...}{further arguments to be passed to specific methods (see above).}

\item{tz}{a time zone name (default: time zone of the POSIXt object \code{x}). See
\code{\link[=OlsonNames]{OlsonNames()}}.}

\item{origin}{a Date object, or something which can be coerced by
\code{as.Date(origin, ...)} to such an object (default: the Unix epoch of
"1970-01-01"). Note that in this instance, \code{x} is assumed to reflect the
number of days since \code{origin} at \code{"UTC"}.}

\item{format}{format argument for character methods. When supplied parsing is
performed by \code{\link[=strptime]{strptime()}}. For this reason consider using specialized
parsing functions in lubridate.}
}
\value{
a vector of \link{Date} objects corresponding to \code{x}.
}
\description{
Convert an object to a date or date-time
}
\section{Compare to base R}{


These are drop in replacements for \code{\link[=as.Date]{as.Date()}} and \code{\link[=as.POSIXct]{as.POSIXct()}}, with a few
tweaks to make them work more intuitively.

\itemize{
\item \code{as_date()} ignores the timezone attribute, resulting in
a more intuitive conversion (see examples)
\item Both functions provide a default origin argument for numeric
vectors.
\item Both functions will generate NAs for invalid date format. A warning message will provide a count of the elements that were not converted
\item \code{as_datetime()} defaults to using UTC.
}
}

\examples{
dt_utc <- ymd_hms("2010-08-03 00:50:50")
dt_europe <- ymd_hms("2010-08-03 00:50:50", tz="Europe/London")
c(as_date(dt_utc), as.Date(dt_utc))
c(as_date(dt_europe), as.Date(dt_europe))
## need not supply origin
as_date(10)
## Will replace invalid date format with NA
dt_wrong <- c("2009-09-29", "2012-11-29", "2015-29-12")
as_date(dt_wrong)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercion.r
\name{as_date}
\alias{as_date}
\alias{as_date,ANY-method}
\alias{as_date,POSIXt-method}
\alias{as_date,numeric-method}
\alias{as_date,character-method}
\alias{as_datetime}
\alias{as_datetime,ANY-method}
\alias{as_datetime,POSIXt-method}
\alias{as_datetime,numeric-method}
\alias{as_datetime,character-method}
\alias{as_datetime,Date-method}
\title{Convert an object to a date or date-time}
\usage{
as_date(x, ...)

\S4method{as_date}{ANY}(x, ...)

\S4method{as_date}{POSIXt}(x, tz = NULL)

\S4method{as_date}{numeric}(x, origin = lubridate::origin)

\S4method{as_date}{character}(x, tz = NULL, format = NULL)

as_datetime(x, ...)

\S4method{as_datetime}{ANY}(x, tz = lubridate::tz(x))

\S4method{as_datetime}{POSIXt}(x, tz = lubridate::tz(x))

\S4method{as_datetime}{numeric}(x, origin = lubridate::origin, tz = "UTC")

\S4method{as_datetime}{character}(x, tz = "UTC", format = NULL)

\S4method{as_datetime}{Date}(x, tz = "UTC")
}
\arguments{
\item{x}{a vector of \link{POSIXt}, numeric or character objects}

\item{...}{further arguments to be passed to specific methods (see above).}

\item{tz}{a time zone name (default: time zone of the POSIXt object \code{x}). See
\code{\link[=OlsonNames]{OlsonNames()}}.}

\item{origin}{a Date object, or something which can be coerced by \code{as.Date(origin, ...)} to such an object (default: the Unix epoch of "1970-01-01"). Note that in
this instance, \code{x} is assumed to reflect the number of days since \code{origin} at
\code{"UTC"}.}

\item{format}{format argument for character methods. When supplied parsing is
performed by \code{parse_date_time(x, orders = formats, exact = TRUE)}. Thus, multiple
formats are supported and are tried in turn.}
}
\value{
a vector of \link{Date} objects corresponding to \code{x}.
}
\description{
Convert an object to a date or date-time
}
\section{Compare to base R}{


These are drop in replacements for \code{\link[=as.Date]{as.Date()}} and \code{\link[=as.POSIXct]{as.POSIXct()}}, with a few
tweaks to make them work more intuitively.

\itemize{

\item Called on a \code{POSIXct} object, \code{as_date()} uses the tzone attribute of
the object to return the same date as indicated by the printed representation
of the object. This differs from as.Date, which ignores the attribute and
uses only the tz argument to \code{as.Date()} ("UTC" by default).

\item Both functions provide a default origin argument for numeric vectors.

\item Both functions will generate NAs for invalid date format. Valid formats are
those described by ISO8601 standard. A warning message will provide a count of the
elements that were not converted.

\item \code{as_datetime()} defaults to using UTC.

}
}

\examples{
dt_utc <- ymd_hms("2010-08-03 00:50:50")
dt_europe <- ymd_hms("2010-08-03 00:50:50", tz = "Europe/London")
c(as_date(dt_utc), as.Date(dt_utc))
c(as_date(dt_europe), as.Date(dt_europe))
## need not supply origin
as_date(10)
## Will replace invalid date format with NA
dt_wrong <- c("2009-09-29", "2012-11-29", "2015-29-12")
as_date(dt_wrong)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-day.r
\name{day}
\alias{day}
\alias{mday}
\alias{wday}
\alias{qday}
\alias{yday}
\alias{day<-}
\alias{mday<-}
\alias{qday<-}
\alias{wday<-}
\alias{yday<-}
\title{Get/set days component of a date-time}
\usage{
day(x)

mday(x)

wday(
  x,
  label = FALSE,
  abbr = TRUE,
  week_start = getOption("lubridate.week.start", 7),
  locale = Sys.getlocale("LC_TIME")
)

qday(x)

yday(x)

day(x) <- value

mday(x) <- value

qday(x) <- value

qday(x) <- value

wday(x, week_start = getOption("lubridate.week.start", 7)) <- value

yday(x) <- value
}
\arguments{
\item{x}{a POSIXct, POSIXlt, Date, chron, yearmon, yearqtr, zoo, zooreg, timeDate,
xts, its, ti, jul, timeSeries, or fts object.}

\item{label}{logical. Only available for wday. TRUE will display the day of the week
as an ordered factor of character strings, such as "Sunday." FALSE will display the
day of the week as a number.}

\item{abbr}{logical. Only available for wday. FALSE will display the day of the week
as an ordered factor of character strings, such as "Sunday." TRUE will display an
abbreviated version of the label, such as "Sun". abbr is disregarded if label =
FALSE.}

\item{week_start}{day on which week starts following ISO conventions: 1 means Monday
and 7 means Sunday (default). When \code{label = FALSE} and \code{week_start = 7}, the number
returned for Sunday is 1, for Monday is 2, etc. When \code{label = TRUE}, the returned
value is a factor with the first level being the week start (e.g. Sunday if
\code{week_start = 7}). You can set \code{lubridate.week.start} option to control this
parameter globally.}

\item{locale}{locale to use for day names. Default to current locale.}

\item{value}{(for \verb{wday<-}) a numeric or a string giving the name of the day in the
current locale or in English. Can be abbreviated. When a string, the value of
\code{week_start} is ignored.}
}
\value{
\code{wday()} returns the day of the week as a decimal number or an ordered factor
if label is \code{TRUE}.
}
\description{
Get/set days component of a date-time
}
\details{
\code{mday()} and \code{yday()} return the day of the month and day of the year
respectively. \code{day()} and \verb{day<-()} are aliases for \code{mday()} and \verb{mday<-()}.
}
\examples{
x <- as.Date("2009-09-02")
wday(x) # 4
wday(x, label = TRUE) # Wed

wday(x, week_start = 1) # 3
wday(x, week_start = 7) # 4

wday(x, label = TRUE, week_start = 7) # Wed (Sun is the first level)
wday(x, label = TRUE, week_start = 1) # Wed (Mon is the first level)

wday(ymd(080101))
wday(ymd(080101), label = TRUE, abbr = FALSE)
wday(ymd(080101), label = TRUE, abbr = TRUE)
wday(ymd(080101) + days(-2:4), label = TRUE, abbr = TRUE)

x <- as.Date("2009-09-02")
yday(x) # 245
mday(x) # 2
yday(x) <- 1 # "2009-01-01"
yday(x) <- 366 # "2010-01-01"
mday(x) > 3
}
\keyword{chron}
\keyword{manip}
\keyword{methods}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-day.r
\name{day}
\alias{day}
\alias{mday}
\alias{wday}
\alias{qday}
\alias{yday}
\alias{day<-}
\alias{mday<-}
\alias{qday<-}
\alias{wday<-}
\alias{yday<-}
\title{Get/set days component of a date-time}
\usage{
day(x)

mday(x)

wday(
  x,
  label = FALSE,
  abbr = TRUE,
  week_start = getOption("lubridate.week.start", 7),
  locale = Sys.getlocale("LC_TIME")
)

qday(x)

yday(x)

day(x) <- value

mday(x) <- value

qday(x) <- value

wday(x, week_start = getOption("lubridate.week.start", 7)) <- value

yday(x) <- value
}
\arguments{
\item{x}{a POSIXct, POSIXlt, Date, chron, yearmon, yearqtr, zoo, zooreg,
timeDate, xts, its, ti, jul, timeSeries, or fts object.}

\item{label}{logical. Only available for wday. TRUE will display the day of
the week as an ordered factor of character strings, such as "Sunday." FALSE
will display the day of the week as a number.}

\item{abbr}{logical. Only available for wday. FALSE will display the day of
the week as an ordered factor of character strings, such as "Sunday." TRUE
will display an abbreviated version of the label, such as "Sun". abbr is
disregarded if label = FALSE.}

\item{week_start}{day on which week starts following ISO conventions - 1
means Monday, 7 means Sunday (default). When \code{label = TRUE}, this will be
the first level of the returned factor. You can set \code{lubridate.week.start}
option to control this parameter globally.}

\item{locale}{locale to use for day names. Default to current locale.}

\item{value}{a numeric object}
}
\value{
\code{wday()} returns the day of the week as a decimal number or an
ordered factor if label is \code{TRUE}.
}
\description{
Get/set days component of a date-time
}
\details{
\code{mday()} and \code{yday()} return the day of the month and day of the
year respectively. \code{day()} and \verb{day<-()} are aliases for \code{mday()} and
\verb{mday<-()}.
}
\examples{
x <- as.Date("2009-09-02")
wday(x) #4

wday(ymd(080101))
wday(ymd(080101), label = TRUE, abbr = FALSE)
wday(ymd(080101), label = TRUE, abbr = TRUE)
wday(ymd(080101) + days(-2:4), label = TRUE, abbr = TRUE)

x <- as.Date("2009-09-02")
yday(x) #245
mday(x) #2
yday(x) <- 1  #"2009-01-01"
yday(x) <- 366 #"2010-01-01"
mday(x) > 3
}
\keyword{chron}
\keyword{manip}
\keyword{methods}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercion.r
\name{reclass_timespan}
\alias{reclass_timespan}
\alias{reclass_timespan,ANY,Duration-method}
\alias{reclass_timespan,ANY,Interval-method}
\alias{reclass_timespan,ANY,Period-method}
\alias{reclass_timespan,ANY,difftime-method}
\title{Convenience method to reclass timespans post-modification.}
\usage{
reclass_timespan(new, orig)
}
\description{
Convenience method to reclass timespans post-modification.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pretty.r
\name{pretty_dates}
\alias{pretty_dates}
\title{Computes attractive axis breaks for date-time data}
\usage{
pretty_dates(x, n, ...)
}
\arguments{
\item{x}{a vector of POSIXct, POSIXlt, Date, or chron date-time objects}

\item{n}{integer value of the desired number of breaks}

\item{...}{additional arguments to pass to function}
}
\value{
a vector of date-times that can be used as axis tick marks or bin breaks
}
\description{
pretty.dates indentifies which unit of time the sub-intervals should be
measured in to provide approximately n breaks. It then chooses a "pretty"
length for the sub-intervals and sets start and endpoints that 1) span the
entire range of the data, and 2) allow the breaks to occur on important
date-times (i.e. on the hour, on the first of the month, etc.)
}
\examples{
x <- seq.Date(as.Date("2009-08-02"), by = "year", length.out = 2)
pretty_dates(x, 12)
}
\keyword{chron}
\keyword{dplot}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/durations.r
\name{duration}
\alias{duration}
\alias{dseconds}
\alias{dminutes}
\alias{dhours}
\alias{ddays}
\alias{dweeks}
\alias{dmonths}
\alias{dyears}
\alias{dmilliseconds}
\alias{dmicroseconds}
\alias{dnanoseconds}
\alias{dpicoseconds}
\alias{is.duration}
\title{Create a duration object.}
\usage{
duration(num = NULL, units = "seconds", ...)

dseconds(x = 1)

dminutes(x = 1)

dhours(x = 1)

ddays(x = 1)

dweeks(x = 1)

dmonths(x = 1)

dyears(x = 1)

dmilliseconds(x = 1)

dmicroseconds(x = 1)

dnanoseconds(x = 1)

dpicoseconds(x = 1)

is.duration(x)
}
\arguments{
\item{num}{the number or a character vector of time units. In string
representation all unambiguous name units and abbreviations and ISO 8601
formats are supported; 'm' stands for month and 'M' for minutes unless ISO
8601 "P" modifier is present (see examples). Fractional units are
supported.}

\item{units}{a character string that specifies the type of units that \code{num}
refers to. When \code{num} is character, this argument is ignored.}

\item{...}{a list of time units to be included in the duration and their
amounts. Seconds, minutes, hours, days, weeks, months and years are
supported. Durations of months and years assume that year consists of
365.25 days.}

\item{x}{numeric value of the number of units to be contained in the
duration.}
}
\value{
a duration object
}
\description{
\code{duration()} creates a duration object with the specified values. Entries
for different units are cumulative. durations display as the number of
seconds in a time span. When this number is large, durations also display an
estimate in larger units,; however, the underlying object is always recorded
as a fixed number of seconds. For display and creation purposes, units are
converted to seconds using their most common lengths in seconds. Minutes = 60
seconds, hours = 3600 seconds, days = 86400 seconds, weeks = 604800. Units
larger than weeks are not used due to their variability.
}
\details{
Durations record the exact number of seconds in a time span. They measure the
exact passage of time but do not always align with measurements
made in larger units of time such as hours, months and years.
This is because the length of larger time units can be affected
by conventions such as leap years
and Daylight Savings Time. Base R provides a second class for measuring
durations, the difftime class.

Duration objects can be easily created with the helper functions \code{\link[=dweeks]{dweeks()}},
\code{\link[=ddays]{ddays()}}, \code{\link[=dminutes]{dminutes()}}, \code{\link[=dseconds]{dseconds()}}. These objects can be added to and
subtracted to date- times to create a user interface similar to object
oriented programming.
}
\examples{

### Separate period and units vectors

duration(90, "seconds")
duration(1.5, "minutes")
duration(-1, "days")

### Units as arguments

duration(day = -1)
duration(second = 90)
duration(minute = 1.5)
duration(mins = 1.5)
duration(second = 3, minute = 1.5, hour = 2, day = 6, week = 1)
duration(hour = 1, minute = -60)

### Parsing

duration("2M 1sec")
duration("2hours 2minutes 1second")
duration("2d 2H 2M 2S")
duration("2days 2hours 2mins 2secs")
# Missing numerals default to 1. Repeated units are added up.
duration("day day")

### ISO 8601 parsing

duration("P3Y6M4DT12H30M5S")
duration("P23DT23H") # M stands for months
duration("10DT10M") # M stands for minutes
duration("P23DT60H 20min 100 sec") # mixing ISO and lubridate style parsing

# Comparison with characters (from v1.6.0)

duration("day 2 sec") > "day 1sec"


## ELEMENTARY CONSTRUCTORS:

dseconds(1)
dminutes(3.5)

x <- ymd_hms("2009-08-03", tz="America/Chicago")
x + ddays(1) + dhours(6) + dminutes(30)
x + ddays(100) - dhours(8)

class(as.Date("2009-08-09") + ddays(1)) # retains Date class
as.Date("2009-08-09") + dhours(12)
class(as.Date("2009-08-09") + dhours(12))
# converts to POSIXt class to accomodate time units

dweeks(1) - ddays(7)
c(1:3) * dhours(1)

# compare DST handling to durations
boundary <- ymd_hms("2009-03-08 01:59:59", tz="America/Chicago")
boundary + days(1) # period
boundary + ddays(1) # duration
is.duration(as.Date("2009-08-03")) # FALSE
is.duration(duration(days = 12.4)) # TRUE
}
\seealso{
\code{\link[=as.duration]{as.duration()}} \linkS4class{Duration}
}
\keyword{chron}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/durations.r
\name{duration}
\alias{duration}
\alias{dseconds}
\alias{dminutes}
\alias{dhours}
\alias{ddays}
\alias{dweeks}
\alias{dmonths}
\alias{dyears}
\alias{dmilliseconds}
\alias{dmicroseconds}
\alias{dnanoseconds}
\alias{dpicoseconds}
\alias{is.duration}
\title{Create a duration object.}
\usage{
duration(num = NULL, units = "seconds", ...)

dseconds(x = 1)

dminutes(x = 1)

dhours(x = 1)

ddays(x = 1)

dweeks(x = 1)

dmonths(x = 1)

dyears(x = 1)

dmilliseconds(x = 1)

dmicroseconds(x = 1)

dnanoseconds(x = 1)

dpicoseconds(x = 1)

is.duration(x)
}
\arguments{
\item{num}{the number or a character vector of time units. In string
representation all unambiguous name units and abbreviations and ISO 8601
formats are supported; 'm' stands for month and 'M' for minutes unless ISO
8601 "P" modifier is present (see examples). Fractional units are
supported.}

\item{units}{a character string that specifies the type of units that \code{num}
refers to. When \code{num} is character, this argument is ignored.}

\item{...}{a list of time units to be included in the duration and their
amounts. Seconds, minutes, hours, days, weeks, months and years are
supported. Durations of months and years assume that year consists of
365.25 days.}

\item{x}{numeric value of the number of units to be contained in the
duration.}
}
\value{
a duration object
}
\description{
\code{duration()} creates a duration object with the specified values. Entries
for different units are cumulative. durations display as the number of
seconds in a time span. When this number is large, durations also display an
estimate in larger units, however, the underlying object is always recorded
as a fixed number of seconds. For display and creation purposes, units are
converted to seconds using their most common lengths in seconds. Minutes = 60
seconds, hours = 3600 seconds, days = 86400 seconds, weeks = 604800. Units
larger than weeks are not used due to their variability.
}
\details{
Durations record the exact number of seconds in a time span. They measure the
exact passage of time but do not always align with measurements
made in larger units of time such as hours, months and years.
This is because the length of larger time units can be affected
by conventions such as leap years
and Daylight Savings Time. Base R provides a second class for measuring
durations, the difftime class.

Duration objects can be easily created with the helper functions \code{\link[=dweeks]{dweeks()}},
\code{\link[=ddays]{ddays()}}, \code{\link[=dminutes]{dminutes()}}, \code{\link[=dseconds]{dseconds()}}. These objects can be added to and
subtracted to date- times to create a user interface similar to object
oriented programming.
}
\examples{

### Separate period and units vectors

duration(90, "seconds")
duration(1.5, "minutes")
duration(-1, "days")

### Units as arguments

duration(day = -1)
duration(second = 90)
duration(minute = 1.5)
duration(mins = 1.5)
duration(second = 3, minute = 1.5, hour = 2, day = 6, week = 1)
duration(hour = 1, minute = -60)

### Parsing

duration("2M 1sec")
duration("2hours 2minutes 1second")
duration("2d 2H 2M 2S")
duration("2days 2hours 2mins 2secs")
# Missing numerals default to 1. Repeated units are added up.
duration("day day")

### ISO 8601 parsing

duration("P3Y6M4DT12H30M5S")
duration("P23DT23H") # M stands for months
duration("10DT10M") # M stands for minutes
duration("P23DT60H 20min 100 sec") # mixing ISO and lubridate style parsing

# Comparison with characters (from v1.6.0)

duration("day 2 sec") > "day 1sec"


## ELEMENTARY CONSTRUCTORS:

dseconds(1)
dminutes(3.5)

x <- ymd("2009-08-03", tz = "America/Chicago")
x + ddays(1) + dhours(6) + dminutes(30)
x + ddays(100) - dhours(8)

class(as.Date("2009-08-09") + ddays(1)) # retains Date class
as.Date("2009-08-09") + dhours(12)
class(as.Date("2009-08-09") + dhours(12))
# converts to POSIXt class to accomodate time units

dweeks(1) - ddays(7)
c(1:3) * dhours(1)

# compare DST handling to durations
boundary <- ymd_hms("2009-03-08 01:59:59", tz = "America/Chicago")
boundary + days(1) # period
boundary + ddays(1) # duration
is.duration(as.Date("2009-08-03")) # FALSE
is.duration(duration(days = 12.4)) # TRUE
}
\seealso{
\code{\link[=as.duration]{as.duration()}} \linkS4class{Duration}
}
\keyword{chron}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/instants.r
\name{now}
\alias{now}
\alias{today}
\title{The current day and time}
\usage{
now(tzone = "")

today(tzone = "")
}
\arguments{
\item{tzone}{a character vector specifying which time zone you would like the
current time in. tzone defaults to your computer's system timezone. You can
retrieve the current time in the Universal Coordinated Time (UTC) with
now("UTC").}
}
\value{
\code{now} - the current datetime as a \code{POSIXct} object
}
\description{
The current day and time
}
\examples{
now()
now("GMT")
now("")
now() == now() # would be TRUE if computer processed both at the same instant
now() < now() # TRUE
now() > now() # FALSE
today()
today("GMT")
today() == today("GMT") # not always true
today() < as.Date("2999-01-01") # TRUE  (so far)
}
\keyword{chron}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decimal-dates.r
\name{decimal_date}
\alias{decimal_date}
\title{Converts a date to a decimal of its year}
\usage{
decimal_date(date)
}
\arguments{
\item{date}{a POSIXt or Date object}
}
\value{
a numeric object where the date is expressed as a fraction of its year
}
\description{
Converts a date to a decimal of its year
}
\examples{
date <- ymd("2009-02-10")
decimal_date(date) # 2009.11
}
\keyword{chron}
\keyword{manip}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decimal-dates.r
\name{decimal_date}
\alias{decimal_date}
\title{Converts a date to a decimal of its year}
\usage{
decimal_date(date)
}
\arguments{
\item{date}{a POSIXt or Date object}
}
\value{
a numeric object where the date is expressed as a fraction of its year
}
\description{
Converts a date to a decimal of its year
}
\examples{
date <- ymd("2009-02-10")
decimal_date(date)  # 2009.11
}
\keyword{chron}
\keyword{manip}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dates.r
\docType{data}
\name{is.Date}
\alias{is.Date}
\alias{Date}
\alias{NA_Date_}
\title{Various date utilities}
\format{
An object of class \code{Date} of length 1.
}
\usage{
is.Date(x)

Date(length = 0L)

NA_Date_
}
\arguments{
\item{x}{an R object}

\item{length}{A non-negative number specifying the desired length. Supplying
an argument of length other than one is an error.}
}
\description{
\code{\link[=Date]{Date()}} mirrors primitive contructors in base R (\code{\link[=double]{double()}}, \code{\link[=character]{character()}}
etc.)
}
\examples{
is.Date(as.Date("2009-08-03")) # TRUE
is.Date(difftime(now() + 5, now())) # FALSE
}
\seealso{
\code{\link[=is.instant]{is.instant()}}, \code{\link[=is.timespan]{is.timespan()}}, \code{\link[=is.POSIXt]{is.POSIXt()}}, \code{\link[=POSIXct]{POSIXct()}}
}
\keyword{chron}
\keyword{datasets}
\keyword{logic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update.r
\name{DateTimeUpdate}
\alias{DateTimeUpdate}
\alias{update.POSIXt}
\title{Changes the components of a date object}
\usage{
\method{update}{POSIXt}(
  object,
  ...,
  roll_dst = c("NA", "post"),
  week_start = getOption("lubridate.week.start", 7),
  roll = NULL,
  simple = NULL
)
}
\arguments{
\item{object}{a date-time object}

\item{...}{named arguments: years, months, ydays, wdays, mdays, days, hours, minutes,
seconds, tzs (time zone component)}

\item{roll_dst}{is a string vector of length one or two. When two values are
supplied they specify how to roll date-times when they fall into "skipped" and
"repeated" DST transitions respectively. A single value is replicated to the
length of two. Possible values are:

\if{html}{\out{<div class="sourceCode">}}\preformatted{* `pre` - Use the time before the transition boundary.
* `boundary` - Use the time exactly at the boundary transition.
* `post` - Use the time after the boundary transition.
* `xfirst` - crossed-first: First time which occurred when crossing the
   boundary. For addition with positive units pre interval is crossed first and
   post interval last. With negative units post interval is crossed first, pre -
   last. For subtraction the logic is reversed.
* `xlast` - crossed-last.
* `NA` - Produce NAs when the resulting time falls inside the problematic interval.
}\if{html}{\out{</div>}}

For example `roll_dst = c("NA", "pre") indicates that for skiped intervals
return NA and for repeated times return the earlier time.

When multiple units are supplied the meaning of "negative period" is determined by
the largest unit. For example \code{time_add(t, days = -1, hours = 2, roll_dst = "xfirst")} would operate as if with negative period, thus crossing the boundary
from the "post" to "pre" side and "xfirst" and hence resolving to "post"
time. As this might result in confusing behavior. See examples.

"xfirst" and "xlast" make sense for addition and subtraction only. An error is
raised if an attempt is made to use them with other functions.}

\item{week_start}{week start day (Default is 7, Sunday. Set \code{lubridate.week.start} to
override). Full or abbreviated names of the days of the week can be in English or
as provided by the current locale.}

\item{simple, roll}{deprecated}
}
\value{
a date object with the requested elements updated. The object will retain its
original class unless an element is updated which the original class does not
support. In this case, the date returned will be a POSIXlt date object.
}
\description{
\code{update.Date()} and \code{update.POSIXt()} return a date with the specified
elements updated.  Elements not specified will be left unaltered. update.Date
and update.POSIXt do not add the specified values to the existing date, they
substitute them for the appropriate parts of the existing date.
}
\examples{
date <- ymd("2009-02-10")
update(date, year = 2010, month = 1, mday = 1)

update(date, year = 2010, month = 13, mday = 1)

update(date, minute = 10, second = 3)
}
\keyword{chron}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update.r
\name{DateTimeUpdate}
\alias{DateTimeUpdate}
\alias{update.POSIXt}
\title{Changes the components of a date object}
\usage{
\method{update}{POSIXt}(
  object,
  ...,
  roll = FALSE,
  week_start = getOption("lubridate.week.start", 7),
  simple = NULL
)
}
\arguments{
\item{object}{a date-time object}

\item{...}{named arguments: years, months, ydays, wdays, mdays, days, hours,
minutes, seconds, tzs (time zone component)}

\item{roll}{logical. If \code{TRUE}, and the resulting date-time lands on a
non-existent civil time instant (DST, 29th February, etc.) roll the date
till next valid point. When \code{FALSE}, the default, produce NA for non
existing date-times.}

\item{week_start}{week starting day (Default is 7, Sunday). Set
\code{lubridate.week.start} option to control this.}

\item{simple}{logical. Deprecated. Same as \code{roll}.}
}
\value{
a date object with the requested elements updated. The object will
retain its original class unless an element is updated which the original
class does not support. In this case, the date returned will be a POSIXlt
date object.
}
\description{
\code{update.Date()} and \code{update.POSIXt()} return a date with the specified
elements updated.  Elements not specified will be left unaltered. update.Date
and update.POSIXt do not add the specified values to the existing date, they
substitute them for the appropriate parts of the existing date.
}
\examples{
date <- ymd("2009-02-10")
update(date, year = 2010, month = 1, mday = 1)

update(date, year =2010, month = 13, mday = 1)

update(date, minute = 10, second = 3)
}
\keyword{chron}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/intervals.r
\name{\%within\%}
\alias{\%within\%}
\alias{\%within\%,Interval,Interval-method}
\alias{\%within\%,ANY,Interval-method}
\alias{\%within\%,Date,list-method}
\alias{\%within\%,POSIXt,list-method}
\title{Does a date (or interval) fall within an interval?}
\usage{
a \%within\% b
}
\arguments{
\item{a}{An interval or date-time object.}

\item{b}{Either an interval vector, or a list of intervals.

If \code{b} is an interval it is recycled to the same length as \code{a}.
If \code{b} is a list of intervals, \code{a} is checked if it falls within \emph{any}
of the intervals, i.e. \code{a \%within\% list(int1, int2)} is equivalent to
\code{a \%within\% int1 | a \%within\% int2}.}
}
\value{
A logical vector.
}
\description{
Check whether \code{a} lies within the interval \code{b}, inclusive of the endpoints.
}
\examples{
int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int2 <- interval(ymd("2001-06-01"), ymd("2002-01-01"))

ymd("2001-05-03") \%within\% int # TRUE
int2 \%within\% int # TRUE
ymd("1999-01-01") \%within\% int # FALSE

## recycling
dates <- ymd(c("2014-12-20", "2014-12-30", "2015-01-01", "2015-01-03"))
blackouts<- c(interval(ymd("2014-12-30"), ymd("2014-12-31")),
              interval(ymd("2014-12-30"), ymd("2015-01-03")))
dates \%within\% blackouts

## within ANY of the intervals of a list
dates <- ymd(c("2014-12-20", "2014-12-30", "2015-01-01", "2015-01-03"))
blackouts<- list(interval(ymd("2014-12-30"), ymd("2014-12-31")),
                 interval(ymd("2014-12-30"), ymd("2015-01-03")))
dates \%within\% blackouts
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/intervals.r
\name{\%within\%}
\alias{\%within\%}
\alias{\%within\%,Interval,Interval-method}
\alias{\%within\%,ANY,Interval-method}
\alias{\%within\%,Date,list-method}
\alias{\%within\%,POSIXt,list-method}
\alias{\%within\%,Interval,list-method}
\title{Does a date (or interval) fall within an interval?}
\usage{
a \%within\% b
}
\arguments{
\item{a}{An interval or date-time object.}

\item{b}{Either an interval vector, or a list of intervals.

If \code{b} is an interval (or interval vector) it is recycled to the same length as \code{a}.
If \code{b} is a list of intervals, \code{a} is checked if it falls within \emph{any}
of the intervals, i.e. \code{a \%within\% list(int1, int2)} is equivalent to
\code{a \%within\% int1 | a \%within\% int2}.}
}
\value{
A logical vector.
}
\description{
Check whether \code{a} lies within the interval \code{b}, inclusive of the endpoints.
}
\examples{
int <- interval(ymd("2001-01-01"), ymd("2002-01-01"))
int2 <- interval(ymd("2001-06-01"), ymd("2002-01-01"))

ymd("2001-05-03") \%within\% int # TRUE
int2 \%within\% int # TRUE
ymd("1999-01-01") \%within\% int # FALSE

## recycling (carefully note the difference between using a vector of
## intervals and list of intervals for the second argument)
dates <- ymd(c("2014-12-20", "2014-12-30", "2015-01-01", "2015-01-03"))
blackout_vector <- c(
  interval(ymd("2014-12-30"), ymd("2014-12-31")),
  interval(ymd("2014-12-30"), ymd("2015-01-03"))
)
dates \%within\% blackout_vector

## within ANY of the intervals of a list
dates <- ymd(c("2014-12-20", "2014-12-30", "2015-01-01", "2015-01-03"))
lst <- list(
  interval(ymd("2014-12-30"), ymd("2014-12-31")),
  interval(ymd("2014-12-30"), ymd("2015-01-03"))
)
dates \%within\% lst

## interval within a list of intervals
int <- interval(
  ymd("2014-12-20", "2014-12-30"),
  ymd("2015-01-01", "2015-01-03")
)
int \%within\% lst
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/periods.r
\name{period_to_seconds}
\alias{period_to_seconds}
\alias{seconds_to_period}
\title{Contrive a period to/from a given number of seconds}
\usage{
period_to_seconds(x)

seconds_to_period(x)
}
\arguments{
\item{x}{A numeric object. The number of seconds to coerce into a period.}
}
\value{
A number (period) that roughly equates to the period (seconds) given.
}
\description{
\code{period_to_seconds()} approximately converts a period to seconds assuming
there are 365.25 days in a calendar year and 365.25/12 days in a month.

\code{seconds_to_period()} create a period that has the maximum number of
non-zero elements (days, hours, minutes, seconds). This computation is exact
because it doesn't involve years or months.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse.r
\name{ymd_hms}
\alias{ymd_hms}
\alias{ymd_hm}
\alias{ymd_h}
\alias{dmy_hms}
\alias{dmy_hm}
\alias{dmy_h}
\alias{mdy_hms}
\alias{mdy_hm}
\alias{mdy_h}
\alias{ydm_hms}
\alias{ydm_hm}
\alias{ydm_h}
\title{Parse date-times with \strong{y}ear, \strong{m}onth, and \strong{d}ay, \strong{h}our,
\strong{m}inute, and \strong{s}econd components.}
\usage{
ymd_hms(
  ...,
  quiet = FALSE,
  tz = "UTC",
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

ymd_hm(
  ...,
  quiet = FALSE,
  tz = "UTC",
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

ymd_h(
  ...,
  quiet = FALSE,
  tz = "UTC",
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

dmy_hms(
  ...,
  quiet = FALSE,
  tz = "UTC",
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

dmy_hm(
  ...,
  quiet = FALSE,
  tz = "UTC",
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

dmy_h(
  ...,
  quiet = FALSE,
  tz = "UTC",
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

mdy_hms(
  ...,
  quiet = FALSE,
  tz = "UTC",
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

mdy_hm(
  ...,
  quiet = FALSE,
  tz = "UTC",
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

mdy_h(
  ...,
  quiet = FALSE,
  tz = "UTC",
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

ydm_hms(
  ...,
  quiet = FALSE,
  tz = "UTC",
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

ydm_hm(
  ...,
  quiet = FALSE,
  tz = "UTC",
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

ydm_h(
  ...,
  quiet = FALSE,
  tz = "UTC",
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)
}
\arguments{
\item{...}{a character vector of dates in year, month, day, hour, minute,
second format}

\item{quiet}{logical. If \code{TRUE}, function evaluates without displaying customary messages.}

\item{tz}{a character string that specifies which time zone to parse the date with. The string
must be a time zone that is recognized by the user's OS.}

\item{locale}{locale to be used, see \link{locales}. On Linux systems you
can use \code{system("locale -a")} to list all the installed locales.}

\item{truncated}{integer, indicating how many formats can be missing. See details.}
}
\value{
a vector of \link{POSIXct} date-time objects
}
\description{
Transform dates stored as character or numeric vectors to POSIXct
objects. The \code{ymd_hms()} family of functions recognizes all non-alphanumeric
separators (with the exception of "." if \code{frac = TRUE}) and correctly
handles heterogeneous date-time representations. For more flexibility in
treatment of heterogeneous formats, see low level parser
\code{\link[=parse_date_time]{parse_date_time()}}.
}
\details{
The \code{ymd_hms()} functions automatically assign the Universal Coordinated Time
Zone (UTC) to the parsed date. This time zone can be changed with
\code{\link[=force_tz]{force_tz()}}.

The most common type of irregularity in date-time data is the truncation
due to rounding or unavailability of the time stamp. If the \code{truncated}
parameter is non-zero, the \code{ymd_hms()} functions also check for truncated
formats. For example, \code{ymd_hms()} with \code{truncated = 3} will also parse
incomplete dates like \verb{2012-06-01 12:23}, \verb{2012-06-01 12} and
\code{2012-06-01}. NOTE: The \code{ymd()} family of functions is based on
\code{\link[base:strptime]{base::strptime()}} which currently fails to parse \verb{\%y-\%m} formats.

In case of heterogeneous date formats the \code{ymd_hms()} family guesses formats
based on a subset of the input vector. If the input vector contains many
missing values or non-date strings, the subset might not contain meaningful
dates and the date-time format won't be guessed resulting in
\verb{All formats failed to parse} error. In such cases please see
\code{\link[=parse_date_time]{parse_date_time()}} for a more flexible parsing interface.

As of version 1.3.0, \pkg{lubridate}'s parse functions no longer return a
message that displays which format they used to parse their input. You can
change this by setting the \code{lubridate.verbose} option to \code{TRUE} with
\code{options(lubridate.verbose = TRUE)}.
}
\examples{

x <- c("2010-04-14-04-35-59", "2010-04-01-12-00-00")
ymd_hms(x)
x <- c("2011-12-31 12:59:59", "2010-01-01 12:00:00")
ymd_hms(x)


## ** heterogeneous formats **
x <- c(20100101120101, "2009-01-02 12-01-02", "2009.01.03 12:01:03",
       "2009-1-4 12-1-4",
       "2009-1, 5 12:1, 5",
       "200901-08 1201-08",
       "2009 arbitrary 1 non-decimal 6 chars 12 in between 1 !!! 6",
       "OR collapsed formats: 20090107 120107 (as long as prefixed with zeros)",
       "Automatic wday, Thu, detection, 10-01-10 10:01:10 and p format: AM",
       "Created on 10-01-11 at 10:01:11 PM")
ymd_hms(x)

## ** fractional seconds **
op <- options(digits.secs=3)
dmy_hms("20/2/06 11:16:16.683")
options(op)

## ** different formats for ISO8601 timezone offset **
ymd_hms(c("2013-01-24 19:39:07.880-0600",
"2013-01-24 19:39:07.880", "2013-01-24 19:39:07.880-06:00",
"2013-01-24 19:39:07.880-06", "2013-01-24 19:39:07.880Z"))

## ** internationalization **
\dontrun{
x_RO <- "Ma 2012 august 14 11:28:30 "
  ymd_hms(x_RO, locale = "ro_RO.utf8")
}

## ** truncated time-dates **
x <- c("2011-12-31 12:59:59", "2010-01-01 12:11", "2010-01-01 12", "2010-01-01")
ymd_hms(x, truncated = 3)
x <- c("2011-12-31 12:59", "2010-01-01 12", "2010-01-01")
ymd_hm(x, truncated = 2)
## ** What lubridate might not handle **
## Extremely weird cases when one of the separators is "" and some of the
## formats are not in double digits might not be parsed correctly:
\dontrun{
ymd_hm("20100201 07-01", "20100201 07-1", "20100201 7-01")}

}
\seealso{
\itemize{
\item \code{\link[=ymd]{ymd()}}, \code{\link[=hms]{hms()}}
\item \code{\link[=parse_date_time]{parse_date_time()}} for the underlying mechanism
}
}
\keyword{POSIXt}
\keyword{parse}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timespans.r
\name{timespan}
\alias{timespan}
\alias{timespans}
\title{Description of time span classes in lubridate}
\description{
A time span can be measured in three ways: as a duration, an interval, or a
period.
\itemize{
\item \link{duration}s record the exact number of seconds in a time span.
They measure the exact passage of time but do not always align with
human measurements like hours, months and years.
\item \link{period}s record the change in the clock time between two date-times.
They are measured in human units: years, months, days, hours, minutes,
and seconds.
\item \link{intervals} are time spans bound by two real date-times. Intervals can be
accurately converted to periods and durations.
}
}
\examples{
duration(3690, "seconds")
period(3690, "seconds")
period(second = 30, minute = 1, hour = 1)
interval(ymd_hms("2009-08-09 13:01:30"), ymd_hms("2009-08-09 12:00:00"))

date <- ymd_hms("2009-03-08 01:59:59") # DST boundary
date + days(1)
date + ddays(1)

date2 <- ymd_hms("2000-02-29 12:00:00")
date2 + years(1)
# self corrects to next real day

date3 <- ymd_hms("2009-01-31 01:00:00")
date3 + c(0:11) * months(1)

span <- date2 \%--\% date # creates interval

date <- ymd_hms("2009-01-01 00:00:00")
date + years(1)
date - days(3) + hours(6)
date + 3 * seconds(10)

months(6) + days(1)
}
\keyword{chron}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timespans.r
\name{timespan}
\alias{timespan}
\alias{timespans}
\title{Description of time span classes in lubridate}
\description{
A time span can be measured in three ways: as a duration, an interval, or a
period.
\itemize{
\item \link{duration}s record the exact number of seconds in a time span.
They measure the exact passage of time but do not always align with
human measurements like hours, months and years.
\item \link{period}s record the change in the clock time between two date-times.
They are measured in human units: years, months, days, hours, minutes,
and seconds.
\item \link{intervals} are time spans bound by two real date-times. Intervals can be
accurately converted to periods and durations.
}
}
\examples{
duration(3690, "seconds")
period(3690, "seconds")
period(second = 30, minute = 1, hour = 1)
interval(ymd_hms("2009-08-09 13:01:30"), ymd_hms("2009-08-09 12:00:00"))

date <- ymd_hms("2009-03-08 01:59:59") # DST boundary
date + days(1)
date + ddays(1)

date2 <- ymd_hms("2000-02-29 12:00:00")
date2 + years(1)
# self corrects to next real day

date3 <- ymd_hms("2009-01-31 01:00:00")
date3 + c(0:11) * months(1)

span <- date2 \%--\% date  #creates interval

date <- ymd_hms("2009-01-01 00:00:00")
date + years(1)
date - days(3) + hours(6)
date + 3 * seconds(10)

months(6) + days(1)
}
\keyword{chron}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/guess.r
\name{guess_formats}
\alias{guess_formats}
\title{Guess possible date-times formats from a character vector}
\usage{
guess_formats(
  x,
  orders,
  locale = Sys.getlocale("LC_TIME"),
  preproc_wday = TRUE,
  print_matches = FALSE
)
}
\arguments{
\item{x}{input vector of date-times.}

\item{orders}{format orders to look for. See examples.}

\item{locale}{locale to use. Defaults to the current locale.}

\item{preproc_wday}{whether to preprocess weekday names. Internal
optimization used by \code{ymd_hms()} family of functions. If \code{TRUE}, weekdays are
substituted with \%a or \%A accordingly, so that there is no need to supply
this format explicitly.}

\item{print_matches}{for development purposes mainly. If \code{TRUE}, prints a matrix
of matched templates.}
}
\value{
a vector of matched formats
}
\description{
Guess possible date-times formats from a character vector.
}
\examples{

x <- c('February 20th 1973',
       "february  14, 2004",
       "Sunday, May 1, 2000",
       "Sunday, May 1, 2000",
       "february  14, 04",
       'Feb 20th 73',
       "January 5 1999 at 7pm",
       "jan 3 2010",
       "Jan 1, 1999",
       "jan 3   10",
       "01 3 2010",
       "1 3 10",
       '1 13 89',
       "5/27/1979",
       "12/31/99",
       "DOB:12/11/00",
       "-----------",
       'Thu, 1 July 2004 22:30:00',
       'Thu, 1st of July 2004 at 22:30:00',
       'Thu, 1July 2004 at 22:30:00',
       'Thu, 1July2004 22:30:00',
       'Thu, 1July04 22:30:00',
       "21 Aug 2011, 11:15:34 pm",
       "-----------",
       "1979-05-27 05:00:59",
       "1979-05-27",
       "-----------",
       "3 jan 2000",
       "17 april 85",
       "27/5/1979",
       '20 01 89',
       '00/13/10',
       "-------",
       "14 12 00",
       "03:23:22 pm")

guess_formats(x, "BdY")
guess_formats(x, "Bdy")
## m also matches b and B; y also matches Y
guess_formats(x, "mdy", print_matches = TRUE)

## T also matches IMSp order
guess_formats(x, "T", print_matches = TRUE)

## b and B are equivalent and match, both, abreviated and full names
guess_formats(x, c("mdY", "BdY", "Bdy", "bdY", "bdy"), print_matches = TRUE)
guess_formats(x, c("dmy", "dbY", "dBy", "dBY"), print_matches = TRUE)


guess_formats(x, c("dBY HMS", "dbY HMS", "dmyHMS", "BdY H"), print_matches = TRUE)

guess_formats(x, c("ymd HMS"), print_matches = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercion.r
\name{as.interval}
\alias{as.interval}
\alias{as.interval,numeric-method}
\alias{as.interval,difftime-method}
\alias{as.interval,Interval-method}
\alias{as.interval,Duration-method}
\alias{as.interval,Period-method}
\alias{as.interval,POSIXt-method}
\alias{as.interval,logical-method}
\title{Change an object to an \code{interval}}
\usage{
as.interval(x, start, ...)
}
\arguments{
\item{x}{a duration, difftime, period, or numeric object that describes the length of the interval}

\item{start}{a POSIXt or Date object that describes when the interval begins}

\item{...}{additional arguments to pass to as.interval}
}
\value{
an interval object
}
\description{
as.interval changes difftime, Duration, Period and numeric class objects to
intervals that begin at the specified date-time. Numeric objects are first
coerced to timespans equal to the numeric value in seconds.
}
\details{
as.interval can be used to create accurate transformations between Period
objects, which measure time spans in variable length units, and Duration objects,
which measure timespans as an exact number of seconds. A start date-
time must be supplied to make the conversion. Lubridate uses
this start date to look up how many seconds each variable
length unit (e.g. month, year) lasted for during the time span
described. See
\code{\link[=as.duration]{as.duration()}}, \code{\link[=as.period]{as.period()}}.
}
\examples{
diff <- make_difftime(days = 31) # difftime
as.interval(diff, ymd("2009-01-01"))
as.interval(diff, ymd("2009-02-01"))

dur <- duration(days = 31) # duration
as.interval(dur, ymd("2009-01-01"))
as.interval(dur, ymd("2009-02-01"))

per <- period(months = 1) # period
as.interval(per, ymd("2009-01-01"))
as.interval(per, ymd("2009-02-01"))

as.interval(3600, ymd("2009-01-01")) # numeric
}
\seealso{
\code{\link[=interval]{interval()}}
}
\keyword{chron}
\keyword{classes}
\keyword{manip}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercion.r
\name{as.interval}
\alias{as.interval}
\alias{as.interval,numeric-method}
\alias{as.interval,difftime-method}
\alias{as.interval,Interval-method}
\alias{as.interval,Duration-method}
\alias{as.interval,Period-method}
\alias{as.interval,POSIXt-method}
\alias{as.interval,logical-method}
\title{Change an object to an \code{interval}}
\usage{
as.interval(x, start, ...)
}
\arguments{
\item{x}{a duration, difftime, period, or numeric object that describes the length of the interval}

\item{start}{a POSIXt or Date object that describes when the interval begins}

\item{...}{additional arguments to pass to as.interval}
}
\value{
an interval object
}
\description{
as.interval changes difftime, Duration, Period and numeric class objects to
intervals that begin at the specified date-time. Numeric objects are first
coerced to timespans equal to the numeric value in seconds.
}
\details{
as.interval can be used to create accurate transformations between Period
objects, which measure time spans in variable length units, and Duration objects,
which measure timespans as an exact number of seconds. A start date-
time must be supplied to make the conversion. Lubridate uses
this start date to look up how many seconds each variable
length unit (e.g. month, year) lasted for during the time span
described. See
\code{\link[=as.duration]{as.duration()}}, \code{\link[=as.period]{as.period()}}.
}
\examples{
diff <- make_difftime(days = 31) #difftime
as.interval(diff, ymd("2009-01-01"))
as.interval(diff, ymd("2009-02-01"))

dur <- duration(days = 31) #duration
as.interval(dur, ymd("2009-01-01"))
as.interval(dur, ymd("2009-02-01"))

per <- period(months = 1) #period
as.interval(per, ymd("2009-01-01"))
as.interval(per, ymd("2009-02-01"))

as.interval(3600, ymd("2009-01-01")) #numeric
}
\seealso{
\code{\link[=interval]{interval()}}
}
\keyword{chron}
\keyword{classes}
\keyword{manip}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stamp.r
\name{stamp}
\alias{stamp}
\alias{stamp_date}
\alias{stamp_time}
\title{Format dates and times based on human-friendly templates}
\usage{
stamp(
  x,
  orders = lubridate_formats,
  locale = Sys.getlocale("LC_TIME"),
  quiet = FALSE
)

stamp_date(x, locale = Sys.getlocale("LC_TIME"))

stamp_time(x, locale = Sys.getlocale("LC_TIME"))
}
\arguments{
\item{x}{a character vector of templates.}

\item{orders}{orders are sequences of formatting characters which might be
used for disambiguation. For example "ymd hms", "aym" etc. See
\code{\link[=guess_formats]{guess_formats()}} for a list of available formats.}

\item{locale}{locale in which \code{x} is encoded. On Linux-like systems use
\code{locale -a} in the terminal to list available locales.}

\item{quiet}{whether to output informative messages.}
}
\value{
a function to be applied on a vector of dates
}
\description{
Stamps are just like \code{\link[=format]{format()}}, but based on human-friendly
templates like "Recorded at 10 am, September 2002" or "Meeting, Sunday May
1, 2000, at 10:20 pm".
}
\details{
\code{stamp()} is a stamping function date-time templates mainly, though it
correctly handles all date and time formats as long as they are
unambiguous. \code{stamp_date()}, and \code{stamp_time()} are the specialized
stamps for dates and times (MHS). These function might be useful when the
input template is unambiguous and matches both a time and a date format.

Lubridate tries hard to guess the formats, but often a given format can be
interpreted in multiple ways. One way to deal with such cases is to provide
unambiguous formats like 22/05/81 instead of 10/05/81 for d/m/y
format. Another way is to use a more specialized \code{\link{stamp_date}} and
\code{\link{stamp_time}}. The core function \code{stamp()} prioritizes longer date-time
formats.

If \code{x} is a vector of values \pkg{lubridate} will choose the format which
"fits" \code{x} the best. Note that longer formats are preferred. If you have
"22:23:00 PM" then "HMSp" format will be given priority to shorter "HMS"
order which also fits the supplied string.

Finally, you can give desired format order directly as \code{orders}
argument.
}
\examples{
D <- ymd("2010-04-05") - days(1:5)
stamp("March 1, 1999")(D)
sf <- stamp("Created on Sunday, Jan 1, 1999 3:34 pm")
sf(D)
stamp("Jan 01")(D)
stamp("Sunday, May 1, 2000", locale = "C")(D)
stamp("Sun Aug 5")(D) #=> "Sun Aug 04" "Sat Aug 04" "Fri Aug 04" "Thu Aug 04" "Wed Aug 03"
stamp("12/31/99")(D)              #=> "06/09/11"
stamp("Sunday, May 1, 2000 22:10", locale = "C")(D)
stamp("2013-01-01T06:00:00Z")(D)
stamp("2013-01-01T00:00:00-06")(D)
stamp("2013-01-01T00:00:00-08:00")(force_tz(D, "America/Chicago"))
}
\seealso{
\code{\link[=guess_formats]{guess_formats()}}, \code{\link[=parse_date_time]{parse_date_time()}}, \code{\link[=strptime]{strptime()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stamp.r
\name{stamp}
\alias{stamp}
\alias{stamp_date}
\alias{stamp_time}
\title{Format dates and times based on human-friendly templates}
\usage{
stamp(
  x,
  orders = lubridate_formats,
  locale = Sys.getlocale("LC_TIME"),
  quiet = FALSE,
  exact = FALSE
)

stamp_date(x, locale = Sys.getlocale("LC_TIME"), quiet = FALSE)

stamp_time(x, locale = Sys.getlocale("LC_TIME"), quiet = FALSE)
}
\arguments{
\item{x}{a character vector of templates.}

\item{orders}{orders are sequences of formatting characters which might be
used for disambiguation. For example "ymd hms", "aym" etc. See
\code{\link[=guess_formats]{guess_formats()}} for a list of available formats.}

\item{locale}{locale in which \code{x} is encoded. On Linux-like systems use
\code{locale -a} in the terminal to list available locales.}

\item{quiet}{whether to output informative messages.}

\item{exact}{logical. If \code{TRUE}, the \code{orders} parameter is interpreted as an
exact \code{\link[base:strptime]{base::strptime()}} format and no format guessing is performed.}
}
\value{
a function to be applied on a vector of dates
}
\description{
Stamps are just like \code{\link[=format]{format()}}, but based on human-friendly
templates like "Recorded at 10 am, September 2002" or "Meeting, Sunday May
1, 2000, at 10:20 pm".
}
\details{
\code{stamp()} is a stamping function date-time templates mainly, though it
correctly handles all date and time formats as long as they are
unambiguous. \code{stamp_date()}, and \code{stamp_time()} are the specialized
stamps for dates and times (MHS). These function might be useful when the
input template is unambiguous and matches both a time and a date format.

Lubridate tries hard to guess the formats, but often a given format can be
interpreted in multiple ways. One way to deal with such cases is to provide
unambiguous formats like 22/05/81 instead of 10/05/81 for d/m/y
format. Another way is to use a more specialized \code{\link{stamp_date}} and
\code{\link{stamp_time}}. The core function \code{stamp()} prioritizes longer date-time
formats.

If \code{x} is a vector of values \pkg{lubridate} will choose the format which
"fits" \code{x} the best. Note that longer formats are preferred. If you have
"22:23:00 PM" then "HMSp" format will be given priority to shorter "HMS"
order which also fits the supplied string.

Finally, you can give desired format order directly as \code{orders}
argument.
}
\examples{
D <- ymd("2010-04-05") - days(1:5)
stamp("March 1, 1999")(D)
sf <- stamp("Created on Sunday, Jan 1, 1999 3:34 pm")
sf(D)
stamp("Jan 01")(D)
stamp("Sunday, May 1, 2000", locale = "C")(D)
stamp("Sun Aug 5")(D) #=> "Sun Aug 04" "Sat Aug 04" "Fri Aug 04" "Thu Aug 04" "Wed Aug 03"
stamp("12/31/99")(D)              #=> "06/09/11"
stamp("Sunday, May 1, 2000 22:10", locale = "C")(D)
stamp("2013-01-01T06:00:00Z")(D)
stamp("2013-01-01T00:00:00-06")(D)
stamp("2013-01-01T00:00:00-08:00")(force_tz(D, "America/Chicago"))
}
\seealso{
\code{\link[=guess_formats]{guess_formats()}}, \code{\link[=parse_date_time]{parse_date_time()}}, \code{\link[=strptime]{strptime()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse.r
\name{ymd}
\alias{ymd}
\alias{ydm}
\alias{mdy}
\alias{myd}
\alias{dmy}
\alias{dym}
\alias{yq}
\alias{ym}
\alias{my}
\title{Parse dates with \strong{y}ear, \strong{m}onth, and \strong{d}ay components}
\usage{
ymd(
  ...,
  quiet = FALSE,
  tz = NULL,
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

ydm(
  ...,
  quiet = FALSE,
  tz = NULL,
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

mdy(
  ...,
  quiet = FALSE,
  tz = NULL,
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

myd(
  ...,
  quiet = FALSE,
  tz = NULL,
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

dmy(
  ...,
  quiet = FALSE,
  tz = NULL,
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

dym(
  ...,
  quiet = FALSE,
  tz = NULL,
  locale = Sys.getlocale("LC_TIME"),
  truncated = 0
)

yq(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"))

ym(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"))

my(..., quiet = FALSE, tz = NULL, locale = Sys.getlocale("LC_TIME"))
}
\arguments{
\item{...}{a character or numeric vector of suspected dates}

\item{quiet}{logical. If \code{TRUE}, function evaluates without displaying
customary messages.}

\item{tz}{Time zone indicator. If \code{NULL} (default), a Date object is
returned. Otherwise a POSIXct with time zone attribute set to \code{tz}.}

\item{locale}{locale to be used, see \link{locales}. On Linux systems you
can use \code{system("locale -a")} to list all the installed locales.}

\item{truncated}{integer. Number of formats that can be truncated.}
}
\value{
a vector of class POSIXct if \code{tz} argument is non-\code{NULL} or Date if tz
is \code{NULL} (default)
}
\description{
Transforms dates stored in character and numeric vectors to Date or POSIXct
objects (see \code{tz} argument). These functions recognize arbitrary
non-digit separators as well as no separator. As long as the order of
formats is correct, these functions will parse dates correctly even when the
input vectors contain differently formatted dates. See examples.
}
\details{
In case of heterogeneous date formats, the \code{ymd()} family guesses formats based
on a subset of the input vector. If the input vector contains many missing
values or non-date strings, the subset might not contain meaningful dates
and the date-time format won't be guessed resulting in
\verb{All formats failed to parse} error. In such cases please see
\code{\link[=parse_date_time]{parse_date_time()}} for a more flexible parsing interface.

If the \code{truncated} parameter is non-zero, the \code{ymd()} functions also check for
truncated formats. For example, \code{ymd()} with \code{truncated = 2} will also
parse incomplete dates like \code{2012-06} and \code{2012}.

NOTE: The \code{ymd()} family of functions is based on \code{parse_date_time()} and thus
directly drop to the internal C parser for numeric months, but uses
\code{\link[base:strptime]{base::strptime()}} for alphabetic months. This implies that some of \code{\link[base:strptime]{base::strptime()}}'s
limitations are inherited by \pkg{lubridate}'s parser. For example, truncated
formats (like \verb{\%Y-\%b}) will not be parsed. Numeric truncated formats (like
\verb{\%Y-\%m}) are handled correctly by \pkg{lubridate}'s C parser.

As of version 1.3.0, \pkg{lubridate}'s parse functions no longer return a
message that displays which format they used to parse their input. You can
change this by setting the \code{lubridate.verbose} option to \code{TRUE} with
\code{options(lubridate.verbose = TRUE)}.
}
\examples{
x <- c("09-01-01", "09-01-02", "09-01-03")
ymd(x)
x <- c("2009-01-01", "2009-01-02", "2009-01-03")
ymd(x)
ymd(090101, 90102)
now() > ymd(20090101)
## TRUE
dmy(010210)
mdy(010210)

yq('2014.2')

## heterogeneous formats in a single vector:
x <- c(20090101, "2009-01-02", "2009 01 03", "2009-1-4",
       "2009-1, 5", "Created on 2009 1 6", "200901 !!! 07")
ymd(x)

## What lubridate might not handle:

## Extremely weird cases when one of the separators is "" and some of the
## formats are not in double digits might not be parsed correctly:
\dontrun{ymd("201002-01", "201002-1", "20102-1")
dmy("0312-2010", "312-2010")}
}
\seealso{
\code{\link[=parse_date_time]{parse_date_time()}} for an even more flexible low level
mechanism.
}
\keyword{chron}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timespans.r, R/intervals.r
\name{time_length}
\alias{time_length}
\alias{time_length,Interval-method}
\title{Compute the exact length of a time span}
\usage{
time_length(x, unit = "second")

\S4method{time_length}{Interval}(x, unit = "second")
}
\arguments{
\item{x}{a duration, period, difftime or interval}

\item{unit}{a character string that specifies with time units to use}
}
\value{
the length of the interval in the specified unit. A negative number
connotes a negative interval or duration
}
\description{
Compute the exact length of a time span
}
\details{
When \code{x} is an \linkS4class{Interval} object and
\code{unit} are years or months, \code{time_length()} takes into account
the fact that all months and years don't have the same number of days.

When \code{x} is a \linkS4class{Duration}, \linkS4class{Period}
or \code{\link[=difftime]{difftime()}} object, length in months or years is based on their
most common lengths in seconds (see \code{\link[=timespan]{timespan()}}).
}
\examples{
int <- interval(ymd("1980-01-01"), ymd("2014-09-18"))
time_length(int, "week")

# Exact age
time_length(int, "year")

# Age at last anniversary
trunc(time_length(int, "year"))

# Example of difference between intervals and durations
int <- interval(ymd("1900-01-01"), ymd("1999-12-31"))
time_length(int, "year")
time_length(as.duration(int), "year")
}
\seealso{
\code{\link[=timespan]{timespan()}}
}
\keyword{chron}
\keyword{math}
\keyword{methods}
\keyword{period}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/durations.r
\docType{class}
\name{Duration-class}
\alias{Duration-class}
\alias{durations}
\title{Duration class}
\description{
Duration is an S4 class that extends the \linkS4class{Timespan} class.
Durations record the exact number of seconds in a time span. They measure the
exact passage of time but do not always align with measurements
made in larger units of time such as hours, months and years.
This is because the exact length of larger time units can be affected
by conventions such as leap years
and Daylight Savings Time.
}
\details{
Durations provide a method for measuring generalized timespans when we wish to
treat time as a mathematical quantity that increases in a uniform, monotone manner
along a continuous numberline. They allow exact comparisons with other durations.
See \linkS4class{Period} for an alternative way to measure timespans that better
preserves clock times.

Durations class objects have one slot: .Data, a numeric object equal to the number
of seconds in the duration.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cyclic_encoding.r
\name{cyclic_encoding}
\alias{cyclic_encoding}
\title{Cyclic encoding of date-times}
\usage{
cyclic_encoding(
  x,
  periods,
  encoders = c("sin", "cos"),
  week_start = getOption("lubridate.week.start", 7)
)
}
\arguments{
\item{x}{a date-time object}

\item{periods}{a character vector of periods. Follows same specification as
\link{period} and \link{floor_date} functions.}

\item{encoders}{names of functions to produce the encoding. Defaults to
"sin" and "cos". Names of any predefined functions accepting a numeric
input are allowed.}

\item{week_start}{day starting a week (used for weekly periods).}
}
\value{
a numeric matrix with number of columns equal \code{length(periods) * length(types)}.
}
\description{
Encode a date-time object into a cyclic coordinate system in which the
distances between two pairs of dates separated by the same time duration are
the same.
}
\details{
Machine learning models don't know that December 31st and January 1st are
close in our human calendar sense. \code{cyclic_encoding} makes it obvious to the
machine learner that two calendar dates are close by mapping the dates onto
the circle.
}
\examples{

times <- ymd_hms("2019-01-01 00:00:00") + hours(0:23)
cyclic_encoding(times, c("day", "week", "month"))
plot(cyclic_encoding(times, "1d"))
plot(cyclic_encoding(times, "2d"), xlim = c(-1, 1))
plot(cyclic_encoding(times, "4d"), xlim = c(-1, 1))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cyclic_encoding.r
\name{cyclic_encoding}
\alias{cyclic_encoding}
\title{Cyclic encoding of date-times}
\usage{
cyclic_encoding(
  x,
  periods,
  encoders = c("sin", "cos"),
  week_start = getOption("lubridate.week.start", 7)
)
}
\arguments{
\item{x}{a date-time object}

\item{periods}{a character vector of periods. Follows same specification as
\link{period} and \link{floor_date} functions.}

\item{encoders}{names of functions to produce the encoding. Defaults to
"sin" and "cos". Names of any predefined functions accepting a numeric
input are allowed.}

\item{week_start}{week start day (Default is 7, Sunday. Set \code{lubridate.week.start} to
override). Full or abbreviated names of the days of the week can be in English or
as provided by the current locale.}
}
\value{
a numeric matrix with number of columns equal \code{length(periods) * length(types)}.
}
\description{
Encode a date-time object into a cyclic coordinate system in which the
distances between two pairs of dates separated by the same time duration are
the same.
}
\details{
Machine learning models don't know that December 31st and January 1st are
close in our human calendar sense. \code{cyclic_encoding} makes it obvious to the
machine learner that two calendar dates are close by mapping the dates onto
the circle.
}
\examples{

times <- ymd_hms("2019-01-01 00:00:00") + hours(0:23)
cyclic_encoding(times, c("day", "week", "month"))
plot(cyclic_encoding(times, "1d"))
plot(cyclic_encoding(times, "2d"), xlim = c(-1, 1))
plot(cyclic_encoding(times, "4d"), xlim = c(-1, 1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/instants.r
\docType{data}
\name{origin}
\alias{origin}
\title{1970-01-01 UTC}
\format{
An object of class \code{POSIXct} (inherits from \code{POSIXt}) of length 1.
}
\usage{
origin
}
\description{
Origin is the date-time for 1970-01-01 UTC in POSIXct format. This date-time
is the origin for the numbering system used by POSIXct, POSIXlt, chron, and
Date classes.
}
\examples{
origin
}
\keyword{chron}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/leap-years.r
\name{leap_year}
\alias{leap_year}
\title{Is a year a leap year?}
\usage{
leap_year(date)
}
\arguments{
\item{date}{a date-time object or a year}
}
\value{
TRUE if x is a leap year, FALSE otherwise
}
\description{
If x is a recognized date-time object, leap_year will return whether x
occurs during a leap year. If x is a number, leap_year returns whether it
would be a leap year under the Gregorian calendar.
}
\examples{
x <- as.Date("2009-08-02")
leap_year(x) # FALSE
leap_year(2009) # FALSE
leap_year(2008) # TRUE
leap_year(1900) # FALSE
leap_year(2000) # TRUE
}
\keyword{chron}
\keyword{logic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse.r
\name{ms}
\alias{ms}
\alias{hm}
\alias{hms}
\title{Parse periods with \strong{h}our, \strong{m}inute, and \strong{s}econd components}
\usage{
ms(..., quiet = FALSE, roll = FALSE)

hm(..., quiet = FALSE, roll = FALSE)

hms(..., quiet = FALSE, roll = FALSE)
}
\arguments{
\item{...}{a character vector of hour minute second triples}

\item{quiet}{logical. If \code{TRUE}, function evaluates without displaying
customary messages.}

\item{roll}{logical. If \code{TRUE}, smaller units are rolled over to higher units
if they exceed the conventional limit. For example,
\code{hms("01:59:120", roll = TRUE)} produces period "2H 1M 0S".}
}
\value{
a vector of period objects
}
\description{
Transforms a character or numeric vector into a period object with the
specified number of hours, minutes, and seconds. \code{hms()} recognizes all
non-numeric characters except '-' as separators ('-' is used for negative
\code{durations}). After hours, minutes and seconds have been parsed, the
remaining input is ignored.
}
\examples{
ms(c("09:10", "09:02", "1:10"))
ms("7 6")
ms("6,5")
hm(c("09:10", "09:02", "1:10"))
hm("7 6")
hm("6,5")

x <- c("09:10:01", "09:10:02", "09:10:03")
hms(x)

hms("7 6 5", "3:23:::2", "2 : 23 : 33", "Finished in 9 hours, 20 min and 4 seconds")
}
\seealso{
\code{\link[=hm]{hm()}}, \code{\link[=ms]{ms()}}
}
\keyword{period}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercion.r
\name{DateCoercion}
\alias{DateCoercion}
\title{Convert a variety of date-time classes to POSIXlt and POSIXct}
\description{
Convert a variety of date-time classes to POSIXlt and POSIXct
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update.r
\name{fit_to_timeline}
\alias{fit_to_timeline}
\title{Fit a POSIXlt date-time to the timeline}
\usage{
fit_to_timeline(lt, class = "POSIXct", simple = FALSE)
}
\arguments{
\item{lt}{a POSIXlt date-time object.}

\item{class}{a character string that describes what type of object to return,
POSIXlt or POSIXct. Defaults to POSIXct. This is an optimization to avoid
needless conversions.}

\item{simple}{if TRUE, \pkg{lubridate} makes no attempt to detect
meaningless time-dates or to correct time zones. No NAs are produced and
the most meaningful valid dates are returned instead. See examples.}
}
\value{
a POSIXct or POSIXlt object that contains no illusory date-times
}
\description{
The POSIXlt format allows you to create instants that do not exist in real
life due to daylight savings time and other conventions. fit_to_timeline
matches POSIXlt date-times to a real times. If an instant does not exist, fit
to timeline will replace it with an NA. If an instant does exist, but has
been paired with an incorrect timezone/daylight savings time combination,
fit_to_timeline returns the instant with the correct combination.
}
\examples{
\dontrun{

tricky <- structure(list(sec   = c(5,    0,    0,    -1),
                         min   = c(0L,   5L,   5L,   0L),
                         hour  = c(2L,   0L,   2L,   2L),
                         mday  = c(4L,   4L,   14L,  4L),
                         mon   = c(10L,  10L,  2L,   10L),
                         year  = c(112L, 112L, 110L, 112L),
                         wday  = c(0L,   0L,   0L,   0L),
                         yday  = c(308L, 308L, 72L,  308L),
                         isdst = c(1L,   0L,   0L,   1L)),
                    .Names = c("sec", "min", "hour", "mday", "mon",
                               "year", "wday", "yday",  "isdst"),
                    class = c("POSIXlt", "POSIXt"),
                    tzone = c("America/Chicago", "CST", "CDT"))

tricky
## [1] "2012-11-04 02:00:00 CDT" Doesn't exist because clocks "fall back" to 1:00 CST
## [2] "2012-11-04 00:05:00 CST" Times are still CDT, not CST at this instant
## [3] "2010-03-14 02:00:00 CDT" DST gap
## [4] "2012-11-04 01:59:59 CDT" Does exist, but has deceptive internal structure

fit_to_timeline(tricky)
## Returns:
## [1] "2012-11-04 02:00:00 CST" instant paired with correct tz & DST combination
## [2] "2012-11-04 00:05:00 CDT" instant paired with correct tz & DST combination
## [3] NA - fake time changed to NA (compare to as.POSIXct(tricky))
## [4] "2012-11-04 01:59:59 CDT" -real instant, left as is

fit_to_timeline(tricky, simple = TRUE)
## Returns valid time-dates by extrapolating CDT and CST zones:
## [1] "2012-11-04 01:00:05 CST" "2012-11-04 01:05:00 CDT"
## [3] "2010-03-14 03:05:00 CDT" "2012-11-04 01:59:59 CDT"
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update.r
\name{fit_to_timeline}
\alias{fit_to_timeline}
\title{Fit a POSIXlt date-time to the timeline}
\usage{
fit_to_timeline(lt, class = "POSIXct", simple = FALSE)
}
\arguments{
\item{lt}{a POSIXlt date-time object.}

\item{class}{a character string that describes what type of object to return,
POSIXlt or POSIXct. Defaults to POSIXct. This is an optimization to avoid
needless conversions.}

\item{simple}{if TRUE, \pkg{lubridate} makes no attempt to detect
meaningless time-dates or to correct time zones. No NAs are produced and
the most meaningful valid dates are returned instead. See examples.}
}
\value{
a POSIXct or POSIXlt object that contains no illusory date-times
}
\description{
The POSIXlt format allows you to create instants that do not exist in real
life due to daylight savings time and other conventions. fit_to_timeline
matches POSIXlt date-times to a real times. If an instant does not exist, fit
to timeline will replace it with an NA. If an instant does exist, but has
been paired with an incorrect timezone/daylight savings time combination,
fit_to_timeline returns the instant with the correct combination.
}
\examples{
\dontrun{

tricky <- structure(list(
  sec = c(5, 0, 0, -1),
  min = c(0L, 5L, 5L, 0L),
  hour = c(2L, 0L, 2L, 2L),
  mday = c(4L, 4L, 14L, 4L),
  mon = c(10L, 10L, 2L, 10L),
  year = c(112L, 112L, 110L, 112L),
  wday = c(0L, 0L, 0L, 0L),
  yday = c(308L, 308L, 72L, 308L),
  isdst = c(1L, 0L, 0L, 1L)
),
.Names = c(
  "sec", "min", "hour", "mday", "mon",
  "year", "wday", "yday", "isdst"
),
class = c("POSIXlt", "POSIXt"),
tzone = c("America/Chicago", "CST", "CDT")
)

tricky
## [1] "2012-11-04 02:00:00 CDT" Doesn't exist because clocks "fall back" to 1:00 CST
## [2] "2012-11-04 00:05:00 CST" Times are still CDT, not CST at this instant
## [3] "2010-03-14 02:00:00 CDT" DST gap
## [4] "2012-11-04 01:59:59 CDT" Does exist, but has deceptive internal structure

fit_to_timeline(tricky)
## Returns:
## [1] "2012-11-04 02:00:00 CST" instant paired with correct tz & DST combination
## [2] "2012-11-04 00:05:00 CDT" instant paired with correct tz & DST combination
## [3] NA - fake time changed to NA (compare to as.POSIXct(tricky))
## [4] "2012-11-04 01:59:59 CDT" -real instant, left as is

fit_to_timeline(tricky, simple = TRUE)
## Returns valid time-dates by extrapolating CDT and CST zones:
## [1] "2012-11-04 01:00:05 CST" "2012-11-04 01:05:00 CDT"
## [3] "2010-03-14 03:05:00 CDT" "2012-11-04 01:59:59 CDT"
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/intervals.r, R/durations.r, R/periods.r,
%   R/hidden.r
\docType{data}
\name{hidden_aliases}
\alias{hidden_aliases}
\alias{Arith,Interval,ANY-method}
\alias{Arith,ANY,Interval-method}
\alias{intersect,Interval,Interval-method}
\alias{union,Interval,Interval-method}
\alias{setdiff,Interval,Interval-method}
\alias{as.numeric,Interval-method}
\alias{show,Interval-method}
\alias{c,Interval-method}
\alias{rep,Interval-method}
\alias{[,Interval-method}
\alias{[<-,Interval,ANY,ANY,ANY-method}
\alias{[[,Interval-method}
\alias{[[<-,Interval,ANY,ANY,ANY-method}
\alias{$,Interval-method}
\alias{$<-,Interval-method}
\alias{as.difftime,Interval-method}
\alias{as.character,Interval-method}
\alias{+,Interval,Duration-method}
\alias{+,Interval,Interval-method}
\alias{+,Interval,Period-method}
\alias{+,Interval,Date-method}
\alias{+,Date,Interval-method}
\alias{+,Interval,difftime-method}
\alias{+,difftime,Interval-method}
\alias{+,Interval,numeric-method}
\alias{+,numeric,Interval-method}
\alias{+,Interval,POSIXct-method}
\alias{+,POSIXct,Interval-method}
\alias{+,Interval,POSIXlt-method}
\alias{+,POSIXlt,Interval-method}
\alias{/,Interval,Duration-method}
\alias{/,Interval,Interval-method}
\alias{/,Interval,Period-method}
\alias{/,Interval,difftime-method}
\alias{/,difftime,Interval-method}
\alias{/,Interval,numeric-method}
\alias{/,numeric,Interval-method}
\alias{*,Interval,ANY-method}
\alias{*,ANY,Interval-method}
\alias{-,Interval,missing-method}
\alias{-,Interval,Interval-method}
\alias{-,Date,Interval-method}
\alias{-,POSIXct,Interval-method}
\alias{-,POSIXlt,Interval-method}
\alias{-,numeric,Interval-method}
\alias{-,Interval,Date-method}
\alias{-,Interval,POSIXct-method}
\alias{-,Interval,POSIXlt-method}
\alias{-,Interval,numeric-method}
\alias{-,Duration,Interval-method}
\alias{-,Period,Interval-method}
\alias{\%\%,Interval,Duration-method}
\alias{\%\%,Interval,Interval-method}
\alias{\%\%,Interval,Period-method}
\alias{\%\%,Interval,Duration}
\alias{\%\%,Interval,Interval}
\alias{\%\%,Interval,Period}
\alias{-,Date,Interval}
\alias{-,Duration,Interval}
\alias{-,Interval,Date}
\alias{-,Interval,Interval}
\alias{-,Interval,POSIXct}
\alias{-,Interval,POSIXlt}
\alias{-,Interval,numeric}
\alias{-,POSIXct,Interval}
\alias{-,POSIXlt,Interval}
\alias{-,numeric,Interval}
\alias{Compare,Duration,ANY-method}
\alias{Compare,Duration,Duration-method}
\alias{Compare,difftime,Duration-method}
\alias{Compare,ANY,Duration-method}
\alias{Compare,Duration,Period-method}
\alias{Compare,Duration,difftime-method}
\alias{Compare,character,Duration-method}
\alias{Compare,Duration,character-method}
\alias{Compare,numeric,Duration-method}
\alias{Compare,Duration,numeric-method}
\alias{as.numeric,Duration-method}
\alias{show,Duration-method}
\alias{c,Duration-method}
\alias{rep,Duration-method}
\alias{[,Duration-method}
\alias{[<-,Duration,ANY,ANY,ANY-method}
\alias{[[,Duration-method}
\alias{[[<-,Duration,ANY,ANY,ANY-method}
\alias{$,Duration-method}
\alias{$<-,Duration-method}
\alias{as.difftime,Duration-method}
\alias{as.character,Duration-method}
\alias{+,Duration,Duration-method}
\alias{+,Duration,Interval-method}
\alias{+,Duration,Period-method}
\alias{+,Duration,Date-method}
\alias{+,Date,Duration-method}
\alias{+,Duration,difftime-method}
\alias{+,difftime,Duration-method}
\alias{+,Duration,numeric-method}
\alias{+,numeric,Duration-method}
\alias{+,Duration,POSIXct-method}
\alias{+,POSIXct,Duration-method}
\alias{+,Duration,POSIXlt-method}
\alias{+,POSIXlt,Duration-method}
\alias{-,Duration,ANY-method}
\alias{/,Duration,Duration-method}
\alias{/,Duration,Interval-method}
\alias{/,Duration,Period-method}
\alias{/,Duration,difftime-method}
\alias{/,difftime,Duration-method}
\alias{/,Duration,numeric-method}
\alias{/,numeric,Duration-method}
\alias{*,Duration,ANY-method}
\alias{*,ANY,Duration-method}
\alias{\%\%,Duration,Duration-method}
\alias{\%\%,Duration,Interval-method}
\alias{\%\%,Duration,Period-method}
\alias{-,Duration,missing-method}
\alias{-,ANY,Duration-method}
\alias{Arith,ANY,Period-method}
\alias{Arith,Duration,Period-method}
\alias{Arith,Period,Duration-method}
\alias{Compare,Period,Duration-method}
\alias{Compare,numeric,Period-method}
\alias{Compare,difftime,Period-method}
\alias{Compare,Period,difftime-method}
\alias{Compare,Period,Period-method}
\alias{Compare,Period,character-method}
\alias{Compare,Period,numeric-method}
\alias{Compare,character,Period-method}
\alias{second,Period-method}
\alias{second<-,Period-method}
\alias{minute,Period-method}
\alias{minute<-,Period-method}
\alias{hour,Period-method}
\alias{hour<-,Period-method}
\alias{Arith,Period,ANY-method}
\alias{day,Period-method}
\alias{day<-,Period-method}
\alias{month,Period-method}
\alias{month<-,Period-method}
\alias{year,Period-method}
\alias{year<-,Period-method}
\alias{date,Period-method}
\alias{date<-,Period-method}
\alias{as.numeric,Period-method}
\alias{show,Period-method}
\alias{c,Period-method}
\alias{rep,Period-method}
\alias{[,Period-method}
\alias{[<-,Period,ANY,ANY,Period-method}
\alias{[[,Period-method}
\alias{[[<-,Period,ANY,ANY,Period-method}
\alias{$,Period-method}
\alias{$<-,Period-method}
\alias{as.difftime,Period-method}
\alias{as.character,Period-method}
\alias{+,Period,Duration-method}
\alias{+,Period,Interval-method}
\alias{+,Period,Period-method}
\alias{+,Period,Date-method}
\alias{+,Date,Period-method}
\alias{+,Period,difftime-method}
\alias{+,difftime,Period-method}
\alias{+,Period,numeric-method}
\alias{+,numeric,Period-method}
\alias{+,Period,POSIXct-method}
\alias{+,POSIXct,Period-method}
\alias{+,Period,POSIXlt-method}
\alias{+,POSIXlt,Period-method}
\alias{/,Period,Duration-method}
\alias{/,Period,Interval-method}
\alias{/,Period,Period-method}
\alias{/,Period,difftime-method}
\alias{/,difftime,Period-method}
\alias{/,Period,numeric-method}
\alias{/,numeric,Period-method}
\alias{*,Period,ANY-method}
\alias{*,ANY,Period-method}
\alias{-,Period,ANY-method}
\alias{-,Period,missing-method}
\alias{-,ANY,Period-method}
\alias{\%\%,Period,Duration-method}
\alias{\%\%,Period,Interval-method}
\alias{\%\%,Period,Period-method}
\alias{>,Period,Period-method}
\alias{>=,Period,Period-method}
\alias{==,Period,Period-method}
\alias{!=,Period,Period-method}
\alias{<=,Period,Period-method}
\alias{<,Period,Period-method}
\alias{>,Period,Duration-method}
\alias{>=,Period,Duration-method}
\alias{==,Period,Duration-method}
\alias{!=,Period,Duration-method}
\alias{<=,Period,Duration-method}
\alias{<,Period,Duration-method}
\alias{>,Duration,Period-method}
\alias{>=,Duration,Period-method}
\alias{==,Duration,Period-method}
\alias{!=,Duration,Period-method}
\alias{<=,Duration,Period-method}
\alias{<,Duration,Period-method}
\alias{>,Period,numeric-method}
\alias{>=,Period,numeric-method}
\alias{==,Period,numeric-method}
\alias{!=,Period,numeric-method}
\alias{<=,Period,numeric-method}
\alias{<,Period,numeric-method}
\alias{>,numeric,Period-method}
\alias{>=,numeric,Period-method}
\alias{==,numeric,Period-method}
\alias{!=,numeric,Period-method}
\alias{<=,numeric,Period-method}
\alias{<,numeric,Period-method}
\alias{!=,Duration,Period}
\alias{!=,Period,Duration}
\alias{!=,Period,Period}
\alias{!=,Period,numeric}
\alias{!=,numeric,Period}
\alias{\%\%,Period,Duration}
\alias{\%\%,Period,Interval}
\alias{\%\%,Period,Period}
\alias{*,ANY,Period}
\alias{*,Period,ANY}
\alias{-,ANY,Period}
\alias{-,Period,Interval}
\alias{-,Period,missing}
\alias{/,numeric,Period}
\alias{<,Duration,Period}
\alias{<,Period,Duration}
\alias{<,Period,Period}
\alias{<,Period,numeric}
\alias{<,numeric,Period}
\alias{<=,Duration,Period}
\alias{<=,Period,Duration}
\alias{<=,Period,Period}
\alias{<=,Period,numeric}
\alias{<=,numeric,Period}
\alias{==,Duration,Period}
\alias{==,Period,Duration}
\alias{==,Period,Period}
\alias{==,Period,numeric}
\alias{==,numeric,Period}
\alias{>,Duration,Period}
\alias{>,Period,Duration}
\alias{>,Period,Period}
\alias{>,Period,numeric}
\alias{>,numeric,Period}
\alias{>=,Duration,Period}
\alias{>=,Period,Duration}
\alias{>=,Period,Period}
\alias{>=,Period,numeric}
\alias{>=,numeric,Period}
\title{Internal page for hidden aliases}
\description{
For S4 methods that require a documentation entry but only clutter the index.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dates.r, R/POSIXt.r, R/intervals.r,
%   R/durations.r, R/periods.r, R/hidden.r
\docType{data}
\name{hidden_aliases}
\alias{hidden_aliases}
\alias{day<-,Date-method}
\alias{hour<-,Date-method}
\alias{minute<-,Date-method}
\alias{month<-,Date-method}
\alias{second<-,Date-method}
\alias{year<-,Date-method}
\alias{day<-,POSIXt-method}
\alias{hour<-,POSIXt-method}
\alias{minute<-,POSIXt-method}
\alias{month<-,POSIXt-method}
\alias{second<-,POSIXt-method}
\alias{year<-,POSIXt-method}
\alias{Arith,Interval,ANY-method}
\alias{Arith,ANY,Interval-method}
\alias{intersect,Interval,Interval-method}
\alias{union,Interval,Interval-method}
\alias{setdiff,Interval,Interval-method}
\alias{as.numeric,Interval-method}
\alias{as.integer,Interval-method}
\alias{show,Interval-method}
\alias{c,Interval-method}
\alias{rep,Interval-method}
\alias{[,Interval-method}
\alias{[<-,Interval,ANY,ANY,ANY-method}
\alias{[[,Interval-method}
\alias{[[<-,Interval,ANY,ANY,ANY-method}
\alias{$,Interval-method}
\alias{$<-,Interval-method}
\alias{as.difftime,Interval-method}
\alias{as.character,Interval-method}
\alias{day<-,Interval-method}
\alias{hour<-,Interval-method}
\alias{minute<-,Interval-method}
\alias{month<-,Interval-method}
\alias{second<-,Interval-method}
\alias{year<-,Interval-method}
\alias{+,Interval,Duration-method}
\alias{+,Interval,Interval-method}
\alias{+,Interval,Period-method}
\alias{+,Interval,Date-method}
\alias{+,Date,Interval-method}
\alias{+,Interval,difftime-method}
\alias{+,difftime,Interval-method}
\alias{+,Interval,numeric-method}
\alias{+,numeric,Interval-method}
\alias{+,Interval,POSIXct-method}
\alias{+,POSIXct,Interval-method}
\alias{+,Interval,POSIXlt-method}
\alias{+,POSIXlt,Interval-method}
\alias{/,Interval,Duration-method}
\alias{/,Interval,Interval-method}
\alias{/,Interval,Period-method}
\alias{/,Interval,difftime-method}
\alias{/,difftime,Interval-method}
\alias{/,Interval,numeric-method}
\alias{/,numeric,Interval-method}
\alias{*,Interval,ANY-method}
\alias{*,ANY,Interval-method}
\alias{-,Interval,missing-method}
\alias{-,Interval,Interval-method}
\alias{-,Date,Interval-method}
\alias{-,POSIXct,Interval-method}
\alias{-,POSIXlt,Interval-method}
\alias{-,numeric,Interval-method}
\alias{-,Interval,Date-method}
\alias{-,Interval,POSIXct-method}
\alias{-,Interval,POSIXlt-method}
\alias{-,Interval,numeric-method}
\alias{-,Duration,Interval-method}
\alias{-,Period,Interval-method}
\alias{\%\%,Interval,Duration-method}
\alias{\%\%,Interval,Interval-method}
\alias{\%\%,Interval,Period-method}
\alias{\%\%,Interval,Duration}
\alias{\%\%,Interval,Interval}
\alias{\%\%,Interval,Period}
\alias{-,Date,Interval}
\alias{-,Duration,Interval}
\alias{-,Interval,Date}
\alias{-,Interval,Interval}
\alias{-,Interval,POSIXct}
\alias{-,Interval,POSIXlt}
\alias{-,Interval,numeric}
\alias{-,POSIXct,Interval}
\alias{-,POSIXlt,Interval}
\alias{-,numeric,Interval}
\alias{-,Interval,Duration-method}
\alias{-,Interval,Period-method}
\alias{Compare,Duration,ANY-method}
\alias{Compare,Duration,Duration-method}
\alias{Compare,difftime,Duration-method}
\alias{Compare,ANY,Duration-method}
\alias{Compare,Duration,Period-method}
\alias{Compare,Duration,difftime-method}
\alias{Compare,character,Duration-method}
\alias{Compare,Duration,character-method}
\alias{Compare,numeric,Duration-method}
\alias{Compare,Duration,numeric-method}
\alias{as.numeric,Duration-method}
\alias{as.integer,Duration-method}
\alias{show,Duration-method}
\alias{c,Duration-method}
\alias{rep,Duration-method}
\alias{[,Duration-method}
\alias{[<-,Duration,ANY,ANY,ANY-method}
\alias{[[,Duration-method}
\alias{[[<-,Duration,ANY,ANY,ANY-method}
\alias{$,Duration-method}
\alias{$<-,Duration-method}
\alias{as.difftime,Duration-method}
\alias{day<-,Duration-method}
\alias{hour<-,Duration-method}
\alias{minute<-,Duration-method}
\alias{month<-,Duration-method}
\alias{second<-,Duration-method}
\alias{year<-,Duration-method}
\alias{as.character,Duration-method}
\alias{+,Duration,Duration-method}
\alias{+,Duration,Interval-method}
\alias{+,Duration,Period-method}
\alias{+,Duration,Date-method}
\alias{+,Date,Duration-method}
\alias{+,Duration,difftime-method}
\alias{+,difftime,Duration-method}
\alias{+,Duration,numeric-method}
\alias{+,numeric,Duration-method}
\alias{+,Duration,POSIXct-method}
\alias{+,POSIXct,Duration-method}
\alias{+,Duration,POSIXlt-method}
\alias{+,POSIXlt,Duration-method}
\alias{-,Duration,ANY-method}
\alias{/,Duration,Duration-method}
\alias{/,Duration,Interval-method}
\alias{/,Duration,Period-method}
\alias{/,Duration,difftime-method}
\alias{/,difftime,Duration-method}
\alias{/,Duration,numeric-method}
\alias{/,numeric,Duration-method}
\alias{*,Duration,ANY-method}
\alias{*,ANY,Duration-method}
\alias{\%\%,Duration,Duration-method}
\alias{\%\%,Duration,Interval-method}
\alias{\%\%,Duration,Period-method}
\alias{-,Duration,missing-method}
\alias{-,ANY,Duration-method}
\alias{-,Duration,Duration-method}
\alias{-,Duration,Date-method}
\alias{-,Duration,POSIXct-method}
\alias{-,Duration,POSIXlt-method}
\alias{-,Duration,Period-method}
\alias{-,Duration,numeric-method}
\alias{-,POSIXct,Duration-method}
\alias{-,numeric,Duration-method}
\alias{-,POSIXlt,Duration-method}
\alias{-,Date,Duration-method}
\alias{Arith,ANY,Period-method}
\alias{Arith,Duration,Period-method}
\alias{Arith,Period,Duration-method}
\alias{Compare,Period,Duration-method}
\alias{Compare,numeric,Period-method}
\alias{Compare,difftime,Period-method}
\alias{Compare,Period,difftime-method}
\alias{Compare,Period,Period-method}
\alias{Compare,Period,character-method}
\alias{Compare,Period,numeric-method}
\alias{Compare,character,Period-method}
\alias{second,Period-method}
\alias{second<-,Period-method}
\alias{minute,Period-method}
\alias{minute<-,Period-method}
\alias{hour,Period-method}
\alias{hour<-,Period-method}
\alias{Arith,Period,ANY-method}
\alias{day,Period-method}
\alias{day<-,Period-method}
\alias{month,Period-method}
\alias{month<-,Period-method}
\alias{year,Period-method}
\alias{year<-,Period-method}
\alias{date,Period-method}
\alias{date<-,Period-method}
\alias{as.numeric,Period-method}
\alias{as.integer,Period-method}
\alias{show,Period-method}
\alias{c,Period-method}
\alias{rep,Period-method}
\alias{[,Period-method}
\alias{[<-,Period,ANY,ANY,Period-method}
\alias{[[,Period-method}
\alias{[[<-,Period,ANY,ANY,Period-method}
\alias{$,Period-method}
\alias{$<-,Period-method}
\alias{as.difftime,Period-method}
\alias{as.character,Period-method}
\alias{+,Period,Duration-method}
\alias{+,Period,Interval-method}
\alias{+,Period,Period-method}
\alias{+,Period,Date-method}
\alias{+,Date,Period-method}
\alias{+,Period,difftime-method}
\alias{+,difftime,Period-method}
\alias{+,Period,numeric-method}
\alias{+,numeric,Period-method}
\alias{+,Period,POSIXct-method}
\alias{+,POSIXct,Period-method}
\alias{+,Period,POSIXlt-method}
\alias{+,POSIXlt,Period-method}
\alias{/,Period,Duration-method}
\alias{/,Period,Interval-method}
\alias{/,Period,Period-method}
\alias{/,Period,difftime-method}
\alias{/,difftime,Period-method}
\alias{/,Period,numeric-method}
\alias{/,numeric,Period-method}
\alias{*,Period,ANY-method}
\alias{*,ANY,Period-method}
\alias{-,Period,ANY-method}
\alias{-,Period,missing-method}
\alias{-,ANY,Period-method}
\alias{-,Period,Period-method}
\alias{\%\%,Period,Duration-method}
\alias{\%\%,Period,Interval-method}
\alias{\%\%,Period,Period-method}
\alias{>,Period,Period-method}
\alias{>=,Period,Period-method}
\alias{==,Period,Period-method}
\alias{!=,Period,Period-method}
\alias{<=,Period,Period-method}
\alias{<,Period,Period-method}
\alias{>,Period,Duration-method}
\alias{>=,Period,Duration-method}
\alias{==,Period,Duration-method}
\alias{!=,Period,Duration-method}
\alias{<=,Period,Duration-method}
\alias{<,Period,Duration-method}
\alias{>,Duration,Period-method}
\alias{>=,Duration,Period-method}
\alias{==,Duration,Period-method}
\alias{!=,Duration,Period-method}
\alias{<=,Duration,Period-method}
\alias{<,Duration,Period-method}
\alias{>,Period,numeric-method}
\alias{>=,Period,numeric-method}
\alias{==,Period,numeric-method}
\alias{!=,Period,numeric-method}
\alias{<=,Period,numeric-method}
\alias{<,Period,numeric-method}
\alias{>,numeric,Period-method}
\alias{>=,numeric,Period-method}
\alias{==,numeric,Period-method}
\alias{!=,numeric,Period-method}
\alias{<=,numeric,Period-method}
\alias{<,numeric,Period-method}
\alias{!=,Duration,Period}
\alias{!=,Period,Duration}
\alias{!=,Period,Period}
\alias{!=,Period,numeric}
\alias{!=,numeric,Period}
\alias{\%\%,Period,Duration}
\alias{\%\%,Period,Interval}
\alias{\%\%,Period,Period}
\alias{*,ANY,Period}
\alias{*,Period,ANY}
\alias{-,ANY,Period}
\alias{-,Period,Interval}
\alias{-,Period,missing}
\alias{/,numeric,Period}
\alias{<,Duration,Period}
\alias{<,Period,Duration}
\alias{<,Period,Period}
\alias{<,Period,numeric}
\alias{<,numeric,Period}
\alias{<=,Duration,Period}
\alias{<=,Period,Duration}
\alias{<=,Period,Period}
\alias{<=,Period,numeric}
\alias{<=,numeric,Period}
\alias{==,Duration,Period}
\alias{==,Period,Duration}
\alias{==,Period,Period}
\alias{==,Period,numeric}
\alias{==,numeric,Period}
\alias{>,Duration,Period}
\alias{>,Period,Duration}
\alias{>,Period,Period}
\alias{>,Period,numeric}
\alias{>,numeric,Period}
\alias{>=,Duration,Period}
\alias{>=,Period,Duration}
\alias{>=,Period,Period}
\alias{>=,Period,numeric}
\alias{>=,numeric,Period}
\alias{-,Period,Date-method}
\alias{-,Period,Duration-method}
\alias{-,Period,POSIXct-method}
\alias{-,Period,POSIXlt-method}
\alias{-,Period,numeric-method}
\alias{-,Date,Period-method}
\alias{-,numeric,Period-method}
\alias{-,POSIXlt,Period-method}
\alias{-,POSIXct,Period-method}
\title{Internal page for hidden aliases}
\description{
For S4 methods that require a documentation entry but only clutter the index.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decimal-dates.r
\name{date_decimal}
\alias{date_decimal}
\title{Converts a decimal to a date}
\usage{
date_decimal(decimal, tz = "UTC")
}
\arguments{
\item{decimal}{a numeric object}

\item{tz}{the time zone required}
}
\value{
a POSIXct object, whose year corresponds to the integer part of
decimal. The months, days, hours, minutes and seconds elements are picked so
the date-time will accurately represent the fraction of the year expressed by
decimal.
}
\description{
Converts a decimal to a date
}
\examples{
date <- ymd("2009-02-10")
decimal <- decimal_date(date) # 2009.11
date_decimal(decimal) # "2009-02-10 UTC"
}
\keyword{chron}
\keyword{manip}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decimal-dates.r
\name{date_decimal}
\alias{date_decimal}
\title{Converts a decimal to a date}
\usage{
date_decimal(decimal, tz = "UTC")
}
\arguments{
\item{decimal}{a numeric object}

\item{tz}{the time zone required}
}
\value{
a POSIXct object, whose year corresponds to the integer part of
decimal. The months, days, hours, minutes and seconds elements are picked so
the date-time will accurately represent the fraction of the year expressed by
decimal.
}
\description{
Converts a decimal to a date
}
\examples{
date <- ymd("2009-02-10")
decimal <- decimal_date(date)  # 2009.11
date_decimal(decimal) # "2009-02-10 UTC"
}
\keyword{chron}
\keyword{manip}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops-m+.r
\name{\%m+\%}
\alias{\%m+\%}
\alias{m+}
\alias{m-}
\alias{\%m-\%}
\alias{\%m+\%,ANY,ANY-method}
\alias{\%m-\%,ANY,ANY-method}
\alias{\%m+\%,Period,ANY-method}
\alias{\%m+\%,ANY,Period-method}
\alias{\%m-\%,Period,ANY-method}
\alias{\%m-\%,ANY,Period-method}
\alias{\%m+\%,Duration,ANY-method}
\alias{\%m+\%,ANY,Duration-method}
\alias{\%m-\%,Duration,ANY-method}
\alias{\%m-\%,ANY,Duration-method}
\alias{\%m+\%,Interval,ANY-method}
\alias{\%m+\%,ANY,Interval-method}
\alias{\%m-\%,Interval,ANY-method}
\alias{\%m-\%,ANY,Interval-method}
\alias{add_with_rollback}
\title{Add and subtract months to a date without exceeding the last day of the new month}
\usage{
e1 \%m+\% e2

add_with_rollback(e1, e2, roll_to_first = FALSE, preserve_hms = TRUE)
}
\arguments{
\item{e1}{A period or a date-time object of class \link{POSIXlt}, \link{POSIXct}
or \link{Date}.}

\item{e2}{A period or a date-time object of class \link{POSIXlt}, \link{POSIXct}
or \link{Date}. Note that one of e1 and e2 must be a period and the other a
date-time object.}

\item{roll_to_first}{rollback to the first day of the month instead of the
last day of the previous month (passed to \code{\link[=rollback]{rollback()}})}

\item{preserve_hms}{retains the same hour, minute, and second information? If
FALSE, the new date will be at 00:00:00 (passed to \code{\link[=rollback]{rollback()}})}
}
\value{
A date-time object of class POSIXlt, POSIXct or Date
}
\description{
Adding months frustrates basic arithmetic because consecutive months have
different lengths. With other elements, it is helpful for arithmetic to
perform automatic roll over. For example, 12:00:00 + 61 seconds becomes
12:01:01. However, people often prefer that this behavior NOT occur with
months. For example, we sometimes want January 31 + 1 month = February 28 and
not March 3. \code{\%m+\%} performs this type of arithmetic. Date \code{\%m+\%} months(n)
always returns a date in the nth month after Date. If the new date would
usually spill over into the n + 1th month, \code{\%m+\%} will return the last day of
the nth month (\code{\link[=rollback]{rollback()}}). Date \code{\%m-\%} months(n) always returns a
date in the nth month before Date.
}
\details{
\code{\%m+\%} and \code{\%m-\%} handle periods with components less than a month by first
adding/subtracting months and then performing usual arithmetics with smaller
units.

\code{\%m+\%} and \code{\%m-\%} should be used with caution as they are not one-to-one
operations and results for either will be sensitive to the order of
operations.
}
\examples{
jan <- ymd_hms("2010-01-31 03:04:05")
jan + months(1:3) # Feb 31 and April 31 returned as NA
# NA "2010-03-31 03:04:05 UTC" NA
jan \%m+\% months(1:3) # No rollover

leap <- ymd("2012-02-29")
"2012-02-29 UTC"
leap \%m+\% years(1)
leap \%m+\% years(-1)
leap \%m-\% years(1)

x <- ymd_hms("2019-01-29 01:02:03")
add_with_rollback(x, months(1))
add_with_rollback(x, months(1), preserve_hms = FALSE)
add_with_rollback(x, months(1), roll_to_first = TRUE)
add_with_rollback(x, months(1), roll_to_first = TRUE, preserve_hms = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/periods.r
\docType{class}
\name{Period-class}
\alias{Period-class}
\title{Period class}
\description{
Period is an S4 class that extends the \linkS4class{Timespan} class.
Periods track the change in the "clock time" between two date-times. They
are measured in common time related units: years, months, days, hours,
minutes, and seconds. Each unit except for seconds must be expressed in
integer values.
}
\details{
The exact length of a period is not defined until the period is placed at a
specific moment of time. This is because the precise length of one year,
month, day, etc. can change depending on when it occurs due to daylight savings,
leap years, and other conventions. A period can be
associated with a specific moment in time by coercing it to an
\linkS4class{Interval} object with \code{\link[=as.interval]{as.interval()}} or by adding
it to a date-time with "+".

Periods provide a method for measuring generalized timespans when we wish to
model clock times. Periods will attain intuitive results at this task even
when leap years, leap seconds, gregorian days, daylight savings changes, and
other events happen during the period.

Because Period represents imprecise amount of time it cannot be compared to
precise timestamps as Durations and Intervals are. You need to explicitely
convert to durations. See \linkS4class{Duration}.

The logic that guides arithmetic with periods can be unintuitive. Starting
with version 1.3.0, \pkg{lubridate} enforces the reversible property of arithmetic
(e.g. a date + period - period = date) by returning an NA if you create an
implausible date by adding periods with months or years units to a date. For
example,  adding one month to January 31st, 2013 results in February 31st,
2013, which is not a real date. \pkg{lubridate} users have argued in the past that
February 31st, 2013 should be rolled over to March 3rd, 2013 or rolled back
to February 28, 2013. However, each of these corrections would destroy the
reversibility of addition (Mar 3 - one month == Feb 3 != Jan 31, Feb 28 - one
month == Jan 28 != Jan 31). If you would like to add and subtract months in a
way that rolls the results back to the last day of a month (when appropriate)
use the special operators, \code{\link{\%m+\%}},  \code{\link{\%m-\%}} or a
bit more flexible \code{\link[=add_with_rollback]{add_with_rollback()}}.

Period class objects have six slots. 1) .Data, a numeric object. The
apparent amount of seconds to add to the period. 2) minute, a numeric object.
The apparent amount of minutes to add to the period. 3) hour, a numeric object.
The apparent amount of hours to add to the period.4) day, a numeric object.
The apparent amount of days to add to the period.5) month, a numeric object.
The apparent amount of months to add to the period. 6) year, a numeric object.
The apparent amount of years to add to the period.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-hour.r
\name{hour}
\alias{hour}
\alias{hour<-}
\title{Get/set hours component of a date-time}
\usage{
hour(x)

hour(x) <- value
}
\arguments{
\item{x}{a date-time object}

\item{value}{numeric value to be assigned to the \code{hour} component}
}
\value{
the hours element of x as a decimal number
}
\description{
Date-time must be a POSIXct, POSIXlt, Date, Period, chron, yearmon, yearqtr, zoo,
zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
}
\examples{
x <- ymd("2012-03-26")
hour(x)
hour(x) <- 1
hour(x) <- 25
hour(x) > 2
}
\keyword{chron}
\keyword{manip}
\keyword{methods}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/intervals.r
\docType{class}
\name{Interval-class}
\alias{Interval-class}
\alias{intervals}
\title{Interval class}
\description{
Interval is an S4 class that extends the \linkS4class{Timespan} class. An
Interval object records one or more spans of time. Intervals record these
timespans as a sequence of seconds that begin at a specified date. Since
intervals are anchored to a precise moment of time, they can accurately be
converted to \linkS4class{Period} or \linkS4class{Duration} class objects. This
is because we can observe the length in seconds of each period that begins on a
specific date. Contrast this to a generalized period, which may not have a
consistent length in seconds (e.g. the number of seconds in a year will change
if it is a leap year).
}
\details{
Intervals can be both negative and positive. Negative intervals progress
backwards from the start date; positive intervals progress forwards.

Interval class objects have two slots: .Data, a numeric object equal to the number
of seconds in the interval; and start, a POSIXct object that specifies the time
when the interval starts.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercion.r
\name{reclass_date}
\alias{reclass_date}
\title{Convenience method to reclass dates post-modification.}
\usage{
reclass_date(new, orig)
}
\description{
Convenience method to reclass dates post-modification.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercion.r
\name{as.period}
\alias{as.period}
\alias{as.period,numeric-method}
\alias{as.period,difftime-method}
\alias{as.period,Interval-method}
\alias{as.period,Duration-method}
\alias{as.period,Period-method}
\alias{as.period,logical-method}
\alias{as.period,character-method}
\title{Change an object to a period}
\usage{
as.period(x, unit, ...)
}
\arguments{
\item{x}{an interval, difftime, or numeric object}

\item{unit}{A character string that specifies which time units to build period in.
unit is only implemented for the as.period.numeric method and the as.period.interval method.
For as.period.interval, as.period will convert intervals to units no larger than the specified
unit.}

\item{...}{additional arguments to pass to as.period}
}
\value{
a period object
}
\description{
as.period changes Interval, Duration, difftime and numeric class objects
to Period class objects with the specified units.
}
\details{
Users must specify which time units to measure the period in. The exact length of
each time unit in a period will depend on when it occurs. See
\linkS4class{Period} and \code{\link[=period]{period()}}.
The choice of units is not trivial; units that are
normally equal may differ in length depending on when the time period
occurs. For example, when a leap second occurs one minute is longer than 60
seconds.

Because periods do not have a fixed length, they can not be accurately
converted to and from Duration objects. Duration objects measure time spans
in exact numbers of seconds, see \linkS4class{Duration}. Hence, a one to one
mapping does not exist between durations and periods. When used with a
Duration object, as.period provides an inexact estimate; the duration is
broken into time units based on the most common lengths of time units, in
seconds. Because the length of months are particularly variable, a period
with a months unit can not be coerced from a duration object. For an exact
transformation, first transform the duration to an interval with
\code{\link[=as.interval]{as.interval()}}.

Coercing an interval to a period may cause surprising behavior if you request
periods with small units. A leap year is 366 days long, but one year long. Such
an interval will convert to 366 days when unit is set to days and 1 year when
unit is set to years. Adding 366 days to a date will often give a different
result than adding one year. Daylight savings is the one exception where this
does not apply. Interval lengths are calculated on the UTC timeline, which does
not use daylight savings. Hence, periods converted with seconds or minutes will not
reflect the actual variation in seconds and minutes that occurs due to daylight
savings. These periods will show the "naive" change in seconds and minutes that is
suggested by the differences in clock time. See the examples below.
}
\examples{
span <- interval(ymd_hms("2009-01-01 00:00:00"), ymd_hms("2010-02-02 01:01:01")) #interval
as.period(span)
as.period(span, unit = "day")
"397d 1H 1M 1S"
leap <- interval(ymd("2016-01-01"), ymd("2017-01-01"))
as.period(leap, unit = "days")
as.period(leap, unit = "years")
dst <- interval(ymd("2016-11-06", tz = "America/Chicago"),
ymd("2016-11-07", tz = "America/Chicago"))
# as.period(dst, unit = "seconds")
as.period(dst, unit = "hours")
per <- period(hours = 10, minutes = 6)
as.numeric(per, "hours")
as.numeric(per, "minutes")

}
\seealso{
\linkS4class{Period}, \code{\link[=period]{period()}}
}
\keyword{chron}
\keyword{classes}
\keyword{manip}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercion.r
\name{as.period}
\alias{as.period}
\alias{as.period,numeric-method}
\alias{as.period,difftime-method}
\alias{as.period,Interval-method}
\alias{as.period,Duration-method}
\alias{as.period,Period-method}
\alias{as.period,logical-method}
\alias{as.period,character-method}
\title{Change an object to a period}
\usage{
as.period(x, unit, ...)
}
\arguments{
\item{x}{an interval, difftime, or numeric object}

\item{unit}{A character string that specifies which time units to build period in.
unit is only implemented for the as.period.numeric method and the as.period.interval method.
For as.period.interval, as.period will convert intervals to units no larger than the specified
unit.}

\item{...}{additional arguments to pass to as.period}
}
\value{
a period object
}
\description{
as.period changes Interval, Duration, difftime and numeric class objects
to Period class objects with the specified units.
}
\details{
Users must specify which time units to measure the period in. The exact length of
each time unit in a period will depend on when it occurs. See
\linkS4class{Period} and \code{\link[=period]{period()}}.
The choice of units is not trivial; units that are
normally equal may differ in length depending on when the time period
occurs. For example, when a leap second occurs one minute is longer than 60
seconds.

Because periods do not have a fixed length, they can not be accurately
converted to and from Duration objects. Duration objects measure time spans
in exact numbers of seconds, see \linkS4class{Duration}. Hence, a one to one
mapping does not exist between durations and periods. When used with a
Duration object, as.period provides an inexact estimate; the duration is
broken into time units based on the most common lengths of time units, in
seconds. Because the length of months are particularly variable, a period
with a months unit can not be coerced from a duration object. For an exact
transformation, first transform the duration to an interval with
\code{\link[=as.interval]{as.interval()}}.

Coercing an interval to a period may cause surprising behavior if you request
periods with small units. A leap year is 366 days long, but one year long. Such
an interval will convert to 366 days when unit is set to days and 1 year when
unit is set to years. Adding 366 days to a date will often give a different
result than adding one year. Daylight savings is the one exception where this
does not apply. Interval lengths are calculated on the UTC timeline, which does
not use daylight savings. Hence, periods converted with seconds or minutes will not
reflect the actual variation in seconds and minutes that occurs due to daylight
savings. These periods will show the "naive" change in seconds and minutes that is
suggested by the differences in clock time. See the examples below.
}
\examples{
span <- interval(ymd_hms("2009-01-01 00:00:00"), ymd_hms("2010-02-02 01:01:01")) # interval
as.period(span)
as.period(span, unit = "day")
"397d 1H 1M 1S"
leap <- interval(ymd("2016-01-01"), ymd("2017-01-01"))
as.period(leap, unit = "days")
as.period(leap, unit = "years")
dst <- interval(
  ymd("2016-11-06", tz = "America/Chicago"),
  ymd("2016-11-07", tz = "America/Chicago")
)
# as.period(dst, unit = "seconds")
as.period(dst, unit = "hours")
per <- period(hours = 10, minutes = 6)
as.numeric(per, "hours")
as.numeric(per, "minutes")
}
\seealso{
\linkS4class{Period}, \code{\link[=period]{period()}}
}
\keyword{chron}
\keyword{classes}
\keyword{manip}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.r
\docType{package}
\name{lubridate-package}
\alias{lubridate}
\alias{lubridate-package}
\title{Dates and times made easy with lubridate}
\description{
Lubridate provides tools that make it easier to parse and
manipulate dates. These tools are grouped below by common
purpose. More information about each function can be found in
its help documentation.
}
\section{Parsing dates}{


Lubridate's parsing functions read strings into R as POSIXct
date-time objects. Users should choose the function whose name
models the order in which the year ('y'), month ('m') and day
('d') elements appear the string to be parsed:
\code{\link[=dmy]{dmy()}}, \code{\link[=myd]{myd()}}, \code{\link[=ymd]{ymd()}},
\code{\link[=ydm]{ydm()}}, \code{\link[=dym]{dym()}}, \code{\link[=mdy]{mdy()}},
\code{\link[=ymd_hms]{ymd_hms()}}). A very flexible and user friendly parser
is provided by \code{\link[=parse_date_time]{parse_date_time()}}.

Lubridate can also parse partial dates from strings into
\linkS4class{Period} objects with the functions
\code{\link[=hm]{hm()}}, \code{\link[=hms]{hms()}} and \code{\link[=ms]{ms()}}.

Lubridate has an inbuilt very fast POSIX parser. Most of the \code{\link[=strptime]{strptime()}}
formats and various extensions are supported for English locales. See
\code{\link[=parse_date_time]{parse_date_time()}} for more details.
}

\section{Manipulating dates}{


Lubridate distinguishes between moments in time (known as
\code{\link[=instants]{instants()}}) and spans of time (known as time spans, see
\linkS4class{Timespan}). Time spans are further separated into
\linkS4class{Duration}, \linkS4class{Period} and
\linkS4class{Interval} objects.
}

\section{Instants}{


Instants are specific moments of time. Date, POSIXct, and
POSIXlt are the three object classes Base R recognizes as
instants. \code{\link[=is.Date]{is.Date()}} tests whether an object
inherits from the Date class. \code{\link[=is.POSIXt]{is.POSIXt()}} tests
whether an object inherits from the POSIXlt or POSIXct classes.
\code{\link[=is.instant]{is.instant()}} tests whether an object inherits from
any of the three classes.

\code{\link[=now]{now()}} returns the current system time as a POSIXct
object. \code{\link[=today]{today()}} returns the current system date.
For convenience, 1970-01-01 00:00:00 is saved to
\link{origin}. This is the instant from which POSIXct
times are calculated. Try \code{unclass(now())} to see the numeric structure that
underlies POSIXct objects. Each POSIXct object is saved as the number of seconds
it occurred after 1970-01-01 00:00:00.

Conceptually, instants are a combination of measurements on different units
(i.e, years, months, days, etc.). The individual values for
these units can be extracted from an instant and set with the
accessor functions \code{\link[=second]{second()}}, \code{\link[=minute]{minute()}},
\code{\link[=hour]{hour()}}, \code{\link[=day]{day()}}, \code{\link[=yday]{yday()}},
\code{\link[=mday]{mday()}}, \code{\link[=wday]{wday()}}, \code{\link[=week]{week()}},
\code{\link[=month]{month()}}, \code{\link[=year]{year()}}, \code{\link[=tz]{tz()}},
and \code{\link[=dst]{dst()}}.
Note: the accessor functions are named after the singular form
of an element. They shouldn't be confused with the period
helper functions that have the plural form of the units as a
name (e.g, \code{\link[=seconds]{seconds()}}).
}

\section{Rounding dates}{


Instants can be rounded to a convenient unit using the
functions \code{\link[=ceiling_date]{ceiling_date()}}, \code{\link[=floor_date]{floor_date()}}
and \code{\link[=round_date]{round_date()}}.
}

\section{Time zones}{


Lubridate provides two helper functions for working with time
zones. \code{\link[=with_tz]{with_tz()}} changes the time zone in which an
instant is displayed. The clock time displayed for the instant
changes, but the moment of time described remains the same.
\code{\link[=force_tz]{force_tz()}} changes only the time zone element of an
instant. The clock time displayed remains the same, but the
resulting instant describes a new moment of time.
}

\section{Timespans}{


A timespan is a length of time that may or may not be connected to
a particular instant. For example, three months is a timespan. So is an hour and
a half. Base R uses difftime class objects to record timespans.
However, people are not always consistent in how they expect time to behave.
Sometimes the passage of time is a monotone progression of instants that should
be as mathematically reliable as the number line. On other occasions time must
follow complex conventions and rules so that the clock times we see reflect what
we expect to observe in terms of daylight, season, and congruence with the
atomic clock. To better navigate the nuances of time, \pkg{lubridate} creates three
additional timespan classes, each with its own specific and consistent behavior:
\linkS4class{Interval}, \linkS4class{Period} and
\linkS4class{Duration}.

\code{\link[=is.difftime]{is.difftime()}} tests whether an object
inherits from the difftime class. \code{\link[=is.timespan]{is.timespan()}}
tests whether an object inherits from any of the four timespan
classes.
}

\section{Durations}{


Durations measure the exact amount of time that occurs between two
instants. This can create unexpected results in relation to clock times if a
leap second, leap year, or change in daylight savings time (DST) occurs in
the interval.

Functions for working with durations include \code{\link[=is.duration]{is.duration()}},
\code{\link[=as.duration]{as.duration()}} and \code{\link[=duration]{duration()}}. \code{\link[=dseconds]{dseconds()}},
\code{\link[=dminutes]{dminutes()}}, \code{\link[=dhours]{dhours()}},  \code{\link[=ddays]{ddays()}},
\code{\link[=dweeks]{dweeks()}} and \code{\link[=dyears]{dyears()}} convenient lengths.
}

\section{Periods}{


Periods measure the change in clock time that occurs between two
instants. Periods provide robust predictions of clock time in the presence of
leap seconds, leap years, and changes in DST.

Functions for working with periods include
\code{\link[=is.period]{is.period()}}, \code{\link[=as.period]{as.period()}} and
\code{\link[=period]{period()}}. \code{\link[=seconds]{seconds()}},
\code{\link[=minutes]{minutes()}}, \code{\link[=hours]{hours()}}, \code{\link[=days]{days()}},
\code{\link[=weeks]{weeks()}}, \code{\link[=months]{months()}} and
\code{\link[=years]{years()}} quickly create periods of convenient
lengths.
}

\section{Intervals}{


Intervals are timespans that begin at a specific instant and
end at a specific instant. Intervals retain complete information about a
timespan. They provide the only reliable way to convert between
periods and durations.

Functions for working with intervals include
\code{\link[=is.interval]{is.interval()}}, \code{\link[=as.interval]{as.interval()}},
\code{\link[=interval]{interval()}}, \code{\link[=int_shift]{int_shift()}},
\code{\link[=int_flip]{int_flip()}}, \code{\link[=int_aligns]{int_aligns()}},
\code{\link[=int_overlaps]{int_overlaps()}}, and
\code{\link{\%within\%}}. Intervals can also be manipulated with
intersect, union, and setdiff().
}

\section{Miscellaneous}{


\code{\link[=decimal_date]{decimal_date()}} converts an instant to a decimal of
its year.
\code{\link[=leap_year]{leap_year()}} tests whether an instant occurs during
a leap year.
\code{\link[=pretty_dates]{pretty_dates()}} provides a method of making pretty
breaks for date-times.
\link{lakers} is a data set that contains information
about the Los Angeles Lakers 2008-2009 basketball season.
}

\references{
Garrett Grolemund, Hadley Wickham (2011). Dates and Times Made
Easy with lubridate. Journal of Statistical Software, 40(3), 1-25.
\url{https://www.jstatsoft.org/v40/i03/}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://lubridate.tidyverse.org}
  \item \url{https://github.com/tidyverse/lubridate}
  \item Report bugs at \url{https://github.com/tidyverse/lubridate/issues}
}

}
\author{
\strong{Maintainer}: Vitalie Spinu \email{spinuvit@gmail.com}

Authors:
\itemize{
  \item Garrett Grolemund
  \item Hadley Wickham
}

Other contributors:
\itemize{
  \item Ian Lyttle [contributor]
  \item Imanuel Costigan [contributor]
  \item Jason Law [contributor]
  \item Doug Mitarotonda [contributor]
  \item Joseph Larmarange [contributor]
  \item Jonathan Boiser [contributor]
  \item Chel Hee Lee [contributor]
  \item Google Inc. [contributor, copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.r
\docType{package}
\name{lubridate-package}
\alias{lubridate}
\alias{lubridate-package}
\title{Dates and times made easy with lubridate}
\description{
Lubridate provides tools that make it easier to parse and
manipulate dates. These tools are grouped below by common
purpose. More information about each function can be found in
its help documentation.
}
\section{Parsing dates}{


Lubridate's parsing functions read strings into R as POSIXct
date-time objects. Users should choose the function whose name
models the order in which the year ('y'), month ('m') and day
('d') elements appear the string to be parsed:
\code{\link[=dmy]{dmy()}}, \code{\link[=myd]{myd()}}, \code{\link[=ymd]{ymd()}},
\code{\link[=ydm]{ydm()}}, \code{\link[=dym]{dym()}}, \code{\link[=mdy]{mdy()}},
\code{\link[=ymd_hms]{ymd_hms()}}). A very flexible and user friendly parser
is provided by \code{\link[=parse_date_time]{parse_date_time()}}.

Lubridate can also parse partial dates from strings into
\linkS4class{Period} objects with the functions
\code{\link[=hm]{hm()}}, \code{\link[=hms]{hms()}} and \code{\link[=ms]{ms()}}.

Lubridate has an inbuilt very fast POSIX parser. Most of the \code{\link[=strptime]{strptime()}}
formats and various extensions are supported for English locales. See
\code{\link[=parse_date_time]{parse_date_time()}} for more details.
}

\section{Manipulating dates}{


Lubridate distinguishes between moments in time (known as
\code{\link[=instants]{instants()}}) and spans of time (known as time spans, see
\linkS4class{Timespan}). Time spans are further separated into
\linkS4class{Duration}, \linkS4class{Period} and
\linkS4class{Interval} objects.
}

\section{Instants}{


Instants are specific moments of time. Date, POSIXct, and
POSIXlt are the three object classes Base R recognizes as
instants. \code{\link[=is.Date]{is.Date()}} tests whether an object
inherits from the Date class. \code{\link[=is.POSIXt]{is.POSIXt()}} tests
whether an object inherits from the POSIXlt or POSIXct classes.
\code{\link[=is.instant]{is.instant()}} tests whether an object inherits from
any of the three classes.

\code{\link[=now]{now()}} returns the current system time as a POSIXct
object. \code{\link[=today]{today()}} returns the current system date.
For convenience, 1970-01-01 00:00:00 is saved to
\link{origin}. This is the instant from which POSIXct
times are calculated. Try \code{unclass(now())} to see the numeric structure that
underlies POSIXct objects. Each POSIXct object is saved as the number of seconds
it occurred after 1970-01-01 00:00:00.

Conceptually, instants are a combination of measurements on different units
(i.e, years, months, days, etc.). The individual values for
these units can be extracted from an instant and set with the
accessor functions \code{\link[=second]{second()}}, \code{\link[=minute]{minute()}},
\code{\link[=hour]{hour()}}, \code{\link[=day]{day()}}, \code{\link[=yday]{yday()}},
\code{\link[=mday]{mday()}}, \code{\link[=wday]{wday()}}, \code{\link[=week]{week()}},
\code{\link[=month]{month()}}, \code{\link[=year]{year()}}, \code{\link[=tz]{tz()}},
and \code{\link[=dst]{dst()}}.
Note: the accessor functions are named after the singular form
of an element. They shouldn't be confused with the period
helper functions that have the plural form of the units as a
name (e.g, \code{\link[=seconds]{seconds()}}).
}

\section{Rounding dates}{


Instants can be rounded to a convenient unit using the
functions \code{\link[=ceiling_date]{ceiling_date()}}, \code{\link[=floor_date]{floor_date()}}
and \code{\link[=round_date]{round_date()}}.
}

\section{Time zones}{


Lubridate provides two helper functions for working with time
zones. \code{\link[=with_tz]{with_tz()}} changes the time zone in which an
instant is displayed. The clock time displayed for the instant
changes, but the moment of time described remains the same.
\code{\link[=force_tz]{force_tz()}} changes only the time zone element of an
instant. The clock time displayed remains the same, but the
resulting instant describes a new moment of time.
}

\section{Timespans}{


A timespan is a length of time that may or may not be connected to
a particular instant. For example, three months is a timespan. So is an hour and
a half. Base R uses difftime class objects to record timespans.
However, people are not always consistent in how they expect time to behave.
Sometimes the passage of time is a monotone progression of instants that should
be as mathematically reliable as the number line. On other occasions time must
follow complex conventions and rules so that the clock times we see reflect what
we expect to observe in terms of daylight, season, and congruence with the
atomic clock. To better navigate the nuances of time, \pkg{lubridate} creates three
additional timespan classes, each with its own specific and consistent behavior:
\linkS4class{Interval}, \linkS4class{Period} and
\linkS4class{Duration}.

\code{\link[=is.difftime]{is.difftime()}} tests whether an object
inherits from the difftime class. \code{\link[=is.timespan]{is.timespan()}}
tests whether an object inherits from any of the four timespan
classes.
}

\section{Durations}{


Durations measure the exact amount of time that occurs between two
instants. This can create unexpected results in relation to clock times if a
leap second, leap year, or change in daylight savings time (DST) occurs in
the interval.

Functions for working with durations include \code{\link[=is.duration]{is.duration()}},
\code{\link[=as.duration]{as.duration()}} and \code{\link[=duration]{duration()}}. \code{\link[=dseconds]{dseconds()}},
\code{\link[=dminutes]{dminutes()}}, \code{\link[=dhours]{dhours()}},  \code{\link[=ddays]{ddays()}},
\code{\link[=dweeks]{dweeks()}} and \code{\link[=dyears]{dyears()}} convenient lengths.
}

\section{Periods}{


Periods measure the change in clock time that occurs between two
instants. Periods provide robust predictions of clock time in the presence of
leap seconds, leap years, and changes in DST.

Functions for working with periods include
\code{\link[=is.period]{is.period()}}, \code{\link[=as.period]{as.period()}} and
\code{\link[=period]{period()}}. \code{\link[=seconds]{seconds()}},
\code{\link[=minutes]{minutes()}}, \code{\link[=hours]{hours()}}, \code{\link[=days]{days()}},
\code{\link[=weeks]{weeks()}}, \code{\link[=months]{months()}} and
\code{\link[=years]{years()}} quickly create periods of convenient
lengths.
}

\section{Intervals}{


Intervals are timespans that begin at a specific instant and
end at a specific instant. Intervals retain complete information about a
timespan. They provide the only reliable way to convert between
periods and durations.

Functions for working with intervals include
\code{\link[=is.interval]{is.interval()}}, \code{\link[=as.interval]{as.interval()}},
\code{\link[=interval]{interval()}}, \code{\link[=int_shift]{int_shift()}},
\code{\link[=int_flip]{int_flip()}}, \code{\link[=int_aligns]{int_aligns()}},
\code{\link[=int_overlaps]{int_overlaps()}}, and
\code{\link{\%within\%}}. Intervals can also be manipulated with
intersect, union, and setdiff().
}

\section{Miscellaneous}{


\code{\link[=decimal_date]{decimal_date()}} converts an instant to a decimal of
its year.
\code{\link[=leap_year]{leap_year()}} tests whether an instant occurs during
a leap year.
\code{\link[=pretty_dates]{pretty_dates()}} provides a method of making pretty
breaks for date-times.
\link{lakers} is a data set that contains information
about the Los Angeles Lakers 2008-2009 basketball season.
}

\references{
Garrett Grolemund, Hadley Wickham (2011). Dates and Times Made
Easy with lubridate. Journal of Statistical Software, 40(3), 1-25.
\url{https://www.jstatsoft.org/v40/i03/}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://lubridate.tidyverse.org}
  \item \url{https://github.com/tidyverse/lubridate}
  \item Report bugs at \url{https://github.com/tidyverse/lubridate/issues}
}

}
\author{
\strong{Maintainer}: Vitalie Spinu \email{spinuvit@gmail.com}

Authors:
\itemize{
  \item Garrett Grolemund
  \item Hadley Wickham
}

Other contributors:
\itemize{
  \item Davis Vaughan [contributor]
  \item Ian Lyttle [contributor]
  \item Imanuel Costigan [contributor]
  \item Jason Law [contributor]
  \item Doug Mitarotonda [contributor]
  \item Joseph Larmarange [contributor]
  \item Jonathan Boiser [contributor]
  \item Chel Hee Lee [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-week.r
\name{week}
\alias{week}
\alias{week<-}
\alias{isoweek}
\alias{epiweek}
\title{Get/set weeks component of a date-time}
\usage{
week(x)

week(x) <- value

isoweek(x)

epiweek(x)
}
\arguments{
\item{x}{a date-time object. Must be a POSIXct, POSIXlt, Date, chron,
yearmon, yearqtr, zoo, zooreg, timeDate, xts, its, ti, jul, timeSeries, or
fts object.}

\item{value}{a numeric object}
}
\value{
the weeks element of x as an integer number
}
\description{
\code{week()} returns the number of complete seven day periods that have
occurred between the date and January 1st, plus one.

\code{isoweek()} returns the week as it would appear in the ISO 8601
system, which uses a reoccurring leap week.

\code{epiweek()} is the US CDC version of epidemiological week. It
follows same rules as \code{isoweek()} but starts on Sunday. In other parts of
the world the convention is to start epidemiological weeks on Monday,
which is the same as \code{isoweek}.
}
\examples{
x <- ymd("2012-03-26")
week(x)
week(x) <- 1
week(x) <- 54
week(x) > 3
}
\references{
\url{https://en.wikipedia.org/wiki/ISO_week_date}
}
\seealso{
\code{\link[=isoyear]{isoyear()}}
}
\keyword{chron}
\keyword{manip}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.r
\name{Deprecated-lubridate}
\alias{Deprecated-lubridate}
\title{Deprecated functions in the lubridate package}
\arguments{
\item{x}{numeric value to be converted into duration}

\item{...}{arguments to be passed to the functions
(obscured to enforce the usage of new functions)}
}
\description{
Deprecated functions in the lubridate package
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/instants.r
\name{make_datetime}
\alias{make_datetime}
\alias{make_date}
\title{Efficient creation of date-times from numeric representations}
\usage{
make_datetime(
  year = 1970L,
  month = 1L,
  day = 1L,
  hour = 0L,
  min = 0L,
  sec = 0,
  tz = "UTC"
)

make_date(year = 1970L, month = 1L, day = 1L)
}
\arguments{
\item{year}{numeric year}

\item{month}{numeric month}

\item{day}{numeric day}

\item{hour}{numeric hour}

\item{min}{numeric minute}

\item{sec}{numeric second}

\item{tz}{time zone. Defaults to UTC.}
}
\description{
\code{make_datetime()} is a very fast drop-in replacement for
\code{\link[base:ISOdatetime]{base::ISOdate()}} and \code{\link[base:ISOdatetime]{base::ISOdatetime()}}. \code{make_date()} produces
objects of class \code{Date}.
}
\details{
Input vectors are silently recycled. All inputs except \code{sec} are
silently converted to integer vectors; \code{sec} can be either integer or
double.
}
\examples{
make_datetime(year = 1999, month = 12, day = 22, sec = 10)
make_datetime(year = 1999, month = 12, day = 22, sec = c(10, 11))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/periods.r
\name{period}
\alias{period}
\alias{periods}
\alias{is.period}
\alias{seconds}
\alias{minutes}
\alias{hours}
\alias{days}
\alias{weeks}
\alias{years}
\alias{milliseconds}
\alias{microseconds}
\alias{nanoseconds}
\alias{picoseconds}
\alias{months.numeric}
\title{Create or parse period objects}
\usage{
period(num = NULL, units = "second", ...)

is.period(x)

seconds(x = 1)

minutes(x = 1)

hours(x = 1)

days(x = 1)

weeks(x = 1)

years(x = 1)

milliseconds(x = 1)

microseconds(x = 1)

nanoseconds(x = 1)

picoseconds(x = 1)

\method{months}{numeric}(x, abbreviate)
}
\arguments{
\item{num}{a numeric or character vector. A character vector can specify
periods in a convenient shorthand format or ISO 8601 specification. All
unambiguous name units and abbreviations are supported, "m" stands for
months, "M" for minutes unless ISO 8601 "P" modifier is present (see
examples). Fractional units are supported but the fractional part is always
converted to seconds.}

\item{units}{a character vector that lists the type of units to be used. The
units in units are matched to the values in num according to their
order. When \code{num} is character, this argument is ignored.}

\item{...}{a list of time units to be included in the period and their
amounts. Seconds, minutes,  hours, days, weeks, months, and years are
supported. Normally only one of \code{num} or \code{...} are present. If both are
present, the periods are concatenated.}

\item{x}{Any R object for \code{is.periods} and a numeric value of the number of
units for elementary constructors. With the exception of seconds(), x must
be an integer.}

\item{abbreviate}{Ignored. For consistency with S3 generic in base namespace.}
}
\value{
a period object
}
\description{
\code{period()} creates or parses a period object with the specified values.
}
\details{
Within a Period object, time units do not have a fixed length (except for
seconds) until they are added to a date-time. The length of each time unit
will depend on the date-time to which it is added. For example, a year that
begins on 2009-01-01 will be 365 days long.  A year that begins on 2012-01-01
will be 366 days long. When math is performed with a period object, each unit
is applied separately. How the length of a period is distributed among its
units is non-trivial. For example, when leap seconds occur 1 minute is longer
than 60 seconds.

Periods track the change in the "clock time" between two date-times. They
are measured in common time related units: years, months, days, hours,
minutes, and seconds. Each unit except for seconds must be expressed in
integer values.

Besides the main constructor and parser \code{\link[=period]{period()}}, period objects can also
be created with the specialized functions \code{\link[=years]{years()}}, \code{\link[=months]{months()}}, \code{\link[=weeks]{weeks()}},
\code{\link[=days]{days()}}, \code{\link[=hours]{hours()}}, \code{\link[=minutes]{minutes()}}, and \code{\link[=seconds]{seconds()}}. These objects can be added
to and subtracted to date-times to create a user interface similar to object
oriented programming.

Note: Arithmetic with periods can result in undefined behavior when
non-existent dates are involved (such as February 29th in non-leap years).
Please see \linkS4class{Period} for more details and \code{\link{\%m+\%}} and
\code{\link[=add_with_rollback]{add_with_rollback()}} for alternative operations.
}
\examples{

### Separate period and units vectors

period(c(90, 5), c("second", "minute"))
#  "5M 90S"
period(-1, "days")
period(c(3, 1, 2, 13, 1), c("second", "minute", "hour", "day", "week"))
period(c(1, -60), c("hour", "minute"))
period(0, "second")

### Units as arguments

period(second = 90, minute = 5)
period(day = -1)
period(second = 3, minute = 1, hour = 2, day = 13, week = 1)
period(hour = 1, minute = -60)
period(second = 0)
period(c(1, -60), c("hour", "minute"), hour = c(1, 2), minute = c(3, 4))

### Lubridate style parsing

period("2M 1sec")
period("2hours 2minutes 1second")
period("2d 2H 2M 2S")
period("2days 2hours 2mins 2secs")
period("2 days, 2 hours, 2 mins, 2 secs")
# Missing numerals default to 1. Repeated units are added up.
duration("day day")

### ISO 8601 parsing

period("P10M23DT23H") # M stands for months
period("10DT10M") # M stands for minutes
period("P3Y6M4DT12H30M5S") # M for both minutes and months
period("P23DT60H 20min 100 sec") # mixing ISO and lubridate style parsing

### Comparison with characters (from v1.6.0)

duration("day 2 sec") > "day 1sec"

### Elementary Constructors

x <- ymd("2009-08-03")
x + days(1) + hours(6) + minutes(30)
x + days(100) - hours(8)

class(as.Date("2009-08-09") + days(1)) # retains Date class
as.Date("2009-08-09") + hours(12)
class(as.Date("2009-08-09") + hours(12))
# converts to POSIXt class to accomodate time units

years(1) - months(7)
c(1:3) * hours(1)
hours(1:3)

# sequencing
y <- ymd(090101) # "2009-01-01 CST"
y + months(0:11)

# compare DST handling to durations
boundary <- ymd_hms("2009-03-08 01:59:59", tz = "America/Chicago")
boundary + days(1) # period
boundary + ddays(1) # duration
is.period(as.Date("2009-08-03")) # FALSE
is.period(period(months = 1, days = 15)) # TRUE
}
\seealso{
\linkS4class{Period}, \code{\link[=period]{period()}}, \code{\link{\%m+\%}},
\code{\link[=add_with_rollback]{add_with_rollback()}}
}
\keyword{chron}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/periods.r
\name{period}
\alias{period}
\alias{periods}
\alias{is.period}
\alias{seconds}
\alias{minutes}
\alias{hours}
\alias{days}
\alias{weeks}
\alias{years}
\alias{milliseconds}
\alias{microseconds}
\alias{nanoseconds}
\alias{picoseconds}
\alias{months.numeric}
\title{Create or parse period objects}
\usage{
period(num = NULL, units = "second", ...)

is.period(x)

seconds(x = 1)

minutes(x = 1)

hours(x = 1)

days(x = 1)

weeks(x = 1)

years(x = 1)

milliseconds(x = 1)

microseconds(x = 1)

nanoseconds(x = 1)

picoseconds(x = 1)

\method{months}{numeric}(x, abbreviate)
}
\arguments{
\item{num}{a numeric or character vector. A character vector can specify
periods in a convenient shorthand format or ISO 8601 specification. All
unambiguous name units and abbreviations are supported, "m" stands for
months, "M" for minutes unless ISO 8601 "P" modifier is present (see
examples). Fractional units are supported but the fractional part is always
converted to seconds.}

\item{units}{a character vector that lists the type of units to be used. The
units in units are matched to the values in num according to their
order. When \code{num} is character, this argument is ignored.}

\item{...}{a list of time units to be included in the period and their
amounts. Seconds, minutes,  hours, days, weeks, months, and years are
supported. Normally only one of \code{num} or \code{...} are present. If both are
present, the periods are concatenated.}

\item{x}{Any R object for \code{is.periods} and a numeric value of the number of
units for elementary constructors. With the exception of seconds(), x must
be an integer.}

\item{abbreviate}{Ignored. For consistency with S3 generic in base namespace.}
}
\value{
a period object
}
\description{
\code{period()} creates or parses a period object with the specified values.
}
\details{
Within a Period object, time units do not have a fixed length (except for
seconds) until they are added to a date-time. The length of each time unit
will depend on the date-time to which it is added. For example, a year that
begins on 2009-01-01 will be 365 days long.  A year that begins on 2012-01-01
will be 366 days long. When math is performed with a period object, each unit
is applied separately. How the length of a period is distributed among its
units is non-trivial. For example, when leap seconds occur 1 minute is longer
than 60 seconds.

Periods track the change in the "clock time" between two date-times. They
are measured in common time related units: years, months, days, hours,
minutes, and seconds. Each unit except for seconds must be expressed in
integer values.

Besides the main constructor and parser \code{\link[=period]{period()}}, period objects can also
be created with the specialized functions \code{\link[=years]{years()}}, \code{\link[=months]{months()}}, \code{\link[=weeks]{weeks()}},
\code{\link[=days]{days()}}, \code{\link[=hours]{hours()}}, \code{\link[=minutes]{minutes()}}, and \code{\link[=seconds]{seconds()}}. These objects can be added
to and subtracted to date-times to create a user interface similar to object
oriented programming.

Note: Arithmetic with periods can result in undefined behavior when
non-existent dates are involved (such as February 29th in non-leap years).
Please see \linkS4class{Period} for more details and \code{\link{\%m+\%}} and
\code{\link[=add_with_rollback]{add_with_rollback()}} for alternative operations.
}
\examples{

### Separate period and units vectors

period(c(90, 5), c("second", "minute"))
#  "5M 90S"
period(-1, "days")
period(c(3, 1, 2, 13, 1), c("second", "minute", "hour", "day", "week"))
period(c(1, -60), c("hour", "minute"))
period(0, "second")

### Units as arguments

period (second = 90, minute = 5)
period(day = -1)
period(second = 3, minute = 1, hour = 2, day = 13, week = 1)
period(hour = 1, minute = -60)
period(second = 0)
period(c(1, -60), c("hour", "minute"), hour = c(1, 2), minute = c(3, 4))

### Lubridate style parsing

period("2M 1sec")
period("2hours 2minutes 1second")
period("2d 2H 2M 2S")
period("2days 2hours 2mins 2secs")
period("2 days, 2 hours, 2 mins, 2 secs")
# Missing numerals default to 1. Repeated units are added up.
duration("day day")

### ISO 8601 parsing

period("P10M23DT23H") # M stands for months
period("10DT10M") # M stands for minutes
period("P3Y6M4DT12H30M5S") # M for both minutes and months
period("P23DT60H 20min 100 sec") # mixing ISO and lubridate style parsing

### Comparison with characters (from v1.6.0)

duration("day 2 sec") > "day 1sec"

### Elementary Constructors

x <- ymd("2009-08-03")
x + days(1) + hours(6) + minutes(30)
x + days(100) - hours(8)

class(as.Date("2009-08-09") + days(1)) # retains Date class
as.Date("2009-08-09") + hours(12)
class(as.Date("2009-08-09") + hours(12))
# converts to POSIXt class to accomodate time units

years(1) - months(7)
c(1:3) * hours(1)
hours(1:3)

# sequencing
y <- ymd(090101) # "2009-01-01 CST"
y + months(0:11)

# compare DST handling to durations
boundary <- ymd_hms("2009-03-08 01:59:59", tz="America/Chicago")
boundary + days(1) # period
boundary + ddays(1) # duration
is.period(as.Date("2009-08-03")) # FALSE
is.period(period(months= 1, days = 15)) # TRUE
}
\seealso{
\linkS4class{Period}, \code{\link[=period]{period()}}, \code{\link{\%m+\%}},
\code{\link[=add_with_rollback]{add_with_rollback()}}
}
\keyword{chron}
\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/round.r
\name{round_date}
\alias{round_date}
\alias{floor_date}
\alias{ceiling_date}
\title{Round, floor and ceiling methods for date-time objects}
\usage{
round_date(
  x,
  unit = "second",
  week_start = getOption("lubridate.week.start", 7)
)

floor_date(
  x,
  unit = "seconds",
  week_start = getOption("lubridate.week.start", 7)
)

ceiling_date(
  x,
  unit = "seconds",
  change_on_boundary = NULL,
  week_start = getOption("lubridate.week.start", 7)
)
}
\arguments{
\item{x}{a vector of date-time objects}

\item{unit}{a character string specifying a time unit or a multiple of a unit
to be rounded to. Valid base units are \code{second}, \code{minute}, \code{hour}, \code{day},
\code{week}, \code{month}, \code{bimonth}, \code{quarter}, \code{season}, \code{halfyear} and
\code{year}. Arbitrary unique English abbreviations as in the \code{\link[=period]{period()}}
constructor are allowed. Rounding to multiples of units (except weeks) is
supported.}

\item{week_start}{when unit is \code{weeks}, specify the reference day.
7 represents Sunday and 1 represents Monday.}

\item{change_on_boundary}{if this is \code{NULL} (the default), instants on the boundary
remain unchanged, but \code{Date} objects are rounded up to the next boundary.
If this is \code{TRUE}, instants on the boundary are rounded up to the next boundary.
If this is \code{FALSE}, nothing on the boundary is rounded up at all. This was the
default for \pkg{lubridate} prior to \code{v1.6.0}.
See section \verb{Rounding Up Date Objects} below for more details.}
}
\description{
\code{round_date()} takes a date-time object and time unit, and rounds it to the nearest value
of the specified time unit. For rounding date-times which are exactly halfway
between two consecutive units, the convention is to round up. Note that this
is in line with the behavior of R's \code{\link[base:round.POSIXt]{base::round.POSIXt()}} function
but does not follow the convention of the base \code{\link[base:Round]{base::round()}} function
which "rounds to the even digit", as per IEC 60559.

Rounding to the nearest unit or multiple of a unit is supported. All
meaningful specifications in the English language are supported - secs, min,
mins, 2 minutes, 3 years etc.

Rounding to fractional seconds is also supported. Please note that rounding to
fractions smaller than 1 second can lead to large precision errors due to the
floating point representation of the POSIXct objects. See examples.

\code{floor_date()} takes a date-time object and rounds it down to the nearest
boundary of the specified time unit.

\code{ceiling_date()} takes a date-time object and rounds it up to the nearest
boundary of the specified time unit.
}
\details{
In \pkg{lubridate}, functions that round date-time objects try to
preserve the class of the input object whenever possible. This is done by
first rounding to an instant, and then converting to the original class as per
usual R conventions.
}
\section{Rounding Up Date Objects}{


By default, rounding up \code{Date} objects follows 3 steps:
\enumerate{
\item Convert to an instant representing lower bound of the Date:
\code{2000-01-01} --> \verb{2000-01-01 00:00:00}
\item Round up to the \strong{next} closest rounding unit boundary. For example,
if the rounding unit is \code{month} then next closest boundary of \code{2000-01-01}
is \verb{2000-02-01 00:00:00}.

The motivation for this is that the "partial" \code{2000-01-01} is conceptually
an interval (\verb{2000-01-01 00:00:00} -- \verb{2000-01-02 00:00:00}) and the day
hasn't started clocking yet at the exact boundary \code{00:00:00}. Thus, it
seems wrong to round a day to its lower boundary.

Behavior on the boundary can be changed by setting
\code{change_on_boundary} to \code{TRUE} or \code{FALSE}.
\item If the rounding unit is smaller than a day, return the instant from step 2
(\code{POSIXct}), otherwise convert to and return a \code{Date} object.
}
}

\examples{

## print fractional seconds
options(digits.secs=6)

x <- ymd_hms("2009-08-03 12:01:59.23")
round_date(x, ".5s")
round_date(x, "sec")
round_date(x, "second")
round_date(x, "minute")
round_date(x, "5 mins")
round_date(x, "hour")
round_date(x, "2 hours")
round_date(x, "day")
round_date(x, "week")
round_date(x, "month")
round_date(x, "bimonth")
round_date(x, "quarter") == round_date(x, "3 months")
round_date(x, "halfyear")
round_date(x, "year")

x <- ymd_hms("2009-08-03 12:01:59.23")
floor_date(x, ".1s")
floor_date(x, "second")
floor_date(x, "minute")
floor_date(x, "hour")
floor_date(x, "day")
floor_date(x, "week")
floor_date(x, "month")
floor_date(x, "bimonth")
floor_date(x, "quarter")
floor_date(x, "season")
floor_date(x, "halfyear")
floor_date(x, "year")

x <- ymd_hms("2009-08-03 12:01:59.23")
ceiling_date(x, ".1 sec") # imprecise representation at 0.1 sec !!!
ceiling_date(x, "second")
ceiling_date(x, "minute")
ceiling_date(x, "5 mins")
ceiling_date(x, "hour")
ceiling_date(x, "day")
ceiling_date(x, "week")
ceiling_date(x, "month")
ceiling_date(x, "bimonth") == ceiling_date(x, "2 months")
ceiling_date(x, "quarter")
ceiling_date(x, "season")
ceiling_date(x, "halfyear")
ceiling_date(x, "year")

## As of R 3.4.2 POSIXct printing of fractional numbers is wrong
as.POSIXct("2009-08-03 12:01:59.3") ## -> "2009-08-03 12:01:59.2 CEST"
ceiling_date(x, ".1 sec") ## -> "2009-08-03 12:01:59.2 CEST"

## behaviour of `change_on_boundary`
## As per default behaviour `NULL`, instants on the boundary remain the
## same but dates are rounded up
ceiling_date(ymd_hms("2000-01-01 00:00:00"), "month")
ceiling_date(ymd("2000-01-01"), "month")

## If `TRUE`, both instants and dates on the boundary are rounded up
ceiling_date(ymd_hms("2000-01-01 00:00:00"), "month", change_on_boundary = TRUE)
ceiling_date(ymd("2000-01-01"), "month")

## If `FALSE`, both instants and dates on the boundary remain the same
ceiling_date(ymd_hms("2000-01-01 00:00:00"), "month", change_on_boundary = FALSE)
ceiling_date(ymd("2000-01-01"), "month")


 x <- ymd_hms("2000-01-01 00:00:00")
 ceiling_date(x, "month")
 ceiling_date(x, "month", change_on_boundary = TRUE)

 ## For Date objects first day of the month is not on the
 ## "boundary". change_on_boundary applies to instants only.
 x <- ymd("2000-01-01")
 ceiling_date(x, "month")
 ceiling_date(x, "month", change_on_boundary = TRUE)
}
\seealso{
\code{\link[base:Round]{base::round()}}
}
\keyword{chron}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/round.r
\name{round_date}
\alias{round_date}
\alias{floor_date}
\alias{ceiling_date}
\title{Round, floor and ceiling methods for date-time objects}
\usage{
round_date(
  x,
  unit = "second",
  week_start = getOption("lubridate.week.start", 7)
)

floor_date(
  x,
  unit = "seconds",
  week_start = getOption("lubridate.week.start", 7)
)

ceiling_date(
  x,
  unit = "seconds",
  change_on_boundary = NULL,
  week_start = getOption("lubridate.week.start", 7)
)
}
\arguments{
\item{x}{a vector of date-time objects}

\item{unit}{a string, \code{Period} object or a date-time object. When a singleton string,
it specifies a time unit or a multiple of a unit to be rounded to. Valid base units
are \code{second}, \code{minute}, \code{hour}, \code{day}, \code{week}, \code{month}, \code{bimonth}, \code{quarter},
\code{season}, \code{halfyear} and \code{year}. Arbitrary unique English abbreviations as in the
\code{\link[=period]{period()}} constructor are allowed. Rounding to multiples of units (except weeks)
is supported. When \code{unit} is a \code{Period} object, units of the period objects are
used. This is equivalent to converting the period object to its string
representation and passing as \code{unit} argument.

When \code{unit} is a date-time object rounding is done to the nearest of the
elements in \code{unit}. If range of \code{unit} vector does not cover the range of
\code{x} \code{ceiling_date()} and \code{floor_date()} round to the \code{max(x)} and \code{min(x)}
for elements that fall outside of \code{range(unit)}.}

\item{week_start}{week start day (Default is 7, Sunday. Set \code{lubridate.week.start} to
override). Full or abbreviated names of the days of the week can be in English or
as provided by the current locale.}

\item{change_on_boundary}{if this is \code{NULL} (the default), instants on the
boundary remain unchanged, but \code{Date} objects are rounded up to the next
boundary. If this is \code{TRUE}, instants on the boundary are rounded up to the
next boundary. If this is \code{FALSE}, nothing on the boundary is rounded up at
all. This was the default for \pkg{lubridate} prior to \code{v1.6.0}. See
section \verb{Rounding Up Date Objects} below for more details.}
}
\value{
When \code{unit} is a string, return a Date object if \code{x} is a Date and
\code{unit} is larger or equal than "day", otherwise a POSIXct object. When
\code{unit} is a date-time object, return a date-time object of the same class
and same time zone as \code{unit}.
}
\description{
\code{round_date()} takes a date-time object and time unit, and rounds it to the nearest value
of the specified time unit. For rounding date-times which are exactly halfway
between two consecutive units, the convention is to round up. Note that this
is in line with the behavior of R's \code{\link[base:round.POSIXt]{base::round.POSIXt()}} function
but does not follow the convention of the base \code{\link[base:Round]{base::round()}} function
which "rounds to the even digit", as per IEC 60559.

Rounding to the nearest unit or multiple of a unit is supported. All
meaningful specifications in the English language are supported - secs, min,
mins, 2 minutes, 3 years etc.

Rounding to fractional seconds is also supported. Please note that rounding to
fractions smaller than 1 second can lead to large precision errors due to the
floating point representation of the POSIXct objects. See examples.

\code{floor_date()} takes a date-time object and rounds it down to the nearest
boundary of the specified time unit.

\code{ceiling_date()} takes a date-time object and rounds it up to the nearest
boundary of the specified time unit.
}
\details{
In \pkg{lubridate}, functions that round date-time objects try to
preserve the class of the input object whenever possible. This is done by
first rounding to an instant, and then converting to the original class as per
usual R conventions.
}
\section{Rounding Up Date Objects}{


By default, rounding up \code{Date} objects follows 3 steps:
\enumerate{
\item Convert to an instant representing lower bound of the Date:
\code{2000-01-01} --> \verb{2000-01-01 00:00:00}
\item Round up to the \strong{next} closest rounding unit boundary. For example,
if the rounding unit is \code{month} then next closest boundary of \code{2000-01-01}
is \verb{2000-02-01 00:00:00}.

The motivation for this is that the "partial" \code{2000-01-01} is conceptually
an interval (\verb{2000-01-01 00:00:00} -- \verb{2000-01-02 00:00:00}) and the day
hasn't started clocking yet at the exact boundary \code{00:00:00}. Thus, it
seems wrong to round a day to its lower boundary.

Behavior on the boundary can be changed by setting
\code{change_on_boundary} to \code{TRUE} or \code{FALSE}.
\item If the rounding unit is smaller than a day, return the instant from step 2
(\code{POSIXct}), otherwise convert to and return a \code{Date} object.
}
}

\examples{

## print fractional seconds
options(digits.secs = 6)

x <- ymd_hms("2009-08-03 12:01:59.23")
round_date(x, ".5s")
round_date(x, "sec")
round_date(x, "second")
round_date(x, "minute")
round_date(x, "5 mins")
round_date(x, "hour")
round_date(x, "2 hours")
round_date(x, "day")
round_date(x, "week")
round_date(x, "month")
round_date(x, "bimonth")
round_date(x, "quarter") == round_date(x, "3 months")
round_date(x, "halfyear")
round_date(x, "year")

x <- ymd_hms("2009-08-03 12:01:59.23")
floor_date(x, ".1s")
floor_date(x, "second")
floor_date(x, "minute")
floor_date(x, "hour")
floor_date(x, "day")
floor_date(x, "week")
floor_date(x, "month")
floor_date(x, "bimonth")
floor_date(x, "quarter")
floor_date(x, "season")
floor_date(x, "halfyear")
floor_date(x, "year")

x <- ymd_hms("2009-08-03 12:01:59.23")
ceiling_date(x, ".1 sec") # imprecise representation at 0.1 sec !!!
ceiling_date(x, "second")
ceiling_date(x, "minute")
ceiling_date(x, "5 mins")
ceiling_date(x, "hour")
ceiling_date(x, "day")
ceiling_date(x, "week")
ceiling_date(x, "month")
ceiling_date(x, "bimonth") == ceiling_date(x, "2 months")
ceiling_date(x, "quarter")
ceiling_date(x, "season")
ceiling_date(x, "halfyear")
ceiling_date(x, "year")

## Period unit argument
floor_date(x, days(2))
floor_date(x, years(1))

## As of R 3.4.2 POSIXct printing of fractional numbers is wrong
as.POSIXct("2009-08-03 12:01:59.3") ## -> "2009-08-03 12:01:59.2 CEST"
ceiling_date(x, ".1 sec") ## -> "2009-08-03 12:01:59.2 CEST"

## behaviour of `change_on_boundary`
## As per default behaviour `NULL`, instants on the boundary remain the
## same but dates are rounded up
ceiling_date(ymd_hms("2000-01-01 00:00:00"), "month")
ceiling_date(ymd("2000-01-01"), "month")

## If `TRUE`, both instants and dates on the boundary are rounded up
ceiling_date(ymd_hms("2000-01-01 00:00:00"), "month", change_on_boundary = TRUE)
ceiling_date(ymd("2000-01-01"), "month")

## If `FALSE`, both instants and dates on the boundary remain the same
ceiling_date(ymd_hms("2000-01-01 00:00:00"), "month", change_on_boundary = FALSE)
ceiling_date(ymd("2000-01-01"), "month")

x <- ymd_hms("2000-01-01 00:00:00")
ceiling_date(x, "month")
ceiling_date(x, "month", change_on_boundary = TRUE)

## For Date objects first day of the month is not on the
## "boundary". change_on_boundary applies to instants only.
x <- ymd("2000-01-01")
ceiling_date(x, "month")
ceiling_date(x, "month", change_on_boundary = TRUE)
}
\seealso{
\code{\link[base:Round]{base::round()}}
}
\keyword{chron}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/intervals.r, R/coercion.r
\docType{import}
\name{reexports}
\alias{reexports}
\alias{intersect}
\alias{union}
\alias{setdiff}
\alias{as.difftime}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{generics}{\code{\link[generics:coercion-time-difference]{as.difftime}}, \code{\link[generics:setops]{intersect}}, \code{\link[generics:setops]{setdiff}}, \code{\link[generics:setops]{union}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/intervals.r
\docType{import}
\name{reexports}
\alias{reexports}
\alias{intersect}
\alias{union}
\alias{setdiff}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{generics}{\code{\link[generics:setops]{intersect}}, \code{\link[generics:setops]{setdiff}}, \code{\link[generics:setops]{union}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors-date.R
\name{date}
\alias{date}
\alias{date<-}
\title{Get/set date component of a date-time}
\usage{
date(x)

date(x) <- value
}
\arguments{
\item{x}{a date-time object}

\item{value}{an object for which the \code{date()} function is defined}
}
\value{
the date of x as a Date
}
\description{
Date-time must be a POSIXct, POSIXlt, Date, chron, yearmon, yearqtr, zoo,
zooreg, timeDate, xts, its, ti, jul, timeSeries, and fts objects.
}
\details{
\code{date()} does not yet support years before 0 C.E.
Also \code{date()} is not defined for Period objects.
}
\section{Base compatibility}{

\code{date()} can be called without any arguments to return a string representing
the current date-time. This provides compatiblity with \code{base:date()} which
it overrides.
}

\examples{
x <- ymd_hms("2012-03-26 23:12:13", tz = "America/New_York")
date(x)
as.Date(x) # by default as.Date assumes you want to know the date in UTC
as.Date(x, tz = "America/New_York")
date(x) <- as.Date("2000-01-02")
x
}
\keyword{chron}
\keyword{manip}
\keyword{methods}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-zones.r
\name{force_tz}
\alias{force_tz}
\alias{force_tz.default}
\alias{force_tzs}
\title{Replace time zone to create new date-time}
\usage{
force_tz(time, tzone = "", ...)

\method{force_tz}{default}(time, tzone = "", roll_dst = c("NA", "post"), roll = NULL, ...)

force_tzs(
  time,
  tzones,
  tzone_out = "UTC",
  roll_dst = c("NA", "post"),
  roll = NULL
)
}
\arguments{
\item{time}{a POSIXct, POSIXlt, Date, chron date-time object, or a data.frame
object. When a data.frame all POSIXt elements of a data.frame are processed
with \code{force_tz()} and new data.frame is returned.}

\item{tzone}{a character string containing the time zone to convert to. R
must recognize the name contained in the string as a time zone on your
system.}

\item{...}{Parameters passed to other methods.}

\item{roll_dst}{is a string vector of length one or two. When two values are
supplied they specify how to roll date-times when they fall into "skipped" and
"repeated" DST transitions respectively. A single value is replicated to the
length of two. Possible values are:

\if{html}{\out{<div class="sourceCode">}}\preformatted{* `pre` - Use the time before the transition boundary.
* `boundary` - Use the time exactly at the boundary transition.
* `post` - Use the time after the boundary transition.
* `xfirst` - crossed-first: First time which occurred when crossing the
   boundary. For addition with positive units pre interval is crossed first and
   post interval last. With negative units post interval is crossed first, pre -
   last. For subtraction the logic is reversed.
* `xlast` - crossed-last.
* `NA` - Produce NAs when the resulting time falls inside the problematic interval.
}\if{html}{\out{</div>}}

For example `roll_dst = c("NA", "pre") indicates that for skiped intervals
return NA and for repeated times return the earlier time.

When multiple units are supplied the meaning of "negative period" is determined by
the largest unit. For example \code{time_add(t, days = -1, hours = 2, roll_dst = "xfirst")} would operate as if with negative period, thus crossing the boundary
from the "post" to "pre" side and "xfirst" and hence resolving to "post"
time. As this might result in confusing behavior. See examples.

"xfirst" and "xlast" make sense for addition and subtraction only. An error is
raised if an attempt is made to use them with other functions.}

\item{roll}{deprecated, same as \code{roll_dst} parameter.}

\item{tzones}{character vector of timezones to be "enforced" on \code{time} time
stamps. If \code{time} and \code{tzones} lengths differ, the smaller one is recycled
in accordance with usual R conventions.}

\item{tzone_out}{timezone of the returned date-time vector (for \code{force_tzs}).}
}
\value{
a POSIXct object in the updated time zone
}
\description{
\code{force_tz} returns the date-time that has the same clock time as input time,
but in the new time zone. \code{force_tzs} is the parallel version of \code{force_tz},
meaning that every element from \code{time} argument is matched with the
corresponding time zone in \code{tzones} argument.
}
\details{
Although the new date-time has the same clock time (e.g. the same values in
the year, month, days, etc. elements) it is a different moment of time than
the input date-time.

As R date-time vectors cannot hold elements with non-uniform time zones,
\code{force_tzs} returns a vector with time zone \code{tzone_out}, UTC by default.
}
\examples{
x <- ymd_hms("2009-08-07 00:00:01", tz = "America/New_York")
force_tz(x, "UTC")
force_tz(x, "Europe/Amsterdam")

## DST skip:
y <- ymd_hms("2010-03-14 02:05:05 UTC")
force_tz(y, "America/New_York", roll_dst = "NA")
force_tz(y, "America/New_York", roll_dst = "pre")
force_tz(y, "America/New_York", roll_dst = "boundary")
force_tz(y, "America/New_York", roll_dst = "post")

## DST repeat
y <- ymd_hms("2014-11-02 01:35:00", tz = "UTC")
force_tz(y, "America/New_York", roll_dst = "NA")
force_tz(y, "America/New_York", roll_dst = "pre")
force_tz(y, "America/New_York", roll_dst = "boundary")
force_tz(y, "America/New_York", roll_dst = "post")

## DST skipped and repeated
y <- ymd_hms("2010-03-14 02:05:05 UTC", "2014-11-02 01:35:00", tz = "UTC")
force_tz(y, "America/New_York", roll_dst = c("NA", "pre"))
force_tz(y, "America/New_York", roll_dst = c("boundary", "post"))

## Heterogeneous time-zones:

x <- ymd_hms(c("2009-08-07 00:00:01", "2009-08-07 01:02:03"))
force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"))
force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"), tzone_out = "America/New_York")

x <- ymd_hms("2009-08-07 00:00:01")
force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"))
}
\seealso{
\code{\link[=with_tz]{with_tz()}}, \code{\link[=local_time]{local_time()}}
}
\keyword{chron}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time-zones.r
\name{force_tz}
\alias{force_tz}
\alias{force_tzs}
\title{Replace time zone to create new date-time}
\usage{
force_tz(time, tzone = "", roll = FALSE)

force_tzs(time, tzones, tzone_out = "UTC", roll = FALSE)
}
\arguments{
\item{time}{a POSIXct, POSIXlt, Date, chron date-time object, or a data.frame
object. When a data.frame all POSIXt elements of a data.frame are processed
with \code{force_tz()} and new data.frame is returned.}

\item{tzone}{a character string containing the time zone to convert to. R
must recognize the name contained in the string as a time zone on your
system.}

\item{roll}{logical. If TRUE, and \code{time} falls into the DST-break, assume
the next valid civil time, otherwise return NA. See examples.}

\item{tzones}{character vector of timezones to be "enforced" on \code{time} time
stamps. If \code{time} and \code{tzones} lengths differ, the smaller one is recycled
in accordance with usual R conventions.}

\item{tzone_out}{timezone of the returned date-time vector (for \code{force_tzs}).}
}
\value{
a POSIXct object in the updated time zone
}
\description{
\code{force_tz} returns the date-time that has the same clock time as input time,
but in the new time zone. \code{force_tzs} is the parallel version of \code{force_tz},
meaning that every element from \code{time} argument is matched with the
corresponding time zone in \code{tzones} argument.
}
\details{
Although the new date-time has the same clock time (e.g. the same values in
the year, month, days, etc. elements) it is a different moment of time than
the input date-time.

As R date-time vectors cannot hold elements with non-uniform time zones,
\code{force_tzs} returns a vector with time zone \code{tzone_out}, UTC by default.
}
\examples{
x <- ymd_hms("2009-08-07 00:00:01", tz = "America/New_York")
force_tz(x, "UTC")
force_tz(x, "Europe/Amsterdam")

## DST skip:

y <- ymd_hms("2010-03-14 02:05:05 UTC")
force_tz(y, "America/New_York", roll=FALSE)
force_tz(y, "America/New_York", roll=TRUE)

## Heterogeneous time-zones:

x <- ymd_hms(c("2009-08-07 00:00:01", "2009-08-07 01:02:03"))
force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"))
force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"), tzone_out = "America/New_York")

x <- ymd_hms("2009-08-07 00:00:01")
force_tzs(x, tzones = c("America/New_York", "Europe/Amsterdam"))
}
\seealso{
\code{\link[=with_tz]{with_tz()}}, \code{\link[=local_time]{local_time()}}
}
\keyword{chron}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timespans.r
\name{is.timespan}
\alias{is.timespan}
\title{Is x a length of time?}
\usage{
is.timespan(x)
}
\arguments{
\item{x}{an R object}
}
\value{
TRUE if x is a period, interval, duration, or difftime object, FALSE otherwise.
}
\description{
Is x a length of time?
}
\examples{
is.timespan(as.Date("2009-08-03")) # FALSE
is.timespan(duration(second = 1)) # TRUE
}
\seealso{
\code{\link[=is.instant]{is.instant()}}, \code{\link[=is.duration]{is.duration()}}, \code{\link[=is.difftime]{is.difftime()}}, \code{\link[=is.period]{is.period()}}, \code{\link[=is.interval]{is.interval()}}
}
\keyword{chron}
\keyword{logic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ops-m+.r
\name{rollbackward}
\alias{rollbackward}
\alias{rollback}
\alias{rollforward}
\title{Roll backward or forward a date the previous, current or next month}
\usage{
rollbackward(dates, roll_to_first = FALSE, preserve_hms = TRUE)

rollback(dates, roll_to_first = FALSE, preserve_hms = TRUE)

rollforward(dates, roll_to_first = FALSE, preserve_hms = TRUE)
}
\arguments{
\item{dates}{A POSIXct, POSIXlt or Date class object.}

\item{roll_to_first}{Rollback to the first day of the month instead of the
last day of the month}

\item{preserve_hms}{Retains the same hour, minute, and second information? If
FALSE, the new date will be at 00:00:00.}
}
\value{
A date-time object of class POSIXlt, POSIXct or Date, whose day has
been adjusted to the last day of the previous month, or to the first day of
the month.
}
\description{
\code{rollbackward()} changes a date to the last day of the previous month or to
the first day of the month. \code{rollforward()} rolls to the last day of the
current month or to the first day of the next month. Optionally, the new date
can retain the same hour, minute, and second information. \code{rollback()} is a
synonym for \code{rollbackward()}.
}
\examples{
date <- ymd("2010-03-03")
rollbackward(date)

dates <- date + months(0:2)
rollbackward(dates)

date <- ymd_hms("2010-03-03 12:44:22")
rollbackward(date)
rollbackward(date, roll_to_first = TRUE)
rollbackward(date, preserve_hms = FALSE)
rollbackward(date, roll_to_first = TRUE, preserve_hms = FALSE)
}
